#include "all.h"
#include "tables.h"

//----------------------------

//#define AIM_MODEL_DISTANCE 0.0f

//----------------------------

static const float view_distances[] = { 0.0f, 3.0f, 6.0f };
static const float view_angles[] = { PI*.0f, PI*.13f, PI*.26f };

//----------------------------

C_game_camera::C_game_camera(PI3D_scene scene):
   mission(NULL),

   mode_dist(1), mode_angle(1), 
   delta_dir(S_vector(0, 0, 1)),
   acquire(true),
   subj_seen_count(0),
   zoom_fov_mult(1.0f),
   last_dest_pos(0, 0, 0),
   look_at_pos(0, 0, 0),
   user(NULL),
   reset(true)
{
   interp = driver->CreateInterpolator(); interp->Release();
   anim = (PI3D_keyframe_anim)driver->CreateAnimation(I3DANIM_KEYFRAME); anim->Release();

   debug_cam = driver->CreateModel(); debug_cam->Release();
   model_cache.Open(debug_cam, "_camera", scene, 0, ErrReport, editor);
   debug_cam->SetName("Sergej");

   cam_hlp = driver->CreateDummy(); cam_hlp->Release();
}

//----------------------------

C_game_camera::~C_game_camera(){

   SetFocus(NULL, NULL, NULL);
}

//----------------------------

void C_game_camera::InitFirst(C_game_mission &m){

   mission = &m;
}

//----------------------------

void C_game_camera::Init(PI3D_camera c){

   real_cam = c;
   real_cam->SetData((dword)this);

   base_fov = real_cam->GetFOV();
   last_dest_pos = real_cam->GetPos();

   debug_cam->SetOn(!acquire);

   cam = acquire ? (PI3D_frame)real_cam : (PI3D_frame)debug_cam;

   interp->SetFrame(cam);
   cam->LinkTo(cam_hlp);
   subj_seen_count = 1;       //let it be 0 in 1st Tick
}

//----------------------------

void C_game_camera::SetToIdealPosition(){

   if(!(focus_pos && focus_dir))
      return;
                              //get ideal position
   S_vector pos, dir;
   GetIdealPos(pos, dir, true);

   last_dest_pos = pos;
   cam->SetPos(pos);
   cam->SetDir(dir);
   cam->Update();
   reset = true;
                              //this will cause checking collisions, etc
   Tick(1);

}

//----------------------------

const S_vector &C_game_camera::GetCameraPos() const{
   return cam->GetWorldPos();
}

//----------------------------

void C_game_camera::SetCameraPos(const S_vector &p, const S_vector &d){

   cam->SetPos(p);
   cam->SetDir(d);
}

//----------------------------

void C_game_camera::GetCameraPos(S_vector &pos, S_vector &dir) const{
   pos = cam->GetWorldPos();
   dir = cam->GetWorldDir();
}

//----------------------------

void C_game_camera::Close(){

   SetZoom(1.0f);
   cam = NULL;
   intp_shake = NULL;
   cam_hlp->SetPos(S_vector(0, 0, 0));
   cam_hlp->Update();
}

//----------------------------

void C_game_camera::Acquire(bool b){

   acquire = b;

   if(cam){
      cam->LinkTo(NULL);
   }

   if(debug_cam) debug_cam->SetOn(!acquire);

   cam = acquire ? (PI3D_frame)real_cam : (PI3D_frame)debug_cam;
   if(!acquire && real_cam){
      cam->SetPos(real_cam->GetWorldPos());
      cam->SetDir(real_cam->GetWorldDir());
      cam->Update();
   }

   if(interp)
      interp->SetFrame(cam);

   if(cam)
      cam->LinkTo(cam_hlp);
}

//----------------------------

void C_game_camera::SetFocus(PI3D_frame fpos, PI3D_frame fdir, LPC_actor a){

   focus_dir = fdir;
   focus_pos = fpos;

   if(!fpos || !fdir){
      reset = true;
   }
   SetZoom(1.0f);
   user = a;
   subj_seen_count = 1;
}

//----------------------------

void C_game_camera::SetZoom(float zoom_mult, bool force){

   if(zoom_fov_mult == zoom_mult && !force)
      return;
   zoom_fov_mult = zoom_mult;
   if(real_cam || force){
      if(real_cam) real_cam->SetFOV(base_fov*zoom_fov_mult);
      if(aim_model){
         PI3D_frame frm = aim_model->GetChild(0);
         if(frm){
            frm->SetScale(zoom_fov_mult*.95f);
            frm->Update();
         }
      }
   }
}

//----------------------------
/*void C_game_camera::SetDistanceMode(dword m){

   assert(m<3);
   skip_fly &= !mode_dist;
   mode_dist=m;
   switch(m){
   case 0:
      intp_shake = NULL;
      cam_hlp->SetPos(S_vector(0, 0, 0));
      cam_hlp->Update();
      break;
   default:
      if(cam)
         last_dest_pos = cam->GetWorldPos();
   }
   SetZoom(1.0f);
}*/

//----------------------------
// You can use skip_fly for move to ideal position immidietly(without animation). Used in vehicles.
void C_game_camera::SetDistanceMode(dword m, bool skip_fly){

   assert(m<3);
   dword old_mode(mode_dist);
   mode_dist=m;
   switch(m){
   case 0:
      intp_shake = NULL;
      cam_hlp->SetPos(S_vector(0, 0, 0));
      cam_hlp->Update();
      break;
   default:
      if(cam){
         if(skip_fly && !old_mode)
            SetToIdealPosition();
         else
            last_dest_pos = cam->GetWorldPos();
      }
   }
   SetZoom(1.0f);
}

//----------------------------

void C_game_camera::SetAimModel(PI3D_frame am, PI3D_frame ar){

   if(aim_model)
      aim_model->LinkTo(NULL);
   aim_model = am;
   aim_ref = ar;
   if(!aim_ref)
      SetZoom(1.0f, true);
}

//----------------------------

void C_game_camera::SetAnim(const char *name, bool loop){

   PI3D_animation_set as;

   as = mission->GetAnim(name);
   if(as){
      if(as->NumAnimations()){
         PI3D_keyframe_anim a = (PI3D_keyframe_anim)as->GetAnimation(0);
         if(!intp_shake){
            intp_shake = driver->CreateInterpolator(); intp_shake->Release();
         }
         intp_shake->SetFrame(cam_hlp);

         intp_shake->SetAnimation(0, a, loop ? I3DANIMOP_LOOP : 0);
      }
   }else
      OUTTEXT(C_fstring("C_game_camera::SetAnim - can't load anim %s", name));
}

//----------------------------
// Get ideal camera position and dir for focused frames.
// Set use_new_pos true if want use param pos as current camera position insted of cam->GetPos().
void C_game_camera::GetIdealPos(S_vector &pos, S_vector &dir, bool use_new_pos){

   assert(focus_pos && focus_dir);
                              //get ideal position
   const S_matrix &m_pos = focus_pos->GetMatrix();
   const S_matrix &m_dir1 = focus_dir->GetMatrix();
   S_vector vdelta(delta_dir.x, 0.0f, delta_dir.z);
   S_matrix m_dir2;
   m_dir2.SetDir(vdelta, 0.0f);
   S_matrix m_dir = m_dir1 * m_dir2;
                              //look at head (few cm above head pivot)
   look_at_pos = m_pos(3);
                              //update pos
   S_vector f_nose = m_dir(2);

   if(mode_dist){
      float xzlen = sqrt(f_nose.x*f_nose.x + f_nose.z*f_nose.z);
      float angle = view_angles[mode_angle];
      float f = Max(-1.0f, Min(1.0f, delta_dir.y));
      angle += asin(f);
      f_nose.y = -tan(angle)*xzlen;
      f_nose.Normalize();
      pos = f_nose * -view_distances[mode_dist] + look_at_pos;
   }else
      pos = look_at_pos;


   const S_vector &curr_pos = use_new_pos ? pos : cam->GetPos();
   pos += S_vector(0, .2f, -.5f).RotateByMatrix(m_dir);
   
                              //update dir
   dir = S_normal(look_at_pos - curr_pos);
}

//----------------------------

static bool I3DAPI CamColResp(I3D_collision_data &cd){

   int mat_id(-1);
   switch(cd.GetHitFrm()->GetType()){
   case FRAME_VOLUME: 
      mat_id = I3DCAST_VOLUME(cd.GetHitFrm())->GetCollisionMaterial(); break;
   case FRAME_VISUAL:
      mat_id = I3DCAST_VISUAL(cd.GetHitFrm())->GetCollisionMaterial(); break;
   }
   if(mat_id != -1){
      if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
         mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, mat_id) ||
         mat_table->ItemB(TAB_B_MAT_CAM_SEE_THROUGH, mat_id) ||
         mat_table->ItemB(TAB_B_MAT_WATER, mat_id))
         return false;
   }
   return true;
}

//----------------------------

static bool I3DAPI CamColResp1(I3D_collision_data &cd){

   int mat_id(-1);
   switch(cd.GetHitFrm()->GetType()){
   case FRAME_VOLUME: 
      mat_id = I3DCAST_VOLUME(cd.GetHitFrm())->GetCollisionMaterial(); break;
   case FRAME_VISUAL:
      mat_id = I3DCAST_VISUAL(cd.GetHitFrm())->GetCollisionMaterial(); break;
   }
   if(mat_id != -1){
      if(mat_table->ItemB(TAB_B_MAT_BORDER, mat_id) ||
         mat_table->ItemB(TAB_B_MAT_AI_SEE_THROUGH, mat_id))
      return false;
   }
   return driver->CollisionResponse(cd);
}

//----------------------------

void C_game_camera::UpdateFPView(int t){

   assert(focus_dir && focus_pos && cam);
   const S_matrix &m_pos  = focus_pos->GetMatrix();
   const S_matrix &m_dir1 = focus_dir->GetMatrix();
   S_vector pos = m_pos(3) + m_pos(1)*.1f;
   S_vector tmp(delta_dir);
   tmp.y = -tmp.y;
   S_vector dir = tmp.RotateByMatrix(m_dir1);
   S_vector displace = dir*.2f;
   cam->SetPos(pos - displace);
   cam->SetDir(dir, 0.0f);
   cam->Update();
   if(aim_model){
      I3DCAST_MODEL(aim_model)->Tick(t);
      if(aim_ref){
         const S_vector &cdir = dir;
         aim_model->SetDir(-cdir, 0.0f);
         const S_matrix &cmat = cam->GetMatrix();

         S_vector apos = aim_ref->GetWorldPos();
         const S_vector &cpos = cmat(3);
         float cam_d = cpos.Dot(cmat(1));
         float aim_d = apos.Dot(cmat(1));
         float bad_dist = cam_d - aim_d - .15f;
         if(bad_dist<0.0f){
            apos += cmat(1)*bad_dist;
         }

         float dist = -(apos-cpos).Dot(cdir);
#ifdef AIM_MODEL_DISTANCE
         dist = AIM_MODEL_DISTANCE + dist;
#endif
         aim_model->SetPos(apos + cdir*dist);
      }else{
         dir.Normalize();
         aim_model->SetPos(pos+dir*.203f - displace);
         aim_model->SetDir(dir, 0.0f);
      }
      aim_model->Update();
   }
}

//----------------------------

void C_game_camera::PrepareFlyAnim(const S_vector &pos, const S_vector &dir){

   const S_vector &curr_pos = cam->GetPos();

   anim->Clear();

   static const dword times[] = { 100, 600, 1000};
   dword time = times[mode_dist];
                           //setup position
   I3D_anim_pos_tcb keys[2];
   keys[0].Clear();
   keys[0].time = 0;
   keys[1].Clear();
   keys[1].time = time;

   //DebugLine(curr_pos, pos, 0, 0xff0000ff);

   keys[0].v = curr_pos;
   keys[1].v = pos;
   keys[1].easy_to = 1.0f;

   anim->SetPositionKeys(keys, 2);
                           //setup directions
   {
                           //make direction faster than position
                           //dependent on distance
      static const byte time_const[] = {1, 3, 6};

      I3D_anim_look_at_tcb keys[2];
      keys[0].Clear();
      keys[0].time = 0;
      keys[0].up_vector = S_vector(0, 1, 0);
      keys[1].Clear();
      keys[1].easy_to = 1.0f;
      keys[1].up_vector = keys[0].up_vector;

      keys[0].v = cam->GetDir();
      keys[1].v = dir;
      keys[1].time = time/time_const[mode_dist];

      anim->SetLookAtKeys(keys, 2);

      anim->SetEndTime(time);
   }
   interp->SetAnimation(0, anim);
}

//----------------------------

#define SUBJ_SEEN_COUNT 2000  //how often we check to see our subject

void C_game_camera::Tick(int t){

   if(!focus_dir || !focus_pos || !cam)
      return;
   if(reset){
                              //setup ideal position
      S_vector pos, dir;
      GetIdealPos(pos, dir, true);
      cam->SetPos(pos);
      cam->SetDir(dir);
      cam->Update();

      interp->SetAnimation(0, NULL);

      reset = false;
   }

   if(!mode_dist){
      UpdateFPView(t);
      return;
   }

   S_vector pos, dir;
   GetIdealPos(pos, dir, false);

   const S_vector &curr_pos = cam->GetPos();

   if((curr_pos-pos).Magnitude() > MRG_ZERO ||
      cam->GetDir().AngleTo(dir) > MRG_ZERO){

      {
                                 //check if ideal position is visible correct
         I3D_collision_data cd(look_at_pos, pos - look_at_pos, 0, focus_dir, CamColResp);
         if(mission->GetScene()->TestCollision(cd)){
            //DebugLine(cd.from, cd.from + cd.dir, 0, 0xff00ffff);
                              //recompute ideal position
            I3D_collision_data cd;
            cd.from = look_at_pos;
            cd.dir = pos - look_at_pos;
            cd.radius = CAM_COL_SPHERE;
            cd.flags = I3DCOL_SPHERE;
            cd.cresp_face = CamColResp1;
            if(mission->GetScene()->TestCollision(cd))
               pos = cd.GetDestinationPos();
                              //check if new pos is ok
            {
               I3D_collision_data cd(curr_pos, pos - curr_pos, 0, focus_dir, CamColResp);
               if(mission->GetScene()->TestCollision(cd))
                  pos = curr_pos;
            }
            //DebugLine(curr_pos, pos, 0, 0xff0000ff);
         }
      }
                              //prepare and set animation
      PrepareFlyAnim(pos, dir);
   }
   if(!subj_seen_count)
      subj_seen_count = SUBJ_SEEN_COUNT;
   else
   if((subj_seen_count -= t) < 0)
      subj_seen_count = 0;

   assert(interp);
   interp->Tick(t);
   if(intp_shake){
      intp_shake->Tick(t);
                              //check if shaking finished
      if(intp_shake->GetCurrTime(0) == intp_shake->GetEndTime(0)){
         intp_shake = NULL;
         cam_hlp->SetPos(S_vector(0, 0, 0));
         cam_hlp->Update();
      }
   }
   CheckCol(cam);
   if(!acquire){
      mission->GetScene()->SetFrameSectorPos(cam_hlp, cam->GetWorldPos());
   }
}

//----------------------------

void C_game_camera::CheckCol(PI3D_frame cam){

   const S_vector &pos = cam->GetPos();
   S_vector new_pos;
   bool reset(false);
   bool force_reset(false);

   I3D_collision_data cd;
   cd.from = last_dest_pos;
   cd.dir = pos-last_dest_pos;
   cd.radius = CAM_COL_SPHERE;
   cd.flags = I3DCOL_SPHERE;
   cd.cresp_face = CamColResp1;
   //DebugLine(cd.from, cd.from + cd.dir, 0, 0xffff00ff);
   reset = mission->GetScene()->TestCollision(cd);
   new_pos = reset ? cd.GetDestinationPos() : pos;


   float dist2move = cd.dir.Magnitude();
   float dist2subject = (look_at_pos - pos).Magnitude();
#if 1
                              //check if covered some distance
   if(reset){
      float covered_dist = (cd.from - new_pos).Magnitude();
      //DEBUG(C_fstring("dist2subject %.2f, covered_dist %.2f, dist2move %.2f", dist2subject, covered_dist, dist2move));
      #define CRITICAL_STUCK_DIST 5
      if((dist2subject > CRITICAL_STUCK_DIST) && (dist2move > .05f) && (covered_dist < .001f)){
         //OUTTEXT("force_reset");
         force_reset = true;
      }
   }
#endif
                              //another hack - sometimes camera stuck when too far from subject,
                              //but ureadable code in tick change destination so can't move
#if 1
   {
      #define TOO_FAR_FOR_STUC 15
      if(IsAbsMrgZero(dist2move) && dist2subject > TOO_FAR_FOR_STUC){
         force_reset = true;
      }
   }
#endif

                              //check if camera's in the same sector as focus
                              //if not, reset subj_seen_count
   /*if(acquire){
      PI3D_frame sct = focus_dir;
      while(bool(sct=sct->GetParent()) && sct->GetType()!=FRAME_SECTOR);
      if(sct!=I3DCAST_CAMERA(cam)->GetSector()) subj_seen_count=0;
   }*/

   //if(!subj_seen_count)
   if(!subj_seen_count || force_reset)
   {
                              //check if we see target
      S_vector from_pos(reset ? new_pos : pos);
      S_vector back_dir(from_pos-look_at_pos);

      I3D_collision_data cd(look_at_pos, back_dir, 0, focus_dir, CamColResp);
      //DebugLine(cd.from, cd.from + cd.dir, 0, 0xff00ffff);
      bool subject_not_seen = mission->GetScene()->TestCollision(cd);
      if(subject_not_seen || force_reset){
         //DebugLine(cd.from, cd.from + cd.dir, 0, 0xff00ffff);
         const S_vector &rpos = cd.GetHitPosition();
         reset = true;
                              //now get position of sphere
         {
            new_pos = rpos;
            I3D_collision_data cd;
            cd.from = look_at_pos;
            cd.dir = rpos - look_at_pos;
            cd.radius = CAM_COL_SPHERE;
            cd.flags = I3DCOL_SPHERE;
            cd.cresp_face = CamColResp1;
            if(mission->GetScene()->TestCollision(cd))
               new_pos = cd.GetDestinationPos();
         }
      }
   }/**/

   //DebugLine(last_dest_pos, reset ? new_pos : pos, 0, 0xff00ff00);
   if(reset){
      cam->SetPos(new_pos);
      last_dest_pos = new_pos;
      cam->Update();
   }else{
      last_dest_pos = pos;
   }
}
//----------------------------
