#ifndef __I3D_FONT_H
#define __I3D_FONT_H

#include <vector>
#include <c_unknwn.hpp>
#include <i3d\i3d2.h>
#include "WideText.h"

#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------

enum E_TEXT_JUSTIFY{
                              //mode:
   TEXT_JUSTIFY_LEFT = 0,
   TEXT_JUSTIFY_CENTER = 1,
   TEXT_JUSTIFY_RIGHT = 2,
   TEXT_JUSTIFY_MASK = 3,
                              //modifiers:
   TEXT_JUSTIFY_CLIP_TO_SCREEN= 0x10000,  //clip to screen borders
   TEXT_JUSTIFY_CONVERT_EOL   = 0x20000,  //convert '\\' char to '\n' char
};

//----------------------------

class C_sprite_image: public C_unknown{
public:
   struct S_spr_rect{
                              //original sprite's rectangle
      int ix, iy, isx, isy, icx, icy;
                              //texture coordinate - in range 0.0f ... 1.0f
      float l, t, sx, sy, cx, cy;
      S_spr_rect(){}
      S_spr_rect(float l1, float t1, float sx1, float sy1, float cx1, float cy1):
         l(l1), t(t1),
         sx(sx1), sy(sy1),
         cx(cx1), cy(cy1)
      { }
   };
private:
                              //texture containing sprites
   C_smart_ptr<I3D_texture> tp;

                              //scale which was applied to texture during loading,
                              // relative to original bitmap
                              //this may be caused by driver texture size limitations
   //float tp_scale[2];
   float image_aspect_ratio;
                              //sprite rectangles
   vector<S_spr_rect> spr_rects;
public:
   C_sprite_image(){}
   C_sprite_image(const C_sprite_image&);
   C_sprite_image &operator =(const C_sprite_image&);
   ~C_sprite_image(){
      Close();
   }

//----------------------------
// Initialize class - create texture and init sprite rectangles.
   bool Init(PI3D_driver, const char *prjname, const char *diff_name = NULL,
      const char *opt_name = NULL, dword txt_create_flags = 0,
      bool use_new_prj_format = false,
      float border_offset = .5f);

   inline float GetAspectRatio() const{ return image_aspect_ratio; }

   void Close();

//----------------------------
// Setup driver for drawing polygons using this sprite image
   inline void Setup(PI3D_driver driver){
      driver->SetTexture(tp);
   }
   
//----------------------------
// Get sprite rectangles.
   inline const vector<S_spr_rect> &GetRects() const{ return spr_rects; }

//----------------------------
// Get image texture.
   inline PI3D_texture GetTexture() const{ return tp; }
};

typedef C_sprite_image *LPC_sprite_image;

//----------------------------
//----------------------------
                              //polygone sprite
class C_poly_sprite: public C_unknown{
   bool on;
   int sort_z;
   S_vector2 pos;             //position in normalized coords
   S_vector2 size;            //size in normalized coords
   S_vector2 hotspot;         //hotspot offset in normalized coords
   S_vector2 screen_pos;      //position in screen coords
   S_vector2 screen_rect[4];  //rectangle in screen coords
   void PrepareDraw();
                              
   dword sprite_index;
   dword color;

                              //dynamic visual for rendering the text, may be NULL if sprite is part of text
   C_smart_ptr<I3D_visual_dynamic> dyn_vis;
   bool vis_contents_dirty;
public:
   C_smart_ptr<C_sprite_image> img;

   struct S_vertex{           //param to DrawPrimitive
      S_vector v;
      float rhw;
      dword diffuse;
      //dword specular;         //padding to 32 bytes
      I3D_text_coor tx;
   } tlv[4];

   C_poly_sprite(float x, float y, float size, class C_sprite_image *img, dword indx, int z = 0, float aspect = 1.0f);
   C_poly_sprite(const C_poly_sprite&);
   C_poly_sprite &operator =(const C_poly_sprite&);

//----------------------------
// Enable/disable sprite.
   inline void SetOn(bool b){ on = b; }

//----------------------------
// Setup new screen position (pixel coordinates).
   void SetScreenPos(float x1, float y1);
   void SetScreenSize(float sx, float sy);

//----------------------------
   void SetTileX(float s, float e);

//----------------------------
   inline dword GetSpriteIndex() const{ return sprite_index; }

//----------------------------
// get position (screen coordinates).
   inline const S_vector2 &GetScreenPos() const{ return screen_pos; }

//----------------------------
// Set position in relative coordinates.
   void SetPos(float x1, float y1);

//----------------------------
// Get position in normalized coords.
   inline const S_vector2 &GetPos() const{ return pos; }

//----------------------------
// Get size of rectangle, in normalized coords.
   inline const S_vector2 &GetSize() const{ return size; }

//----------------------------
// Get hotspot offset in normalized coords.
   inline const S_vector2 &GetHotSpot() const{ return hotspot; }

   inline int GetZ() const{ return sort_z; }

//----------------------------
// Shear - italic-like text.
   void Shear(float angle = 0.0f, float base_y = 0.0f);

//----------------------------
// Set/Get color.
   void SetColor(dword c);
   //dword GetColor() const{ return (alpha<<24) | ((int)(v_color[0]*255.0f)<<16) | ((int)(v_color[1]*255.0f)<<8) | ((int)(v_color[0]*255.0f)); }
   inline dword GetColor() const{ return color; }

//----------------------------
// Render on screen.
   void Render();

//----------------------------
// Get text extent, in normalized screen coordinates.
   void GetExtent(float &l, float &t, float &r, float &b) const;

//----------------------------
   void Rotate(float angle);

   inline bool IsOn() const{ return on; }
};

typedef C_poly_sprite *LPC_poly_sprite;

//----------------------------
//----------------------------
                              //Group of sprites - usualy ones using the same sprite image.
                              //All these sprites are rendered together.
                              //Usable for text consisting from more letters - sprites. 
class C_sprite_group: public C_unknown{
   vector<C_smart_ptr<C_poly_sprite> > sprs;
   S_vector2 pos;             //position in normalized coords
   bool resort;

   friend class C_poly_text;  //direct access to our sprites
   friend class C_text;  //direct access to our sprites
public:
   C_sprite_group();
   C_sprite_group(const C_sprite_group&);
   C_sprite_group &operator =(const C_sprite_group&);

   void AddSprite(LPC_poly_sprite);
   void RemoveSprite(LPC_poly_sprite);

   void Render();

//----------------------------
// Set color of all sprites.
   void SetColor(dword);

//----------------------------
// Get group's position.
   inline const S_vector2 & GetPos() const{ return pos; }

//----------------------------
// Enable/disable all sprites.
   void SetOn(bool);

//----------------------------
// Get sprites.
   inline const vector<C_smart_ptr<C_poly_sprite> > &GetSPrites() const{ return sprs; }
};

typedef C_sprite_group *LPC_sprite_group;

//----------------------------
//----------------------------
                              //2D text class - made out of textured polygons (single-char),
                              // or colorkeyed images (wide-char)
class C_text: public C_unknown{
   C_poly_text *texts;
                              //single-char support
   C_smart_ptr<C_sprite_group> poly_text;
                              //wide-char support
   C_smart_ptr<C_wide_text> wide_text;
                              //dynamic visual for rendering the text
   C_smart_ptr<I3D_visual_dynamic> dyn_vis;

   int z;                     //sort value
   bool on;
   dword color;

   void CreateVisual();

   friend class C_poly_text;
public:
   C_text(): texts(NULL){}
   C_text(LPC_sprite_group pg, C_poly_text *t):
      texts(t),
      poly_text(pg),
      on(true)
   {}
   C_text(C_wide_text *wt, C_poly_text *t):
      texts(t),
      wide_text(wt),
      on(true)
   {}

   C_text(const C_text&);
   C_text &operator =(const C_text&);

   ~C_text();

//----------------------------
// Get text extent, in normalized screen coordinates.
   void GetExtent(float &l, float &t, float &r, float &b) const;

//----------------------------
// Render as 3D polygons.
   void Render() const;

//----------------------------
// Enable/disable text drawing.
   inline void SetOn(bool b){ on = b; }

//----------------------------
   inline C_sprite_group *GetAnsiText() const{ return poly_text; }

//----------------------------
// Get/Set color of text.
   void SetColor(dword c);
   inline dword GetColor() const{ return color; }

//----------------------------
// Get/Set position of text;
   void SetPos(float x, float y);
   void GetPos(float &x, float &y);
};

typedef C_text *LPC_text;

//----------------------------
//----------------------------
                              //Class to keep all visible 2D texts so that they may be
                              // rendered together.
class C_poly_text: public C_unknown{
   bool resort;

                              //loaded font data:
   struct S_font_size{
      byte l, r;
   };
   vector<S_font_size> font_sizes;
   vector<int> font_size_offsets;
   //C_smart_ptr<C_sprite_image> font_img;

//----------------------------
// Fit two letters together - compute smallest distance for which letters don't overlap.
   word LetterFit(byte l, byte r);

//----------------------------
// For given string, build spacing table - distances of letters.
   void BuildSpaceTable(const char *text, short *tab, int tablen);

//----------------------------
// All texts to be rendered.
   vector<LPC_text> texts;
   
   static bool TextLess(const LPC_text t1, const LPC_text t2){
      return (t1->z < t2->z);
   }
public:
   C_poly_text():
      resort(true)
   {
      //font_img = new C_sprite_image;
      //font_img->Release();
   }
   C_poly_text(const C_poly_text&);
   C_poly_text &operator =(const C_poly_text&);
   ~C_poly_text();

//----------------------------
// Call with param diff_name relative to dir "Maps\\".
// 'font_image' is name of image to be used for building font spacing table.
//----------------------------
   bool Init();
   bool Init(const C_poly_text &pt){
      font_sizes = pt.font_sizes;
      font_size_offsets = pt.font_size_offsets;
      texts.clear();
   }
   inline bool IsInit() const{ return (font_sizes.size()!=0); }

   void Close();

//----------------------------
// Create text object, user cannot release this pointer, rather he must call
// DeleteText when the text object is no longer needed.
   LPC_text CreateText(const void *tp, dword color, float x, float y, float size,
      dword justify = 0, bool wide = false, float fmt_rect[4] = NULL,
      float *xret = NULL, float *yret = NULL, int *num_lines = NULL,
      int z = 0, float italic = 0.0f, float scale_x = 1.0f, float scale_y = 1.0f);
   bool RemoveText(C_text *tp);

//----------------------------
// Render texts created by this class.
   void Render();
};

//----------------------------

#endif