/*******************************************************
* File: c_cache.h
* Purpose: cached read/write using file or memory
* Created by: Michal Bacik
*******************************************************/
#ifndef __C_CACHE_H
#define __C_CACHE_H

#include <dta_read.h>
#include <memory.h>
#include <stdlib.h>
#include <assert.h>


enum E_CACHE_MODE{
   CACHE_NO,
   CACHE_READ,
   CACHE_WRITE,
   CACHE_READ_MEM,
   CACHE_WRITE_MEM,
};

#pragma pack(push,4)

class C_cache{
   enum{DEFAULT_CACHE_SIZE = 0x8000 };

   size_t buffer_size;
   byte *base, *top, *curr;   //cache memory
   int h;                     //read/write handle
   E_CACHE_MODE mode;         //open mode
   size_t curr_pos;           //real file pos
   bool b_fail;               //fail bit

                              //memory read/write:
   size_t mem_size, *sizep;
   byte **memp;
public:
                              //virtual functions provide that this class may be used freely
                              //between DLLs, as only on module manages allocation
   virtual byte *AllocMem(size_t sz){
      return (byte*)malloc(sz);
   }
   virtual byte *ReallocMem(void *bp, size_t sz){
      return (byte*)realloc(bp, sz);
   }
   virtual void FreeMem(byte *bp){
      free(bp);
   }
   C_cache():
      buffer_size(DEFAULT_CACHE_SIZE),
      mode(CACHE_NO)
   {}
   C_cache(size_t sz):
      buffer_size(sz),
      mode(CACHE_NO)
   {}
   ~C_cache(){
      close();
   }

   bool open(const char *fname, E_CACHE_MODE m){
      close();
      switch(m){
      case CACHE_READ:
         h = dtaOpen(fname);
         if(h==-1){
            b_fail = true;
            return false;
         }
         curr = top = base = AllocMem(buffer_size);
         if(!base){
            dtaClose(h);
/*#ifdef C_CACHE_USE_EXCEPTIONS
            throw C_except_mem("C_cache::open (CACHE_READ)");
#endif*/
            return false;
         }
         break;
      case CACHE_WRITE:
         h = dtaOpenWrite(fname);
         if(h==-1){
            b_fail = true;
            return false;
         }
         curr = top = base = AllocMem(buffer_size);
         if(!base){
            dtaClose(h);
/*#ifdef C_CACHE_USE_EXCEPTIONS
            throw C_except_mem("C_cache::open (CACHE_WRITE)");
#endif*/
            return false;
         }
         break;
      }
      mode = m;
      curr_pos = 0;
      b_fail = false;
      return true;
   }

//----------------------------
// Open cache for read from / write to memory. Memory is managed by virtual functions
// of C_cache.
// In case of writing to memory, user is responsible for freeing memory after closing cache
// by calling C_cache::FreeMem method.
// When opening cache for write to memory, parameter default_size specifies initial size of
// buffer being allocated.
//----------------------------
   bool open(byte **mem, size_t *size1, E_CACHE_MODE m, size_t default_size = 0){
      close();
      switch(m){
      case CACHE_READ_MEM:
         base = *mem;
         mem_size = *size1;
         break;
      case CACHE_WRITE_MEM:
         sizep = size1;
         memp = mem;
         base = NULL;
         mem_size = 0;
         if(default_size){
            if(base = AllocMem(default_size))
               mem_size = default_size;
         }
         break;
      default: 
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::open: invalid open mode");
#endif*/
         assert(0);
         return false;
      }
      mode = m;
      curr_pos = 0;
      b_fail = false;
      return true;
   }

   void close(){
      switch(mode){
      case CACHE_WRITE:
         flush();
         dtaClose(h);
         FreeMem(base);
         base = NULL;
         break;
      case CACHE_READ:
         dtaClose(h);
         FreeMem(base);
         base = NULL;
         break;
      case CACHE_WRITE_MEM:
         *memp = base;
         *sizep = curr_pos;
                              //note: user is responsible for freeing memory
         break;
      case CACHE_READ_MEM:
         break;
      }
      mode = CACHE_NO;
   }

   size_t read(char *mem, size_t len){
      size_t rl;
      switch(mode){
      case CACHE_READ_MEM:
         rl = Min(len, mem_size-curr_pos);
         memcpy(mem, base+curr_pos, rl);
         curr_pos += rl;
         return rl;
      case CACHE_READ:
         rl = 0;
         while(curr+len>top){
            memcpy(mem, curr, top-curr);
            int rl1 = top-curr;
            len -= rl1;
            mem += rl1;
            size_t sz = dtaRead(h, curr=base, buffer_size);
            top = base + sz;
            curr_pos += sz;
            rl += rl1;
            if(!sz) return rl;
         }
         memcpy(mem, curr, len);
         curr += len;
         return rl + len;
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::read: invalid mode");
#endif*/
      }
      return -1;
   }

   inline size_t read(byte *mem, size_t len){ return read((char*)mem, len); }

   bool write(const char *buf, size_t len){
      switch(mode){
      case CACHE_WRITE_MEM:
                              //check if we need re-allocate
         if(len + curr_pos > mem_size){
            base = ReallocMem(base, mem_size = (len + curr_pos)*2);
         }
         memcpy(base+curr_pos, buf, len);
         curr_pos += len;
         break;
      case CACHE_WRITE:
                              //put into cache
         while((curr-base)+len > buffer_size){
            size_t sz = buffer_size - (curr-base);
            memcpy(curr, buf, sz);
            curr += sz;
            top = curr;
            buf += sz;
            len -= sz;
            flush();
         }
         memcpy(curr, buf, len);
         curr += len;
         if(top<curr) top = curr;
         break;
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::write: invalid mode");
#endif*/
      }
      return true;
   }

   C_cache &getline(char *buf, int len, char='\n'){
      switch(mode){
      case CACHE_READ:
      case CACHE_READ_MEM:
         while(--len){
            char c;
            read(&c, 1);
            if(c=='\r'){
               read(&c, 1);
               break;
            }
            if(c=='\n') break;
            *buf++ = c;
         }
         *buf = 0;
         break;
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::getline: invalid mode");
#endif*/
      }
      return *this;
   }

   C_cache &flush(){
      switch(mode){
      case CACHE_WRITE: 
         if(top>base) dtaWrite(h, base, top-base);
         curr_pos += top-base;
         curr = top = base;
         break;
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::flush: invalid mode");
#endif*/
      }
      return (*this);
   }

   bool eof() const{
      switch(mode){
      //case CACHE_WRITE: return os.eof();
      case CACHE_WRITE: return true;
      case CACHE_READ: return (curr==top && filesize()==tellg());
      case CACHE_READ_MEM: return (curr_pos==mem_size);
      }
      return false;
   }

   size_t filesize() const{
      switch(mode){
      case CACHE_READ_MEM: return mem_size;
      case CACHE_WRITE_MEM: return curr_pos;
      case CACHE_READ:
         {
            int cp = dtaSeek(h, 0, DTA_SEEK_CUR);
            int len = dtaSeek(h, 0, DTA_SEEK_END);
            dtaSeek(h, cp, DTA_SEEK_SET);
            return len;
         }
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::filesize: invalid mode");
#endif*/
      }
      return -1;
   }

   size_t tellg() const{
      switch(mode){
      //case CACHE_WRITE: return ((fstream&)os).tellg();
      case CACHE_WRITE: return dtaSeek(h, 0, DTA_SEEK_CUR) + curr - base;
      case CACHE_READ: return curr_pos+curr-top;
      case CACHE_READ_MEM: return curr_pos;
      case CACHE_WRITE_MEM: return curr_pos;
      }
      return -1;
   }

   inline bool fail() const{
      return b_fail;
   }

   inline bool bad() const{ return fail(); }

   inline bool good() const{ return !fail(); }

   bool seekg(size_t pos){
      switch(mode){
      case CACHE_WRITE:
         {
            int delta = pos - curr_pos - (curr-base);
            if(delta<0){
               if(curr-base >= -delta) curr += delta;
               else{
                  flush();
                  dtaSeek(h, curr_pos = pos, DTA_SEEK_SET);
               }
            }else{
               if(((size_t)delta) <= (buffer_size-(curr-base))){
                              //fill by zeros
                  int delta1 = pos - curr_pos - (top-base);
                  if(delta1>0) memset(top, 0, delta1);
                              //read original contents (if any)
                  int file_len = dtaSeek(h, 0, DTA_SEEK_END);
                  if(file_len > curr_pos+top-base){
                              //read-in the rest
                     dtaSeek(h, curr_pos+top-base, DTA_SEEK_SET);
                     dtaRead(h, curr, pos - curr_pos-(top-base));
                  }
                  dtaSeek(h, curr_pos, DTA_SEEK_SET);

                  curr += delta;
                  if(top<curr) top = curr;
               }else{
                  flush();
                  dtaSeek(h, curr_pos = pos, DTA_SEEK_SET);
               }
            }
            //flush();
            //dtaSeek(h, pos, DTA_SEEK_SET);
         }
         break;
      case CACHE_READ:
         if(pos<tellg()){
            if(pos>=curr_pos-(top-base)){
               curr=top-(curr_pos-pos);
            }else{
               curr=top=base;
               dtaSeek(h, pos, DTA_SEEK_SET);
               curr_pos=pos;
            }
         }else
         if(pos>curr_pos){
            curr=top=base;
            dtaSeek(h, pos, DTA_SEEK_SET);
            curr_pos=pos;
         }else curr=top+pos-curr_pos;
         return true;
      case CACHE_READ_MEM:
         curr_pos = Max((size_t)0, Min(mem_size, pos));
         return true;
      case CACHE_WRITE_MEM:
         curr_pos = Max((size_t)0, Min(mem_size, pos));
         return true;
      default:
         assert(0);
/*#ifdef C_CACHE_USE_EXCEPTIONS
         throw C_except_general("C_cache::seekg: invalid mode");
#endif*/
      }
      return false;
   }

   inline bool seekp(size_t pos){ return seekg(pos); }

   inline int GetHandle() const{ return h; }
};

#pragma pack(pop)

//----------------------------

#endif   //__C_CACHE_H
