#ifndef __C_CHUNK_H_
#define __C_CHUNK_H_

#include <rules.h>
#include <vector>
#include <c_string.hpp>
#include <i3d\i3d_math.h>
#include <c_cache.h>

typedef word CK_TYPE;

#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------

enum E_BASIC_CHUNK_TYPE{
   CT_VERSION  =0x0001,
      //word maj
      //word min
      //copyright info text
   CT_COPYRIGHT=0x0002,
      //text
   CT_NAME     =0x0010,
      //null-terminated string
   CT_POSITION =0x0020,
      //S_vector
   CT_DIRECTION=0x0021,
      //S_vector
   CT_QUATERNION  =0x0022,
      //S_quat
   CT_SCALE    =0x0023,
      //S_vector
   CT_NUSCALE  =CT_SCALE,
      //S_vector
   CT_FLOAT    =0x0024,
   CT_INT      =0x0025,
   CT_COLOR    =0x0026,
      //S_vector
   CT_USCALE   =0x002a,
      //float
   CT_PIVOT    =0x002b,
      //S_vector
   CT_WORLD_POSITION =0x002c,
      //S_vector

   CT_ROTATION =0x0030,
      //S_vector axis / float angle
};

//----------------------------

class C_chunk{
                              //use allocator with virtual allocation functions,
                              // so that we may use C_chunk accross modules.
   template<class T>
   class c_c_allocator: public allocator<T>{
   public:
      virtual pointer allocate(size_type n, const void *hint){
         return (pointer)new byte[n*sizeof(T)];
      }
      virtual void deallocate(pointer p, size_type n){
         delete[] (byte*)p;
      }
   };

   vector<size_t, c_c_allocator<size_t> > level;
   C_cache cache;
   bool out, init;
#pragma pack(push, 1)
   struct S_ckhdr{
      word type;
      size_t size;
   };
#pragma pack(pop)

   bool OpenFile(const char *fname, bool write_access){
      if(!IsInit()){
         cache.open(fname, !write_access ? CACHE_READ : CACHE_WRITE);
         if(!cache.fail()){
            out = write_access;
            return (init = true);
         }
      }
      return false;
   }
   bool OpenMem(byte **mem, dword *size, bool write_access){
      if(!IsInit()){
         cache.open(mem, (size_t*)size, !write_access ? CACHE_READ_MEM : CACHE_WRITE_MEM);
         if(!cache.fail()){
            out = write_access;
            return (init = true);
         }
      }
      return false;
   }

public:
   C_chunk(): init(false){}

//----------------------------
// Init class with specified cache size
   C_chunk(dword sz): cache(sz), init(false){}

   ~C_chunk(){ Close(); }
                              //read
//----------------------------
// Open disk file for reading.
   inline bool ROpen(const char *fname){
      return OpenFile(fname, false);
   }

//----------------------------
// Open memory location for reading.
   inline bool ROpen(const byte *mem, dword size){
      return OpenMem((byte**)&mem, &size, false);
   }

//----------------------------
// Enter chunk in read mode and retrieve chunk type.
// The returned value is false if chunk can't be entered (end of file, or end of chunk).
   bool RAscend(CK_TYPE &t){

      assert(!out);
      if(IsInit()){
         if(level.size() && Size() < sizeof(S_ckhdr))
            return false;
         S_ckhdr chkdr;
         size_t rl = cache.read((char*)&chkdr, sizeof(S_ckhdr));
         if(rl<sizeof(S_ckhdr)) return false;
         level.push_back(chkdr.size+cache.tellg()-sizeof(S_ckhdr));
         t = chkdr.type;
         return true;
      }
      return false;
   }

//----------------------------
// Read data.
   inline bool Read(void *data, size_t sz){
      assert(!out);
      if(IsInit() && !out){
         if(level.size()){    //don't read data after the block size
            size_t bl_left = Max(0, (int)(level.back() - cache.tellg()));
            if(sz > bl_left) return false;
         }
         if(cache.read((char*)data, sz)<sz) return false;
         //return cache.good();
         return true;
      }
      return false;
   }

//----------------------------
// Get size left from current location to the end of current chunk.
   inline size_t Size() const{
      if(!IsInit()) return 0;
      assert(level.size());
      return level.back()-cache.tellg();
   }

//----------------------------
// Open disk file for writting.
   inline bool WOpen(const char *fname){
      return OpenFile(fname, true);
   }

//----------------------------
// Open memory location for writting.
   inline bool WOpen(byte **mem, dword *size){
      return OpenMem(mem, size, true);
   }

//----------------------------
// Create chunk of specified type in write mode.
   bool WAscend(CK_TYPE ckt){
      assert(out);
      if(IsInit()){
         S_ckhdr chkdr;
         chkdr.type = ckt;
         level.push_back(cache.tellg());
                                 //write partial header, write rest later
         cache.write((char*)&chkdr, sizeof(S_ckhdr));
         return true;
      }
      return false;
   }

//----------------------------
// Write data.
   bool Write(const void *data, size_t sz){
      assert(out);
      if(IsInit() && out){
         cache.write((char*)data, sz);
         return cache.good();
      }
      return false;
   }

//----------------------------
// Get pointer of C_cache we're working on.
   inline C_cache *GetHandle(){ return &cache; }
   
//----------------------------
// Descend currently entered chunk - works for both read and write modes.
   bool Descend(){
      if(IsInit()){
         assert(level.size());
         if(!out){
                              //skip to rest of chunk
            cache.seekp(level.back());
         }else{               
                              //write chunk header
            size_t curr_pos = cache.tellg();
            size_t chk_pos = level.back();
            size_t chk_size = curr_pos - chk_pos;
            cache.seekp(chk_pos + offsetof(S_ckhdr, size));
            cache.write((char*)&chk_size, sizeof(size_t));
            cache.seekp(curr_pos);
         }
         level.pop_back();
         return true;
      }
      return false;
   }

//----------------------------
// Close class - descend all currently opened chunks (in write mode).
   bool Close(){
      if(IsInit()){
         if(out)
            while(level.size()) Descend();
         cache.close();
         init = false;
         return true;
      }
      return false;
   }

//----------------------------
// Check if class is initialized (opened).
   inline bool IsInit() const{ return init; }

//----------------------------
// Following clas of Write???Chunk functions create specified chunk and write associated data into that.
   void WStringChunk(dword ct, const char *cp){
      WAscend(ct);
      Write(cp, strlen(cp)+1);
      Descend();
   }
   void WVectorChunk(dword ct, const S_vector &v){
      WAscend(ct);
      Write(&v, sizeof(S_vector));
      Descend();
   }
   void WRotChunk(dword ct, const S_vector &axis, float angle){
      WAscend(ct);
      Write(&axis, sizeof(S_vector));
      Write(&angle, sizeof(float));
      Descend();
   }
   void WRotChunk(dword ct, const S_quat &q){
      WAscend(ct);
      Write(&q, sizeof(q));
      Descend();
   }
   void WFloatChunk(dword ct, float f){
      WAscend(ct);
      Write(&f, sizeof(float));
      Descend();
   }
   void WIntChunk(dword ct, int i){
      WAscend(ct);
      Write(&i, sizeof(int));
      Descend();
   }
   void WWordChunk(dword ct, word w){
      WAscend(ct);
      Write(&w, sizeof(w));
      Descend();
   }
   void WByteChunk(dword ct, byte b){
      WAscend(ct);
      Write(&b, sizeof(byte));
      Descend();
   }
   void WBoolChunk(dword ct, bool b){
      WAscend(ct);
      Write(&b, sizeof(byte));
      Descend();
   }
   void WArrayChunk(dword ct, const void *mem, dword len){
      WAscend(ct);
      Write(mem, len);
      Descend();
   }
   template<class T>
   void WriteChunk(dword ct, const T &t){
      WAscend(ct);
      Write(&t, sizeof(T));
      Descend();
   }

//----------------------------
// Following clas of Read???Chunk functions read specified data from already ascended chunk
// and descend out of the chunk after this.

   void ReadString(C_string &name){
      char buf[256];
      for(int i=0; i<256; ){
         Read(&buf[i], sizeof(char));
         if(!buf[i++])
            break;
      }
      buf[i-1] = 0;
      name = buf;
   }

   bool ReadString(dword ct, C_string &name){
      CK_TYPE t;
      if(RAscend(t)){
         if(t==ct){
            char buf[256];
            Read(buf, Min(sizeof(buf), Size()));
            name = buf;
            Descend();
            return true;
         }
         Descend();
      }
      name = NULL;
      return false;
   }
   C_string RStringChunk(){
      char buf[256];
      Read(buf, Min(sizeof(buf), Size()));
      Descend();
      return buf;
   }
   S_vector RVectorChunk(){
      S_vector v;
      Read(&v, sizeof(S_vector));
      Descend();
      return v;
   }
   S_quat RQuaternionChunk(){
      S_quat q;
      Read(&q, sizeof(q));
      Descend();
      return q;
   }
   float RFloatChunk(){
      float f;
      Read(&f, sizeof(f));
      Descend();
      return f;
   }
   int RIntChunk(){
      int i;
      Read(&i, sizeof(i));
      Descend();
      return i;
   }
   int RWordChunk(){
      word w;
      Read(&w, sizeof(w));
      Descend();
      return w;
   }
   byte RByteChunk(){
      byte b;
      Read(&b, sizeof(byte));
      Descend();
      return b;
   }
   bool RBoolChunk(){
      bool b = false;
      Read(&b, sizeof(byte));
      Descend();
      return b;
   }
   void RArrayChunk(dword ct, void *mem, dword len){
      Read(mem, len);
      Descend();
   }
   template<class T>
   T ReadChunk(){
      T t;
      Read(&t, sizeof(T));
      Descend();
      return t;
   }
};

//----------------------------
#endif
