#ifndef __C_STRING_HPP
#define __C_STRING_HPP
#pragma once

/*----------------------------------------------------------------\
Copyright (c) 1998 Michal Bacik  All rights reserved
Abstract:
 C_string class - reference-counting string class, minimizing memory copying
 as possible, keeping reference to other string after copying.
 Safe access - making its own copy on non-constant access.
 Keeping valid NULL-terminated "C" string even if not initialized.
 Allowing wild-card compare with other string.
 Support for single-byte and double-byte contents.
   (use #define CSTRING_WCHAR to enable wide-char support)
 Usable for shared memory data.

 C_fstring - inherited from C_string, having additional constructor for
 creating string from formatted string and variable-length arguments.

|\----------------------------------------------------------------*/
#include <string.h>
#include <rules.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#define CFSTRING_MAX_FMT_BUFFER 1024  //size of formatting buffer, in characters

//#define CSTRING_WCHAR

/*--------------------------------------------------------
   String base - concrete string data. This class' size is
   variable, depending on size of string it contains.
--------------------------------------------------------*/
#pragma pack(push,1)
class C_str_base{
   dword ref, str_len;
#ifdef CSTRING_WCHAR
   char data[2];           //wide-char string terminator
#else
   char data[1];
#endif
public:
   inline C_str_base(const void *cp, int len): ref(1), str_len(len){
      memcpy(data, cp, len);
#ifdef CSTRING_WCHAR
      *(word*)(data+len) = 0;
#else
      data[len] = 0;
#endif
   }
   inline C_str_base(int len): ref(1), str_len(len){ }
   inline C_str_base(const char *cp1, const char *cp2, int l1, int l2): ref(1), str_len(l1+l2){
      memcpy(data, cp1, l1);
      memcpy(data+l1, cp2, l2);
#ifdef CSTRING_WCHAR
      *(word*)(data+str_len) = 0;
#else
      data[str_len] = 0;
#endif
   }
   inline void AddRef(){ ++ref; }
   inline void Release(){ if(!--ref) delete this; }
   inline dword Count() const{ return ref; }

   inline char *GetData(){ return data; }
   inline const char *GetData() const{ return data; }
   inline dword Size() const{ return str_len; }
                              //memory management
   void *operator new(size_t sz, size_t str_len){
      return new byte[sz+str_len];
   }
   void operator delete(void *vp, size_t){
      delete[] (byte*)vp;
   }
};
#pragma pack(pop)

/*--------------------------------------------------------
   String - keeping reference to string data.
--------------------------------------------------------*/

class C_string{
   C_str_base *rep;

//----------------------------
// Make sure this string is the only owner of actual string data.
   void MakeUniqueCopy(){

      if(rep->Count()!=1){    
         C_str_base *new_rep = new(rep->Size()) C_str_base(rep->GetData(), rep->Size());
         rep->Release();
         rep=new_rep;
      }
   }
public:
//----------------------------
// Empty constructor.
   inline C_string(): rep(NULL) {}

//----------------------------
// Constructor from C string.
   inline C_string(const char *cp){
      if(!cp) rep = NULL;
      else{
         int len = strlen(cp);
         rep = new(len) C_str_base(cp, len);
      }
   }

//----------------------------
// Copy constructor.
   inline C_string(const C_string &s){ if(rep=s.rep) rep->AddRef(); }

//----------------------------
// Destructor.
   inline ~C_string(){ if(rep) rep->Release(); }

//----------------------------
// Getting pointer to C string. This is always valid NULL-terminated string,
// even if the string is empty. It is valid until any other operation is done on
// contents of string.
   inline operator const char *() const{ return rep ? rep->GetData() : (const char*)&rep; }

//----------------------------
// Assignment operator - fast copy of reference from given string.
   C_string &operator =(const C_string &s){
      if(s.rep) s.rep->AddRef();
      if(rep) rep->Release();
      rep = s.rep;
      return (*this);
   }

//----------------------------
// Assignment from C string.
   C_string &operator =(const char *cp){
      C_str_base *new_rep;
      if(!cp) new_rep = NULL;
      else{
         int len = strlen(cp);
         new_rep = new(len) C_str_base(cp, len);
      }
      if(rep) rep->Release();
      rep = new_rep;
      return (*this);
   }

//----------------------------
// Assign raw data to string. Data may be anything, no '\0' character at the end is searched.
// If 'cp' is NULL, the string allocates uninitialized buffer for holding the string,
// which may be filled further by array-access functions.
   C_string &Assign(const char *cp, dword size){
      C_str_base *new_rep;
      if(!cp) new_rep = new(size) C_str_base(size);
      else
      new_rep = new(size) C_str_base(cp, size);
      if(rep) rep->Release();
      rep = new_rep;
      return (*this);
   }

//----------------------------
// Concanetate string with another C string and store result in this string.
   C_string &operator +=(const char *cp){
      if(!rep) return operator =(cp);
      int len = rep->Size(), len1 = strlen(cp);
      C_str_base *new_rep = new(len+len1) C_str_base(rep->GetData(), cp, len, len1);
      rep->Release();
      rep = new_rep;
      return (*this);
   }

//----------------------------
// Concanetate string with another string and store result in this string.
   C_string &operator +=(const C_string &s){
      if(!rep) return operator =(s);
      int len = Size(), len1 = s.Size();
      C_str_base *new_rep = new(len+len1) C_str_base(rep->GetData(), s.rep->GetData(), len, len1);
      rep->Release();
      rep = new_rep;
      return (*this);
   }

//----------------------------
// Add this string with another string and return result as a string.
   C_string operator +(const C_string &s) const{
      if(!rep) return s;
      int l1 = Size(), l2 = s.Size();
      if(!l2) return *this;
      C_string ret;
      ret.rep = new(l1+l2) C_str_base(rep->GetData(), s.rep->GetData(), l1, l2);
      return ret;
   }

//----------------------------
// Add this string with another C string and return result as a string.
   C_string operator +(const char *cp) const{
      if(!rep) return cp;
      int l1 = Size(), l2 = strlen(cp);
      C_string ret;
      ret.rep = new(l1+l2) C_str_base(rep->GetData(), cp, l1, l2);
      return ret;
   }

//----------------------------
// Get const character reference inside of string on particular position.
   inline const char &operator [](int pos) const{ assert(pos<((int)Size()+1)); return rep->GetData()[pos]; }

//----------------------------
// Get non-const character reference inside of string on particular position.
// If contents of string is shared among multiple strings, it is made unique.
   char &operator [](int pos){
      assert(pos<((int)Size()+1));
      MakeUniqueCopy();
      return rep->GetData()[pos];
   }

//----------------------------
// Compare this string with another C string for equality. Slow, depends on sizes of
// both strings.
   inline bool operator ==(const char *cp) const{ return !strcmp(*this, cp); }

//----------------------------
// Compare this string with another string for equality. Optimizations based on sizes of strings.
   bool operator ==(const C_string &s) const{
      if(rep==s.rep) return true;
      if(Size()!=s.Size()) return false;
      return !memcmp((const char*)*this, (const char*)s, Size());
   }

//----------------------------
// Compare this string with another strings.
   inline bool operator <(const C_string &s) const{
      const char *cp = rep ? rep->GetData() : (const char*)&rep;
      return (strcmp(cp, s) < 0);
   }

//----------------------------
// Compare this string with another C strings.
   inline bool operator <(const char *s) const{
      const char *cp = rep ? rep->GetData() : (const char*)&rep;
      return (strcmp(cp, s) < 0);
   }

//----------------------------
// Get size of string data, without terminating NULL character.
   inline size_t Size() const{ return rep ? rep->Size() : 0; }

//----------------------------
// Compare this string with another wild-char string. Following wildcards are recognized:
//    * - ignore the contents of string to the end
//    ? - ignore character on current position
   bool Match(const C_string &s) const{
      size_t l1 = Size(), l2 = s.Size();
      if(!l1 || !l2) return (l1==l2);
      for(const char *cp1=operator const char*(), *cp2=s; ; ++cp1, ++cp2){
         switch(*cp2){
         case 0: return !(*cp1);
         case '*': return true;
         case '?':
            if(!*cp1) return false;
            break;
         default: if(*cp1 != *cp2) return false;
         }
      }
   }

//----------------------------
// Wildcard compare, but ignore case.
   bool Matchi(const C_string &s) const{
      size_t l1 = Size(), l2 = s.Size();
      if(!l1 || !l2) return (l1==l2);
      for(const char *cp1=operator const char*(), *cp2=s; ; ++cp1, ++cp2){
         switch(*cp2){
         case 0: return !(*cp1);
         case '*': return true;
         case '?':
            if(!*cp1) return false;
            break;
         default: if(tolower(*cp1) != tolower(*cp2)) return false;
         }
      }
   }

//----------------------------
// Convert all characters to lower case.
   void ToLower(){
      size_t l = Size();
      if(!l)
         return;
      MakeUniqueCopy();
      char *buf = rep->GetData();
      while(l--)
         buf[l] = tolower(buf[l]);
   }

//----------------------------
// Convert all characters to upper case.
   void ToUpper(){
      size_t l = Size();
      if(!l)
         return;
      MakeUniqueCopy();
      char *buf = rep->GetData();
      while(l--)
         buf[l] = toupper(buf[l]);
   }

//----------------------------
                              //double-byte specialization
#ifdef CSTRING_WCHAR
   C_string(const wchar_t *wp){
      if(!wp) rep = NULL;
      else{
         int len = wcslen(wp);
         rep = new(len*2) C_str_base(wp, len*2);
      }
   }
   operator const wchar_t *() const{ return rep ? (const wchar_t*)rep->GetData() : (const wchar_t*)&rep; }
   C_string &operator =(const wchar_t *wp){
      C_str_base *new_rep;
      if(!wp) new_rep = NULL;
      else{
         int len = wcslen(wp)*2;
         new_rep = new(len) C_str_base(wp, len);
      }
      if(rep) rep->Release();
      rep = new_rep;
      return (*this);
   }
   C_string &operator +=(const wchar_t *wp){
      if(!rep) return operator =(wp);
      int len = rep->Size(), len1 = wcslen(wp)*2;
      C_str_base *new_rep = new(len+len1) C_str_base(rep->GetData(), (const char*)wp, len, len1);
      rep->Release();
      rep = new_rep;
      return (*this);
   }
   C_string operator +(const wchar_t *wp) const{
      C_string ret;
      int l1 = rep->Size(), l2 = wcslen(wp)*2;
      ret.rep = new(l1+l2) C_str_base(rep->GetData(), (const char*)wp, l1, l2);
      return ret;
   }
   bool operator ==(const wchar_t *wp) const{ return !wcscmp((const wchar_t*)(const char*)(*this), wp); }
#endif//CSTRING_WCHAR
};

class C_fstring: public C_string{
public:
   C_fstring(const char *cp, ...){
      char buf[CFSTRING_MAX_FMT_BUFFER];
      va_list arglist;
      va_start(arglist, cp);
#ifdef _MSC_VER
#ifndef NDEBUG
      int num = 
#endif
         _vsnprintf(buf, sizeof(buf), cp, arglist);
      assert(num>=0);         //not enough big buffer? report it
#else
      vsprintf(buf, cp, arglist);
#endif
      va_end(arglist);
      C_string::operator=(buf);
   }
#ifdef CSTRING_WCHAR
   C_fstring(const wchar_t *wp, ...){
      wchar_t buf[CFSTRING_MAX_FMT_BUFFER];
      va_list arglist;
      va_start(arglist, wp);
      vswprintf(buf, sizeof(buf), wp, arglist);
      va_end(arglist);
      C_string::operator =(buf);
   }
#endif//CSTRING_WCHAR
};

//----------------------------

#endif
