#ifndef __EDITOR_H
#define __EDITOR_H

#include <IGraph2.h>
#include <i3d\i3d2.h>
#include <c_unknwn.hpp>
#include <c_string.hpp>
#include <vector>
#include <smartptr.h>
#include <list>
#include <i3d\insanity.h>

#ifdef _MSC_VER
using namespace std;
#endif

/*****************************************************************************************
                                    - Insanity editor -  
   Written by: Michal Bacik
   Date: november 1999
   Purpose: Full editing support for Insanity 3D engine. Non-portable, not to be
     included in final release of game.
   Note: All referenced to editor must be removed in release compilation of product.
   
   Interfaces:

   C_editor_item - editor plugin, interface performing simple or complex editing tasks,
      for example "Selection" plugin takes care about keeping, loading, saving, rendering
      current selection of no, one or more 3D objects in 3D scene. Plugins may communicate
      one with another by C_editor_item::Action() method.
      
   C_editor - editor, managing all installed plugins. Communicates with game and plugins,
      registers plugins, manages menus, sends user input to plugins, etc.
*****************************************************************************************/

                              //Following are descriptions of plugins, with enumeration
                              // of their actions, parameters (optional) and return values
                              // of the actions (optional).
                              //Each decription contains short info about plugin's
                              // purpose, dependence list, followed by symbolic names
                              // of the actions.
//----------------------------
//----------------------------
                              //Plugin: "Undo"
                              //Dependence list: no
                              //Purpose: undo/redo functionality - keeping and performing
                              //    reactions on actions made by plugins
                              //Notes: doesn't need any specialized information about any
                              //    particular plugin

enum E_ACTION_UNDO{           
   E_UNDO_UNDO,               // - undo until break
   E_UNDO_REDO,               // - redo until break
   E_UNDO_SAVE,               //(S_edit_Undo_2*) - save undo information
   E_UNDO_BREAK,              //insert break code - to be called by C_editor 
   E_UNDO_CLEAR,              // - clear both undo and redo buffers
   E_UNDO_GET_TOP_PLUGIN,     //ret: const char *plg_name; NULL if undo empty
   E_UNDO_GET_TOP_ACTION,     //ret: int action; -1 if undo empty
   E_UNDO_GET_TOP_CONTEXT,    //ret: void *context; NULL if undo empty
   E_UNDO_POP_TOP,            //ret: bool success - removes top undo entry
};

struct S_edit_Undo_2{         //undo save information - reaction info
   const char *plugin_name;   //plugin name which will be called
   int action_id;             //id passed to C_editor_item::Action()
   void *data;                //data passed to C_editor_item::Action()
   dword data_size;           //size of stored data

   S_edit_Undo_2():
      plugin_name(NULL),
      data(NULL),
      data_size(0)
   {}
   S_edit_Undo_2(const char *pn1, int ai1 = 0, void *d1 = NULL, dword ds1 = 0):
      plugin_name(pn1),
      action_id(ai1),
      data(d1),
      data_size(ds1)
   {}
};

//----------------------------
                              //Plugin: "Grid"
                              //Dependence list: no
                              //Purpose: draws axis-aligned grid

enum E_ACTION_GRID{          
   E_GRID_TOGGLE = 1000,      // - toggle grid on/off
   E_GRID_TOGGLE_X,           // - toggle grid on x axis
   E_GRID_TOGGLE_Y,           // - toggle grid on y axis
   E_GRID_TOGGLE_Z,           // - toggle grid on z axis
   E_GRID_CONFIG,             // - configure grid 
};

//----------------------------
                              //Plugin: "Selection"
                              //Dependence list: "Undo", 
                              //Run-time use: "MouseEdit"
                              //Purpose: keeps current selection
                              //Notifications: each time selection changes, notification is sent
                              // to registered plugins, using registered action ID and passing
                              // vector<PI3D_frame>* as context parameter
typedef const char *(t_GetFrameScript)(PI3D_frame);
typedef const char *(t_GetFrameActor)(PI3D_frame);

enum E_ACTION_SELECTION{      
   E_SELECTION_GETVECTOR = 2000, // - get selection, ret: vector<PI3D_frame>*
   E_SELECTION_GET_SINGLE,    // - get single selected frame, or NULL if not single selection; ret: PI3D_frame
   E_SELECTION_CLEAR,         // - clear selection
   E_SELECTION_CLEAR2,        // - clear selection with message
   E_SELECTION_INVERT,        // - invert selection with message
   E_SELECTION_ADD_FRAME,     //(PI3D_frame) - add frame to selection
   E_SELECTION_DEL_FRAME,     //(PI3D_frame) - remove frame from selection
   E_SELECTION_ADD_FRAMES,    //(char *names - separated by \0) - add frames to selection, by names
   E_SELECTION_DEL_FRAMES,    //(char *names - separated by \0) - add frames to selection, by names
   E_SELECTION_BROWSE,        // - select by browsing
   E_SELECTION_PARENTS,       // - select parent frames of currently selected frames
   E_SELECTION_CHILDREN,      // - select child frames of current selected frames
   E_SELECTION_FLASH_FRAME,   //(PI3D_frame) - flash selected frame (visual effect)
   E_SELECTION_BLINK_FRAME,   //(PI3D_frame) - blink selected frame forever (visual effect)
   E_SELECTION_CLEAR_FLASH,   // - clear list of flashing frames
   E_SELECTION_PROMPT,        //(vector<PI3D_frame>*); ret: bool OK - let user select from list of frames
   E_SELECTION_CONFIG,        // - configure params
   E_SELECTION_SET_GET_SCR,   //(t_GetFrameScript*) - set function which returns frame's script
   E_SELECTION_SET_GET_ACT,   //(t_GetFrameActor*) - set function which returns frame's actor name

   E_SELECTION_ADD_NOTIFY,    //(pair<LPC_editor_item, dword message>*) - set notification to send when selection changes
   E_SELECTION_DEL_NOTIFY,    //(LPC_editor_item)

   E_SELECTION_TOGGLE_LINK,   // - toggle drawing link on selection
   E_SELECTION_SEL_FLASH_NEXT,// - select next frame from the flash list
   E_SELECTION_SEL_FLASH_PREV,// -   ''   prev     ;;
};

//----------------------------
                              //Plugin: "Modify"
                              //Dependence list: "Undo", "Selection"
                              //Purpose: keeps modifications on frames
                              //Notifications: each time modification on frame changes, 
                              // notification is sent to registered plugins, 
                              // using registered action ID and passing
                              // pair<PI3D_frame, dword flags>* as context parameter
#pragma pack(push,1)

struct S_edit_Modify_undo_pos{
   int num_frames;
   struct S_frm{
      S_vector pos;
      char frm_name[1];
   };
   S_frm frm_info[1];
};

struct S_edit_Modify_undo_rot{
   int num_frames;
   struct S_frm{
      S_vector pos;           //for case of rotation of multi-selection
      S_quat rot;
      bool use_pos;
      char frm_name[1];
   };
   S_frm frm_info[1];
};

struct S_edit_Modify_undo_scl{
   int num_frames;
   struct S_frm{
      S_vector scl;
      bool use_non_uniform;
      char frm_name[1];
   };
   S_frm frm_info[1];
};

struct S_edit_Modify_undo_lnk{
   int num_frames;
   struct S_frm{
      S_vector orig_pos;
      S_quat orig_rot;
      float orig_scl;
      char frm_names[1];      //frm_name, link_name
   };
   S_frm frm_info[1];
};

struct S_edit_Modify_undo_light{
   int num_frames;
   struct S_frm{
      I3D_LIGHTTYPE ltype;
      S_vector color, specular_color;
      float power, cone_i, cone_o, range_n, range_f, specular_power;//, specular_angle;
      dword mode;
      char frm_names[1];
   } frm_info[1];
};

struct S_edit_Modify_undo_sound{
   int num_frames;
   struct S_frm{
      I3D_SOUNDTYPE stype;
      float vol, out_vol, fade_ratio, cone_i, cone_o, range_n, range_f;
      bool loop, streaming, on;
      char frm_names[1];      //frm_name, snd_name
   } frm_info[1];
};

struct S_edit_Modify_undo_volume{
   int num_frames;
   struct S_frm{
      I3D_VOLUMETYPE vtype;
      int material_id;
      char frm_name[1];      
   } frm_info[1];
};

struct S_edit_Modify_undo_visual{
   int num_frames;
   struct S_frm{
      dword visual_type;
      int material_id;
      word data_size;
                              //NULL-terminated visual name, followed by all properties
                              // with values, each property stored in its actual size
      char data[1];       
   } frm_info[1];
};

struct S_edit_Modify_undo_sector{
   int num_frames;
   struct S_frm{
      dword env_id;
                              //NULL-terminated frame name
      char frm_name[1];       
   } frm_info[1];
};

struct S_edit_Modify_undo_frmflags{
   int num_frames;
   struct S_frm{
      dword frm_flags;
      char frm_name[1];      
   } frm_info[1];
};

#pragma pack(pop)

typedef void t_modify_enum(PI3D_frame, dword flags, dword context);

enum E_ACTION_MODIFY{         
   E_MODIFY_ADD_FLAGS = 3000, //(pair<PI3D_frame, dword flags>*) - add flags to frame, saves original values
   E_MODIFY_DEL_FLAGS,        //(pair<PI3D_frame, dword flags>*) - removes flags of frame
   E_MODIFY_DEL_FRAME,        //(PI3D_frame); ret: (bool)status - removes frame from list
   E_MODIFY_GET_FLAGS,        //(PI3D_frame); ret: flags - get frame's flags
   E_MODIFY_REPLACE_FRAME,    //(pair<PI3D_frame orig, PI3D_frame new>); ret: (bool)success - replace one frame for another (due casting)
   E_MODIFY_RESET,            // - reset modifications on selection 
   E_MODIFY_GET_NUM_ITEMS,    //ret: dword number_of_items - get number of items in list
   E_MODIFY_GET_ITEM_LINK,    //(PI3D_frame); ret: const char *link_name - get name of frame's parent
   E_MODIFY_ENUM,             //(pair<t_modify_enum*, dword contex> ) - enum modified frames
   E_MODIFY_ADD_NAMED,        //internal
   E_MODIFY_RESET_NAMED,      //internal
   E_MODIFY_TRANSLATE_NAMED,  // - internal (undo)
   E_MODIFY_ROTATE_NAMED,     // - internal (undo)
   E_MODIFY_ROT_TRANS_NAMED,  // - internal (undo)
   E_MODIFY_SCALE_NAMED,      // - internal (undo)
   E_MODIFY_LINK_NAMED,       // - internal (undo)
   E_MODIFY_LIGHT,            //(S_edit_Modify_undo_light) - internal (undo)
   E_MODIFY_SOUND,            //(S_edit_Modify_undo_sound) - internal (undo)
   E_MODIFY_VOLUME,           //(S_edit_Modify_undo_volume) - internal (undo)
   E_MODIFY_VISUAL,           //(S_edit_Modify_undo_visual) - internal (undo)
   E_MODIFY_SECTOR,           // - internal (undo)
   E_MODIFY_FRMFLAGS,         //(S_edit_Modify_undo_frmflags) - internal (undo)

   E_MODIFY_ADD_MODEL_DEL_CHILD, //(pair<PI3D_frame, const char *child_name>*) - add frame which should be deleted from model
   E_MODIFY_DEL_MODEL_DEL_CHILD, //(pair<PI3D_frame, const char *child_name>*) - remove frame which should be deleted from model, if .second is NULL, remove all existing frames

   E_MODIFY_ADD_NOTIFY,       //(pair<LPC_editor_item, dword message>*) - set notification to send when modification is made on frame(s)
   E_MODIFY_DEL_NOTIFY,       //(LPC_editor_item)

                              //following are modification flags:
   E_MODIFY_FLG_POSITION   = 0x1,      //I3D_frame
   E_MODIFY_FLG_ROTATION   = 0x2,      // ''
   E_MODIFY_FLG_SCALE      = 0x4,      // ''
   E_MODIFY_FLG_NU_SCALE   = 0x8,      //I3D_visual
   E_MODIFY_FLG_LINK       = 0x10,     //I3D_frame
   E_MODIFY_FLG_LIGHT      = 0x20,     //I3D_light
   E_MODIFY_FLG_SOUND      = 0x40,     //I3D_sound
   E_MODIFY_FLG_VOLUME     = 0x80,     //I3D_volume
   E_MODIFY_FLG_PARTICLE   = 0x200,    //I3D_particle
   E_MODIFY_FLG_VISUAL     = 0x400,    //I3D_visual
   E_MODIFY_FLG_SECTOR     = 0x800,    //I3D_sector
   E_MODIFY_FLG_MATERIAL   = 0x1000,   //collision material(I3D_visual/I3D_volume)

   E_MODIFY_FLG_HIDE       = 0x10000,  //hidden frame
   E_MODIFY_FLG_CREATE     = 0x20000,  //frame created in editor
   E_MODIFY_FLG_TEMPORARY  = 0x40000,  //temporary frame, modifications not saved on disk
   E_MODIFY_FLG_ALL        = 0x7fffffff,
};

//----------------------------
                              //Plugin: "Create"
                              //Dependence list: "Undo", "Selection", "Modify"
                              //Purpose: create various frame types

enum E_ACTION_CREATE{         
   E_CREATE_MODEL = 4000,     // - create model and initialize it by loading from a file
   E_CREATE_LIGHT,            
   E_CREATE_SOUND,             
   E_CREATE_VOLUME,            
   E_CREATE_CAMERA,            
   E_CREATE_DUMMY,            
   E_CREATE_TARGET,
   E_CREATE_OCCLUDER,
   E_CREATE_SECTOR,
   E_CREATE_DUPLICATE_SELECTION, // - duplicate all duplicate-able frames from current selection
   E_CREATE_DUPLICATE_REPEAT, 
   E_CREATE_DELETE_SELECTION, // - delete all delete-able frames from current selection
   E_CREATE_DELETE_FRAME,     //(PI3D_frame) - delete specified frame
   E_CREATE_BROWSE_MODEL,     //(char[256]); ret: bool ok; - browse model filename in common directories
   E_CREATE_BROWSE_SOUND,     //(char[256]); ret: bool ok; - browse sound filename in common directories
   E_CREATE_FRAME_NAMED,      // - internal (undo) - re-create frame
   E_CREATE_CHILD_NAMED,      // - internal (undo) - re-create child of model
   E_CREATE_DELETE_NAMED,     //(const char *frm_name) - internal (undo)
   E_CREATE_NOTIFY_SELECTION, //(vector<PI3D_frame> *sel_list) - notification received from selection
   E_CREATE_PICK_OCCL_VERTEX, // - user-pick mode
   E_CREATE_VERT_SELECTION,   // - undo info for occluder creation
   E_CREATE_DEL_VERT_SEL,     // - undo info for occluder creation
};

//----------------------------
                              //Plugin: "Hide"
                              //Dependence list: "Undo", "Selection", "Modify"
                              //Purpose: hide/unhide frames

enum E_ACTION_HIDE{           
   E_HIDE_SELECTION = 5000,   // - hide selected frames
   E_HIDE_UNHIDE_ALL,         // - unhide all hidden frames
   E_HIDE_UNHIDE_PROMPT,      // - ask user which frames to unhide
   E_HIDE_HIDE_NAMED,         //internal (undo)
   E_HIDE_UNHIDE_NAMED,       //internal (undo)

   E_HIDE_NOTIFY_SELECTION,   //(vector<PI3D_frame> *sel_list) - notification received from selection
};

//----------------------------
                              //Plugin: "Screen"
                              //Dependence list: no
                              //Purpose: changes screen resolution
enum E_ACTION_SCREEN{
   E_SCREEN_FULLSCREEN = 6000,// - toggle full-screen mode
   E_SCREEN_NEXT_RES,         // - cycle forward through available resolutions
   E_SCREEN_PREV_RES,         // - cycle backward through available resolutions
};

//----------------------------
                              //Plugin: "MouseEdit"
                              //Dependence list: "Undo", "Selection", "Modify"
                              //Purpose: main edit utility, picking objects, making
                              //    basic 3d operations on selection (position, rotation, scale), 
                              //    linking, camera movement
                              //Notes:
                              // User pick mode: any plugin may set exclusive mode, when picking
                              //    is done for this plugin. When in this mode, cursor is changed
                              //    and any mouse-click evaluates to single frame (or NULL),
                              //    which is sent to registered plugin as:
                              //    pick_plugin->Action(pick_action, (S_MouseEdit_picked*)pick_info);
                              //    When registered plugin returns false on this call,
                              //    it is unregistered from further picking, otherwise it
                              //    remains registered with last parameters.
                              // Override picking: a message EDIT_ACTION_PICK with pointer to S_MouseEdit_picked
                              //    is broadcast to all plugins by "MouseEdit" plugin to let them
                              //    override results of clicking on frame.
                              //    Plugins may write to 'frm_picked', 'pick_fist' and 'pick_action'
                              //    to pick non-visual frames, or override default action
                              //    resulting from picking a frame.

struct S_MouseEdit_pick{      //struct used with E_MOUSE_SET_USER_PICK
   class C_editor_item *ei;   //plugin called on pick
   int action_id;             //action sent on pick
   void *hcursor;             //handle to windows cursor, or NULL for default
};

struct S_MouseEdit_picked{    //struct passed to a plugin, when picking occured
   PI3D_frame frm_picked;    //NULL if no frame collided
   int screen_x, screen_y;    //screen pick (mouse) position
   S_vector pick_from, pick_dir; //picking ray, pick_dir is normalized
   float pick_dist;           //distance of collision, valid only if frm_picked is not NULL
   S_vector pick_norm;        //normal of collision, valid only if frm_picked is not NULL
   enum E_PICK_ACTION{        //action being performed (controlled by Ctrl/Shift/Alt held during pick)
      PA_NULL,                // - no action
      PA_DEFAULT,             // - classical pick
      PA_TOGGLE,              // - toggle current selection
      PA_LINK,                // - link picked
      PA_VOLUME,              // - select volume
      PA_EDIT_ONLY,           // - only ediable frames
      PA_NONPICKABLE,         // - default test plus nonpickable visuals
   } pick_action;
};

struct S_MouseEdit_user_mode{ //struct used with E_MOUSE_SET_USER_MODE
   class C_editor_item *ei;   //plugin which receives events
   struct S_mode{
      void *hcursor;          //cursor to use with this mode
      int action_id;          //ID to call ei->Action with
      const char *desc;       //friendly mode description
   };
   enum E_MODE{               //user action called with params:
      MODE_DEFAULT,           //(void)
      MODE_SELECT,            //(S_MouseEdit_picked*)
      MODE_MOVE_XZ,           //(S_vector *delta) - camera XZ coordinates
      MODE_LINK,              //(S_MouseEdit_picked*)
      MODE_MOVE_Y,            //(S_vector *delta) - screen Y
      MODE_SCALE,             //(float) relative scale
      MODE_ROTATE,            //(S_quat*)
      MODE_PICK_SPEC,         //(S_MouseEdit_picked*)
      MODE_DOUBLECLICK,       //(S_MouseEdit_picked*)
      MODE_DELETE,            //(void)
      MODE_LAST
   };
   S_mode modes[MODE_LAST];
};

enum E_ACTION_MOUSE{          
   E_MOUSE_GET_LOC_SYSTEM = 7000,   //ret: bool - get status of local system flag
   E_MOUSE_GET_UNIFORM,       //ret: bool - get status of uniform scale flag
   E_MOUSE_GET_ROT_AXIS,      //ret: dword axis_bits - get selected rotation axis, represented by first 3 bits, one for each axis
   E_MOUSE_GET_SPEED,         //ret: float - get movement speed
   E_MOUSE_TOGGLE_MODE,       // - switch between editing modes
   E_MOUSE_SET_MODE,          //(dword mode) [0=edit, 1=view] - set editing mode
   E_MOUSE_GET_MODE,          // - get editing mode [0=edit, 1=view]
   E_MOUSE_SET_USER_PICK,     //(S_MouseEdit_pick*) - set user pick mode, or cancel if NULL passed
   E_MOUSE_GET_USER_PICK,     //ret: LPC_editor_item user_pick_plugin - get currently set user-pick plugin, or NULL if no
   E_MOUSE_GET_EDIT_CAMERA,   //ret: PI3D_camera
   E_MOUSE_SET_CURR_CAMERA,   //(PI3D_frame) - setup current camera used by editor
   E_MOUSE_SET_USER_MODE,     //(S_MouseEdit_user_mode*) - set user edit mode, or cancel if NULL passed
   E_MOUSE_GET_USER_MODE,     //ret: S_MouseEdit_user_mode* - get currently set user edit mode, or NULL if no
   E_MOUSE_CANCEL,            // - ESC key to cancel user pick mode
   E_MOUSE_ROT_45_LEFT,       // - rotate selection 45 degrees to the left
   E_MOUSE_ROT_45_RIGHT,      // - rotate selection 45 degrees to the right
   E_MOUSE_POLAR_MODE,        //(bool) - set polar mode
   E_MOUSE_GET_POLAR_MODE,    //ret: bool - get polar mode
   E_MOUSE_POLAR_ALIGN,       //align selection to polar coordinates
   E_MOUSE_POLAR_ALIGN_FRAME, // - internal (PI3D_frame) - align frame to polar coordinates, no undo!
   E_MOUSE_ALIGN_TO_NORMAL,   //align selection to picked frame
   E_MOUSE_ALIGN_PICK,        // - internal - picked frame to alight with
   E_MOUSE_EDIT_SUBOBJECT,    // - toggle subobject edit mode

   E_MOUSE_CAMERA_MOVE,       //internal (undo)
};

//----------------------------
                              //Plugin: "LightMap"
                              //Dependence list: "Undo", "Selection", "Modify"
                              //Purpose: construction and destruction of light-maps

enum E_ACTION_LMAP{           //Plugin: "LightMap"
   E_LMAP_SELECT_NOTC = 8000, // - construct all unconstructed
   E_LMAP_CONSTRUCT_SEL,      // - (re)construct selected
   E_LMAP_CONSTRUCT_ALL,      // - (re)construct all
   E_LMAP_DESTRUCT_SEL,       // - destruct selected
   E_LMAP_DESTRUCT_ALL,       // - destruct all
   E_LMAP_DESTRUCT_NAMED,     //(const char *frm_name) - destruct named frame
   E_LMAP_CONSTRUCT_NAMED,    //(data: frm_name, '\0', lightmap) - re-construct named frame
   E_LMAP_TOGGLE_BLUR,
   E_LMAP_TOGGLE_MIPMAP,
   E_LMAP_AA_1,               // - setup 1x1 antialiasing
   E_LMAP_AA_2,
   E_LMAP_AA_3,
   E_LMAP_AA_4,
   E_LMAP_AA_5,
   E_LMAP_AA_6,
   E_LMAP_RECONSTRUCT,        // - reconstruct all constructed LMaps using current settings
};

//----------------------------
                              //Plugin: "Stats"
                              //Dependence list: no
                              //Purpose: statistics about various items
enum E_ACTION_STATS{
   E_STATS_FPS = 9000,        // - toggles FPS counter
   E_STATS_DISABLE,           // - hides all stats
   E_STATS_VIDEO,             // - toggle video stats
   E_STATS_SYSMEM,            // - toggle sysmem stats
   E_STATS_COLLISIONS,        // - toggle collision stats
   E_STATS_SCENE,             // - toggle scene stats
   E_STATS_SELECTION,         // - toggle stats about current scene
   E_STATS_SOUND,             // - toggle sound stats 
   E_STATS_NET,               // - toggle network stats
   E_STATS_DETAILED,          // - toggle detailed stats
   E_STATS_BSP,               // - toggle bsp stats
   E_STATS_VOLTREE,           // - toggle volume tree stats
   E_STATS_SET_NET_TEXT,      //(const char*) - save network stats string, for displaying if net stats are selected
   /*
   E_STATS_STATUS_BAR,        // - toggle status bar
   E_STATS_ADD_SB_TEXT,       //(const char*) - internal - set SB text
   E_STATS_SET_MODIFIED,      //(bool) - internal - set modified flag
   */
};

//----------------------------
                              //Plugin: "View"
                              //Dependence list: no
                              //Purpose: toggles Insanity 3D engine states
enum E_ACTION_VIEW{
};

//----------------------------
                              //Plugin: "Properties"
                              //Dependence list: "Undo", "Selection", "Modify"
                              //Run-time use: "MouseEdit", "Create"
                              //Purpose: provide properties window

enum E_ACTION_PROPERTIES{     
   E_PROP_TOGGLE = 10000,     // - toggle properties on/off
   E_PROP_ADD_SHEET,          //(HWND); ret: bool success
   E_PROP_DEL_SHEET,          //(HWND); ret: bool success
   E_PROP_SHOW_SHEET,         //(HWND) - select sheet
   E_PROP_ACTIVATE,           // - set focus to Property window
   E_PROP_RENAME_FRAME,       //(char *old_name "\0" char *new_name) - internal (undo)
   E_PROP_RELOAD_MODEL,       //(char *frm_name "\0" char *model_name) - internal (undo)
   E_PROP_RELOAD_SOUND,       //(char *frm_name "\0" char *sound_name) - internal (undo)

   E_PROP_NOTIFY_SELECTION,   //(vector<PI3D_frame> *sel_list) - notification received from selection
   E_PROP_NOTIFY_MODIFY,      //(pair<PI3D_frame, dword flags>*) - notification received from modify
   E_PROP_RESET_SHEET,        //reset current sheet
   E_PROP_NOTIFY_PICK_TARGET, //(PI3D_frame) - notification received from MouseEdit
   E_PROP_NOTIFY_PICK_MODEL,  //(PI3D_frame) - notification received from MouseEdit
   E_PROP_NOTIFY_PICK_LIGHT,  //(PI3D_frame) - notification received from MouseEdit
   E_PROP_NOTIFY_PICK_LSECT,  //(PI3D_frame) - notification received from MouseEdit
   E_PROP_NOTIFY_PICK_SSECT,  //(PI3D_frame) - notification received from MouseEdit
   E_PROP_NOTIFY_PICK_VISUAL, //(PI3D_frame) - notification received from MouseEdit

   E_PROP_SAVE_UNDO_VISUALS,  // - internal - save undo info for all viusals
};

//----------------------------
                              //Plugin: "Log"
                              //Dependence list: "Properties"
                              //Purpose: show log

enum E_ACTION_LOG{
   E_LOG_ADD_TEXT = 11000,    //(const char*) - add text line to log window
   E_LOG_ADD_RICH_TEXT,       //(pair<const char *text, dword color>*) - add text line (in rich-edit format) to log window
   E_LOG_CLOSE,               // - close log window 
   E_LOG_ADD_TEXT_QUIET,      //(const char*) - add text line to log window, don't activate log
   E_LOG_SET_DUMP_TEXT,       //(const char*) - set contents of dump window
   E_LOG_CLOSE_DUMP,          // - close dump window
};

//----------------------------
                              //Plugin: "Gravitation"

enum E_ACTION_GRAVITATION{
   E_GRAV_SELECTED_DOWN = 12000,//fall selected objects
   E_GRAV_UNDO_FALL,          // - internal - undo
   E_GRAV_TOGGLE_DIALOG,      //toggle gravitation dialog on/off
   E_GRAV_LINK_TO_BACKGROUND, //link to background
   E_GRAV_SET_PARENT,         // - internal (undo link) 
};

//----------------------------
                              //Plugin: "Seed"

enum E_ACTION_SEED{
   E_SEED_TOGGLE_DIALOG = 13000, //toggle seed dialog on/off
   E_SEED_PICK_MODEL,         // - internal callback, pick model
   E_SEED_PUT_MODEL,          // - internal callback, put random model
};

//----------------------------
                              //Plugin: "Exit"
                              //Dependence list: 
                              //Purpose: exit from editor
enum E_ACTION_EXIT{
   E_EXIT_GET_STATE = 14000,  //ret: bool do_exit
   E_EXIT_SET_STATE,          //(bool do_exit)
};

//----------------------------
                              //Plugin: "DebugLine"
                              //Dependence list: 
                              //Purpose: draw debugging lines of following types:
                              // - one-time
                              // - persistent
                              // - fading
struct S_add_line{
   S_vector v0, v1;
   dword color, time;
   S_add_line(){}
   S_add_line(S_vector v_0, S_vector v_1, dword color1 = 0xffffffff, dword time1 = 500):
      v0(v_0), v1(v_1), color(color1), time(time1)
   {}
};

enum E_ACTION_DEBUGLINE{
   E_DEBUGLINE_ADD = 15000,   //(S_add_line*)
   E_DEBUGLINE_ADD_SINGLE,    //(S_add_line*)
   E_DEBUGLINE_ADD_TIMED,     //(S_add_line*)
   E_DEBUGLINE_CLEAR,

                              //toggle debug collision testing (between frames "from" and "to")
                              // old style:
   E_DEBUGLINE_GEO_COLLISION, // - geo line test
   E_DEBUGLINE_VOL_COLLISION, // - vol line test
   E_DEBUGLINE_SPH_COLLISION, // - vol sphere test
                              // new style:
   E_DEBUGLINE_COLTEST_STANDARD, // - standard test
   E_DEBUGLINE_COLTEST_EXACT,    // - exact test
   E_DEBUGLINE_COLTEST_SPHERE,   // - sphere test
   E_DEBUGLINE_COLTEST_STATIC_SPHERE, // - static sphere test
};

//----------------------------
                              
enum E_ACTION_LIGHT{          //Plugin: "Light"
   E_LIGHT_TOGGLE_HELP = 16000,  // - toggle helping light
   E_LIGHT_MAKE_HELP,         // - create helping light
   E_LIGHT_DESTROY_HELP,      // - destroy helping light
};

//----------------------------

enum E_ACTION_STATUSBAR{
   E_STATUSBAR_GET_HWND = 17000, //ret: HWND
};

//----------------------------
                              //Plugin: "Bsp-tools"
enum E_ACTION_BSP{
   E_BSP_UNDO_TOGGLE_STATIC = 18000,
   E_BSP_CREATE,              //ret: bool success; build BSP tree on scene
   E_BSP_TOGGLE_SHOW,
   E_BSP_DESTROY,
   E_BSP_TOGGLE_STATIC_SELECTION,
   E_BSP_INFO,
};

//----------------------------
                              //Plugin: "Mission"
enum E_ACTION_MISSION{
   E_MISSION_BROWSE = 19000,  //(pair<HWND,char[256]>); ret: bool ok; - browse mission filename in common directories
   E_MISSION_LAST
};

//----------------------------
//----------------------------
                              //plug-in system
class C_editor_item{
   dword ref;
   struct S_menu_item{
                              //keyboard shortcut used to invoke command
      char key;
      byte key_modifier;
      int id;                 //action ID
                              //menu access
      void *hmenu_parent;
      int menu_id;            //unique menu item identifier, 0 if menu item not created
      bool enabled;
   };

   template<class T> 
   class C_allocator: public allocator<T> {
   public:
      virtual pointer allocate(size_type n, const void *hint){
         //return (pointer)operator new(n * sizeof(T));
         return (pointer)malloc(n * sizeof(T));
      }
      virtual void deallocate(pointer p, size_type n){
         //operator delete(p);
         free(p);
      }
   };    

   vector<S_menu_item, C_allocator<S_menu_item> > menu_items;
   virtual const char *GetName() const = 0;
   friend class C_editor_imp;
protected:
   class C_editor *ed;
public:
   C_editor_item(): ref(1){}
   virtual ~C_editor_item(){ ed = NULL; }
   dword AddRef(){ return ++ref; }
   dword Release(){ if(--ref) return ref; Close(); delete this; return 0; }

   virtual bool Init() = 0;
   virtual void Close(){ menu_items.clear(); }

   virtual dword Action(int id, void *context = NULL){ return 0; }
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt)
   {}
   virtual void Render(){}

   virtual bool LoadState(class C_chunk*){ return false; }
   virtual bool SaveState(class C_chunk*) const{ return false; }
};
typedef C_editor_item *LPC_editor_item;

//----------------------------
//----------------------------
                              //editor plugin template
/*
                              //editor plugin: Grid
class C_edit_Grid: public C_editor_item{
                              //returns plugin's name
   virtual const char *GetName() const{ return "Grid"; }
public:
                              //initialize plugin
   virtual bool Init(){
                              //add menu items (and keyboard short-cuts)
      ed->AddShortcut(this, 1000, "&Stats\\&Video\tCtrl+1", K_1, SKEY_CTRL);
      return true;
   }
                              //uninitialize plugin
   virtual void Close(){ }
                              //tick - performed in every rendering frame
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}
                              //action - called from menu/keypress, or by other plugins
   virtual dword Action(int id, void *context){
      switch(id){
      case 0:
         break;
      }
      return 0;
   }
                              //called for registered plugins during rendering
   virtual void Render(){}
                              //store/restore plugin's state
   virtual bool LoadState(class C_chunk *ck){ return false; }
   virtual bool SaveState(class C_chunk *ck) const{ return false; }
};
*/

//----------------------------
//----------------------------

enum E_EDIT_MESSAGE{
   EM_MESSAGE,
   EM_WARNING,
   EM_ERROR,
};

enum E_EDIT_ACTION{           //pre-defined actions
   EDIT_ACTION_AFTER_LOAD = 0x40000000,      //after scene was loaded
   EDIT_ACTION_BEFORE_FREE,   //before scene is being freed
   EDIT_ACTION_MISSION_SAVE_0,//(C_chunk*) scene is being saved
   EDIT_ACTION_MISSION_SAVE_1,// like EDIT_ACTION_MISSION_SAVE_0, but called after it
   EDIT_ACTION_MISSION_SAVE_2,// like EDIT_ACTION_MISSION_SAVE_1, but called after it
   EDIT_ACTION_MISSION_LOAD,  //(C_chunk*) scene is loaded - found chunk for plugin
   EDIT_ACTION_WINDOW_RESET,  //(LPIGraph) - main window is re-created
   EDIT_ACTION_FRAME_DELETE,  //(PI3D_frame) - frame is being deleted
   EDIT_ACTION_FRAME_DUPLICATE,//(pair<PI3D_frame old, PI3D_frame new>*) - frame is being duplicated
   EDIT_ACTION_PICK,          //(S_MouseEdit_picked*) - picking - let plugins override it, they may write to 'frm_picked' and 'pdist' members
   EDIT_ACTION_SET_ACTIVE,    //(bool) - activate/deactivate editor
   EDIT_ACTION_QUERY_MODIFY,  //(bool quiet); ret: int; query if scene may be modified (ret: 0=yes, 1=no)
   EDIT_ACTION_VALIDATE,      //ret: bool; scene validation, plugin returning true marks error and stop further processing
   EDIT_ACTION_SUBOBJ_EDIT,   //(PI3D_frame); ret: bool; switch to subobject edit mode
   EDIT_ACTION_SUBOBJ_CANCEL, //cancel sub-object edit mode
   EDIT_ACTION_FEED_MATERIALS,//(pair<int, const char*>[]) - feed material table, terminated by pair<>(-1, NULL)
   EDIT_ACTION_FEED_ENVIRONMENTS,//(pair<int, const char*>[]) - feed sound environment table, terminated by pair<>(-1, NULL)
};

class C_editor{
public:
   virtual dword AddRef() = 0;
   virtual dword Release() = 0;

   virtual void MenuHit(int menu_id, char key, byte key_modifier) = 0;
   virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt) = 0;
   virtual void Render() = 0;
   virtual int CreateStatusIndex(int width) = 0;
   virtual void Message(const char *str, int sb_index = 0, E_EDIT_MESSAGE = EM_MESSAGE, bool immediate = false) = 0;
   virtual void BroadcastAction(E_EDIT_ACTION, void *context) = 0;
   virtual void EnumPlugins(bool(*)(LPC_editor_item, void *context), void *context) = 0;

   virtual bool LoadState(const char *bin_name) = 0;
   virtual bool SaveState(const char *bin_name) const = 0;

//----------------------------
// Check/uncheck modify flag.
   virtual void SetModified(bool = true) = 0;

//----------------------------
// Check if edited scene is modified.
   virtual bool IsModified() const = 0;

//----------------------------
// Query if modifications can be made.
   virtual bool CanModify(bool quiet = false) = 0;

   virtual bool InstallPlugin(LPC_editor_item) = 0;
   virtual bool RemovePlugin(LPC_editor_item) = 0;
   virtual bool RegForRender(LPC_editor_item, int render_order = 1000) = 0;
   virtual bool UnregForRender(LPC_editor_item) = 0;

                              //called by plugins during insitialization
   virtual bool AddShortcut(LPC_editor_item ei, int id, const char *menu_item, char key, byte key_modifier) = 0;
   virtual bool DelShortcut(LPC_editor_item ei, int id) = 0;

   virtual bool IsInit() const = 0;
   virtual LPIGraph GetIGraph() const = 0;
   virtual PI3D_driver GetDriver() const = 0;
   virtual PI3D_scene GetScene() const = 0;

   virtual LPC_editor_item FindPlugin(const char *plugin_name) const = 0;
                              //functions to be called bu plug-ins
   virtual void CheckMenu(LPC_editor_item, int id, bool) = 0;
   virtual void EnableMenu(LPC_editor_item, int id, bool) = 0;

   virtual void SetActive(bool) = 0;
   virtual bool IsActive() const = 0;

   virtual C_I3D_cache_base<I3D_model> &GetModelCache() = 0;
   virtual C_I3D_cache_base<I3D_sound> &GetSoundCache() = 0;
};

typedef C_editor *LPC_editor;

//----------------------------

#pragma comment (lib,"IEditor")

//----------------------------
// Editor creation.
// Note on plugins initialization:
// - plugins are initialized by C_editor::InstallPlugin
// - plugin DLLs are searched in ".\Plugins\*.dll" path, in each found DLL
//    an initialization function is searched, which is declared as:
//    extern "C" void __stdcall Initialize(LPC_editor)
//    this function may add any amount of plugins
//----------------------------
LPC_editor __declspec(dllexport) CreateEditor(PI3D_driver, PI3D_scene, C_I3D_cache_base<I3D_model> *model_cache,
   C_I3D_cache_base<I3D_sound> *sound_cache);

//----------------------------

#endif
