/*--------------------------------------------------------
   Copyright (c) 2002 Lonely Cat Games
   All rights reserved.

   File: I3D_eng.h
   Content: Insanity 3D II interfaces include file
--------------------------------------------------------*/

#ifndef __ENG_3DI_
#define __ENG_3DI_
#pragma once

#include <I3D\I3D_math.h>
#include <memory.h>
#include <smartptr.h>

#pragma comment (lib,"i3d2_thunk")

# define I3DAPI __stdcall
# define I3DMETHOD_(type,method) virtual type I3DAPI method
# define I3DMETHOD(method) virtual I3D_RESULT I3DAPI method

//----------------------------

typedef class I3D_scene *PI3D_scene;
typedef const I3D_scene *CPI3D_scene;

typedef class I3D_driver *PI3D_driver;
typedef const I3D_driver *CPI3D_driver;

class C_cache;

//----------------------------
                              //result codes
typedef long I3D_RESULT;
enum{
   I3D_OK,
   I3D_DONE,

   I3DERR_NOFILE = 0x80000001,//unable to open file
   I3DERR_OUTOFMEM = 0x80000002, //out of memory
   I3DERR_TEXTURESNOTSUPPORTED = 0x80000003,
   I3DERR_GENERIC = 0x80000006,  //internal error
   I3DERR_OBJECTNOTFOUND = 0x80000007, //find/delete - no such object in list
   I3DERR_INVALIDPARAMS = 0x80000008,
   I3DERR_CANTSETSECTOR = 0x8000000b,
   I3DERR_NOTINITIALIZED,
   I3DERR_FILECORRUPTED,
   I3DERR_NOFILE1,            //additional file (opacity map)
   I3DERR_CYCLICLINK,         //the linking would cause cyclic hierarchy
   I3DERR_CANCELED,           //operation has been cancelled by user (enumeration, loading, lightmap computation, etc)
   I3DERR_UNSUPPORTED,
};

#ifndef I3D_SUCCESS
# define I3D_SUCCESS(n) ((I3D_RESULT)n>=0)
#endif
#ifndef I3D_FAIL
# define I3D_FAIL(n) ((I3D_RESULT)n<0)
#endif

//----------------------------
                              //enumeration

                              //enum frames flags
#define ENUMF_VISUAL    1
#define ENUMF_LIGHT     2
#define ENUMF_CAMERA    4
#define ENUMF_SOUND     8
#define ENUMF_SECTOR    0x10
#define ENUMF_DUMMY     0x20
#define ENUMF_TARGET    0x40
#define ENUMF_MODEL     0x80
#define ENUMF_USER      0x100
#define ENUMF_VOLUME    0x200
#define ENUMF_JOINT     0x400
#define ENUMF_OCCLUDER  0x800
#define ENUMF_ALL       0x0ffff
#define ENUMF_WILDMASK  0x10000
#define ENUMF_USEMASK   ENUMF_WILDMASK    //obsolete
#define ENUMF_CASESENS  0x20000
#define ENUMF_BACKDROP  0x40000

                              //casting
#define I3DCAST_LIGHT(f)      (static_cast<PI3D_light>((PI3D_frame)f))
#define I3DCAST_SOUND(f)      (static_cast<PI3D_sound>((PI3D_frame)f))
#define I3DCAST_CAMERA(f)     (static_cast<PI3D_camera>((PI3D_frame)f))
#define I3DCAST_OBJECT(f)     (static_cast<PI3D_object>((PI3D_frame)f))
#define I3DCAST_LIT_OBJECT(f) (static_cast<PI3D_lit_object>((PI3D_frame)f))
#define I3DCAST_DUMMY(f)      (static_cast<PI3D_dummy>((PI3D_frame)f))
#define I3DCAST_JOINT(f)      (static_cast<PI3D_joint>((PI3D_frame)f))
#define I3DCAST_USER(f)       (static_cast<PI3D_user>((PI3D_frame)f))
#define I3DCAST_VOLUME(f)     (static_cast<PI3D_volume>((PI3D_frame)f))
#define I3DCAST_SECTOR(f)     (static_cast<PI3D_sector>((PI3D_frame)f))
#define I3DCAST_TARGET(f)     (static_cast<PI3D_target>((PI3D_frame)f))
#define I3DCAST_MODEL(f)      (static_cast<PI3D_model>((PI3D_frame)f))
#define I3DCAST_VISUAL(f)     (static_cast<PI3D_visual>((PI3D_frame)f))
#define I3DCAST_OCCLUDER(f)   (static_cast<PI3D_occluder>((PI3D_frame)f))

                              //enum frames return value
enum I3DENUMRET{
   I3DENUMRET_OK,
   I3DENUMRET_SKIPCHILDREN,
   I3DENUMRET_CANCEL,
};

//----------------------------
                              //bounding box defined by 2 boundary points
struct I3D_bbox{
   S_vector min, max;
   I3D_bbox(){}
   I3D_bbox(const S_vector &n, const S_vector &x){
      min = n;
      max = x;
   }
   inline const S_vector &operator[](int i) const{ return (&min)[i]; }

//----------------------------
// Make bounding-box 'invalid' - expand min to positive infinity and max to negative infinity.
// In this state the bounding-box is prepared for extens expansion.
   inline void Invalidate(){
      min = S_vector(1e+16f, 1e+16f, 1e+16f);
      max = S_vector(-1e+16f, -1e+16f, -1e+16f);
   }

//----------------------------
// Check if bounding-box is valid.
   inline bool IsValid() const{ return(min.x <= max.x); }

//----------------------------
// Expand AA bounding-box, defined by 2 extreme points, into a bounding-box defined by
// 8 corner points.   
   void Expand(S_vector bbox_full[8]) const{
                              //expand bound-box (2 corner points) to full bbox (8 points)
      bbox_full[0].x = min.x; bbox_full[0].y = min.y; bbox_full[0].z = min.z;
      bbox_full[1].x = max.x; bbox_full[1].y = min.y; bbox_full[1].z = min.z;
      bbox_full[2].x = min.x; bbox_full[2].y = max.y; bbox_full[2].z = min.z;
      bbox_full[3].x = max.x; bbox_full[3].y = max.y; bbox_full[3].z = min.z;
      bbox_full[4].x = min.x; bbox_full[4].y = min.y; bbox_full[4].z = max.z;
      bbox_full[5].x = max.x; bbox_full[5].y = min.y; bbox_full[5].z = max.z;
      bbox_full[6].x = min.x; bbox_full[6].y = max.y; bbox_full[6].z = max.z;
      bbox_full[7].x = max.x; bbox_full[7].y = max.y; bbox_full[7].z = max.z;
   }
};

typedef I3D_bbox *PI3D_bbox;
typedef const I3D_bbox *CPI3D_bbox;

//----------------------------
                              //boundnig sphere
struct I3D_bsphere{
   S_vector pos;
   float radius;
   I3D_bsphere(){}
   I3D_bsphere(const S_vector &p, float r):
      pos(p),
      radius(r)
   {}
};

typedef I3D_bsphere *PI3D_bsphere;
typedef const I3D_bsphere *CPI3D_bsphere;

//----------------------------
                              //bounding volume - consisting of minimal box and minimal sphere
struct I3D_bound_volume{
   I3D_bbox bbox;
   I3D_bsphere bsphere;
};

typedef I3D_bound_volume *PI3D_bound_volume;
typedef const I3D_bound_volume *CPI3D_bound_volume;

//----------------------------
                              //rectangle
struct I3D_rectangle{
   union{
      struct{
         int l, t, r, b;
      };
      struct{
         int m[4];
      };
   };
   I3D_rectangle(){}
   I3D_rectangle(int l1, int t1, int r1, int b1):
      l(l1), t(t1), r(r1), b(b1)
   {}
};

//----------------------------
                              //frame types
enum I3D_FRAME_TYPE{
   FRAME_NULL, FRAME_VISUAL, FRAME_LIGHT, FRAME_CAMERA,
   FRAME_SOUND, FRAME_SECTOR, FRAME_DUMMY, FRAME_TARGET,
   FRAME_USER, FRAME_MODEL, FRAME_JOINT, FRAME_VOLUME,
   FRAME_OCCLUDER,
   FRAME_LAST,
};

#define I3D_VISUAL_OBJECT     0x5f4a424f  //OBJ_
#define I3D_VISUAL_PARTICLE   0x43545250  //PRTC
#define I3D_VISUAL_NURBS      0x4252554e  //NURB
#define I3D_VISUAL_LIT_OBJECT 0x50414d4c  //LMAP
#define I3D_VISUAL_SINGLEMESH 0x4d474e53  //SNGM
#define I3D_VISUAL_BILLBOARD  0x44524242  //BBRD
#define I3D_VISUAL_MIRROR     0x5252494d  //MIRR
#define I3D_VISUAL_MORPH      0x4850524d  //MRPH
#define I3D_VISUAL_DYNAMIC    0x434d4e44  //DNMC
#define I3D_VISUAL_UV_SHIFT   0x5356554f  //OUVS
#define I3D_VISUAL_FLARE      0x45524c46  //FLRE
#define I3D_VISUAL_CAMVIEW    0x564d4143  //CAMV
#define I3D_VISUAL_MIRROR     0x5252494d  //MIRR


struct I3D_note_callback{     //struct used to pass params when note callback encountered
   int curr_time;             //current time
   int note_time;             //time on which the note key is set
   int stage;                 //stage on which the animation is processed
   class I3D_frame *frm;      //frame in which the animation is processed
   const char *note;          //note string
};

                              //frame callback messages
enum FRM_CALLBACK_MESSAGE{
   CBM_ANIM_FINISH   = 2,     //(dword stage)
   CBM_ANIMNOTE      = 3,     //(const I3D_note_callback*)
   CBM_LODCHANGE     = 4,     //(float old_LOD, float new_LOD)
};

                              //frame callback functions
typedef dword I3DAPI I3D_FRMCALLBACK(class I3D_frame*,
   FRM_CALLBACK_MESSAGE msg, dword prm1, dword prm2);
typedef I3D_FRMCALLBACK *PI3D_FRMCALLBACK;


                              //loader callback messages
enum I3D_LOADMESSAGE{
   CBM_PROGRESS      = 0,     //(float progress, dword status); status: 0=beg, 1=in, 2=end
   CBM_ERROR         = 1,     //(const char *msg, int importance (0=max, 1=warning, ...))
   CBM_LOG           = 2,     //(const char *msg)
};

//----------------------------
// Loader - callback function.
// If the returned value for CBM_PROGRESS message is true, the loading is cancelled,
// and the loader method returns I3DERR_CANCELED.
typedef bool I3DAPI I3D_LOAD_CB_PROC(I3D_LOADMESSAGE msg, dword prm1, dword prm2, void *context);
typedef I3D_LOAD_CB_PROC *PI3D_LOAD_CB_PROC;

                              //loading flags - for all Open methods
#define I3DLOAD_LOG              1
#define I3DLOAD_PROGRESS         2  
#define I3DLOAD_ERRORS           4
#define I3DLOAD_SOUND_FREQUENCY  8  //frequency of sound will adjustable
#define I3DLOAD_BUILD_SECTORS    0x10  //model only - build sectors after loading
#define I3DLOAD_SOUND_SOFTWARE   0x20  //play sound in software buffer
#define I3DLOAD_ANIM_COMPATIBILITY  0x40  //retain animations compatibility - treat special scale keys as note and visibility keys

                              //frame callback flags
#define FRMCALLBACK_ANIMNOTE        0x10000000
#define FRMCALLBACK_LODCHANGE       0x20000000
#define FRMCALLBACK_ANIMFINISH      0x40000000
#define FRMCALLBACK_ALL             0x70000000

                              //link flags
#define I3DLINK_UPDMATRIX  1

                              //maximal frame scale, setting greater than this will fail
#define I3D_MAX_FRAME_SCALE 1e+4f
//----------------------------
                              //statistics:

typedef struct{               //collisions
   dword sphere_sphere, sphere_rectangle, sphere_box, sphere_face;
   dword line_sphere, line_rectangle, line_box, line_face;
} I3D_stats_volume, *PI3D_stats_volume;

                              //bsp
typedef struct{
   dword num_nodes;           //number of tree nodes
   dword depth;               //depth of tree
   dword num_planes;          //number of partion planes( <= num_nodes)
   dword num_faces;           //number of faces in tree
   dword face_references;     //number of references to faces ( >= num_faces)
   dword mem_used;            //memory used by tree in bytes.
} I3D_stats_bsp, *PI3D_stats_bsp;

typedef struct{
   dword total_num;
   dword root_num;
   dword moving_static;
   dword reserved2;
} I3D_stats_voltree, *PI3D_stats_voltree;

typedef struct{               //rendering
   dword triangle;            //triangles drawn
   dword line;                //lines drawn
   dword vert_trans;          //vertices transformed
} I3D_stats_render, *PI3D_stats_render;

typedef struct{               //frames
   dword frame_count[FRAME_LAST];   //frames seen/heard
   dword num_vis_computed;    //visuals checked for on-screen
   dword dyn_shd_casters;     //dynamic shadow casters rendered
} I3D_stats_scene, *PI3D_stats_scene;

typedef struct{               //textures
   dword txt_total;           //number of textures
   dword txt_vidmem;          //number of textures loaded into vidmem
   dword txt_sys_mem;         //system memory used by textures
   dword txt_video_mem;       //video memory used by textures
   dword vidmem_total;        //total video memory
   dword vidmem_free;         //free video memory
   dword txtmem_total;        //total texture video memory
   dword txtmem_free;         //free texture video memory
} I3D_stats_video, *PI3D_stats_video;

typedef struct{               //memory
   dword available;           //memory available (by system)
   dword used;                //memory used by I3D
   dword used_other;          //memory used by other resources
   dword num_blocks;          //allocated memory blocks
   dword dbase_size;          //total bytes available in database
   dword dbase_used;          //bytes used in database
} I3D_stats_memory, *PI3D_stats_memory;

typedef struct{               //sounds
   dword buf_hw_all, buf_hw_free;   //hardware sound buffers
   dword buf_3d_all, buf_3d_free;   //hardware 3D sound buffers
   dword hw_mem_all, hw_mem_free;   //sound card memory
   dword snds_all, snds_audible;    //sound frames total, audible
   dword snds_in_card;        //number of sounds uploaded to sound card
   dword mem_used;            //memory used for all sounds
} I3D_stats_sound, *PI3D_stats_sound;

typedef struct{               //I3D_frame
   dword class_mem;           //direct memory allocated for concrete frame (without pointed interfaces)
   dword lmap_mem;
} I3D_stats_frame, *PI3D_stats_frame;

typedef struct{               //I3D_mesh_object, I3D_mesh_level
   dword class_mem;           //direct memory allocated for concrete frame (without pointed interfaces)
   dword vertex_count;        //sum of...
   dword vertex_mem;          //memory occupied by...
   dword face_count;          //sum of...
   dword face_mem;            //memory occupied by...
} I3D_stats_mesh, *PI3D_stats_mesh;

typedef struct{               //I3D_material
   dword class_mem;           //direct memory allocated for class (without pointed interfaces)
} I3D_stats_material, *PI3D_stats_material;

typedef struct{               //I3D_texture
   dword class_mem;           //direct memory allocated for class (without pointed interfaces)
   dword num_texels;
   dword mem_size;
} I3D_stats_texture, *PI3D_stats_texture;


enum I3DSTATSINDEX{           //input index for GetStats() method of various interfaces
            //I3D_driver:
   I3DSTATS_VOLUME,           //PI3D_stats_volume
   I3DSTATS_RENDER,           //I3D_stats_render
   I3DSTATS_SCENE,            //PI3D_stats_scene
   I3DSTATS_VIDEO,            //PI3D_stats_video
   I3DSTATS_MEMORY,           //PI3D_stats_memory
   I3DSTATS_SOUND,            //PI3D_stats_sound
   I3DSTATS_BSP,              //I3D_stats_bsp
   I3DSTATS_VOLTREE,          //I3D_stats_voltree
            //I3D_frame:
   I3DSTATS_FRAME,            //PI3D_stats_frame
            //I3D_mesh_object, I3D_mesh_level:
   I3DSTATS_MESH,             //PI3D_stats_mesh
            //I3D_material:
   I3DSTATS_MATERIAL,         //PI3D_stats_material
            //I3D_texture:
   I3DSTATS_TEXTURE,          //PI3D_stats_texture
            //generic:
   I3DSTATS_CLASS_MEM,        //dword*
};

//----------------------------

enum{
   I3D_FRMF_NOSHADOW = 1,     //don't cast shadows on lightmaps (visuals)
   I3D_FRMF_NOCOLLISION = 2,  //don't detect collisions on this frame (visuals and volumes)
   I3D_FRMF_NO_FLARE_COLLISION = 4, //let flares no detect collisions with this frame (volumes)
   I3D_FRMF_SHADOW_CAST = 8,  //this frame (model) casts shadows
   I3D_FRMF_SHADOW_RECEIVE = 0x10,  //this frame (visual) receives shadows
   I3D_FRMF_STATIC_COLLISION = 0x20,//this frame (visual || volume) is considered static for collisions
};

//----------------------------

enum I3D_ON_MODE{
   I3D_OFF = 0,               //off/disabled/invisible
   I3D_ON = 1,                //on/enabled/visible
   I3D_ON_CONDITIONAL = 2,    //on except in primary camera
};

//----------------------------

class I3D_unknown{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
};

//----------------------------

typedef class I3D_frame *PI3D_frame;
typedef const I3D_frame *CPI3D_frame;

typedef I3DENUMRET I3DAPI I3D_ENUMPROC(PI3D_frame, dword context);

#if !defined I3D_FULL
class I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(I3D_FRAME_TYPE,GetType)() const = 0;
                              //setup
   I3DMETHOD(SetPos)(const S_vector&) = 0;
   I3DMETHOD(SetScale)(float) = 0;             
   I3DMETHOD(SetRot)(const S_quat&) = 0;
   I3DMETHOD(SetDir)(const S_vector&, float roll = 0.0f) = 0;
   I3DMETHOD(SetDir1)(const S_vector &dir, const S_vector &up) = 0;

   I3DMETHOD_(const S_vector&,GetPos)() const = 0;
   I3DMETHOD_(const S_vector&,GetWorldPos)() const = 0;
   I3DMETHOD_(float,GetScale)() const = 0;        
   I3DMETHOD_(float,GetWorldScale)() const = 0;   
   I3DMETHOD_(const S_quat&,GetRot)() const = 0;
   I3DMETHOD_(S_quat,GetWorldRot)() const = 0;
   I3DMETHOD_(void,GetRot1)(S_vector &axis, float &angle) const = 0;
   I3DMETHOD_(const S_vector&,GetDir)() const = 0;
   I3DMETHOD_(const S_vector&,GetWorldDir)() const = 0;

   I3DMETHOD_(const S_matrix&,GetFrameMatrix)() const = 0;
   I3DMETHOD_(const S_matrix&,GetMatrix)() const = 0;
   I3DMETHOD_(const S_matrix&,GetInvMatrix)() const = 0;

   I3DMETHOD_(void,Update)() = 0;

   I3DMETHOD_(void,SetCallback)(PI3D_FRMCALLBACK cb, dword cbf) = 0;
                              //user data
   I3DMETHOD_(void,SetData)(dword) = 0;
   I3DMETHOD_(dword,GetData)() const = 0;

   I3DMETHOD_(void,SetUserFlags)(dword) = 0;
   I3DMETHOD_(dword,GetUserFlags)() const = 0;
                              //visibility control
   I3DMETHOD_(void,SetOn)(dword) = 0;
   I3DMETHOD_(I3D_ON_MODE,IsOn)() const = 0;

   I3DMETHOD_(void,SetName)(const char *cp) = 0;
   I3DMETHOD_(const char*,GetName)() const = 0;
   I3DMETHOD_(const char*,GetOrigName)() const = 0;
   I3DMETHOD_(void,ComputeHRBoundVolume)(PI3D_bound_volume) const = 0;
                              //hierarchy
   I3DMETHOD_(PI3D_frame,GetParent)() const = 0;
   I3DMETHOD(LinkTo)(PI3D_frame, dword flags = 0) = 0;
   I3DMETHOD_(dword,NumChildren)() const = 0;
   I3DMETHOD_(PI3D_frame,GetChild)(int indx) const = 0;

   I3DMETHOD(EnumFrames)(I3D_ENUMPROC*, dword user,
      dword flags = ENUMF_ALL, const char *mask = NULL) const = 0;
   I3DMETHOD_(PI3D_frame,FindChildFrame)(const char *name, dword data=ENUMF_ALL) const = 0;

   I3DMETHOD(Duplicate)(CPI3D_frame) = 0;
   I3DMETHOD(DebugDraw)(PI3D_scene) const;
                              //frame flags
   I3DMETHOD_(dword,SetFlags)(dword new_flags, dword flags_mask = 0xffffffff) = 0;
   I3DMETHOD_(dword,GetFlags)() const = 0;
private:
   I3DMETHOD_(void,FrmReserved00)() = 0;
   I3DMETHOD_(void,FrmReserved01)() = 0;
   I3DMETHOD_(void,FrmReserved02)() = 0;
   I3DMETHOD_(void,FrmReserved03)() = 0;
   I3DMETHOD_(void,FrmReserved04)() = 0;
   I3DMETHOD_(void,FrmReserved05)() = 0;
   I3DMETHOD_(void,FrmReserved06)() = 0;
   I3DMETHOD_(void,FrmReserved07)() = 0;
};

#endif

//----------------------------
typedef class I3D_sector *PI3D_sector;
typedef const I3D_sector *CPI3D_sector;
                              //light
enum I3D_LIGHTTYPE{
   I3DLIGHT_NULL, I3DLIGHT_POINT, I3DLIGHT_SPOT, I3DLIGHT_DIRECTIONAL,
   I3DLIGHT_AMBIENT, I3DLIGHT_FOG, I3DLIGHT_POINTAMBIENT, I3DLIGHT_POINTFOG,
   I3DLIGHT_LAYEREDFOG,
};

                              //flags specifying if this light...
#define I3DLIGHTMODE_VERTEX   1     //...illuminates vertices
#define I3DLIGHTMODE_LIGHTMAP 2     //...affects lightmap
#define I3DLIGHTMODE_SHADOW   4     //...casts shadow
#define I3DLIGHTMODE_DYNAMIC_LM  8  //...cast dynamic circles on lightmaps

#if !defined I3D_FULL 
class I3D_light: public I3D_frame{
public:
   I3DMETHOD(SetLightType)(I3D_LIGHTTYPE) = 0;
   I3DMETHOD(SetColor)(const S_vector&) = 0;
   I3DMETHOD(SetPower)(float) = 0;
   I3DMETHOD(SetMode)(dword) = 0;
   I3DMETHOD(SetRange)(float near1, float far1) = 0;
   I3DMETHOD(SetCone)(float inner, float outer) = 0;
   I3DMETHOD(SetSpecularColor)(const S_vector&) = 0;
   I3DMETHOD(SetSpecularPower)(float) = 0;
                              //sectors
   I3DMETHOD_(dword,NumLightSectors)() const = 0;
   I3DMETHOD_(PI3D_sector const*,GetLightSectors)() const = 0;

   I3DMETHOD_(I3D_LIGHTTYPE,GetLightType)() const = 0;
   I3DMETHOD_(const S_vector&,GetColor)() const = 0;
   I3DMETHOD_(float,GetPower)() const = 0;
   I3DMETHOD_(dword,GetMode)() const = 0;
   I3DMETHOD_(void,GetRange)(float &near1, float &far1) const = 0;
   I3DMETHOD_(void,GetCone)(float &inner, float &outer) const = 0;
   I3DMETHOD_(const S_vector&,GetSpecularColor)() const = 0;
   I3DMETHOD_(float,GetSpecularPower)() const = 0;
};
#endif

typedef class I3D_light *PI3D_light;
typedef const I3D_light *CPI3D_light;

//----------------------------
                              //sound caps
enum I3D_SOUNDTYPE{
   I3DSOUND_NULL, I3DSOUND_POINT, I3DSOUND_SPOT, I3DSOUND_AMBIENT,
   I3DSOUND_VOLUME, I3DSOUND_POINTAMBIENT,
};


#if !defined I3D_FULL 
class I3D_sound: public I3D_frame{
public:
   I3DMETHOD(Open)(const char *fname, dword flags = 0, I3D_LOAD_CB_PROC* = NULL,
      void *context = NULL, PI3D_scene scene_hlp = NULL) = 0;
   I3DMETHOD_(bool,IsPlaying)() const = 0;
                              //setup
   I3DMETHOD(SetSoundType)(I3D_SOUNDTYPE) = 0;
   I3DMETHOD(SetRange)(float min, float max, float ambient_facrot = 0.0f) = 0;
   I3DMETHOD(SetCone)(float in_angle, float out_angle) = 0;
   I3DMETHOD(SetOutVol)(float) = 0;
   I3DMETHOD(SetVolume)(float) = 0;
   I3DMETHOD(SetLoop)(bool) = 0;
   I3DMETHOD(SetStreaming)(bool) = 0;
   I3DMETHOD(SetCurrTime)(dword) = 0;
   I3DMETHOD(SetFrequency)(float) = 0;

   I3DMETHOD_(I3D_SOUNDTYPE,GetSoundType)() const = 0;
   I3DMETHOD_(void,GetRange)(float &min, float &max, float &ambient_facrot) const = 0;
   I3DMETHOD_(void,GetCone)(float &in_angle, float &out_angle) const = 0;
   I3DMETHOD_(float,GetOutVol)() const = 0;
   I3DMETHOD_(float,GetVolume)() const = 0;
   I3DMETHOD_(bool,IsLoop)() const = 0;
   I3DMETHOD_(bool,IsStreaming)() const = 0;
   I3DMETHOD_(dword,GetCurrTime)() const = 0;
   I3DMETHOD_(dword,GetPlayTime)() const = 0;
   I3DMETHOD_(float,GetFrequency)() const = 0;

   I3DMETHOD_(const char*,GetFileName)() const = 0;
   I3DMETHOD_(dword,GetOpenFlags)() const = 0;
   I3DMETHOD_(class ISND_source*,GetSoundSource)() const = 0;
                              //sectors
   I3DMETHOD_(dword,NumSoundSectors)() const = 0;
   I3DMETHOD_(PI3D_sector const*,GetSoundSectors)() const = 0;
};
#endif

typedef class I3D_sound *PI3D_sound;
typedef const I3D_sound *CPI3D_sound;

//----------------------------

#if !defined I3D_FULL 
class I3D_camera: public I3D_frame{
public:
                              //setup:
   I3DMETHOD(SetFOV)(float) = 0;
   I3DMETHOD(SetRange)(float near1, float far1) = 0;
   I3DMETHOD(SetOrthoScale)(float scale) = 0;
   I3DMETHOD(SetOrthogonal)(bool) = 0;

   I3DMETHOD_(float,GetFOV)() const = 0;
   I3DMETHOD_(void,GetRange)(float &near1, float &far1) const = 0;
   I3DMETHOD_(float,GetOrthoScale)() const = 0;
   I3DMETHOD_(bool,GetOrthogonal)() const = 0;

   I3DMETHOD_(const S_matrix&,GetViewMatrix)() const = 0;
   I3DMETHOD_(const S_matrix&,GetProjectionMatrix)() const = 0;
   I3DMETHOD_(PI3D_sector,GetSector)() const = 0;
};
#endif

typedef class I3D_camera *PI3D_camera;
typedef const I3D_camera *CPI3D_camera;

//----------------------------

struct I3D_text_coor{
   float u, v;
   inline I3D_text_coor(){}
   inline I3D_text_coor(float u1, float v1):
      u(u1),
      v(v1)
   {}
   inline float operator[](int i) const{ return (&u)[i]; }
};

typedef I3D_text_coor *PI3D_text_coor;
typedef const I3D_text_coor *CPI3D_text_coor;

//----------------------------

#if !defined I3D_FULL 
class I3D_texture{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(const char*,GetFileName)(int index = 0) const = 0;
   I3DMETHOD_(dword,SizeX)() const = 0;
   I3DMETHOD_(dword,SizeY)() const = 0;
   I3DMETHOD_(const class IImage*,GetSysmemImage)() const = 0;
};
#endif

typedef class I3D_texture *PI3D_texture;
typedef const I3D_texture *CPI3D_texture;

//----------------------------
                              //index of texture within material
enum I3D_MATERIAL_TEXTURE_INDEX{
   MTI_DIFFUSE,
   MTI_ENVIRONMENT,
   MTI_BUMP,
   MTI_LAST
};

//----------------------------
                              //material
#if !defined I3D_FULL 
class I3D_material{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(void,SetName)(const char *cp) = 0;
   I3DMETHOD_(const char*,GetName)() const = 0;

   I3DMETHOD(SetAmbient)(const S_vector&) = 0;
   I3DMETHOD(SetDiffuse)(const S_vector&) = 0;
   I3DMETHOD(SetSpecularColor)(const S_vector&) = 0;
   I3DMETHOD(SetSpecularPower)(float) = 0;
   I3DMETHOD(SetSpecularAngle)(float) = 0;
   I3DMETHOD(SetEmissive)(const S_vector&) = 0;
   I3DMETHOD(SetAlpha)(float) = 0;
   I3DMETHOD(SetTexture)(PI3D_texture) = 0;
   I3DMETHOD_(void,Update)() = 0;
   I3DMETHOD(Duplicate)(const I3D_material*) = 0;

   I3DMETHOD(Set2Sided)() = 0;

   I3DMETHOD_(const S_vector&,GetAmbient)() const = 0;
   I3DMETHOD_(const S_vector&,GetDiffuse)() const = 0;
   I3DMETHOD_(const S_vector&,GetSpecularColor)() const = 0;
   I3DMETHOD_(float,GetSpecularPower)() const = 0;
   I3DMETHOD_(float,GetSpecularAngle)() const = 0;
   I3DMETHOD_(const S_vector&,GetEmmisive)() const = 0;
   I3DMETHOD_(float,GetAlpha)() const = 0;

   I3DMETHOD_(PI3D_texture,GetTexture)(I3D_MATERIAL_TEXTURE_INDEX) const = 0;
   I3DMETHOD_(bool,IsCkey)() const = 0;
};
#endif

typedef class I3D_material *PI3D_material;
typedef const I3D_material *CPI3D_material;

//----------------------------

struct I3D_edge{
   word v[2];
   I3D_edge(){}
   I3D_edge(word v0, word v1){ v[0]=v0; v[1]=v1; }
   inline const word &operator[](int i) const{ return v[i]; }
   inline word &operator[](int i){ return v[i]; }
   inline void Invalidate(){ v[0] = 0xffff; }
   inline bool IsValid() const{ return (v[0]!=0xffff); }
};

//----------------------------

#pragma pack(push,1)
struct I3D_triface{
   word vi[3];
public:
   I3D_triface(){}
   I3D_triface(word i0, word i1, word i2){ vi[0]=i0, vi[1]=i1, vi[2]=i2; }
   inline word &operator [](int i){ return vi[i]; }
   inline const word &operator [](int i) const{ return vi[i]; }
   inline void Invalidate(){ vi[0] = 0xffff; }
   inline bool IsValid() const{ return (vi[0]!=0xffff); }
};
#pragma pack(pop)

typedef I3D_triface *PI3D_triface;
typedef const I3D_triface *CPI3D_triface;

//----------------------------

#if !defined I3D_FULL 
class I3D_face_group{
public:
   C_smart_ptr<I3D_material> mat;
   dword base_index;
   dword num_faces;           //number of faces in this group

   inline PI3D_material GetMaterial() const{ return mat; }
   inline void SetMaterial(PI3D_material m){ mat = m; }
};
#endif

typedef class I3D_face_group *PI3D_face_group;
typedef const I3D_face_group *CPI3D_face_group;

//----------------------------

#if !defined I3D_FULL 
class I3D_visual: public I3D_frame{
public:
   I3DMETHOD_(dword,GetVisualType)() const = 0;

   I3DMETHOD_(const I3D_bound_volume&,GetBoundVolume)() const = 0;
   I3DMETHOD_(const I3D_bound_volume&,GetHRBoundVolume)() const = 0;
   I3DMETHOD_(dword,GetLastRenderTime)() const = 0;
   I3DMETHOD_(class I3D_mesh_base*,GetMesh)() const = 0;
   I3DMETHOD(SetProperty)(dword index, dword param) = 0;
   I3DMETHOD_(dword,GetProperty)(dword index) const = 0;
   I3DMETHOD_(dword,GetCollisionMaterial)() const = 0;
   I3DMETHOD_(void,SetCollisionMaterial)(dword id) = 0;
private:
   I3DMETHOD_(void,VisReserved00)() = 0;
   I3DMETHOD_(void,VisReserved01)() = 0;
   I3DMETHOD_(void,VisReserved02)() = 0;
   I3DMETHOD_(void,VisReserved03)() = 0;
   I3DMETHOD_(void,VisReserved04)() = 0;
   I3DMETHOD_(void,VisReserved05)() = 0;
   I3DMETHOD_(void,VisReserved06)() = 0;
   I3DMETHOD_(void,VisReserved07)() = 0;
   I3DMETHOD_(void,VisReserved08)() = 0;
   I3DMETHOD_(void,VisReserved09)() = 0;
   I3DMETHOD_(void,VisReserved10)() = 0;
   I3DMETHOD_(void,VisReserved11)() = 0;
   I3DMETHOD_(void,VisReserved12)() = 0;
   I3DMETHOD_(void,VisReserved13)() = 0;
   I3DMETHOD_(void,VisReserved14)() = 0;
   I3DMETHOD_(void,VisReserved15)() = 0;
public:
};
#endif

typedef class I3D_visual *PI3D_visual;
typedef const I3D_visual *CPI3D_visual;

//----------------------------

struct I3D_vertex_mesh{
   S_vector xyz;
   S_vector normal;
   I3D_text_coor tex[1];
};

typedef I3D_vertex_mesh *PI3D_vertex_mesh;
typedef const I3D_vertex_mesh *CPI3D_vertex_mesh;

//----------------------------

#if !defined I3D_FULL
class I3D_mesh_base{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

                              //vertices
   I3DMETHOD_(dword,NumVertices)() const = 0;
   I3DMETHOD_(const S_vector*,LockVertices)() = 0;
   I3DMETHOD_(void,UnlockVertices)() = 0;
   I3DMETHOD_(dword,GetSizeOfVertex)() const = 0;
                              //faces and groups
   I3DMETHOD_(dword,NumFGroups)() const = 0;
   I3DMETHOD_(dword,NumFaces)() const = 0;
   I3DMETHOD_(PI3D_face_group,GetFGroup)(int indx) = 0;
   I3DMETHOD_(CPI3D_face_group,GetFGroup)(int indx) const = 0;
   I3DMETHOD_(const I3D_triface*,LockFaces)() = 0;
   I3DMETHOD_(void,UnlockFaces)() = 0;
private:
   I3DMETHOD_(void,VisReserved00)() = 0;
   I3DMETHOD_(void,VisReserved01)() = 0;
   I3DMETHOD_(void,VisReserved02)() = 0;
   I3DMETHOD_(void,VisReserved03)() = 0;
   I3DMETHOD_(void,VisReserved04)() = 0;
   I3DMETHOD_(void,VisReserved05)() = 0;
   I3DMETHOD_(void,VisReserved06)() = 0;
   I3DMETHOD_(void,VisReserved07)() = 0;
};
#endif

typedef class I3D_mesh_base *PI3D_mesh_base;
typedef const I3D_mesh_base *CPI3D_mesh_base;

//----------------------------
                              //lock flags
#define I3DLOCK_XYZ        1        //position modified during lock
#define I3DLOCK_NORMAL     2        //normal modified during lock
#define I3DLOCK_UV         4        //UV coordinates modified during lock
#define I3DLOCK_DIFFUSE    8        //diffuse modified during lock
#define I3DLOCK_SPECULAR   8        //specular modified during lock
#define I3DLOCK_ALPHA      0x10     //alpha modified during lock
#define I3DLOCK_VERTEX     0xff     //all elements of vertex
#define I3DLOCK_INDEX      0x100    //face index modified during lock
#define I3DLOCK_NO_LIGHT_RESET 0x10000 //don't reset lighting info in locked vertices

//----------------------------
                              //vertex component flags
                              // the flags, if present, must appear in vertex
                              // in the order as they're defined here
#define I3DVC_XYZ          2        //S_vector - position
#define I3DVC_XYZRHW       4        //S_vectorw - homogeneous position
#define I3DVC_ALPHA        8        //float - alpha multiplier
#define I3DVC_NORMAL       0x10     //S_vector - normal (normalized)
#define I3DVC_DIFFUSE      0x40     //dword - diffuse in ARGB format
#define I3DVC_SPECULAR     0x80     //dword - specular in ARGB format
#define I3DVC_TEXCOUNT_MASK   0xf00 //mask of bit containing UV set count
#define I3DVC_TEXCOUNT_SHIFT  8     //shift of integer

//----------------------------
                              //one level of mesh detail
#if !defined I3D_FULL
class I3D_mesh_level: public I3D_mesh_base{
public:
   I3DMETHOD(Build2)(const I3D_vertex_mesh *verts, dword num_verts,
      const I3D_triface *faces, dword num_faces,
      const I3D_face_group *fgroups, dword num_fgroups,
      dword flags = 0) = 0;
   I3DMETHOD(Duplicate2)(const I3D_mesh_level*) = 0;
                              //vertices
   I3DMETHOD_(const I3D_vertex_mesh*,LockVertices2)() = 0;
   I3DMETHOD_(void,UnlockVertices2)() = 0;
};
#endif

typedef class I3D_mesh_level *PI3D_mesh_level;
typedef const I3D_mesh_level *CPI3D_mesh_level;

//----------------------------

#if !defined I3D_FULL 
class I3D_mesh_object{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(PI3D_mesh_level,GetLOD)(int indx) const = 0;
   I3DMETHOD(SetLOD)(PI3D_mesh_level, int indx) = 0;
   I3DMETHOD_(dword,NumLevels)() const = 0;
   I3DMETHOD(Duplicate)(const I3D_mesh_object*) = 0;

   I3DMETHOD_(const I3D_bound_volume&,GetBoundVolume)() const = 0;
                              //vertex access
   I3DMETHOD_(dword,NumVertices)(int lod_index = 0) const = 0;
   I3DMETHOD_(PI3D_vertex_mesh,LockVertices)(int lod_index = 0, dword flags = I3DLOCK_VERTEX) = 0;
   I3DMETHOD(UnlockVertices)(int lod_index = 0) = 0;

   I3DMETHOD_(PI3D_triface,LockFaces)(int lod_index = 0, dword flags = I3DLOCK_INDEX) = 0;
   I3DMETHOD(UnlockFaces)() = 0;
};
#endif

typedef class I3D_mesh_object *PI3D_mesh_object;
typedef const I3D_mesh_object *CPI3D_mesh_object;

//----------------------------

#if !defined I3D_FULL 
class I3D_object: public I3D_visual{
public:
   I3DMETHOD(SetMeshObject)(PI3D_mesh_object) = 0;
   I3DMETHOD_(PI3D_mesh_object,GetMeshObject)() const = 0;

   I3DMETHOD_(void,SetUpdate)() = 0;
   I3DMETHOD(SetMesh)(PI3D_mesh_level) = 0;
};
#endif

typedef class I3D_object *PI3D_object;
typedef const I3D_object *CPI3D_object;

//----------------------------

                              //light computation flags:
#define I3D_LIGHT_GEO_COLLISIONS  1     //use geometry collisions
#define I3D_LIGHT_VOL_COLLISIONS  2     //use volume collisions
#define I3D_LIGHT_BLUR            4     //blur destination after make
#define I3D_LIGHT_ALLOW_CANCEL    8     //allow ESC key to cancell light-map creation
#define I3D_LIGHT_MAKEMIPMAPS     0x10  //create mipmaps in LM surface

                              //lit-object mode
#if !defined I3D_FULL
class I3D_lit_object: public I3D_visual{
public:
   I3DMETHOD(Construct)(PI3D_scene, dword flags = 0,
      I3D_LOAD_CB_PROC* = NULL, void *context = NULL) = 0;
   I3DMETHOD(Destruct)(dword flags = 0) = 0;
   I3DMETHOD(Load)(int handle) = 0;
   I3DMETHOD(Save)(int handle) = 0;
   I3DMETHOD_(bool,IsConstructed)() const = 0;
};
#endif

typedef class I3D_lit_object *PI3D_lit_object;
typedef const I3D_lit_object *CPI3D_lit_object;

//----------------------------
                              //dynamic visual
                              // - containing own non-inheritable mesh
                              // - used in cases when frequent updates of geometry
                              //   are necessary
                              // - accept lock flags and updates elements by these
#if !defined I3D_FULL
class I3D_visual_dynamic: public I3D_visual{
public:
   I3DMETHOD_(PI3D_mesh_base,GetMesh)() const = 0;
   I3DMETHOD_(I3D_RESULT,Build)(const void *verts, dword num_verts, dword vc_flags,
      const I3D_triface *faces, dword num_faces,
      const I3D_face_group *fgroups, dword num_fgroups,
      dword flags = 0) = 0;
   I3DMETHOD_(I3D_RESULT,BuildAutoLODs)(float min_dist, float max_dist, dword num_lods, float lowest_quality) = 0;
   I3DMETHOD_(I3D_RESULT,DuplicateMesh)(CPI3D_mesh_base, dword vc_flags) = 0;
   I3DMETHOD_(I3D_RESULT,ResetLightInfo)() = 0;

   I3DMETHOD_(void*,LockVertices)(dword lock_flags) = 0;
   I3DMETHOD_(I3D_RESULT,UnlockVertices)() = 0;

   I3DMETHOD_(PI3D_triface,LockFaces)(dword lock_flags) = 0;
   I3DMETHOD_(I3D_RESULT,UnlockFaces)() = 0;

//----------------------------
// Render - only XYZRHW (transformed) vertex formats.
   I3DMETHOD_(I3D_RESULT,Render)() = 0;

   I3DMETHOD_(dword,NumVertices)() const = 0;
};
#endif

typedef class I3D_visual_dynamic *PI3D_visual_dynamic;
typedef const I3D_visual_dynamic *CPI3D_visual_dynamic;

//----------------------------

#if !defined I3D_FULL 
class I3D_dummy: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(const I3D_bbox*,GetBBox)() const = 0;
   I3DMETHOD(SetBBox)(const I3D_bbox&) = 0;
};
#endif

typedef class I3D_dummy *PI3D_dummy;
typedef const I3D_dummy *CPI3D_dummy;

//----------------------------

#if !defined I3D_FULL
class I3D_joint: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
};
#endif

typedef class I3D_joint *PI3D_joint;
typedef const I3D_joint *CPI3D_joint;

//----------------------------

#if !defined I3D_FULL 
class I3D_user: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
                              //additional user data
   I3DMETHOD_(void,SetData2)(dword) = 0;
   I3DMETHOD_(dword,GetData2)() const = 0;
};
#endif

typedef class I3D_user *PI3D_user;
typedef const I3D_user *CPI3D_user;

//----------------------------

#if !defined I3D_FULL 
class I3D_target: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
                              //frame for which it will be target
   I3DMETHOD_(void,SetFrame)(PI3D_frame) = 0;
   I3DMETHOD_(PI3D_frame,GetFrame)() const = 0;
};
#endif

typedef class I3D_target *PI3D_target;
typedef const I3D_target *CPI3D_target;

//----------------------------

enum I3D_VOLUMETYPE{
   I3DVOLUME_NULL,
   I3DVOLUME_SPHERE, I3DVOLUME_RECTANGLE, I3DVOLUME_BOX,// I3DVOLUME_FACE,
};

#if !defined I3D_FULL 
class I3D_volume: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD(SetNUScale)(const S_vector&) = 0;
   I3DMETHOD_(const S_vector&,GetNUScale)() = 0;

   I3DMETHOD(SetVolumeType)(I3D_VOLUMETYPE) = 0;
   I3DMETHOD(SetOwner)(PI3D_frame) = 0;
   I3DMETHOD(SetScene)(PI3D_scene) = 0;
   I3DMETHOD_(void,SetCollisionMaterial)(dword) = 0;

   I3DMETHOD_(I3D_VOLUMETYPE,GetVolumeType)() const = 0;
   I3DMETHOD_(PI3D_frame,GetOwner)() const = 0;
   I3DMETHOD_(dword,GetCollisionMaterial)() const = 0;
   I3DMETHOD_(void, SetForceDynamic)(bool) = 0;
};
#endif

typedef class I3D_volume *PI3D_volume;
typedef const I3D_volume *CPI3D_volume;

//----------------------------

enum I3D_OCCLUDERTYPE{
   I3DOCCLUDER_MESH, I3DOCCLUDER_SPHERE,
};


#if !defined I3D_FULL 
class I3D_occluder: public I3D_frame{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD(Build)(const S_vector *verts, dword num_verts, dword vertex_pitch = sizeof(S_vector)) = 0;
                              //vertices
   I3DMETHOD_(dword,NumVertices)() const = 0;
   I3DMETHOD_(S_vector*,LockVertices)(dword flags = 0) = 0;
   I3DMETHOD_(void,UnlockVertices)() = 0;

   I3DMETHOD(SetOccluderType)(I3D_OCCLUDERTYPE) = 0;
   I3DMETHOD_(I3D_OCCLUDERTYPE,GetOccluderType)() const = 0;
};
#endif

typedef class I3D_occluder *PI3D_occluder;
typedef const I3D_occluder *CPI3D_occluder;

//----------------------------

typedef class I3D_portal *PI3D_portal;
typedef const I3D_portal *CPI3D_portal;

#if !defined I3D_FULL
class I3D_portal{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(dword,NumVertices)() const = 0;
   I3DMETHOD_(const S_vector*,GetVertices)() const = 0;
   I3DMETHOD_(const S_plane&,GetPlane)() const = 0;
   I3DMETHOD_(PI3D_sector,GetSector)() const = 0;
   I3DMETHOD_(PI3D_sector,GetConnectSector)() const = 0;

   I3DMETHOD_(void,SetOn)(bool) = 0;
   I3DMETHOD_(bool,IsOn)() const = 0;
};
#endif

//----------------------------

#if !defined I3D_FULL 
class I3D_sector: public I3D_frame{
public:
   I3DMETHOD_(bool,CheckPoint)(const S_vector &v) = 0;
   I3DMETHOD_(bool,IsPrimary)() const = 0;
                              //portals
   I3DMETHOD_(dword,NumPortals)() const = 0;
   I3DMETHOD_(const PI3D_portal*,GetPortals)() const = 0;
   I3DMETHOD_(dword,NumConnectPortals)() const = 0;
   I3DMETHOD_(const PI3D_portal*,GetConnectPortals)() const = 0;

                              //lights
   I3DMETHOD(AddLight)(PI3D_light) = 0;
   I3DMETHOD(DeleteLight)(PI3D_light) = 0;
   I3DMETHOD_(dword,NumLights)() const = 0;
   I3DMETHOD_(PI3D_light,GetLight)(int index) const = 0;

                              //sounds
   I3DMETHOD(AddSound)(PI3D_sound) = 0;
   I3DMETHOD(DeleteSound)(PI3D_sound) = 0;
   I3DMETHOD_(dword,NumSounds)() const = 0;
   I3DMETHOD_(const PI3D_sound*,GetSounds)() const = 0;

                              //environment
   I3DMETHOD_(void,SetEnvironmentID)(dword) = 0;
   I3DMETHOD_(dword,GetEnvironmentID)() const = 0;

   I3DMETHOD_(dword,NumBoundPlanes)() const = 0;
   I3DMETHOD_(const S_plane*,GetBoundPlanes)() const = 0;
};
#endif

//----------------------------
                              //animation options and key definition

                              //value applying mode:
                              // SET = replace previous value by new value
                              // ADD = add new value (scaled by scale factor) with previous value
                              // BLEND = blend new value (scaled by scale factor) with previous value using blend factor

                              //position:
#define I3DANIMOP_POS_SET     0x0
#define I3DANIMOP_POS_ADD     0x1
#define I3DANIMOP_POS_BLEND   0x2
#define I3DANIMOP_POS_MASK    0x3
                              //rotation:
#define I3DANIMOP_ROT_SET     0x0
#define I3DANIMOP_ROT_ADD     0x4
#define I3DANIMOP_ROT_BLEND   0x8
#define I3DANIMOP_ROT_MASK    0xc
                              //power (light/volume/scale):
#define I3DANIMOP_POW_SET     0x00
#define I3DANIMOP_POW_ADD     0x10
#define I3DANIMOP_POW_BLEND   0x20
#define I3DANIMOP_POW_MASK    0x30
                              //combination of position/rotation/power:
#define I3DANIMOP_SET         0
#define I3DANIMOP_ADD         0x15
#define I3DANIMOP_BLEND       0x2a
#define I3DANIMOP_APPLY_MASK  0x3f  //mask of apply mode
                              //precedence braces count
#define I3DANIMOP_BRACE_MASK  0x03c0
#define I3DANIMOP_BRACE_SHIFT 6

                              //modifier flags:
#define I3DANIMOP_LOOP 0x10000   //loop animation (if not specified, anim is played only once and stopped)
#define I3DANIMOP_KEEPDONE 0x20000  //do not erase finished animations
#define I3DANIMOP_BLEND_WITH_NEXT 0x40000 //blend current stage with next first, then result with current

                              //animation structs

                              //time unit = 1/1000 sec
                              //tension:    -1 .. 1
                              //continuity: -1 .. 1
                              //bias:       -1 .. 1
                              //easy_from:   0 .. 1
                              //easy_to:     0 .. 1
struct I3D_easiness{
   float easy_from, easy_to;
};
struct I3D_bezier{
   S_vector in_tan, out_tan;
};
struct I3D_tcb{
   float tension, continuity, bias;
};
struct I3D_anim_vector_tcb: public I3D_easiness, public I3D_tcb{
   int time;
   S_vector v;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_vector_tcb)); }
};
typedef I3D_anim_vector_tcb I3D_anim_pos_tcb;
struct I3D_anim_vector_bezier: public I3D_easiness, public I3D_bezier{
   int time;
   S_vector v;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_vector_bezier)); }
};
typedef I3D_anim_vector_bezier I3D_anim_pos_bezier;

struct I3D_anim_look_at_tcb: public I3D_anim_vector_tcb{
   S_vector up_vector;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_look_at_tcb)); }
};

struct I3D_anim_rot: public I3D_easiness{
   int time;
   S_vector axis;
   float angle;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_rot)); }
};

typedef I3D_anim_vector_tcb I3D_anim_scale_tcb;
typedef I3D_anim_vector_bezier I3D_anim_scale_bezier;

struct I3D_anim_note{
   int time;
   const char *note;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_note)); }
};

struct I3D_anim_visibility{
   int time;
   bool on;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_visibility)); }
};

struct I3D_anim_color: public I3D_easiness{
   int time;
   S_vector color;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_color)); }
};

struct I3D_anim_power: public I3D_easiness{
   int time;
   float power;
   inline void Clear(){ memset(this, 0, sizeof(I3D_anim_power)); }
};

//----------------------------

enum I3D_ANIMATION_TYPE{
   I3DANIM_KEYFRAME,          //animation consisting of key-frame tracks
   I3DANIM_POSE,              //static pose
};

//----------------------------

#if !defined I3D_FULL
class I3D_animation_base{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(I3D_ANIMATION_TYPE,GetType)() const = 0;
};
#endif

typedef class I3D_animation_base *PI3D_animation_base;
typedef const I3D_animation_base *CPI3D_animation_base;

//----------------------------

#if !defined I3D_FULL
class I3D_keyframe_anim: public I3D_animation_base{
public:
   I3DMETHOD_(void,Clear)() = 0;
   I3DMETHOD(SetPositionKeys)(I3D_anim_pos_tcb*, dword num_keys) = 0;
   I3DMETHOD(SetRotationKeys)(I3D_anim_rot*, dword num_keys) = 0;
   I3DMETHOD(SetLookAtKeys)(I3D_anim_look_at_tcb*, dword num_keys) = 0;
   I3DMETHOD(SetScaleKeys)(I3D_anim_scale_tcb*, dword num_keys) = 0;
   I3DMETHOD(SetNoteKeys)(I3D_anim_note*, dword num_keys) = 0;
   I3DMETHOD(SetVisibilityKeys)(I3D_anim_visibility*, dword num_keys) = 0;
   I3DMETHOD(SetColorKeys)(I3D_anim_color*, dword num_keys) = 0;
   I3DMETHOD(SetPowerKeys)(I3D_anim_power*, dword num_keys) = 0;
   I3DMETHOD(SetEndTime)(dword) = 0;

   I3DMETHOD(GetPositionKeys)(const I3D_anim_pos_bezier**, dword *num_keys) const = 0;
   I3DMETHOD(GetRotationKeys)(const I3D_anim_rot**, dword *num_keys) const = 0;
   I3DMETHOD(GetLookAtKeys)(const I3D_anim_pos_bezier**, dword *num_keys) const = 0;
   I3DMETHOD(GetScaleKeys)(const I3D_anim_pos_bezier**, dword *num_keys) const = 0;
   I3DMETHOD(GetNoteKeys)(const I3D_anim_note**, dword *num_keys) const = 0;
   I3DMETHOD(GetVisibilityKeys)(const I3D_anim_visibility**, dword *num_keys) const = 0;
   I3DMETHOD(GetColorKeys)(const I3D_anim_color**, dword *num_keys) const = 0;
   I3DMETHOD(GetPowerKeys)(const I3D_anim_power**, dword *num_keys) const = 0;
   I3DMETHOD_(dword,GetEndTime)() const = 0;
};
#endif

typedef class I3D_keyframe_anim *PI3D_keyframe_anim;
typedef const I3D_keyframe_anim *CPI3D_keyframe_anim;

//----------------------------

#ifndef I3D_FULL
class I3D_anim_pose: public I3D_animation_base{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

//----------------------------
// Clear pose - reset all elements to uninitialized.
   I3DMETHOD_(void,Clear)() = 0;

//----------------------------
// Set position of pose. When param is NULL, the position is set to uninitialized.
   I3DMETHOD_(void,SetPos)(const S_vector*) = 0;

//----------------------------
// Set power of pose. When param is NULL, the power is set to uninitialized.
// Power means scale/volume/lightness, dependent on frame type.
   I3DMETHOD_(void,SetPower)(const float*) = 0;             

//----------------------------
// Set rotation of pose. When param is NULL, the rotation is set to uninitialized.
   I3DMETHOD_(void,SetRot)(const S_quat*) = 0;

//----------------------------
// Set look-at direction of pose. When any param is NULL, the direction is set to uninitialized.
   I3DMETHOD_(void,SetDir)(const S_vector *look_dir, const S_vector *up_vector) = 0;

   I3DMETHOD_(const S_vector*,GetPos)() const = 0;
   I3DMETHOD_(const S_quat*,GetRot)() const = 0;
   I3DMETHOD_(const float*,GetPower)() const = 0;
   I3DMETHOD_(void,GetDir)(const S_vector **look_dir, const S_vector **up) = 0;
};
#endif

typedef class I3D_anim_pose *PI3D_anim_pose;
typedef const I3D_anim_pose *CPI3D_anim_pose;

//----------------------------

#if !defined I3D_FULL 
class I3D_animation_set{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD(Open)(const char *fname, dword flags = 0, I3D_LOAD_CB_PROC* = NULL,
      void *context = NULL, PI3D_scene scene_hlp = NULL) = 0;
   I3DMETHOD_(dword,GetTime)() const = 0;

   I3DMETHOD_(void,AddAnimation)(PI3D_animation_base, const char *link) = 0;
   I3DMETHOD(DeleteAnimation)(PI3D_animation_base) = 0;

   I3DMETHOD_(dword,NumAnimations)() const = 0;
   I3DMETHOD_(PI3D_animation_base,GetAnimation)(int index) const = 0;
   I3DMETHOD_(const char*,GetAnimLink)(int index) const = 0;

   I3DMETHOD(Duplicate)(const I3D_animation_set*) = 0;
};
#endif

typedef class I3D_animation_set *PI3D_animation_set;
typedef const I3D_animation_set *CPI3D_animation_set;

//----------------------------

#if !defined I3D_FULL
class I3D_interpolator{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

//----------------------------
// Update interpolator by specified time - animate frame.
// When interpolator is finished with animation of all stages, the returned value is I3D_DONE.
   I3DMETHOD_(I3D_RESULT,Tick)(int time) = 0;

//----------------------------
// Set frame on which this interpolator operates (i.e. frame to which it applies results of animation).
   I3DMETHOD_(I3D_RESULT,SetFrame)(PI3D_frame) = 0;

//----------------------------
// Setup animation for particular stage.
   I3DMETHOD_(I3D_RESULT,SetAnimation)(dword stage, PI3D_animation_base,
      dword options = I3DANIMOP_SET, float scale_factor = 1.0f,
      float blend_factor = 1.0f, float speed = 1.0f) = 0;

//----------------------------
// Set scale factor of particular stage.
   I3DMETHOD_(I3D_RESULT,SetScaleFactor)(dword stage, float factor) = 0;

//----------------------------
// Set scale factor of particular stage.
   I3DMETHOD_(I3D_RESULT,SetBlendFactor)(dword stage, float factor) = 0;

//----------------------------
// Set speed multiplier of particular stage.
   I3DMETHOD_(I3D_RESULT,SetSpeed)(dword stage, float speed) = 0;

   //I3DMETHOD_(I3D_RESULT,SetPose)(PI3D_pose) = 0;
//----------------------------
// Set end time of interpolator, in specified stage.
   I3DMETHOD_(I3D_RESULT,SetEndTime)(dword stage, dword time) = 0;

//----------------------------
// Set current time of interpolator, in specified stage.
   I3DMETHOD_(I3D_RESULT,SetCurrTime)(dword stage, dword time) = 0;

//----------------------------
// Get frame on which this interpolator operates.
   I3DMETHOD_(PI3D_frame,GetFrame)() const = 0;

//----------------------------
// Get animation set on specified stage.
   I3DMETHOD_(PI3D_animation_base,GetAnimation)(dword stage) const = 0;

//----------------------------
// Get animation options associated with specified stage.
   I3DMETHOD_(dword,GetOptions)(dword stage) const = 0;

//----------------------------
   I3DMETHOD_(float,GetScaleFactor)(dword stage) const = 0;

//----------------------------
   I3DMETHOD_(float,GetBlendFactor)(dword stage) const = 0;

//----------------------------
   I3DMETHOD_(float,GetSpeed)(dword stage) const = 0;

//----------------------------
   I3DMETHOD_(dword,GetEndTime)(dword stage) const = 0;

//----------------------------
   I3DMETHOD_(dword,GetCurrTime)(dword stage) const = 0;
};
#endif

typedef class I3D_interpolator *PI3D_interpolator;
typedef const I3D_interpolator *CPI3D_interpolator;

//----------------------------
                              //particle parameters
enum{
   PRT_OUT_TIME,              //int 
   PRT_OUTT_THRESH,           //int
   PRT_LIFE_LEN,              //int
   PRT_LIFEL_THRESH,          //int
   PRT_SCALE_GROW_MIN,        //float
   PRT_SCALE_GROW_MAX,        //float
   PRT_INIT_POS_THRESH,       //float
   PRT_INIT_DIR_THRESH,       //float
   PRT_AFFECT_TIME,           //int
   PRT_AFFECTT_THRESH,        //int
   PRT_AFFECT_DIR_THRESH,     //float
   PRT_OPACITY_IN,            //float
   PRT_OPACITY_OUT,           //float
   PRT_BASE_DIR_X, PRT_BASE_DIR_Y, PRT_BASE_DIR_Z, //float
   PRT_DEST_DIR_X, PRT_DEST_DIR_Y, PRT_DEST_DIR_Z, //float
   PRT_DRAW_MODE,             //int
   PRT_SETCOUNT,              //int
   PRT_BASE_DIR_WORLD,        //bool
   PRT_DEST_DIR_WORLD,        //bool
   PRT_EMIT_BASE,             //int
   PRT_ROT_SPEED_MIN,         //float
   PRT_ROT_SPEED_MAX,         //float
   PRT_ELEMENT_SCALE,         //float
   PRT_OPACITY,               //float
   PRT_MODE,                  //0=billb, 1=XY, 2=XZ, 3=YZ
   PRT_OPTIMIZED,
   PRT_ASPECT_RATIO,          //float
   PRT_RESERVED2,
                              //read-only
   PRT_NUMELEMENTS,           //int
};

//----------------------------

#if !defined I3D_FULL
class I3D_particle: public I3D_visual{
public:
   I3DMETHOD(SetMaterial)(PI3D_material) = 0;
   I3DMETHOD_(PI3D_material,GetMaterial)() const = 0;
};
#endif

typedef class I3D_particle *PI3D_particle;
typedef const I3D_particle *CPI3D_particle;

//----------------------------
                              //container - base class keeping references to frames and interpolators,
                              // allows for loading scene files
                              // used as base for I3D_scene and I3D_model classes

#define I3D_CONTAINER(END_LINE) \
/* Open scene hierarchy file from disk. If 'scene_hlp' is set, it's used for*/ \
/* volume setting and sector creation */ \
   I3DMETHOD_(I3D_RESULT,Open)(const char* fname, dword flags = 0, I3D_LOAD_CB_PROC* = NULL, \
      void *context = NULL, PI3D_scene scene_hlp = NULL) END_LINE \
/* Close - erase contents (frames, interpolators, etc).*/ \
   I3DMETHOD_(void,Close)() END_LINE \
/* Get name which was last used to Open method, regardless of success.*/ \
   I3DMETHOD_(const char*,GetFileName)() const END_LINE \
/* Store filename. */ \
   I3DMETHOD_(void,StoreFileName)(const char*) END_LINE \
/* Add/delete frame from interal list.*/ \
   I3DMETHOD_(I3D_RESULT,AddFrame)(PI3D_frame frm) END_LINE \
   I3DMETHOD_(I3D_RESULT,DeleteFrame)(PI3D_frame frm) END_LINE \
/* Add frame into internal list - it is kept there until not Close-d or Release-d.*/ \
   I3DMETHOD_(dword,NumFrames)() const END_LINE \
/* Get access to internal frame list - number and pointer to frames.*/ \
   I3DMETHOD_(const PI3D_frame*,GetFrames)() const END_LINE \
/* Get access to internal list of interpolators - add, delete, number and pointer to them.*/ \
   I3DMETHOD_(I3D_RESULT,AddInterpolator)(PI3D_interpolator) END_LINE \
   I3DMETHOD_(I3D_RESULT,DeleteInterpolator)(PI3D_interpolator) END_LINE \
   I3DMETHOD_(dword,NumInterpolators)() const END_LINE \
   I3DMETHOD_(const PI3D_interpolator*,GetInterpolators)() const END_LINE \
/* Update all interpolators.*/ \
   I3DMETHOD_(I3D_RESULT,Tick)(int time) END_LINE \

#if !defined I3D_FULL
class I3D_container{
public:
   I3D_CONTAINER( = 0;)
};
#endif

typedef class I3D_container *PI3D_container;
typedef const I3D_container *CPI3D_container;

//----------------------------

#if !defined I3D_FULL 
class I3D_model: public I3D_frame{
public:
//----------------------------
// Get minimal bounding volume enclosing all visual children of model.
// This may be invalidated bounding volume if there's no geometry in the hierarchy.
   I3DMETHOD_(const I3D_bound_volume&,GetHRBoundVolume)() const = 0;

//----------------------------
// Set animation-set on children of model. This function resolves links stored in
// animset, sets up interpolators processing the animation in specified mode.
   I3DMETHOD_(I3D_RESULT,SetAnimation)(dword stage, PI3D_animation_set,
      dword options = I3DANIMOP_SET, float scale_factor = 1.0f,
      float blend_factor = 1.0f, float speed = 1.0f) = 0;

//----------------------------
// Set scale factor of all interpolators working with specified stage.
   I3DMETHOD_(I3D_RESULT,SetAnimScaleFactor)(dword stage, float scale_factor) = 0;

//----------------------------
// Set blend factor of all interpolators working with specified stage.
   I3DMETHOD_(I3D_RESULT,SetAnimBlendFactor)(dword stage, float blend_factor) = 0;

//----------------------------
// Set speed multiplier of all interpolators working with specified stage.
   I3DMETHOD_(I3D_RESULT,SetAnimSpeed)(dword stage, float speed) = 0;

//----------------------------
// Stop animation playback on specified stage.
   I3DMETHOD_(I3D_RESULT,StopAnimation)(dword stage) = 0;

//----------------------------
// Get animation set loaded from file when I3D_model::Open was last called.
   I3DMETHOD_(PI3D_animation_set,GetAnimationSet)() const = 0;

//----------------------------
// The rest is container.
   I3D_CONTAINER( = 0;)
   I3DMETHOD_(PI3D_container,GetContainer)() = 0;
};
#endif

typedef class I3D_model *PI3D_model;
typedef const I3D_model *CPI3D_model;

//----------------------------
                              //create texture flags - used with I3D_CREATETEXTURE structure
#define TEXTMAP_DIFFUSE          1        //text_name valid, specifying diffuse bitmap
#define TEXTMAP_OPACITY          2        //text_name valid, specifying opacity bitmap
#define TEXTMAP_BUMPMAP          4        //text_name valid, specifying bump bitmap
#define TEXTMAP_TRANSP           0x40     //texture with transparent (color-keyed) texels
#define TEXTMAP_INVALPHA         0x80     //invert alpha channel during loading
#define TEXTMAP_MIPMAP           0x100    //generate mipmap levels
#define TEXTMAP_NOMIPMAP         0x200    //disable auto-mipmap feature
#define TEXTMAP_COMPRESS         0x800    //create compressed texture (if supported by hw)
#define TEXTMAP_USEPIXELFORMAT   0x1000   //use specified pixel format
#define TEXTMAP_HINTDYNAMIC      0x2000   //dymanic texture
#define TEXTMAP_TRUECOLOR        0x4000   //choose true-color pixel format, if available
#define TEXTMAP_MEM_PTR          0x8000   //diffuse and/or opacity maps specified by memory pointers, rather than by filenames
#define TEXTMAP_RENDERTARGET     0x10000  //request texture which may be used as rendertarget

                              //find texture format flags (hint only)
#define FINDTF_PALETIZED   1        //paletized textures
#define FINDTF_ALPHA       2        
#define FINDTF_ALPHA1      4
#define FINDTF_COMPRESSED  8
#define FINDTF_TRUECOLOR   0x10     //true-color
#define FINDTF_BUMPMAP     0x20     //only bump-map formats
#define FINDTF_RENDERTARGET 0x40    //render target surfaces
                              //create texture struct
struct I3D_CREATETEXTURE{
   dword flags;               //TEXTMAP_ flags
   union{                     //diffuse bitmap - valid with TEXTMAP_DIFFUSE flag
      const char *text_name;  //file name, used with TEXTMAP_DIFFUSE flag
      const char *bump_name;  //file name, used with TEXTMAP_BUMPMAP flag
      const void *text_mem;   //pointer to data in memory (BMP format)
      const void *proc_name;  //name of procedure, for procedural texture
   };
   union{                     //opacity bitmap - valid with TEXTMAP_OPACITY flag
      const char *opt_name;   //file name
      const void *opt_mem;    //pointer to data in memory (BMP format)
   };
   dword size_x, size_y;
   struct S_pixelformat *pf;  //requested pixel format, used with TEXTMAP_USEPIXELFORMAT
};

enum SCN_CALLBACK_MESSAGE{
   CBM_VOLTREEBUILD  = 2,     //
   CBM_SECTORSWITCH  = 3,     //old_sector, new_sector
   CBM_MESSAGE       = 4,     //const char*, bool still
   CBM_DEBUG_MESSAGE = 5,     //(const char*, int type) - debug message, types: 0=one-time, 1=console, 2=run-time
   CBM_DEBUG_CODE    = 6,     //dword code, void *data
   CBM_ANIM_NOTE     = 7,     //const I3D_note_callback*
};

                              //driver callback message is the same
#define DRV_CALLBACK_MESSAGE SCN_CALLBACK_MESSAGE

                              //scene callback function
typedef dword I3DAPI I3D_SCNCALLBACK(SCN_CALLBACK_MESSAGE msg, dword prm1, dword prm2, void *context);
typedef I3D_SCNCALLBACK *PI3D_SCNCALLBACK;
                              //scene callback flags
#define SCNCALLBACK_BEFORERENDER 1
#define SCNCALLBACK_AFTERRENDER  2
#define SCNCALLBACK_SECTSWITCH   4

//----------------------------

                              //collision test flags
#define I3DCOL_EDIT_FRAMES       2           //'pick' lights and sounds in way
#define I3DCOL_TEXTURE           0x20000     //check transparent texels
#define I3DCOL_RAY               0x40000     //test ray instead of line segment
#define I3DCOL_INVISIBLE         0x100000    //test also invisible frames
#define I3DCOL_FORCE2SIDE        0x200000    //always consider faces to be 2-sided
#define I3DCOL_SPHERE            0x400000    //test sphere move along line
#define I3DCOL_SOLVE_SLIDE       0x800000    //slide object along collision plane
#define I3DCOL_EXACT             0x1000000   //no aproximation, ignore volumes
#define I3DCOL_STATIC_SPHERE     0x2000000   //test sphere against the bsp tree
#define I3DCOL_SPHERE_GROUP      0x4000000   //test move of all owner sphere volumes
#define I3DCOL_EDIT_FRMS_ONLY    0x8000000   //test only edit frames, nothing more
#define I3DCOL_EDIT_NONPICKABLE  0x10000000  //test also on nonpickable visuals (valid only when I3DCOL_EDIT_FRAMES is also set)
#define I3DCOL_JOINT_VOLUMES     0x20000000  //test also joint volumes on singlemash. valid only for line tests.

                              //object picking flags(for old style col compatability)
#define I3DGEOCOL_FACE        0           //default
#define I3DGEOCOL_EDIT_FRAMES I3DCOL_EDIT_FRAMES           //'pick' lights and sounds in way
#define I3DGEOCOL_TEXTURE     I3DCOL_TEXTURE      //check transparent texels
#define I3DGEOCOL_RAY         I3DCOL_RAY      //test ray instead of line segment
#define I3DGEOCOL_INVISIBLE   I3DCOL_INVISIBLE    //test also invisible frames
#define I3DGEOCOL_FORCE2SIDE  I3DCOL_FORCE2SIDE    //always consider faces to be 2-sided
#define I3DGEOCOL_NO_BILLBOARD 0 //not supported yet


                              //collision response function
typedef void I3DAPI I3D_collision_response_old(const S_vector &limit_normal,
   float limit_d, const S_vector &from, S_vector &to, float d_to,
   PI3D_volume vol_to, PI3D_volume vol_from);
typedef I3D_collision_response_old *PI3D_collision_response_old;

typedef bool I3DAPI I3D_line_collision_response(const S_vector &collision_normal,
   float collision_dist, PI3D_frame, void *context, void *special);
typedef I3D_line_collision_response *PI3D_line_collision_response;


#define I3DCHECKCOL_NORESPONSE   1  //don't perform response on collision, only check if any collision occured

typedef bool I3DAPI I3D_BSP_BUILD_CB_PROC(dword prm1, dword prm2, void *context);

typedef bool I3DAPI I3D_collision_response(struct I3D_collision_data&);
typedef I3D_collision_response *PI3D_collision_response;

                              //note: currently invisible frames are skipped only if are set off,
                              //parent  doesn't enough for it. this is for speed up RayTest.
                              //note: when testing SPHERE_GROUP, each face if tested for entire group only ones
                              //thus if you reject volume which collide with sphere1, you cannot expect to
                              //response it for volume_source sphere2 in same test.
struct I3D_collision_data{
   void *reserved;            //for internal purposes - do not modify!
                              //in:
   S_vector from;
   S_vector dir;
   dword flags;
   PI3D_line_collision_response cresp_line_old, cresp_texel_old; //tmp: old response support
   PI3D_collision_response cresp_face, cresp_texel;
   void *cresp_context;
   PI3D_frame frm_ignored;    //this frame and it's children are skipped. used only with dymanic tree tests.
   float radius;              //used only with sphere/cylinder tests
   PI3D_frame frm_root;       //root frame will be enumerated for all sphere volumes and those will be tested.
                              //used only with sphere group

                              //out:
private:
   float closest_hit;
   PI3D_frame closest_frm;
   S_vector hit_normal;
   S_vector dest_pos;         //valid only for moving objects
                              //response:
   bool continue_test;        //flag returned by response which say if re-test new direction
   S_vector primary_velocity; //original solved velocity(before any modify by clip plane)
   S_vector orig_dir;         //original solved velocity(before any modify by clip plane)
   PI3D_volume vol_source;    //which volume is source of collison (valid only for group response)
   S_vector face_normal;      //inormal of impact face
                              

   friend class I3D_scene;
   friend class C_bsp_tree;
   friend class I3D_volume;
   friend inline bool FastTest_L_S(struct I3D_collision_data &cd, const S_vector &sphere_pos, float &radius, float radius_2);

                              //temporary help
private:
   S_vector normalized_dir;   //valid only in test collision functions
   float dir_magnitude;       //valid only in test collision functions
   inline bool PrepareTest(){
      dest_pos = from + dir;
      dir_magnitude = dir.Magnitude();
      if(IsMrgZero(dir_magnitude))
         return false;
      normalized_dir = dir / dir_magnitude;
      return true;
   }

public:
   I3D_collision_data(){
      Reset();
   }
                              //common constructor
   I3D_collision_data(const S_vector &from1, const S_vector &dir1, dword flags1, PI3D_frame frm_ignored1 = NULL,
      PI3D_collision_response cresp_face1 = NULL, void *c1 = NULL) : 
   from(from1), dir(dir1), flags(flags1), frm_ignored(frm_ignored1), cresp_face(cresp_face1), cresp_context(c1)
   {
      cresp_texel = NULL;
      cresp_line_old = NULL, cresp_texel_old = NULL; 
      closest_hit = 1e+16f;
      closest_frm = NULL;
      vol_source = NULL;
      frm_root = NULL;
      reserved = NULL;
   }
   void Reset(){
      flags = 0;
      cresp_line_old = NULL, cresp_texel_old = NULL; 
      cresp_face = NULL,  cresp_texel = NULL;
      cresp_context = NULL;
      frm_ignored = NULL;
      continue_test = true;
      reserved = NULL;
      frm_root = NULL;

      closest_hit = 1e+16f;
      closest_frm = NULL;
      vol_source = NULL;
   }
                              //return hit distance. valid only for line tests.
   inline const float &GetClosestHit() const { return closest_hit; }
                              //return calculated position of hit. valid only for line tests.
   inline S_vector GetHitPosition() const { return from + dir*closest_hit / dir.Magnitude(); }
                              //return closest frame which was hit. 
   inline PI3D_frame GetHitFrm() const { return closest_frm; }
                              //return normal of impact plane.
   inline S_vector GetHitNormal() const { return hit_normal; }
                              //return solved position after response(default sliding). valid only for sphere/group.
   inline S_vector GetDestinationPos() const { return dest_pos; }
                              //change destination position. used in response functions.
   inline void SetDestination(const S_vector &dp) { dest_pos = dp; }
                              //set common return values. used internaly by collision functions.
   void SetReturnValues(float h_dist, PI3D_frame h_frm, const S_vector *h_norm = NULL, const S_vector *h_pos = NULL);
                              //change hit normal. used in response functions and only for spheres/groups.
   inline void SetHitNormal(const S_vector &n) { hit_normal = n; }
                              //sphere/group tests after response check if new destination is reachable,
                              //and response any next collisions. set to false in respose to avoid next tests.
                              //used only with sphere/group tests in response.
   inline void SetContinueTest(bool b) { continue_test = b; }
                              //this is primary input test vector from call of test collision. 
                              //used only with spheres/groups in response.
   inline const S_vector &GetPrimaryVelocity() const { return primary_velocity; }
                              //this is curently tested dir before modify in response.
                              //in difference to primary dir, this can multiple change,
                              //since some fragments was covered sucessfully.
                              //used only with spheres/groups in response.
   inline const S_vector &GetOrigDir() const { return orig_dir; }
                              //return sphere volume which collided. valid only in sphere/group tests.
   inline PI3D_volume GetVolumeSource() const { return vol_source; }
                              //return normal of impact face
   inline const S_vector &GetFaceNormal() const{ return face_normal; }
};

//----------------------------
                              //rendering flags
#define I3DRENDER_UPDATELISTENER    1     //update 3D sound listener during this rendering

//----------------------------
                              //bsp load flags:
                              //perform checksum, delete tree if failed. use only in editor,
                              //don't use in final game where BSP MUST be valid as is.
#define FBSPLOAD_CHECKSUM     1
                              //check consistecy of tree - report which frame are marked static but not in tree.
                              //don't use in final game, this slow down loading.
#define FBSPLOAD_CONSISTENCE  2

//----------------------------

#if !defined I3D_FULL

class I3D_scene{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
                              //general
   I3DMETHOD_(PI3D_driver,GetDriver)() const = 0;
   I3DMETHOD_(PI3D_SCNCALLBACK,SetCallback)(PI3D_SCNCALLBACK cb, dword cbf) = 0;
                              //rendering
   I3DMETHOD(Render)(dword flags = 0, void *context = NULL) = 0;
   I3DMETHOD(SetViewport)(dword l, dword t, dword r, dword b, bool aspect=false) = 0;
   I3DMETHOD_(void,GetViewport)(dword &l, dword &t, dword &r, dword &b) = 0;
                              //frames
   I3DMETHOD(EnumFrames)(I3D_ENUMPROC*, dword user,
      dword flags = ENUMF_ALL, const char *mask = NULL) const = 0;
   I3DMETHOD_(PI3D_frame,FindFrame)(const char *name, dword data=ENUMF_ALL) const = 0;
                              //lights
   I3DMETHOD(SetGlobalColor)(const S_vector&) = 0;
   I3DMETHOD_(S_vector,GetLightness)(const S_vector &pos, const S_vector *norm = NULL, dword light_mode_flags = I3DLIGHTMODE_VERTEX) const = 0;
                              //cameras
   I3DMETHOD(SetActiveCamera)(PI3D_camera) = 0;
   I3DMETHOD_(PI3D_camera,GetActiveCamera)() const = 0;
   I3DMETHOD(UnmapScreenPoint)(int x, int y, S_vector &pos, S_vector &dir) const = 0;
   I3DMETHOD(TransformPoints)(const S_vector *in, S_vectorw *out, dword num) const = 0;
                              //sectors
   I3DMETHOD_(PI3D_sector,CreateSector)() = 0;
   I3DMETHOD_(PI3D_sound,CreateSound)() = 0;
   I3DMETHOD_(PI3D_sector,GetSector)(const S_vector&, PI3D_sector suggest_sector = NULL) const = 0;
   I3DMETHOD_(PI3D_sector,GetPrimarySector)() const = 0;
   I3DMETHOD_(PI3D_sector,GetBackdropSector)() const = 0;
   I3DMETHOD_(void,SetBackdropRange)(float near1, float far1) = 0;
   I3DMETHOD_(void,GetBackdropRange)(float &near1, float &far1) const = 0;
   I3DMETHOD(SetFrameSector)(PI3D_frame) = 0;
   I3DMETHOD(SetFrameSectorPos)(PI3D_frame, const S_vector &pos) = 0;
                              //background
   I3DMETHOD(SetBgndColor)(const S_vector&) = 0;
   I3DMETHOD_(const S_vector&,GetBgndColor)() const = 0;

                              //collisions
   I3DMETHOD_(void, ResetDynamicVolTreeEx)() = 0;
   I3DMETHOD_(bool, TestCollision)(struct I3D_collision_data&) const = 0;

   I3DMETHOD_(void,AttenuateSounds)() = 0;
   I3DMETHOD_(dword,GetLastRenderTime)() const = 0;

   I3DMETHOD(SetRenderMatrix)(const S_matrix&) = 0;
   I3DMETHOD(DrawLine)(const S_vector &v1, const S_vector &v2,
      const S_vector &color, dword flags = 0) = 0;
   I3DMETHOD(DrawLines)(const S_vector *v, dword numv, const word *indx,
      dword numi, const S_vector &color, dword v_stride=sizeof(S_vector),
      dword flags=0) = 0;
   I3DMETHOD(DrawTriangles)(const void *v, dword numv, const word *indx,
      dword numi, const S_vector &color, dword v_stride = sizeof(S_vector),
      dword flags = 0, const I3D_text_coor* = NULL) = 0;
   I3DMETHOD(DrawSprite)(const S_vector &pos, PI3D_material, byte inv_alpha = 0, float size = 1.0f) = 0;

   I3DMETHOD(CreateBspTree)(I3D_BSP_BUILD_CB_PROC *cbP = NULL, void *context = NULL) = 0;
   I3DMETHOD_(void, DestroyBspTree)() = 0;

   I3DMETHOD_(bool, SaveBsp)(C_cache *cp) = 0;
   I3DMETHOD_(bool, LoadBsp)(C_cache *cp, I3D_LOAD_CB_PROC* = NULL, void *context = NULL, dword check_flags = 0) = 0;
   I3DMETHOD_(bool, IsBspBuild)() = 0;

//----------------------------
// The rest is container.
   I3D_CONTAINER( = 0;)
   I3DMETHOD_(PI3D_container,GetContainer)() = 0;
};
#endif

//----------------------------
                              //init driver struct
struct I3DINIT{
   class IGraph *lp_graph;
   dword flags;
   void *guid;
   class ISND_driver *lp_sound;
   typedef void __stdcall t_log_func(const char*);
   t_log_func *log_func;
};
                              //init driver flags
#define I3DINIT_VERBOSE       2
#define I3DINIT_DIRECT_TRANS  4     //direct transformation, no ProcessVertices calls
#define I3DINIT_WBUFFER       8     //use w-buffer instead of z-buffer
#define I3DINIT_WARN_OUT      0x10  //display warnings about unreleased interfaces

typedef I3DINIT *PI3DINIT;
typedef const I3DINIT *CPI3DINIT;

//----------------------------
                              //driver rendering state
enum I3D_RENDERSTATE{
   RS_LINEARFILTER = 0,       //linear texture filtering
   RS_MIPMAP = 1,             //mipmap texture filtering
   RS_BUMPMAPPING = 2,        //use bump mapping
   RS_USEZB = 3,              //use z-buffer during rendering
   RS_CLEAR = 4,              //clear screen/zb before rendering
   RS_DITHER = 5,             //dither primitives during rendering
   RS_POINTFOG = 6,           //allow point-fog computation
   RS_WIREFRAME = 7,          //use wire-rendering
   RS_DRAWBOUNDBOX = 8,       //debug draw object bound boxes
   RS_DRAWHRBOUNDBOX = 9,     //debug draw hierarchy bound boxes
   RS_DRAWLINKS = 10,         //debug draw hierarchy links
   RS_LOADMIPMAP = 11,        //build mipmap levels during texture loading
   RS_DRAWMIRRORS = 12,       //render mirrors
   RS_USESHADOWS = 13,        //render shadows
   RS_DEBUGDRAWSHADOWS = 14,  //debug draw shadow textures and volumes
   RS_DRAWSECTORS = 15,       //debug draw sectors
   RS_DRAWPORTALS = 16,       //debug draw portals
   RS_TEXTUREDITHER = 17,     //texture dither during loading
   RS_DRAWVISUALS = 18,
   RS_FOG = 19,               //set fog on/off
   RS_LOD_QUALITY = 20,       //float number
   RS_DEBUGDRAWSHDRECS = 21,  //draw shadow receivers
   RS_TEXTURELOWDETAIL = 22,  //scale all loaded textures to 1/4 of size
   RS_DRAWVOLUMES = 23,       //debug draw collision volumes
   RS_DRAWLIGHTS = 24,        //debug draw lights ranges and cones
   RS_DEBUGMATVISUALIZE = 25, //material of which geometry and visuals are drawn, -1 to switch off
   RS_DRAWSOUNDS = 26,        //draw sound ranges and cones
   RS_TEXTURECOMPRESS = 27,   //use texture compression if available
   RS_USELMAPPING = 28,       //use light-mapping
   RS_DRAWLMTEXTURES = 29,    //debug draw light-map textures
   RS_DRAWCAMERAS = 30,       //debug draw cameras
   RS_DRAWDUMMYS = 31,        //debug draw dummy frames
   RS_DRAWTARGETS = 32,       //debug draw targets
   RS_DRAWOCCLUDERS = 33,     //debug draw occluders
   RS_DRAWTEXTURES = 34,      //draw textures
   RS_LMTRUECOLOR = 35,       //use true-color lightmaps
   RS_LMDITHER = 36,          //dither highcolor textures
   RS_LOD_SCALE = 37,         //float scale (1.0f = default, lower=better LOD, higher=worse LOD)
   RS_DEBUGDRAWBSP = 38,      //draw bsp tree polygons
   RS_DRAWJOINTS = 39,        //debug draw joints
   RS_DEBUGDRAWDYNAMIC = 40,  //draw dynamic volume tree hiearchy
   RS_DEBUGDRAWSTATIC = 41,   //draw static frames
   RS_ANISO_FILTERING = 42,   //enable anisotropic filtering



   RS_LOD_INDEX = 46,         //force automatic LOD, if -1, use default (computed)
   RS_USE_OCCLUSION = 47,     //enable occlusion testing (default)
   RS_LM_AA_RATIO = 48,       //LightMaps creation - antialias subpixel division (1 = no, 2 = 4x, 3 = 9x, ect)
   RS_ENVMAPPING = 49,        //use environment mapping
   RS_DEBUG_INT0 = 50,        //debugging purposes
   RS_DEBUG_INT1 = 51,
   RS_DEBUG_INT2 = 52,
   RS_DEBUG_INT3 = 53,
   RS_DEBUG_INT4 = 54,
   RS_DEBUG_INT5 = 55,
   RS_DEBUG_INT6 = 56,
   RS_DEBUG_INT7 = 57,
   RS_DEBUG_FLOAT0 = 58,      ///debugging purposes
   RS_DEBUG_FLOAT1 = 59,
   RS_DEBUG_FLOAT2 = 60,
   RS_DEBUG_FLOAT3 = 61,
   RS_DEBUG_FLOAT4 = 62,
   RS_DEBUG_FLOAT5 = 63,
   RS_DEBUG_FLOAT6 = 64,
   RS_DEBUG_FLOAT7 = 65,
};

//----------------------------
                              //blend modes for I3D_driver::SetupBlend
enum{
   I3DBLEND_OPAQUE      = 0,  //ONE:ZERO
   I3DBLEND_ALPHABLEND  = 1,  //SRCALPHA:INVSRCALPHA
   I3DBLEND_ADD         = 2,  //ONE:ONE
   I3DBLEND_ADDALPHA    = 3,  //SRCALPHA:ONE
   I3DBLEND_BLEND       = 4,  //ZERO:SRCCOLOR
   I3DBLEND_INVBLEND    = 5,  //ZERO:INVSRCCOLOR
                              //modifiers:
   I3DBLEND_ALPHAFACTOR = 0x20000,
   I3DBLEND_VERTEXALPHA = 0x40000,
};

//----------------------------

#define I3DDRAW_TRANPARENCY_MASK 0xff
#define I3DDRAW_TRANSFORMED   0x100    //points are already transformed to screen space
#define I3DDRAW_NOCULL        0x200
#define I3DDRAW_INVCULL       0x400

//----------------------------

enum I3D_SOUND_ENV_DATA_INDEX{
                              //EAX:
   I3D_SENV_F_ROOM_LEVEL_LF,     //(float) - room level low freq., ratio (0.0 ... 1.0)
   I3D_SENV_F_ROOM_LEVEL_HF,     //(float) -    ''      high      ''
   I3D_SENV_I_REVERB_DECAY_TIME, //(int) - reverbation decay time, in msec
   I3D_SENV_F_DECAY_HF_RATIO,    //(float) - HF to LF decay time, ratio (0.0 ... 1.0)
   I3D_SENV_F_REFLECTIONS_RATIO, //(float) - early reflections level - relative to room effect, ratio (0.0 ... 1.0)
   I3D_SENV_I_REFLECTIONS_DELAY, //(int) - early reflections delay, in msec
   I3D_SENV_F_REVERB_RATIO,      //(float) - late reverberation level - relative to room effect, (0.0 ... 1.0)
   I3D_SENV_I_REVERB_DELAY,      //(int) - late reverberation delay time relative to initial reflection, in msec
   I3D_SENV_F_ENV_DIFFUSION,     //(float) - environment diffusion, ratio (0.0 ... 1.0)
                              //other:
   I3D_SENV_S_NAME,           //(const char*) - friendly name of environment
   I3D_SENV_F_EMUL_VOLUME,       //(float) - sector's volume in non-EAX mode (emulation)

   I3D_SENV_LAST,
};

//----------------------------

enum I3D_DIRECTORY_TYPE{
   I3DDIR_MAPS,
   I3DDIR_SOUNDS,
   I3DDIR_PROCEDURALS,
   I3DDIR_LAST
};

enum I3D_PROPERTYTYPE{        //size (in bytes)    notes
   I3DPROP_NULL,              //0
   I3DPROP_BOOL,              //1
   I3DPROP_INT,               //4
   I3DPROP_FLOAT,             //4
   I3DPROP_VECTOR3,           //12                 passed as pointer
   I3DPROP_VECTOR4,           //16                 ''
   I3DPROP_COLOR,             //12                 ''
   I3DPROP_STRING,            //strlen(val)        ''
   I3DPROP_ENUM,              //4
   I3DPROP_LAST
};

//----------------------------
typedef I3DENUMRET I3DAPI I3D_ENUMVISUALPROC(dword visual_type, 
   const char *friendly_name, dword context);

#if !defined I3D_FULL 
class I3D_driver{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;
                              //initialization
   I3DMETHOD(Init)(CPI3DINIT) = 0;
   I3DMETHOD(Close)() = 0;
   I3DMETHOD_(bool,IsInit)() const = 0;
   I3DMETHOD(GetStats)(I3DSTATSINDEX, void *data) = 0;
                              //directories
   I3DMETHOD(AddDir)(I3D_DIRECTORY_TYPE, const char*) = 0;
   I3DMETHOD(DeleteDir)(I3D_DIRECTORY_TYPE, const char*) = 0;
   I3DMETHOD(ClearDirs)(I3D_DIRECTORY_TYPE) = 0;
   I3DMETHOD_(int,NumDirs)(I3D_DIRECTORY_TYPE) const = 0;
   I3DMETHOD_(const char*,GetDir)(I3D_DIRECTORY_TYPE, int indx) const = 0;
                              //interfaces
   I3DMETHOD_(class ISND_driver*,GetSoundInterface)() const = 0;
   I3DMETHOD_(class IGraph*,GetGraphInterface)() const = 0;
                              //render state
   I3DMETHOD(SetState)(I3D_RENDERSTATE, dword value) = 0;
   I3DMETHOD_(dword,GetState)(I3D_RENDERSTATE) const = 0;
   I3DMETHOD(FindTextureFormat)(S_pixelformat&, dword flags) const = 0;
   I3DMETHOD(SetTexture)(PI3D_texture) = 0;
                              //rendering
   I3DMETHOD(SetViewport)(dword l, dword t, dword r, dword b) = 0;
   I3DMETHOD(BeginScene)() = 0;
   I3DMETHOD(EndScene)() = 0;
   I3DMETHOD(Clear)(dword color = 0) = 0;
                              //creation - frames
   I3DMETHOD_(PI3D_scene,CreateScene)() = 0;
   I3DMETHOD_(PI3D_user,CreateUserFrame)() = 0;
   I3DMETHOD_(PI3D_dummy,CreateDummy)() = 0;
   I3DMETHOD_(PI3D_target,CreateTarget)() = 0;
   I3DMETHOD_(PI3D_visual,CreateVisual)(dword visual_type) = 0;
   I3DMETHOD_(PI3D_model,CreateModel)() = 0;
   I3DMETHOD_(PI3D_light,CreateLight)() = 0;
   I3DMETHOD_(PI3D_camera,CreateCamera)() = 0;
   I3DMETHOD_(PI3D_volume,CreateVolume)() = 0;
   I3DMETHOD_(PI3D_joint,CreateJoint)() = 0;
   I3DMETHOD_(PI3D_frame,CreateFrame)(dword frm_type, dword sub_type = 0) = 0;
                              //creation - other types
   I3DMETHOD(CreateTexture)(const I3D_CREATETEXTURE*, PI3D_texture*) = 0;
   I3DMETHOD_(PI3D_material,CreateMaterial)() = 0;
   I3DMETHOD_(PI3D_mesh_level,CreateMeshLevel)() = 0;
   I3DMETHOD_(PI3D_mesh_object,CreateMesh)() = 0;
   I3DMETHOD_(PI3D_animation_base,CreateAnimation)(I3D_ANIMATION_TYPE) = 0;
   I3DMETHOD_(PI3D_animation_set,CreateAnimationSet)() = 0;
   I3DMETHOD_(PI3D_interpolator,CreateInterpolator)() = 0;
                              //visual types
   I3DMETHOD(EnumVisualTypes)(I3D_ENUMVISUALPROC*, dword user) const = 0;

   I3DMETHOD_(PI3D_SCNCALLBACK,SetCallback)(PI3D_SCNCALLBACK cb, dword cbf) = 0;
   I3DMETHOD(SetDataBase)(const char *db_filename, dword size, int max_keep_days = 0) = 0;
   I3DMETHOD(FlushDataBase)() = 0;

   I3DMETHOD_(dword,NumProperties)(dword visual_type) const = 0;
   I3DMETHOD_(const char*,GetPropertyName)(dword visual_type, dword index) const = 0;
   I3DMETHOD_(I3D_PROPERTYTYPE,GetPropertyType)(dword visual_type, dword index) const = 0;
   I3DMETHOD_(bool,CollisionResponse)(I3D_collision_data&) const = 0;
   I3DMETHOD_(void,DisableLeakWarning)() const = 0;

                              //sound properties
   I3DMETHOD(SetSoundEnvData)(int set_id, I3D_SOUND_ENV_DATA_INDEX, dword data) = 0;
   I3DMETHOD_(dword, GetSoundEnvData)(int set_id, I3D_SOUND_ENV_DATA_INDEX) const = 0;
};

typedef class I3D_driver *PI3D_driver;
typedef const I3D_driver *CPI3D_driver;
#endif

//----------------------------
                              //located in thunk static lib:
I3D_RESULT I3DAPI I3DCreate2(PI3D_driver *drvp, bool debug);

//----------------------------


#endif
