/*--------------------------------------------------------
   Copyright (c) 1999 Michal Bacik.  All rights reserved.

   File: I3D_math.h     Version: 2.0
   Content: Insanity 3D math classes include file
--------------------------------------------------------*/

#ifndef __MATH_3D_
#define __MATH_3D_
#pragma once

#include <rules.h>
#include <math.h>

#ifndef I3D_MATH_NO_LIB_REF
#ifdef _DEBUG
#pragma comment (lib,"i3d2_math_d")
#else
#pragma comment (lib,"i3d2_math")
#endif
#endif

                              //PI definition
#define PI 3.1415926536f

inline float GetPI(){ return PI; }
inline float GetPI2(){ return PI*2.0f; }
inline float SqrtFast(float f){ return (float)sqrt(f); }

#define MRG_ZERO 1e-8f

#define I3DAPI __stdcall
#define I3DMATHMETHOD_(type,method) type I3DAPI method

//----------------------------

                              //checking against marginal zero
#if defined _MSC_VER

#define MRG_ZERO_BITMASK 0x322BCC77
#pragma warning(disable:4035) //disable func returns no value, we're returning in eax
inline bool IsMrgZero(float f){
   __asm{
      xor eax, eax
      cmp f, MRG_ZERO_BITMASK
      setl al
   }
}
inline bool IsAbsMrgZero(float f){
   __asm{
      fld f
      fabs
      fstp f
      xor eax, eax
      cmp f, MRG_ZERO_BITMASK
      setl al
   }
}
#pragma warning(default:4035)
#else                         //_MSC_VER
                              //compare float value againt near-zero
inline bool IsMrgZero(float f){ return (f<MRG_ZERO); }
                              //compare absolute float value againt near-zero
inline bool IsAbsMrgZero(float f){ return ((float)fabs(f)<MRG_ZERO); }
#endif                        //!_MSC_VER

//----------------------------
                              //2d vector
struct S_vector2{
   union{
      struct{
         float x, y;
      };
      float f[2];
   };
   inline S_vector2(){}
   inline S_vector2(float x1, float y1): x(x1), y(y1){}
   inline float &operator [](int i){ return f[i]; }
   inline const float &operator [](int i) const{ return f[i]; }
   bool operator ==(const S_vector2 &v) const{ return(x==v.x && y==v.y); }
   bool operator !=(const S_vector2 &v) const{ return !(*this==v); }
   inline S_vector2 operator *(float f) const{ return S_vector2(x*f, y*f); }
   inline S_vector2 &operator *=(float f){
      x *= f, y *= f;
      return *this;
   }
   inline S_vector2 operator -() const{ return S_vector2(-x, -y); }
   inline S_vector2 operator /(float f) const{ return S_vector2(x/f, y/f); }
   inline S_vector2 &operator /=(float f){
      x /= f, y /= f;
      return *this;
   }
   inline S_vector2 operator +(const S_vector2& v) const{ return S_vector2(x+v.x, y+v.y); }
   inline S_vector2 &operator +=(const S_vector2& v){
      x += v.x, y += v.y;
      return *this;
   }
   inline S_vector2 operator -(const S_vector2& v) const{ return S_vector2(x-v.x, y-v.y); }
   inline S_vector2& operator -=(const S_vector2& v){
      x -= v.x, y -= v.y;
      return *this;
   }
   inline void Zero(){ x=0.0f, y=0.0f; }
   inline float Magnitude() const{ return SqrtFast(x*x + y*y); }

//----------------------------
// Normalize vector - make its length to be one. If the vector is too small to normalize,
// The returned value is false.
   bool Normalize(){
      float m=Magnitude();
      if(m<MRG_ZERO) return false;
      operator /=(m);
      return true;
   }
   inline float Dot(const S_vector2 &v) const{ return x*v.x + y*v.y; }
   inline float Sum() const{ return x+y; }
   inline void Invert(){ x=-x, y=-y; }
   inline bool IsNull() const{ return ((x*x + y*y) < (MRG_ZERO*MRG_ZERO)); }
   float I3DAPI AngleTo(const S_vector2&) const;
                              // >0 - right, 0 - on line, <0 - left
                              //line is given by point and direction vector
                              //using line equation: ax+by+c, where
                              //(a,b,c) is vector vertical to line
                              //(x,y) is any point on the plane
                              //note: x goes right & y goes down
   float DistanceToLine(const S_vector2 &p, const S_vector2 &d) const{
      S_vector2 n(-d.y, d.x); //get right-turned vector
      n.Normalize();
      return Dot(n) - p.Dot(n);
   }
   inline S_vector2 operator *(const struct S_matrix2&) const;
   inline S_vector2 &operator *=(const struct S_matrix2&);

                              //Get intersection point of 2 lines, both specified by 
                              // a starting point and a direction vector.
                              //If point is found, the return value is true, otherwise 
                              // the lines are parallel or coincident.
   bool GetIntersection(const S_vector2 &p1, const S_vector2 &d1,
      const S_vector2 &p2, const S_vector2 &d2){

      float den = d2.y * d1.x - d2.x * d1.y;
      if(IsAbsMrgZero(den)) return false;
      float u = (d2.x * (p1.y - p2.y) - d2.y * (p1.x - p2.x)) / den;
      x = p1.x + d1.x * u;
      y = p1.y + d1.y * u;
      return true;
   }
};

//----------------------------

struct S_matrix2{
   float m[2][2];

   inline float &operator()(int y, int x){ return m[y][x]; }
   inline const float &operator()(int y, int x) const{ return m[y][x]; }
   inline S_vector2 &operator()(int r){ return (S_vector2&)m[r][0]; }
   inline const S_vector2 &operator()(int r) const{ return (S_vector2&)m[r][0]; }
   void Identity(){
      m[0][0] = 0.0f;
      m[0][1] = 0.0f;
      m[1][0] = 0.0f;
      m[1][1] = 1.0f;
   }
   void Zero(){
      m[0][0] = 0.0f;
      m[1][1] = 0.0f;
      m[0][1] = 0.0f;
      m[1][0] = 0.0f;
   }
                              //base is at (0, 1), PI/2 is at (-1, 0)
   void SetDir(float a){
      m[0][0] = m[1][1] = (float)cos(a);
      m[1][0] = -(m[0][1] = (float)sin(a));
   }
   S_matrix2 operator *(const S_matrix2 &m1){
      S_matrix2 r;
      r.m[0][0] = m[0][0]*m1.m[0][0] + m[0][1]*m1.m[1][0];
      r.m[0][1] = m[0][0]*m1.m[0][1] + m[0][1]*m1.m[1][1];
      r.m[1][0] = m[1][0]*m1.m[0][0] + m[1][1]*m1.m[1][0];
      r.m[1][1] = m[1][0]*m1.m[0][1] + m[1][1]*m1.m[1][1];
      return r;
   }
};

inline S_vector2 S_vector2::operator *(const struct S_matrix2 &m) const{
   return S_vector2(x*m(0, 0)+y*m(1, 0), x*m(0, 1)+y*m(1, 1));
}

inline S_vector2 &S_vector2::operator *=(const struct S_matrix2 &m){
   x = x*m(0, 0) + y*m(1, 0);
   y = x*m(0, 1) + y*m(1, 1);
   return *this;
}

//----------------------------
                              //3D vector
struct S_vector{
   union{
      struct{
         float x, y, z;
      };
      float f[3];
   };
   inline S_vector(){}
   inline S_vector(float x1, float y1, float z1): x(x1), y(y1), z(z1) {}

   inline float &operator [](int i){ return f[i]; }
   inline const float &operator [](int i) const{ return f[i]; }
   inline bool operator ==(const S_vector &v) const{ return(x==v.x && y==v.y && z==v.z); }
   inline bool operator !=(const S_vector &v) const{ return !(*this==v); }
   inline S_vector operator *(float f) const{ return S_vector(x*f, y*f, z*f); }
   inline S_vector &operator *=(const S_vector &v){
      x *= v.x, y *= v.y, z *= v.z;
      return (*this);
   }
   inline S_vector &operator *=(float f){
      x *= f, y *= f, z *= f;
      return (*this);
   }
   S_vector &I3DAPI operator *=(const struct S_matrix&);
   S_vector I3DAPI operator *(const struct S_matrix&) const;
   inline S_vector operator -() const{ return S_vector(-x, -y, -z); }
   inline S_vector operator /(float f) const{
      f = 1.0f / f;
      return S_vector(x*f, y*f, z*f);
   }
   inline S_vector &operator /=(float f){
      f = 1.0f / f;
      x *= f, y *= f, z *= f;
      return *this;
   }
   inline S_vector operator +(const S_vector& v) const{ return S_vector(x+v.x, y+v.y, z+v.z); }
   inline S_vector &operator +=(const S_vector& v){
      x += v.x, y += v.y, z += v.z;
      return *this;
   }
   inline S_vector operator -(const S_vector& v) const{ return S_vector(x-v.x, y-v.y, z-v.z); }
   inline S_vector& operator -=(const S_vector& v){
      x -= v.x, y -= v.y, z -= v.z;
      return *this;
   }

//----------------------------
// Rotate vector by matrix (3x3 part of matrix used only).
   S_vector I3DAPI RotateByMatrix(const struct S_matrix&) const;
   inline S_vector operator %(const struct S_matrix &m) const{
      return RotateByMatrix(m);
   }
   inline S_vector &operator %=(const struct S_matrix &m){
      *this = RotateByMatrix(m);
      return *this;
   }

   inline void Zero(){ x=0.0f, y=0.0f, z=0.0f; }

   inline S_vector operator *(const S_vector &v) const{ return S_vector(x*v.x, y*v.y, z*v.z); }

   inline float Dot(const S_vector &v) const{ return x*v.x + y*v.y + z*v.z; }
   inline float Square() const{ return x*x + y*y + z*z; }

   inline float Magnitude() const{
      float sq = Dot(*this);
      if(IsMrgZero(sq)) return 0.0f;
      return SqrtFast(sq);
   }

//----------------------------
// Normalize vector - make its length to be one. If the vector is too small to normalize,
// The returned value is false.
   bool Normalize(){
      float m = Magnitude();
      if(IsMrgZero(m))
         return false;
      operator /=(m);
      return true;
   }

   inline S_vector Cross(const S_vector &v) const{
      return S_vector(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
   }
   inline float Sum() const{ return x+y+z; }
   inline void Invert(){ x=-x, y=-y, z=-z; }
   inline bool IsNull() const{ return ((x*x + y*y + z*z) < (MRG_ZERO*MRG_ZERO)); }
                              // >0 - sharp, 0 - PI/2, <0 - not sharp
   float I3DAPI AngleTo(const S_vector&) const;
                              // >0 - in front of, 0 - on, <0 - behind
                              //plane is given by point and normal
                              //using plane equation: ax+by+cz+d, where
                              //(a,b,c) is plane normal (length=1 !) and
                              //(x,y,z) is any point on the plane
   float DistanceToPlane(const S_vector &p, const S_vector &n) const{
      return Dot(n) - p.Dot(n);
   }
   inline float DistanceToPlane(const struct S_plane&) const;
//----------------------------
// Compute position of vector to line. The returned value is a multiplier, which
// gives closest position to vector, when used in formula: p + dir * multiplier.
   inline float PositionOnLine(const S_vector &p, const S_vector &dir) const{
      float len_2 = dir.Dot(dir);
      if(len_2<MRG_ZERO) return 0.0f;
      return dir.Dot((*this) -p) / len_2;
   }

//----------------------------
//Compute distance to line.
// Line is given by a starting point and a direction vector.
// If direction vector is zero, the return value is negative
// to mark an error.
   inline float DistanceToLine(const S_vector &p, const S_vector &dir) const{
      float len_2 = dir.Dot(dir);
      if(len_2<MRG_ZERO) return -1e+16f;
      float u = dir.Dot((*this) -p) / len_2;
      return ((p + dir * u) - *this).Magnitude();
   }
   inline void GetNormal(const S_vector &u1, const S_vector &v1, const S_vector &w1){
      float ux = w1.x - v1.x, uy = w1.y - v1.y, uz = w1.z - v1.z;
      float vx = u1.x - v1.x, vy = u1.y - v1.y, vz = u1.z - v1.z;
      x = uy * vz - uz * vy;
      y = uz * vx - ux * vz;
      z = ux * vy - uy * vx;
   }
   inline S_vector Min1(const S_vector &v) const{
      return S_vector(Min(x, v.x), Min(y, v.y), Min(z, v.z));
   }
   inline S_vector Max1(const S_vector &v) const{
      return S_vector(Max(x, v.x), Max(y, v.y), Max(z, v.z));
   }
   inline S_vector &Minimal(const S_vector &v){
      x = Min(x, v.x); y = Min(y, v.y); z = Min(z, v.z);
      return *this;
   }
   inline S_vector &Maximal(const S_vector &v){
      x = Max(x, v.x); y = Max(y, v.y); z = Max(z, v.z);
      return *this;
   }
};

typedef S_vector *LPS_vector;

//----------------------------

struct S_normal: public S_vector{
   inline S_normal(){}

                              //Construct unit-length normal from 2 direction vectors (using cross-product).
   inline S_normal(const S_vector &u, const S_vector &v){
      x = u.y * v.z - u.z * v.y;
      y = u.z * v.x - u.x * v.z;
      z = u.x * v.y - u.y * v.x;
      if(!Normalize())
         Zero();
   }

                              //Construct unit-length normal from vector.
   inline S_normal(const S_vector &v):
      S_vector(v)
   {
      if(!Normalize())
         Zero();
   }
};

typedef S_normal *LPS_normal;

//----------------------------
                              //homogeneous vector, containing w
struct S_vector4{
   S_vector xyz;
   float w;

   S_vector4 I3DAPI operator *(const struct S_matrix&) const;
   inline void Zero(){
      xyz.Zero();
      w = 0.0f;
   }
};

//----------------------------

struct S_vectorw: public S_vector{
   float w;
   S_vectorw(){}
   S_vectorw(float x1, float y1, float z1, float w1): S_vector(x1, y1, z1), w(w1){}
   S_vectorw(const S_vector &v): S_vector(v), w(1.0f){}
   S_vectorw I3DAPI operator *(const struct S_matrix&) const;
};

//----------------------------
struct S_quat;

const S_matrix &I3DAPI GetIdentityMatrix();
                              //matrix 4x4
struct S_matrix{
   float m[4][4];
   S_matrix(){}   

   inline float &operator()(int y, int x){ return m[y][x]; }
   inline const float &operator()(int y, int x) const{ return m[y][x]; }
   inline S_vector &operator()(int r){ return (S_vector&)m[r][0]; }
   inline const S_vector &operator()(int r) const{ return (S_vector&)m[r][0]; }
   void I3DAPI Identity();
   void I3DAPI Zero();
   bool I3DAPI Invert();
   inline bool Inverse(){ return Invert(); }
   void I3DAPI Transpose();

//----------------------------
// Set rotation part of matrix from given direction vector and rotation around this vector.
// If the direction vector is too small, the returned value is false.
   bool I3DAPI SetDir(const S_vector &dir, float roll);

//----------------------------
// Set rotation part of matrix from given direction vector and rotation around this vector.
// If any of input vectors is too small, the returned value is false.
   bool I3DAPI SetDir(const S_vector &dir, const S_vector &up);

//----------------------------
// Multiplication assignment.
   I3DMATHMETHOD_(S_matrix&,operator *=)(const S_matrix&);
   inline S_matrix &operator *=(const S_quat &q){ return operator *=(S_matrix(q)); }

   S_matrix I3DAPI operator *(const S_matrix&) const;
   S_matrix I3DAPI Mult4X4(const S_matrix&) const;

//----------------------------
// Rotate matrix by matrix (3x3 part of matrices used only).
   //S_matrix I3DAPI RotateByMatrix(const S_matrix&) const;
   S_matrix I3DAPI operator %(const S_matrix&) const;
   I3DMATHMETHOD_(S_matrix&,operator %=)(const S_matrix &m){
      *this = operator %(m);
      return *this;
   }

   S_vector I3DAPI GetScale() const;

//----------------------------
// Make simple rotations around basic 3 axes, the rest of matrix is set to identity.
   void I3DAPI RotationX(float angle);
   void I3DAPI RotationY(float angle);
   void I3DAPI RotationZ(float angle);

//----------------------------
// Construct matrix from quaternion.
   explicit I3DAPI S_matrix(const S_quat&);
   
//----------------------------
// Get inversed matrix.
   inline S_matrix operator ~() const{
      S_matrix rtn = *this;
      rtn.Invert();
      return rtn;
   }

//----------------------------
// Make resulting matrix by multiplaying two input matrices, using 3x4 parts.
   I3DMATHMETHOD_(S_matrix&,Make)(const S_matrix&, const S_matrix&);

//----------------------------
// Make resulting matrix by multiplaying two input matrices, using 4x4 parts.
   I3DMATHMETHOD_(S_matrix&,Make4X4)(const S_matrix&, const S_matrix&);

//----------------------------
// Same as Make4X4, but transpose matrix after transformations.
   I3DMATHMETHOD_(S_matrix&,Make4X4Transposed)(const S_matrix&, const S_matrix&);
};

typedef S_matrix *LPS_matrix;

//----------------------------
                              //3D quaternion
struct S_quat{
   float s;
   S_vector v;
                              //constructors
   inline S_quat(){}
   inline S_quat(float s1, float x1, float y1, float z1): s(s1){
      v.x = x1, v.y = y1, v.z = z1;
   }
   inline explicit S_quat(float s1, const S_vector &v1): s(s1), v(v1) {}
   inline explicit S_quat(const S_vector &axis, float angle){ Make(axis, angle); }
   explicit I3DAPI S_quat(const S_matrix&);

   void I3DAPI Make(const S_vector &axis, float angle);

//----------------------------
// Set rotation from given direction vector and rotation around this vector.
// If the direction vector is too small, the returned value is false.
   bool I3DAPI SetDir(const S_vector &dir, float roll);

//----------------------------
// Set rotation from given direction vector and rotation around this vector.
// If any of input vectors is too small, the returned value is false.
   bool I3DAPI SetDir(const S_vector &dir, const S_vector &up);

//----------------------------
   void I3DAPI Inverse(S_vector &axis, float &angle) const;
   //S_matrix I3DAPI RotationMatrix() const;
   inline S_quat operator +(const S_quat &q) const{ return S_quat(s+q.s, v+q.v); }
   inline S_quat operator -(const S_quat &q) const{ return S_quat(s-q.s, v-q.v); }
   inline S_quat operator -() const{ return S_quat(s, -v); }
   inline S_quat operator *(const S_quat &q) const{
      return S_quat(s * q.s - v.Dot(q.v), q.v * s + v * q.s + v.Cross(q.v));
   }

//----------------------------
// Make indentity quaternion.
   inline void Identity(){ s = 1.0f; v.Zero(); }

//----------------------------
// Get direction vector from quaternion.
   inline S_vector GetDir() const{
      float sx = v.x * 2.0f;
      float sy = v.y * 2.0f;
      float sz = v.z * 2.0f;
      return S_vector((v.x*sz) - (s*sy), (v.y*sz) + (s*sx), 1.0f - (v.x*sx) - (v.y*sy));
   }

//----------------------------
// Get angle of rotation which this quaterion represents.
   inline float GetAngle() const{ return (float)acos(Max(-1.0f, Min(1.0f, s))) * 2.0f; }

//----------------------------
// Multiplication assignment.
   inline S_quat &operator *=(const S_quat &q){
      float s1 = s;
      s = s * q.s - v.Dot(q.v);
      v = q.v * s1 + v * q.s + v.Cross(q.v);
      return *this;
   }
   inline S_quat &operator *=(const S_matrix &m){ return operator *=(S_quat(m)); }

   inline S_quat operator *(float f) const{
      return S_quat(s * f, v * f);
   }
   inline bool operator ==(const S_quat &q) const{ return(s==q.s && v==q.v); }
   inline float Dot(const S_quat q) const{ return s*q.s + v.x*q.v.x + v.y*q.v.y + v.z*q.v.z; }
   S_quat I3DAPI Slerp(const S_quat &q, float t, bool shorten=false) const;

//----------------------------
// Scale quaternion - this is the same as linear interpolation (Slerp) with unit quaternion.
   void I3DAPI Scale(float);

   S_quat I3DAPI RotateByMatrix(const struct S_matrix&) const;
};

typedef S_quat *LPS_quat;

//----------------------------
                              //plane - parametric
struct S_plane{
   S_vector normal;
   float d;
   S_plane(){}
   S_plane(const S_vector &n, float d1): normal(n), d(d1){}
   S_plane(float x, float y, float z, float d1): normal(x, y, z), d(d1){}
   S_plane(const S_vector &v0, const S_vector &v1, const S_vector &v2){
      normal.GetNormal(v0, v1, v2);
      normal.Normalize();
      d = -normal.Dot(v0);
   }
                              //intersection with another plane
                              //return: xpt=point on line
                              //        xdir=normalized line vector
   bool I3DAPI IntersectionLine(const S_plane&, S_vector &xpt, S_vector &xdir) const;

//----------------------------
// Get intersection of line and plane - return ratio on line, where intersection occured.
   bool IntersectionPosition(const S_vector &pos, const S_vector &dir, float &ret) const{
      float f = normal.Dot(dir);
      if(fabs(f)<MRG_ZERO)
         return false;
      ret = -(normal.Dot(pos)+d) / f;
      return true;
   }

//----------------------------
// Get intersection of line and plane - return intersection point.
   inline bool Intersection(const S_vector &pos, const S_vector &dir, S_vector &ret) const{
      float f;
      if(!IntersectionPosition(pos, dir, f))
         return false;
      ret = pos + dir * f;
      return true;
   }

//----------------------------
// Compute plane transformed by a matrix.
   S_plane operator *(const S_matrix &m) const{
      S_plane rtn;
      rtn.normal = normal.RotateByMatrix(m);
      rtn.normal.Normalize();
      S_vector p = (-normal * d) * m;
      rtn.d = -p.Dot(rtn.normal);
      return rtn;
   }

//----------------------------
// Compute inversive plane.
   inline S_plane operator -() const{ return S_plane(-normal.x, -normal.y, -normal.z, -d); }

//----------------------------
// Multiply all elements of plane by scalar.
   S_plane &operator *=(float f){
      normal *= f;
      d *= f;
      return *this;
   }

//----------------------------
// Invert plane.
   inline void Invert(){
      normal.Invert();
      d = -d;
   }
};

typedef S_plane *LPS_plane;

//----------------------------

inline float S_vector::DistanceToPlane(const S_plane &pl) const{
   return Dot(pl.normal) + pl.d;
}

//----------------------------

struct S_int_random{
   int i;
   I3DAPI S_int_random(int max);
   inline operator int() const{ return i; }
};

//----------------------------

struct S_float_random{
   float f;
   I3DAPI S_float_random();
   I3DAPI S_float_random(float max);
   inline operator float() const{ return f; }
};

//----------------------------
                              //math functions
//----------------------------
// Get intersection (closest connection) of 2 lines in 3D.
// The returned u1 and u2 are positions on line1 and line2 where 2 closest points are located.
// If lines are perpendicular, the call fails.
bool I3DAPI LineIntersection(const S_vector &p1, const S_vector &dir1,
   const S_vector &p2, const S_vector &dir2,
   float &u1, float &u2);

//----------------------------

#endif
