#ifndef __INSANITY_H
#define __INSANITY_H


#include <i3d\i3d2.h>
#include <direct.h>
#include <smartptr.h>
#include <c_string.hpp>
#include <vector>
#include <debug.h>
#include <c_unknwn.hpp>

#ifdef _MSC_VER
using namespace std;
#endif


/************************************
  Created By: Michal Bacik
  Date: 11/22/99 2:02:57 PM
  Description:
   Template used to cache any Insanity 3D interface objects, which support
      AddRef(), Release()
 ************************************/


template <class T>
class C_I3D_cache_base{
public:
   virtual void SetCacheSize(int sz) = 0;
   virtual I3D_RESULT Open(T *t1, const char *filename, PI3D_scene, dword open_flags = 0,
      void (*err_fnc)(const char *msg, void *context) = NULL, void *err_fnc_context = NULL) = 0;
   virtual I3D_RESULT Create(const char *filename, T **ret, PI3D_scene, void (*err_fnc)(const char *msg, void *context) = NULL, void *err_fnc_context = NULL) = 0;
   virtual T *CreateNew(PI3D_scene) const{ return NULL; }
   virtual void Clear() = 0;
   virtual ~C_I3D_cache_base(){};
   virtual void AddDir(const char*) = 0;
   virtual void DelDir(const char*) = 0;
   virtual void ClearDirs() = 0;
   virtual int NumDirs() const = 0;
   virtual const char *GetDir(int i) const = 0;
   virtual bool GetRelativeDirectory(char *fname) = 0;
   virtual int GetCacheSize() const{ return -1; }
   virtual int GetMemSize() const{ return -1; }
};

//----------------------------


                              //cache implementation

template <class T>
class C_I3D_cache: public C_I3D_cache_base<T>{
   struct S_itm{
      C_smart_ptr<T> item;
      C_string filename;
      S_itm(){}
      S_itm(const S_itm &si){
         operator =(si);
      }
      S_itm &operator =(const S_itm &si){
         item = si.item;
         filename = si.filename;
         return (*this);
      };
   };
   vector<S_itm> cache;       //cache of items
   vector<C_string> dirs;     //dirs in which we search for file
   int cache_size;            //max. number of items kept in cache

   struct S_err_call{
      void(*err_fnc)(const char*, void*);
      void *err_fnc_context;
   };
   static bool I3DAPI cbErr(I3D_LOADMESSAGE msg, dword prm1, dword prm2, void *context){
      if(msg==CBM_ERROR){
         S_err_call *ec = (S_err_call*)context;
         (*ec->err_fnc)((const char*)prm1, ec->err_fnc_context);
      }
      return false;
   }
public:
   C_I3D_cache():
      cache_size(128)
   {}

//----------------------------

   ~C_I3D_cache(){
      Clear();
   }

//----------------------------

   virtual void SetCacheSize(int sz){ cache_size = sz; }

//----------------------------

   virtual void Duplicate(T *dst, T *src) = 0;

//----------------------------

   virtual I3D_RESULT OpenFile(T *t, const char *dir, const char *filename, dword flags,
      I3D_LOAD_CB_PROC *cb_proc, void *cb_context){

      return t->Open(C_fstring("%s%s", dir, filename), flags, cb_proc, cb_context);
   }

//----------------------------

   virtual I3D_RESULT Open(T *t1, const char *filename, PI3D_scene scn,
      dword open_flags = 0, void (*err_fnc)(const char*, void*) = NULL, void *err_fnc_context = NULL){

      int i;
                              //if it exists in cache, remove it first
      for(i=cache.size(); i--; ) if(cache[i].item==t1) break;
      if(i!=-1) cache.erase(&cache[i]);
                              //look into cache first
      for(i=0; i<cache.size(); i++){
         if(!cache[i].filename.Matchi(filename)) continue;
         Duplicate(t1, cache[i].item);
         return I3D_OK;
      }
      I3D_RESULT ir = I3DERR_NOFILE;   //used if no dir is registered

      S_err_call err_call = {err_fnc, err_fnc_context};

                              //always build sectors if model is loaded
      dword flags = open_flags | I3DLOAD_BUILD_SECTORS;
      if(err_fnc) flags |= I3DLOAD_ERRORS;
                              //look into all dirs
      for(i=0; i<dirs.size(); i++){
         ir = OpenFile(t1, dirs[i], filename, flags, cbErr, &err_call);
         switch(ir){
         case I3DERR_NOFILE: continue;
         case I3DERR_OUTOFMEM: 
         case I3DERR_FILECORRUPTED:
            return ir;
         }
         if(I3D_SUCCESS(ir)){
                              //put into cache
            if(cache.size()>=cache_size){
                              //try to release oldest items
               for(i=0; i<cache.size(); i++){
                  cache[i].item->AddRef();
                  if(cache[i].item->Release()==1){
                     cache.erase(&cache[i]);
                     break;
                  }
               }
            }
            cache.push_back(S_itm());
            cache.back().filename = filename;
                              //keep unique item in cache, so that it's not affected
                              // by modifications
            T *tu = CreateNew(scn);
            assert(tu);
            Duplicate(tu, t1);
            cache.back().item = tu;
            tu->Release();
            return ir;
         }
      }
      if(err_fnc)             //report error
         err_fnc(C_fstring("Failed to open file: '%s'", filename), err_fnc_context);
      return ir;
   }

//----------------------------

   I3D_RESULT Create(const char *filename, T **ret, PI3D_scene scn,
      void (*err_fnc)(const char*, void*) = NULL, void *err_fnc_context = NULL){

      *ret = NULL;
      int i;
                              //look into cache first
      for(i=0; i<cache.size(); i++){
         if(!cache[i].filename.Matchi(filename)) continue;
         cache[i].item->AddRef();
         *ret = cache[i].item;
         return I3D_OK;
      }
      I3D_RESULT ir = I3DERR_NOFILE;   //used if no dir is registered

      T *tp = CreateNew(scn);
      if(!tp) return NULL;
      S_err_call err_call = {err_fnc, err_fnc_context};
                              //look into all dirs
      for(i=0; i<dirs.size(); i++){
         ir = OpenFile(tp, dirs[i], filename, err_fnc ? I3DLOAD_ERRORS : 0, cbErr, &err_call);
         switch(ir){
         case I3DERR_NOFILE: continue;
         case I3DERR_OUTOFMEM: 
         case I3DERR_FILECORRUPTED:
            {
               tp->Release();
               return ir;
            }
         }
         if(I3D_SUCCESS(ir)){
                              //put into cache
            if(cache.size()>=cache_size){
                              //try to release oldest items
               for(i=0; i<cache.size(); i++){
                  cache[i].item->AddRef();
                  if(cache[i].item->Release()==1){
                     cache.erase(&cache[i]);
                     break;
                  }
               }
            }
            cache.push_back(S_itm());
            cache.back().item = tp;
            cache.back().filename = filename;
            *ret = tp;
            return ir;
         }
      }
      tp->Release();
      if(err_fnc)             //report error
         err_fnc(C_fstring("Failed to load file: '%s'", filename), err_fnc_context);
      return ir;
   }

//----------------------------

   virtual void Clear(){
      cache.clear();
   }

//----------------------------

   virtual int GetCacheSize() const{ return cache.size(); }

//----------------------------
                              //directory functions

   void AddDir(const char *cp){ 
      DelDir(cp);
      C_fstring str = cp;
      if(str[(int)str.Size()-1] != '\\') str += "\\";
      dirs.push_back(str); 
   }
   void DelDir(const char *cp){
      C_fstring str = cp;
      if(str[(int)str.Size()-1] != '\\') str += "\\";
      for(int i=dirs.size(); i--; ) if(dirs[i]==str){
         dirs.erase(&dirs[i]);
         break;
      }
   }
   void ClearDirs(){ dirs.clear(); }
   int NumDirs() const{ return dirs.size(); } 
   const char *GetDir(int i) const{ return dirs[i]; }

//----------------------------
                              //check is full-path filename 'fname' is relative of
                              //any registered directories
                              //if match found, func returns true and fname is stripped
                              //by removing base directory

   bool GetRelativeDirectory(char *fn1){
                              //try to strip current working directory from fname
      char cd[260];
      _getcwd(cd, sizeof(cd));
      C_fstring wild_dir("%s\\*", cd);
      C_string fname(fn1);
      if(fname.Matchi(wild_dir))
         fname = &fname[(int)wild_dir.Size()-1];

      for(int i=0; i<dirs.size(); i++){
         C_string wild_dir = dirs[i] + "*";
         if(fname.Matchi(wild_dir)){
            //fname = &fname[(int)dirs[i].size()];
            strcpy(fn1, &fname[(int)dirs[i].Size()]);
            return true;
         }
      }
      return false;
   }
};

//----------------------------

class C_I3D_model_cache: public C_I3D_cache<I3D_model>{
   
   virtual void Duplicate(PI3D_model dst, PI3D_model src){
                              //we're not duplicating basic I3D_frame values
                              // thus we save it now and restore after duplication
      C_string save_name = dst->GetName();
      S_vector pos = dst->GetPos();
      S_quat rot = dst->GetRot();
      float scale = dst->GetScale();
      dword save_data = dst->GetData();
      dword frm_flags = dst->GetFlags();

      dst->Duplicate(src);
      dst->SetName(save_name);
      dst->SetPos(pos);
      dst->SetRot(rot);
      dst->SetScale(scale);
      dst->SetData(save_data);
      dst->SetFlags(frm_flags);
                              //zero user data of duplicated model's frames
      const PI3D_frame *mod_frms = dst->GetFrames();
      for(int i=dst->NumFrames(); i--; ){
         mod_frms[i]->SetData(0);
      }
   }
   virtual PI3D_model CreateNew(PI3D_scene scn) const{
      return scn->GetDriver()->CreateModel();
   }
};

//----------------------------

class C_I3D_sound_cache: public C_I3D_cache<I3D_sound>{

   static void snd_err_fnc(const char *str, void *context){

      C_string &err_string = *(C_string*)context;
      err_string = str;
   }

   virtual void Duplicate(PI3D_sound dst, PI3D_sound src){
                              //duplication means letting all params except sound source
      dst->Open(src->GetFileName(), src->GetOpenFlags());
   }
   virtual PI3D_sound CreateNew(PI3D_scene scn) const{
      return scn->CreateSound();
   }
public:
   virtual I3D_RESULT Open(PI3D_sound t1, const char *filename, PI3D_scene scn,
      dword open_flags = 0, void (*err_fnc)(const char*, void*) = NULL, void *err_fnc_context = NULL){

      C_string err_string;
      I3D_RESULT ir;
#ifdef _DEBUG                 //make sure filename doesn't contain extension
      for(int i=strlen(filename); i--; ){
         char c = filename[i];
         if(c=='\\')
            break;
         if(c=='.'){
            assert(0);
            break;
         }
      }
#endif
                              //try 2 extensions
      ir = C_I3D_cache<I3D_sound>::Open(t1, C_fstring("%s.ogg", filename),
         scn, open_flags, err_fnc ? snd_err_fnc : NULL, &err_string);
      if(I3D_SUCCESS(ir))
         return ir;

      ir = C_I3D_cache<I3D_sound>::Open(t1, C_fstring("%s.wav", filename),
         scn, open_flags, err_fnc ? snd_err_fnc : NULL, &err_string);
      if(I3D_SUCCESS(ir))
         return ir;
      if(err_fnc && err_string.Size())
         err_fnc(err_string, err_fnc_context);
      return ir;
   }

};

//----------------------------

class C_I3D_anim_cache: public C_I3D_cache<I3D_animation_set>{
   virtual void Duplicate(PI3D_animation_set dst, PI3D_animation_set src){}
   virtual PI3D_animation_set CreateNew(PI3D_scene scn) const{
      return scn->GetDriver()->CreateAnimationSet();
   }
};

//----------------------------

                              //structure containing all necessary info
                              // for updating anything
struct S_tick_context{
   int time;                  //time from last update
   char key;                  //recently pressed key
   byte modify_keys;          //key modifiers (Shift, Alt, Ctrl)
   int mouse_rel[3];          //mouse relative movement since last frame
   byte mouse_buttons;        //combination of mouse buttons currently held down
   class C_controller *p_ctrl;//pointer to controller (in updated state)

//----------------------------
// Clear this struct.
   void Zero(){
      memset(this, 0, sizeof(S_tick_context));
   }

//----------------------------
// Clear input parameters, just leave time.
   void ClearInput(){
      key = -1;
      modify_keys = 0;
      mouse_rel[0] = 0;
      mouse_rel[1] = 0;
      mouse_rel[2] = 0;
      mouse_buttons = 0;
      p_ctrl = NULL;
   }
};

//----------------------------
                              //base class for any class used in game loop for main processing
class C_class_to_be_ticked: public C_unknown{
public:
   virtual ~C_class_to_be_ticked(){}

//----------------------------
// Get identifier of this class. Each inherited class of C_class_to_be_ticked has specific ID,
// by which it may be identified. ID of particular class defined with the class.
   virtual dword GetID() const = 0;

//----------------------------
// Update - to be called once per frame.
   virtual void Tick(const S_tick_context&) = 0;

//----------------------------
// Render frame.
   virtual void Render() = 0;

//----------------------------
// Screen reset, update changes (when change in resolution, bitdepth, or device is reset).
   virtual void ResetScreen(){}

//----------------------------
// Process incomming network chunk.
   virtual void NetIn(enum E_NETWORK_MESSAGE msg_id,
      class C_net_in_msg *in_msg){ }

//----------------------------
   //virtual void ProcessSystemNetMessage(PI_net inet, PC_net_in_msg msg, dword context){}
   virtual void OnPlayerDisconnect(dword pid, const char *name){}
   virtual void OnSessionLost(){}
};

typedef C_class_to_be_ticked *PC_class_to_be_ticked;

//----------------------------
                              //Insanity editor binary file chunks
                              // used by Insanity.dll for savinf into
                              // binary file, using C_chunk class (c_chunk.h)

enum E_INSANITY_EDITOR_CHUNK_TYPE{

   CT_BASECHUNK   =0x4c53,    //mission data

//----------------------------   *** COMPATIBILITY ***
      CT_MODELS   =0x2000,    //list of models
         CT_MODEL =0x2010,
            CT_MODELNAME   =0x2011,
            CT_MODELFILENAME  =0x2012,
            //CT_POSITION pos
            //CT_QUATERNION rot
            //CT_SCALE scale
            //CT_MODEL_FLAGS,
            CT_MODEL_DEL_FRAME =0x2013,
               //const char[]
               //word flags [optional]
      CT_CAMERA   =0x3000,
         //CT_POSITION
         //CT_QUATERNION
         CT_CAMERA_FOV     =0x3010,
            //float
         CT_CAMERA_RANGE   =0x3011,
            //float far_range
         CT_CAMERA_ORTHOGONAL    =0x3012,
         CT_CAMERA_ORTHO_SCALE   =0x3013,

      CT_SCENE_BGND_COLOR  =0x3200,
         //S_vector scene_background_color
      CT_SCENE_BACKDROP    =0x3211,
         //float near_range, far_range

      CT_MODIFICATIONS=0x4000,
         CT_MODIFICATION =0x4010,
            //CT_NAME
            //CT_POSITION
            //CT_QUATERNION
         CT_FRAME_TYPE   =0x4011,
         CT_FRAME_SUBTYPE   =0x4012,
            //dword I3D_FRAME_TYPE
         CT_FRAME_FLAGS = 0x4014,
            //dword frame_flags
         CT_LINK  =0x4020,
            //CT_NAME   (parent name)
         //CT_LIGHTCOLOR  =0x4030,
         //CT_LIGHTPOWER  =0x4031,
         CT_LSECTOR     =0x4032,

         CT_HIDDEN      =0x4033,
         CT_MODIFY_LIGHT   =0x4040,       //light params, followed by params:
            CT_LIGHT_TYPE  =0x4041,
               //dword type
            //CT_COLOR color
            CT_LIGHT_POWER =0x4042,
               //float power
            CT_LIGHT_CONE  =0x4043,
               //float cone_in, cone_out
            CT_LIGHT_RANGE =0x4044,
               //float range_n, range_f
            CT_LIGHT_MODE  =0x4045,
               //dword mode
            CT_LIGHT_SECTOR=0x4046,
               //char[] sector_name
            CT_LIGHT_SPECULAR_COLOR =0x4047,
               //S_vector color
            CT_LIGHT_SPECULAR_POWER =0x4048,
               //float power
            //CT_LIGHT_SPECULAR_ANGLE =0x4049,
               //float angle (in radians)
         CT_MODIFY_DUMMY   =0x4050,
            //I3D_bbox
         CT_MODIFY_SOUND   =0x4060,
            //CT_NAME file_name
            CT_SOUND_TYPE  =0x4061,
               //dword type
            CT_SOUND_VOLUME   =0x4062,
               //float volume
            CT_SOUND_OUTVOL   =0x4063,
               //float outside_volume
            CT_SOUND_CONE  =0x4064,
               //float cone_in, cone_out
            CT_SOUND_RANGE =0x4065,
               //float range_n, range_f, range_fade
            CT_SOUND_LOOP  =0x4066,
            CT_SOUND_ENABLE   =0x4067,
            CT_SOUND_SECTOR   =0x4068,
            CT_SOUND_STREAMING  =0x4069,
               //char[] sector_name
         CT_MODIFY_VOLUME  =0x4070,
            CT_VOLUME_TYPE =0x4071,
               //dword type
            CT_VOLUME_MATERIAL =0x4072,
               //dword material_id
         CT_MODIFY_OCCLUDER   =0x4080,
            CT_OCCLUDER_VERTICES =0x4082,
               //dword num_vertices
               //S_vector[num_vertices] vertex_pool
         CT_MODIFY_VISUAL     =0x4090,
            //dword visual_type
            CT_VISUAL_MATERIAL =0x4092,
               //dword material_id
         CT_MODIFY_LIT_OBJECT=0x40a0,
            //lit-object data
         CT_MODIFY_SECTOR  =0x40b0,
            CT_SECTOR_SOUND_DATA =0x40b1,
               //word I3D_SECTOR_SOUND_DATA_INDEX
               //dword data
            CT_SECTOR_ENVIRONMENT =0x40b2,
               //word env_id
         CT_MODIFY_VISUAL_PROPERTY  =0x40c0,
            //byte I3D_PROPERTYTYPE
            //word property_index
            //property data (variable length)
         CT_MODIFY_TARGET =0x40d0,
            //string target_name
      CT_BSP_FILE = 0x6800,

      CT_EDITOR_PLUGIN = 0x7800,
         //CT_NAME plugin_name
         //plugin-specified data

//----------------------------
                              //FREE: 8000 - BFFF
                              // (H&D reserved: 8000 - 9FFF)
                              // (gangsters reserved: A000 - BFFF)

//----------------------------
                              //EDITOR: C000 - CFFF

   CT_EDITOR_STATE = 0xc000,  //editor plugins state
      CT_EDITOR_PLUGIN_STATE = 0xc010,
         //CT_NAME
         //plugin custom data
//----------------------------
                              //FREE: E000 - FFFF

};

//----------------------------

//----------------------------

#endif
