#include "all.h"
#include "resource.h"
#include <dta_read.h>
#include <cstdlib>
#include <vector>
#include <map>
#include <c_string.hpp>

using namespace std;

/*--------------------------------------------------------
   Copyright (c) 2002 Lonely Cat Games,  All rights reserved.
   Data reader - read-only access to data packages.
   Contents:
      - data initalizer class
      - I/O functions simulating standard open, close, read and seek

   Access:
      28. 3. 1999    Michal Bacik
      - re-designed to support openinng the same file multiple times - there was a bug
         that each file kept only one current position regardles of number of open handles
         on them
      13. 9. 2000    Michal Bacik
      - support for file time, info read from DTA file.
--------------------------------------------------------*/

#include <rules.h>
#include <c_unknwn.hpp>


#define COMPCNT_VERSION '6'

typedef class C_dta_read *LPC_dta_read;

#define DTA_SEEK_SET 0
#define DTA_SEEK_CUR 1
#define DTA_SEEK_END 2

#define DTACREATE_SHOWINFO 1  //show dta file scanning during content creation

//----------------------------
                              //help functions

                              //compare 2 string without case respect
inline bool StrcmpLower(const C_string &str1, const C_string &str2){

   if(str1.Size() != str2.Size())
      return false;
   int i = str1.Size();
   while(i--){
      char c(tolower(str2[i]));
      if(c != tolower(str1[i]))
         return false;
   }
   return true;
}

//----------------------------
                              //single file in open DTA
                              //NOTE: This structure is written onto disk into content file
                              //    in order to cache file information.
                              //    If there's  change in this struct, increase the
                              //    COMPCNT_VERSION constant.
#pragma pack(push, 1)
struct S_file_entry{
   dword filepos;             //position within DTA file
   dword length;              //length of file entry
   dword file_time;           //in MS-DOS format
};
#pragma pack(pop)

typedef S_file_entry *LPS_file_entry;

//----------------------------

class C_unk_base{
public:
   virtual unsigned long DTAAPI AddRef() = 0;
   virtual unsigned long DTAAPI Release() = 0;
};

//----------------------------
                              //internal class declaration

class C_dta_read_imp: public C_dta_read{
   dword ref;
   typedef map<C_string, S_file_entry> t_file_map;
   t_file_map file_map;
   HANDLE dta_handle;         //filename opened by operating system
   dword curr_file_pos;       //real position within DTA file

   C_dta_read_imp():
      ref(1),
      dta_handle(NULL),
      curr_file_pos(0)
   {
   }
   ~C_dta_read_imp(){
      Close();
   }
                              //internal - called by destructor
   void Close();

   friend LPC_dta_read DTAAPI dtaCreate(const char *, dword, const char*);
public:
   bool Open(const char *name, dword flags, const char *info_title);

//----------------------------
// Find entry in file table, identified by name. The name is assumed to be in lower-case.
   LPS_file_entry FindEntry(const C_string &fname){

      t_file_map::iterator it = file_map.find(fname);
      if(it!=file_map.end())
         return &(*it).second;
      return NULL;
   }

//----------------------------

   void SetPointer(dword p){
      if(curr_file_pos==p)
         return;
      SetFilePointer(dta_handle, curr_file_pos=p, NULL, FILE_BEGIN);
   }

//----------------------------

   bool ReadFile(void *buffer, dword len, dword *ret){

      if(!::ReadFile(dta_handle, buffer, len, ret, NULL))
         return false;
                              //adjust file position
      curr_file_pos += *ret;
      return true;
   }

//----------------------------
public:
   virtual unsigned long DTAAPI AddRef(){ return ++ref; }
   virtual unsigned long DTAAPI Release(){ if(--ref) return ref; delete this; return 0; }
};


//----------------------------
                              //statics

                              //list of open dta
                              //filled front-to-back
                              //entries are searched for file back-to-front, so
                              //later opened, earlier searched
static vector<C_dta_read_imp*> *dta_list;

//----------------------------

struct S_dlg_init{
   const char *title;
   const char *message;
};

//----------------------------

static HWND hwnd_dlg;

static BOOL CALLBACK DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

   switch(uMsg){
   case WM_INITDIALOG:
      {
         const S_dlg_init &dlg_init = *(S_dlg_init*)lParam;
         RECT rc;
         GetWindowRect(hwnd, &rc);
         int x = (GetSystemMetrics(SM_CXSCREEN)-(rc.right-rc.left))/2;
         int y = (GetSystemMetrics(SM_CYSCREEN)-(rc.bottom-rc.top))/2;
         SetWindowPos(hwnd, NULL, x, y, 0, 0, SWP_NOACTIVATE | SWP_NOREDRAW | SWP_NOSIZE);
         if(dlg_init.title)
            SendMessage(hwnd, WM_SETTEXT, 0, (LPARAM)dlg_init.title);
         SetDlgItemText(hwnd, IDC_I, dlg_init.message);

         ShowWindow(hwnd, SW_SHOW);
         UpdateWindow(hwnd);
      }
      return 1;
   case WM_USER:
      {
         char **cp = (char**)lParam;
         if(cp)
            SetDlgItemText(hwnd, IDC_I, cp[1]);
      }
      break;
   }
   return 0;
}

//----------------------------

static void ShowInfo(const char *title, const char *text){

   const char *cp[2] = {
      title, text
   };
   if(!hwnd_dlg){
      S_dlg_init dlg_init = {title, text };
      hwnd_dlg = CreateDialogParam(GetModuleHandle("icom_rd.dll"),
         "DLG_INFO", NULL, (DLGPROC)DialogProc, (LPARAM)&dlg_init);
   }else{
      SendMessage(hwnd_dlg, WM_USER, 0, (LPARAM)cp);
   }
}

//----------------------------

static void HideInfo(){

   if(hwnd_dlg){
      DestroyWindow(hwnd_dlg);
      hwnd_dlg = NULL;
   }
}

//----------------------------

#pragma pack(push, 1)
struct S_cnt_header{
   char header[8];
   dword size;
   FILETIME ftWrite;
   dword filenum;
};

struct S_cnt_entry{
   dword filepos, length;
   dword file_time;
   word name_len;
};
#pragma pack(pop)

//----------------------------

bool C_dta_read_imp::Open(const char *name, dword flags, const char *info_title){

   Close();

   int i;
   ulong ul;
   char temp_name[512];
   const char *full_name;
   char drive;

                              //check if full path in filename
   if(name[0] && name[1]==':' && name[2]=='\\'){
      drive = name[0];
      full_name = name;
   }else{
      GetCurrentDirectory(sizeof(temp_name), temp_name);
      drive = temp_name[0];
      full_name = name;
   }

                              //open file
   HANDLE h = CreateFile(full_name, GENERIC_READ, FILE_SHARE_READ, NULL,
      OPEN_EXISTING,
      //FILE_FLAG_SEQUENTIAL_SCAN,
      FILE_FLAG_RANDOM_ACCESS,
      NULL);
                              //open error?
   if(h == INVALID_HANDLE_VALUE)
      return false;

   char cnt_name[512];
   S_cnt_header cnt_header = {
      'C', 'O', 'M', 'P', 'C', 'N', 'T', COMPCNT_VERSION
   }, test_header;
   HANDLE cnt_h;

   bool is_writable_disk = false;
   switch(GetDriveType(C_fstring("%c:\\", drive))){
   case DRIVE_RAMDISK:
   case DRIVE_FIXED:
      is_writable_disk = true;
      break;
   }
   bool cnt_ok = false;

   for(int ci=0; ci<2; ci++){
                              //get contents name
      if(!ci){
                              //use current dir
         strcpy(cnt_name, full_name);
      }else{
                              //use temp dir
         GetTempPath(sizeof(cnt_name), cnt_name);
         i = strlen(cnt_name);
         if(i && cnt_name[i-1] != '\\') cnt_name[i] = '\\';
         i = strlen(name);
         while(name[i-1] != '\\') if(!--i) break;
         strcat(cnt_name, &full_name[i]);
      }

                              //change extension
      for(i = strlen(cnt_name); i--; ){
         if(cnt_name[i]=='.')
            break;
      }
      strcpy(&cnt_name[i], ".cnt");
                              //try open contents file
      cnt_h = CreateFile(cnt_name, GENERIC_READ, FILE_SHARE_READ, NULL,
         OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
      if(cnt_h != INVALID_HANDLE_VALUE){
                                 //get file info
         cnt_header.size = GetFileSize(h, NULL);
         GetFileTime(h, NULL, NULL, &cnt_header.ftWrite);
                                 //read contents header
         if(::ReadFile(cnt_h, &test_header, sizeof(test_header), &ul, 0) && ul==sizeof(test_header)){

                                 //allow small time differences (few secs), due to time precision losing
                                 // during files installation
            __int64 i64_tm0 = *(__int64*)&cnt_header.ftWrite;
            __int64 i64_tm1 = *(__int64*)&test_header.ftWrite;
            __int64 time_delta = i64_tm0 - i64_tm1;
#if defined _DEBUG && 0
                                 //convert file times to human-readable form (for debugging)
            SYSTEMTIME st_cnt, st_file;
            FileTimeToSystemTime(&cnt_header.ftWrite, &st_cnt);
            FileTimeToSystemTime(&test_header.ftWrite, &st_file);
#endif
                                 //compare headers (without filenum member)
            if(!memcmp(&cnt_header, &test_header, offsetof(S_cnt_header, ftWrite)) &&
               time_delta < __int64(0x4000000) && time_delta>-__int64(0x4000000)){
                                 //valid contents file, read
               for(i=0; i<(int)test_header.filenum; i++){
                  S_cnt_entry ce;
                  if(!::ReadFile(cnt_h, &ce, sizeof(ce), &ul, 0)) break;
                  if(ul!=sizeof(ce)) break;
                  char filename[512];
                  if(!::ReadFile(cnt_h, filename, ce.name_len, &ul, 0)) break;
                  if(ul != ce.name_len) break;
                  filename[ce.name_len] = 0;

                  S_file_entry fe;
                  fe.filepos = ce.filepos;
                  fe.length = ce.length;
                  fe.file_time = ce.file_time;
                  file_map.insert(pair<C_string, S_file_entry>(filename, fe));
               }
               if(i == (int)test_header.filenum)
                  cnt_ok = true;
               else
                  file_map.clear();
            }
         }
         CloseHandle(cnt_h);
         if(!cnt_ok && is_writable_disk){
                                 //delete contents file
            SetFileAttributes(cnt_name, FILE_ATTRIBUTE_ARCHIVE);
            DeleteFile(cnt_name);
         }
      }
      if(cnt_ok || is_writable_disk)
         break;
   }

   if(!cnt_ok){
      if(flags&DTACREATE_SHOWINFO)
         ShowInfo(info_title, name);
                              //get # of files
#pragma pack(push, 1)
                              //RAR file system
      struct S_dta_header{
         dword id;
         dword res[4];
      } dta_header;

      struct S_file_header{
         word header_crc;

                              //HEAD_TYPE=0x72          marker block
                              //HEAD_TYPE=0x73          archive header
                              //HEAD_TYPE=0x74          file header
                              //HEAD_TYPE=0x75          comment header
                              //HEAD_TYPE=0x76          extra information
                              //HEAD_TYPE=0x77          subblock
                              //HEAD_TYPE=0x78          recovery record
         byte block_type;

                              //0x01 - file continued from previous volume
                              //0x02 - file continued in next volume
                              //0x04 - file encrypted with password
                              //0x08 - file comment present
                              //0x10 - information from previous files is used (solid flag)
                              // (for RAR 2.0 and later)
         word block_flags;

                              //File header full size including file name and comments
         word header_size;
         dword file_len_packed;
         dword file_len_unpacked;

                              //Operating system used for archiving
                              //0 - MS DOS
                              //1 - OS/2
                              //2 - Win32
                              //3 - Unix
         byte host_os;

         dword file_crc;

                              //file date and time in MS-DOS format
         dword file_time;        

                              //10 * Major version + minor version
         byte rar_version;

                              //Packing method encoding:
                              //0x30 - storing
                              //0x31 - fastest compression
                              //0x32 - fast compression
                              //0x33 - normal compression
                              //0x34 - good compression
                              //0x35 - best compression
         byte pack_method;

         word fname_len;

         dword file_attributes;
      } file_header;
#pragma pack(pop)

      dword file_pos = 0;

                              //read file header
      ::ReadFile(h, &dta_header, sizeof(dta_header), &ul, 0);
      bool ok = true;
                              //support 2 header types: Rar and generic DTA
      if(ul != sizeof(dta_header)){
         ok = false;
      }else
      switch(dta_header.id){
      case 0x21726152:        //"Rar!"
      case 0x5f415444:        //"DTA_"
         break;
      default:
         ok = false;
      }
      if(!ok){
         CloseHandle(h);
         return false;
      }
      file_pos += ul;
      int win_update_count = 0;  //file counter after which we'll let windows breath
                              //just for easy break-ing out
      while(true){
         ::ReadFile(h, &file_header, sizeof(file_header), &ul, 0);
         if(ul<sizeof(file_header))
            break;
         file_pos += ul;
         char filename[512];
         if(file_header.fname_len < sizeof(filename)-1){

            ::ReadFile(h, &filename, file_header.fname_len, &ul, 0);
            if(ul!=file_header.fname_len && file_header.file_len_packed != file_header.file_len_unpacked){
               file_map.clear();
               break;
            }
            filename[ul] = 0;
            file_pos += ul;

            S_file_entry fe;
            fe.length = file_header.file_len_packed;
            fe.file_time = file_header.file_time;
            fe.filepos = file_pos;
            C_string name = filename;
            name.ToLower();
            file_map.insert(pair<C_string, S_file_entry>(name, fe));

            SetFilePointer(h, file_header.file_len_packed, 0, FILE_CURRENT);
            file_pos += file_header.file_len_packed;
         }
                                       //let windows breath
         if(++win_update_count==20){
            win_update_count = 0;
            MSG msg;
            while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE)){
               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         }
      }

      if(!file_map.size()){   //no files?
         CloseHandle(h);
         return false;
      }

                              //write contents file
      i = 0;
      cnt_h = CreateFile(cnt_name, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
         FILE_ATTRIBUTE_NORMAL, NULL);
      if(cnt_h != INVALID_HANDLE_VALUE){
         cnt_header.size = GetFileSize(h, NULL);
         GetFileTime(h, NULL, NULL, &cnt_header.ftWrite);
         cnt_header.filenum = file_map.size();
                              //write file info
         bool ok;
         ok = WriteFile(cnt_h, &cnt_header, sizeof(cnt_header), &ul, NULL);
         if(ok)
            ok = (ul==sizeof(cnt_header));
         for(t_file_map::const_iterator it = file_map.begin(); it!=file_map.end(); it++){
            const S_file_entry &fe = (*it).second;
            const C_string &name = (*it).first;

            S_cnt_entry ce;
            ce.filepos = fe.filepos;
            ce.length = fe.length;
            ce.file_time = fe.file_time;
            ce.name_len = name.Size();
            WriteFile(cnt_h, &ce, sizeof(ce), &ul, NULL);
            if(!(ok=(ul==sizeof(ce))))
               break;
            WriteFile(cnt_h, &name[0], ce.name_len, &ul, NULL);
            if(!(ok=(ul==ce.name_len)))
               break;
         }
         CloseHandle(cnt_h);

         if(!ok)
            DeleteFile(cnt_name);
      }
   }
   dta_handle = h;

   if(!dta_list)
      dta_list = new vector<C_dta_read_imp*>;
   dta_list->push_back(this);
   curr_file_pos = 0;
   return true;
}

//----------------------------

void C_dta_read_imp::Close(){

   file_map.clear();
   CloseHandle(dta_handle);
   dta_handle = NULL;

                              //remove from static list
   if(dta_list)
   for(int i=dta_list->size(); i--; ){
      if((*dta_list)[i]==this){
         dta_list->erase(&(*dta_list)[i]);
         if(!dta_list->size()){
            delete dta_list;
            dta_list = NULL;
         }
         break;
      }
   }
}

//----------------------------
//----------------------------

class C_dta_handle: public C_unknown{
public:
   HANDLE win_h;              //valid if file opened by Windows file system

   LPS_file_entry fe;         //non-null if file opened by dta function
   C_dta_read_imp *dta;

   dword dta_file_pos;
   C_dta_handle():
      win_h(INVALID_HANDLE_VALUE),
      fe(NULL),
      dta(NULL),
      dta_file_pos(0)
   {}
};
typedef C_dta_handle *LPC_dta_handle;

//----------------------------

int DTAAPI dtaOpen(const char *name){

#if defined _DEBUG && 0
   OutputDebugString(name); OutputDebugString("\n");
#endif

   if(hwnd_dlg)
      HideInfo();
   LPC_dta_handle handle = new C_dta_handle;

   C_string str(name);
   str.ToLower();
   if(dta_list)
   for(int i = dta_list->size(); i--; ){
                              //search all dta-s
      handle->fe = (*dta_list)[i]->FindEntry(str);
      if(handle->fe){
         handle->dta = (*dta_list)[i];
         handle->dta->AddRef();
         break;
      }
   }
   if(!handle->fe){
      handle->win_h = CreateFile(name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
         FILE_FLAG_SEQUENTIAL_SCAN, NULL);
      if(handle->win_h==INVALID_HANDLE_VALUE){
         handle->Release();
         return -1;
      }
   }
   return (int)handle;
}

//----------------------------

bool DTAAPI dtaGetTime(int h1, dword time[2]){
   
   if(h1==-1)
      return false;
   LPC_dta_handle handle = (LPC_dta_handle)h1;
   if(!handle)
      return false;

   if(handle->fe){
      DosDateTimeToFileTime(handle->fe->file_time>>16, handle->fe->file_time&0xffff,
         (LPFILETIME)time);
   }else
   if(handle->win_h!=INVALID_HANDLE_VALUE){
      GetFileTime(handle->win_h, NULL, NULL, (LPFILETIME)time);
   }
   return true;
}

//----------------------------

int DTAAPI dtaOpenWrite(const char *name){

   if(hwnd_dlg)
      HideInfo();
   LPC_dta_handle handle = new C_dta_handle;
   handle->win_h = CreateFile(name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, 
      NULL, CREATE_ALWAYS,
      FILE_ATTRIBUTE_NORMAL, NULL);
   if(handle->win_h==INVALID_HANDLE_VALUE){
      handle->Release();
      return -1;
   }
   return (int)handle;
}

//----------------------------

int DTAAPI dtaWrite(int h1, const void *buffer, dword len1){

   LPC_dta_handle handle = (LPC_dta_handle)h1;
   dword ret = -1;

   if(handle->win_h!=INVALID_HANDLE_VALUE){
      if(!WriteFile(handle->win_h, buffer, len1, &ret, 0))
         ret = -1;
   }
   return ret;
}

//----------------------------

bool DTAAPI dtaClose(int h1){

   if(h1==-1)
      return false;
   LPC_dta_handle handle = (LPC_dta_handle)h1;
   if(!handle)
      return false;
   bool ret = false;
   if(handle->fe){
      handle->dta->Release();
      ret = true;
   }else
   if(handle->win_h!=INVALID_HANDLE_VALUE){
      ret = CloseHandle(handle->win_h);
   }
   handle->Release();
   return ret;
}


//----------------------------

int DTAAPI dtaRead(int h1, void *buffer, dword len1){

   if(h1==-1)
      return -1;
   LPC_dta_handle handle = (LPC_dta_handle)h1;
   if(!handle)
      return false;
   dword ret = -1;
   if(handle->fe){
                              //validate read length
      len1 = Min((int)len1, (int)handle->fe->length - (int)handle->dta_file_pos);
                                 //seek to file
      handle->dta->SetPointer(handle->fe->filepos + handle->dta_file_pos);
      if(!handle->dta->ReadFile(buffer, len1, &ret))
         return -1;
                                 //adjust file position
      handle->dta_file_pos += ret;
   }else
   if(handle->win_h!=INVALID_HANDLE_VALUE){
      if(!ReadFile(handle->win_h, buffer, len1, &ret, 0))
         ret = -1;
   }
   return ret;
}

//----------------------------

int DTAAPI dtaSeek(int h1, int pos1, int whence){

   if(h1==-1)
      return -1;
   LPC_dta_handle handle = (LPC_dta_handle)h1;
   if(!handle)
      return false;
   int ret = -1;

   if(handle->fe){
      switch(whence){
      case DTA_SEEK_CUR:
         pos1 += handle->dta_file_pos;
         break;
      case DTA_SEEK_END:
         pos1 += handle->fe->length;
         break;
      }
                              //clamp to area of file
      pos1 = Max(Min(pos1, (int)handle->fe->length), 0);
      ret =
      handle->dta_file_pos = pos1;
   }else
   if(handle->win_h!=INVALID_HANDLE_VALUE){
      ret = SetFilePointer(handle->win_h, pos1, NULL, whence);
   }
   return ret;
}

//----------------------------
                              //convert binary buffer to text format
                              //removes '\r' characters from buffer
                              //returns size of new buffer
int DTAAPI dtaBin2Text(void *mem, dword len){

   char *src=(char*)mem, *dst=(char*)mem;
   if(len)
   do{
      if(*src!='\r') *dst++=*src;
   }while(++src, --len);
   return dst - (char*)mem;
}

//----------------------------
                              //create dta object - open dta file,
                              //if open fails, creation fails too and NULL is returned
LPC_dta_read DTAAPI dtaCreate(const char *fname, dword flags, const char *info_title){

   C_dta_read_imp *dta = new C_dta_read_imp;
   if(!dta->Open(fname, flags, info_title)){
      dta->Release();
      return NULL;
   }
   return dta;
}

//----------------------------

LPC_dta_read DTAAPI dtaGetDtaPack(int h){

   if(h==-1)
      return NULL;
   LPC_dta_handle handle = (LPC_dta_handle)h;
   if(!handle)
      return NULL;
   return handle->dta;
}

//----------------------------
