#ifndef __ANIM_H
#define __ANIM_H

#include <vector>
#include <c_string.hpp>
#include <math\bezier.hpp>
#include <math\ease.hpp>
#include "frame.h"


//----------------------------

class I3D_animation_base{
   I3D_ANIMATION_TYPE type;
protected:
   dword ref;
   PI3D_driver drv;

   I3D_animation_base(PI3D_driver d1, I3D_ANIMATION_TYPE t1):
      ref(1),
      drv(d1),
      type(t1)
   {
      drv->AddCount(I3D_CLID_ANIMATION_BASE);
   }
   ~I3D_animation_base(){
      drv->DecCount(I3D_CLID_ANIMATION_BASE);
   }
public:
   inline I3D_ANIMATION_TYPE GetType1() const{ return type; }
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(I3D_ANIMATION_TYPE,GetType)() const{ return type; }
};

//----------------------------

class I3D_keyframe_anim: public I3D_animation_base{
   struct S_anim_note{
      int time;
      C_string note;
      S_anim_note(){}
      S_anim_note(int t1, const char *cp1): time(t1), note(cp1) {}
      S_anim_note(const S_anim_note &an){ operator =(an); }
      S_anim_note &operator =(const S_anim_note &an){
         time = an.time;
         note = an.note;
         return *this;
      }
   };
   
   int end_time;
   S_quat base_rot;           //keep base rotation, since rotation track values are relative
                              //anim keys for pos/rot/scl/not/vis
   vector<I3D_anim_pos_bezier> prs_keys[3];
   vector<I3D_anim_rot> rot_keys;
   vector<S_anim_note> note_keys;
   vector<I3D_anim_visibility> vis_keys;
   vector<I3D_anim_color> color_keys;
   vector<I3D_anim_power> power_keys;

   friend I3D_scene;
   friend I3D_driver;
   friend I3D_model;
   friend class I3D_interpolator;
   friend class I3D_interpolator_PRS;

   I3D_keyframe_anim(PI3D_driver);
   I3D_keyframe_anim(const I3D_keyframe_anim&);
   I3D_keyframe_anim& operator=(const I3D_keyframe_anim&);
public:
   inline const vector<I3D_anim_pos_bezier> &GetPositionKeysVector() const{ return prs_keys[0]; }
   inline const vector<I3D_anim_pos_bezier> &GetDirectionKeysVector() const{ return prs_keys[1]; }
   inline const vector<I3D_anim_rot> &GetRotationKeysVector() const{ return rot_keys; }

   inline const vector<S_anim_note> &GetNoteKeysVector() const{ return note_keys; }
   inline const vector<I3D_anim_visibility> &GetVisibilityKeysVector() const{ return vis_keys; }
   inline const vector<I3D_anim_color> &GetColorKeysVector() const{ return color_keys; }
   inline const vector<I3D_anim_power> &GetPowerKeysVector() const{ return power_keys; }

   inline const S_quat &GetBaseRotation() const{ return base_rot; }
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(void,Clear)();
   I3DMETHOD(SetPositionKeys)(I3D_anim_pos_tcb*, dword num_keys);
   I3DMETHOD(SetRotationKeys)(I3D_anim_rot*, dword num_keys);
   I3DMETHOD(SetLookAtKeys)(I3D_anim_look_at_tcb*, dword num_keys);
   I3DMETHOD(SetScaleKeys)(I3D_anim_scale_tcb*, dword num_keys);
   I3DMETHOD(SetNoteKeys)(I3D_anim_note*, dword num_keys);
   I3DMETHOD(SetVisibilityKeys)(I3D_anim_visibility*, dword num_keys);
   I3DMETHOD(SetColorKeys)(I3D_anim_color*, dword num_keys);
   I3DMETHOD(SetPowerKeys)(I3D_anim_power*, dword num_keys);
   I3DMETHOD(SetEndTime)(dword);

   I3DMETHOD(GetPositionKeys)(const I3D_anim_pos_bezier **kp, dword *nk) const{
      *kp = prs_keys[0].begin();
      *nk = prs_keys[0].size();
      return I3D_OK;
   }
   I3DMETHOD(GetRotationKeys)(const I3D_anim_rot **kp, dword *nk) const{
      *kp = rot_keys.begin();
      *nk = rot_keys.size();
      return I3D_OK;
   }
   I3DMETHOD(GetLookAtKeys)(const I3D_anim_pos_bezier **kp, dword *nk) const{
      *kp = prs_keys[1].begin();
      *nk = prs_keys[1].size();
      return I3D_OK;
   }
   I3DMETHOD(GetScaleKeys)(const I3D_anim_pos_bezier **kp, dword *nk) const{
      *kp = prs_keys[2].begin();
      *nk = prs_keys[2].size();
      return I3D_OK;
   }
   I3DMETHOD(GetNoteKeys)(const I3D_anim_note **kp, dword *nk) const{
      /*
      *kp = note_keys.begin();
      *nk = note_keys.size();
      return I3D_OK;
      */
      return I3DERR_UNSUPPORTED;
   }
   I3DMETHOD(GetVisibilityKeys)(const I3D_anim_visibility **kp, dword *nk) const{
      *kp = vis_keys.begin();
      *nk = vis_keys.size();
      return I3D_OK;
   }
   I3DMETHOD(GetColorKeys)(const I3D_anim_color **kp, dword *nk) const{
      *kp = color_keys.begin();
      *nk = color_keys.size();
      return I3D_OK;
   }
   I3DMETHOD(GetPowerKeys)(const I3D_anim_power **kp, dword *nk) const{
      *kp = power_keys.begin();
      *nk = power_keys.size();
      return I3D_OK;
   }
   I3DMETHOD_(dword,GetEndTime)() const{ return end_time; }
};

//----------------------------

class I3D_anim_pose: public I3D_animation_base{

#define APOSF_POS       1     //position defined
#define APOSF_ROT       2     //rotation defined
#define APOSF_POW       4     //power defined
#define APOSF_DIR       8     //direction defined

   PI3D_driver drv;
   dword apos_flags;

   S_vector position, direction, up_vector;
   S_quat rotation;
   float power;

   I3D_anim_pose(PI3D_driver d1):
      I3D_animation_base(d1, I3DANIM_POSE),
      apos_flags(0)
   {
   }
   friend I3D_driver;
public:
   inline dword GetFlags() const{ return apos_flags; }
   inline const S_vector &GetPos1() const{ return position; }
   inline const S_quat &GetRot1() const{ return rotation; }
   inline const float GetPower1() const{ return power; }
   inline const S_vector &GetDir1() const{ return direction; }
   inline const S_vector &GetUpVector1() const{ return up_vector; }
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(void,Clear)(){
      apos_flags = 0;
   }

   I3DMETHOD_(void,SetPos)(const S_vector *v){
      apos_flags &= ~APOSF_POS;
      if(v){
         position = *v;
         apos_flags |= APOSF_POS;
      }
   }
   I3DMETHOD_(void,SetPower)(const float *f){
      apos_flags &= ~APOSF_POW;
      if(f){
         power = *f;
         apos_flags |= APOSF_POW;
      }

   }
   I3DMETHOD_(void,SetRot)(const S_quat *q){
      apos_flags &= ~APOSF_ROT;
      if(q){
         rotation = *q;
         apos_flags |= APOSF_ROT;
      }
   }
   I3DMETHOD_(void,SetDir)(const S_vector *look_dir, const S_vector *up){
      apos_flags &= ~APOSF_DIR;
      if(look_dir && up){
         direction = *look_dir;
         up_vector = *up;
         apos_flags |= APOSF_DIR;
      }
   }
   I3DMETHOD_(const S_vector*,GetPos)() const{
      return (apos_flags&APOSF_POS) ? &position : NULL;
   }
   I3DMETHOD_(const S_quat*,GetRot)() const{
      return (apos_flags&APOSF_ROT) ? &rotation : NULL;
   }
   I3DMETHOD_(const float*,GetPower)() const{
      return (apos_flags&APOSF_POW) ? &power : NULL;
   }
   I3DMETHOD_(void,GetDir)(const S_vector **look_dir, const S_vector **up){
      if(apos_flags&APOSF_DIR){
         *look_dir = &direction;
         *up = &up_vector;
      }else{
         *look_dir = NULL;
         *up = NULL;
      }
   }
};

//----------------------------

class I3D_animation_set{
protected:
   dword ref;
public:
   struct S_anim_link{
      C_smart_ptr<I3D_animation_base> anim;
      C_string link;

      S_anim_link(){}
      S_anim_link(const S_anim_link &al){ operator =(al); }
      S_anim_link &operator =(const S_anim_link &al){
         anim = al.anim;
         link = al.link;
         return *this;
      }
   };
private:
   vector<S_anim_link> anims;
   PI3D_driver drv;

   I3D_animation_set(PI3D_driver);

   friend I3D_driver;
public:
   ~I3D_animation_set();
   const S_anim_link *GetAnimLinks() const{ return &anims[0]; }
   dword NumLinks() const{ return anims.size(); }
   void AddAnimation1(PI3D_animation_base, const C_string &link);
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(Open)(const char *fname, dword flags = 0, PI3D_LOAD_CB_PROC = NULL,
      void *context = NULL, PI3D_scene scene_hlp = NULL);
   I3DMETHOD_(dword,GetTime)() const;
   
   I3DMETHOD_(void,AddAnimation)(PI3D_animation_base anim, const char *link){
      AddAnimation1(anim, link);
   }
   I3DMETHOD(DeleteAnimation)(PI3D_animation_base);

   I3DMETHOD_(dword,NumAnimations)() const{ return anims.size(); }
   I3DMETHOD_(PI3D_animation_base,GetAnimation)(int i) const{ return anims[i].anim; }
   I3DMETHOD_(const char*,GetAnimLink)(int i) const{ return anims[i].link; }

   I3DMETHOD(Duplicate)(PI3D_animation_set aset){
      anims = aset->anims;
      return I3D_OK;
   }
};

typedef I3D_animation_set *PI3D_animation_set;

//----------------------------

class I3D_interpolator{
protected:
   dword ref;

   PI3D_driver drv;

   C_smart_ptr<I3D_frame> frm;  //frame to which it applies
public:
   I3D_interpolator(PI3D_driver d):
      ref(1),
      drv(d)
   {
      drv->AddCount(I3D_CLID_INTERPOLATOR);
   }

   ~I3D_interpolator(){
      drv->DecCount(I3D_CLID_INTERPOLATOR);
   }

   inline PI3D_frame GetFrame1() const{ return frm; }
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
                              //controll
   I3DMETHOD_(I3D_RESULT,Tick)(int time) = 0;
                              //setup
   I3DMETHOD_(I3D_RESULT,SetFrame)(PI3D_frame in_frm){ frm = in_frm; return I3D_OK; }

   I3DMETHOD_(I3D_RESULT,SetAnimation)(dword stage, PI3D_animation_base ab, dword options = I3DANIMOP_SET,
      float scale_factor = 1.0f, float blend_factor = 1.0f, float speed = 1.0f){
      return I3DERR_UNSUPPORTED;
   }

   I3DMETHOD_(I3D_RESULT,SetScaleFactor)(dword stage, float factor){ return I3DERR_UNSUPPORTED; }
   I3DMETHOD_(I3D_RESULT,SetBlendFactor)(dword stage, float factor){ return I3DERR_UNSUPPORTED; }
   I3DMETHOD_(I3D_RESULT,SetSpeed)(dword stage, float speed){ return I3DERR_UNSUPPORTED; }
   I3DMETHOD_(I3D_RESULT,SetEndTime)(dword stage, dword time){ return I3DERR_UNSUPPORTED; }
   I3DMETHOD_(I3D_RESULT,SetCurrTime)(dword stage, dword time){ return I3DERR_UNSUPPORTED; }
                              //retrieve
   I3DMETHOD_(PI3D_frame,GetFrame)() const{ return frm; }
   I3DMETHOD_(PI3D_animation_base,GetAnimation)(dword stage) const{ return NULL; }
   I3DMETHOD_(dword,GetOptions)(dword stage) const{ return 0; }
   I3DMETHOD_(float,GetScaleFactor)(dword stage) const{ return 0.0f; }
   I3DMETHOD_(float,GetBlendFactor)(dword stage) const{ return 0.0f; }
   I3DMETHOD_(float,GetSpeed)(dword stage) const{ return 0.0f; }
   I3DMETHOD_(dword,GetEndTime)(dword stage = 0) const{ return 0; }
   I3DMETHOD_(dword,GetCurrTime)(dword stage = 0) const{ return 0; }
public:
   virtual int NumActiveStages() const{ return false; }

//----------------------------
// Returns true if it is PRS interpolator.
   virtual bool IsPRS() const{ return false; }
};

typedef I3D_interpolator *PI3D_interpolator;

//----------------------------
//----------------------------

#endif
