/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.
   All rights reserved.

   File: Anim_3d.cpp
   Content: All animation classes, and all about animations.
--------------------------------------------------------*/

#include "all.h"
#include "anim.h"
#include "light.h"
#include "particle.h"
#include "model.h"
#include "soundi3d.h"
#include <malloc.h>           //alloca
#include <map>

#define USE_ALLOCA            //use alloca for local storage instead of new[] / delete[]

//----------------------------
//----------------------------

static const S_quat ZERO_QUATERNION(1.0f, 0.0f, 0.0f, 0.0f);

//----------------------------
//----------------------------

I3D_keyframe_anim::I3D_keyframe_anim(PI3D_driver d):
   I3D_animation_base(d, I3DANIM_KEYFRAME),
   end_time(0),
   base_rot(0.0f, 0.0f, 0.0f, 1.0f)
{
   
}

//----------------------------

void I3D_keyframe_anim::Clear(){

   end_time = 0;
   base_rot = S_quat(0.0f, 0.0f, 0.0f, 1.0f);
                              //clear all keys
   for(int i=0; i<3; i++) prs_keys[i].clear();
   rot_keys.clear();
   note_keys.clear();
}

//----------------------------
                              //convert 3 position key into tangent values
                              //for middle key
static void Tcb2Bezier(
   const I3D_anim_vector_tcb &prev_key,
   const I3D_anim_vector_tcb &curr_key,
   const I3D_anim_vector_tcb &next_key,
   I3D_bezier &bezier){
/*
   float speed_ratio;
   if(!border_key){
      float speed_in, speed_out;
      speed_in=(curr_key.v - prev_key.v).Magnitude() / (float)(curr_key.time - prev_key.time);
      speed_out=(next_key.v - curr_key.v).Magnitude() / (float)(next_key.time - curr_key.time);
      //speed_ratio=speed_in/speed_out;
      speed_ratio = float(curr_key.time - prev_key.time) / float(next_key.time - curr_key.time);
   }else speed_ratio=1.0f;
   */

   S_vector delta = prev_key.v - next_key.v;
   bezier.in_tan = delta;
   bezier.out_tan = -delta;
   float size, mag;
   float tension = (1.0f - curr_key.tension) / 3;
   float continuity = -curr_key.continuity;
   S_vector dir;
                        //compute in-tangent
   dir = prev_key.v - curr_key.v;
   bezier.in_tan += (dir-bezier.in_tan) * continuity;
   mag = bezier.in_tan.Magnitude();
   if(mag > MRG_ZERO){
      size = tension * dir.Magnitude();
      bezier.in_tan *= (size/mag);
      /*
      if(speed_ratio<1.0f) bezier.in_tan *= speed_ratio;
      else bezier.in_tan *= (2.0f-(1.0f/speed_ratio));
      */
      bezier.in_tan += curr_key.v;
   }else bezier.in_tan = curr_key.v;
                        //compute out-tangent
   dir = next_key.v - curr_key.v;
   bezier.out_tan += (dir-bezier.out_tan) * continuity;
   mag = bezier.out_tan.Magnitude();
   if(mag > MRG_ZERO){
      size = tension * dir.Magnitude();
      bezier.out_tan *= (size/mag);
      /*
      speed_ratio = 1.0f/speed_ratio;
      if(speed_ratio<1.0f) bezier.out_tan *= speed_ratio;
      else bezier.out_tan *= (2.0f-(1.0f/speed_ratio));
      */
      bezier.out_tan += curr_key.v;
   }else bezier.out_tan = curr_key.v;
}

//----------------------------

static void SetupVectorKeys(vector<I3D_anim_vector_bezier> &v_keys,
   I3D_anim_vector_tcb *keys, dword num_keys){

   v_keys.clear();
   if(!keys || !num_keys) return;

   v_keys.reserve(num_keys);
   for(int i=0; i<num_keys; i++){
                              //setup key base
      v_keys.push_back(I3D_anim_vector_bezier());
      I3D_anim_vector_bezier &key=v_keys.back();
      key.time=keys[i].time;
      key.v=keys[i].v;
      key.easy_from=keys[i].easy_from;
      key.easy_to=keys[i].easy_to;
                              //convert TCB into BEZIER
      Tcb2Bezier(keys[i ? i-1 : i], keys[i], keys[(i<num_keys-1) ? i+1 : i], key);
   }
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetPositionKeys(I3D_anim_pos_tcb *keys, dword num_keys){

   SetupVectorKeys(prs_keys[0], keys, num_keys);
   return I3D_OK;
}

//----------------------------
                              //look-at keys collide with rotation keys,
                              // setting these erases the the others
I3D_RESULT I3D_keyframe_anim::SetLookAtKeys(I3D_anim_look_at_tcb *keys, dword num_keys){

#ifdef USE_ALLOCA
   I3D_anim_pos_tcb *tmp = (I3D_anim_pos_tcb*)alloca(num_keys*sizeof(I3D_anim_pos_tcb));
#else
   I3D_anim_pos_tcb *tmp = new I3D_anim_pos_tcb[num_keys];
#endif
   rot_keys.assign(num_keys);
                              //normalize all look-at keys
   for(int i=0; i<num_keys; i++){
                              //copy look-at
      tmp[i] = keys[i];
      if(!tmp[i].v.IsNull())
         tmp[i].v.Normalize();
                              //copy angle (in power vector)
      rot_keys[i].time = keys[i].time;
      rot_keys[i].axis = keys[i].up_vector;
   }

   SetupVectorKeys(prs_keys[1], tmp, num_keys);
#ifndef USE_ALLOCA
   delete[] tmp;
#endif

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetRotationKeys(I3D_anim_rot *keys, dword num_keys){

   rot_keys.clear();
   prs_keys[1].clear();
   if(!keys || !num_keys) return I3D_OK;

   rot_keys.reserve(num_keys);
   for(int i=0; i<num_keys; i++) rot_keys.push_back(keys[i]);

   base_rot.Make(keys[0].axis, keys[0].angle);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetScaleKeys(I3D_anim_scale_tcb *keys, dword num_keys){

   SetupVectorKeys(prs_keys[2], keys, num_keys);
   return I3D_OK;
}

//----------------------------


I3D_RESULT I3D_keyframe_anim::SetNoteKeys(I3D_anim_note *keys, dword num_keys){

   note_keys.clear();
   if(!keys || !num_keys)
      return I3D_OK;

   note_keys.reserve(num_keys);
   for(int i=0; i<num_keys; i++)
      note_keys.push_back(S_anim_note(keys[i].time, keys[i].note));
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetColorKeys(I3D_anim_color *keys, dword num_keys){

   color_keys.clear();
   if(!keys || !num_keys)
      return I3D_OK;

   color_keys.reserve(num_keys);
   for(int i=0; i<num_keys; i++)
      color_keys.push_back(keys[i]);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetVisibilityKeys(I3D_anim_visibility *keys, dword num_keys){

   vis_keys.clear();
   if(!keys || !num_keys)
      return I3D_OK;

   vis_keys.reserve(num_keys);
   for(int i=0; i<num_keys; i++) vis_keys.push_back(keys[i]);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetPowerKeys(I3D_anim_power *keys, dword num_keys){

   power_keys.clear();
   if(!keys || !num_keys)
      return I3D_OK;

   power_keys.assign(num_keys);
   for(int i=0; i<num_keys; i++){
      power_keys[i] = keys[i];
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_keyframe_anim::SetEndTime(dword e){

   end_time = e;
   return I3D_OK;
}

//----------------------------
//----------------------------

class I3D_interpolator_particle: public I3D_interpolator{
public:
   I3D_interpolator_particle(PI3D_driver d): I3D_interpolator(d)
   {}

   I3DMETHOD(Tick)(int time){
      PI3D_particle prt = (PI3D_particle)((PI3D_frame)frm);
      prt->Tick(Min(2000, time));
      return I3D_OK;
   }

//----------------------------
// Always return 1, since we don't support stages, but always animate particle in one stage.
   virtual int NumActiveStages() const{ return 1; }
};

//----------------------------
                              //interpolator used to update model by Tick-ing it
                              // this interpolator is contained in scenes and models
                              // owning model frams, so that when these are Ticked,
                              // models owned by them are Ticked too
class I3D_interpolator_model: public I3D_interpolator{
public:
   I3D_interpolator_model(PI3D_driver d): I3D_interpolator(d)
   {}
   I3DMETHOD(Tick)(int time){
      PI3D_model mod = I3DCAST_MODEL((PI3D_frame)frm);
      return mod->GetContainer1()->Tick(time);
   }

//----------------------------
// Always return 1, since we don't support stages, but always animate particle in one stage.
   virtual int NumActiveStages() const{ return 1; }
};

//----------------------------
//----------------------------
                              //interpolator of note track - used to work with given animation and time,
                              // calling the registered frame callback proc
                              // with given note and time when it happened in animation
class C_interpolator_note{
                              //index to next key
   int next_key;
public:
   C_interpolator_note():
   next_key(0)
   {}

   void Restart(){ next_key = 0; }

   void Tick(int curr_time, int stage_index, PI3D_frame frm, PI3D_keyframe_anim anim){

      {
         const vector<I3D_keyframe_anim::S_anim_note> &note_keys = anim->GetNoteKeysVector();
         int i = next_key;
         if(i < note_keys.size()){
            while(curr_time >= note_keys[i].time){

               I3D_note_callback nc = { curr_time, note_keys[i].time, stage_index, frm, note_keys[i].note};

               if(frm->GetFrameFlags()&FRMCALLBACK_ANIMNOTE){
                              //notify frame
                  frm->DoCallBack(CBM_ANIMNOTE, (dword)&nc, 0);
               }
                              //notify driver
               PI3D_driver drv = frm->GetDriver1();
               if(drv->cb_proc){
                  drv->cb_proc(CBM_ANIM_NOTE, (dword)&nc, 0, NULL);
               }
               if(++i == note_keys.size())
                  break;
            }
            next_key = i;
         }
      }
   }
};

//----------------------------
//----------------------------
                              //interpolator of visibility
class C_interpolator_visibility{
                              //index to next key
   int next_key;
public:
   C_interpolator_visibility():
   next_key(0)
   {}

   void Restart(){ next_key = 0; }

   void Tick(int curr_time, PI3D_frame frm, PI3D_keyframe_anim anim){

      const vector<I3D_anim_visibility> &vis_keys = anim->GetVisibilityKeysVector();
      int i = next_key;
      if(i<vis_keys.size()){
         while(curr_time >= vis_keys[i].time){
            frm->SetOn(vis_keys[i].on);
            if(++i == vis_keys.size())
               break;
         }
         next_key = i;
      }
   }
};

//----------------------------
//----------------------------
                              //color interpolator - used to work with given animation and time,
                              // applying result to given frame (light only)
class C_interpolator_color{
                              //index to next key
   int next_key;
public:
   C_interpolator_color():
   next_key(0)
   {}

   void Restart(){ next_key = 0; }

   void Tick(int curr_time, PI3D_frame frm, PI3D_keyframe_anim anim){

      const vector<I3D_anim_color> &color_keys = anim->GetColorKeysVector();
      int i = next_key;
      if(i<color_keys.size()){
         bool last = false;
         while(curr_time >= color_keys[i].time){
            if(++i == color_keys.size()){
               last = true;
               break;
            }
         }
         if(i){
            S_vector color;
            if(last) color = color_keys[i-1].color;
            else{
               const I3D_anim_color &key0 = color_keys[i-1],
                  &key1 = color_keys[i];
               const int &t0 = key0.time;
               const int &t1 = key1.time;
               float f = ((float)(curr_time - t0)) / ((float)(t1-t0));
               color = key1.color - key0.color;
               color *= f;
               color += key0.color;
            }
            switch(frm->GetType1()){
            case FRAME_LIGHT:
               I3DCAST_LIGHT(frm)->SetColor(color);
               break;
            }
         }
         next_key = i;
      }
   }
};

//----------------------------
//----------------------------
                              //position/rotation/scale interpolator
class I3D_interpolator_PRS: public I3D_interpolator{

//----------------------------
// Base of any controller processed by this interpolator.
   class C_ctrl_base{
   protected:
                              //index to next key
      int next_key;
                              //easy times (0..1)
      float easy_in_end, easy_out_beg, easy_out_range;
      bool use_ease;          //true to use easiness
   public:

      C_ctrl_base():
      next_key(0)
      {}

//----------------------------
// Setup ease sub-interpolator.
      void EaseSetup(float from, float to){
         use_ease = (from!=0.0f || to!=0.0f);
         if(!use_ease)
            return;

         easy_in_end = from;
         easy_out_range = to;
         float f = easy_in_end + easy_out_range;
         if(f>1.0f){
            easy_in_end /= f;
            easy_out_range /= f;
         }
         easy_out_beg = 1.0f - easy_out_range;
      }

//----------------------------
// Evaluate easiness - convert linear key time (0.0 ... 1.0) to
// easy-interpoated value.
      inline float GetEase(float time) const{
         if(use_ease){
            if(time<easy_in_end){
               return (cos(GetPI() + time/easy_in_end * GetPI() * .5f) + 1.0f) *
                  easy_in_end;
            }else
            if(time>easy_out_beg){
               float out = 1.0f - time;
               return easy_out_beg + easy_out_range *
                  cos(-out/easy_out_range * GetPI() * .5f);
            }
         }
         return time;
      }

//----------------------------
// Restart controller.
      inline void Restart(){ next_key = 0; }
   };

//----------------------------
// Bezier controller, used for position/direction.
   class C_ctrl_bezier: public C_ctrl_base{
   public:
      C_bezier_curve<S_vector> bezier;

      bool Evaluate(const vector<I3D_anim_vector_bezier> &keys, float curr_time, S_vector &v){

         if(next_key >= keys.size())
            return false;

         int i_curr_time = FloatToInt(curr_time-.5f);
         bool overflow = false;
         if(i_curr_time >= keys[next_key].time){
            do{
               if(++next_key == keys.size()){
                  overflow = true;
                  --next_key;
                  break;
               }
            }while(i_curr_time >= keys[next_key].time);

            const I3D_anim_vector_bezier &prev = keys[next_key-1];
            bezier.Init(prev.v, prev.out_tan, keys[next_key].in_tan, keys[next_key].v);
                                    //compute easiness
            EaseSetup(prev.easy_from, keys[next_key].easy_to);
         }else{
            if(!next_key)
               return false;
         }
         float t;
         if(overflow){
            t = 1.0f;
         }else{
            t = (curr_time - (float)keys[next_key-1].time) / (float)(keys[next_key].time - keys[next_key-1].time);
            t = GetEase(t);
         }
         bezier.Evaluate(t, v);

         return true;
      }
   };

//----------------------------
// Rotational controller.
   class C_ctrl_rot: public C_ctrl_base{
   public:
      S_quat q_segment_start, q_segment_end, q_segment_end_180;
      float t_segment_start, rc_t_segment_delta;
      float pow;

      void Restart(const S_quat &base_rot){
         q_segment_start = base_rot;
         q_segment_end = base_rot;
         C_ctrl_base::Restart();
      }

      void Evaluate(const vector<I3D_anim_rot> &keys, const S_quat &base_rot, float curr_time, S_quat &q){

         int num_keys = keys.size();
         assert(next_key<=num_keys);
         if(next_key == num_keys){
                                    //already after end of track, return last segment's end rotation
            q = q_segment_end;
            return;
         }
         int i_curr_time = FloatToInt(curr_time-.5f);
         if(i_curr_time >= keys[next_key].time){
                                    //passed through another key
                                    //find segment in which rotation is now
            do{
               q_segment_start = q_segment_end;
               if(++next_key == num_keys){
                                    //passed through last key
                  q = q_segment_end;
                  return;
               }
                                    //compute rotation at this key
               q_segment_end *= S_quat(keys[next_key].axis, keys[next_key].angle);
            }while(i_curr_time >= keys[next_key].time);

            t_segment_start = keys[next_key-1].time;

            S_quat q(keys[next_key].axis, GetPI());
            pow = keys[next_key].angle / GetPI();
            q_segment_end_180 = q_segment_start * q;
            rc_t_segment_delta = 1.0f / (keys[next_key].time - t_segment_start);
                                    //compute easiness
            EaseSetup(keys[next_key-1].easy_from, keys[next_key].easy_to);
         }
         if(!next_key){
            q = q_segment_start;
            return;
         }
                                    //compute actual key
         float t = (curr_time - t_segment_start) * rc_t_segment_delta;
         t = GetEase(t);
         q = q_segment_start.Slerp(q_segment_end_180, t * pow);
      }
   };

//----------------------------
// Up-vector controller, used for up-vector of direction.
   class C_ctrl_up_dir: public C_ctrl_base{
   public:
      void Evaluate(const vector<I3D_anim_rot> &keys, float curr_time, S_vector &out_up_dir){
         assert(next_key<=keys.size());

         int i_curr_time = FloatToInt(curr_time-.5f);
         bool overflow = false;
         if(i_curr_time >= keys[next_key].time){
            do{
               if(++next_key == keys.size()){
                  overflow = true;
                  --next_key;
                  break;
               }
            }while(i_curr_time >= keys[next_key].time);
         }else
         if(!next_key){
            out_up_dir = keys.front().axis;
            return;
         }

         float t;
         if(overflow){
            t = 1.0f;
         }else{
            t = (curr_time - (float)keys[next_key-1].time) / (float)(keys[next_key].time - keys[next_key-1].time);
         }
         out_up_dir = keys[next_key-1].axis + (keys[next_key].axis - keys[next_key-1].axis) * t;
      }
   };

//----------------------------
// Power controller - used to work with given animation and time,
// applying result to given frame power means different things, depending on frame type
   class C_ctrl_power{
                              //index to next key
      int next_key;
   public:
      C_ctrl_power():
         next_key(0)
      {}

      void Restart(){ next_key = 0; }

      bool Evaluate(const vector<I3D_anim_power> &keys, float curr_time, float &pow){

         bool ok = false;

         int i_curr_time = FloatToInt(curr_time-.5f);

         int i = next_key;
         if(i<keys.size()){
            bool last = false;
            while(i_curr_time >= keys[i].time){
               if(++i == keys.size()){
                  last = true;
                  break;
               }
            }
            if(i){
               if(last)
                  pow = keys[i-1].power;
               else{
                  const I3D_anim_power &key0 = keys[i-1],
                     &key1 = keys[i];
                  const int &t0 = key0.time;
                  const int &t1 = key1.time;
                  float f = (curr_time - (float)t0) / (float)(t1-t0);
                  pow = key1.power - key0.power;
                  pow *= f;
                  pow += key0.power;
               }
               ok = true;
            }
            next_key = i;
         }
         return ok;
      }
   };
                              //results of animation processing
                              // initially values are not valid,
                              // after processing some values may become valid
                              // valid values are applied to the animated frame
   class C_anim_results{
   public:
      S_vector pos, dir, upv;
      S_quat rot;
      float pow;
      bool pos_ok, rot_ok, pow_ok, dir_ok;

      C_anim_results():
         pos_ok(false),
         rot_ok(false),
         pow_ok(false),
         dir_ok(false)
      {}

//----------------------------
// Mix values with another values. 'options' specify mode (combination of I3DANIMOP_? flags).
      void Merge(dword options, float scale_factor, float blend_factor, C_anim_results &results) const{

         bool is_unit_scale = (scale_factor==1.0f);
         if(pos_ok){
            switch(options&I3DANIMOP_POS_MASK){
            case I3DANIMOP_POS_SET:
               results.pos = pos;
               break;
            case I3DANIMOP_POS_ADD:
               if(results.pos_ok)
                  results.pos += is_unit_scale ? pos : pos * scale_factor;
               else
                  results.pos = pos * (scale_factor*blend_factor);
               break;
            case I3DANIMOP_POS_BLEND:
               if(results.pos_ok)
                  results.pos = results.pos * (1.0f-blend_factor) + pos * (blend_factor*scale_factor);
               else
                  results.pos = pos * (blend_factor*scale_factor);
               break;
            default:
               assert(0);
            }
            results.pos_ok = true;
         }
         if(rot_ok){
            switch(options&I3DANIMOP_ROT_MASK){
            case I3DANIMOP_ROT_SET:
               results.rot = rot;
               break;
            case I3DANIMOP_ROT_ADD:
               if(results.rot_ok){
                  if(is_unit_scale)
                     results.rot = results.rot * rot;
                  else{
                     S_quat tmp = rot;
                     tmp.Scale(scale_factor);
                     results.rot = results.rot * tmp;
                  }
               }else{
                  if(is_unit_scale)
                     results.rot = rot;
                  else{
                     results.rot = rot;
                     results.rot.Scale(scale_factor);
                  }
               }
               break;
            case I3DANIMOP_ROT_BLEND:
               if(results.rot_ok){
                  if(is_unit_scale)
                     results.rot = results.rot.Slerp(rot, blend_factor, true);
                  else{
                     S_quat tmp = rot;
                     tmp.Scale(scale_factor);
                     results.rot = results.rot.Slerp(tmp, blend_factor, true);
                  }
               }else{
                  results.rot = rot;
                  if(is_unit_scale)
                     results.rot.Scale(blend_factor);
                  else{
                     results.rot.Scale(scale_factor*blend_factor);
                  }
               }
               break;
            default:
               assert(0);
            }
            results.rot_ok = true;
         }else
         if(dir_ok){
            switch(options&I3DANIMOP_ROT_MASK){
            case I3DANIMOP_ROT_SET:
               results.dir = dir;
               results.upv = upv;
               break;
               /*
            case I3DANIMOP_ROT_ADD:
               break;
            case I3DANIMOP_ROT_BLEND:
               break;
               */
            default:
               assert(0);
            }
            results.dir_ok = true;
         }

         if(pow_ok){
            switch(options&I3DANIMOP_POW_MASK){
            case I3DANIMOP_POW_SET:
               results.pow = pow;
               break;
            case I3DANIMOP_POW_ADD:
               results.pow += pow * scale_factor - 1.0f;
               break;
            case I3DANIMOP_POW_BLEND:
               results.pow = results.pow * (1.0f-blend_factor) + pow * (blend_factor*scale_factor);
               break;
            default:
               assert(0);
            }
            results.pow_ok = true;
         }
      }
   };

                              //stage processing class
                              // stage processing depends on animation type

                              //C_stage is a common base for any stage type
   class C_stage{
   public:
      C_smart_ptr<I3D_animation_base> anim;
      dword options;
      float scale_factor;
      float blend_factor;

                              //run-time:
      bool is_unit_scale, is_unit_blend;  //little optim - set if value is 1.0f

      C_stage(PI3D_animation_base ab):
         anim(ab)
      {
      }

//----------------------------
// Update stage by given time, evaluate values and mix with with current results.
// If the returned value is true, the stage still continues to run.
      virtual bool Evaluate(int time, int stage_index, PI3D_frame,
         dword options, float scale_factor, float blend_factor, C_anim_results&) = 0;

      virtual void Restart(){}
   };

                              //keyframe animation stage processing
   class C_KF_stage: public C_stage{
   public:
      float curr_time, end_time;
      float speed_mult;

      bool use_look_at;
      C_ctrl_bezier ctrl_position, ctrl_direction;
      C_ctrl_rot ctrl_rot;
      C_ctrl_up_dir ctrl_up_dir;
      C_ctrl_power ctrl_power;

      C_interpolator_note i_note;
      C_interpolator_visibility i_visibility;
      C_interpolator_color i_color;

      C_KF_stage(PI3D_animation_base ab):
         C_stage(ab),
         curr_time(0.0f), end_time(0.0f),
         use_look_at(false)
      {
         ctrl_rot.Restart(((PI3D_keyframe_anim)(PI3D_animation_base)anim)->GetBaseRotation());
      }

      virtual void Restart(){
         curr_time = 0.0f;
         if(anim){
            ctrl_position.Restart();
            ctrl_direction.Restart();
            ctrl_rot.Restart(((PI3D_keyframe_anim)(PI3D_animation_base)anim)->GetBaseRotation());
            ctrl_up_dir.Restart();
            ctrl_power.Restart();

            i_note.Restart();
            i_visibility.Restart();
            i_color.Restart();
         }
      }

      virtual bool Evaluate(int time, int stage_index, PI3D_frame frm,
         dword options, float scale_factor, float blend_factor, C_anim_results &results){

         PI3D_keyframe_anim ka = (PI3D_keyframe_anim)(PI3D_animation_base)anim;

         bool done = false;
         curr_time += time * speed_mult;
         if(curr_time >= end_time){
            curr_time = end_time;
            done = true;
         }

         S_vector tmp;
                           //make position
         const vector<I3D_anim_pos_bezier> &pos_keys = ka->GetPositionKeysVector();
         if(ctrl_position.Evaluate(pos_keys, curr_time, tmp)){

            switch(options&I3DANIMOP_POS_MASK){
            case I3DANIMOP_POS_SET:
               results.pos = tmp;
               break;
            case I3DANIMOP_POS_ADD:
               if(results.pos_ok){
                  results.pos += is_unit_scale ? tmp : tmp * scale_factor;
               }else
                  results.pos = is_unit_scale ? tmp : tmp * scale_factor;
               break;
            case I3DANIMOP_POS_BLEND:
               if(results.pos_ok)
                  results.pos = results.pos * (1.0f-blend_factor) + tmp * (blend_factor*scale_factor);
               else
                  results.pos = tmp * (blend_factor*scale_factor);
               break;
            default:
               assert(0);
            }
            results.pos_ok = true;
         }

                        //make rotation
                        // - use either quaternion rotation, or
                        // look-at position (only one of them)
                        // use controller #1 in any case
         if(!use_look_at){
            const vector<I3D_anim_rot> &rot_keys = ka->GetRotationKeysVector();
            if(rot_keys.size()){
               S_quat tmp;
               ctrl_rot.Evaluate(rot_keys, ka->GetBaseRotation(), curr_time, tmp);

               switch(options&I3DANIMOP_ROT_MASK){
               case I3DANIMOP_ROT_SET:
                  results.rot = tmp;
                  break;
               case I3DANIMOP_ROT_ADD:
                  {
                     if(!is_unit_scale)
                        tmp.Scale(scale_factor);
                     if(results.rot_ok)
                        results.rot = results.rot * tmp;
                     else
                        results.rot = tmp;
                  }
                  break;
               case I3DANIMOP_ROT_BLEND:
                  {
                     if(results.rot_ok){
                        if(!is_unit_scale)
                           tmp.Scale(scale_factor);
                        results.rot = results.rot.Slerp(tmp, blend_factor, false);
                     }else{
                        results.rot = tmp;
                        if(!is_unit_scale && !is_unit_blend)
                           results.rot.Scale(scale_factor * blend_factor);
                     }
                  }
                  break;
               default:
                  assert(0);
               }
               results.rot_ok = true;
            }
         }else{
            const vector<I3D_anim_pos_bezier> &dir_keys = ka->GetDirectionKeysVector();
            if(ctrl_direction.Evaluate(dir_keys, curr_time, tmp)){

               switch(options&I3DANIMOP_ROT_MASK){
               case I3DANIMOP_ROT_SET:
                  results.dir = tmp;
                  ctrl_up_dir.Evaluate(ka->GetRotationKeysVector(), curr_time, results.upv);
                  results.dir_ok = true;
                  break;
                  /*
               case I3DANIMOP_ROT_ADD:
                  break;
               case I3DANIMOP_ROT_BLEND:
                  break;
                  */
               default:
                  assert(0);
               }
            }
         }

                        //power
         const vector<I3D_anim_power> &power_keys = ka->GetPowerKeysVector();
         if(power_keys.size()){
            if(ctrl_power.Evaluate(power_keys, curr_time, tmp.x)){

               switch(options&I3DANIMOP_POW_MASK){
               case I3DANIMOP_POW_SET:
                  results.pow = tmp.x;
                  break;
               case I3DANIMOP_POW_ADD:
                  if(results.pow_ok)
                     results.pow += tmp.x * scale_factor - 1.0f;
                  else
                     results.pow = tmp.x * scale_factor;
                  break;
               case I3DANIMOP_POW_BLEND:
                  if(results.pow_ok)
                     results.pow = results.pow * (1.0f - blend_factor) + tmp.x * (blend_factor*scale_factor);
                  else
                     results.pow = tmp.x * (blend_factor*scale_factor);
                  break;
               default:
                  assert(0);
               }
               results.pow_ok = true;
            }
         }
         {
                              //make visibility, color and note controllers
            int i_curr_time = FloatToInt(curr_time-.5f);
            i_visibility.Tick(i_curr_time, frm, ka);
            i_color.Tick(i_curr_time, frm, ka);
            i_note.Tick(i_curr_time, stage_index, frm, ka);
         }

                              //run until time's not over
         return (!done);
      }
   };

                              //pose animation stage processing
   class C_pose_stage: public C_stage{
   public:
      C_pose_stage(PI3D_animation_base ab):
         C_stage(ab)
      {
      }

      virtual bool Evaluate(int time, int stage_index, PI3D_frame frm,
         dword options, float scale_factor, float blend_factor, C_anim_results &results){

         PI3D_anim_pose ap = (PI3D_anim_pose)(PI3D_animation_base)anim;
         dword pose_flags = ap->GetFlags();

         if(pose_flags&APOSF_POS){
            const S_vector &tmp = ap->GetPos1();
            switch(options&I3DANIMOP_POS_MASK){
            case I3DANIMOP_POS_SET:
               results.pos = tmp;
               break;
            case I3DANIMOP_POS_ADD:
               if(results.pos_ok)
                  results.pos += is_unit_scale ? tmp : tmp * scale_factor;
               else
                  results.pos = tmp * (scale_factor*blend_factor);
               break;
            case I3DANIMOP_POS_BLEND:
               if(results.pos_ok)
                  results.pos = results.pos * (1.0f-blend_factor) + tmp * (blend_factor*scale_factor);
               else
                  results.pos = tmp * (blend_factor*scale_factor);
               break;
            default:
               assert(0);
            }
            results.pos_ok = true;
         }
         if(pose_flags&APOSF_ROT){
            const S_quat &tmp = ap->GetRot1();
            switch(options&I3DANIMOP_ROT_MASK){
            case I3DANIMOP_ROT_SET:
               results.rot = tmp;
               break;
            case I3DANIMOP_ROT_ADD:
               if(results.rot_ok){
                  if(is_unit_scale)
                     results.rot = results.rot * tmp;
                  else{
                     S_quat tmp1 = tmp;
                     tmp1.Scale(scale_factor);
                     results.rot = results.rot * tmp1;
                  }
               }else{
                  if(is_unit_scale)
                     results.rot = tmp;
                  else{
                     results.rot = tmp;
                     results.rot.Scale(scale_factor);
                  }
               }
               break;
            case I3DANIMOP_ROT_BLEND:
               if(results.rot_ok){
                  if(is_unit_scale)
                     results.rot = results.rot.Slerp(tmp, blend_factor, true);
                  else{
                     S_quat tmp1 = tmp;
                     tmp1.Scale(scale_factor);
                     results.rot = results.rot.Slerp(tmp1, blend_factor, true);
                  }
               }else{
                  results.rot = tmp;
                  if(is_unit_scale)
                     results.rot.Scale(blend_factor);
                  else{
                     results.rot.Scale(scale_factor*blend_factor);
                  }
               }
               break;
            default:
               assert(0);
            }
            results.rot_ok = true;
         }else
         if(pose_flags&APOSF_DIR){
            const S_vector &dir = ap->GetDir1();
            const S_vector &upv = ap->GetUpVector1();
            switch(options&I3DANIMOP_ROT_MASK){
            case I3DANIMOP_ROT_SET:
               results.dir = dir;
               results.upv = upv;
               break;
               /*
            case I3DANIMOP_ROT_ADD:
               break;
            case I3DANIMOP_ROT_BLEND:
               break;
               */
            default:
               assert(0);
            }
            results.dir_ok = true;
         }

         if(pose_flags&APOSF_POW){
            const float pow = ap->GetPower1();
            switch(options&I3DANIMOP_POW_MASK){
            case I3DANIMOP_POW_SET:
               results.pow = pow;
               break;
            case I3DANIMOP_POW_ADD:
               results.pow += pow * scale_factor - 1.0f;
               break;
            case I3DANIMOP_POW_BLEND:
               results.pow = results.pow * (1.0f-blend_factor) + pow * (blend_factor*scale_factor);
               break;
            default:
               assert(0);
            }
            results.pow_ok = true;
         }
         return true;
      }
   };

                              //dynamic array of stages
                              // inactive stages have NULL pointer and are not processed
                              // the array is kept as small as possible
   typedef map<int, C_stage*> t_stage_cont;
   t_stage_cont stages;

//----------------------------
// Stage is finished with animation
   void FinishStage(t_stage_cont::iterator it, bool &is_running){

                           //at the end - determine what to do next
      C_stage *stg = (*it).second;
      if(stg->options&I3DANIMOP_LOOP){
         stg->Restart();
         is_running = true;
      }else{
         if(frm->GetFrameFlags()&FRMCALLBACK_ANIMFINISH){
            frm->DoCallBack(CBM_ANIM_FINISH, (*it).first, 0);
         }
         if(stg->options&I3DANIMOP_KEEPDONE)
            is_running = true;
      }
   }

//----------------------------
// Evaluate single controller stage. This means merging the stage with provided results.
// Depending on flags, the stage may be mixed directly with results and recursively call this func again,
// or it may mix with next stage first.
// Note about precedence:
//    - all stages are evauated left-to-right by default
//    - if options for specified stage contain number of braces,
//       such number of successive stages are belnded together prior to blending with current results
//    - param 'brace_count' is recursively passed to next stages,
//       so that they know if they should blend with successive stages

   bool EvalSingleStage(t_stage_cont::iterator it, int time, C_anim_results &results, int brace_count = 0){

      C_stage *stg = (*it).second;
      assert(stg && stg->anim);

      int stg_index = (*it).first;
      bool stg_running;
      int new_brace_count = stg->options&I3DANIMOP_BRACE_MASK;
      if(new_brace_count){
         new_brace_count >>= I3DANIMOP_BRACE_SHIFT;
         assert(!brace_count || new_brace_count >= brace_count);

                                 //try to blend with next stage first, then with current results
         t_stage_cont::iterator it_next = it; ++it_next;
         if(it_next!=stages.end()){
            int next_stg_index = (*it_next).first;
            int index_delta = next_stg_index - stg_index;
            if(index_delta <= new_brace_count){
                              //eval this stage fully
               C_anim_results tmp_results;
               stg_running = stg->Evaluate(time, stg_index, frm, I3DANIMOP_SET, 1.0f, 1.0f, tmp_results);
                              //eval next stage(s) with these results
               bool next_running = EvalSingleStage(it_next, time, tmp_results, new_brace_count - index_delta + 1);
               if(!next_running)
                  FinishStage(it_next, next_running);
               (byte&)stg_running |= (byte)next_running;
                              //now merge bracketed results with current ones
               tmp_results.Merge(stg->options, stg->scale_factor, stg->blend_factor, results);
                              //make sure we skip stages already blended
               it = it_next;
               while(it!=stages.end() && (*it).first < stg_index+index_delta)
                  ++it;
            }else{
                              //eval normally (next stage is out of our brackets)
               stg_running = stg->Evaluate(time, stg_index, frm, stg->options, stg->scale_factor, stg->blend_factor, results);
            }
         }else{
                              //eval normally (this is last stage)
            stg_running = stg->Evaluate(time, stg_index, frm, stg->options, stg->scale_factor, stg->blend_factor, results);
         }
         if(brace_count)
            brace_count = Max(0, brace_count - new_brace_count);
      }else{
         stg_running = stg->Evaluate(time, stg_index, frm, stg->options, stg->scale_factor, stg->blend_factor, results);
      }
      if(!stg_running)
         FinishStage(it, stg_running);
                              //make next stage, if not at the end
      ++it;
      if(it!=stages.end()){
         if(brace_count){
                              //check if we'are allowed to blend with next (precedence)
            int index_delta = (*it).first - stg_index;
            if(index_delta < brace_count){
               bool next_running = EvalSingleStage(it, time, results, brace_count-index_delta);
               (byte&)stg_running |= (byte)next_running;
            }
         }else{
            bool next_running = EvalSingleStage(it, time, results);
            (byte&)stg_running |= (byte)next_running;
         }
      }
      return stg_running;
   }

//----------------------------
public:
   I3D_interpolator_PRS(PI3D_driver d1):
      I3D_interpolator(d1)
   {
   }

   ~I3D_interpolator_PRS(){
                              //delete all allocated stages
      for(t_stage_cont::iterator it = stages.begin(); it!=stages.end(); it++){
         delete (*it).second;
      }
   }
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

//----------------------------

   I3D_RESULT I3DAPI Tick(int time){

      if(!frm || !stages.size())
         return I3DERR_NOTINITIALIZED;

      bool upd = false;
                                 //since callback gives user possibility to
                                 // release the frame, we must hold ref to our interface
      //AddRef();
      ++ref;

      C_anim_results results;
      bool running = EvalSingleStage(stages.begin(), time, results);

      if(results.pos_ok){
         frm->SetPos(results.pos);
         upd = true;
      }
      if(results.rot_ok){
         frm->SetRot(results.rot);
         upd = true;
      }
      if(results.dir_ok){
         assert(!results.rot_ok);
         frm->SetDir1(results.dir, results.upv);
         upd = true;
      }
      if(results.pow_ok){
         switch(frm->GetType1()){
         case FRAME_LIGHT:
            I3DCAST_LIGHT(frm)->SetPower(results.pow);
            break;
         case FRAME_SOUND:
            I3DCAST_SOUND(frm)->SetVolume(results.pow);
            break;
         default: 
            frm->SetScale(results.pow);
            upd = true;
         }
      }

      if(upd)
         frm->Update();

      if(!--ref)
         delete this;
      return (running ? I3D_OK : I3D_DONE);
   }

//----------------------------

   I3DMETHOD(Reset)(dword si){
      assert(0);
      return I3DERR_UNSUPPORTED;
   }

   I3DMETHOD(SetAnimation)(dword stage, PI3D_animation_base, dword options, float scale_factor,
      float blend_factor, float time_scale);

   I3DMETHOD(SetScaleFactor)(dword si, float factor){
      t_stage_cont::iterator it = stages.find(si);
      if(it==stages.end())
         return I3DERR_OBJECTNOTFOUND;
      C_stage *stg = (*it).second;
      stg->scale_factor = factor;
      stg->is_unit_scale = (factor==1.0f);
      return I3D_OK;
   }

   I3DMETHOD(SetBlendFactor)(dword si, float factor){
      t_stage_cont::iterator it = stages.find(si);
      if(it==stages.end())
         return I3DERR_OBJECTNOTFOUND;
      C_stage *stg = (*it).second;
      stg->blend_factor = factor;
      stg->is_unit_blend = (factor==1.0f);
      return I3D_OK;
   }

   I3DMETHOD_(I3D_RESULT,SetSpeed)(dword si, float speed){
      t_stage_cont::iterator it = stages.find(si);
      if(it==stages.end())
         return I3DERR_OBJECTNOTFOUND;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return I3DERR_INVALIDPARAMS;
      ((C_KF_stage*)(*it).second)->speed_mult = speed;
      return I3D_OK;
   }

   I3DMETHOD(SetEndTime)(dword si, dword time){ 
      t_stage_cont::iterator it = stages.find(si);
      if(it==stages.end())
         return I3DERR_OBJECTNOTFOUND;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return I3DERR_INVALIDPARAMS;
      ((C_KF_stage*)(*it).second)->end_time = time;
      return I3D_OK; 
   }

   I3DMETHOD(SetCurrTime)(dword si, dword time){
      t_stage_cont::iterator it = stages.find(si);
      if(it==stages.end())
         return I3DERR_OBJECTNOTFOUND;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return I3DERR_INVALIDPARAMS;
      ((C_KF_stage*)(*it).second)->curr_time = time;
      return I3D_OK;
   }

   I3DMETHOD_(PI3D_animation_base,GetAnimation)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return NULL;
      return (*it).second->anim;
   }

   I3DMETHOD_(dword,GetOptions)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0;
      return (*it).second->options;
   }

   I3DMETHOD_(float,GetScaleFactor)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0.0f;
      return (*it).second->scale_factor;
   }

   I3DMETHOD_(float,GetBlendFactor)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0.0f;
      return (*it).second->blend_factor;
   }

   I3DMETHOD_(float,GetSpeed)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0.0f;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return 0.0f;
      return ((C_KF_stage*)(*it).second)->speed_mult;
   }

   I3DMETHOD_(dword,GetEndTime)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return 0;
      return FloatToInt(((C_KF_stage*)(*it).second)->end_time);
   }

   I3DMETHOD_(dword,GetCurrTime)(dword si) const{
      t_stage_cont::const_iterator it = stages.find(si);
      if(it==stages.end())
         return 0;
      if((*it).second->anim->GetType1()!=I3DANIM_KEYFRAME)
         return 0;
      return FloatToInt(((C_KF_stage*)(*it).second)->curr_time);
   }
public:
   virtual int NumActiveStages() const{
      return stages.size();
   }

   virtual bool IsPRS() const{ return true; }
};

//----------------------------

I3D_RESULT I3D_interpolator_PRS::SetAnimation(dword si, PI3D_animation_base ap,
   dword options, float scale_factor, float blend_factor, float speed){

   if(ap) ap->AddRef();

   t_stage_cont::iterator it = stages.find(si);
   if(it!=stages.end()){
      delete (*it).second;
      stages.erase(it);
   }

   if(ap){
      C_stage *stg_base;
      switch(ap->GetType1()){
      case I3DANIM_KEYFRAME:
         {
            C_KF_stage *stg = new C_KF_stage(ap);
            PI3D_keyframe_anim ka = (PI3D_keyframe_anim)ap;
            stg->end_time = ka->end_time;
            stg->use_look_at = (ka->prs_keys[1].size() != 0);
            stg->speed_mult = speed;
            stg_base = stg;
         }
         break;
      case I3DANIM_POSE:
         {
            C_pose_stage *stg = new C_pose_stage(ap);
            stg_base = stg;
         }
         break;
      default: assert(0);
      }
      stg_base->options = options;
      stg_base->scale_factor = scale_factor;
      stg_base->is_unit_scale = (scale_factor==1.0f);
      stg_base->blend_factor = blend_factor;
      stg_base->is_unit_blend = (blend_factor==1.0f);
      stages[si] = stg_base;
   }
   if(ap) ap->Release();
   return I3D_OK;
}

//----------------------------
//----------------------------
                              //animation set
I3D_animation_set::I3D_animation_set(PI3D_driver d):
   ref(1),
   drv(d)
{
   drv->AddCount(I3D_CLID_ANIMATION_SET);
}

//----------------------------

I3D_animation_set::~I3D_animation_set(){
   drv->DecCount(I3D_CLID_ANIMATION_SET);
}

//----------------------------

void I3D_animation_set::AddAnimation1(PI3D_animation_base ap, const C_string &str){

   anims.push_back(S_anim_link());
   anims.back().anim = ap;
   anims.back().link = str;
}

//----------------------------

I3D_RESULT I3D_animation_set::DeleteAnimation(PI3D_animation_base ap){

   for(int i=0; i<anims.size(); i++)
   if(anims[i].anim == ap){
      anims[i]=anims.back(); anims.pop_back();
      return I3D_OK;
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

dword I3D_animation_set::GetTime() const{
                              //time of anims is synced to max of them
                              //so return time of 1st anim
   if(!anims.size())
      return 0;
   PI3D_animation_base ab = anims.front().anim;
   if(ab->GetType1()!=I3DANIM_KEYFRAME)
      return 0;
   return ((PI3D_keyframe_anim)ab)->GetEndTime();
}

//----------------------------
//----------------------------

PI3D_interpolator I3D_driver::CreateInterpolator(){
   
   return new I3D_interpolator_PRS(this);
}

//----------------------------
//----------------------------

I3D_RESULT I3D_container::AddInterpolator(PI3D_interpolator ip){

                              //don't need the same interpolator twice in list
   if(intps.size()){
      int i = FindPointerInArray((void**)&intps.front(), intps.size(), ip);
      if(i!=-1)
         return I3DERR_INVALIDPARAMS;
   }
   intps.push_back(ip);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_container::DeleteInterpolator(PI3D_interpolator ip){

   for(int i=intps.size(); i--; )
   if(intps[i]==ip){
      intps[i] = intps.back(); intps.pop_back();
      return I3D_OK;
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

I3D_RESULT I3D_container::AddAnimFrame(PI3D_frame frm){

   DeleteAnimFrame(frm);

   PI3D_interpolator ib;
   switch(frm->GetType1()){
   case FRAME_VISUAL:
      if(!(I3DCAST_VISUAL((PI3D_frame)frm)->GetVisualType()==I3D_VISUAL_PARTICLE))
         return I3DERR_INVALIDPARAMS;
      ib = new I3D_interpolator_particle(frm->GetDriver1());
      break;

   case FRAME_MODEL:
      ib = new I3D_interpolator_model(frm->GetDriver1());
      break;

   default:
      return I3DERR_INVALIDPARAMS;
   }
   intps.push_back(ib);
   ib->SetFrame(frm);
   ib->Release();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_container::DeleteAnimFrame(PI3D_frame frm){

   for(int i=intps.size(); i--; ){
      if(intps[i]->GetFrame()==frm){
         intps[i] = intps.back(); intps.pop_back();
         return I3D_OK;
      }
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

I3D_RESULT I3D_container::Tick(int time){

   int i;
   int sz;
   for(i=0; i<(sz=intps.size()); i++){
      PI3D_interpolator *ip = (PI3D_interpolator*)intps.begin();
      PI3D_interpolator intp = ip[i];
#ifdef USE_PREFETCH
      if(i<sz-1){
         Prefetch1(ip[i+1]);
      }
#endif
      I3D_RESULT ir = intp->Tick(time);
      if(ir==I3D_DONE){
         if(i<intps.size() && intp==intps[i]){
            intps[i]=intps.back(); intps.pop_back();
            --i;
         }
      }
   }
   if(!NumInterpolators1())
      return I3D_DONE;
   return I3D_OK;
}

//----------------------------
//----------------------------
