/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.
   All rights reserved.

   File: Camera.cpp
   Content: Camera frame.
--------------------------------------------------------*/

#include "all.h"
#include "camera.h"
#include "scene.h"

//----------------------------
                              //camera class
I3D_camera::I3D_camera(PI3D_driver d):
   I3D_frame(d),
   fov_a(PI*.5f),
   ncp(1.0f), fcp(1000.0f),
   ortho_scale(1.0f),
   cam_flags(CAMF_RESET_PROJ | CAMF_RESET_VIEW)
{
   drv->AddCount(I3D_CLID_CAMERA);
   curr_sector = NULL;
   type = FRAME_CAMERA;
   enum_mask = ENUMF_CAMERA;
   m_view.Identity();
}

//----------------------------

I3D_camera::~I3D_camera(){
   drv->DecCount(I3D_CLID_CAMERA);
}

//----------------------------

I3D_RESULT I3D_camera::Duplicate(CPI3D_frame frm){

   if(frm->GetType1()==FRAME_CAMERA){
      PI3D_camera cp = I3DCAST_CAMERA(frm);
      fov_a = cp->fov_a;
      ncp = cp->ncp;
      fcp = cp->fcp;
      ortho_scale = cp->ortho_scale;
      SetOrthogonal(cp->GetOrthogonal());

      cam_flags |= CAMF_RESET_PROJ | CAMF_RESET_VIEW;
   }
   return I3D_frame::Duplicate(frm);
}

//----------------------------

I3D_RESULT I3D_camera::SetFOV(float angle){

   if(IsAbsMrgZero(angle) || angle<=0.0f)
      return I3DERR_INVALIDPARAMS;
   fov_a = angle;
   if(!(cam_flags&CAMF_ORTHOGONAL)) cam_flags |= CAMF_RESET_PROJ;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_camera::SetRange(float n, float f){

   if((f-n)<.001f)
      return I3DERR_INVALIDPARAMS;
   ncp = n;
   fcp = f;
   cam_flags |= CAMF_RESET_PROJ;
   return I3D_OK;
}

//----------------------------

void I3D_camera::Update(){

   I3D_frame::Update();
   cam_flags |= CAMF_RESET_VIEW;
}

//----------------------------

void I3D_camera::GetRange(float &n, float &f) const{
   n = ncp;
   f = fcp;
}

//----------------------------

const S_matrix &I3D_camera::GetViewMatrix() const{ return GetViewMatrix1(); }

//----------------------------

const S_matrix &I3D_camera::GetProjectionMatrix() const{ return m_proj; }

//----------------------------

I3D_sector *I3D_camera::GetSector() const{ return curr_sector; }

//----------------------------

I3D_RESULT I3D_camera::SetOrthoScale(float scale){

   if(scale <= 0.0f)
      return I3DERR_INVALIDPARAMS;
   ortho_scale = scale;
   if(cam_flags&CAMF_ORTHOGONAL)
      cam_flags |= CAMF_RESET_PROJ;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_camera::SetOrthogonal(bool b){

   if(b){
      if(!(cam_flags&CAMF_ORTHOGONAL))
         cam_flags |= (CAMF_ORTHOGONAL|CAMF_RESET_PROJ);
   }else{
      if(cam_flags&CAMF_ORTHOGONAL){
         cam_flags &= ~CAMF_ORTHOGONAL;
         cam_flags |= CAMF_RESET_PROJ;
      }
   }
   return I3D_OK;
}

//----------------------------

void I3D_camera::UpdateMatrices(float aspect_ratio){

   if(cam_flags&CAMF_RESET_PROJ){
                              //critical error
      if(IsAbsMrgZero(fov_a))
         return;

      m_proj.Zero();

      if(!(cam_flags&CAMF_ORTHOGONAL)){
                              //perspective projection matrix
         float fov_a_half = fov_a * .5f;
         float ca = cos(fov_a_half), sa = sin(fov_a_half);
         float w = ca / sa;
         float h = ca / (sa * aspect_ratio);
         float Q = fcp / (fcp - ncp);
         m_proj(0, 0) = w;
         m_proj(1, 1) = h;
         m_proj(2, 2) = Q;
         m_proj(2, 3) = 1.0f;
         m_proj(3, 2) = -Q * ncp;
      }else{
                              //orthogonal projection matrix
         m_proj(0, 0) = ortho_scale;
         m_proj(1, 1) = ortho_scale / aspect_ratio;
         m_proj(2, 2) = 1.0f / (fcp - ncp);
         m_proj(3, 2) = -ncp / (fcp - ncp);
         m_proj(3, 3) = 1.0f;
      }

      cam_flags &= ~CAMF_RESET_PROJ;
   }

   if((cam_flags&CAMF_RESET_VIEW) || ChangedMatrix()){
      const S_matrix &mat = GetMatrix();
                              //direct inverse frame matrix
      m_view(0, 0) = mat(0, 0);
      m_view(0, 1) = mat(1, 0);
      m_view(0, 2) = mat(2, 0);

      m_view(1, 0) = mat(0, 1);
      m_view(1, 1) = mat(1, 1);
      m_view(1, 2) = mat(2, 1);

      m_view(2, 0) = mat(0, 2);
      m_view(2, 1) = mat(1, 2);
      m_view(2, 2) = mat(2, 2);
                              //rotate position by 3x3 part of view matrix
      m_view(3) = -(mat(3).RotateByMatrix(m_view));

      cam_flags &= ~CAMF_RESET_VIEW;
   }
}

//----------------------------

void I3D_camera::Mirror(const S_plane &pl){

   GetMatrix();
   S_matrix &m = matrix;
                              //make mirror
   m(0) = m(0) - pl.normal * (m(0).Dot(pl.normal) * 2.0f);
   m(1) = m(1) - pl.normal * (m(1).Dot(pl.normal) * 2.0f);
   m(2) = m(2) - pl.normal * (m(2).Dot(pl.normal) * 2.0f);
   m(3) = m(3) - pl.normal * (m(3).DistanceToPlane(pl) * 2.0f);

}

//----------------------------

void I3D_camera::Draw1(PI3D_scene scene, bool strong) const{

   byte alpha(strong ? 0 : 0xa0);

                              //draw cone
   {
      scene->SetRenderMatrix(GetMatrix());

      S_vector v1;
      float f = fov_a*.5f;
      v1.x = sin(f);
      v1.z = cos(f);
      v1.y = v1.x * .75f;
      v1 *= 2.0f / v1.Magnitude();
      S_vector v[5];
      v[0].Zero();
      v[1] = v1;
      v[2] = v1, v[2].y = -v[2].y;
      v[3] = v1, v[3].x = -v[3].x;
      v[4] = v[3], v[4].y = -v[4].y;
      static const word indx[] = {0, 1, 0, 2, 0, 3, 0, 4, 1, 2, 2, 4, 4, 3, 3, 1};

      static const S_vector color(.8f, .8f, .8f);
      scene->DrawLines(v, 5, indx, sizeof(indx)/sizeof(word), 
         color, sizeof(S_vector), alpha);
   }
   scene->DrawIcon(GetWorldPos(), 9, alpha);
}

//----------------------------

I3D_RESULT I3D_camera::DebugDraw(PI3D_scene scene) const{

   Draw1(scene, true);
   return I3D_OK;
}

//----------------------------
//----------------------------
