/*--------------------------------------------------------
   Copyright (c) 2000 Lonely Cat Games.
   All rights reserved.

   File: Mirror.cpp
   Content: Mirror frame.
--------------------------------------------------------*/

#include "all.h"
#include "object.h"
#include "camera.h"

//----------------------------
//----------------------------

class I3D_object_camview: public I3D_object{
   C_smart_ptr<I3D_texture> texture;
   C_smart_ptr<IDirect3DSurface8> zbuf;
   C_smart_ptr<I3D_camera> cam;

   float alpha;
   dword alpha_factor;
   C_string cam_name;
   dword txt_x, txt_y;

//----------------------------

   void SetTextureSize(dword x, dword y){

      if(txt_x!=x || txt_y!=y){
         txt_x = x;
         txt_y = y;
         texture = NULL;
         zbuf = NULL;
      }
   }

//----------------------------

   void SetCameraName(const char *name){

      cam_name = name;
      cam = NULL;
   }

//----------------------------

   void CreateRenderTarget(){

      texture = NULL;
      zbuf = NULL;

      PI3D_texture tp;
      I3D_CREATETEXTURE ct;
      memset(&ct, 0, sizeof(ct));
      ct.flags = TEXTMAP_HINTDYNAMIC | TEXTMAP_RENDERTARGET;
      ct.size_x = txt_x;
      ct.size_y = txt_y;

      if(I3D_SUCCESS(drv->CreateTexture(&ct, &tp))){
         texture = tp;
         tp->Release();

         D3DFORMAT fmt = drv->FindZStenFormat(texture->GetSysmemImage()->GetPixelFormat());
         if(fmt==D3DFMT_UNKNOWN){
            texture = NULL;
         }else{
            LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
            HRESULT hr = lpDev8->CreateDepthStencilSurface(texture->SizeX(), texture->SizeY(), fmt,
               D3DMULTISAMPLE_NONE, (IDirect3DSurface8**)&zbuf);
            if(FAILED(hr)){
               texture = NULL;
            }
         }
      }
   }

//----------------------------

   void ResetCallback(I3D_driver::E_RESET_MSG msg){

      switch(msg){
      case I3D_driver::RESET_RELEASE:
         texture = NULL;
         zbuf = NULL;
         break;
      }
   }

   static void ResetCallback_thunk(I3D_driver::E_RESET_MSG msg, void *c){ ((I3D_object_camview*)c)->ResetCallback(msg); }

public:
   I3D_object_camview(PI3D_driver d):
      I3D_object(d),
      alpha(.5f), alpha_factor(128),
      txt_x(256), txt_y(256)
   {
      visual_type = I3D_VISUAL_CAMVIEW;
      drv->RegResetCallback(ResetCallback_thunk, this);
   }

   ~I3D_object_camview(){
      drv->UnregResetCallback(ResetCallback_thunk, this);
   }


   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

//----------------------------

   I3DMETHOD(Duplicate)(CPI3D_frame frm){

      if(frm->GetType1()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==GetVisualType()){
         I3D_object_camview *cv = (I3D_object_camview*)frm;
         alpha = cv->alpha;
         alpha_factor = cv->alpha_factor;
         SetCameraName(cv->cam_name);
         SetTextureSize(cv->txt_x, cv->txt_y);
      }
      return I3D_object::Duplicate(frm);
   }

//----------------------------

   I3DMETHOD(SetProperty)(dword index, dword value){

      switch(index){
      case 0:
         alpha = IntAsFloat(value);
         alpha_factor = Max(0, Min(255, FloatToInt(alpha*255.0f)));
         break;
      case 1: SetTextureSize(value, txt_y); break;
      case 2: SetTextureSize(txt_x, value); break;
      case 3: SetCameraName((const char*)value); break;
      default:
         return I3DERR_INVALIDPARAMS;
      }
      return I3D_OK;
   }

//----------------------------

   I3DMETHOD_(dword,GetProperty)(dword index) const{
      
      switch(index){
      case 0: return FloatAsInt(alpha);
      case 1: return txt_x;
      case 2: return txt_y;
      case 3: return (dword)(const char*)cam_name;
      }
      return 0xcdcdcdcd;
   }
};

//----------------------------

extern const dword vs_decl_vertex_mesh[];

//----------------------------

void I3D_object_camview::AddPrimitives(S_preprocess_context &pc){

   if(pc.mode==RV_SHADOW_CASTER)
      return;

   PI3D_mesh_base mb = GetMesh();
   if(!mb)
      return;
   assert(mb);
   const I3D_face_group &fg = mb->GetFGroupVector().front();
   dword vertex_count = mb->vertex_buffer.NumVertices();
   assert(vertex_count);
   word face_count = fg.NumFaces1();
   drv->render_stats.triangle += face_count;
   drv->render_stats.vert_trans += vertex_count;

   int curr_auto_lod = drv->force_lod_index;
   const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
   if(curr_auto_lod==-1 && auto_lods.size()){
                           //compute automatic LOD based on distance
      curr_auto_lod = GetAutoLOD(pc, &auto_lods[0].ratio, sizeof(C_auto_lod),
         auto_lods.size(), mb->GetAutoLODDists());
   }else{
      curr_auto_lod = Min(curr_auto_lod, (int)auto_lods.size()-1);
   }

   pc.prim_list.push_back(S_render_primitive(curr_auto_lod));
   S_render_primitive &p = pc.prim_list.back();
   
   p.vis = this;
   p.clip = pc.clip;
   p.sector = pc.curr_sector;
   p.scene = pc.scene;
                              
   p.sort_value = PRIM_SORT_OPAQUE;
   ++pc.opaque;

   if(!vertex_buffer.GetD3DVertexBuffer())
      PrepareDestVB(mb);
}

//----------------------------

void I3D_object_camview::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   I3D_mesh_base *mb = mesh->GetLOD1();

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   bool cam_mode = (pc.mode==RV_NORMAL);

   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      if(pc.mode==RV_NORMAL && cam_name.Size()){
         if(!texture){
            CreateRenderTarget();
            if(!texture){
               drv->DEBUG(C_fstring("CamView '%s': can't create render target", (const char*)name), 2);
            }
         }
         if(!cam){
            cam = I3DCAST_CAMERA(pc.scene->FindFrame(cam_name, ENUMF_CAMERA));
            if(!cam){
               drv->DEBUG(C_fstring("CamView '%s': can't find camera '%s'", (const char*)name, (const char*)cam_name), 2);
            }
         }
         if(cam){
                              //render view to the texture
            if(texture){
               C_smart_ptr<IDirect3DSurface8> save_rt, save_zs;
               drv->SaveRenderTarget(save_rt, save_zs);

               IDirect3DSurface8 *rt;
               hr = texture->GetD3DTexture()->GetSurfaceLevel(0, &rt);
               CHECK_D3D_RESULT("GetSurfaceLevel", hr);
               drv->SetRenderTarget(rt, zbuf);
               rt->Release();

               dword vp[4];
               pc.scene->GetViewport(vp[0], vp[1], vp[2], vp[3]);
               pc.scene->SetViewport(0, 0, texture->SizeX(), texture->SizeY());
               C_smart_ptr<I3D_camera> cam_save = pc.scene->GetActiveCamera1();
               pc.scene->SetActiveCamera(cam);
               pc.scene->RenderView(pc.render_context, pc.render_flags, RV_CAM_TO_TEXTURE);

                                 //restore back
               pc.scene->SetActiveCamera(cam_save);
               pc.scene->SetMViewProj(cam_save->GetViewMatrix1(), cam_save->GetProjectionMatrix1());
               hr = lpDev8->SetTransform(D3DTS_PROJECTION, (const D3DMATRIX*)&cam_save->GetProjectionMatrix1());
               CHECK_D3D_RESULT("SetTransform", hr);
               drv->SetRenderTarget(save_rt, save_zs);
               const int *vport = pc.scene->GetViewport1();
               pc.scene->SetViewport(vp[0], vp[1], vp[2], vp[3]);
               drv->UpdateViewport(vport[0], vport[1], vport[2], vport[3]);
            }
         }else{
            cam_mode = false;
         }
      }

      dest_vb_trans_id = pc.render_view_id;

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }

      dword vertex_count;
      const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
      if(rp.curr_auto_lod < 0){
         vertex_count = mb->vertex_buffer.NumVertices();
      }else{
         vertex_count = auto_lods[rp.curr_auto_lod].vertex_count;
      }

                              //transform vertices
      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX;

      if(drv->IsDirectTransform()){
         if(pc.mode!=RV_SHADOW_CASTER)
            prep_flags |= VSPREP_MAKELIGHTING;
         prep_flags |= VSPREP_COPY_UV;
      }else{
         if(pc.mode!=RV_SHADOW_CASTER){
#ifdef DEBUG_NO_CACHE_LIGHT
            vis_flags &= ~VISF_DEST_LIGHT_VALID;  
#endif
            if(!(vis_flags&VISF_DEST_LIGHT_VALID)){
               prep_flags |= VSPREP_MAKELIGHTING;
               vis_flags |= VISF_DEST_LIGHT_VALID;
            }
         }
         if(!(vis_flags&VISF_DEST_UV0_VALID)){
            prep_flags |= VSPREP_COPY_UV;
            vis_flags |= VISF_DEST_UV0_VALID;
         }
      }
      if(pc.mode==RV_SHADOW_CASTER)
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;

      I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);
      PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      if(!drv->IsDirectTransform()){
         drv->SetClipping(rp.clip);
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_mesh));
         hr = lpDev8->ProcessVertices(mb->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }

   if(!drv->IsDirectTransform()){
      drv->SetVertexShader(vertex_buffer.GetFVFlags());
   }
   drv->SetClipping(rp.clip);

   const I3D_face_group *fg;
   int vertex_count;
   int base_index;
   if(rp.curr_auto_lod < 0){
                              //direct mesh
      fg = &mb->GetFGroupVector()[0];
      vertex_count = mb->vertex_buffer.NumVertices();
      base_index = mb->index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(mb->index_buffer.GetD3DIndexBuffer(),
         vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }else{
      const C_auto_lod &al = mb->GetAutoLODs()[rp.curr_auto_lod];
      fg = &al.fgroups[0];
      vertex_count = al.vertex_count;
      base_index = al.index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(al.index_buffer.GetD3DIndexBuffer(),
         vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }

   PI3D_material mat = fg->GetMaterial1();
   if(pc.mode!=RV_SHADOW_CASTER){
      if(drv->last_blend_flags!=I3DBLEND_OPAQUE)
         drv->SetupBlend(I3DBLEND_OPAQUE);
      drv->SetupTextureStage(0, D3DTOP_SELECTARG1);
                           //apply environment effect
      E_ENVIROMENTAL_EFFECT ee = mat->GetEnvironmetEffect();
      if(ee!=EE_NULL && envmap_uv_set!=-1)
         SetupEnvMapping(ee, mat, envmap_uv_set);
      else
         drv->DisableTextureStage(1);
      drv->EnableFog(vis_flags&VISF_FOG_VALID);
      drv->EnableAnisotropy(0, true);
   }else{
      drv->SetTexture1(0, NULL);
   }

                           //debug mode - without textures
   if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES))
      drv->SetTexture1(0, NULL);

   drv->SetStreamSource(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.GetSizeOfVertex());

   if(cam_mode){
      bool is_wire = (drv->GetFlags()&DRVF_WIREFRAME);
      if(is_wire) lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

      int stage = 0;
      PI3D_texture_base txt_bump = mat->GetTexture1(MTI_BUMP);
      if(txt_bump){
         drv->SetTextureCoordIndex(1, 0);
         drv->SetupTextureStage(stage, D3DTOP_BUMPENVMAP);
         drv->SetBumpScale(stage, mat->GetBumpOpacity());
         drv->SetTexture1(stage++, txt_bump);
      }
      drv->SetTexture1(stage, texture);
      drv->SetupBlend(I3DBLEND_OPAQUE);
      drv->SetupTextureStage(0, D3DTOP_SELECTARG1);

      drv->DisableTextureStage(stage+1);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
         (base_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);

      drv->SetTextureCoordIndex(1, 1);
      drv->DisableTextureStage(1);
      if(is_wire) lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

                              //re-render with original texture
      if(mat->GetTexture1(MTI_DIFFUSE)){
         drv->SetRenderMat(mat);

         dword blend_mode = I3DBLEND_ALPHABLEND | I3DBLEND_ALPHAFACTOR;
         if(mat->IsDiffuseAlpha()) 
            blend_mode |= I3DBLEND_VERTEXALPHA;
         drv->SetupBlend(blend_mode);
         drv->SetupTextureStage(0, drv->vertex_light_blend_op);
         drv->SetTextureFactor(alpha_factor<<24);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, fg->num_faces);
         CHECK_D3D_RESULT("DrawIndexedPrimitive", hr);
      }
   }else{
      drv->SetRenderMat(mat);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
         (base_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);
   }
}

//----------------------------

static const S_visual_property props[] = {
   I3DPROP_FLOAT, "Alpha",
   I3DPROP_INT, "Texture X",
   I3DPROP_INT, "Texture Y",
   I3DPROP_STRING, "Camera name",
};

static struct S_auto_init_camview{
   static I3D_visual *CreateCamView(PI3D_driver drv){
      return new I3D_object_camview(drv);
   }
   S_auto_init_camview(){
      RegisterVisual(CreateCamView, I3D_VISUAL_CAMVIEW, "Camera View",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_camview;

//----------------------------
//----------------------------


