
/*--------------------------------------------------------
   Copyright (c) 2001 LonelyCatGames.
   All rights reserved.
   Author: J.Vrubel

   File: Collision.cpp
   Content: Collision testing functions, combine bsp tree for static geometry and hiearchy 
            volumes for dynamic actors.
--------------------------------------------------------*/

#include "all.h"
#include "scene.h"
#include "volume.h"           
#include "bsp.h"
#include "visual.h"
                              //for edit frames
#include "target.h"
#include "occluder.h"
#include "dummy.h"
#include "joint.h"

                              //minimal count of volumes neaded for create hiearchy autobox
#define DYNAMIC_VOLUME_BOUND_LIMIT 2
//#define DYNAMIC_VOLUME_BOUND_LIMIT 3

//#define SPHERE_VOLUME_DETECT_INSIDE    //detect line-sphere collision from inside of sphere
//#define BOX_VOLUME_DETECT_INSIDE    //detect line-box collision from inside of sphere
#define VOLUME_STATISTICS

                              //DEBUG:
#ifdef _DEBUG
//#define PERFORMANECE_COUNT 5000
                              //global I3D_driver variable for debug output in response functions
CPI3D_driver log_drv;

//#define COL_DEBUG(n) assert(log_drv); log_drv->DEBUG(n)
#define COL_DEBUG(n) 
//#define S_B_DEBUG_DYN(n) log_drv->DEBUG(n)
#define S_B_DEBUG_DYN(n) 
//#define S_R_DEBUG_DYN(n) log_drv->DEBUG(n)
#define S_R_DEBUG_DYN(n)
//#define S_S_DEBUG_DYN(n) log_drv->DEBUG(n)
#define S_S_DEBUG_DYN(n)

#else
#define COL_DEBUG(n)
#define S_B_DEBUG_DYN(n)
#define S_R_DEBUG_DYN(n)
#define S_S_DEBUG_DYN(n)

#endif //_DEBUG


//----------------------------
//----------------------------

bool pickBBoxTest(const S_matrix &mat, const I3D_bbox &bb1,
   const S_vector &from1, const S_vector &norm_dir1, const float len1,
   float *hit_dist_ret){

                              //consider we're in the box
   bool in = true;
                              //cached distances, keeping min/max extremes
   float hmin = 0.0f;
   float lmax = 1.0e+10f;
                              //rotate the bounding box by matrix
   I3D_bbox bbox;
   bbox.min = bb1.min * mat;
   bbox.max = bb1.max * mat;
                              //check all 3 axes
   for(int j=0; j<3; j++){
      float d_min =  mat(j).Dot(bbox.min);
      float d_max = -mat(j).Dot(bbox.max);

                              //distances in which our ray collides with
                              //min/max sides of box
      float u_min, u_max;

      float dot_pos = mat(j).Dot(from1);

      u_max = -(dot_pos + d_max);
      u_min = -(dot_pos - d_min);

      float f = mat(j).Dot(norm_dir1);

                              //check if we aren't parallel with this side
      if(IsAbsMrgZero(f)){
                              //check if any of 'from' or 'to' is inside
                              //fast quit 1 - missing parallel side
         if(u_min>=0.0f || u_max<0.0f){
            return false;
         }
         continue;
      }
      float r_dot_dir = 1.0f / f;
      u_max *= r_dot_dir;
      u_min *= r_dot_dir;

                              //determine which side (min/max) we've hit first
      if(u_max < u_min)
         swap(u_max, u_min);

                              //both sides behind our back, exit
      if(u_max <= 0.0f)
         return false;

      if(u_min > 0.0f){
                              //both sides in front of us, we can't be in
         hmin = Max(hmin, u_min);
         in = false;
      }
      lmax = Min(lmax, u_max);
   }
                              //we're inside we're colliding in lmax distance
   if(in){
      if(hit_dist_ret)
         *hit_dist_ret = 0.0f;
      return true;
   }
                              //test if our collision computations end up on the box
   if(hmin > lmax)
      return false;

                              //we've hit the box
                              //check if we have already better collision detected
   if(len1 <= hmin)
      return false;

                              //collision detected
   if(hit_dist_ret)
      *hit_dist_ret = hmin;
   return true;
}

//----------------------------

bool __stdcall FastTest_S_S(const S_trace_help &th, const S_vector &sphere_pos, float sphere_radius){

   assert(CHECK_ZERO_GREATER(sphere_radius));
   float radius_sum = th.radius + sphere_radius;
                              //get distance of 'from' from volume
   const S_vector vol_from_dir = sphere_pos - th.from;
   float d_from2 = vol_from_dir.Dot(vol_from_dir);
                              //get distance to 'to' plus radius
   float d_sum2 = th.dir_magn + radius_sum;
   d_sum2 *= d_sum2;
                              //compare with (sum of radies)^2
   //if(d_sum2 < d_from2) return false;
   if(FLOAT_BITMASK(d_sum2) < FLOAT_BITMASK(d_from2)) 
      return false;
   float radius_sum2 = radius_sum * radius_sum;
   //if(radius_sum2 > d_from2) return true;
   if(FLOAT_BITMASK(radius_sum2) > FLOAT_BITMASK(d_from2)) 
      return true;
                              //get distance of volume pos to direction line
   if(!th.dir_2_sum)
      return false;
   float u = vol_from_dir.Dot(th.dir) / th.dir_2_sum;
   if(u <= 0.0f) return false;
   S_vector col_normal;
   if(u >= 1.0f) col_normal = th.to - sphere_pos;
   else col_normal = (th.from + th.dir * u) - sphere_pos;
   float d_to2 = col_normal.Dot(col_normal);
   if(radius_sum2 > d_to2) return true;
   return false;
}

//----------------------------
// I3D_collision_data
//----------------------------

void I3D_collision_data::SetReturnValues(float h_dist, PI3D_frame h_frm, const S_vector *h_norm, const S_vector *h_pos){

   closest_hit = h_dist;
   closest_frm = h_frm;
   if(h_norm)
      hit_normal = *h_norm;
   if(h_pos)
      dest_pos = *h_pos;
}


//----------------------------
// I3D_scene collision functions
//----------------------------

bool I3D_scene::TestPickEditFrames(I3D_collision_data &cd, const float test_length) const{

                              //try picking non-visuals 
   dword enum_flags = 0;
   if(drv->GetFlags2()&DRVF2_DRAWLIGHTS) enum_flags |= (ENUMF_LIGHT | ENUMF_TARGET);
   if(drv->GetFlags2()&DRVF2_DRAWSOUNDS) enum_flags |= (ENUMF_SOUND | ENUMF_TARGET);
   if(drv->GetFlags2()&DRVF2_DRAWCAMERAS) enum_flags |= (ENUMF_CAMERA | ENUMF_TARGET);
   if(drv->GetFlags2()&DRVF2_DRAWDUMMYS) enum_flags |= ENUMF_DUMMY;
   if(drv->GetFlags2()&DRVF2_DRAWJOINTS) enum_flags |= ENUMF_JOINT;
   if(drv->GetFlags2()&DRVF2_DRAWOCCLUDERS) enum_flags |= ENUMF_OCCLUDER;
   if(cd.flags&I3DCOL_EDIT_NONPICKABLE)
      enum_flags |= ENUMF_VISUAL;
   if(enum_flags){
      struct S_hlp{
         I3D_collision_data *pick_data;
         dword flags;
         S_vector from;
         S_vector norm_dir;
         float test_length;
         bool picked;
         static I3DENUMRET I3DAPI cbPick(PI3D_frame frm, dword c){

            S_hlp *hp = (S_hlp*)c;
            float icon_size = .5f;
            
            switch(frm->GetType1()){
            case FRAME_JOINT:
               {
                  PI3D_joint jnt = I3DCAST_JOINT(frm);
                  icon_size = jnt->GetIconSize() * jnt->GetMatrix()(0).Magnitude() * .3f;
                  //icon_size = .06f;
               }
               break;
            case FRAME_TARGET:
               {
                           //check if it's target for out type
                  PI3D_frame tf = I3DCAST_TARGET(frm)->GetFrame();
                  bool ok = false;
                  if(tf)
                  switch(tf->GetType1()){
                  case FRAME_LIGHT: ok = (hp->flags&ENUMF_LIGHT); break;
                  case FRAME_SOUND: ok = (c&ENUMF_SOUND); break;
                  case FRAME_CAMERA: ok = (c&ENUMF_CAMERA); break;
                  case FRAME_DUMMY: ok = (c&ENUMF_DUMMY); break;
                  case FRAME_JOINT: ok = (c&ENUMF_JOINT); break;
                  }
                  if(!ok) return I3DENUMRET_OK;
                  icon_size = .3f;
               }
               break;

            case FRAME_OCCLUDER:
               {
                              //perform bounding-box test
                  PI3D_occluder occ = I3DCAST_OCCLUDER(frm);

                  float hit_dist;
                  bool b = false;

                  switch(occ->GetOccluderType()){
                  case I3DOCCLUDER_MESH:
                              //first make bounding-box test
                     b = pickBBoxTest(frm->GetMatrix(), occ->GetBound().bound_local.bbox,
                        hp->from, hp->norm_dir, hp->test_length, &hit_dist);
                     if(b){
                        hit_dist = hp->test_length;
                              //bb test passed, make detailed test agains
                        b = occ->TestCollision(hp->from, hp->norm_dir, hit_dist);
                     }
                     break;
                  case I3DOCCLUDER_SPHERE:
                     {
                                          //check if it's better than what we already have                        
                        float d = (occ->GetWorldPos() - hp->from).Dot(hp->norm_dir);
                        //hit_dist = d; //jirka: isn't should be this here ??
                        if(d < hp->test_length){
                           S_vector pol = hp->from + hp->norm_dir * d;
                           d = (pol - occ->GetWorldPos()).Magnitude();
                                          //check if picked
                           if(d < occ->GetWorldScale()){
                              b = true;
                           }
                        }
                     }
                     break;
                  }

                  if(b){
                     hp->test_length = hit_dist;
                     hp->pick_data->SetReturnValues(hit_dist, frm, &S_vector(.0f, .0f, .0f));
                     hp->picked = true;
                  }
                  return I3DENUMRET_OK;
               }
               break;
            case FRAME_DUMMY:
               {
                  PI3D_dummy dum = I3DCAST_DUMMY(frm);
                  const I3D_bbox &bbox = *dum->GetBBox();
                  icon_size = ((bbox.max - bbox.min) * frm->GetMatrix().GetScale()).Magnitude() * .5f;
               }
               break;

            case FRAME_VISUAL:
               {
                              //pick known "unpickable" visual types
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  switch(vis->GetVisualType()){
                  case I3D_VISUAL_PARTICLE:
                  case I3D_VISUAL_BILLBOARD:
                  case I3D_VISUAL_FLARE:
                     {
                        if(!frm->IsOn1())
                           break;
                        if(!(vis->vis_flags&VISF_BOUNDS_VALID))
                           vis->ComputeBounds();
                        float hit_dist;
                        bool b = pickBSphereTest(vis->bound.GetBoundSphereTrans(vis, FRMFLAGS_BSPHERE_TRANS_VALID),
                           hp->from, hp->norm_dir, hp->test_length);
                        if(b){
                           b = pickBBoxTest(vis->GetMatrix(), vis->GetBoundVolume().bbox,
                              hp->from, hp->norm_dir, hp->test_length, &hit_dist);
                        }
                        if(b){
                           hp->test_length = hit_dist;
                           hp->pick_data->SetReturnValues(hit_dist, frm, &S_vector(.0f, .0f, .0f));
                           hp->picked = true;
                        }
                     }
                     break;
                  }
                  return I3DENUMRET_OK;
               }
               break;
            }
                           //test distance to line
            const S_vector &pos = frm->GetWorldPos();
            float u;
            u = (pos-hp->from).Dot(hp->norm_dir) / (hp->norm_dir.Dot(hp->norm_dir));
            if(u>=0.0f && (!hp->pick_data->GetHitFrm() || u < hp->test_length)){
               S_vector pol = hp->from + hp->norm_dir * u;
               float dist = (pol-pos).Magnitude();
               if(dist < icon_size){
                           //picked!
                  hp->test_length = u;
                  hp->pick_data->SetReturnValues(u, frm, &S_vector(.0f, .0f, .0f));
                  hp->picked = true;
               }
            }
            return I3DENUMRET_OK;
         }
      } hlp;
      hlp.pick_data = &cd;
      hlp.flags = enum_flags;
      hlp.from = cd.from;
      hlp.norm_dir = cd.dir; hlp.norm_dir.Normalize();
      hlp.test_length = test_length;
      hlp.picked = false;
      I3D_scene::GetPrimarySector1()->EnumFrames(S_hlp::cbPick, (dword)&hlp, enum_flags, NULL);
      return hlp.picked;
   }
   return false;
}

//----------------------------

PI3D_frame I3D_scene::TestColBspLine(I3D_collision_data& cd) const{

#ifdef PERFORMANECE_COUNT
   BegProf();
   for(int i=PERFORMANECE_COUNT; i--; ){
      bsp_builder.RayIntersection(cd);
                                 //support for editor non-visual
      if(cd.flags&I3DCOL_EDIT_FRAMES){
         TestPickEditFrames(cd, cd.GetClosestHit());
      }
   }
   drv->DEBUG(EndProf());
#else
   bsp_builder.RayIntersection(cd);
                              //support for editor non-visual
   if(cd.flags&I3DCOL_EDIT_FRAMES){
      TestPickEditFrames(cd, cd.GetClosestHit());
   }
#endif //PERFORMANECE_COUNT

   return cd.closest_frm;
}
//----------------------------

template<class T>
void operator += (vector<T> &v1, const vector<T> &v2){

   v1.reserve(v1.size()+v2.size());
   for(int i=0, j=v2.size(); i<j; i++)
      v1.push_back(v2[i]);
}

//----------------------------
//----------------------------
                              //recursive function which builds dynamic volume tree

static void cbEnumDynVol(PI3D_driver drv,  //driver for auto-volume creation
   const I3D_scene *scene,               //scene to be assigned to each volume
   PI3D_frame prnt1,                     //current parent frame
   I3D_bbox &up_level_box,                //bound-box which will get expanded by all children
   t_volume_container &up_level_list,    //list which will collect all volumes in this hierarchy level
   t_volume_container &auto_volume_list, //list keeping all auto-created volumes
   t_volume_container &root_volume_list){//list keeping all volumes which are not hierarchical - for root frame it equals to up_level_list - non-hierarchical frames end up in root level list

   int i;

   I3D_bbox bbox;             //bounding box of sub-hierarchy
   bbox.Invalidate();
   t_volume_container sub_tree;

   const PI3D_frame *children = &prnt1->GetChildren1().front();
                              //try find boundary volume on the level
   for(i=prnt1->NumChildren1(); i--; ){

      PI3D_frame frm = children[i];

      switch(frm->GetType1()){
      case FRAME_VOLUME:
         {
            PI3D_volume vol = I3DCAST_VOLUME(frm);

            if(vol->hr_scene!=scene) break;
            vol->Prepare(); //validate all volumes for drawing(how could it be done better?)

            //if(vol->GetFrameFlags()&I3D_FRMF_STATIC_COLLISION)
            if(vol->GetFrameFlags()&I3D_FRMF_STATIC_COLLISION && !(vol->GetVolFlags()&VOLF_EDIT_FORCE_DYNAMIC))
            {
                              //skip static volumes which are supposed in bsp tree
                              //sphere are always in dynamic (for now)
               if(vol->GetVolumeType() != I3DVOLUME_SPHERE)
                  break;         
            }

            drv->voltree_stats.total_num++;
#ifdef DISABLE_HIEARCHY_FOR_NON_STATIC
            if(!(vol->GetFrameFlags()&I3D_FRMF_STATIC_COLLISION))
            {
                              //not static, put to root list.
                              //(to hiearchy goes only static spheres)
                              //don't put disabled volumes into list
               if(!vol->IsOn1()) break;
               if(!vol->Prepare()) break;
                              //commnet out: currently used for special root_volumes
//#ifndef ENABLE_SPHERE_SPACE //volumes goes into root list only when sphere space disabled.                              
               root_volume_list.push_back(vol);
               drv->voltree_stats.root_num++;
//#endif
            }else
#endif

            {
                              //don't put disabled volumes into hierarchy
               if(!vol->IsOn1()) break;
               if(!vol->Prepare()) break;

               //vol->hr_scene = scene;
                              //get bound box of volume
               I3D_bbox vol_bbox;
               vol->GetBoundBox(&vol_bbox);
                              //enlarge this level bbox
               bbox.min.Minimal(vol_bbox.min);
               bbox.max.Maximal(vol_bbox.max);

               sub_tree.push_back(vol);
            }
         }
         //break; //22.01.2002: comment out for support volumes linked to volumes.
      default:
                  //22.01.2002: comment out for support volumes linked to volumes.
                              //collect sub-hierarchy into local list
         //if(frm->NumChildren1())
         //   cbEnumDynVol(drv, scene, frm, bbox, sub_tree, auto_volume_list, root_volume_list);
         break;
      }
                              //collect sub-hierarchy into local list
      if(frm->NumChildren1())
         cbEnumDynVol(drv, scene, frm, bbox, sub_tree, auto_volume_list, root_volume_list);
   }
   int num_vols = sub_tree.size();
   if(!num_vols) return;      //fast quit

   if(num_vols >= DYNAMIC_VOLUME_BOUND_LIMIT){

                              //it's worth to create bound volume
      PI3D_volume vol = drv->CreateVolume();
                              //setup volume

      vol->SetupAutoBoxDyn(sub_tree, &bbox);
                              //store in auto list
      auto_volume_list.push_back(vol);
                              //put to caller's level
      up_level_list.push_back(vol);
      vol->Release();
   }else{
                              //too small list, pass it directly to caller
      up_level_list += sub_tree;
   }
                              //enlarge bound box of caller
   up_level_box.min.Minimal(bbox.min);
   up_level_box.max.Maximal(bbox.max);
}

//----------------------------

static void cbDrawDynVol(const t_volume_container &vtree, PI3D_scene sc){

   for(int i=vtree.size(); i--; ){
      PI3D_volume vol = vtree[i];
      if(vol->IsOn1())
         vol->Draw1(sc, false);
      if(vol->dyn_vol_tree.size()) 
         cbDrawDynVol(vol->dyn_vol_tree, sc);
   }
}

//----------------------------

void I3D_scene::DebugDrawDynamicVolumes() const{

   bool zw_en = drv->IsZWriteEnabled();
   if(zw_en) drv->EnableZWrite(false);

   BuildDynamicVolTree();
   PI3D_scene sc = (PI3D_scene)this;
   cbDrawDynVol(dynamic_volume_tree, sc);
   if(zw_en) drv->EnableZWrite(true);
}

//----------------------------

void I3D_volume::SetupAutoBoxDyn(const t_volume_container &sub_tree, PI3D_bbox bbox1){

                              //auto-created box
   SetVolumeType(I3DVOLUME_BOX);
   vol_flags |= VOLF_AUTO_GEN;
   vol_flags &= ~VOLF_RESET;  //never reset
                              //assign volume hierarchy
   dyn_vol_tree = sub_tree;

   int i;

   normal[0] = S_vector(1.0f, 0.0f, 0.0f);
   normal[1] = S_vector(0.0f, 1.0f, 0.0f);
   normal[2] = S_vector(0.0f, 0.0f, 1.0f);
   bbox = *bbox1;
   const float EXPAND_VALUE = .001f;

   for(i=0; i<3; i++){
                              //expand the bounding box slightly
      bbox.min[i] -= EXPAND_VALUE;
      bbox.max[i] += EXPAND_VALUE;

      d_min[i] =  bbox.min[i];
      d_max[i] = -bbox.max[i];

      side[i].Zero();
      side[i][i] = (float)fabs(d_min[i] + d_max[i]);

      side_dot[i] = side[i].Dot(side[i]);
      side_size_2[i] = side[i][i];
      side_size_2[i] *= side_size_2[i];
   }

                              //make bounding sphere from bounding box
   {
      S_vector mag_half = (bbox.max-bbox.min)*.5f;
      w_sphere.pos = bbox.min + mag_half;
      w_sphere.radius = mag_half.Magnitude();
   }
                              //test - cumpute smallest possible sphere

                              //algo: center of bbox, limit to farhtest sphere
   {
      float max_dist = 0.0f;
      for(i=dyn_vol_tree.size(); i--; ){
         PI3D_volume vol = dyn_vol_tree[i];
         float dist = (w_sphere.pos - vol->w_sphere.pos).Magnitude() + vol->w_sphere.radius;
         max_dist = Max(max_dist, dist);
      }
      w_sphere.radius = Min(w_sphere.radius, max_dist);
   }
}

//----------------------------

void I3D_scene::BuildDynamicVolTree() const{

                              //don't build twice between two frames
   if(GetLastRenderTime() == last_dynamic_tree_build_time)
      return;
//PR_BEG;
   if(scn_flags&SCNF_RESET_DYNAMIC_VOL_TREE){

      memset(&drv->voltree_stats, 0, sizeof(drv->voltree_stats));
      last_dynamic_tree_build_time = GetLastRenderTime();

      DestroyDynamicVolTree();
      I3D_bbox bbox;
      bbox.Invalidate();

#ifdef ENABLE_SPHERE_SPACE
                              //special root volume list - used for line test(if remove, check also cbEnumDynVol)
      cbEnumDynVol(drv, this, primary_sector, bbox, dynamic_volume_tree, dynamic_auto_volume_list, root_volumes);
#else
      cbEnumDynVol(drv, this, primary_sector, bbox, dynamic_volume_tree, dynamic_auto_volume_list, dynamic_volume_tree);
#endif

      scn_flags &= ~SCNF_RESET_DYNAMIC_VOL_TREE;
      scn_flags |= SCNF_DYNAMIC_VOL_TREE_BUILT;

      if(cb_proc)
         cb_proc(CBM_VOLTREEBUILD, 0, 0, NULL);
   }
//PR_END;
}

//----------------------------

void I3D_scene::DestroyDynamicVolTree() const{

   if(scn_flags&SCNF_DYNAMIC_VOL_TREE_BUILT){
      struct S_hlp{
         static void cbDestroyDynVolTree(t_volume_container &vt){

            for(int i=vt.size(); i--; ){
               assert(i < vt.size());
               PI3D_volume vol = vt[i];
               assert(vol);
               cbDestroyDynVolTree(vol->dyn_vol_tree);
            }
            vt.clear();
         }
      };
                              //invalidate tree now, avoid recursive call of DestroyDynamicVolTree
                              //when volume is released.
      scn_flags &= ~SCNF_DYNAMIC_VOL_TREE_BUILT;
      scn_flags |= SCNF_RESET_DYNAMIC_VOL_TREE;

      S_hlp::cbDestroyDynVolTree(dynamic_volume_tree);
      dynamic_auto_volume_list.clear();
#ifdef ENABLE_SPHERE_SPACE
      root_volumes.clear();
#endif
   }
}

//----------------------------
//----------------------------

void I3D_volume::CheckCol_L_S_dyn(struct I3D_collision_data &cd){

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_L_S];
#endif

                              //check if direct dist to sphere closest than what we have
   S_vector dir_to_sphere = w_sphere.pos - cd.from;
   {
      float dist_2 = dir_to_sphere.Dot(dir_to_sphere);
      float u = cd.GetClosestHit() + w_sphere.radius;
      u *= u;
      if(FLOAT_BITMASK(u) < FLOAT_BITMASK(dist_2))
                              //we can't reach the sphere anyway
         return;
   }
                              //the collision of line and sphere is solved by equation:
                              // d = -b +- sqrt(b^2 - 4 * a * c) / (2 * a)

                              //compute a (squared line's direction)
                              //optimization: since a is always 1.0f (we've got normalized
                              // direction), all we need is to scale top part by 0.5f 

                              //compute b (2 * line's direction dot direction to sphere center)
   const S_vector &norm_dir = cd.normalized_dir;

   float b = norm_dir.Dot(dir_to_sphere) * 2.0f;
                              //compute c (sphere center squared + line's P1 squared
                              // - 2 * line's p1 dot sphere center - squared radius
   float from_dot = cd.from.Dot(cd.from);
   float c = sphere_pos_dot + from_dot -
      2.0f * (cd.from.Dot(w_sphere.pos)) - radius_2;

                              //solve the equation (a is ommited, since it is 1.0)
   float d = b * b - 4.0f * c;
                              //now we have semi-result, by which sign we may detect:
                              //when d is less than zero, the ray doesn't intersect line,
                              // when d is zero, the ray touches line (which we may dismiss too)
   if(CHECK_ZERO_LESS(d))
      return;

   d = SqrtFast(d);
                              //solve both cases of the equation
                              // postpone division (i.e. multiply by 0.5) for later
   float u  = (b + d);
   float u1 = (b - d);
     
                              //get smaller positive value (intersection distance)
#ifdef SPHERE_VOLUME_DETECT_INSIDE
   if(CHECK_ZERO_LESS(u)){
      if(CHECK_ZERO_LESS(u1)){
                              //both intersection points behind us
         return false;
      }
      u = u1;
   }else
   if(CHECK_ZERO_GREATER(u1)){
      u = Min(u, u1);
   }
#else
                              //if any of them is less than zero, we can't hit the sphere
   if(CHECK_ZERO_LESS(u) || CHECK_ZERO_LESS(u1))
      return;
   (int&)u = Min((int&)u, (int&)u1);
#endif

   u *= .5f;
                              //now u is distance of closest collision point

                              //check if closest hit is better than what we have
   if(FLOAT_BITMASK(cd.GetClosestHit()) < FLOAT_BITMASK(u))
      return;

   //S_vector hit_normal = (pick_data.from + pick_data.normalized_dir * u) - w_sphere.pos;
   S_vector hit_normal = (cd.from + norm_dir * u) - w_sphere.pos;

                              //confirm collision
   if(cd.cresp_face){
                              //store last valid values
      S_vector last_norm = cd.GetHitNormal();
      float last_dist = cd.GetClosestHit();
      PI3D_frame last_frm = cd.GetHitFrm();
                              //set values for collision response
      cd.SetReturnValues(u, this, &hit_normal);
      //if(!cd.cresp_face(hit_normal, u, this, cd.cresp_context, NULL))
      bool b = cd.cresp_face(cd);
                              //restore last values
      cd.SetReturnValues(last_dist, last_frm, &last_norm);

      if(!b)
         return;
   }
                              //collision detected
                              //save closest distance
   cd.SetReturnValues(u, this, &hit_normal);

   vol_flags |= VOLF_DEBUG_TRIGGER;
   return;
}

//----------------------------

bool I3D_volume::CheckCol_L_B_dyn(struct I3D_collision_data &cd, bool passive){

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_L_B];
#endif

                              //consider we're in the box
   bool in = true;
                              //cached distances, keeping min/max extremes
   float hmin = 0.0f;
   float lmax = 1.0e+10f;

                              //keep indicies into best inside and outside normals
                              //so that we may return it as collision result
   int best_norm;
#ifdef _DEBUG
   best_norm = -1;
#endif
                              //flip flags help us determine if normal need
                              //to be flipped
   bool flip_normal;

#ifdef BOX_VOLUME_DETECT_INSIDE
   int best_norm_in;
   bool flip_normal_in;
#endif

                              //check all 3 axes
   for(int j=0; j<3; j++){
                              //distances in which our ray collides with
                              //min/max sides of box
      float u_min, u_max;

      float dot_pos = normal[j].Dot(cd.from);

      u_max = -(dot_pos + d_max[j]);
      u_min = -(dot_pos - d_min[j]);

      S_vector &norm_dir = cd.normalized_dir;
      float f = normal[j].Dot(norm_dir);

//      L_B_DEBUG(C_fstring("u_min: %.2f, u_max: %.2f, f: %.2f", u_min, u_max, f));
                              //check if we aren't parallel with this side
      if(IsAbsMrgZero(f)){
                              //check if any of 'from' or 'to' is inside
                              //fast quit 1 - missing parallel side
         if(CHECK_ZERO_GREATER(u_min) || CHECK_ZERO_LESS(u_max)){
//            L_B_DEBUG("Fast quit 1 - missing parallel side");
            return false;
         }
         continue;
      }
      float r_dot_dir = 1.0f / f;
      u_max *= r_dot_dir;
      u_min *= r_dot_dir;

      bool swapped = false;
                              //determine which side (min/max) we've hit first
      if(u_max < u_min){
         swap(u_max, u_min);
         swapped = true;
      }

                              //both sides behind our back, exit
      if(CHECK_ZERO_LESS(u_max))
         return false;

      if(CHECK_ZERO_GREATER(u_min)){
                              //both sides in front of us, we can't be in
         if(hmin <= u_min){
                              //this is out farest (outside) hit position
                              //save this
            hmin = u_min;
            best_norm = j;
            flip_normal = swapped;
         }
         in = false;
      }
      if(lmax > u_max){
                              //this is out closest (inside) hit position
                              //save this
         lmax = u_max;
#ifdef BOX_VOLUME_DETECT_INSIDE
         best_norm_in = j;
         flip_normal_in = swapped;
#endif
      }
   }

   if(in){
                              //we're inside
                              // if passive test is made, it's succeeded,
                              // otherwise it's failed
#ifdef BOX_VOLUME_DETECT_INSIDE
                              //passive check ends here
      if(passive)
         return true;
                              //we're colliding with best_norm_in side
                              //in lmax distance
      hmin = lmax;
      best_norm = best_norm_in;
      flip_normal = flip_normal_in;
#else
      return (passive);
#endif
   }
#ifdef BOX_VOLUME_DETECT_INSIDE
   else
#endif
                              //test if our collision computations
                              //end up on the box
   if(hmin > lmax)
      return false;
                              //we've hit the box
                              //check if we have already better collision detected
   if(FLOAT_BITMASK(cd.GetClosestHit()) <= FLOAT_BITMASK(hmin))
      return false;

                              //collision detected
                              //passive check ends here
   if(passive)
      return true;

   assert(best_norm!=-1);

   S_vector hit_normal = normal[best_norm];
   if(!flip_normal)
      hit_normal.Invert();

   if(cd.cresp_face){
                              //store last valid values
      S_vector last_norm = cd.GetHitNormal();
      float last_dist = cd.GetClosestHit();
      PI3D_frame last_frm = cd.GetHitFrm();
                              //set values for collision response
      cd.SetReturnValues(hmin, this, &hit_normal);
      //if(!cd.cresp_face(hit_normal, hmin, this, cd.cresp_context, NULL))
      bool b = cd.cresp_face(cd);
                              //restore last values
      cd.SetReturnValues(last_dist, last_frm, &last_norm);
      if(!b)
         return false;
  }

   cd.SetReturnValues(hmin, this, &hit_normal);

   vol_flags |= VOLF_DEBUG_TRIGGER;
   return true;
}

//----------------------------

void I3D_volume::CheckCol_L_R_dyn(struct I3D_collision_data &cd){

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_L_R];
#endif

                              //get ray/plane collision point
   S_vector &norm_dir = cd.normalized_dir;
   float u = normal[2].Dot(norm_dir);
   if(IsAbsMrgZero(u))
      return;

   float f = -normal[2].Dot(cd.from);
   float length = (f - d_max[2]) / u;
                              //don't look back
   if(CHECK_ZERO_LESS(length))
      return;
                              //reject more far collisions than
                              //what we already have
   if(FLOAT_BITMASK(cd.GetClosestHit()) <= FLOAT_BITMASK(length))
      return;
                              //get point on plane
   S_vector p = cd.from + norm_dir * length;

                              //check point's position against edges
                              //check X
   u = normal[0].Dot(p);
   if(u <= d_min[0] || u >= -d_max[0])
      return;
                              //check Y
   u = normal[1].Dot(p);
   if(u <= d_min[1] || u >= -d_max[1])
      return;

                              //collision detected
   S_vector hit_normal = normal[2];
   if(f > d_max[2])
      hit_normal.Invert();

   if(cd.cresp_face){
                              //store last valid values
      S_vector last_norm = cd.GetHitNormal();
      float last_dist = cd.GetClosestHit();
      PI3D_frame last_frm = cd.GetHitFrm();
                              //set values for collision response
      cd.SetReturnValues(length, this, &hit_normal);
      //if(!pick_data.cresp_line(hit_normal, length, this, pick_data.cresp_context, NULL))
      bool b = cd.cresp_face(cd);
                              //restore last values
      cd.SetReturnValues(last_dist, last_frm, &last_norm);

      if(!b)
         return;
   }

   cd.SetReturnValues(length, this, &hit_normal);

   vol_flags |= VOLF_DEBUG_TRIGGER;
   return;
}

//----------------------------

/*void I3D_scene::CheckColLineRecursiveDyn(I3D_collision_data &cd, const t_volume_container &vtree) const{

   const PI3D_volume *vols = (const PI3D_volume*)vtree.begin();

   for(int i=vtree.size(); i--; ){
      PI3D_volume vol = *vols++;
      if(i){
         PI3D_volume next = *vols;
         Prefetch2ptr(next, &next->vol_flags);
         Prefetch1(&next->w_sphere);
      }
      if(!vol->IsOn1())
         continue;

      if(cd.frm_ignored && cd.frm_ignored==vol->GetOwner1())
         continue;

      if(vol->GetVolFlags()&VOLF_RESET)
         vol->Prepare();

      switch(vol->GetVolumeType1()){

      case I3DVOLUME_SPHERE:
         vol->CheckCol_L_S_dyn(cd);
         break;

      case I3DVOLUME_BOX:
         if(vol->GetVolFlags()&VOLF_AUTO_GEN){
                              //skip hierarchy if no collision
            if(vol->CheckCol_L_B_dyn(cd, true)){
                              //repeat on children
               CheckColLineRecursiveDyn(cd, vol->dyn_vol_tree);
            }
         }else{
#ifdef VOLUME_STATISTICS
            ++drv->col_stats[I3D_driver::DRV_COL_STATS_L_S];
#endif
            if(!FastTest_L_S(cd, vol->w_sphere.pos, vol->w_sphere.radius, vol->radius_2))
               break;

            vol->CheckCol_L_B_dyn(cd, false);
         }
         break;

      case I3DVOLUME_RECTANGLE:
         vol->CheckCol_L_R_dyn(cd);
         break;
      }
   }
}*/

//----------------------------

void I3D_scene::CheckColLineList(I3D_collision_data &cd, const PI3D_volume *vols, int num_vols, bool children) const{

   assert(num_vols >= 0);
   for(int i=num_vols; i--; ){
      PI3D_volume vol = *vols++;
#ifdef USE_PREFETCH
      if(i){
         PI3D_volume next = *vols;
         Prefetch2ptr(next, &next->vol_flags);
         Prefetch1(&next->w_sphere);
      }
#endif
      if(!vol->IsOn1())
         continue;

      if(cd.frm_ignored && cd.frm_ignored==vol->GetOwner1())
         continue;

      if(vol->GetVolFlags()&VOLF_RESET)
         vol->Prepare();

      switch(vol->GetVolumeType1()){

      case I3DVOLUME_SPHERE:
         vol->CheckCol_L_S_dyn(cd);
         break;

      case I3DVOLUME_BOX:
         if(vol->GetVolFlags()&VOLF_AUTO_GEN){
                              //skip hierarchy if no collision
            if(vol->CheckCol_L_B_dyn(cd, true)){
                              //repeat on children
               //CheckColLineRecursiveDyn(cd, vol->dyn_vol_tree);
               const t_volume_container &sublist = vol->dyn_vol_tree;
               assert(children || !sublist.size()); //velidity: root volumes should not have children
               if(children && sublist.size())
                  CheckColLineList(cd, (const PI3D_volume*)sublist.begin(), sublist.size(), children);
            }
         }else{
#ifdef VOLUME_STATISTICS
            ++drv->col_stats[I3D_driver::DRV_COL_STATS_L_S];
#endif
            if(!FastTest_L_S(cd, vol->w_sphere.pos, vol->w_sphere.radius, vol->radius_2))
               break;

            vol->CheckCol_L_B_dyn(cd, false);
         }
         break;

      case I3DVOLUME_RECTANGLE:
         vol->CheckCol_L_R_dyn(cd);
         break;
      }
   }
}


//----------------------------
// Test collision on dynamic tree an sphere space.
void I3D_scene::CheckColLineDyn(I3D_collision_data &cd, const t_volume_container &vtree) const{

                              //first check static spheres
   CheckColLineList(cd, (const PI3D_volume*)vtree.begin(), vtree.size(), true);

                              //second check spheres space map
#ifdef ENABLE_SPHERE_SPACE

   CheckColLineList(cd, (const PI3D_volume*)root_volumes.begin(), root_volumes.size(), false);

   /*t_spheres_space::const_iterator it;
   for(it = spheres_space.begin(); it != spheres_space.end(); it++){

      const vector<CPI3D_volume> &vol_list = (*it).second;
      CheckColLineList(cd, (const PI3D_volume*)vol_list.begin(), vol_list.size());
   }*/
                              //check only spaces which can line intersect
   /*float space_min = cd.from.Magnitude();
   S_vector to = cd.from + cd.dir;
   S_vector to = cd.from + cd.normalized_dir * cd.closest_hit;
   float space_max = to.Magnitude();
   if(space_min > space_max)
      swap(space_min, space_max);
   int sp_max_i = FLOAT_2_SPHERE_SPACE(space_max);
   int sp_min_i = FLOAT_2_SPHERE_SPACE(space_min);

   for(int space_i = sp_min_i; space_i <= sp_max_i; space_i++){
      t_spheres_space::iterator it = spheres_space.find(space_i);
      if(it == spheres_space.end())
         continue;

      vector<CPI3D_volume> &vol_list = spheres_space[space_i];
      CheckColLineList(cd, (const PI3D_volume*)vol_list.begin(), vol_list.size());
   }*/
#endif
}

//----------------------------

PI3D_volume I3D_scene::TestColLineDynamic(struct I3D_collision_data &cd) const{


   //if(cd.dir.IsNull()) 
   if(IsAbsMrgZero(cd.dir_magnitude) || (IsAbsMrgZero(cd.closest_hit)))
      return NULL;

   if(scn_flags&SCNF_RESET_DYNAMIC_VOL_TREE) 
      BuildDynamicVolTree();

//PR_BEG;
                              //use line segment, default is ray
   if(!(cd.flags&I3DCOL_RAY))
      cd.closest_hit = Min(cd.dir_magnitude, cd.GetClosestHit());

#ifdef PERFORMANECE_COUNT
   BegProf();
   for(int i=PERFORMANECE_COUNT; i--; )
   //CheckColLineRecursiveDyn(cd, dynamic_volume_tree);
   CheckColLineDyn(cd, dynamic_volume_tree);
   drv->DEBUG(EndProf());
#else
   //CheckColLineRecursiveDyn(cd, dynamic_volume_tree);
   CheckColLineDyn(cd, dynamic_volume_tree);
#endif

//PR_END;

   return I3DCAST_VOLUME(cd.closest_frm);
}

//----------------------------



//----------------------------

bool I3D_scene::TestCollision(struct I3D_collision_data &cd) const{

#ifdef _DEBUG
   log_drv = drv;  //!!total debug, used for COL_DEBUG
#endif

   cd.PrepareTest();
                              //don't test zero length line
   if(IsAbsMrgZero(cd.dir_magnitude))
      return false;

   if(cd.flags&I3DCOL_EDIT_FRMS_ONLY){
      return TestPickEditFrames(cd, cd.GetClosestHit());
   }
                              //static sphere test
   if(cd.flags&I3DCOL_STATIC_SPHERE)
   {
      bool b = bsp_builder.SphereStaticIntersection(cd);
      return b;
   }
                              //sphere group
   if(cd.flags&I3DCOL_SPHERE_GROUP){
      assert(cd.frm_root);
      if(!cd.frm_root) 
         return false;
//PR_BEG;
      bool b = SphereMove(cd, cd.frm_root);
//PR_END;
      return b;
   }
                              //sphere test
   if(cd.flags&I3DCOL_SPHERE){
//PR_BEG;
      bool b = SphereMove(cd);
//PR_END;
      return b;
   }
                              //exact test on geometry
   if(cd.flags&I3DCOL_EXACT){
      float closest_hit;
      S_vector hit_normal;
      PI3D_frame closest_frm = TestColGeoInternal(cd.from,
         cd.dir,
         cd.flags, &closest_hit,
         &hit_normal, cd.cresp_line_old,
         cd.cresp_context, cd.cresp_texel_old);
      if(closest_frm){
         cd.SetReturnValues(closest_hit, closest_frm, &hit_normal);
      }
      return closest_frm;
   }
                              //test dynamic hr tree, result is stored in cd
   TestColLineDynamic(cd);
                              //test joints volumes
   if(cd.flags&I3DCOL_JOINT_VOLUMES)
      TestVolumeSM(cd);
                              //check bsp tree, result is stored in cd
   return TestColBspLine(cd);                              //
}

//----------------------------
// Sphere - BSP intersection (static, move)
//----------------------------

//----------------------------
// Clip velocity by plane. Both (velocity and plane) has their origin in O.
void ClipVelocity(S_vector &velocity, const S_vector &normal){

   //const float stop_epsilon = 1e-5f;
   const float stop_epsilon = MRG_ZERO;

	float backoff = velocity.Dot(normal);
   velocity -= normal*backoff;

   for(int i = 3; i--;){
      if(velocity[i] > -stop_epsilon && velocity[i] < stop_epsilon)
         velocity[i] = .0f;
   }
}

//----------------------------
// Slide move along collision planes.
// Return false if move was stopped.
// Algo:
// Testing line (from, to). Trace for closest collision, accept first part (from, impact_point).
// Now response velocity (impact_point, to), ie second part of line. We clip this velocity
// vector by plane normal and recieve new_direction. New to we get by equtation
// new_to = impact_point + new_direction. 
// This relative solution don't require impact plane's d or to_pos (both can be computed from
// impact_point), which is main adventage when trying response group of objects, where 
// every volume collide on different line (parallel).
bool SolveCollision(S_vector &solve_dir, const vector<S_vector> &clip_planes,
                    const S_vector &orig_dir, const S_vector &primary_velocity){

   int num_planes = clip_planes.size();
   const S_vector *p_clip = clip_planes.begin();
                  //modify dir so it parallels all of the clip planes
   for(int i = num_planes; i--; ){
      const S_vector &clip = p_clip[i];
      ClipVelocity(solve_dir, clip);
      COL_DEBUG(C_fstring("use clip plane [%.2f, %.2f, %.2f], move dir change to: [%.2f, %.2f, %.2f]", 
         clip.x, clip.y, clip.z, solve_dir.x, solve_dir.y, solve_dir.z));
                              //check parallel/back
      for (int j = num_planes; j--; ){
		   if(j != i){
            float f = solve_dir.Dot(p_clip[j]);
            //if(CHECK_ZERO_LESS(f))
            if(f < -MRG_ZERO)
            {
               COL_DEBUG(C_fstring("velocity-clip dot: %.8f", f));
				   break;	// not ok
            }
		   }
      }
      if(j == -1) 
         break;               //all clip planes parallel, break
   }

   if(i != -1){
                        // ok, slide along plane, just keep modified dir
      COL_DEBUG(C_fstring("move_dir clipped from [%.2f, %.2f, %.2f] to [%.2f, %.2f, %.2f]",
         orig_dir.x, orig_dir.y, orig_dir.z, solve_dir.x, solve_dir.y, solve_dir.z));
   }
   else{
                        // planes forming line, slide it
	   if(num_planes != 2){
                              //blocked by more planes
         //solve_dir = orig_dir;
         solve_dir.Zero();
         COL_DEBUG(C_fstring("move bloked by %i planes", num_planes));
         return false;
	   }

      int last_clip = 0;
      int prev_clip = 1;
      S_vector line_dir = p_clip[prev_clip].Cross(p_clip[last_clip]);
      float d = line_dir.Dot(orig_dir);
      solve_dir = line_dir * d;
      COL_DEBUG(C_fstring("forming slide line: [%.2f, %.2f, %.2f]", solve_dir.x, solve_dir.y, solve_dir.z));
   }

                        // if new velocity is against the original velocity, stop to avoid occilations in corners
#if 1
   float f3 = primary_velocity.Dot(solve_dir);
	if(CHECK_ZERO_LESS(f3))
   {
      COL_DEBUG(C_fstring("velocity against primary dir[%.2f, %.2f, %.2f], dot: %.8f",
         primary_velocity.x, primary_velocity.y, primary_velocity.z, f3));
      solve_dir.Zero();
		return false;
	}
#endif
   return true;
}


//----------------------------
//Default collision response, slide along the clip planes.
//Always return true (collison was accepted).
bool I3DAPI I3D_DefaultColResponse(struct I3D_collision_data &cd){

   vector<S_vector> &clip_planes_global = *(vector<S_vector>*)cd.reserved;

   clip_planes_global.push_back(cd.GetHitNormal());

   S_vector solve_dir = cd.dir;
   bool stop_move = !SolveCollision(solve_dir, clip_planes_global, 
      cd.GetOrigDir(), cd.GetPrimaryVelocity());
   cd.SetDestination(stop_move ? cd.from : cd.from + solve_dir);
   return true;
}

//----------------------------

bool I3D_driver::CollisionResponse(I3D_collision_data &cd) const{
   return I3D_DefaultColResponse(cd);
}

//----------------------------
//----------------------------

bool C_bsp_tree::TraceSphere(const S_trace_help &th)const{

#if 1

   if(!root)
      return false;

   if(IsMrgZero(th.dir_magn))
      return false;

   bool b = NodeIntersection_MS(*root, .0f, th.dir_magn, th);
   return b;

#else
   bool ret = false;
   const S_bsp_triface *p_trifaces = faces.begin();
   for(int i = 0; i < faces.size(); i++)
   //int i = 3;
   {
      bool b = CheckCol_MS_F(p_trifaces[i], th.from, th.to, th.radius, th.collision_list);
      if(b){
         p_trifaces[i].highlight = true;
         ret = true;
      }
   }
   return ret;
#endif
}

//----------------------------
                              //sphere group support

struct S_sphere_hlp{
   float radius;              //radius of sphere
   S_vector vol_delta;        //delta position against master frame
   PI3D_volume vol;           //pinter to sphere volume
};

//----------------------------
                              //used for collision of group of sphere
struct S_col_group_hlp{
   S_vector master_pos;
   vector<S_sphere_hlp> sphere_list;
};

//----------------------------

/*
static I3DENUMRET I3DAPI cbGroupCol(PI3D_frame frm, dword hlp){

   PI3D_volume vol = (PI3D_volume)frm;

   if(vol->GetVolumeType() == I3DVOLUME_SPHERE){

      S_col_group_hlp &cgh = *(S_col_group_hlp*)hlp;
      cgh.sphere_list.push_back(S_sphere_hlp());
      S_sphere_hlp &sh = cgh.sphere_list.back();

      const S_matrix &mv = frm->GetMatrix();
      sh.radius = mv(0).Magnitude();
      sh.vol_delta = mv(3) - cgh.master_pos;
      sh.vol = vol;
   }
   return I3DENUMRET_SKIPCHILDREN;
}
*/
static void cbCollectSpheres(PI3D_frame root, const S_vector &master_pos, vector<S_sphere_hlp> &sphere_list){

   const vector<PI3D_frame> &children = root->GetChildren1();
   PI3D_frame const *cptr = children.begin();
   for(int i=children.size(); i--; ){
      PI3D_frame frm = *cptr++;
      if(frm->GetType1()==FRAME_VOLUME){
         PI3D_volume vol = I3DCAST_VOLUME(frm);
         if(vol->GetVolumeType1()==I3DVOLUME_SPHERE){
            sphere_list.push_back(S_sphere_hlp());
            S_sphere_hlp &sh = sphere_list.back();

            if(vol->GetVolFlags()&VOLF_RESET)
               vol->Prepare();
            const I3D_bsphere bs = vol->GetBoundSphere();
            sh.radius = bs.radius;
            sh.vol_delta = bs.pos - master_pos;
            sh.vol = vol;
         }
      }
      if(frm->NumChildren1())
         cbCollectSpheres(frm, master_pos, sphere_list);
   }
}
                              
//----------------------------

bool I3D_scene::SphereMove(I3D_collision_data &cd, PI3D_frame root) const{

   if(cd.flags&I3DCOL_SOLVE_SLIDE){
      if(!cd.cresp_face)
         cd.cresp_face = I3D_DefaultColResponse;
   }

                              //sphere group support:
   vector<S_sphere_hlp> sphere_list;
   //S_col_group_hlp cg_hlp;
   if(root){
                              //get info about spheres group
      //root->EnumFrames(cbGroupCol, (dword)&cg_hlp, ENUMF_VOLUME, NULL);
      sphere_list.reserve(16);
      cbCollectSpheres(root, root->GetWorldPos(), sphere_list);
   }
                              //(end: sphere group support)

   const int NUM_BUMPS = 6;       //how much parts of move dir we accept

   bool collided = false;          //if any collision accured

   S_vector primary_velocity = cd.dir; //primary dir, keep it to the end
   S_vector orig_dir = cd.dir;       //currenly solved segment

   S_vector solve_dir = cd.dir;      //current segment modified by clip planes
   S_vector solved_from = cd.from;
   S_vector solved_to = cd.from+cd.dir;
   vector<S_collision_info> affect_list;
   affect_list.reserve(NUM_BUMPS);
   
   vector<S_vector> clip_planes_global;
   clip_planes_global.reserve(NUM_BUMPS);
   cd.reserved = &clip_planes_global;
                              //default ene of move is wanted position (for case that callback don't set it by own)
   cd.SetDestination(solved_to);

                              //we used this list for skip response col twice on one frame

   S_trace_help trace_hlp;
   trace_hlp.collision_list.reserve(128); //ususally small number of collisions expected (10-20)
   trace_hlp.radius = cd.radius;
   trace_hlp.frm_ignored = cd.frm_ignored;
   trace_hlp.flags = cd.flags;
   for(int i = NUM_BUMPS; i--; ){

      trace_hlp.collision_list.clear();
      trace_hlp.from = solved_from;
      trace_hlp.to = solved_to;
      if(!trace_hlp.Prepare())
         break;
                              //sphere group support:
      if(root){
         //for(int kk = cg_hlp.sphere_list.size(); kk--; )
         for(int kk = sphere_list.size(); kk--; )
         {
            //const S_sphere_hlp &sh = cg_hlp.sphere_list[kk];
            const S_sphere_hlp &sh = sphere_list[kk];
            trace_hlp.from = solved_from + sh.vol_delta;
            trace_hlp.to = solved_to + sh.vol_delta;
            trace_hlp.radius = sh.radius;
            trace_hlp.vol_source = sh.vol;
            bsp_builder.TraceSphere(trace_hlp);
//PR_BEG;
            trace_hlp.PrepareMoveSpace(); //keep it after trace_hlp.Prepare(), it use dir_magnitude
            TestColSphereDynamic(trace_hlp);
//PR_END;
         }
      }else
                              //(end:sphere group support)
      {
         //trace_hlp.vol_source = cd.vol; //we don't pass sphere volume to single sphere tests
         bsp_builder.TraceSphere(trace_hlp);
         trace_hlp.PrepareMoveSpace();    //keep it after trace_hlp.Prepare(), it use dir_magnitude
         TestColSphereDynamic(trace_hlp);
      }
                              //if any collison accure
      if(!trace_hlp.collision_list.size())
         break;               //no collision, entire segment pass, end test

                              //sort collected collisions from closest to farthest
      sort(trace_hlp.collision_list.begin(), trace_hlp.collision_list.end());
                              //get closest collision, which plane was not tested
      const S_collision_info *p_col = trace_hlp.collision_list.begin();
      const int num_collisions = trace_hlp.collision_list.size();
      int responsing_col_id = 0;
test_next_collision:
      for(int j = responsing_col_id; j < num_collisions; j++){
         const S_collision_info *p_affect = affect_list.begin();
         for(int k = affect_list.size(); k--; ){
                              //skip faces already tested
            if((p_col[j].face_id != -1) && (p_col[j].face_id == p_affect[k].face_id))
               break;
                              //skip planes on dynamic frames which was tested
            if((p_col[j].plane_id == -1) && (p_col[j].hit_frm == p_affect[k].hit_frm))
               break;
         }
         if(k == -1)
            break;
      }
      if(j == num_collisions)
         break;               //no new plane found, end of tests

                              //remember last response, user may want reject it
      responsing_col_id = j;

                              //put little back from collision plane
      trace_hlp.collision_list[j].in_dist = Max(.0f, p_col[j].in_dist-.001f);
      const S_collision_info &ci = p_col[j];
      S_vector curr_solved_from = solved_from;

                              //part of segment was passed succefully
      if(ci.in_dist > MRG_ZERO)
      {
                              //actualy covered some distance, response only rest of line
         affect_list.clear();
         clip_planes_global.clear();
         S_vector solved_dir_norm = solved_to - solved_from;
         float len = solved_dir_norm.Magnitude();
         assert(!IsAbsMrgZero(len));
         solved_dir_norm.Normalize();
         curr_solved_from += (solved_dir_norm * ci.in_dist);
         float rest_len = (len-ci.in_dist);
         if(IsAbsMrgZero(rest_len)){
            solved_to = curr_solved_from;
            collided |= true;
            break;
         }
         else{
                              //can't be negative, ci.in_dist must be less then entire dir
            //assert(CHECK_ZERO_GREATER(rest_len));
            assert(rest_len > -1e-6f);
            if(!CHECK_ZERO_GREATER(rest_len)){
               //rest_len = .0f;
               solved_to = curr_solved_from;
               collided |= true;
               break;
            }
            solve_dir = solved_dir_norm * rest_len;
            orig_dir = solve_dir;
         }
      }

      affect_list.push_back(ci);
                              //response to collided planes
      {
         I3D_collision_data resp_data;
         resp_data.from = curr_solved_from;
         resp_data.dir = solve_dir;
         resp_data.hit_normal = ci.plane.normal;
         resp_data.face_normal = ci.modified_face_normal ? ci.face_normal : ci.plane.normal;
         resp_data.dest_pos = curr_solved_from + solve_dir;
         resp_data.flags = cd.flags;
         resp_data.reserved = cd.reserved;
         resp_data.vol_source = ci.vol_source;
                              //FIXME:const pointers
         resp_data.closest_frm = (PI3D_frame)ci.hit_frm;
         resp_data.primary_velocity = primary_velocity; 
         resp_data.orig_dir = orig_dir;

         if(!cd.cresp_face){
            cd.closest_frm = (PI3D_frame)ci.hit_frm;
            if(i == NUM_BUMPS-1){
               cd.closest_hit = ci.in_dist;
               cd.hit_normal = ci.plane.normal;
               cd.face_normal = ci.face_normal;
            }
            solved_from = curr_solved_from;
            solved_to = solved_from;
            collided = true;
            break;            //no response, accept closest hit
         }
         resp_data.cresp_face = cd.cresp_face;
         resp_data.cresp_context = cd.cresp_context;
         bool accepted = 
            cd.cresp_face(resp_data);
         collided |= accepted;
#if 1 //USER REJECTION
         if(!accepted){
                              //remove hit plane, which was ignored by response
            affect_list.pop_back();
                              //continue with next collision
            responsing_col_id++;
            goto test_next_collision;
         }
#endif //USER REJECTION
         solved_to = resp_data.GetDestinationPos();
         solved_from = curr_solved_from;
                              //FIXME:this can be returned directly by response function
         solve_dir = solved_to-solved_from;
         cd.closest_frm = (PI3D_frame)ci.hit_frm;
         if(!resp_data.continue_test)
            break;
      }
                              //FIXME:this can be computed elsewhere
      bool end_move = IsAbsMrgZero(solve_dir.Magnitude());

      if(end_move)
         break;
   }
   if(i == -1){
                              //last segmnet didn't pass clearly
      COL_DEBUG(C_fstring("sphere move: collided with more then %i planes, quit", NUM_BUMPS));
      solved_to = solved_from;
   }
   cd.SetDestination(solved_to);
   clip_planes_global.clear(); //free response planes
   return collided;
}

//----------------------------

bool C_bsp_tree::SphereStaticIntersection(I3D_collision_data &cd) const{
 
   bool any_hit = false;
   if(cd.flags &I3DCOL_STATIC_SPHERE){
      if(faces.size() > face_cache.Size())
         face_cache.Reserve(faces.size());
      face_cache.Reset();
      any_hit = root ? NodeIntersection_S(*root, cd) : false;
   }
   return any_hit;
}

//----------------------------

float C_bsp_tree::ClassifySphere(BSP_index plane_indx, const S_vector &pos, float radius, E_PLANE_SIDE &near_side, E_PLANE_SIDE &far_side)const{

   assert(plane_indx < planes.size());
   const S_plane &partion_plane = planes[plane_indx];

   float dist = pos.DistanceToPlane(partion_plane);

   if(dist < -MRG_ZERO){
      near_side = PS_BACK;
      far_side = PS_FRONT;
   }
   else{
      near_side = PS_FRONT;
      far_side = PS_BACK;
   }
   return dist;
}

//----------------------------

bool C_bsp_tree::FaceListItersection_S(const S_bsp_node &node, I3D_collision_data &cd, float plane_dist) const{

   bool ret(false);
   const BSP_index *pflist = node.GetFaces();

   const S_bsp_triface *pfaces = faces.begin();
   for(int i = node.NumFaces(); i--;)
   {
      const S_bsp_triface &tf = pfaces[pflist[i]];

                              //check if we tested it before
      if(face_cache.IsOn(pflist[i]))
         continue;
      else
         face_cache.SetOn(pflist[i]);

                              //recompute distance of sphere to face plane
      bool dist_valid = true;
      float sf_dist;
      if(tf.plane_id != node.plane_id){
         dist_valid = false;
         S_plane face_plane = planes[tf.plane_id];
         if(!tf.plane_side)
            face_plane.Invert();
         sf_dist = cd.from.DistanceToPlane(face_plane);
                                    //fast quit: sphere not touch the plane
         if(fabs(sf_dist) > cd.radius){
            continue;
         }
      }
      bool b = CheckCol_S_F(tf, cd, dist_valid ? plane_dist : sf_dist);
      if(b){
         bool accept = true;
         if(cd.cresp_face){
            accept =  cd.cresp_face(cd);
         }
         tf.highlight = accept;
         ret |= accept;
      }
   }
   return ret;
}

//----------------------------

bool C_bsp_tree::NodeIntersection_S(const S_bsp_node &node, I3D_collision_data &cd) const{


   const S_bsp_node *tree_near = NULL; //child of Node for half-space containing the origin of sphere
   const S_bsp_node *tree_far = NULL;  //the "other" child of Node -- i.e. not equal to near

                              //signed distance of sphere origin to plane defined by Node
   E_PLANE_SIDE near_side, far_side;
   float dist = ClassifySphere(node.plane_id, cd.from, cd.radius, near_side, far_side);
   assert(near_side != far_side); //debug
   tree_near = (near_side == PS_FRONT) ? node.front : (near_side == PS_BACK) ? node.back : NULL;
   tree_far = (far_side == PS_FRONT) ? node.front : (far_side == PS_BACK) ? node.back : NULL;

                              
   if((dist > cd.radius) || (dist < -cd.radius)){
                              //whole sphere is on near side
      if(tree_near){
         bool hit = NodeIntersection_S(*tree_near, cd);
         return hit;
      }
   }
   else{
                              //sphere intersect plane
      bool any_hit = false;
      if(tree_far){
         bool hit = NodeIntersection_S(*tree_far, cd);
         any_hit |= hit;
      }
                           //test if sphere collide with faces in node (on partion plane)
      bool hit = FaceListItersection_S(node, cd, dist);
      any_hit |= hit;

      if(tree_near){
         bool hit = NodeIntersection_S(*tree_near, cd);
         any_hit |= hit;
         return any_hit;
      }
   }
   return false;
}

//----------------------------

bool C_bsp_tree::CheckCol_S_F(const S_bsp_triface &tf, I3D_collision_data &cd, float plane_dist)const{

                              // count the number of triangle vertices inside the sphere
   for(int i=3; i--; ){
      float point_dist = (tf.verticies[i] - cd.from).Magnitude();
                              //fast quit: point inside sphere
      if(point_dist < cd.radius)
         return true;
   }
                              //all point are outside of sphere, 
                              //get closest point on triangle plane
   S_plane face_plane = planes[tf.plane_id];
   if(!tf.plane_side)
      face_plane.Invert();
   S_vector point_on_plane = cd.from - face_plane.normal*plane_dist;

                              //distance of point to edge planes
   bool inside_ep[3];
   for(int j = 3; j--; ){
                              //original precomputed edges are inverted, correct it
      float ep_dist = point_on_plane.DistanceToPlane(-tf.edge_planes[j]);
      inside_ep[j] = CHECK_ZERO_GREATER(ep_dist);
      if(!inside_ep[j])
         break;
   }
                              //fast quit: point inside of triangle
   if(inside_ep[0]&&inside_ep[1]&&inside_ep[2])
      return true;

                              //check intersection of edges and sphere
   float edge_magnitude[3];
   S_vector edge_dir[3];
   for(j = 3; j--; ){

      edge_magnitude[j] = tf.edges[j].Magnitude();
      edge_dir[j] = -tf.edges[j]; edge_dir[j].Normalize();
      float dist_to_edge = cd.from.DistanceToLine(tf.verticies[j], edge_dir[j]);
      float pos_on_line = cd.from.PositionOnLine(tf.verticies[j], edge_dir[j]);

      assert(CHECK_ZERO_GREATER(dist_to_edge));
      bool edge_intersect = (dist_to_edge < cd.radius) &&
         CHECK_ZERO_GREATER(pos_on_line) &&  (pos_on_line < edge_magnitude[j]);
      if(edge_intersect)
         return true;
   }

   return false;
}

//----------------------------
// sphere testing
//----------------------------

bool I3D_scene::TestColSphereDynamic(const S_trace_help &th) const{

//PR_BEG;
   if(scn_flags&SCNF_RESET_DYNAMIC_VOL_TREE) 
      BuildDynamicVolTree();
//PR_END;

#ifdef PERFORMANECE_COUNT
   BegProf();
   PI3D_volume vol;
   for(int i=PERFORMANECE_COUNT; i--; )
      vol = CheckColSDyn(th, dynamic_volume_tree);
   drv->DEBUG(EndProf());
#else
//PR_BEG;
   PI3D_volume vol = CheckColSDyn(th, dynamic_volume_tree);
//PR_END;
#endif

   return vol;
}

//----------------------------

/*PI3D_volume I3D_scene::CheckColSDyn(const S_trace_help &th, const t_volume_container &vtree) const{

   PI3D_volume col_vol = NULL;

#ifndef ENABLE_SPHERE_SPACE
   const PI3D_volume *vols = (const PI3D_volume*)vtree.begin();
#else
   int sp_max_i = FLOAT_2_SPHERE_SPACE(th.space_max);
   int sp_min_i = FLOAT_2_SPHERE_SPACE(th.space_min);
   for(int space_i = sp_min_i; space_i <= sp_max_i; space_i++)
#endif
   {

#ifdef ENABLE_SPHERE_SPACE
      t_spheres_space::iterator it = spheres_space.find(space_i);
      if(it == spheres_space.end()) 
         continue;

      vector<CPI3D_volume> &vtree = spheres_space[space_i];
      const PI3D_volume *vols = (const PI3D_volume*)vtree.begin();
#endif

      for(int i=vtree.size(); i--; ){

         //PI3D_volume vol = vols[i];
         PI3D_volume vol = *vols++;
         if(!vol->IsOn1())
            continue;

         if(th.frm_ignored && ((th.frm_ignored==vol->GetOwner1()) || (th.frm_ignored==vol)))
            continue;

         if(vol->GetVolFlags()&VOLF_RESET)
            vol->Prepare();
         assert(!(vol->GetVolFlags()&VOLF_RESET));

         bool col=false;

         switch(vol->GetVolumeType1()){

         case I3DVOLUME_SPHERE:
            if(vol->CheckCol_S_S_dyn(&th)){
               col = true;
               if(!col_vol) 
                  col_vol = vol;
            }
            break;

         case I3DVOLUME_BOX:

            if(vol->GetVolFlags()&VOLF_AUTO_GEN){
                                 //skip hierarchy if no collision
                                 //perform sphere test, followed by box test
   #ifdef VOLUME_STATISTICS
               ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_S];
   #endif
               //float radius1 = th.radius;
               //bool hr_col = FastTest_S_S(th.from, th.to, radius1, vol->w_sphere.pos, vol->w_sphere.radius);
               bool hr_col = FastTest_S_S(th, vol->w_sphere.pos, vol->w_sphere.radius);
               if(hr_col){
                  hr_col = vol->CheckCol_S_B_dyn(&th, true);
                  if(hr_col){
                     PI3D_volume cv = CheckColSDyn(th, vol->dyn_vol_tree);
                     if(cv && !col_vol) col_vol = cv;
                  }
               }
            }else
            {
   #ifdef VOLUME_STATISTICS
               ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_S];
   #endif
               //float radius1 = th.radius;
               //if(!FastTest_S_S(th.from, th.to, radius1, vol->w_sphere.pos, vol->w_sphere.radius))
               if(!FastTest_S_S(th, vol->w_sphere.pos, vol->w_sphere.radius))
                  break;
               col = vol->CheckCol_S_B_dyn(&th, false);
               if(col){
                  if(!col_vol) col_vol = vol;
               }
            }
            break;

         case I3DVOLUME_RECTANGLE:

            col = vol->CheckCol_S_R_dyn(&th);
            if(col){
               if(!col_vol) col_vol = vol;
            }
            break;
         }
      }
   }
   return col_vol;
}*/

//----------------------------

PI3D_volume I3D_scene::CheckColSDyn(const S_trace_help &th, const t_volume_container &vtree) const{

   PI3D_volume col_vol = NULL;
                              //first check static spheres
   col_vol = CheckColSlist(th, (const PI3D_volume*)vtree.begin(), vtree.size());

                              //second check spheres space map
#ifdef ENABLE_SPHERE_SPACE
   int sp_max_i = FLOAT_2_SPHERE_SPACE(th.space_max);
   int sp_min_i = FLOAT_2_SPHERE_SPACE(th.space_min);
   for(int space_i = sp_min_i; space_i <= sp_max_i; space_i++){
      t_spheres_space::iterator it = spheres_space.find(space_i);
      if(it == spheres_space.end())
         continue;

      vector<CPI3D_volume> &vol_list = spheres_space[space_i];
      col_vol = CheckColSlist(th, (const PI3D_volume*)vol_list.begin(), vol_list.size());
   }
#endif

   return col_vol;
}

//----------------------------

PI3D_volume I3D_scene::CheckColSlist(const S_trace_help &th, const PI3D_volume *vols, int num_vols) const{

   PI3D_volume col_vol = NULL;

   for(int i = num_vols; i--; ){
      PI3D_volume vol = *vols++;
      if(!vol->IsOn1())
         continue;

      if(th.frm_ignored && ((th.frm_ignored==vol->GetOwner1()) || (th.frm_ignored==vol)))
         continue;

      if(vol->GetVolFlags()&VOLF_RESET)
         vol->Prepare();
      assert(!(vol->GetVolFlags()&VOLF_RESET));

      bool col=false;

      switch(vol->GetVolumeType1()){

      case I3DVOLUME_SPHERE:
         if(vol->CheckCol_S_S_dyn(&th)){
            col = true;
            if(!col_vol) 
               col_vol = vol;
         }
         break;

      case I3DVOLUME_BOX:

         if(vol->GetVolFlags()&VOLF_AUTO_GEN){
                              //skip hierarchy if no collision
                              //perform sphere test, followed by box test
#ifdef VOLUME_STATISTICS
            ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_S];
#endif
            bool hr_col = FastTest_S_S(th, vol->w_sphere.pos, vol->w_sphere.radius);
            if(hr_col){
               hr_col = vol->CheckCol_S_B_dyn(&th, true);
               if(hr_col){
//                  PI3D_volume cv = CheckColSDyn(th, vol->dyn_vol_tree);
                  const t_volume_container &sublist = vol->dyn_vol_tree;
                  PI3D_volume cv = NULL;
                  if(sublist.size())
                     cv = CheckColSlist(th, (const PI3D_volume*)sublist.begin(), sublist.size());
                  if(cv && !col_vol) col_vol = cv;
               }
            }
         }else
         {
#ifdef VOLUME_STATISTICS
            ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_S];
#endif
            if(!FastTest_S_S(th, vol->w_sphere.pos, vol->w_sphere.radius))
               break;
            col = vol->CheckCol_S_B_dyn(&th, false);
            if(col){
               if(!col_vol) col_vol = vol;
            }
         }
         break;

      case I3DVOLUME_RECTANGLE:

         col = vol->CheckCol_S_R_dyn(&th);
         if(col){
            if(!col_vol) col_vol = vol;
         }
         break;
      }
   }
   return col_vol;
}

//----------------------------

bool I3D_volume::CheckCol_S_S_dyn(const struct S_trace_help *th)const{

   //assert(th);
   assert(th && th->space_computed);
                              //fast quit if they can't intersect
   if(!(frm_flags&I3D_FRMF_STATIC_COLLISION)){
      if((FLOAT_BITMASK(sphere_space_max) < FLOAT_BITMASK(th->space_min))
         || (FLOAT_BITMASK(sphere_space_min) > FLOAT_BITMASK(th->space_max)))
         return false;
   }

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_S];
#endif

                              //get distance of 'from' from volume
   const S_vector vol_from_dir = w_sphere.pos - th->from;

   float d_from2 = vol_from_dir.Dot(vol_from_dir);

                              //get distance to 'to' plus radius
   float radius_sum = w_sphere.radius + th->radius;
   float d_sum2 = th->dir_magn + radius_sum;
   d_sum2 *= d_sum2;

   S_S_DEBUG_DYN(C_fstring("dist from-sphere^2 = %.2f, dist from-to^2 = %.2f", d_from2, d_sum2));
                              //compare with (sum of radies)^2
   //assert(!CHECK_ZERO_LESS(d_sum2) && !CHECK_ZERO_LESS(d_from2));
   if(FLOAT_BITMASK(d_sum2) < FLOAT_BITMASK(d_from2))
   {
      S_S_DEBUG_DYN("not in range - exit");
      return false;
   }
                              //get distance of volume pos to line
   if(!th->dir_2_sum) return false;
   const float pos_on_line = (w_sphere.pos-th->from).Dot(th->dir) / th->dir_2_sum;
   S_S_DEBUG_DYN(C_fstring("dist w_sphere.pos-vect_to at %.2f", pos_on_line));
   if(pos_on_line <= 0.0f){
      S_S_DEBUG_DYN("bad dir - exit");
      return false;
   }

   S_vector col_normal;
   if(pos_on_line >= 1.0f){
      S_S_DEBUG_DYN("facing sphere");
      col_normal = th->to - w_sphere.pos;
   }else{
      S_S_DEBUG_DYN("crossing sphere");
      S_vector new_to = th->from + th->dir * pos_on_line;
      col_normal = new_to - w_sphere.pos;
   }

   const float d_p2 = col_normal.Dot(col_normal); 

   float radius_sum2 = radius_sum * radius_sum;

   if(d_p2 >= radius_sum2){
      S_S_DEBUG_DYN("no collision - exit");
      return false;
   }
                              //perform response
   float dist = IsMrgZero(d_p2) ? .0f : SqrtFast(d_p2);
   if(dist > MRG_ZERO){
      col_normal /= dist;
      float col_d = -(col_normal.Dot(w_sphere.pos)) - radius_sum;
      {
                              //save collision for later response
         //S_S_DEBUG_DYN(C_fstring("col_d: %.2f, d_to: %.2f, u: %.2f", col_d, d_to, u));
         th->collision_list.push_back(S_collision_info());
         S_collision_info &ci = th->collision_list.back();
         ci.plane.normal = col_normal;
         ci.plane.d = col_d;
         ci.hit_frm = this;
         ci.vol_source = th->vol_source;
                              //dynamic volumes has not plane in bsp tree, assign one negative
         ci.plane_id = -1;
                              //calculate hit distance(by pythagor formula)
         S_vector a_side = (th->from + th->dir * pos_on_line) - w_sphere.pos;
         float dist_to_line2 = a_side.Dot(a_side);
         float pos_on_line_normalized = pos_on_line * th->dir_magn;
                              //b2 = c2 - a2
         float b2 = radius_sum2 - dist_to_line2;
         float impact_dist2 = IsMrgZero(b2) ? .0f : SqrtFast(b2);

         ci.in_dist = Max(.0f, pos_on_line_normalized - impact_dist2);
         //assert(ci.in_dist <= (th->dir_magn + MRG_ZERO));
         S_S_DEBUG_DYN(C_fstring("in_dist %.2f", ci.in_dist));
      }
   }
   //S_S_DEBUG_DYN("collision done");

//#ifdef _DEBUG
   vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif

   return true;
}

//----------------------------

bool I3D_volume::CheckCol_S_B_dyn(const struct S_trace_help *th, bool passive) const{

   assert(th);

   float radius = th->radius;

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_B];
#endif

   int i;

   enum E_SIDE_MODE{          //bits
      SM_MIN_OUT = 1,         //absolutely outside
      SM_MIN_OUT_CROSS = 2,   //crossing, center outside
      SM_MIN_IN_CROSS = 4,    //crossing, center inside
      SM_IN = 8,              //absolutely inside
      SM_MAX_IN_CROSS = 0x10, //crossing, center inside
      SM_MAX_OUT_CROSS = 0x20,//crossing, center outside
      SM_MAX_OUT = 0x40,      //absolutely outside
      SM_MIN_CENTER_OUT = 0x80,  //center less than min axis
      SM_MAX_CENTER_OUT = 0x100, //center greater than max axis
   };

   struct S_sphere_info{
                              //info for all 3 volume-aligned axes
      float d[3];
      dword side_mode[3];
      int num_sides_center_out;
      int solid_side_index;
      S_sphere_info():
         num_sides_center_out(0)
      {}
   };
   S_sphere_info si_from, si_to;

   S_vector edge_col_to;      //must contain copy of (adjusted) th->to before entering edge_col

                              //float bitmask of radius (radius is always positive)
   int radius_i = FLOAT_BITMASK(radius);

                              //compute what's necessary
   S_B_DEBUG_DYN(C_fstring("S_B collision on %s:", GetName()));
   for(i=0; i<3; i++){
      float d_f_min, d_f_max;
      float d_t_min, d_t_max;

                              //compute info for 'from'
      float &d_from = si_from.d[i];
      float &d_to = si_to.d[i];

      d_from = - (normal[i].Dot(th->from));
      d_to = - (normal[i].Dot(th->to));

                              //fast quits - absolute bottleneck - optimized
      d_f_min = d_min[i] + d_from;
      d_t_min = d_min[i] + d_to;
      if(FLOAT_BITMASK(d_f_min) >= radius_i){
         if(FLOAT_BITMASK(d_t_min) >= radius_i){
            S_B_DEBUG_DYN("both 'from' & 'to' out of side, quitting");
            if(FLOAT_BITMASK(d_f_max) < th->critical_dist_i){
               //AddCriticalVolume((PI3D_volume)this);
               S_B_DEBUG_DYN("'from' closer than critical_dist, adding to critical list");
            }
            return false;
         }
                              //no way, compute the rest
                              //'from'
         si_from.side_mode[i] = SM_MIN_OUT | SM_MIN_CENTER_OUT;
         ++si_from.num_sides_center_out;
         si_from.solid_side_index = i;
                              //'to'
         if(CHECK_ZERO_GREATER(d_t_min)){
            si_to.side_mode[i] = SM_MIN_OUT_CROSS | SM_MIN_CENTER_OUT;
            ++si_to.num_sides_center_out;
            si_to.solid_side_index = i;
         }else
         if(CHECK_ZERO_GREATER(d_t_max=d_max[i] - d_to)){
            si_to.side_mode[i] = ((FLOAT_BITMASK(d_t_max) >= radius_i) ? SM_MAX_OUT : SM_MAX_OUT_CROSS) | SM_MAX_CENTER_OUT;
            ++si_to.num_sides_center_out;
            si_to.solid_side_index = i;
         }else
            si_to.side_mode[i] = (CHECK_ZERO_GREATER(d_t_min) || FLOAT_BITMASK(d_t_min) < radius_i) ? SM_MIN_IN_CROSS :
               (CHECK_ZERO_GREATER(d_t_max) || FLOAT_BITMASK(d_t_max) < radius_i) ? SM_MAX_IN_CROSS : SM_IN;
      }else{
         d_f_max = d_max[i] - d_from;
         d_t_max = d_max[i] - d_to;
         if(FLOAT_BITMASK(d_f_max) >= radius_i){
            if(FLOAT_BITMASK(d_t_max) >= radius_i){
               S_B_DEBUG_DYN("both 'from' & 'to' out of side, quitting");
               if(FLOAT_BITMASK(d_f_max) < th->critical_dist_i){
                  //AddCriticalVolume((PI3D_volume)this);
                  S_B_DEBUG_DYN("'from' closer than critical_dist, adding to critical list");
               }
               return false;
            }
                              //no way, compute the rest
                              //'from'
            si_from.side_mode[i] = SM_MAX_OUT | SM_MAX_CENTER_OUT;
            ++si_from.num_sides_center_out;
            si_from.solid_side_index = i;
         }else{
                              //no way, compute the rest
                              //'from'
            if(CHECK_ZERO_GREATER(d_f_min)){
               si_from.side_mode[i] = SM_MIN_OUT_CROSS | SM_MIN_CENTER_OUT;
               ++si_from.num_sides_center_out;
               si_from.solid_side_index = i;
            }else
            if(CHECK_ZERO_GREATER(d_f_max)){
               si_from.side_mode[i] = SM_MAX_OUT_CROSS | SM_MAX_CENTER_OUT;
               ++si_from.num_sides_center_out;
               si_from.solid_side_index = i;
            }else
               si_from.side_mode[i] = (CHECK_ZERO_GREATER(d_f_min) || FLOAT_BITMASK(d_f_min) < radius_i) ? SM_MIN_IN_CROSS :
                  (CHECK_ZERO_GREATER(d_f_max) || FLOAT_BITMASK(d_f_max) < radius_i) ? SM_MAX_IN_CROSS : SM_IN;
         }
                              //'to'
         if(CHECK_ZERO_GREATER(d_t_min)){
            si_to.side_mode[i] = ((FLOAT_BITMASK(d_t_min) >= radius_i) ? SM_MIN_OUT : SM_MIN_OUT_CROSS) | SM_MIN_CENTER_OUT;
            ++si_to.num_sides_center_out;
            si_to.solid_side_index = i;
         }else
         if(CHECK_ZERO_GREATER(d_t_max)){
            si_to.side_mode[i] = ((FLOAT_BITMASK(d_t_max) >= radius_i) ? SM_MAX_OUT : SM_MAX_OUT_CROSS) | SM_MAX_CENTER_OUT;
            ++si_to.num_sides_center_out;
            si_to.solid_side_index = i;
         }else
            si_to.side_mode[i] = (CHECK_ZERO_GREATER(d_t_min) || FLOAT_BITMASK(d_t_min) < radius_i) ? SM_MIN_IN_CROSS :
               (CHECK_ZERO_GREATER(d_t_max) || FLOAT_BITMASK(d_t_max) < radius_i) ? SM_MAX_IN_CROSS : SM_IN;
      }
                              //dump info
      S_B_DEBUG_DYN(C_fstring("%i: d_from: %.2f d_to: %.2f dfmax: %.2f dfmin: %.2f dtmax: %.2f dtmin: %.2f, from.side_mode = 0x%x, to.side_mode = 0x%x",
         i, si_from.d[i], si_to.d[i], d_f_max, d_f_min, d_t_max, d_t_min,
         si_from.side_mode[i], si_to.side_mode[i]));
      S_B_DEBUG_DYN(C_fstring("   from.side_mode = 0x%x, to.side_mode = 0x%x",         
         si_from.side_mode[i], si_to.side_mode[i]));
   }
   S_B_DEBUG_DYN(C_fstring("from.num_sides_center_out = %i, to.num_sides_center_out = %i", si_from.num_sides_center_out, si_to.num_sides_center_out));

   S_vector coll_point;       //used much much later

                              //info for solid collision
   int solid_axis_index = -1; //put invalid value here
   bool solid_max_side;

   if(passive){
      if(si_from.num_sides_center_out<2)
      if(!((si_from.side_mode[0]|si_from.side_mode[1]|si_from.side_mode[2])&(SM_MIN_OUT|SM_MAX_OUT))){
         S_B_DEBUG_DYN("Fast quit 2 - 'from' touching box, passive test succeeded");
//#ifdef _DEBUG
         vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
         return true;
      }
      if(si_to.num_sides_center_out<2)
      if(!((si_to.side_mode[0]|si_to.side_mode[1]|si_to.side_mode[2])&(SM_MIN_OUT|SM_MAX_OUT))){
         S_B_DEBUG_DYN("Fast quit 2 - 'to' touching box, passive test succeeded");
//#ifdef _DEBUG
         vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
         return true;
      }
   }else{   //passive
                              //fast quit 3 - comming from inside
      if(!si_from.num_sides_center_out){
         S_B_DEBUG_DYN("Fast quit 3 - comming from inside");
         return false;
      }
      if(si_from.num_sides_center_out==1){
         if(!si_to.num_sides_center_out){
            solid_axis_index = si_from.solid_side_index;
            solid_max_side = (si_from.side_mode[solid_axis_index]&SM_MAX_CENTER_OUT);
            goto solid_col;
         }
         if(si_to.num_sides_center_out==1 && (si_from.solid_side_index == si_to.solid_side_index)){
                              //check if we're comming towards the side
            float d_delta = si_to.d[si_from.solid_side_index]-si_from.d[si_from.solid_side_index];
            if(si_from.side_mode[si_from.solid_side_index]&SM_MAX_CENTER_OUT){
               if(d_delta<=0.0f){
                  S_B_DEBUG_DYN("Fast quit - going out of side");
                  return false;
               }
            }else{
               if(d_delta>=0.0f){
                  S_B_DEBUG_DYN("Fast quit - going out of side");
                  return false;
               }
            }
            solid_axis_index = si_from.solid_side_index;
            solid_max_side = (si_from.side_mode[solid_axis_index]&SM_MAX_CENTER_OUT);
            goto solid_col;
         }
      }
   }
   int hit_index[3];
   float hit_dist[3];         //with min|max side
   bool hit_front[3];

   int hit_sorted_axes[3];

   for(i=0; i<3; i++){
      if(si_from.side_mode[i]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
         hit_index[i] = bool(si_from.side_mode[i]&SM_MAX_CENTER_OUT);
         hit_front[i] = true;
      }else{
         hit_index[i] = bool(si_from.d[i]>si_to.d[i]);
         hit_front[i] = false;
      }

      float u;
      u = normal[i].Dot(th->dir);
      if(IsAbsMrgZero(u)){
         hit_dist[i] = 1e+10f;
      }else
         hit_dist[i] = (si_from.d[i] - (!hit_index[i] ? -d_min[i] : d_max[i])) / u;
      S_B_DEBUG_DYN(C_fstring("hit_index = %i, side hit at distance %.2f, hit_front = %i", hit_index[i], hit_dist[i], hit_front[i]));
   }
                              //sort axes by hit distance
   if(hit_dist[0]<=hit_dist[1]){
      if(hit_dist[0]<=hit_dist[2]){
         hit_sorted_axes[0] = 0;

         if(hit_dist[1]<=hit_dist[2]) hit_sorted_axes[1]=1, hit_sorted_axes[2]=2;
         else hit_sorted_axes[1]=2, hit_sorted_axes[2]=1;
      }else{
         hit_sorted_axes[0] = 2;
         hit_sorted_axes[1]=0, hit_sorted_axes[2]=1;
      }
   }else
   if(hit_dist[0]<=hit_dist[2]){
      hit_sorted_axes[0] = 1;
      hit_sorted_axes[1]=0, hit_sorted_axes[2]=2;
   }else{
      if(hit_dist[1]<=hit_dist[2]){
         hit_sorted_axes[0] = 1;
         hit_sorted_axes[1] = 2;
      }else{
         hit_sorted_axes[0] = 2;
         hit_sorted_axes[1] = 1;
      }
      hit_sorted_axes[2]=0;
   }
   S_B_DEBUG_DYN(C_fstring("sorted axes: %i %i %i", hit_sorted_axes[0], hit_sorted_axes[1], hit_sorted_axes[2]));

   if(si_from.num_sides_center_out==1){
                              //we're comming from inside of single side

                              //check if closest side is the one 'out' - if yes, it's solid collision
      if(si_from.side_mode[hit_sorted_axes[0]]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
         if(passive){
            S_B_DEBUG_DYN("hitting cube - test successfull");
//#ifdef _DEBUG
            vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
            return true;
         }
         solid_axis_index = hit_sorted_axes[0];
         solid_max_side = (hit_index[solid_axis_index]);
         goto solid_col;
      }
                              //this may be edge or corner collision now,

                              //get edge of first closest sides
      S_vector edge_pt = bbox.min;
      if(hit_index[hit_sorted_axes[0]]) edge_pt += side[hit_sorted_axes[0]];
      if(hit_index[hit_sorted_axes[1]]) edge_pt += side[hit_sorted_axes[1]];
      const S_vector &edge_dir = side[hit_sorted_axes[2]];
      S_B_DEBUG_DYN(C_fstring("checking against edge, point = %.2f, %.2f, %.2f, dir = %.2f, %.2f, %.2f",
         edge_pt.x, edge_pt.y, edge_pt.z,
         edge_dir.x, edge_dir.y, edge_dir.z));

                              //check if fahrest side is the one 'out' - if yes, it may be corner collision
      if(si_from.side_mode[hit_sorted_axes[2]]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
         S_B_DEBUG_DYN("corner collision");
                              //get closest point of 'to' and the point
         if(th->dir_2_sum < MRG_ZERO) return false;
         coll_point = edge_pt;
         float u = th->dir.Dot(coll_point-th->from) / th->dir_2_sum;
         if(u<=.005f){  //passed the corner?
            S_B_DEBUG_DYN("Fast quit 6 - past the corner");
            return false;
         }
         S_B_DEBUG_DYN(C_fstring("Closest point to corner at distance %.2f", u));
         edge_col_to = th->to;
         if(u < 1.0f) edge_col_to = th->from + th->dir*u;
      }else{
         S_B_DEBUG_DYN("edge collision");
         float u, u1;
                              //should always pass
         if(!LineIntersection(th->from, th->dir, edge_pt, edge_dir, u, u1)){
            //todo("test this!!!");
                              //too small th->dir...? solve as distance of point to line
            u = side_dot[hit_sorted_axes[2]];
            if(IsAbsMrgZero(u))  //is this possible???
               return false;
            u1 = edge_dir.Dot(th->to-edge_pt) / u;
            u = 1.0f;
         }
         S_B_DEBUG_DYN(C_fstring("Collision of edge and th->dir at distances %.2f, %.2f", u1, u));
         edge_col_to = th->to;
         if(u <= 1.0f){
            if(IsAbsMrgZero(u)) return false;
            coll_point = edge_pt + edge_dir * u1;
            u = 1.0f/u;
            coll_point = th->from + (coll_point-th->from)*u;
            radius *= u;
         }else{
                        //re-compute closest point on edge
            u1 = edge_dir.Dot(th->to-edge_pt) / side_dot[hit_sorted_axes[2]];
            S_B_DEBUG_DYN(C_fstring("re-computing point on edge at distance %.2f", u1));
            coll_point = edge_pt + edge_dir * u1;
         }
      }
   }else{
      int master_axis;        //facing and farer
      int slave_axis;         //facing and closer
      int third_axis;

      for(i=3; i--; ) if(hit_front[hit_sorted_axes[i]]) break;
      master_axis = hit_sorted_axes[i];
      if(hit_dist[master_axis]<=0.0f){
         S_B_DEBUG_DYN("Fast quit - going out of master axis");
         return false;
      }

      for(; i--; ) if(hit_front[hit_sorted_axes[i]]) break;
      assert(i!=-1);
      slave_axis = hit_sorted_axes[i];
      third_axis = 3-(master_axis+slave_axis);

      S_B_DEBUG_DYN(C_fstring("master_axis = %i, slave_axis = %i", master_axis, slave_axis));

                              //edge_side_1 is alway master_axis
      int edge_side_2;
      bool var_limit_to_side;
                              //for case of solid col
      solid_axis_index = master_axis;
      solid_max_side = (si_from.side_mode[master_axis]&SM_MAX_CENTER_OUT);

                              //detect if collision with master/slave edge
      if(si_from.side_mode[slave_axis]&si_to.side_mode[slave_axis]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
         edge_side_2 = slave_axis;
         //var_limit_to_side=true;
         var_limit_to_side=false;
         edge_col_to = th->to;
         goto edge_col;
      }
                              //detect if collision with master/slave-1 edge
                              //if 'to' is on opposite side of box...
      if(si_to.side_mode[slave_axis]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
                              //...and dist to master is greater than to slave-1

         float u = normal[slave_axis].Dot(th->dir);
         assert(!IsAbsMrgZero(u));
         if((float)fabs(u)>.001f){
            hit_dist[slave_axis] = -(normal[slave_axis].Dot(th->from) + (hit_index[slave_axis] ? -d_min[slave_axis] : d_max[slave_axis])) / u;

            if(hit_dist[slave_axis] < hit_dist[master_axis]){
               hit_index[slave_axis] = !hit_index[slave_axis];

               edge_side_2 = slave_axis;
               var_limit_to_side=false;
               edge_col_to = th->to;
               goto edge_col;
            }
         }
      }
                              //detect if collision with master/third edge
      if(si_to.side_mode[third_axis]&(SM_MIN_CENTER_OUT|SM_MAX_CENTER_OUT)){
                              //...and dist to master is greater than to third

                              //flip third axis if necessary
         if(si_from.num_sides_center_out==3)
         if(!(bool((si_to.side_mode[third_axis]&SM_MAX_CENTER_OUT)&hit_index[third_axis]))){
                              //recompute hit dist on third axis
            float u = normal[third_axis].Dot(th->dir);
            assert(!IsAbsMrgZero(u));
            if((float)fabs(u)>.001f){
               hit_dist[third_axis] = -(normal[third_axis].Dot(th->from) + (hit_index[third_axis] ? -d_min[third_axis] : d_max[third_axis])) / u;
            }else goto solid_col_p;
            hit_index[third_axis] = !hit_index[third_axis];
         }

         if(hit_dist[third_axis] < hit_dist[master_axis]){
            edge_side_2 = third_axis;
            var_limit_to_side=false;
            edge_col_to = th->to;
            goto edge_col;
         }
      }
      goto solid_col_p;

edge_col:
      S_vector edge_pt = bbox.min;
      if(hit_index[master_axis]) edge_pt += side[master_axis];
      if(hit_index[edge_side_2]) edge_pt += side[edge_side_2];
      i = 3-(master_axis+edge_side_2);
      const S_vector &edge_dir = side[i];
      const float &edge_dot = side_dot[i];

      S_B_DEBUG_DYN("edge collision");
      float u, u1;
                           //should always pass
      if(!LineIntersection(th->from, th->dir, edge_pt, edge_dir, u, u1)){
                              //too small th->dir...? solve as distance of point to line
         S_B_DEBUG_DYN("LineIntersection failed, solving point-to-line distance");

         u = edge_dot;
         if(IsAbsMrgZero(u)) return false;   //is this possible???
         u1 = edge_dir.Dot(edge_col_to-edge_pt) / u;
         u = 1.0f;
      }
      S_B_DEBUG_DYN(C_fstring("Collision of edge and th->dir at distances %.2f, %.2f", u1, u));

      if(u < 1.0f){
                           //limit edge range - we get corners for free at extremities
         u1 = Max(0.0f, Min(1.0f, u1));
         coll_point = edge_pt + edge_dir * u1;

         if(var_limit_to_side){
                           //fail to solid
            goto solid_col_p;
         }else
         if(u<0.0f){
            S_B_DEBUG_DYN("Fast quit - going out of edge");
            return false;
         }else{
            if(IsMrgZero(u)) return false;
            //edge_col_to = th->from + th->dir * u;
            edge_col_to = th->to;

            u = 1.0f/u;
            coll_point = th->from + (coll_point-th->from)*u;
            radius *= u;
         }
      }else{
                     //re-compute closest point on edge - avoid cutting into the edge
         u = edge_dot;
         if(IsAbsMrgZero(u)) return false;   //is this possible???
         u1 = edge_dir.Dot(edge_col_to-edge_pt) / u;
         S_B_DEBUG_DYN(C_fstring("re-computing point on edge at distance %.2f", u1));
                           //limit edge range - we get corners for free at extremities
         u1 = Max(0.0f, Min(1.0f, u1));
         coll_point = edge_pt + edge_dir * u1;
      }
   }

                              //solve collision to given point
   {
      S_vector col_normal = edge_col_to - coll_point;

      float dist_2 = col_normal.Dot(col_normal);
      if(dist_2<MRG_ZERO){
                              //this shoudn't happen
         if(solid_axis_index==-1) return false;
         goto solid_col_p;
      }
      float dist = SqrtFast(dist_2);

                              //fast quit 5 - missing closest point
      if(dist >= radius){
         S_B_DEBUG_DYN("Fast quit 5 - missing closest point");
         return false;
      }
      if(passive){
         S_B_DEBUG_DYN("hitting edge/corner - test successful");
//#ifdef _DEBUG
   vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
         return true;
      }
      S_B_DEBUG_DYN("solving edge/corner collision");
                              //normalize collision normal
      col_normal /= dist;
      float col_d = - col_normal.Dot(coll_point);
                              //collision response
                              //save collision for later response
      {
         //S_B_DEBUG_DYN(C_fstring("col_d: %.2f, d_to: %.2f, u: %.2f", col_d, d_to, u));
         th->collision_list.push_back(S_collision_info());
         S_collision_info &ci = th->collision_list.back();
         ci.plane.normal = col_normal;
         ci.plane.d = col_d - radius;
         ci.hit_frm = this;
         ci.vol_source = th->vol_source;
                              //dynamic volumes has not plane in bsp tree, assign one negative
         ci.plane_id = -1;
//----------------------------
         {
                              //compute in distace:
            float f = ci.plane.normal.Dot(th->dir_norm);
            if(fabs(f)<MRG_ZERO){
               ci.in_dist = .0f;
            }
            else{
               float in_dist = -(ci.plane.normal.Dot(th->from)+ci.plane.d) / f;
               ci.in_dist = Min(Max(.0f, in_dist), th->dir_magn);
            }
         }
//----------------------------
      }
   }

//#ifdef _DEBUG
   vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
   return true;

solid_col_p:
   if(passive){
//#ifdef _DEBUG
      vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
      return true;
   }
solid_col:
   assert(!passive);
   S_B_DEBUG_DYN(C_fstring("solving solid collision in axis %i", solid_axis_index));
                              //perform solid response on side
                              //save collision for later response
   {
      //S_B_DEBUG_DYN(C_fstring("col_d: %.2f, d_to: %.2f, u: %.2f", col_d, d_to, u));
      th->collision_list.push_back(S_collision_info());
      S_collision_info &ci = th->collision_list.back();
      ci.plane.normal = solid_max_side ? normal[solid_axis_index] : -normal[solid_axis_index];
      ci.plane.d = solid_max_side ? d_max[solid_axis_index] - radius : d_min[solid_axis_index] - radius;
      ci.hit_frm = this;
      ci.vol_source = th->vol_source;
                           //dynamic volumes has not plane in bsp tree, assign one negative
      ci.plane_id = -1;
//----------------------------
      {
                              //compute in distace:
         float f = ci.plane.normal.Dot(th->dir_norm);
         if(fabs(f)<MRG_ZERO){
            ci.in_dist = .0f;
         }
         else{
            float in_dist = -(ci.plane.normal.Dot(th->from)+ci.plane.d) / f;
            ci.in_dist = Min(Max(.0f, in_dist), th->dir_magn);
         }
      }
//----------------------------
   }
//#ifdef _DEBUG
   vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
   return true;
}

//----------------------------

bool I3D_volume::CheckCol_S_R_dyn(const struct S_trace_help *th) const{

   assert(th);

#ifdef VOLUME_STATISTICS
   ++drv->col_stats[I3D_driver::DRV_COL_STATS_S_R];
#endif

   int i;
   float u;

   float d_from[3];
   float d_to[3];
   float dist_from_max[3];
   float dist_from_min[3];
   float dist_to_max[3];
   float dist_to_min[3];

                              //compute what's necessary
                              //check sides out
   S_R_DEBUG_DYN(C_fstring("volume %s distances:", GetName()));
   for(i=3; i--; ){
      d_from[i] = - normal[i].Dot(th->from);
      dist_from_max[i] = d_max[i] - d_from[i];
      dist_from_min[i] = d_min[i] + d_from[i];

      d_to[i] = - normal[i].Dot(th->to);
      dist_to_max[i] = d_max[i] - d_to[i];
      dist_to_min[i] = d_min[i] + d_to[i];

      S_R_DEBUG_DYN(C_fstring("%i: d_from: %.2f d_to: %.2f dfmax: %.2f dfmin: %.2f dtmax: %.2f dtmin: %.2f",
         i, d_from[i], d_to[i], dist_from_max[i], dist_from_min[i], dist_to_max[i], dist_to_min[i]));

      if((dist_from_max[i] >= th->radius && dist_to_max[i] >= th->radius) ||
         (dist_from_min[i] >= th->radius && dist_to_min[i] >= th->radius)){

         S_R_DEBUG_DYN(C_fstring("out of axis %i - exit", i));
         return false;
      }
   }

   bool crossing_z = (dist_from_min[2]*dist_to_min[2])<0.0f;

   int col_edge_count=0;
   bool max_side[3];
   int coll_axis[2];

   max_side[2] = (d_max[2] > d_from[2]);
   bool z_max_dir = (d_to[2] < d_from[2]);

   S_R_DEBUG_DYN(C_fstring("crossing_z = %i z_max_dir = %i", crossing_z, z_max_dir));

   for(i=0; i<2; i++){

      if(crossing_z){
                              ////get point on plane
         S_vector pp;
         u = normal[2].Dot(th->dir);
         if(!IsAbsMrgZero(u)){
            u = - (normal[2].Dot(th->from) + d_max[2]) / u;
            pp = th->from + th->dir * u;
            S_R_DEBUG_DYN(C_fstring("getting point on plane at dist. %.2f", u));
         }else{
            S_R_DEBUG_DYN("dir parallel with plane");
         }

         float d = -normal[i].Dot(pp);
         float p_dist_from_max = d_max[i] - d;
         float p_dist_from_min = d_min[i] + d;

         if(p_dist_from_max>0.0f){
                                 //out of rect
                                 //test edge
            coll_axis[col_edge_count++] = i;
            max_side[i] = true;
         }else
         if(p_dist_from_min>0.0f){
            coll_axis[col_edge_count++] = i;
            max_side[i] = false;
         }
      }else{
                              //now it depends on from we are coming
         if(dist_from_min[i] >= 0.0f){
            if(z_max_dir == max_side[2]){
                              //hit closest edge
               if(dist_from_min[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = false;
               }else
               if(dist_from_max[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = true;
               }
            }else
                              //going from min side
            if(dist_to_min[i]>0.0f){
               coll_axis[col_edge_count++] = i;
               max_side[i] = false;
            }else
            if(dist_to_max[i]>0.0f){
               coll_axis[col_edge_count++] = i;
               max_side[i] = true;
            }
         }else
         if(dist_from_max[i] >= 0.0f){
            if(z_max_dir == max_side[2]){
                              //hit closest edge
               if(dist_from_max[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = true;
               }else
               if(dist_from_min[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = false;
               }
            }else
                              //going from max side
            if(dist_to_max[i]>0.0f){
               coll_axis[col_edge_count++] = i;
               max_side[i] = true;
            }else
            if(dist_to_min[i]>0.0f){
               coll_axis[col_edge_count++] = i;
               max_side[i] = false;
            }
         }else{
            if(z_max_dir == max_side[2]) continue;

            if(dist_from_max[i] > dist_to_max[i]){
               if(dist_to_min[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = false;
               }
            }else{
               if(dist_to_max[i]>0.0f){
                  coll_axis[col_edge_count++] = i;
                  max_side[i] = true;
               }
            }
         }
      }
   }

   coll_axis[1] = 1-coll_axis[0];   //for case it's not filled

   S_R_DEBUG_DYN(C_fstring("col_edge_count = %i max_side = %i %i",
      col_edge_count, max_side[0], max_side[1]));
   S_R_DEBUG_DYN(C_fstring("sorted axes: %i %i", coll_axis[0], coll_axis[1]));

   if(col_edge_count){

      S_vector coll_pt(bbox.min);
      S_vector col_normal;
      float u1, u2;
      float dist;
      float col_d;
      float d_to1;

      --col_edge_count;
      if(!col_edge_count){
                              //pure edge collision
         if(i=coll_axis[0], max_side[i]) coll_pt += side[i];

         bool b = LineIntersection(coll_pt, side[coll_axis[1]], th->from, th->dir, u1, u2);
         if(!b)
            return false;

         S_R_DEBUG_DYN(C_fstring("solving edge collision: line/line col. at %.2f, %.2f", u1, u2));
         if(u2 <= 0.0f) return false;

         if(u2<1.0f) col_normal = th->from + th->dir*u2;
         else{
            col_normal = th->to;
                              //recompute closest point on edge
            u1 = side[coll_axis[1]].Dot(col_normal-coll_pt) / side_dot[coll_axis[1]];
            S_R_DEBUG_DYN(u);
         }
         coll_pt += side[coll_axis[1]]*u1;
      }else{
                              //corner collision (probably)
                              //check both edges
         S_vector coll_pt1 = coll_pt;
         if(i=coll_axis[0], max_side[i]) coll_pt += side[i];
         if(i=coll_axis[1], max_side[i]) coll_pt1 += side[i];

         bool out1, out2;

         out1 = !LineIntersection(coll_pt, side[coll_axis[1]], th->from, th->dir, u1, u2);

         float u1a, u2a;
         out2 = !LineIntersection(coll_pt1, side[coll_axis[0]], th->from, th->dir, u1a, u2a);

                              //check which edge is out
         if(!out1) out1 = (u1<0.0f || u1>1.0f);
         if(!out2) out2 = (u1a<0.0f || u1a>1.0f);
         if(out1 && out2){
            if(IsMrgZero(th->dir_2_sum)) return false;
                              //corner
            if(i=coll_axis[1], max_side[i]) coll_pt += side[i];
                              //get dist to the line
            u2 = (coll_pt-th->from).Dot(th->dir) / th->dir_2_sum;

            if(u2 <= 0.0f) return false;

            S_R_DEBUG_DYN(C_fstring("solving corner collision: line/pt dist at %.2f", u2));
            u2 = Min(1.0f, u2);
            col_normal = th->from + th->dir*u2;
         }else{
                              //pure edge
            if(out1){
                              //swap edges - shouldn't happen in ideal conditions
               u1=u1a;
               u2=u2a;
               coll_pt=coll_pt1;
               swap(coll_axis[1], coll_axis[0]);
            }
            if(u2 <= 0.0f) return false;
            coll_pt += side[coll_axis[1]]*u1;
            if(u2 < 1.0f) col_normal = th->from + th->dir*u2;
            else col_normal = th->to;
         }
      }

      col_normal -= coll_pt;
      dist = col_normal.Magnitude();
      if(dist >= th->radius){
         S_R_DEBUG_DYN("too far, exiting");
         return false;
      }
      if(dist<MRG_ZERO) goto edge_close;
                              //normalize collision normal
      col_normal /= dist;

      col_d = - col_normal.Dot(coll_pt);
      d_to1 = - col_normal.Dot(th->to);
      col_d -= th->radius;

      if(u2<1.0f){
         float delta = d_to1-col_d;
         delta /= u2;
         if(delta > th->dir_magn){
            S_R_DEBUG_DYN("delta truncated");
            delta = th->dir_magn;
         }
         d_to1 = col_d + delta*1.05f;
      }
                              //collision response
      {
         th->collision_list.push_back(S_collision_info());
         S_collision_info &ci = th->collision_list.back();
         ci.plane.normal = col_normal;
         ci.plane.d = col_d;
         ci.hit_frm = this;
         ci.vol_source = th->vol_source;
                              //dynamic volumes has not plane in bsp tree, assign one negative
         ci.plane_id = -1;
                                 //compute in distace:
         {
            float f = ci.plane.normal.Dot(th->dir_norm);
            if(fabs(f)<MRG_ZERO){
               ci.in_dist = .0f;
            }
            else{
               float in_dist = -(ci.plane.normal.Dot(th->from)+ci.plane.d) / f;
               ci.in_dist = Min(Max(.0f, in_dist), th->dir_magn);
            }
         }
      }
      goto exit_true;
   }
edge_close:
   if(!crossing_z){
      if(z_max_dir == max_side[2]) return false;
   }

   S_R_DEBUG_DYN("solving solid z-side collision");
   {
      th->collision_list.push_back(S_collision_info());
      S_collision_info &ci = th->collision_list.back();
      ci.plane.normal = max_side[2] ? normal[2] : -normal[2];
      ci.plane.d = max_side[2] ? d_max[2] - th->radius : d_min[2] - th->radius;
      ci.hit_frm = this;
      ci.vol_source = th->vol_source;
                           //dynamic volumes has not plane in bsp tree, assign one negative
      ci.plane_id = -1;

      {
                              //compute in distace:
         float f = ci.plane.normal.Dot(th->dir_norm);
         if(fabs(f)<MRG_ZERO){
            ci.in_dist = .0f;
         }
         else{
            float in_dist = -(ci.plane.normal.Dot(th->from)+ci.plane.d) / f;
            ci.in_dist = Min(Max(.0f, in_dist), th->dir_magn);
         }
      }
   }

exit_true:

//#ifdef _DEBUG
   vol_flags |= VOLF_DEBUG_TRIGGER;
//#endif
   return true;
}

//----------------------------