/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.

   File: Driver.cpp
   Content: Insanity 3D driver.
--------------------------------------------------------*/

#include "all.h"
#include "mesh.h"
#include <math.h>
#include <conv_rgb.hpp>
#include <isound2.h>
#include "material.h"
#include "light.h"
#include "soundi3d.h"
#include "camera.h"
#include "sector.h"
#include "volume.h"
#include "scene.h"
#include "driver.h"
#include "model.h"
#include "visual.h"
#include "joint.h"
#include "user.h"
#include "dummy.h"
#include "target.h"
#include "procedural.h"
#include "occluder.h"
#include "texture.h"
#include "cpu.h"
#ifdef _MSC_VER
#include <malloc.h>           //alloca
#endif

//----------------------------


//#define USE_PALETIZED_TEXTURES//if source texture is paletized, try to find paletized texture format
//#define OVERRIDE_F_TO_L       //override calls to _ftol with our (faster) version

                              //how many recent shaders we keep in cache
#define VSHADER_CACHE_SIZE 96
#define PSHADER_CACHE_SIZE 32

//#define USE_OWN_FRAME_ALLOCATION 128   //specify heap size (kb), undef to disable own allocation


#ifdef _DEBUG

                              //debugging:
//#define DEBUG_NO_LEAK_WARNING //no warning on unreleased interfaces

//#define DEBUG_NO_MULTI_TEXTURING//use multi-texturing if available by hardware

                              //simulate low of video memory, the number specifies bytes
                              // to grab when initializing
                              // (undef to disable)
//#define DEBUG_SIM_LOWVIDMEM (25*1024*1024)
//#define DEBUG_DISABLE_BUMPMAPPING  //disable using of bump texture formats

//#define DEBUG_FORCE_DIRECT_TRANSFORM   //direct transfrormations, force inclusion of DRVF2_DIRECT_TRANSFORM

//#define DEBUG_FORCE_TXT_CLIPPING //force clipping by textures

//#define DEBUG_NO_NVLINK_OPTIM //disable shader optimizations of VN Linker

//#define DEBUG_NO_MODULATE2X   //disable support for MODULATE2X

#endif

//----------------------------

#ifdef DEBUG_SIM_LOWVIDMEM
static vector<C_smart_ptr<IImage> > low_mem_textures;
#endif

//----------------------------
//----------------------------
static dword num_alloc_bytes;

/*
void *__cdecl operator new(size_t sz){
   void *vp = malloc(sz);
   if(vp) num_alloc_bytes += _msize(vp);
   return vp;
}
void *operator new[](size_t sz){ return operator new(sz);}
void *operator new[](size_t sz, void*vp){ return vp;}

void __cdecl operator delete(void *vp){
   if(vp) num_alloc_bytes -= _msize(vp);
   free(vp);
}
void operator delete[](void *vp){ operator delete(vp); }
*/

//----------------------------

dword GetSizeOfVertex(dword fvf_flags){

   dword rtn = 0;
   switch(fvf_flags&D3DFVF_POSITION_MASK){
   case D3DFVF_XYZ: rtn += sizeof(float) * 3; break;
   case D3DFVF_XYZRHW: rtn += sizeof(float) * 4; break;
   case D3DFVF_XYZB1: rtn += sizeof(float) * 4; break;
   case D3DFVF_XYZB2: rtn += sizeof(float) * 5; break;
   case D3DFVF_XYZB3: rtn += sizeof(float) * 6; break;
   case D3DFVF_XYZB4: rtn += sizeof(float) * 7; break;
   case D3DFVF_XYZB5: rtn += sizeof(float) * 8; break;
   default: assert(0);
   }

   if(fvf_flags&D3DFVF_NORMAL) rtn += sizeof(float) * 3;
   if(fvf_flags&D3DFVF_DIFFUSE) rtn += sizeof(dword);
   if(fvf_flags&D3DFVF_SPECULAR) rtn += sizeof(dword);

   rtn += sizeof(I3D_text_coor) * ((fvf_flags&D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT);

   return rtn;
}

//----------------------------
                              //plugin system - visual creation
typedef PI3D_visual t_CreateVisual(PI3D_driver drv);

struct S_plugin_entry{
   t_CreateVisual *create_proc;
   const char *friendly_name;
   dword visual_type;
   const S_visual_property *props;
   dword num_props;
};

#define MAX_VISUAL_TYPES 32
static S_plugin_entry visual_plugins[MAX_VISUAL_TYPES];
static int num_visual_plugins;

void RegisterVisual(t_CreateVisual *create_proc, dword visual_type, 
   const char *friendly_name, const S_visual_property *props, 
   dword num_props){

   assert(num_visual_plugins < MAX_VISUAL_TYPES);
   visual_plugins[num_visual_plugins].create_proc = create_proc;
   visual_plugins[num_visual_plugins].visual_type = visual_type;
   visual_plugins[num_visual_plugins].friendly_name = friendly_name;
   visual_plugins[num_visual_plugins].props = props;
   visual_plugins[num_visual_plugins].num_props = num_props;
   ++num_visual_plugins;
}


//----------------------------
//----------------------------

#ifdef USE_OWN_FRAME_ALLOCATION

static class C_mem_alloc_manager{
   class C_heap{
      byte *base;
      dword heap_size;

      dword max_free_block;

      struct S_alloc_block{
         enum E_TYPE{
            FREE,
            USED
         } type;
         dword size;
         S_alloc_block(E_TYPE t, dword sz):
            type(t), size(sz)
         {}
      };
                              //list of blocks
      list<S_alloc_block> blocks;
      typedef list<S_alloc_block>::iterator bl_iter;
   public:
      C_heap(dword hs = USE_OWN_FRAME_ALLOCATION * 1024):
         base(new byte[heap_size = hs]),
         max_free_block(hs)
      {
                              //create base allocation frame
         blocks.push_back(S_alloc_block(S_alloc_block::FREE, heap_size));
      }
      ~C_heap(){
         delete[] base;
      }
                              //copying - not implemented
      C_heap(const C_heap&);
      C_heap &operator =(const C_heap&);

      dword GetMaxFreeBlock() const{
         return max_free_block;
      }

      void *Allocate(dword size){
                              //search blocks, find free
         byte *p = base;
         dword max_free_bl = 0;
         for(bl_iter it = blocks.begin(); it!=blocks.end(); it++){
            dword block_size = (*it).size;
            if((*it).type == S_alloc_block::FREE){
               if(block_size >= size){
                              //use this block
                  (*it).type = S_alloc_block::USED;
                  dword rest_size = block_size - size;
                  if(rest_size >= 32){
                              //split block - create new after this
                     (*it).size = size;
                     ++it;
                     it = blocks.insert(it, S_alloc_block(S_alloc_block::FREE, rest_size));
                  }
                  if(max_free_block==block_size){
                              //find different max-sized block
                     for(; it!=blocks.end(); it++){
                        if((*it).type == S_alloc_block::FREE){
                           max_free_bl = Max(max_free_bl, (*it).size);
                        }
                     }
                     max_free_block = max_free_bl;
                  }
                  break;
               }
               max_free_bl = Max(max_free_bl, block_size);
            }
            p += block_size;
         }
         return p;
      }

      bool Free(void *vp){
                              //fast test if it's in our area
         byte *bp = (byte*)vp;
         if(bp < base || bp >= (base + heap_size))
            return false;

                              //probably ours pointer, try to delete it
         byte *p = base;
         for(bl_iter it = blocks.begin(); it!=blocks.end(); it++){
            dword block_size = (*it).size;
            if(p==bp){
               assert((*it).type == S_alloc_block::USED);
                              //this is ours block, free it and concanetate with neighbours
               (*it).type = S_alloc_block::FREE;
               bl_iter bl_next = it;
               ++bl_next;
               if(bl_next!=blocks.end() && (*bl_next).type==S_alloc_block::FREE){
                              //join with next block
                  (*it).size += (*bl_next).size;
                  blocks.erase(bl_next);
               }
               if(it!=blocks.begin()){
                  bl_iter bl_prev = it;
                  --bl_prev;
                  if((*bl_prev).type==S_alloc_block::FREE){
                              //join with next block
                     (*bl_prev).size += (*it).size;
                     blocks.erase(it);
                     it = bl_prev;
                  }
               }
               max_free_block = Max(max_free_block, (*it).size);
               break;
            }
            p += block_size;
         }
         assert(it!=blocks.end());
         return true;
      }

      bool IsEmpty() const{
         return (blocks.size()==1);
      }
   };

   list<C_heap*> all_heaps;
   typedef list<C_heap*>::iterator heap_iter;
public:
   ~C_mem_alloc_manager(){
                              //delete all currently allocated heaps,
                              // although we should finish with empty list
      assert(!all_heaps.size());
      for(heap_iter it=all_heaps.begin(); it!=all_heaps.end(); it++){
         delete (*it);
      }
   }

   void *Allocate(dword size){
                              //search free space in all heaps
      for(heap_iter it=all_heaps.begin(); it!=all_heaps.end(); it++){
         if((*it)->GetMaxFreeBlock() >= size)
            break;
      }
      if(it==all_heaps.end()){
                              //create new heap
                              // push to front, so that we search fresh heaps first
         all_heaps.push_front(new C_heap);
         it = all_heaps.begin();
      }
      void *p = (*it)->Allocate(size);
      assert(p);
      return p;
   }

   void Free(void *p){

      for(heap_iter it=all_heaps.begin(); it!=all_heaps.end(); it++){
         if((*it)->Free(p)){
                              //erase empty heaps
            if((*it)->IsEmpty()){
               delete (*it);
               all_heaps.erase(it);
            }
            return;
         }
      }
                              //pointer not found in our heaps???
      assert(0);
   }

} mem_alloc_manager;
#endif

//----------------------------

/*
void *I3D_visual::operator new(size_t sz){
#ifdef USE_OWN_FRAME_ALLOCATION
   return mem_alloc_manager.Allocate(sz);
#else
   return ::operator new(sz);
#endif
}
void I3D_visual::operator delete(void *p){
#ifdef USE_OWN_FRAME_ALLOCATION
   mem_alloc_manager.Free(p);
#else
   ::operator delete(p);
#endif
}

//----------------------------

void *I3D_model::operator new(size_t sz){
#ifdef USE_OWN_FRAME_ALLOCATION
   return mem_alloc_manager.Allocate(sz);
#else
   return ::operator new(sz);
#endif
}
void I3D_model::operator delete(void *p){
#ifdef USE_OWN_FRAME_ALLOCATION
   mem_alloc_manager.Free(p);
#else
   ::operator delete(p);
#endif
}
*/

//----------------------------
//----------------------------

I3D_RESULT __declspec(dllexport) I3DAPI __stdcall I3DCreate(I3D_driver **drvp, const char *dll_filename){

   *drvp = new I3D_driver(dll_filename);
   return *drvp ? I3D_OK : I3DERR_OUTOFMEM;
}

//----------------------------
//----------------------------

dword CountBits(dword val){
   dword mask = 0x80000000;
   dword num = 0;
   do{
      if(val&mask) ++num;
   }while(mask>>=1);
   return num;
}

//----------------------------

dword CountBits(const S_pixelformat &fmt){

   if(fmt.flags&PIXELFORMAT_PALETTE)
      return 8;
   dword val = (fmt.a_mask | fmt.r_mask | fmt.g_mask | fmt.b_mask);
   dword mask = 0x80000000;
   dword num = 0;
   do{
      if(val&mask) ++num;
   }while(mask>>=1);
   return num;
}

//----------------------------

inline dword Rotate(dword dw, int num_bits){
   if(num_bits>=0) return dw<<num_bits;
   else return dw>>(-num_bits);
}

//----------------------------

static const char *inteface_name[I3D_CLID_LAST] = {
   "I3D_driver",
   "I3D_scene",
   "I3D_frame",
   "I3D_dummy",
   "I3D_visual",
   "I3D_model",
   "I3D_light",
   "I3D_sound",
   "I3D_sector",
   "I3D_camera",
   "I3D_target",
   "I3D_volume",
   "I3D_user",
   "I3D_bone",
   "I3D_occluder",
   "I3D_singlemesh",

   "I3D_texture",
   "I3D_material",
   "I3D_mesh_base",
   "I3D_mesh_object",
   "I3D_portal",
   "I3D_animation_base",
   "I3D_animation_set",
   "I3D_interpolator",
   "I3D_anim_interpolator",
   "I3D_procedural",
};

//----------------------------
//----------------------------

inline byte BitCount(dword v){
   dword r = 0;
   dword mask = 1;
   do{
      if(v&mask) ++r;
   }while(mask<<=1);
   return r;
}

//----------------------------

static char copyright_info[] =
   " ***   Copyright (c) 2000 Lonely Cat Games, s.r.o.. All rights reserved. "
   "This library is part of Insanity graphics engine. ***"
;

//----------------------------
//----------------------------

static const D3DFORMAT supported_texture_formats[] = {
   D3DFMT_R8G8B8,
   D3DFMT_A8R8G8B8,
   D3DFMT_X8R8G8B8,
   D3DFMT_R5G6B5,
   D3DFMT_X1R5G5B5,
   D3DFMT_A1R5G5B5,
   D3DFMT_A4R4G4B4,
   D3DFMT_R3G3B2,
   D3DFMT_A8R3G3B2,
   D3DFMT_X4R4G4B4,
   D3DFMT_P8,
                              //compressed
   D3DFMT_DXT1,
   D3DFMT_DXT2,
   D3DFMT_DXT3,
   D3DFMT_DXT4,
   D3DFMT_DXT5,
                              //bump
                              // note: we don't support bump formats with more
                              // than 8 bits per element
   D3DFMT_V8U8,
   D3DFMT_L6V5U5,
   D3DFMT_X8L8V8U8,
   D3DFMT_Q8W8V8U8,
}, supported_zsten_formats[] = {
   D3DFMT_D16,
   D3DFMT_D32,
   D3DFMT_D24S8,
   D3DFMT_D24X4S4,
   D3DFMT_D15S1,
   D3DFMT_D24X8,
   //D3DFMT_D16_LOCKABLE,
};

//----------------------------

enum{
   AF_ALPHA = 1,
   AF_PREMULT = 2,
   AF_INTERPOLATED = 4,
};

I3D_RESULT I3D_driver::FindTextureFormat(S_pixelformat &pf, dword flags) const{

   if(flags&FINDTF_ALPHA1)
      flags &= ~FINDTF_ALPHA;
                              //can't support paletized alpha-textures
   if(flags&FINDTF_ALPHA)
      flags &= ~FINDTF_PALETIZED;
   int i;
   const S_pixelformat *best_fmt = NULL;

   const vector<S_pixelformat> &fmts = texture_formats[!(flags&FINDTF_RENDERTARGET) ? 0 : 1];
   for(i=0; i<fmts.size(); i++){
      const S_pixelformat &new_fmt = fmts[i];

      if(flags&FINDTF_BUMPMAP){
                              //
         if(new_fmt.flags&PIXELFORMAT_BUMPMAP){
            //if(i==15) break;  //debug - force using specified texture format
            if(!best_fmt)
               best_fmt = &new_fmt;
            else{
               //int orig_depth = CountBits(*best_fmt);
               int orig_bpp = best_fmt->bytes_per_pixel;
               //int new_depth = CountBits(new_fmt);
               int new_bpp = new_fmt.bytes_per_pixel;
               if(orig_bpp == new_bpp){
                  int orig_uv_depth = CountBits(best_fmt->u_mask) + CountBits(best_fmt->v_mask);
                  int new_uv_depth = CountBits(new_fmt.u_mask) + CountBits(new_fmt.v_mask);
                  if(orig_uv_depth < new_uv_depth){
                              //trade for bigger UV coverage, if luminance is not requested
                     best_fmt = &new_fmt;
                  }
               }else
               switch(orig_bpp){
               case 2:
                  switch(new_bpp){
                  case 3: case 4:
                     //if(flags&FINDTF_TRUECOLOR) 
                        //best_fmt = &new_fmt;
                     break;
                  }
                  break;
               case 4:
                  //if(!(flags&FINDTF_TRUECOLOR))
                  {
                              //trade for smaller bitdepth
                     if(new_bpp<4 && new_bpp>=2)
                        best_fmt = &new_fmt;
                  }
                  break;
               }
            }
         }
         continue;
      }
#ifndef USE_PALETIZED_TEXTURES
      if(new_fmt.flags&PIXELFORMAT_PALETTE)
         continue;
#endif
                              //skip non-RGB formats
      if(new_fmt.flags&PIXELFORMAT_BUMPMAP)
         continue;

      if(new_fmt.flags&PIXELFORMAT_COMPRESS){
                              //that's compressed format
         if(flags&FINDTF_COMPRESSED){
            //if(i==6) break;  //debug - force using specified texture format
            struct S_hlp{
               static bool GetAlphaFlags(dword four_cc, dword &ret){
                  ret = AF_ALPHA;
                  switch(four_cc){
                  case MAKEFOURCC('D', 'X', 'T', '3'): break;
                  case MAKEFOURCC('D', 'X', 'T', '1'): ret = 0; break;
                  case MAKEFOURCC('D', 'X', 'T', '2'): ret |= AF_PREMULT; break;
                  case MAKEFOURCC('D', 'X', 'T', '4'): ret |= AF_PREMULT;
                              //flow...
                  case MAKEFOURCC('D', 'X', 'T', '5'): ret |= AF_INTERPOLATED; break;
                  default:    //unknown
                     return false;
                  }
                  return true;
               }
            };

            dword new_flags;
            if(!S_hlp::GetAlphaFlags(new_fmt.four_cc, new_flags))
               continue;
            if(new_flags&(AF_PREMULT|AF_INTERPOLATED))
               continue;

            if(!best_fmt || !(best_fmt->flags&PIXELFORMAT_COMPRESS))
               best_fmt = &new_fmt;
            else{
               dword curr_flags;
               S_hlp::GetAlphaFlags(best_fmt->four_cc, curr_flags);
               bool take = false;
                           //strategy:
                           // prefer non-premultiplied explicit alpha
                           // for alpha-textures choose formats fith alpha bits
                           // for non-alpha or alpha1 textures choose color-key format
               if(flags&FINDTF_ALPHA){
                  if(new_flags&AF_ALPHA){
                     if(!(curr_flags&AF_ALPHA)) take = true;
                     else
                     if(!(new_flags&AF_PREMULT) && (curr_flags&AF_PREMULT)) take = true;
                     else
                     if(!(new_flags&AF_INTERPOLATED) && (curr_flags&AF_INTERPOLATED)) take = true;
                  }
               }else{
                  if(!(new_flags&AF_ALPHA)) take = true;
                  else
                  if(!(new_flags&AF_PREMULT) && (curr_flags&AF_PREMULT)) take = true;
                  else
                  if(!(new_flags&AF_INTERPOLATED) && (curr_flags&AF_INTERPOLATED)) take = true;
               }
               if(take)
                  best_fmt = &new_fmt;
            }
         }
      }else{
#if 0
         if(i==19){
            best_fmt = &new_fmt;
            break;  //debug - force using specified texture format
         }
#endif
         if(flags&FINDTF_COMPRESSED){
                              //if we have already compressed format, skip
            if(best_fmt && (best_fmt->flags&PIXELFORMAT_COMPRESS))
               continue;
         }

         if(flags&FINDTF_ALPHA){
            if(new_fmt.flags&PIXELFORMAT_ALPHA){
               if(!best_fmt || !(best_fmt->flags&PIXELFORMAT_ALPHA)){
                  best_fmt = &new_fmt;
               }else{
                  int orig_depth = CountBits(*best_fmt);
                  int new_depth = CountBits(new_fmt);
                  if(orig_depth==new_depth){
                                 //keep all masks sizes ballanced
                     if(abs(CountBits(new_fmt.r_mask)-CountBits(new_fmt.a_mask)) <
                        abs(CountBits(best_fmt->r_mask)-CountBits(best_fmt->a_mask))){
                        best_fmt = &new_fmt;
                     }
                  }else
                  switch(orig_depth){
                  case 8:        //don't know such format, use the other
                  default:
                     best_fmt = &new_fmt;
                     break;
                  case 16:
                     switch(new_depth){
                     case 24: case 32:
                                 //support higher depths only if alpha insufficient
                        if((flags&FINDTF_TRUECOLOR) || (CountBits(best_fmt->a_mask)<3)) 
                           best_fmt = &new_fmt;
                        break;
                     }
                     break;
                  case 24:
                     if(!(flags&FINDTF_TRUECOLOR))
                     switch(new_depth){
                     case 16:    //check aplha sufficiency
                        if(CountBits(new_fmt.a_mask)>=3)
                           best_fmt = &new_fmt;
                        break;
                     }
                     break;
                  case 32:
                     if(!(flags&FINDTF_TRUECOLOR))
                     switch(new_depth){
                     case 16:
                        if(CountBits(new_fmt.a_mask) >=3 )
                           best_fmt = &new_fmt;
                        break;
                     case 24:
                        best_fmt = &new_fmt;
                        break;
                     }
                     break;
                  }
               }
            }
            if(best_fmt && (best_fmt->flags&PIXELFORMAT_ALPHA))
               continue;
                                 //alpha not found yet, try non-alpha
         }
         if(flags&FINDTF_PALETIZED){
            if(!best_fmt){
                              //1st found
               best_fmt = &new_fmt;
            }else
            if(new_fmt.flags&PIXELFORMAT_PALETTE){
               if((flags&FINDTF_ALPHA1) && CountBits(best_fmt->a_mask)==1){
                                 //we've got alpha1 format already
               }else{
                                 //paletized found
                  best_fmt = &new_fmt;
               }
            }else{
                                 //paletized requested, non-pal found
                                 //check RGB bit-count, support
                                 //16-bit, 8-bit, 24-bit, 32-bit
               int orig_depth = CountBits(*best_fmt);
               int new_depth = CountBits(new_fmt);
               if(orig_depth==new_depth){
                  if((flags&FINDTF_ALPHA1) && CountBits(new_fmt.a_mask) ==1 ){
                                 //get alpha1 format
                     best_fmt = &new_fmt;
                  }else          //don't drop alpha1 format
                  if(!((flags&FINDTF_ALPHA1) && CountBits(best_fmt->a_mask)==1)){
                     if(orig_depth==8){
                                 //prefer paletized (new is not)

                     }else       //check for wider resolution
                     if((CountBits(new_fmt.r_mask) + CountBits(new_fmt.g_mask) + CountBits(new_fmt.b_mask)) >
                        (CountBits(best_fmt->r_mask) + CountBits(best_fmt->r_mask) + CountBits(best_fmt->r_mask))){

                        best_fmt = &new_fmt;
                     }
                  }
               }else
               switch(orig_depth){
               case 8:           //RGB format
               default:
                                 //don't drop alpha1 format
                  if(!((flags&FINDTF_ALPHA1) && CountBits(best_fmt->a_mask)==1)){
                                 //don't drop paletized format
                     if(!(best_fmt->flags&PIXELFORMAT_PALETTE))
                     switch(new_depth){
                     case 16:
                        best_fmt = &new_fmt;
                        break;
                     }
                  }
                  break;
               case 24:
                  switch(new_depth){
                  case 8:
                  case 16:
                     best_fmt = &new_fmt;
                     break;
                  }
                  break;
               case 32:
                  switch(new_depth){
                  case 8:
                  case 16:
                  case 24:
                     best_fmt = &new_fmt;
                     break;
                  }
                  break;
               }
            }
         }else{
            if(!best_fmt){
                                 //any is good
               best_fmt = &new_fmt;
            }else
                                 //non-paletized requested
            if(!(new_fmt.flags&PIXELFORMAT_PALETTE)){
                                 //check RGB bit-count, support
                                 //16-bit, 8-bit, 24-bit, 32-bit
               int orig_depth = CountBits(*best_fmt);
               int new_depth = CountBits(new_fmt);
               if(orig_depth==new_depth){
                  if((flags&FINDTF_ALPHA1) && CountBits(new_fmt.a_mask)==1){
                     best_fmt = &new_fmt;
                  }else          //don't drop alpha1 format
                  if(!((flags&FINDTF_ALPHA1) && CountBits(best_fmt->a_mask)==1)){
                                 //check for wider resolution
                     if((CountBits(new_fmt.r_mask) + CountBits(new_fmt.g_mask) + CountBits(new_fmt.b_mask)) >
                        (CountBits(best_fmt->r_mask) + CountBits(best_fmt->r_mask) + CountBits(best_fmt->r_mask))){
                        best_fmt = &new_fmt;
                     }else    //drop alpha format, if not requested
                     if((best_fmt->flags&PIXELFORMAT_ALPHA) && !(new_fmt.flags&PIXELFORMAT_ALPHA)){
                        best_fmt = &new_fmt;
                     }
                  }
               }else
                                 //don't drop alpha1 format
               if(!((flags&FINDTF_ALPHA1) && CountBits(new_fmt.a_mask)==1))
               switch(orig_depth){
               case 8:           //RGB format
               default:
                  switch(new_depth){
                  case 24: case 16: 
                     best_fmt = &new_fmt;
                     break;
                  case 8:        //take this, as it yelds same results (3-3-2)
                  case 32:
                     if(best_fmt->flags&PIXELFORMAT_PALETTE)
                        best_fmt = &new_fmt;
                     break;
                  }
                  break;
               case 16:
                  switch(new_depth){
                  case 24: case 32:
                     if(flags&FINDTF_TRUECOLOR)
                        best_fmt = &new_fmt;
                     break;
                  }
                  break;
               case 24:
                  switch(new_depth){
                  case 16: 
                     if(!(flags&FINDTF_TRUECOLOR))
                        best_fmt = &new_fmt;
                     break;
                  }
                  break;
               case 32:
                  switch(new_depth){
                  case 16:
                     if(!(flags&FINDTF_TRUECOLOR))
                        best_fmt = &new_fmt;
                     break;
                  case 24: 
                     best_fmt = &new_fmt;
                     break;
                  }
                  break;
               }
            }
         }
      }
   }
   if(!best_fmt)
      return I3DERR_TEXTURESNOTSUPPORTED;
   pf = *best_fmt;
   return I3D_OK;
}

//----------------------------

D3DFORMAT I3D_driver::FindZStenFormat(const S_pixelformat *pf) const{

                              //find which format it is
   const vector<S_pixelformat> &fmts = texture_formats[1];
   for(int i=fmts.size(); i--; ){
      if(!memcmp(&fmts[i], pf, sizeof(S_pixelformat)))
         break;
   }
   assert(i!=-1);
   return rtt_zbuf_fmts[i].front();
}

//----------------------------
//----------------------------

#ifdef USE_PREFETCH
static __declspec(naked) void Prefetch3DNow(const void *ptr){
                              //opcode: 0x0f, 0xd, W(&0x8) | reg(&0x7)
   __asm{
      __emit 0x0f
      __emit 0x0d
      __emit 0x01 
      ret
   }
}
static __declspec(naked) void Prefetch3DNow2(const void*, const void*){
   __asm{
      __emit 0x0f
      __emit 0x0d
      __emit 0x01 
      __emit 0x0f
      __emit 0x0d
      __emit 0x02
      ret
   }
}
static __declspec(naked) void Prefetch3DNow64(const void*){
   __asm{
      __emit 0x0f
      __emit 0x0d
      __emit 0x01
      __emit 0x0f
      __emit 0x0d
      __emit 0x41 
      __emit 0x20
      ret
   }
}

//----------------------------

static __declspec(naked) void PrefetchSIMD(const void *ptr){
                              //opcode: 0x0f, 0x18, type(&0x18) | reg(&0x7)
                              // type: x00 = prefetchnta
                              //       x08 = prefetcht0
                              //       x10 = prefetcht1
                              //       x18 = prefetcht2
   __asm{
      __emit 0x0f
      __emit 0x18
      __emit 0x09
      ret
   }
}
static __declspec(naked) void PrefetchSIMD2(const void*, const void*){
   __asm{
      __emit 0x0f
      __emit 0x18
      __emit 0x09
      __emit 0x0f
      __emit 0x18
      __emit 0x0a
      ret
   }
}
static __declspec(naked) void PrefetchSIMD64(const void*){
   __asm{
      __emit 0x0f
      __emit 0x18
      __emit 0x09
      __emit 0x0f
      __emit 0x18
      __emit 0x49
      __emit 0x20
      ret
   }
}

//----------------------------

static __declspec(naked) void PrefetchNop(const void *ptr){
   __asm ret
}

t_Prefetch1 *Prefetch1 = PrefetchNop;
t_Prefetch2 *Prefetch2ptr = (t_Prefetch2*)PrefetchNop;
t_Prefetch64 *Prefetch64Bytes = (t_Prefetch64*)PrefetchNop;

//t_SphereInFrustum SphereInFrustum_x86, SphereInFrustum_3DNow, SphereInFrustum_SSE, *SphereInFrustum = SphereInFrustum_x86;

struct S_CPU_init{
   S_CPU_init(){
                              //detect CPU type
      if(GetCPUCaps(HAS_3DNOW)){
         //MessageBox(NULL, "3DNow! detected!", "", MB_OK);
         Prefetch1 = Prefetch3DNow;
         Prefetch2ptr = Prefetch3DNow2;
         Prefetch64Bytes = Prefetch3DNow64;
         if(GetCPUCaps(HAS_MMX)){
            //SphereInFrustum = SphereInFrustum_3DNow;
         }
      }else
      if(GetCPUCaps(HAS_SIMD)){
         //MessageBox(NULL, "SIMD detected!", "", MB_OK);
         Prefetch1 = PrefetchSIMD;
         Prefetch2ptr = PrefetchSIMD2;
         Prefetch64Bytes = PrefetchSIMD64;
         //SphereInFrustum = SphereInFrustum_SSE;
      }
   }
} CPU_init;

#endif//USE_PREFETCH

//----------------------------
//----------------------------

I3D_driver::I3D_driver(const char *dfn):
   ref(1),
   dll_filename(dfn),
   lod_scale(1.0f),
   clip_txt_stage(-1),
   debug_draw_mat(-1),
   lpD3D8(NULL),
   is_hal(false),
   in_scene_count(0),
   last_vs_handle(0),
   texture_sort_id(0),
   force_lod_index(-1),
   drv_flags(0),
   drv_flags2(0),
   cb_flags(0),
   cb_proc(NULL),
   rgbconv_lm(NULL),
   lm_create_aa_ratio(1),
   other_memory_size(NULL),
   render_view_id(0),

                              //cache
   srcblend(D3DBLEND_ONE),
   dstblend(D3DBLEND_ZERO),
   alpha_mode(0),
   render_time(0),
                              //cache
   //last_render_mat(NULL),
   b_cullnone(false),
   cull_default(D3DCULL_CCW),
   last_blend_flags(I3DBLEND_OPAQUE),
   last_stream0_vb(NULL),
   last_stream0_stride(0),
   shadow_range_n(20.0f),
   shadow_range_f(30.0f)
{
   memset(last_dir_index, 0, sizeof(last_dir_index));
   memset(stage_anisotropy, 0, sizeof(stage_anisotropy));
   stage_txt_op[0] = D3DTOP_MODULATE;
   for(int i=1; i<MAX_TEXTURE_STAGES; i++){
      stage_txt_op[i] = D3DTOP_DISABLE;
   }

   memset(debug_int, 0, sizeof(debug_int));
   memset(debug_float, 0, sizeof(debug_float));
   memset(interface_count, 0, sizeof(interface_count));
   memset(last_texture, 0, sizeof(last_texture));

   //shadow_range_f = 1000.0f;
}

//----------------------------

I3D_driver::~I3D_driver(){

   Close();
}

//----------------------------

#ifdef OVERRIDE_F_TO_L
extern "C" long __cdecl _ftol (float);

#pragma warning(disable:4035)
__declspec(naked) long __cdecl _ftol (float f){
   __asm{
      push eax
      fistp dword ptr [esp]
      pop eax
      ret
   }
}
#pragma warning(disable:4035)
#endif   //OVERRIDE_F_TO_L

//----------------------------

I3D_RESULT I3D_driver::Init(CPI3DINIT isp){

   if(IsInit())
      Close();

   I3D_RESULT ir;

   if(!isp){
      assert(0);
      ir = I3DERR_INVALIDPARAMS;
      goto out;
   }
   {
      h_nvlinker = LoadLibrary("nvlinker.dll");
      if(!h_nvlinker){
         assert(0);
         ir = I3DERR_GENERIC;
         goto out;
      }
      LPCREATELINKER pCreate = (LPCREATELINKER)GetProcAddress(h_nvlinker, "CreateLinker");
      if(!pCreate){
         assert(0);
         ir = I3DERR_GENERIC;
         goto out;
      }
      eOPTIMIZERLEVEL nv_create_flags = OPTIMIZERLEVEL_BASIC;
#ifdef DEBUG_NO_NVLINK_OPTIM
      nv_create_flags = OPTIMIZERLEVEL_NONE;
#endif
                              //create 2 linkers, for both vertex and pixel shaders
      for(int i=0; i<2; i++){
         nv_linker[i] = (*pCreate)(0, i==0 ? nv_create_flags : OPTIMIZERLEVEL_NONE);
         nv_linker[i]->ReserveConstantRange(0, VSC_FIXED_LAST - 1);
      }
                              //find and open shader file
      HRESULT hr;
      bool ok = false;
      {
         HMODULE hi = GetModuleHandle(dll_filename);
         if(hi){
            HRSRC h;
            h = FindResource(hi, (const char*)1000, "DATA");
            if(h){
               int sz = SizeofResource(hi, h);
               HGLOBAL hgl = LoadResource(hi, h);
               if(hgl){
                              //note: there's no way to unlock resources
                  const void *data = LockResource(hgl);
                  if(data){
                     hr = nv_linker[0]->AddFragments(data, sz);
                     if(SUCCEEDED(hr)){
                        h = FindResource(hi, (const char*)1001, "DATA");
                        if(h){
                           sz = SizeofResource(hi, h);
                           hgl = LoadResource(hi, h);
                           if(hgl){
                              data = LockResource(hgl);
                              if(data){
                                 hr = nv_linker[1]->AddFragments(data, sz);
                                 ok = SUCCEEDED(hr);
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
      if(!ok){
         assert(0);
         ir = I3DERR_GENERIC;
         goto out;
      }
      static const char *vs_fragment_names[VSF_LAST] = {
         "transform",
         "transform_matrix_palette_blend",
         "transform_weighted_morph1",
         "transform_weighted_morph2",
         "transform_mirror",
         "test",
         "copy_uv0",
         "copy_uv1",
         "copy_uv0to1",
         "copy_uv1to0",
         "generate_envmap_uv_os_1",
         "generate_envmap_uv_ws_1",
         "generate_envmap_uv_os_2",
         "generate_envmap_uv_ws_2",
         "shift_uv_0",
         "texture_project",
         "txt_clip_os_1",
         "txt_clip_os_2",
         "txt_clip_os_3",
         "txt_clip_ws_1",
         "txt_clip_ws_2",
         "txt_clip_ws_3",
         "fog_simple_os",
         "fog_simple_ws",
         "fog_begin",
         "fog_end",
         "fog_end_halved",
         "fog_height",
         "fog_height_ws",
         "fog_layered_os",
         "fog_layered_ws",
         "fog_point_os",
         "fog_point_ws",
         "light_begin",
         "light_end",
         "light_copy",
         "light_end_alpha",
         "specular_begin",
         "specular_end",
         "light_directional_os",
         "light_directional_ws",
         NULL,
         "light_point_os",
         "light_point_ws",
         NULL,
         "light_point_ambient_os",
         "light_point_ambient_ws",
         "diffuse_copy",
         "diffuse_mul_alpha",
         "copy_alpha",
         "simple_dir_light",
      }, *constant_names[VSC_LAST] = {
         "c_cam_loc_pos",
         "c_light_dir_normal",
         "c_light_dir_diffuse",
         "c_light_point_pos_fr",
         "c_light_point_diffuse_rrd",
         "c_light_layered_plane",
         "c_light_layered_color_r",
         "c_light_layered_pos",
         "c_light_pfog_conetop_rcca",
         "c_light_pfog_conedir_mult",
         "c_light_pfog_color_power",
         "c_txt_clip",
      };
      for(i=0; i<VSF_LAST; i++){
         if(!vs_fragment_names[i]) continue;   //allow NULL names (unused)
         hr = nv_linker[0]->GetFragmentID(vs_fragment_names[i], &vs_fragment_ids[i]);
         CHECK_D3D_RESULT(C_fstring("NVLink: can't find fragment: %s", vs_fragment_names[i]), hr);
      }
      for(i=0; i<VSC_LAST; i++){
         hr = nv_linker[0]->GetConstantID(constant_names[i], &vs_constant_ids[i]);
         if(FAILED(hr)){
            vs_constant_ids[i] = -1;
         }
      }
#ifdef _DEBUG

      static const char *ps_fragment_names[PSF_LAST] = {
         "ps_test",
         "tex_0",
         "tex_01",
         "tex_012",
         "tex_0123",
         "diffuse_copy",
         "texture_copy",
         "mod_stage0",
      };

      for(i=0; i<PSF_LAST; i++){
         if(!ps_fragment_names[i]) continue;   //allow NULL names (unused)
         hr = nv_linker[1]->GetFragmentID(ps_fragment_names[i], &ps_fragment_ids[i]);
         CHECK_D3D_RESULT(C_fstring("NVLink: can't find fragment: %s", ps_fragment_names[i]), hr);
      }
#endif
   }

   drv_flags = 0;
   drv_flags2 = 0;

   lpD3D8 = isp->lp_graph->GetD3DInterface();
   lpD3D8->AddRef();

   lpDev8 = isp->lp_graph->GetD3DDevice();
   lpDev8->AddRef();

   HRESULT hr;
   hr = lpDev8->GetDeviceCaps(&d3d_caps8);
   CHECK_D3D_RESULT("GetDeviceCaps", hr);

   isp->lp_graph->AddCallback(GraphCallback_thunk, this);

   init_data = (*isp);

   drv_flags2 |= DRVF2_LMTRUECOLOR;

   InitD3DResources();

                              //keep references to interfaces
   if(init_data.lp_sound) init_data.lp_sound->AddRef();
   if(init_data.lp_graph) init_data.lp_graph->AddRef();
                              //setup default state
   SetState(RS_LINEARFILTER, true);
   SetState(RS_USEZB, true);
   SetState(RS_CLEAR, true);
   SetState(RS_TEXTUREDITHER, true);
   SetState(RS_DRAWVISUALS, true);
   SetState(RS_DRAWMIRRORS, true);
   SetState(RS_DRAWTEXTURES, true);
   SetState(RS_MIPMAP, true);
   SetState(RS_FOG, true);
   SetState(RS_POINTFOG, true);
   SetState(RS_DITHER, true);
   SetState(RS_USELMAPPING, true);
   SetState(RS_ENVMAPPING, true);
   SetState(RS_BUMPMAPPING, true);
   SetState(RS_USE_OCCLUSION, true);
   SetState(RS_USESHADOWS, true);
   SetState(RS_LOADMIPMAP, true);

                              //reset stats count
   memset(&render_stats, 0, sizeof(render_stats));
   memset(&scene_stats, 0, sizeof(scene_stats));
   memset(&col_stats, 0, sizeof(col_stats));
   memset(&voltree_stats, 0, sizeof(voltree_stats));

                              //init default sound environment
   env_properties.insert(pair<int, S_sound_env_properties>(0, S_sound_env_properties()));

   ir = I3D_OK;

out:
   return ir;
}

//----------------------------

I3D_RESULT I3D_driver::Close(){

   if(!IsInit())
      return I3DERR_NOTINITIALIZED;

   CloseD3DResources();
   managed_textures.clear();
   //gone_textures.clear();

   dbase.Close();

   int i;
   //assert(!used_palettes.size());

   for(i=vs_decl_cache.size(); i--; ){
      delete vs_decl_cache[i];
   }
   vs_decl_cache.clear();


   if(lpDev8){
      lpDev8->Release();
      lpDev8 = NULL;
   }

   init_data.lp_graph->RemoveCallback(GraphCallback_thunk, this);

   lpD3D8->Release();
   lpD3D8 = NULL;

   for(i=0; i<I3DDIR_LAST; i++)
      ClearDirs((I3D_DIRECTORY_TYPE)i);

   if(h_nvlinker){
      for(i=0; i<2; i++){
         if(nv_linker[i]){
            nv_linker[i]->Release();
            nv_linker[i] = NULL;
         }
      }
      FreeLibrary(h_nvlinker);
      h_nvlinker = NULL;
   }

   if(init_data.lp_sound) init_data.lp_sound->Release();
   if(init_data.lp_graph) init_data.lp_graph->Release();

#ifndef DEBUG_NO_LEAK_WARNING

   bool rel_warned = false;
   if(init_data.flags&I3DINIT_WARN_OUT){
                              //check interface counting
      C_string str;
      for(i=0; i<I3D_CLID_LAST; i++)
      if(interface_count[i]){
         if(!str.Size()) str="I3D_driver::Release() - unreleased interfaces:\n";
         str += C_fstring("%s:\t%i\n", inteface_name[i], interface_count[i]);
      }
      if(str.Size()){
         GetGraphInterface()->FlipToGDI();
         MessageBox((HWND)GetGraphInterface()->GetHWND(), str,
            "I3D Driver Destruction", MB_OK);
         rel_warned = true;
      }
   }
   if(!rel_warned){
      assert(!vb_manager_list.size());
      assert(!ib_manager_list.size());
      assert(!hw_vb_list.size());
   }
   last_stream0_vb = NULL;
#endif//DEBUG_NO_LEAK_WARNING

   memset(&init_data, 0, sizeof(init_data));

   return I3D_OK;
}

//----------------------------

dword *I3D_driver::RegisterVSDeclarator(const vector<dword> &decl){

                              //check if such declarator exists
   for(int i=vs_decl_cache.size(); i--; ){
      if(*vs_decl_cache[i]==decl){
         return vs_decl_cache[i]->begin();
      }
   }
   vs_decl_cache.push_back(new vector<dword>(decl));
   return vs_decl_cache.back()->begin();
}

//----------------------------

const I3D_driver::S_vs_shader_entry *I3D_driver::GetVSHandle(const S_vs_shader_entry_in &se){

                              //find shader in cache
                              //a little trick here - pass in S_vs_shader_entry_in&
                              // cast to a S_vs_shader_entry&
                              // both inherit from same base class, and find() only
                              // uses base's operator< for comparing
   vs_set::iterator it = vs_cache.find(*(const S_vs_shader_entry*)&se);
   if(it != vs_cache.end()){
      (*it).last_render_time = render_time;
      return &(*it);
   }

   assert(se.num_fragments <= MAX_VS_FRAGMENTS);

                              //manage cache, don't allow to grow too big
   if(vs_cache.size() >= VSHADER_CACHE_SIZE){
                              //find oldest shader, remove this
      vs_set::iterator it_oldest = vs_cache.end();
      dword oldest_time = 4000;  //don't erase too young shaders
      for(it=vs_cache.begin(); it!=vs_cache.end(); it++){
         dword shd_time = render_time - (*it).last_render_time;
         if(oldest_time < shd_time){
            oldest_time = shd_time;
            it_oldest = it;
         }
      }
      if(it_oldest != vs_cache.end())
         vs_cache.erase(it_oldest);
   }

                              //convert internal fragment IDs into NVLinker IDs
   vector<dword> nv_fragment_ids(se.num_fragments + 1);
   for(int i=se.num_fragments; i--; )
      nv_fragment_ids[i] = vs_fragment_ids[se.fragment_code[i]];
   nv_fragment_ids[se.num_fragments] = 0; //terminator

   C_smart_ptr<INVLinkBuffer> nv_shader;
   HRESULT hr;
   hr = nv_linker[0]->CreateBinaryShader(nv_fragment_ids.begin(), (INVLinkBuffer**)&nv_shader, 0);
   assert(SUCCEEDED(hr));
   if(FAILED(hr))
      return NULL;

#if defined _DEBUG && 0
   {
      C_smart_ptr<INVLinkBuffer> src;
      nv_linker[0]->GetShaderSource((INVLinkBuffer**)&src);
      char *cp = (char*)src->GetPointer();
      int len = src->GetBufferSize();
      len = dtaBin2Text(cp, len);
      cp[len] = 0;
      DEBUG(cp, 3);
   }
#endif
   dword new_vs_handle;

   dword d3d_usage = D3DUSAGE_SOFTWAREPROCESSING;
   if(IsDirectTransform())
      d3d_usage = 0;

   hr = lpDev8->CreateVertexShader(se.shd_decl, (const dword*)nv_shader->GetPointer(), &new_vs_handle, d3d_usage);
   CHECK_D3D_RESULT("CreateVertexShader", hr);
   if(FAILED(hr))
      return NULL;

   pair<vs_set::iterator, bool> pit = vs_cache.insert(se);
   assert(pit.second);
                              //save values
   S_vs_shader_entry &se_out = (*pit.first);
   se_out.vs_handle = new_vs_handle;
   se_out.last_render_time = render_time;
#if defined _DEBUG && 0
                              //init slots by invalid values
   se_out.vscs_cam_loc_pos = -1;
   memset(se_out.vscs_light_param, -1, sizeof(se_out.vscs_light_param));
#endif


   {
                              //parse fragments, get slots of constants which are used
      dword dir_light_count = 0;
      dword pnt_light_count = 0;
      dword lfog_light_count = 0;
      dword pfog_light_count = 0;
      dword constant_count = 0;
      for(i=0; i<se.num_fragments; i++){
         E_VS_FRAGMENT vsf = se.fragment_code[i];

         switch(vsf){
         case VSF_GENERATE_ENVUV_OS_1: case VSF_GENERATE_ENVUV_WS_1:
         case VSF_GENERATE_ENVUV_OS_2: case VSF_GENERATE_ENVUV_WS_2:
            hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_CAM_LOC_POS], 0, &se_out.vscs_cam_loc_pos);
            assert(SUCCEEDED(hr) && se_out.vscs_cam_loc_pos<96);
            break;

         case VSF_LIGHT_DIR_OS:
         case VSF_LIGHT_DIR_WS:
            {
               assert(dir_light_count < MAX_VS_LIGHTS);
               dword &c0 = se_out.vscs_light_param[constant_count++],
                  &c1 = se_out.vscs_light_param[constant_count++];
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_DIR_NORMAL], dir_light_count, &c0);
               assert(SUCCEEDED(hr) && c0<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_DIR_DIFFUSE], dir_light_count, &c1);
               assert(SUCCEEDED(hr) && c1<96);
               ++dir_light_count;
            }
            break;

         case VSF_LIGHT_POINT_OS:
         case VSF_LIGHT_POINT_WS:
         case VSF_LIGHT_POINT_AMB_OS:
         case VSF_LIGHT_POINT_AMB_WS:
            {
               assert(pnt_light_count < MAX_VS_LIGHTS);
               dword &c0 = se_out.vscs_light_param[constant_count++],
                  &c1 = se_out.vscs_light_param[constant_count++];
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_POINT_POS_FR], pnt_light_count, &c0);
               assert(SUCCEEDED(hr) && c0<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_POINT_DIFFUSE_RRD], pnt_light_count, &c1);
               assert(SUCCEEDED(hr) && c1<96);
               ++pnt_light_count;
            }
            break;

         case VSF_FOG_LAYERED_OS:
         case VSF_FOG_LAYERED_WS:
            {
               assert(lfog_light_count < MAX_VS_LIGHTS);
               dword &c0 = se_out.vscs_light_param[constant_count++],
                  &c1 = se_out.vscs_light_param[constant_count++],
                  &c2 = se_out.vscs_light_param[constant_count++];
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_LAYERED_PLANE], lfog_light_count, &c0);
               assert(SUCCEEDED(hr) && c0<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_LAYERED_COL_R], lfog_light_count, &c1);
               assert(SUCCEEDED(hr) && c1<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_LAYERED_POS], lfog_light_count, &c2);
               assert(SUCCEEDED(hr) && c2<96);
               ++lfog_light_count;
            }
            break;

         case VSF_FOG_POINT_OS:
         case VSF_FOG_POINT_WS:
            {
               assert(pfog_light_count < MAX_VS_LIGHTS);
               dword &c0 = se_out.vscs_light_param[constant_count++],
                  &c1 = se_out.vscs_light_param[constant_count++],
                  &c2 = se_out.vscs_light_param[constant_count++];
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_PFOG_CONE_TOP], pfog_light_count, &c0);
               assert(SUCCEEDED(hr) && c0<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_PFOG_CONE_DIR], pfog_light_count, &c1);
               assert(SUCCEEDED(hr) && c1<96);
               hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_LIGHT_PFOG_COLOR], pfog_light_count, &c2);
               assert(SUCCEEDED(hr) && c2<96);
               ++pfog_light_count;
            }
            break;

         case VSF_TEXT_CLIP_OS_1: case VSF_TEXT_CLIP_OS_2: case VSF_TEXT_CLIP_OS_3:
         case VSF_TEXT_CLIP_WS_1: case VSF_TEXT_CLIP_WS_2: case VSF_TEXT_CLIP_WS_3:
            hr = nv_linker[0]->GetConstantSlot(vs_constant_ids[VSC_TXT_CLIP_PLANE], 0, &se_out.vscs_txt_clip);
            assert(SUCCEEDED(hr) && se_out.vscs_txt_clip<96);
            break;
         }
      }
   }
   return &se_out;
}

//----------------------------

const I3D_driver::S_ps_shader_entry *I3D_driver::GetPSHandle(const S_ps_shader_entry_in &se){

                              //find shader in cache
                              //a little trick here - pass in S_ps_shader_entry_in&
                              // cast to a S_ps_shader_entry&
                              // both inherit from same base class, and find() only
                              // uses base's operator< for comparing
   ps_set::iterator it = ps_cache.find(*(const S_ps_shader_entry*)&se);
   if(it != ps_cache.end()){
      (*it).last_render_time = render_time;
      return &(*it);
   }

   assert(se.num_fragments <= 8);

                              //manage cache, don't allow to grow too big
   if(ps_cache.size() >= PSHADER_CACHE_SIZE){
                              //find oldest shader, remove this
      ps_set::iterator it_oldest = ps_cache.end();
      dword oldest_time = 4000;  //don't erase too young shaders
      for(it=ps_cache.begin(); it!=ps_cache.end(); it++){
         dword shd_time = render_time - (*it).last_render_time;
         if(oldest_time < shd_time){
            oldest_time = shd_time;
            it_oldest = it;
         }
      }
      if(it_oldest != ps_cache.end())
         ps_cache.erase(it_oldest);
   }

                              //convert internal fragment IDs into NVLinker IDs
   vector<dword> nv_fragment_ids(se.num_fragments + 1);
   for(int i=se.num_fragments; i--; )
      nv_fragment_ids[i] = ps_fragment_ids[se.fragment_code[i]];
   nv_fragment_ids[se.num_fragments] = 0; //terminator

   C_smart_ptr<INVLinkBuffer> nv_shader;
   HRESULT hr;
   hr = nv_linker[1]->CreateBinaryShader(nv_fragment_ids.begin(), (INVLinkBuffer**)&nv_shader, 0);
   assert(SUCCEEDED(hr));
   if(FAILED(hr))
      return NULL;

#if defined _DEBUG && 1
   {
      C_smart_ptr<INVLinkBuffer> src;
      nv_linker[1]->GetShaderSource((INVLinkBuffer**)&src);
      char *cp = (char*)src->GetPointer();
      int len = src->GetBufferSize();
      len = dtaBin2Text(cp, len);
      cp[len] = 0;
      DEBUG(cp, 3);
   }
#endif
   dword new_ps_handle;

   dword d3d_usage = D3DUSAGE_SOFTWAREPROCESSING;
   if(IsDirectTransform())
      d3d_usage = 0;

   hr = lpDev8->CreatePixelShader((const dword*)nv_shader->GetPointer(), &new_ps_handle);
   //CHECK_D3D_RESULT("CreatePixelShader", hr);
   if(FAILED(hr))
      return NULL;

   pair<ps_set::iterator, bool> pit = ps_cache.insert(se);
   assert(pit.second);
                              //save values
   S_ps_shader_entry &se_out = (*pit.first);
   se_out.shd_handle = new_ps_handle;
   se_out.last_render_time = render_time;

   return &se_out;
}

//----------------------------

void I3D_driver::SetLegacyPixelShader(){

   S_ps_shader_entry_in se;

   if(!last_texture[0]){
                              //pure diffuse color
      se.AddFragment(PSF_V0_COPY);
   }else{
      se.AddFragment(PSF_TEX_0);
      switch(stage_txt_op[0]){
      case D3DTOP_SELECTARG1:
         se.AddFragment(PSF_T0_COPY);
         break;
      case D3DTOP_SELECTARG2:
         se.AddFragment(PSF_V0_COPY);
         break;
      case D3DTOP_MODULATE:
         se.AddFragment(PSF_MOD_STAGE0);
         break;
      default: assert(0);
      }
   }
   const S_ps_shader_entry *se_out = GetPSHandle(se);
   if(se_out){
      HRESULT hr;
      hr = lpDev8->SetPixelShader(se_out->shd_handle);
      CHECK_D3D_RESULT("SetPixelShader", hr);
   }
}

//----------------------------
//----------------------------

static void ConvertD3DpfToI3Dpf(const D3DFORMAT d3d_fmt, S_pixelformat &pf){

   memset(&pf, 0, sizeof(pf));
                              //setup bpp
   switch(d3d_fmt){

   case D3DFMT_P8:
   case D3DFMT_R3G3B2:
      pf.bytes_per_pixel = 1;
      break;

   case D3DFMT_R5G6B5:
   case D3DFMT_X1R5G5B5:
   case D3DFMT_A1R5G5B5:
   case D3DFMT_A4R4G4B4:
   case D3DFMT_A8R3G3B2:
   case D3DFMT_X4R4G4B4:
   case D3DFMT_V8U8:
   case D3DFMT_L6V5U5:
      pf.bytes_per_pixel = 2;
      break;

   case D3DFMT_R8G8B8:
      pf.bytes_per_pixel = 3;
      break;

   case D3DFMT_A8R8G8B8:
   case D3DFMT_X8R8G8B8:
   case D3DFMT_X8L8V8U8:
   case D3DFMT_Q8W8V8U8:
   case D3DFMT_V16U16:
   case D3DFMT_W11V11U10:
      pf.bytes_per_pixel = 4;
      break;
   }
                              //setup alpha/palette/compression/bump
   switch(d3d_fmt){

   case D3DFMT_DXT1:
   case D3DFMT_DXT2:
   case D3DFMT_DXT3:
   case D3DFMT_DXT4:
   case D3DFMT_DXT5:
      {
         pf.flags = PIXELFORMAT_COMPRESS;
         pf.flags |= PIXELFORMAT_ALPHA;
         byte b;
         switch(d3d_fmt){
         case D3DFMT_DXT1: pf.flags &= ~PIXELFORMAT_ALPHA; b = 1; break;
         case D3DFMT_DXT2: b = 2; break;
         case D3DFMT_DXT3: b = 3; break;
         case D3DFMT_DXT4: b = 4; break;
         case D3DFMT_DXT5: b = 5; break;
         default: assert(0);
         }
         pf.four_cc = MAKEFOURCC('D', 'X', 'T', '0' + b);
      }
      break;

   case D3DFMT_P8:
      pf.flags = PIXELFORMAT_PALETTE;
      break;

   case D3DFMT_A8R8G8B8:
   case D3DFMT_A1R5G5B5:
   case D3DFMT_A4R4G4B4:
   case D3DFMT_A8R3G3B2:
      pf.flags = PIXELFORMAT_ALPHA;
      break;

   case D3DFMT_V8U8:
   case D3DFMT_L6V5U5:
   case D3DFMT_X8L8V8U8:
   case D3DFMT_Q8W8V8U8:
   case D3DFMT_V16U16:
   case D3DFMT_W11V11U10:
      pf.flags = PIXELFORMAT_BUMPMAP;
      break;
   }

                              //setup RGB/UV components
   switch(d3d_fmt){
   case D3DFMT_R8G8B8:
   case D3DFMT_X8R8G8B8:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00ff0000;
      pf.g_mask = 0x0000ff00;
      pf.b_mask = 0x000000ff;
      break;
   case D3DFMT_A8R8G8B8:
      pf.a_mask = 0xff000000;
      pf.r_mask = 0x00ff0000;
      pf.g_mask = 0x0000ff00;
      pf.b_mask = 0x000000ff;
      break;
   case D3DFMT_R5G6B5:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x0000f800;
      pf.g_mask = 0x000007e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_X1R5G5B5:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00007c00;
      pf.g_mask = 0x000003e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_A1R5G5B5:
      pf.a_mask = 0x00008000;
      pf.r_mask = 0x00007c00;
      pf.g_mask = 0x000003e0;
      pf.b_mask = 0x0000001f;
      break;
   case D3DFMT_A4R4G4B4:
      pf.a_mask = 0x0000f000;
      pf.r_mask = 0x00000f00;
      pf.g_mask = 0x000000f0;
      pf.b_mask = 0x0000000f;
      break;
   case D3DFMT_X4R4G4B4:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x00000f00;
      pf.g_mask = 0x000000f0;
      pf.b_mask = 0x0000000f;
      break;
   case D3DFMT_R3G3B2:
      pf.a_mask = 0x00000000;
      pf.r_mask = 0x000000e0;
      pf.g_mask = 0x0000001c;
      pf.b_mask = 0x00000003;
      break;
   case D3DFMT_A8R3G3B2:
      pf.a_mask = 0x0000ff00;
      pf.r_mask = 0x000000e0;
      pf.g_mask = 0x0000001c;
      pf.b_mask = 0x00000003;
      break;

   case D3DFMT_V8U8:
      pf.u_mask = 0x00ff;
      pf.v_mask = 0xff00;
      pf.l_mask = 0x0000;
      break;

   case D3DFMT_L6V5U5:
      pf.u_mask = 0x001f;
      pf.v_mask = 0x03e0;
      pf.l_mask = 0xfc00;
      break;      

   case D3DFMT_X8L8V8U8:
      pf.u_mask = 0x000000ff;
      pf.v_mask = 0x0000ff00;
      pf.l_mask = 0x00ff0000;
      break;      

   case D3DFMT_Q8W8V8U8:
      pf.u_mask = 0x000000ff;
      pf.v_mask = 0x0000ff00;
      pf.l_mask = 0x00000000;
      break;      

   case D3DFMT_V16U16:
      pf.u_mask = 0x0000ffff;
      pf.v_mask = 0xffff0000;
      pf.l_mask = 0x00000000;
      break;      

   case D3DFMT_W11V11U10:
      pf.u_mask = 0x000003ff;
      pf.v_mask = 0x001ffc00;
      pf.l_mask = 0x00000000;
      break;      
   }
}

//----------------------------

static const S_vertex_flare flare_coor[NUM_FLARE_VERTS] = {
   -.5f,  .5f,       0.0f, 0.0f,
    .5f,  .5f,       1.0f, 0.0f,
   -.5f, -.5f,       0.0f, 1.0f,
    .5f, -.5f,       1.0f, 1.0f
};

//----------------------------

void I3D_driver::InitD3DResources(){

   {                          //setup blend table
      dword caps = d3d_caps8.SrcBlendCaps;
                              //setup source
      blend_table[0][0] = D3DBLEND_ZERO;
      blend_table[0][D3DBLEND_ZERO] = caps&D3DPBLENDCAPS_ZERO ? D3DBLEND_ZERO : D3DBLEND_ONE;
      blend_table[0][D3DBLEND_ONE] = D3DBLEND_ONE;
      blend_table[0][D3DBLEND_SRCCOLOR] = caps&D3DPBLENDCAPS_SRCCOLOR ? D3DBLEND_SRCCOLOR :
         caps&D3DPBLENDCAPS_SRCALPHA ? D3DBLEND_SRCALPHA : D3DBLEND_ONE;
      blend_table[0][D3DBLEND_INVSRCCOLOR] = caps&D3DPBLENDCAPS_INVSRCCOLOR ? D3DBLEND_INVSRCCOLOR :
         caps&D3DPBLENDCAPS_INVSRCCOLOR ? D3DBLEND_INVSRCALPHA : D3DBLEND_ONE;
      blend_table[0][D3DBLEND_SRCALPHA] = caps&D3DPBLENDCAPS_SRCALPHA ? D3DBLEND_SRCALPHA :
         blend_table[0][D3DBLEND_SRCCOLOR];
      blend_table[0][D3DBLEND_INVSRCALPHA] = caps&D3DPBLENDCAPS_INVSRCALPHA ? D3DBLEND_INVSRCALPHA :
         blend_table[0][D3DBLEND_INVSRCCOLOR];
      blend_table[0][D3DBLEND_DESTCOLOR] = caps&D3DPBLENDCAPS_DESTCOLOR ? D3DBLEND_DESTCOLOR :
         caps&D3DPBLENDCAPS_DESTALPHA ? D3DBLEND_DESTALPHA : D3DBLEND_ONE;
      blend_table[0][D3DBLEND_INVDESTCOLOR] = caps&D3DPBLENDCAPS_INVDESTCOLOR ? D3DBLEND_INVDESTCOLOR :
         caps&D3DPBLENDCAPS_INVDESTCOLOR ? D3DBLEND_INVDESTALPHA : D3DBLEND_ONE;
      blend_table[0][D3DBLEND_DESTALPHA] = caps&D3DPBLENDCAPS_DESTALPHA ? D3DBLEND_DESTALPHA :
         blend_table[0][D3DBLEND_DESTCOLOR];
      blend_table[0][D3DBLEND_INVDESTALPHA] = caps&D3DPBLENDCAPS_INVDESTALPHA ? D3DBLEND_INVDESTALPHA :
         blend_table[0][D3DBLEND_INVDESTCOLOR];
      blend_table[0][D3DBLEND_SRCALPHASAT] = D3DBLEND_SRCALPHASAT;
      blend_table[0][D3DBLEND_SRCALPHASAT] = D3DBLEND_BOTHSRCALPHA;
      blend_table[0][D3DBLEND_BOTHINVSRCALPHA] = D3DBLEND_BOTHINVSRCALPHA;
                              //setup dest
      memcpy(&blend_table[1], &blend_table[0], sizeof(blend_table[0]));
      blend_table[1][0] = D3DBLEND_ZERO;
      blend_table[1][D3DBLEND_ZERO] = D3DBLEND_ZERO;
      blend_table[1][D3DBLEND_ONE] = caps&D3DPBLENDCAPS_ONE ? D3DBLEND_ONE : blend_table[1][D3DBLEND_INVSRCALPHA];
   }
                              //get list of supported texture formats
   {
      HRESULT hr;

      texture_formats[0].clear();
      texture_formats[1].clear();
      rtt_zbuf_fmts.clear();
      D3DDISPLAYMODE disp_mode;
      hr = lpDev8->GetDisplayMode(&disp_mode);
      CHECK_D3D_RESULT("GetDisplayMode", hr);

      D3DDEVICE_CREATION_PARAMETERS cp;
      hr = lpDev8->GetCreationParameters(&cp);
      CHECK_D3D_RESULT("GetCreationParameters", hr);

      vector<D3DFORMAT> rtt_formats;
      for(int i = 0 ; i < sizeof(supported_texture_formats)/sizeof(D3DFORMAT); i++){
         for(int j=0; j<2; j++){
            D3DFORMAT fmt = supported_texture_formats[i];
            hr = lpD3D8->CheckDeviceFormat(cp.AdapterOrdinal, cp.DeviceType, disp_mode.Format,
               !j ? 0 : D3DUSAGE_RENDERTARGET,
               D3DRTYPE_TEXTURE, fmt);
            if(SUCCEEDED(hr)){
               texture_formats[j].push_back(S_pixelformat());
               S_pixelformat &pf = texture_formats[j].back();
               ConvertD3DpfToI3Dpf(fmt, pf);
               if(j==1){
                              //store also D3D fmt, for z-buf checking below
                  rtt_formats.push_back(fmt);
                  rtt_zbuf_fmts.push_back(vector<D3DFORMAT>());
               }
            }
         }
      }
      for(i=0; i<sizeof(supported_zsten_formats)/sizeof(D3DFORMAT); i++){
         D3DFORMAT fmt = supported_zsten_formats[i];
         hr = lpD3D8->CheckDeviceFormat(cp.AdapterOrdinal, cp.DeviceType, disp_mode.Format,
            D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, fmt);
         if(SUCCEEDED(hr)){
                              //check if this format matches some render targets
            for(int j=rtt_formats.size(); j--; ){
               hr = lpD3D8->CheckDepthStencilMatch(cp.AdapterOrdinal, cp.DeviceType, disp_mode.Format,
                  rtt_formats[j], fmt);
               if(SUCCEEDED(hr)){
                  rtt_zbuf_fmts[j].push_back(fmt);
               }
            }
         }
      }
                              //now remove all RTT formats which has no compatible depth formats
      for(i=texture_formats[1].size(); i--; ){
         if(!rtt_zbuf_fmts[i].size()){
            texture_formats[1][i] = texture_formats[1].back(); texture_formats[1].pop_back();
            rtt_zbuf_fmts[i] = rtt_zbuf_fmts.back(); rtt_zbuf_fmts.pop_back();
         }
      }
      is_hal = (cp.DeviceType == D3DDEVTYPE_HAL);
   }

   HRESULT hr;
                              //init D3D defaults
   hr = lpDev8->SetTextureStageState(0, D3DTSS_COLOROP, stage_txt_op[0] = D3DTOP_MODULATE);
   CHECK_D3D_RESULT("SetTextureStageState", hr);
   for(int i=1; i<MAX_TEXTURE_STAGES; i++){
      hr = lpDev8->SetTextureStageState(i, D3DTSS_COLOROP, stage_txt_op[i] = D3DTOP_DISABLE);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      hr = lpDev8->SetTextureStageState(i, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      hr = lpDev8->SetTextureStageState(i, D3DTSS_COLORARG2, D3DTA_CURRENT);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAOP, stage_alpha_op[i] = D3DTOP_DISABLE);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
   }

   hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAOP, stage_alpha_op[0] = D3DTOP_SELECTARG1);
   CHECK_D3D_RESULT("SetTextureStageState", hr);
   hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG2, stage0_alpha_arg2 = D3DTA_CURRENT);
   CHECK_D3D_RESULT("SetTextureStageState", hr);

   hr = lpDev8->SetRenderState(D3DRS_FOGCOLOR, d3d_fog_color = 0x808080);
   CHECK_D3D_RESULT("SetRenderState", hr);

   hr = lpDev8->SetRenderState(D3DRS_TEXTUREFACTOR, last_texture_factor = 0);
   CHECK_D3D_RESULT("SetRenderState", hr);
   last_alpha_ref = -1;

                              //re-apply filtering
   bool is_filter = (drv_flags&DRVF_LINFILTER);
   drv_flags ^= DRVF_LINFILTER;
   SetState(RS_LINEARFILTER, is_filter);

   memset(last_bump_scale, 0, sizeof(last_bump_scale));
   for(i=0; i<MAX_TEXTURE_STAGES; i++){
      last_txt_coord_index[i] = i;
      SetTexture1(i, NULL);
      stage_minf[i] = D3DTEXF_POINT;
      EnableAnisotropy(i, true);
   }

   hr = lpDev8->SetRenderState(D3DRS_SPECULARENABLE, specular_enable = false);
   CHECK_D3D_RESULT("SetRenderState", hr);
   hr = lpDev8->SetRenderState(D3DRS_ALPHATESTENABLE, alpharef_enable = false);
   CHECK_D3D_RESULT("SetRenderState", hr);

                              //init fog
   hr = lpDev8->SetRenderState(D3DRS_FOGENABLE, fog_enable = false);
   CHECK_D3D_RESULT("SetRenderState", hr);

   hr = lpDev8->SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL);
   CHECK_D3D_RESULT("SetRenderState", hr);
   hr = lpDev8->SetRenderState(D3DRS_LIGHTING, false);
   CHECK_D3D_RESULT("SetRenderState", hr);
   hr = lpDev8->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, false);
   CHECK_D3D_RESULT("SetRenderState", hr);
   //lpDev8->SetRenderState(D3DRS_EDGEANTIALIAS, false);

   hr = lpDev8->SetRenderState(D3DRS_SRCBLEND, srcblend = D3DBLEND_ONE);
   CHECK_D3D_RESULT("SetRenderState", hr);
   hr = lpDev8->SetRenderState(D3DRS_DESTBLEND, dstblend = D3DBLEND_ZERO);
   CHECK_D3D_RESULT("SetRenderState", hr);
                              //setup clipping
   hr = lpDev8->SetRenderState(D3DRS_CLIPPING, d3d_clipping = true);
   CHECK_D3D_RESULT("SetRenderState", hr);

   alpha_mode = 0;
   last_blend_flags = I3DBLEND_OPAQUE;
   b_cullnone = false;
   zw_enable = true;
   last_vs_handle = 0;

                              //check basic blending caps
   drv_flags2 &= ~(DRVF2_CAN_MODULATE2X | DRVF2_CAN_TXTOP_ADD);
#ifndef DEBUG_NO_MODULATE2X
   if(d3d_caps8.TextureOpCaps&D3DTEXOPCAPS_MODULATE2X){
                              //make sure we really can use this (some drivers report it, but can't use it)

      I3D_RESULT ir;
      S_pixelformat pf;
      ir = FindTextureFormat(pf, 0);
      assert(I3D_SUCCESS(ir));

      I3D_CREATETEXTURE ct;
      memset(&ct, 0, sizeof(ct));
      ct.flags = 0;
      ct.flags |= TEXTMAP_NOMIPMAP;
      ct.size_x = 16;
      ct.size_y = 16;
      ct.flags |= TEXTMAP_USEPIXELFORMAT;
      ct.pf = &pf;

      PI3D_texture tp;
      if(I3D_SUCCESS(CreateTexture(&ct, &tp))){
         SetupTextureStage(0, D3DTOP_MODULATE2X);
         SetTexture1(0, tp);

         dword num_passes = 0;
         HRESULT hr = lpDev8->ValidateDevice(&num_passes);
         if(SUCCEEDED(hr) && num_passes==1){
            drv_flags2 |= DRVF2_CAN_MODULATE2X;
         }
         SetTexture1(0, NULL);
         tp->Release();
      }
   }
#endif
   if(d3d_caps8.TextureOpCaps&D3DTEXOPCAPS_ADD){
      drv_flags2 |= DRVF2_CAN_TXTOP_ADD;
   }

                              //determine vertex blending mode
   vertex_light_blend_op = D3DTOP_MODULATE;
   vertex_light_mult = 1.0f;
   if(drv_flags2&DRVF2_CAN_MODULATE2X){
      vertex_light_blend_op = D3DTOP_MODULATE2X;
      vertex_light_mult = .5f;
   }


   InitLMConvertor();

                              //determine if hardware is capable of
                              // hardware transforming
   drv_flags2 &= ~DRVF2_DIRECT_TRANSFORM;
   if(d3d_caps8.DevCaps&D3DDEVCAPS_HWTRANSFORMANDLIGHT){
      D3DDEVICE_CREATION_PARAMETERS cp;
      hr = lpDev8->GetCreationParameters(&cp);
      CHECK_D3D_RESULT("GetCreationParameters", hr);
      if(cp.BehaviorFlags&(D3DCREATE_HARDWARE_VERTEXPROCESSING|D3DCREATE_MIXED_VERTEXPROCESSING)){
         drv_flags2 |= DRVF2_DIRECT_TRANSFORM;
      }
   }
#ifdef DEBUG_FORCE_DIRECT_TRANSFORM
   drv_flags2 |= DRVF2_DIRECT_TRANSFORM;
#endif
   if(init_data.flags&I3DINIT_DIRECT_TRANS)
      drv_flags2 |= DRVF2_DIRECT_TRANSFORM;

   if(!vb_flare){
                              //init vertex buffer for flares
      HRESULT hr;
      dword d3d_usage = D3DUSAGE_WRITEONLY;
      if(!IsDirectTransform())
         d3d_usage |= D3DUSAGE_SOFTWAREPROCESSING;
      hr = lpDev8->CreateVertexBuffer(sizeof(S_vertex_flare)*NUM_FLARE_VERTS, d3d_usage,
         0, D3DPOOL_DEFAULT, (IDirect3DVertexBuffer8**)&vb_flare);
      CHECK_D3D_RESULT("CreateVertexBuffer", hr);
      if(SUCCEEDED(hr)){
         byte *p_dst;
         hr = vb_flare->Lock(0, sizeof(S_vertex_flare) * NUM_FLARE_VERTS, &p_dst, 0);
         if(SUCCEEDED(hr)){
            memcpy(p_dst, flare_coor, sizeof(S_vertex_flare) * NUM_FLARE_VERTS);
            vb_flare->Unlock();
         }else{
            vb_flare = NULL;
         }
      }
   }

   SetState(RS_USELMAPPING, GetState(RS_USELMAPPING));

                              //detect support for bump-mapping
   drv_flags &= ~DRVF_BUMPMAPPING;

#ifndef DEBUG_DISABLE_BUMPMAPPING
                              //does this device support the two bump mapping blend operations?
                              //does this device support up to 2 blending stages?
   if((d3d_caps8.TextureOpCaps&(D3DTEXOPCAPS_BUMPENVMAP|D3DTEXOPCAPS_BUMPENVMAPLUMINANCE)) &&
      d3d_caps8.MaxTextureBlendStages >= 2){

      drv_flags |= DRVF_BUMPMAPPING;
   }
#endif//!DEBUG_DISABLE_BUMPMAPPING

                              //determine if stencil-buffer is attached
   drv_flags &= ~DRVF_HASSTENCIL;
   {
      IDirect3DSurface8 *lpZS;
      HRESULT hr = lpDev8->GetDepthStencilSurface(&lpZS);
      if(SUCCEEDED(hr)){
         D3DSURFACE_DESC desc;
         hr = lpZS->GetDesc(&desc);
         if(SUCCEEDED(hr)){
            switch(desc.Format){
            case D3DFMT_D15S1:
            case D3DFMT_D24S8:
            case D3DFMT_D24X4S4:
               drv_flags |= DRVF_HASSTENCIL;
               break;
            }
         }
         lpZS->Release();
      }
   }

#ifndef DEBUG_FORCE_TXT_CLIPPING
   if(!d3d_caps8.MaxUserClipPlanes)
#endif
   {
      S_pixelformat pf;
      if(I3D_SUCCESS(FindTextureFormat(pf, FINDTF_ALPHA1))){

         I3D_CREATETEXTURE ct;
         memset(&ct, 0, sizeof(ct));
         ct.flags = TEXTMAP_USEPIXELFORMAT | TEXTMAP_NOMIPMAP;
         ct.size_x = 2;
         ct.size_y = 1;
         ct.pf = &pf;
         if(I3D_SUCCESS(CreateTexture(&ct, (I3D_texture**)&tp_clip))){
            const IImage *img = tp_clip->GetSysmemImage();
            IDirect3DTexture8 *dst = img->GetDirecXTexture();
            D3DLOCKED_RECT lrc;
            HRESULT hr = dst->LockRect(0, &lrc, NULL, 0);
            CHECK_D3D_RESULT("LockRect", hr);
            if(SUCCEEDED(hr)){
               int bpp = img->GetPixelFormat()->bytes_per_pixel;
               memset(lrc.pBits, 0, bpp);
               memset(((byte*)lrc.pBits)+bpp, 0xff, bpp);
               hr = dst->UnlockRect(0);
               CHECK_D3D_RESULT("UnlockRect", hr);
            }else{
               tp_clip = NULL;
            }
         }
      }
   }
   clip_txt_stage = -1;

#ifdef DEBUG_SIM_LOWVIDMEM
   {
      S_pixelformat pf;
      memset(&pf, 0, sizeof(pf));
      pf.bytes_per_pixel = 2;
      pf.r_mask = 0xf800;
      pf.g_mask = 0x07e0;
      pf.b_mask = 0x001f;
      int num_txts = DEBUG_SIM_LOWVIDMEM / (256*256*2);
      for(int i=0; i<num_txts; i++){
         LPIImage img;
         bool b = GetGraphInterface()->CreateImage(&img);
         if(b){
            b = img->Open(NULL,
               IMGOPEN_TEXTURE|IMGOPEN_VIDMEM|IMGOPEN_EMPTY|IMGOPEN_MIPMAP|IMGOPEN_HINTSTATIC,
               256, 256, &pf, 0);
            if(b)
               low_mem_textures.push_back(img);
            img->Release();
         }
      }
   }
#endif


   {
                              //setup shader constants
      hr = lpDev8->SetVertexShaderConstant(VSC_CONSTANTS, &S_vectorw(0.0f, 0.5f, 1.0f, 2.0f), 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
      hr = lpDev8->SetVertexShaderConstant(VSC_CONSTANTS1, &S_vectorw(.5f, -.5f, .0f, .0f), 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   }

   hr = lpDev8->GetRenderTarget((IDirect3DSurface8**)&curr_render_target); curr_render_target->Release();
   CHECK_D3D_RESULT("GetRenderTarget", hr);
   hr = lpDev8->GetDepthStencilSurface((IDirect3DSurface8**)&curr_z_sten); curr_z_sten->Release();
   CHECK_D3D_RESULT("GetDepthStencilSurface", hr);

   InitShadowResources();
}

//----------------------------

void I3D_driver::InitShadowResources(){

   HRESULT hr;
                              //init shadow rendering
   drv_flags &= ~DRVF_CANRENDERSHADOWS;
                              //multitexturing must be supported
   if(NumSimultaneousTextures() >= 2)
   do{
                              //D3DPBLENDCAPS_INVSRCCOLOR must be reported
      if(!(d3d_caps8.DestBlendCaps&D3DPBLENDCAPS_INVSRCCOLOR))
         break;

      const int SHD_RECT_SIZE = 128;

      I3D_CREATETEXTURE ct;
      memset(&ct, 0, sizeof(ct));
      ct.flags = TEXTMAP_NOMIPMAP | TEXTMAP_HINTDYNAMIC | TEXTMAP_RENDERTARGET | TEXTMAP_TRUECOLOR;
      for(int i=2; i--; ){
         ct.size_x = ct.size_y = SHD_RECT_SIZE;
#if 0
         if(!i){
                              //make RTT texture 2x bigger, which will effectively
                              // produce 2x2 antialiasing, if using texture filtering
            ct.size_x = ct.size_y = SHD_RECT_SIZE * 2;
         }
#endif
         I3D_RESULT ir = CreateTexture(&ct, (I3D_texture**)&tp_shadow[i]);
         if(I3D_FAIL(ir))
            break;
      }
      if(i!=-1)
         break;
      {
                              //create depth buffer
         D3DDISPLAYMODE disp_mode;
         hr = lpDev8->GetDisplayMode(&disp_mode);
         CHECK_D3D_RESULT("GetDisplayMode", hr);

         D3DDEVICE_CREATION_PARAMETERS cp;
         hr = lpDev8->GetCreationParameters(&cp);
         CHECK_D3D_RESULT("GetCreationParameters", hr);

         D3DSURFACE_DESC rt_desc;
         hr = tp_shadow[0]->GetSysmemImage()->GetDirecXTexture()->GetLevelDesc(0, &rt_desc);
         if(SUCCEEDED(hr)){
            static const D3DFORMAT zb_fmts[] = {
               D3DFMT_D16,
               D3DFMT_D15S1,
               D3DFMT_D32,
               D3DFMT_D24S8,
               D3DFMT_D24X4S4,
               D3DFMT_D24X8,
               D3DFMT_UNKNOWN
            };
            IDirect3DSurface8 *zb;
            for(i=0; zb_fmts[i]!=D3DFMT_UNKNOWN; i++){
               hr = lpD3D8->CheckDepthStencilMatch(cp.AdapterOrdinal, cp.DeviceType, disp_mode.Format,
                  rt_desc.Format, zb_fmts[i]);
               if(SUCCEEDED(hr)){
                  hr = lpDev8->CreateDepthStencilSurface(SHD_RECT_SIZE, SHD_RECT_SIZE,
                     zb_fmts[i], D3DMULTISAMPLE_NONE, &zb);
                  if(SUCCEEDED(hr)){
                     zb_shadow = zb;
                     zb->Release();
                     break;
                  }
               }
            }
         }
         if(FAILED(hr))
            break;
      }
      {
                              //create clamp texture
         ct.flags = TEXTMAP_TRUECOLOR | TEXTMAP_NOMIPMAP;
         ct.size_x = 256;
         ct.size_y = 1;
         I3D_RESULT ir = CreateTexture(&ct, (I3D_texture**)&tp_shadow[2]);
         if(I3D_FAIL(ir))
            break;
         byte range[768];
         for(int i=0; i<256; i++){
            //byte b = 256 - i;
            byte b = i;
            range[i*3+0] = range[i*3+1] = range[i*3+2] = b;
         }
         range[765] = range[766] = range[767] = 0;

         IDirect3DTexture8 *dst = tp_shadow[2]->GetSysmemImage()->GetDirecXTexture();
         D3DLOCKED_RECT lrc;
         HRESULT hr = dst->LockRect(0, &lrc, NULL, 0);
         CHECK_D3D_RESULT("LockRect", hr);
         if(FAILED(hr))
            break;
         C_rgb_conversion *rgbconv;
         GetGraphInterface()->CreateRGBConv(&rgbconv);
         rgbconv->Init(*tp_shadow[2]->GetSysmemImage()->GetPixelFormat());

         void *dstp = lrc.pBits;

         rgbconv->Convert(range, dstp,
            tp_shadow[2]->SizeX1(), tp_shadow[2]->SizeY1(),
            0,                //srcpitch
            lrc.Pitch,        //dstpitch
            3,                //srcbpp
            NULL,             //pal
            0,                //flags
            0);               //colorkey

         rgbconv->Release();

         hr = dst->UnlockRect(0);
         CHECK_D3D_RESULT("UnlockRect", hr);
         if(FAILED(hr))
            break;
      }

      HRESULT hr;
      hr = lpDev8->CreateVertexBuffer(sizeof(S_blur_vertex) * 4 * NUM_SHADOW_BLUR_SAMPLES,
         //D3DUSAGE_DONOTCLIP |
         D3DUSAGE_WRITEONLY,
         D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1, D3DPOOL_DEFAULT, (IDirect3DVertexBuffer8**)&vb_blur);
      CHECK_D3D_RESULT("CreateVertexBuffer", hr);
      if(FAILED(hr))
         break;

      hr = lpDev8->CreateIndexBuffer(sizeof(word) * 6 * NUM_SHADOW_BLUR_SAMPLES,
         //D3DUSAGE_DONOTCLIP |
         D3DUSAGE_WRITEONLY,
         D3DFMT_INDEX16, D3DPOOL_DEFAULT, (IDirect3DIndexBuffer8**)&ib_blur);
      CHECK_D3D_RESULT("CreateIndexBuffer", hr);
      if(FAILED(hr))
         break;
                              //fill-in blur buffers
      {
         S_blur_vertex *bv;
         word *indx;
         hr = vb_blur->Lock(0, 0, (byte**)&bv, 0);
         CHECK_D3D_RESULT("Lock", hr);
         if(FAILED(hr))
            break;

         hr = ib_blur->Lock(0, 0, (byte**)&indx, 0);
         CHECK_D3D_RESULT("Lock", hr);
         if(FAILED(hr)){
            vb_blur->Unlock();
            break;
         }

         struct S_blur_matrix{
            float du, dv;
            float amount;
         };
         static const S_blur_matrix bm_4[] = {
            {-.5f, -.5f, 4.0f / 16.0f},
            { .5f, -.5f, 4.0f / 16.0f},
            {-.5f,  .5f, 4.0f / 16.0f},
            { .5f,  .5f, 4.0f / 16.0f},
         };

         const float F5 = .7f;
         static const S_blur_matrix bm_5[] = {
            {-F5, -F5, .2f},
            { F5, -F5, .2f},
            {-F5,  F5, .2f},
            { F5,  F5, .2f},
            {  0,   0, .2f},
         };

         //const float F9a = 1.0f, F9b = .707f;
         const float F9a = 1.5f, F9b = 1.0f;
         //const float F9a = 2.0f, F9b = 1.41f;
         /*
         static const S_blur_matrix bm_9[] = {
            {-F9b, -F9b, 1.0f / 16.0f},
            {   0, -F9a, 2.0f / 16.0f},
            { F9b, -F9b, 1.0f / 16.0f},
            {-F9a,    0, 2.0f / 16.0f},
            {   0,    0, 4.0f / 16.0f},
            { F9a,    0, 2.0f / 16.0f},
            {-F9b,  F9b, 1.0f / 16.0f},
            {   0,  F9a, 2.0f / 16.0f},
            { F9b,  F9b, 1.0f / 16.0f},
         };
         */
         static const S_blur_matrix bm_9[] = {
            {-F9b, -F9b, 1.5f / 16.0f},
            {   0, -F9a, 2.0f / 16.0f},
            { F9b, -F9b, 1.5f / 16.0f},
            {-F9a,    0, 2.0f / 16.0f},
            {   0,    0, 2.0f / 16.0f},
            { F9a,    0, 2.0f / 16.0f},
            {-F9b,  F9b, 1.5f / 16.0f},
            {   0,  F9a, 2.0f / 16.0f},
            { F9b,  F9b, 1.5f / 16.0f},
         };

         float textel_size = 1.0f / (float)tp_shadow[0]->SizeX1();

         const S_blur_matrix *bm = (NUM_SHADOW_BLUR_SAMPLES==9) ? bm_9 : bm_5;

                                       //add half-texel to correct rounding errors
         float half_texel = .5f / (float)tp_shadow[0]->SizeX1();

         for(int i=NUM_SHADOW_BLUR_SAMPLES; i--; bv += 4, indx += 6){
                              //make indicies
            indx[0] = i*4;
            indx[1] = indx[0] + 1;
            indx[2] = indx[0] + 2;
            indx[3] = indx[2];
            indx[4] = indx[1];
            indx[5] = indx[0] + 3;
                              //make vertices
            for(int j=4; j--; ){
               S_blur_vertex &v = bv[j];
               v.v.Zero();
               v.v.w = 1.0f;
               v.uv.u = half_texel;
               v.uv.v = half_texel;
               if(j&1){
                  v.v.x += tp_shadow[1]->SizeX1();
                  v.uv.u += 1.0f;
               }
               if(j&2){
                  v.v.y += tp_shadow[1]->SizeY1();
                  v.uv.v += 1.0f;
               }
               v.uv.u += textel_size * bm[i].du;
               v.uv.v += textel_size * bm[i].dv;
               v.diffuse = (FloatToInt(255.0f * bm[i].amount)) << 24;
            }
         }
         hr = vb_blur->Unlock();
         CHECK_D3D_RESULT("Unlock", hr);
         hr = ib_blur->Unlock();
         CHECK_D3D_RESULT("Unlock", hr);
      }
                              //create dynamic index buffer
      IDirect3DIndexBuffer8 *ib;
      dword d3d_usage = D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY;
      if(!IsDirectTransform())
         d3d_usage |= D3DUSAGE_SOFTWAREPROCESSING;
      hr = lpDev8->CreateIndexBuffer(8192*sizeof(I3D_triface), d3d_usage,
         D3DFMT_INDEX16, D3DPOOL_DEFAULT, &ib);
      CHECK_D3D_RESULT("CreateIndexBuffer", hr);
      ib_shadow = ib;
      ib->Release();
      ib_shadow_base_index = 0;

      drv_flags |= DRVF_CANRENDERSHADOWS;
   }while(false);

   if(!(drv_flags&DRVF_CANRENDERSHADOWS)){
      for(int i=0; i<3; i++){
         tp_shadow[i] = NULL;
      }
      zb_shadow = NULL;
      vb_blur = NULL;
      ib_blur = NULL;
      ib_shadow = NULL;
   }
}

//----------------------------

void I3D_driver::InitLMConvertor(){

                              //determine how much memory we may spend on light-maps
   {
      I3D_stats_video sv;
      GetStats(I3DSTATS_VIDEO, &sv);
                              //
      max_lm_memory = sv.txtmem_total/6;
      //max_lm_memory = sv.txtmem_free / 6;
   }

                              //find texture format for light-mapping (standard format)
   dword findtf_flags = 0;

   if(drv_flags2&DRVF2_LMTRUECOLOR){
                              //allow only if we have enough LM memory, use default depth otherwise
      if(max_lm_memory > 3*1024*1024)
         findtf_flags |= FINDTF_TRUECOLOR;
   }

   I3D_RESULT ir;
   ir = FindTextureFormat(pf_light_maps, findtf_flags);
   assert(I3D_SUCCESS(ir));

   {
                              //check if single-pass multitexturing possible
      drv_flags &= ~DRVF_SINGLE_PASS_MODULATE | DRVF_SINGLE_PASS_MODULATE2X;
#ifndef DEBUG_NO_MULTI_TEXTURING
      if(NumSimultaneousTextures() >= 2){

         PI3D_texture tp[2] = {NULL, NULL};
         I3D_CREATETEXTURE ct;
         ct.flags = 0;
         ct.flags |= TEXTMAP_NOMIPMAP;
         ct.size_x = 16;
         ct.size_y = 16;
         ct.flags |= TEXTMAP_USEPIXELFORMAT | TEXTMAP_HINTDYNAMIC;

                              //base texture
         {
            S_pixelformat pf;
            FindTextureFormat(pf, 0);
            ct.pf = &pf;
            CreateTexture(&ct, &tp[0]);
         }
                              //light-map texture
         ct.pf = &pf_light_maps;
         CreateTexture(&ct, &tp[1]);
         if(tp[0] && tp[1]){
            //SetState(RS_MIPMAP, false);
            if(stage_txt_op[0]!=D3DTOP_SELECTARG1)
               lpDev8->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);

            SetTexture1(0, tp[0]);
            SetTexture1(1, tp[1]);

            dword num_passes;
                              //check MODULATE
            SetupTextureStage(1, D3DTOP_MODULATE);
            drv_flags &= ~DRVF_SINGLE_PASS_MODULATE;
            HRESULT hr = lpDev8->ValidateDevice(&num_passes);
            if(SUCCEEDED(hr) && num_passes==1)
               drv_flags |= DRVF_SINGLE_PASS_MODULATE;

                              //check MODULATE2X
            SetupTextureStage(1, D3DTOP_MODULATE2X);
            drv_flags &= ~DRVF_SINGLE_PASS_MODULATE2X;
            hr = lpDev8->ValidateDevice(&num_passes);
            if(SUCCEEDED(hr) && num_passes==1)
               drv_flags |= DRVF_SINGLE_PASS_MODULATE2X;

            SetTexture1(0, NULL);
            SetTexture1(1, NULL);
            DisableTextureStage(1);
            if(stage_txt_op[0]!=D3DTOP_SELECTARG1)
               lpDev8->SetTextureStageState(0, D3DTSS_COLOROP, stage_txt_op[0]);
         }
         if(tp[0]) tp[0]->Release();
         if(tp[1]) tp[1]->Release();
                              //todo: if this is really possible (try also with 3-linear filtering)
      }
#endif   //!DEBUG_NO_MULTI_TEXTURING
   }

                              //init rgb convertor
   if(!rgbconv_lm)
      GetGraphInterface()->CreateRGBConv(&rgbconv_lm);
   S_pixelformat pf = pf_light_maps;
   if(pf.flags&PIXELFORMAT_PALETTE){
      pf.r_mask = 0xe0;
      pf.g_mask = 0x1c;
      pf.b_mask = 0x03;
      pf.a_mask = 0x00;
      pf.flags &= ~PIXELFORMAT_PALETTE;

      pf_light_maps = pf;
   }else{
      static const dword bit_mask[] = {0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};
                              //make all components same width, so that we don't have
                              // colorized gray scale
      int num = Min(CountBits(pf.r_mask), Min(CountBits(pf.g_mask), CountBits(pf.b_mask)));
      if(num<=8){
         dword mask = 0xff&bit_mask[num];
         pf.r_mask = Rotate(mask, FindLastBit(pf.r_mask)-7);
         pf.g_mask = Rotate(mask, FindLastBit(pf.g_mask)-7);
         pf.b_mask = Rotate(mask, FindLastBit(pf.b_mask)-7);
      }
   }
   rgbconv_lm->Init(pf);
}

//----------------------------

void I3D_driver::CloseD3DResources(){


#ifdef DEBUG_SIM_LOWVIDMEM
   low_mem_textures.clear();
#endif
   //gone_textures.clear();
   int i;

                              //call all registered callbacks
   for(i=reset_callbacks.size(); i--; ){
      const pair<t_ResetCallback*, void*> &p = reset_callbacks[i];
      (*p.first)(RESET_RELEASE, p.second);
   }

   for(i=managed_textures.size(); i--; ){
      PI3D_texture tp = managed_textures[i];
      tp->Manage(I3D_texture::MANAGE_FREE);
   }
   texture_formats[0].clear();
   texture_formats[1].clear();

   for(i=MAX_TEXTURE_STAGES; i--; )
      SetTexture1(i, NULL);

   for(vs_set::const_iterator it = vs_cache.begin(); it!=vs_cache.end(); it++){
      HRESULT hr;
      hr = lpDev8->DeleteVertexShader((*it).vs_handle);
      CHECK_D3D_RESULT("DeleteVertexShader", hr);
   }
   vs_cache.clear();

   vb_blur = NULL;
   ib_blur = NULL;
   ib_shadow = NULL;
   for(i=0; i<3; i++)
      tp_shadow[i] = NULL;
   zb_shadow = NULL;

   tp_clip = NULL;

   curr_render_target = NULL;
   curr_z_sten = NULL;

   tp_dyn_light_circle = NULL;

   vb_flare = NULL;
                     //modify all references to HW vertex buffers
   for(i = hw_vb_list.size(); i--; ){
      FreeVertexBuffer(hw_vb_list[i].vb, hw_vb_list[i].vertex_index);
      S_VB_use_ref &ur = hw_vb_list[i];
      *ur.ref = NULL;
   }
                     //now no allocated VBs and IBs may be in hw
#ifdef _DEBUG
   for(i=vb_manager_list.size(); i--; )
      assert(vb_manager_list[i]->GetPool() != D3DPOOL_DEFAULT);
   for(i=ib_manager_list.size(); i--; )
      assert(ib_manager_list[i]->GetPool() != D3DPOOL_DEFAULT);
#endif
                     //forget all references, they'll be recreated
   hw_vb_list.clear();
   SetStreamSource(NULL, 0);

   if(rgbconv_lm){
      rgbconv_lm->Release();
      rgbconv_lm = NULL;
   }

                              //release editing icons
   for(i=icon_materials.size(); i--; ){
      if(icon_materials[i])
         icon_materials[i]->Release();
   }
   icon_materials.clear();
}

//----------------------------

dword I3D_driver::GraphCallback(dword msg, dword par1, dword par2){

   switch(msg){
   case CM_RECREATE:
      switch(par2){
      case 0:
         {
                              //release all hardware resources
            if(!IsInit())
               return 0;
            CloseD3DResources();
         }
         break;
      case 1:                 //construct
         {
            InitD3DResources();
            RefreshAfterReset();
         }
         break;
      }
      break;
   }
   return 0;
}

//----------------------------

dword I3D_driver::GraphCallback_thunk(dword msg, dword par1, dword par2, void *context){

   PI3D_driver drv = (PI3D_driver)context;
   assert(drv);
   return drv->GraphCallback(msg, par1, par2);
}

//----------------------------

void I3D_driver::RefreshAfterReset(){


   SetViewport(0, 0, init_data.lp_graph->Scrn_sx(), init_data.lp_graph->Scrn_sy());

   int i;
                              //textures
   
   for(i=managed_textures.size(); i--; ){
      PI3D_texture tp = managed_textures[i];
                           //don't refresh LMap textures
      if(UnregLMTexture(tp)){
         tp->SetDirty();
         managed_textures[i] = managed_textures.back(); managed_textures.pop_back();
         continue;
      }
      tp->Reload();
   }
   lm_textures.clear();    //don't keep dirty LM textures
   map_channel_2_img = NULL;

                              //call all registered callbacks
   for(i=reset_callbacks.size(); i--; ){
      const pair<t_ResetCallback*, void*> &p = reset_callbacks[i];
      (*p.first)(RESET_RECREATE, p.second);
   }

                              //refresh lost state
   static const I3D_RENDERSTATE states[] = {
      RS_LINEARFILTER, RS_MIPMAP, RS_USEZB, RS_DITHER,
      RS_FOG, RS_WIREFRAME, 
   };
   for(i=0; i<sizeof(states)/sizeof(I3D_RENDERSTATE); i++){
      bool b = GetState(states[i]);
      SetState(states[i], !b);
      SetState(states[i], b);
   }
}

//----------------------------

PI3D_material I3D_driver::CreateMaterial(){

   PI3D_material mp = new I3D_material(this);
   return mp;
}

//----------------------------

PI3D_scene I3D_driver::CreateScene(){

   PI3D_scene sp = new I3D_scene(this);
   if(sp) sp->Init(this);

   return sp;
}

//----------------------------

PI3D_user I3D_driver::CreateUserFrame(){

   return new I3D_user(this);
}

//----------------------------

PI3D_dummy I3D_driver::CreateDummy(){

   return new I3D_dummy(this);
}

//----------------------------

PI3D_target I3D_driver::CreateTarget(){

   return new I3D_target(this);
}

//----------------------------

PI3D_visual I3D_driver::CreateVisual(dword visual_type){

                              //scan through registered plugins
   for(int i=0; i<num_visual_plugins; i++)
   if(visual_plugins[i].visual_type == visual_type)
      return (*visual_plugins[i].create_proc)(this);
   return NULL;
}

//----------------------------

I3D_RESULT I3D_driver::EnumVisualTypes(I3D_ENUMVISUALPROC *ep, dword user) const{

   for(int i=0; i<num_visual_plugins; i++){
      I3DENUMRET er = (*ep)(visual_plugins[i].visual_type, visual_plugins[i].friendly_name, user);
      if(er==I3DENUMRET_CANCEL) return I3DERR_CANCELED;
   }
   return I3D_OK;
}

//----------------------------

PI3D_model I3D_driver::CreateModel(){

   return new I3D_model(this);
}

//----------------------------

PI3D_light I3D_driver::CreateLight(){

   PI3D_light lp = new I3D_light(this);
   return lp;
}

//----------------------------

PI3D_camera I3D_driver::CreateCamera(){

   return new I3D_camera(this);
}

//----------------------------

PI3D_volume I3D_driver::CreateVolume(){

   return new I3D_volume(this);
}

//----------------------------

PI3D_joint I3D_driver::CreateJoint(){

   return new I3D_joint(this);
}

//----------------------------

PI3D_frame I3D_driver::CreateFrame(dword t, dword sub_type){

   switch(t){
   case FRAME_VISUAL: 
      {
         int i;
         dword num_plg = num_visual_plugins;
                                    //scan through registered plugins
         for(i=0; i<num_plg; i++)
         if(visual_plugins[i].visual_type == sub_type)
            return (*visual_plugins[i].create_proc)(this);
      }
      return NULL;
   case FRAME_LIGHT:
      {
         PI3D_light lp = new I3D_light(this);
         //if(lp) lights.push_back(lp);
         return lp;
      }
      break;
   case FRAME_CAMERA: return new I3D_camera(this);
      /*
   case FRAME_SOUND:
      {
         PI3D_sound sp = new I3D_sound(this);
         if(sp) sounds.push_back(sp);
         return sp;
      }
      break;
      */
   case FRAME_DUMMY: return new I3D_dummy(this);
   case FRAME_TARGET: return new I3D_target(this);
   case FRAME_USER: return new I3D_user(this);
   case FRAME_MODEL: return new I3D_model(this);
   case FRAME_JOINT: return new I3D_joint(this);
   case FRAME_VOLUME: return new I3D_volume(this);
   case FRAME_OCCLUDER: return new I3D_occluder(this);
   }
   return NULL;
}

//----------------------------

PI3D_mesh_level I3D_driver::CreateMeshLevel(){

   return new I3D_mesh_level(this);
}

//----------------------------

PI3D_mesh_object I3D_driver::CreateMesh(){

   return new I3D_mesh_object(this);
}

//----------------------------

PI3D_animation_base I3D_driver::CreateAnimation(I3D_ANIMATION_TYPE type){

   switch(type){
   case I3DANIM_KEYFRAME:
      return new I3D_keyframe_anim(this);
   case I3DANIM_POSE:
      return new I3D_anim_pose(this);
   }
   return NULL;
}

//----------------------------

PI3D_animation_set I3D_driver::CreateAnimationSet(){
   
   return new I3D_animation_set(this);
}

//----------------------------
                              //internal

I3D_RESULT I3D_driver::CreateTexture1(const I3D_CREATETEXTURE *ct, PI3D_texture *tp1, PI3D_LOAD_CB_PROC cb_load, void *cb_context){

   if(ct->flags&(TEXTMAP_DIFFUSE|TEXTMAP_OPACITY|TEXTMAP_BUMPMAP)){
                              //check if such texture already exists, re-use
      for(int i=managed_textures.size(); i--; ){
         if(managed_textures[i]->Match(ct)){
            *tp1 = managed_textures[i];
            (*tp1)->AddRef();
            return I3D_OK;
         }
      }
   }

   PI3D_texture tp = new I3D_texture(this);
   if(!tp)
      return I3DERR_OUTOFMEM;

   I3D_RESULT ir;
   ir = tp->Open(ct, cb_load, cb_context);
   if(I3D_FAIL(ir)){
      tp->Release();
      *tp1 = NULL;
      return ir;
   }
   *tp1 = tp;
   //managed_textures.push_back(tp);
   return ir;
}

//----------------------------

void I3D_driver::RegisterManagedTexture(PI3D_texture tp){

   int i;
   i = VectorIndex(managed_textures, tp);
   assert(i==-1);
   managed_textures.push_back(tp);
}

//----------------------------

void I3D_driver::UnregisterManagedTexture(PI3D_texture tp){

   int i = VectorIndex(managed_textures, tp);
   //assert(i!=-1);           //lm textures hit this... why?
   if(i!=-1){
      managed_textures[i] = managed_textures.back(); managed_textures.pop_back();
   }
   for(i=MAX_TEXTURE_STAGES; i--; ){
      if(last_texture[i]==tp)
         SetTexture1(i, NULL);
   }
}

//----------------------------

bool I3D_driver::AllocVertexBuffer(dword fvf_flags, D3DPOOL mem_pool, dword d3d_usage, int num_verts,
   LPDIRECT3DVERTEXBUFFER8 *vb_ret, int *beg_indx_ret, bool allow_cache){

                              //if, for some reason, FVF is XYZRHW (transformed),
                              // force include clipping info
   if((fvf_flags&D3DFVF_POSITION_MASK)==D3DFVF_XYZRHW)
      d3d_usage &= ~D3DUSAGE_DONOTCLIP;

                              //find VB manager which would be suitable
   for(int i=vb_manager_list.size(); i--; )
   if(vb_manager_list[i]->GetFVFlags()==fvf_flags && vb_manager_list[i]->GetUsage()==d3d_usage &&
      vb_manager_list[i]->GetPool()==mem_pool)
      break;

   if(i==-1){
                              //alloc new
      dword size_of_element = GetSizeOfVertex(fvf_flags);
      C_RB_manager<IDirect3DVertexBuffer8> *mgr = new C_RB_manager<IDirect3DVertexBuffer8>(mem_pool, d3d_usage, size_of_element, fvf_flags);
      vb_manager_list.push_back(mgr);
      mgr->Release();         //smart pointer's keeping ref
      i = vb_manager_list.size() - 1;
   }
   bool b = vb_manager_list[i]->Allocate(num_verts, this, vb_ret, beg_indx_ret, allow_cache);
   assert(b);

                              //keep reference of pointer
   if(mem_pool == D3DPOOL_DEFAULT){
      hw_vb_list.push_back(S_VB_use_ref());
      S_VB_use_ref &vbr = hw_vb_list.back();
      vbr.vb = *vb_ret;
      vbr.vertex_index = *beg_indx_ret;
      vbr.ref = vb_ret;
   }
   return b;
}

//----------------------------

bool I3D_driver::FreeVertexBuffer(LPDIRECT3DVERTEXBUFFER8 vb, int vertex_indx){

                              
   for(int i=vb_manager_list.size(); i--; ){
      C_RB_manager<IDirect3DVertexBuffer8> *vb_mgr = vb_manager_list[i];
      if(vb_mgr->Free(vb, vertex_indx)){
                              //check pool, remove hw vb from internal list
         if(vb_mgr->GetPool() == D3DPOOL_DEFAULT){
            S_VB_use_ref *vb_ref = hw_vb_list.end();
            for(int i=hw_vb_list.size(); --vb_ref, i--; ){
               //if(hw_vb_list[i].vb == vb && hw_vb_list[i].vertex_index == vertex_indx)
               if(vb_ref->vb == vb && vb_ref->vertex_index == vertex_indx)
                  break;
            }
            assert(i!=-1);
            if(i!=-1){
               hw_vb_list[i] = hw_vb_list.back(); hw_vb_list.pop_back();
            }
         }

                              //destroy empty managers
         if(vb_mgr->IsEmpty()){
            vb_manager_list[i] = vb_manager_list.back(); vb_manager_list.pop_back();
                              //forget cache
            if(last_stream0_vb==vb)
               last_stream0_vb = NULL;
         }
         return true;
      }
   }
   assert(0);
   return false;
}

//----------------------------

bool I3D_driver::AllocIndexBuffer(D3DPOOL mem_pool, dword d3d_usage, int num_faces,
   IDirect3DIndexBuffer8 **ib_ret, int *beg_indx_ret){

   //return ib_manager.Allocate(num_faces, this, ib_ret, beg_indx_ret);

                              //find VB manager which would be suitable
   for(int i=ib_manager_list.size(); i--; ){
      C_RB_manager<IDirect3DIndexBuffer8> *ib = ib_manager_list[i];
      if(ib->GetUsage()==d3d_usage && ib->GetPool()==mem_pool)
         break;
   }

   if(i==-1){
                              //alloc new
      C_RB_manager<IDirect3DIndexBuffer8> *mgr = new C_RB_manager<IDirect3DIndexBuffer8>(mem_pool, d3d_usage, sizeof(I3D_triface));
      ib_manager_list.push_back(mgr);
      mgr->Release();         //smart pointer's keeping ref
      i = ib_manager_list.size() - 1;
   }
   bool b = ib_manager_list[i]->Allocate(num_faces, this, ib_ret, beg_indx_ret);
   assert(b);

                              //keep reference of pointer
   if(mem_pool == D3DPOOL_DEFAULT){
      assert(0);
      /*
      hw_vb_list.push_back(S_VB_use_ref());
      S_VB_use_ref &vbr = hw_vb_list.back();
      vbr.vb = *vb_ret;
      vbr.vertex_index = *beg_indx_ret;
      vbr.ref = vb_ret;
      */
   }
   return b;
}

//----------------------------

bool I3D_driver::FreeIndexBuffer(IDirect3DIndexBuffer8 *ib, int base_index){

   //return ib_manager.Free(ib, base_index);
   for(int i=ib_manager_list.size(); i--; ){
      C_RB_manager<IDirect3DIndexBuffer8> *ib_mgr = ib_manager_list[i];
      if(ib_mgr->Free(ib, base_index)){
                              //check pool, remove hw vb from internal list
         if(ib_mgr->GetPool() == D3DPOOL_DEFAULT){
            assert(0);
            /*
            S_VB_use_ref *vb_ref = hw_vb_list.end();
            for(int i=hw_vb_list.size(); --vb_ref, i--; ){
               //if(hw_vb_list[i].vb == vb && hw_vb_list[i].vertex_index == vertex_indx)
               if(vb_ref->vb == vb && vb_ref->vertex_index == vertex_indx)
                  break;
            }
            assert(i!=-1);
            if(i!=-1){
               hw_vb_list[i] = hw_vb_list.back(); hw_vb_list.pop_back();
            }
            */
         }
                              //destroy empty managers
         if(ib_mgr->IsEmpty()){
            ib_manager_list[i] = ib_manager_list.back(); ib_manager_list.pop_back();
            /*
                              //forget cache
            if(last_stream0_ib==ib)
               last_stream0_ib = NULL;
               */
         }
         return true;
      }
   }
   assert(0);
   return false;
}

//----------------------------

void I3D_driver::UpdateViewport(int l, int t, int r, int b){

                              //limit to viewport size
   int r1 = Min(r, (int)GetGraphInterface()->Scrn_sx());
   int b1 = Min(b, (int)GetGraphInterface()->Scrn_sy());

   D3DVIEWPORT8 vport;
   vport.X = l;
   vport.Y = t;
   vport.Width = r1 - l;
   vport.Height = b1 - t;
   vport.MinZ = 0.0f;
   vport.MaxZ = 1.0f;

   HRESULT hr;
   hr = lpDev8->SetViewport(&vport);
   CHECK_D3D_RESULT("SetViewport", hr);
}

//----------------------------

I3D_RESULT I3D_driver::SetViewport(dword l, dword t, dword r, dword b){

   UpdateViewport(l, t, r, b);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_driver::BeginScene(){

   if(!in_scene_count){
      HRESULT hr = lpDev8->BeginScene();
      if(FAILED(hr)){
         CHECK_D3D_RESULT("BeginScene", hr);
         return I3DERR_GENERIC;
      }
#if defined _DEBUG
      //lpDev8->SetVertexShaderConstant(VSC_DEBUG, &S_vectorw(S_float_random(), S_float_random(), S_float_random(), S_float_random()), 1);
      lpDev8->SetVertexShaderConstant(VSC_DEBUG, debug_float, 1);

      lpDev8->SetPixelShaderConstant(PSC_DEBUG, &S_vectorw(S_float_random(), S_float_random(), S_float_random(), S_float_random()), 1);
#endif
   }
   ++in_scene_count;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_driver::EndScene(){

   if(!in_scene_count)
      return I3DERR_INVALIDPARAMS;
   if(!--in_scene_count){
      SetStreamSource(NULL, 0);
      HRESULT hr;
      hr = lpDev8->EndScene();
      CHECK_D3D_RESULT("EndScene", hr);
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_driver::Clear(dword color){

   dword clear_flags = D3DCLEAR_TARGET;
   if((drv_flags&DRVF_USEZB))
      clear_flags |= D3DCLEAR_ZBUFFER;
   if(GetFlags()&DRVF_HASSTENCIL)
      clear_flags |= D3DCLEAR_STENCIL;

   HRESULT hr;
   hr = lpDev8->Clear(0, NULL, clear_flags, color, 1.0f, 0);
   CHECK_D3D_RESULT("Clear", hr);

   return SUCCEEDED(hr) ? I3D_OK : I3DERR_GENERIC;
}

//----------------------------

void I3D_driver::SetupBlend(dword flags){

   if(last_blend_flags==flags)
      return;

   last_blend_flags = flags;

   dword base_flags = flags&0xffff;
   bool alpha = (base_flags!=I3DBLEND_OPAQUE) ||
      (flags&(I3DBLEND_ALPHAFACTOR|I3DBLEND_VERTEXALPHA));

   HRESULT hr;
   byte src, dst;

   switch(base_flags){
   case I3DBLEND_OPAQUE:
      src = blend_table[0][D3DBLEND_ONE];
      dst = blend_table[1][D3DBLEND_ZERO];
      break;
   case I3DBLEND_ALPHABLEND:
      src = blend_table[0][D3DBLEND_SRCALPHA];
      dst = blend_table[1][D3DBLEND_INVSRCALPHA];
      break;
   case I3DBLEND_ADD:
      src = blend_table[0][D3DBLEND_ONE];
      dst = blend_table[1][D3DBLEND_ONE];
      break;
   case I3DBLEND_ADDALPHA:
      src = blend_table[0][D3DBLEND_SRCALPHA];
      dst = blend_table[1][D3DBLEND_ONE];
      break;
   case I3DBLEND_BLEND:
      src = D3DBLEND_ZERO;
      dst = D3DBLEND_SRCCOLOR;
      break;
   case I3DBLEND_INVBLEND:
      src = D3DBLEND_ZERO;
      dst = D3DBLEND_INVSRCCOLOR;
      break;
   }

   if(!alpha){
      if(alpha_mode&ALPHAMODE_ENABLED){
         alpha_mode &= ~ALPHAMODE_ENABLED;
         hr = lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, false);
         CHECK_D3D_RESULT("SetRenderState", hr);
         if(stage_alpha_op[0] != D3DTOP_SELECTARG1){
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAOP, stage_alpha_op[0] = D3DTOP_SELECTARG1);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
   }else{
      dword new_mode = 0;
      if(!(alpha_mode&ALPHAMODE_ENABLED)){
         alpha_mode |= ALPHAMODE_ENABLED;
         hr = lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }else
         new_mode |= ALPHAMODE_ENABLED;

      if(base_flags==I3DBLEND_ALPHABLEND) 
         new_mode |= ALPHAMODE_TEXTURE;

      if(flags&I3DBLEND_ALPHAFACTOR) 
         new_mode |= ALPHAMODE_FACTOR;
      if(flags&I3DBLEND_VERTEXALPHA) 
         new_mode |= ALPHAMODE_VERTALPHA;

      if(alpha_mode != new_mode)
      {
         alpha_mode = new_mode;

         static const D3DTEXTUREOP alpha_ops[] = {
            D3DTOP_SELECTARG1,
            D3DTOP_SELECTARG1,//pure texture alpha
            D3DTOP_SELECTARG2,//pure factor
            D3DTOP_MODULATE,  //blend of texture alpha and factor
            D3DTOP_MODULATE,
            D3DTOP_MODULATE,
            D3DTOP_MODULATE,
            D3DTOP_MODULATE,
         };
         D3DTEXTUREOP alpha_op = alpha_ops[alpha_mode&(ALPHAMODE_TEXTURE|ALPHAMODE_FACTOR|ALPHAMODE_VERTALPHA)];
         if(stage_alpha_op[0] != alpha_op){
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAOP, stage_alpha_op[0] = alpha_op);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
         dword alpha_arg2 = ((flags&I3DBLEND_ALPHAFACTOR) ? D3DTA_TFACTOR : D3DTA_DIFFUSE);
         if(stage0_alpha_arg2 != alpha_arg2){
            hr = lpDev8->SetTextureStageState(0, D3DTSS_ALPHAARG2, stage0_alpha_arg2 = alpha_arg2);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
      alpha_mode |= ALPHAMODE_ENABLED;
   }

   if(srcblend != src){
      hr = lpDev8->SetRenderState(D3DRS_SRCBLEND,  srcblend=(D3DBLEND)src);
      CHECK_D3D_RESULT("SetRenderState", hr);
   }
   if(dstblend != dst){
      hr = lpDev8->SetRenderState(D3DRS_DESTBLEND, dstblend=(D3DBLEND)dst);
      CHECK_D3D_RESULT("SetRenderState", hr);
   }
}

//----------------------------

void I3D_driver::SetClippingPlane(const S_plane &pl, const S_matrix &m_view_proj_hom){

   if(d3d_caps8.MaxUserClipPlanes){
      S_matrix m = m_view_proj_hom;
      m.Invert();
      m.Transpose();

      S_plane c_clip_plane;
      *(S_vectorw*)&c_clip_plane = (*(S_vectorw*)&pl) * m;
      c_clip_plane.Invert();

      HRESULT hr;
      hr = lpDev8->SetClipPlane(0, (const float*)&c_clip_plane);
      CHECK_D3D_RESULT("SetClipPlane", hr);
      hr = lpDev8->SetRenderState(D3DRS_CLIPPLANEENABLE, 1);
      CHECK_D3D_RESULT("SetRenderState", hr);
   }
}

//----------------------------

void I3D_driver::DisableClippingPlane(){

   if(d3d_caps8.MaxUserClipPlanes){
      HRESULT hr;
      hr = lpDev8->SetRenderState(D3DRS_CLIPPLANEENABLE, 0);
      CHECK_D3D_RESULT("SetRenderState", hr);
   }
}

//----------------------------

/*
void I3D_driver::AdjustTxtClipping(){

   assert(drv_flags2&DRVF2_TEXCLIP_ON);

   for(int i=1; i<MAX_TEXTURE_STAGES; i++){
      if(stage_txt_op[i]==D3DTOP_DISABLE)
         break;
   }
   if(clip_txt_stage!=i){
      //lpDev8->SetTextureStageState(clip_txt_stage, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
      if(drv_flags&DRVF_LINFILTER){
         lpDev8->SetTextureStageState(clip_txt_stage, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
         lpDev8->SetTextureStageState(clip_txt_stage, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
      }
      {
         for(int i=clip_txt_stage; --i > 0; )
            lpDev8->SetTextureStageState(i, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
      }
      clip_txt_stage = -1;

      if(i<NumSimultaneousTextures()){
         clip_txt_stage = i;
         lpDev8->SetTextureStageState(i, D3DTSS_COLOROP, D3DTOP_SELECTARG2);
         lpDev8->SetTextureStageState(i, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
         lpDev8->SetTexture(i, tp_clip->GetD3DTexture());

         if(drv_flags&DRVF_LINFILTER){
            lpDev8->SetTextureStageState(i, D3DTSS_MINFILTER, D3DTEXF_POINT);
            lpDev8->SetTextureStageState(i, D3DTSS_MAGFILTER, D3DTEXF_POINT);
         }
         while(--i > 0)
            lpDev8->SetTextureStageState(i, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
      }
   }
}
*/

//----------------------------

I3D_RESULT I3D_driver::SetTexture(PI3D_texture tp){

   SetTexture1(0, tp);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_driver::SetState(I3D_RENDERSTATE st, dword value){

   HRESULT hr = D3D_OK;

   switch(st){
   case RS_LINEARFILTER:
      {
         bool is_on = (drv_flags&DRVF_LINFILTER);
         if((bool)value==is_on)
            break;
         drv_flags &= ~DRVF_LINFILTER;
         if(value)
            drv_flags |= DRVF_LINFILTER;

         D3DTEXTUREFILTERTYPE ftmin, ftmag;
         dword anisotropy = 1;
         switch(value){
         case 0: 
            ftmin = ftmag = D3DTEXF_POINT;
            min_filter[0] = min_filter[1] = ftmin;
            break;
         case 1: 
            {
               ftmin = D3DTEXF_LINEAR;
               ftmag = D3DTEXF_LINEAR;
                              //choose anisotropic filtering, if available
               if((drv_flags&DRVF_USE_ANISO_FILTER) && 
                  (d3d_caps8.RasterCaps&D3DPRASTERCAPS_ANISOTROPY) &&
                  (d3d_caps8.TextureFilterCaps&D3DPTFILTERCAPS_MINFANISOTROPIC) &&
                  d3d_caps8.MaxAnisotropy){

                  ftmin = D3DTEXF_ANISOTROPIC;
                  anisotropy = d3d_caps8.MaxAnisotropy;
               }
               min_filter[0] = D3DTEXF_LINEAR;
               min_filter[1] = ftmin;
            }
            break;
         }
         HRESULT hr;
         for(int i=0; i<MAX_TEXTURE_STAGES; i++){
            D3DTEXTUREFILTERTYPE minf = min_filter[stage_anisotropy[i]];
            hr = lpDev8->SetTextureStageState(i, D3DTSS_MINFILTER, stage_minf[i] = minf);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(i, D3DTSS_MAGFILTER, ftmag);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            hr = lpDev8->SetTextureStageState(i, D3DTSS_MAXANISOTROPY, anisotropy);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
      break;

   case RS_ANISO_FILTERING:
      {
         drv_flags &= ~DRVF_USE_ANISO_FILTER;
         if(value)
            drv_flags |= DRVF_USE_ANISO_FILTER;
                              //apply change
         bool filter_on = (drv_flags&DRVF_LINFILTER);
         if(filter_on){
            drv_flags &= ~DRVF_LINFILTER;
            SetState(RS_LINEARFILTER, true);
         }
      }
      break;

   case RS_MIPMAP:
      {
         bool is_on = (drv_flags&DRVF_MIPMAP);
         if((bool)value==is_on)
            break;

         if(value)
            drv_flags |= DRVF_MIPMAP;
         else
            drv_flags &= ~DRVF_MIPMAP;

         D3DTEXTUREFILTERTYPE ft;
         if(!value)
            ft = D3DTEXF_NONE;
         else
            ft = D3DTEXF_LINEAR;
         for(int i=0; i<MAX_TEXTURE_STAGES; i++){
            hr = lpDev8->SetTextureStageState(i, D3DTSS_MIPFILTER, ft);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
      break;

   case RS_LOADMIPMAP:
      drv_flags &= ~DRVF_LOADMIPMAP;
      if(value) drv_flags |= DRVF_LOADMIPMAP;
      break;

      /*
   case RS_POINTFOG:
      drv_flags &= ~DRVF_POINTFOG;
      if(value) drv_flags |= DRVF_POINTFOG;
      break;
      */

   case RS_ENVMAPPING:
      drv_flags2 &= ~DRVF2_ENVMAPPING;
      if(value) drv_flags2 |= DRVF2_ENVMAPPING;
      break;

   case RS_BUMPMAPPING:
      drv_flags2 &= ~DRVF2_BUMPMAPPING;
      if(value) drv_flags2 |= DRVF2_BUMPMAPPING;
      break;

   case RS_TEXTURECOMPRESS:
      drv_flags &= ~DRVF_TEXTURECOMPRESS;
      if(value) drv_flags |= DRVF_TEXTURECOMPRESS;
      break;

   case RS_USEZB:
      {
         bool is_on = (drv_flags&DRVF_USEZB);
         bool on = value;
         if(is_on!=on){
            if(on) drv_flags |= DRVF_USEZB;
            else drv_flags &= ~DRVF_USEZB;
            D3DZBUFFERTYPE zb_type;
            if(!on){
               zb_type = D3DZB_FALSE;
            }else{
               if((init_data.flags&I3DINIT_WBUFFER) &&
                  (d3d_caps8.RasterCaps&D3DPRASTERCAPS_WBUFFER)){

                  zb_type = D3DZB_USEW;
               }else{
                  zb_type = D3DZB_TRUE;
               }
            }
            hr = lpDev8->SetRenderState(D3DRS_ZENABLE, zb_type);
            CHECK_D3D_RESULT("SetRenderState", hr);
         }
      }
      break;

   case RS_CLEAR:
      drv_flags &= ~DRVF_CLEAR;
      if(value) drv_flags |= DRVF_CLEAR;
      break;

   case RS_DITHER:
      {
         bool is_on = (drv_flags&DRVF_DITHER);
         if((bool)value==is_on)
            break;

         drv_flags &= ~DRVF_DITHER;
         if(value)
            drv_flags |= DRVF_DITHER;

         hr = lpDev8->SetRenderState(D3DRS_DITHERENABLE, value);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
      break;

   case RS_WIREFRAME:
      {
         bool is_on = (drv_flags&DRVF_WIREFRAME);
         if((bool)value==is_on)
            break;

         drv_flags &= ~DRVF_WIREFRAME;
         if(value)
            drv_flags |= DRVF_WIREFRAME;

         hr = lpDev8->SetRenderState(D3DRS_FILLMODE, value ? D3DFILL_WIREFRAME : D3DFILL_SOLID);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
      break;

   case RS_DRAWBOUNDBOX:
      drv_flags &= ~DRVF_DRAWBBOX;
      if(value) drv_flags |= DRVF_DRAWBBOX;
      break;

   case RS_DRAWHRBOUNDBOX:
      drv_flags &= ~DRVF_DRAWHRBBOX;
      if(value) drv_flags |= DRVF_DRAWHRBBOX;
      break;

   case RS_DRAWLINKS:
      drv_flags &= ~DRVF_DRAWHRLINK;
      if(value) drv_flags |= DRVF_DRAWHRLINK;
      break;

   case RS_DRAWPORTALS:
      drv_flags &= ~DRVF_DRAWPORTALS;
      if(value) drv_flags |= DRVF_DRAWPORTALS;
      break;

   case RS_DRAWSECTORS:
      drv_flags &= ~DRVF_DRAWSECTORS;
      if(value) drv_flags |= DRVF_DRAWSECTORS;
      break;

   case RS_TEXTUREDITHER:
      drv_flags &= ~DRVF_TEXTDITHER;
      if(value) drv_flags |= DRVF_TEXTDITHER;
      break;

   case RS_DRAWVISUALS:
      drv_flags &= ~DRVF_DRAWVISUALS;
      if(value) drv_flags |= DRVF_DRAWVISUALS;
      break;

   case RS_FOG:
      drv_flags &= ~DRVF_USEFOG;
      if(value) drv_flags |= DRVF_USEFOG;
      break;

   case RS_USESHADOWS:
      drv_flags &= ~DRVF_USESHADOWS;
      if(value) drv_flags |= DRVF_USESHADOWS;
      break;

   case RS_DEBUGDRAWSHADOWS:
      drv_flags &= ~DRVF_DEBUGDRAWSHADOWS;
      if(value) drv_flags |= DRVF_DEBUGDRAWSHADOWS;
      break;

   case RS_DEBUGDRAWSHDRECS:
      drv_flags &= ~DRVF_DEBUGDRAWSHDRECS;
      if(value) drv_flags |= DRVF_DEBUGDRAWSHDRECS;
      break;

   case RS_DEBUGDRAWSTATIC:
      drv_flags2 &= ~DRVF2_DEBUGDRAWSTATIC;
      if(value) drv_flags2 |= DRVF2_DEBUGDRAWSTATIC;
      break;

   case RS_DRAWVOLUMES:
      drv_flags2 &= ~DRVF2_DRAWVOLUMES;
      if(value) drv_flags2 |= DRVF2_DRAWVOLUMES;
      break;

   case RS_DRAWLIGHTS:
      drv_flags2 &= ~DRVF2_DRAWLIGHTS;
      if(value) drv_flags2 |= DRVF2_DRAWLIGHTS;
      break;

   case RS_DRAWSOUNDS:
      drv_flags2 &= ~DRVF2_DRAWSOUNDS;
      if(value) drv_flags2 |= DRVF2_DRAWSOUNDS;
      break;

   case RS_DRAWMIRRORS:
      drv_flags &= ~DRVF_DRAWMIRRORS;
      if(value) drv_flags |= DRVF_DRAWMIRRORS;
      break;

   case RS_USELMAPPING:
      {
         drv_flags &= ~DRVF_USELMAPPING;
         if(value){
            if(!(drv_flags&DRVF_SINGLE_PASS_MODULATE2X)){
                              //check if required blending mode supported
               if(!((d3d_caps8.SrcBlendCaps&D3DPBLENDCAPS_DESTCOLOR) &&
                  (d3d_caps8.DestBlendCaps&D3DPBLENDCAPS_SRCCOLOR))){
                              //can't use desired blend mode, light-mapping not supported
                  break;
               }
            }
            drv_flags |= DRVF_USELMAPPING;
         }
      }
      break;

   case RS_DRAWLMTEXTURES:
      drv_flags &= ~DRVF_DRAWLMTEXTURES;
      if(value) drv_flags |= DRVF_DRAWLMTEXTURES;
      break;

   case RS_DEBUGDRAWBSP:
      drv_flags &= ~DRVF_DEBUGDRAWBSP;
      if(value) drv_flags |= DRVF_DEBUGDRAWBSP;
      break;

   case RS_DEBUGDRAWDYNAMIC:
      drv_flags &= ~DRVF_DEBUGDRAWDYNAMIC;
      if(value) drv_flags |= DRVF_DEBUGDRAWDYNAMIC;
      break;

   case RS_DRAWCAMERAS:
      drv_flags2 &= ~DRVF2_DRAWCAMERAS;
      if(value) drv_flags2 |= DRVF2_DRAWCAMERAS;
      break;

   case RS_DRAWDUMMYS:
      drv_flags2 &= ~DRVF2_DRAWDUMMYS;
      if(value) drv_flags2 |= DRVF2_DRAWDUMMYS;
      break;

   case RS_DRAWTARGETS:
      drv_flags2 &= ~DRVF2_DRAWTARGETS;
      if(value) drv_flags2 |= DRVF2_DRAWTARGETS;
      break;

   case RS_DRAWOCCLUDERS:
      drv_flags2 &= ~DRVF2_DRAWOCCLUDERS;
      if(value) drv_flags2 |= DRVF2_DRAWOCCLUDERS;
      break;

   case RS_DRAWTEXTURES:
      drv_flags2 &= ~DRVF2_DRAWTEXTURES;
      if(value) drv_flags2 |= DRVF2_DRAWTEXTURES;
      break;

   case RS_LMTRUECOLOR:
      {
         bool was_on = (drv_flags2&DRVF2_LMTRUECOLOR);
         if((bool)value!=was_on){
            drv_flags2 &= ~DRVF2_LMTRUECOLOR;
            if(value) drv_flags2 |= DRVF2_LMTRUECOLOR;
            InitLMConvertor();
            PurgeLMTextures();
         }
      }
      break;

   case RS_LMDITHER:
      {
         bool was_on = (drv_flags2&DRVF2_LMDITHER);
         if((bool)value!=was_on){
            drv_flags2 &= ~DRVF2_LMDITHER;
            if(value) drv_flags2 |= DRVF2_LMDITHER;
            if(pf_light_maps.bytes_per_pixel <= 2){
               InitLMConvertor();
               PurgeLMTextures();
            }
         }
      }
      break;

   case RS_LM_AA_RATIO:
      if((int)value <= 0 || value > 100)
         return I3DERR_INVALIDPARAMS;
      lm_create_aa_ratio = value;
      break;

   case RS_DRAWJOINTS:
      drv_flags2 &= ~DRVF2_DRAWJOINTS;
      if(value) drv_flags2 |= DRVF2_DRAWJOINTS;
      break;

   case RS_DEBUG_INT0: case RS_DEBUG_INT1: case RS_DEBUG_INT2: case RS_DEBUG_INT3:
   case RS_DEBUG_INT4: case RS_DEBUG_INT5: case RS_DEBUG_INT6: case RS_DEBUG_INT7:
      debug_int[st-RS_DEBUG_INT0] = value;
      break;

   case RS_DEBUG_FLOAT0: case RS_DEBUG_FLOAT1: case RS_DEBUG_FLOAT2: case RS_DEBUG_FLOAT3:
   case RS_DEBUG_FLOAT4: case RS_DEBUG_FLOAT5: case RS_DEBUG_FLOAT6: case RS_DEBUG_FLOAT7:
      debug_float[st-RS_DEBUG_FLOAT0] = *(float*)&value;
      break;

   case RS_LOD_INDEX:
      force_lod_index = Max(-1, (int)value);
      break;

   case RS_LOD_SCALE:
      lod_scale = IntAsFloat(value);
      break;

   case RS_USE_OCCLUSION:
      drv_flags2 &= ~DRVF2_USE_OCCLUSION;
      if(value) drv_flags2 |= DRVF2_USE_OCCLUSION;
      break;

   case RS_DEBUGMATVISUALIZE:
      debug_draw_mat = value;
      break;
   }

   if(FAILED(hr))
      return I3DERR_GENERIC;
   return I3D_OK;
}

//----------------------------

dword I3D_driver::GetState(I3D_RENDERSTATE st) const{

   switch(st){
   case RS_LINEARFILTER: return (bool)(drv_flags&DRVF_LINFILTER);
   case RS_ANISO_FILTERING: return (bool)(drv_flags&DRVF_USE_ANISO_FILTER); break;
   case RS_MIPMAP: return (bool)(drv_flags&DRVF_MIPMAP);
   case RS_TEXTURECOMPRESS: return (bool)(drv_flags&DRVF_TEXTURECOMPRESS);
   case RS_USEZB: return (bool)(drv_flags&DRVF_USEZB);
   case RS_CLEAR: return (bool)(drv_flags&DRVF_CLEAR);
   case RS_DITHER: return (bool)(drv_flags&DRVF_DITHER);
   case RS_WIREFRAME: return (bool)(drv_flags&DRVF_WIREFRAME);
   case RS_DRAWBOUNDBOX: return (bool)(drv_flags&DRVF_DRAWBBOX);
   case RS_DRAWHRBOUNDBOX: return (bool)(drv_flags&DRVF_DRAWHRBBOX);
   case RS_DRAWLINKS: return (bool)(drv_flags&DRVF_DRAWHRLINK);
   case RS_LOADMIPMAP: return (bool)(drv_flags&DRVF_LOADMIPMAP);
   //case RS_POINTFOG: return (bool)(drv_flags&DRVF_POINTFOG);
   case RS_ENVMAPPING: return (bool)(drv_flags2&DRVF2_ENVMAPPING);
   case RS_BUMPMAPPING: return (bool)(drv_flags2&DRVF2_BUMPMAPPING);
   case RS_DRAWSECTORS: return (bool)(drv_flags&DRVF_DRAWSECTORS);
   case RS_DRAWPORTALS: return (bool)(drv_flags&DRVF_DRAWPORTALS);
   case RS_TEXTUREDITHER: return (bool)(drv_flags&DRVF_TEXTDITHER);
   case RS_DRAWVISUALS: return (bool)(drv_flags&DRVF_DRAWVISUALS);
   case RS_FOG: return (bool)(drv_flags&DRVF_USEFOG);
   //case RS_TEXTURELOWDETAIL: return (bool)(drv_flags&DRVF_TEXTURELOWDET);
   case RS_DRAWVOLUMES: return (bool)(drv_flags2&DRVF2_DRAWVOLUMES);
   case RS_DRAWLIGHTS: return (bool)(drv_flags2&DRVF2_DRAWLIGHTS);
   case RS_DRAWSOUNDS: return (bool)(drv_flags2&DRVF2_DRAWSOUNDS);
   case RS_USELMAPPING: return (bool)(drv_flags&DRVF_USELMAPPING);
   case RS_DRAWLMTEXTURES: return (bool)(drv_flags&DRVF_DRAWLMTEXTURES);
   case RS_DRAWCAMERAS: return (bool)(drv_flags2&DRVF2_DRAWCAMERAS);
   case RS_DRAWDUMMYS: return (bool)(drv_flags2&DRVF2_DRAWDUMMYS);
   case RS_DRAWTARGETS: return (bool)(drv_flags2&DRVF2_DRAWTARGETS);
   case RS_DRAWOCCLUDERS: return (bool)(drv_flags2&DRVF2_DRAWOCCLUDERS);
   case RS_DRAWTEXTURES: return (bool)(drv_flags2&DRVF2_DRAWTEXTURES);
   case RS_LMTRUECOLOR: return (bool)(drv_flags2&DRVF2_LMTRUECOLOR);
   case RS_LMDITHER: return (bool)(drv_flags2&DRVF2_LMDITHER);
   case RS_LM_AA_RATIO: return lm_create_aa_ratio;
   case RS_DRAWJOINTS: return (bool)(drv_flags2&DRVF2_DRAWJOINTS);
   case RS_DEBUG_INT0: case RS_DEBUG_INT1: case RS_DEBUG_INT2: case RS_DEBUG_INT3:
   case RS_DEBUG_INT4: case RS_DEBUG_INT5: case RS_DEBUG_INT6: case RS_DEBUG_INT7:
      return debug_int[st-RS_DEBUG_INT0];
   case RS_DEBUG_FLOAT0: case RS_DEBUG_FLOAT1: case RS_DEBUG_FLOAT2: case RS_DEBUG_FLOAT3:
   case RS_DEBUG_FLOAT4: case RS_DEBUG_FLOAT5: case RS_DEBUG_FLOAT6: case RS_DEBUG_FLOAT7:
      return FloatAsInt(debug_float[st-RS_DEBUG_FLOAT0]);
   case RS_LOD_INDEX: return force_lod_index;
   case RS_LOD_SCALE: return FloatAsInt(lod_scale); break;
   case RS_USE_OCCLUSION: return (bool)(drv_flags2&DRVF2_USE_OCCLUSION);
   case RS_DEBUGMATVISUALIZE: return debug_draw_mat;

   }
   return 0;
}

//----------------------------

I3D_RESULT I3D_driver::GetStats(I3DSTATSINDEX indx, void *data){

   switch(indx){

   case I3DSTATS_VOLUME:
      {
         PI3D_stats_volume st = (PI3D_stats_volume)data;
         st->sphere_sphere = col_stats[DRV_COL_STATS_S_S];
         st->sphere_rectangle = col_stats[DRV_COL_STATS_S_R];
         st->sphere_box = col_stats[DRV_COL_STATS_S_B];
         st->sphere_face = col_stats[DRV_COL_STATS_S_F];
         st->line_sphere = col_stats[DRV_COL_STATS_L_S];
         st->line_rectangle = col_stats[DRV_COL_STATS_L_R];
         st->line_box = col_stats[DRV_COL_STATS_L_B];
         st->line_face = col_stats[DRV_COL_STATS_L_F];
         memset(&col_stats, 0, sizeof(col_stats));
      }
      break;

   case I3DSTATS_RENDER:
      {
         PI3D_stats_render st = (PI3D_stats_render)data;
         *st = render_stats;
         memset(&render_stats, 0, sizeof(render_stats));
      }
      break;

   case I3DSTATS_SCENE:
      {
         PI3D_stats_scene st = (PI3D_stats_scene)data;
         *st = scene_stats;
         memset(&scene_stats, 0, sizeof(scene_stats));
      }
      break;

   case I3DSTATS_VIDEO:
      {
         PI3D_stats_video st = (PI3D_stats_video)data;
         st->txt_total = managed_textures.size();
         st->txt_vidmem = 0;
         st->txt_sys_mem = 0;
         st->txt_video_mem = 0;
         for(int i=managed_textures.size(); i--; ){
            PI3D_texture tp = managed_textures[i];
            dword txtf = tp->GetTxtFlags();
            //if(!(txtf&TXTF_UPLOAD_NEEDED))
            if(!(txtf&TXTF_UPLOAD_NEEDED))
               ++st->txt_vidmem;
            const IImage *img = tp->GetSysmemImage();
                              //compute memory of image
            const S_pixelformat *pf = img->GetPixelFormat();
            dword img_bytes;
            if(pf->flags&PIXELFORMAT_COMPRESS){
               img_bytes = img->SizeX() * img->SizeY() * 1;
               if(pf->flags&PIXELFORMAT_ALPHA)
                  img_bytes *= 2;
            }else{
               img_bytes = img->SizeX() * img->SizeY() * pf->bytes_per_pixel;
            }
                              //add mipmap levels
            //img_bytes += img_bytes*.3333f;
            st->txt_sys_mem += img_bytes;

            if(!(txtf&TXTF_UPLOAD_NEEDED))
               st->txt_video_mem += img_bytes;
         }
                              //get vidmem stats from D3D
         st->vidmem_free = lpDev8->GetAvailableTextureMem();
         st->vidmem_total = st->txt_video_mem + st->vidmem_free;

         st->txtmem_total = st->vidmem_total;
         st->txtmem_free = st->vidmem_free;
      }
      break;

   case I3DSTATS_MEMORY:
      {
         PI3D_stats_memory st = (PI3D_stats_memory)data;
         st->used_other = other_memory_size;
         st->available = 0;
         dbase.GetStats(st->dbase_size, st->dbase_used);

#ifdef _MSC_VER
         st->num_blocks = 0;
         st->used = num_alloc_bytes;
         /*
         st->used = 0;
         _heapinfo hi;
         hi._pentry = NULL;
         while(true){
            int status = _heapwalk(&hi);
            if(status!=_HEAPOK) break;
            if(hi._useflag == _USEDENTRY){
               ++st->num_blocks;
               st->used += hi._size;
            }
         }
         */
#else
         memset(st, 0, sizeof(I3D_stats_memory));
         //return I3DERR_UNSUPPORTED;
#endif
      }
      break;

   case I3DSTATS_SOUND:
      {
         PISND_driver lpis = GetSoundInterface();
         PI3D_stats_sound st = (PI3D_stats_sound)data;
         if(lpis){
            set<PISND_source> snd_srcs;
            st->snds_audible = scene_stats.frame_count[FRAME_SOUND];
            scene_stats.frame_count[FRAME_SOUND] = 0;
                              //count in-card sounds
            st->snds_in_card = 0;
            st->mem_used = 0;
            for(int i=sounds.size(); i--; ){
               PI3D_sound snd = sounds[i];
               if(snd->voice){
                  if(snd->snd_flags&SNDF_HARDWARE)
                     ++st->snds_in_card;
               }
               if(snd->GetSoundSource()){
                  if(!snd->IsStreaming())
                     snd_srcs.insert(snd->GetSoundSource());
                  else{
                     const S_wave_format &wf = *snd->GetSoundSource()->GetFormat();
                     st->mem_used += (((int)(wf.samples_per_sec*wf.bytes_per_sample*1.0f)) & -16);
                  }
               }
            }
                              //collect size of samples
            for(set<PISND_source>::const_iterator it = snd_srcs.begin(); it!=snd_srcs.end(); ++it){
               const S_wave_format *wf = (*it)->GetFormat();
               if(wf)
                  st->mem_used += wf->size;
            }
                              //get info from sound driver
            dword caps[6];
            lpis->GetCaps(caps);
            st->buf_hw_all = caps[0];
            st->buf_hw_free = caps[1];
            st->buf_3d_all = caps[2];
            st->buf_3d_free = caps[3];
            st->hw_mem_all = caps[4];
            st->hw_mem_free = caps[5];
         }else{
            memset(st, 0, sizeof(I3D_stats_sound));
         }
         st->snds_all = sounds.size();
      }
      break;
   case I3DSTATS_BSP:
      {
         PI3D_stats_bsp st = (PI3D_stats_bsp)data;
         *st = bsp_stats;
         //memset(&bsp_stats, 0, sizeof(bsp_stats));
      }
      break;
   case I3DSTATS_VOLTREE:
      {
         PI3D_stats_voltree st = (PI3D_stats_voltree)data;
         *st = voltree_stats;
         //memset(&voltree_stats, 0, sizeof(voltree_stats));
         voltree_stats.moving_static = 0;
      }
      break;

   default:
      return I3DERR_INVALIDPARAMS;
   }
   return I3D_OK;
}

//----------------------------

PI3D_material I3D_driver::GetIcon(int res_id){

   while(icon_materials.size() <= res_id)
      icon_materials.push_back(NULL);

   if(!icon_materials[res_id]){
      PI3D_texture txt;
      HINSTANCE hinst = GetModuleHandle(dll_filename);
      HGLOBAL hg = LoadResource(hinst, FindResource(hinst, (const char*)res_id, RT_BITMAP));
      if(!hg)
         return NULL;
      /*
      {
         BITMAPINFO *bi = (BITMAPINFO*)hg;
         bi->bmiHeader.biXPelsPerMeter = 0;
      }
      */

      I3D_CREATETEXTURE ct;
      memset(&ct, 0, sizeof(ct));
      ct.flags = TEXTMAP_DIFFUSE | TEXTMAP_TRANSP |
         //TEXTMAP_NOMIPMAP |
         TEXTMAP_MIPMAP |
         TEXTMAP_MEM_PTR;
      ct.text_name = (const char*)hg;
      CreateTexture(&ct, &txt);

      PI3D_material mat = CreateMaterial();
      mat->SetTexture(txt);
      if(txt)
         txt->Release();

      icon_materials[res_id] = mat;
   }
   return icon_materials[res_id];
}

//----------------------------

PI3D_SCNCALLBACK I3D_driver::SetCallback(PI3D_SCNCALLBACK cb, dword cbf){

   PI3D_SCNCALLBACK ret = cb_proc;
   if(cb){
      if(cb_proc!=cb) cb_flags = 0;
      cb_proc = cb;
      cb_flags |= cbf;
   }else{
      cb_flags &= ~cbf;
      if(!cb_flags) cb_proc=NULL;
   }
   return ret;
}

//----------------------------

void I3D_driver::PurgeLMTextures(){

   for(int i=managed_textures.size(); i--; ){
      PI3D_texture tp = managed_textures[i];
                           //don't refresh LMap textures
      if(UnregLMTexture(tp))
         tp->SetDirty();
   }
}

//----------------------------

I3D_RESULT I3D_driver::SetDataBase(const char *db_filename, dword size, int max_keep_days){

   if(db_filename){
      bool b = dbase.Open(db_filename, size, max_keep_days);
      if(!b)
         return I3DERR_GENERIC;
      /*
      dword ft[2] = {0x29462c10, 0x01c15c76};
      C_cache *h = dbase.OpenRecord("Models.0\\GERMAN1.i3d+Hlva*", ft);
      dbase.CloseRecord(h);
      */
      return I3D_OK;
   }else{
      dbase.Close();
      return I3D_OK;
   }
}

//----------------------------

I3D_RESULT I3D_driver::FlushDataBase(){

   if(!dbase.IsOpen()) return I3DERR_NOTINITIALIZED;
   dbase.Flush();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_driver::GetVersion(dword &v0, dword &v1){

   v0 = 0x35fe9d7c;
   v1 = 0x0ac576b3;
   return I3D_OK;
}

//----------------------------

/*
int I3D_driver::AllocatePaletteNumber(){

                              //find lowest possible number not in the list
   int new_num = 0;
   if(used_palettes.size()){
      set<int>::iterator it, it_next;
      for(it=used_palettes.begin(); ; it = it_next){
         it_next = it; ++it_next;
         if(it_next == used_palettes.end() || ((*it)+1) != (*it_next)){
            new_num = (*it) + 1;
            break;
         }
      }
   }
   assert(used_palettes.find(new_num) == used_palettes.end());
   used_palettes.insert(new_num);
   return new_num;
}

//----------------------------

void I3D_driver::FreePaletteNumber(int p_num){

   set<int>::iterator it = used_palettes.find(p_num);
   if(it == used_palettes.end()){
      assert(0);
      return;
   }
   used_palettes.erase(it);
}
*/

//----------------------------

dword I3D_driver::NumProperties(dword visual_type) const{

   for(int i=num_visual_plugins; i--; ){
      if(visual_plugins[i].visual_type == visual_type){
         return visual_plugins[i].num_props;
      }
   }
   return -1;
}

//----------------------------

const char *I3D_driver::GetPropertyName(dword visual_type, dword index) const{

   for(int i=num_visual_plugins; i--; ){
      if(visual_plugins[i].visual_type == visual_type){
         dword num = visual_plugins[i].num_props;
         if(index >= num)
            return NULL;
         return visual_plugins[i].props[index].prop_name;
      }
   }
   return NULL;
}

//----------------------------

I3D_PROPERTYTYPE I3D_driver::GetPropertyType(dword visual_type, dword index) const{

   for(int i=num_visual_plugins; i--; ){
      if(visual_plugins[i].visual_type == visual_type){
         dword num = visual_plugins[i].num_props;
         if(index >= num)
            return I3DPROP_NULL;
         return visual_plugins[i].props[index].prop_type;
      }
   }
   return I3DPROP_NULL;
}

//----------------------------

PI3D_procedural_base I3D_driver::GetProcedural(const char *filename,
   const S_load_context &lc, const char *frame_name, const char *mat_name){

                              //search in cache
   for(int i=procedural_cache.size(); i--; ){
      PI3D_procedural_base pb = procedural_cache[i];
      if(pb->GetName().Match(filename)){
         pb->AddRef();
         return pb;
      }
   }
                              //not found in cache, create new
   
                              //open file and create procedural from there
   C_cache ck;
   for(i=dirs[I3DDIR_PROCEDURALS].size(); i--; ){
      if(ck.open(C_fstring("%s\\%s.txt", (const char*)dirs[I3DDIR_PROCEDURALS][i], filename), CACHE_READ))
         break;
   }
   if(i==-1){
      lc.REPORT_ERR(C_fstring("Cannot create procedural '%s': file not found. ('%s', in '%s')",
         filename, (const char*)lc.file_name, frame_name ? frame_name : mat_name));
      return NULL;
   }

   PI3D_procedural_base pb = NULL;

#define SKIP_WS while(*cp && isspace(*cp)) ++cp;

   I3D_RESULT ir = I3D_OK;
                              //read procedural declaration
   while(!ck.eof()){
      char line[256];
      ck.getline(line, sizeof(line));
      const char *cp = line;
                              //skip leading space
      SKIP_WS;
      
                              //ignore comments
      if(*cp == ';' || *cp==0) continue;
                              //read keyword
      if(!strncmp(cp, "Class", 5)){
         cp += 5;
         SKIP_WS;
         char class_name[128];
         if(sscanf(cp, "%128s", class_name) == 1){
                              //create specified procedural
            ir = CreateProcedural(this, filename, class_name, ck, lc, &pb);
            break;
         }
      }
   }
   ck.close();
   if(I3D_FAIL(ir)){
      const char *err = NULL;
      switch(ir){
      case I3DERR_NOFILE: err = "file not found"; break;
      }
      if(err){
         lc.REPORT_ERR(C_fstring("Procedural '%s': creation failed - %s. ('%s', in '%s')",
            filename, err, (const char*)lc.file_name, frame_name ? frame_name : mat_name));
      }
   }

   if(pb)
      procedural_cache.push_back(pb);
   return pb;
}

//----------------------------

void I3D_driver::UnregisterProcedural(PI3D_procedural_base pb){

   for(int i=procedural_cache.size(); i--; ){
      if(procedural_cache[i]==pb){
         procedural_cache[i] = procedural_cache.back(); procedural_cache.pop_back();
         break;
      }
   }
}

//----------------------------

                              //sound properties
I3D_RESULT I3D_driver::SetSoundEnvData(int set_id, I3D_SOUND_ENV_DATA_INDEX indx, dword data){

   if(indx>=I3D_SENV_LAST)
      return I3DERR_INVALIDPARAMS;

   t_env_properties::iterator it = env_properties.find(set_id);
                              //create new set, if not exists
   if(it==env_properties.end())
      it = env_properties.insert(pair<int, S_sound_env_properties>(set_id, S_sound_env_properties())).first;

   S_sound_env_properties &ep = (*it).second;

   bool raw_store = true;
   switch(indx){
   case I3D_SENV_F_ROOM_LEVEL_LF:
   case I3D_SENV_F_ROOM_LEVEL_HF:
   case I3D_SENV_F_DECAY_HF_RATIO:
   case I3D_SENV_F_REFLECTIONS_RATIO:
   case I3D_SENV_F_REVERB_RATIO:
   case I3D_SENV_F_ENV_DIFFUSION:
   case I3D_SENV_F_EMUL_VOLUME:
      {
         float f = IntAsFloat(data);
         f = Min(1.0f, Max(0.0f, f));
         data = FloatAsInt(f);
      }
      break;

#ifdef USE_EAX
   case I3D_SENV_I_REVERB_DECAY_TIME:
      data = (dword)Max((int)(EAXLISTENER_MINDECAYTIME*1000), Min((int)(EAXLISTENER_MAXDECAYTIME*1000), (int)data));
      break;

   case I3D_SENV_I_REFLECTIONS_DELAY:
      data = (dword)Max((int)(EAXLISTENER_MINREFLECTIONSDELAY*1000), Min((int)(EAXLISTENER_MAXREFLECTIONSDELAY*1000), (int)data));
      break;

   case I3D_SENV_I_REVERB_DELAY:
      data = (dword)Max((int)(EAXLISTENER_MINREVERBDELAY*1000), Min((int)(EAXLISTENER_MAXREVERBDELAY*1000), (int)data));
      break;
#endif

   case I3D_SENV_S_NAME:
      raw_store = false;
      ep.name = (const char*)data;
      break;

   default:
      //assert(0);
      return I3DERR_INVALIDPARAMS;
   }
                              //save value
   if(raw_store)
      ((dword*)&ep)[indx] = data;
   drv_flags |= DRVF_SOUND_ENV_RESET;

   return I3D_OK;
}

//----------------------------

dword I3D_driver::GetSoundEnvData(int set_id, I3D_SOUND_ENV_DATA_INDEX indx) const{

   if(indx>=I3D_SENV_LAST)
      return -1;

   t_env_properties::const_iterator it = env_properties.find(set_id);
   if(it==env_properties.end())
      return -1;
   const S_sound_env_properties &ep = (*it).second;

   switch(indx){
   case I3D_SENV_S_NAME:
      return (dword)(const char*)ep.name;
   }
   return ((dword*)&ep)[indx];
}

//----------------------------

void I3D_driver::DisableLeakWarning() const{

}

//----------------------------

