#ifndef __DRIVER_H
#define __DRIVER_H

#include <vector>
#include <d3d8.h>
#include <c_string.hpp>
#include "texture.h"
#include "material.h"
#include "vb_mgr.h"
#include "database.h"

#include "nvidia\nvlink\nvlink.h"   //vertex shader stuff


//----------------------------
                              //I3D class identificators
enum I3D_CLID{
   I3D_CLID_DRIVER,
   I3D_CLID_SCENE,

   I3D_CLID_FRAME,
   I3D_CLID_DUMMY,
   I3D_CLID_VISUAL,
   I3D_CLID_MODEL,
   I3D_CLID_LIGHT,
   I3D_CLID_SOUND,
   I3D_CLID_SECTOR,
   I3D_CLID_CAMERA,
   I3D_CLID_TARGET,
   I3D_CLID_VOLUME,
   I3D_CLID_USER,
   I3D_CLID_JOINT,
   I3D_CLID_OCCLUDER,
   I3D_CLID_SINGLE_MESH,

   I3D_CLID_TEXTURE,
   I3D_CLID_MATERIAL,
   I3D_CLID_MESH_BASE,
   I3D_CLID_MESH_OBJECT,
   I3D_CLID_PORTAL,
   I3D_CLID_ANIMATION_BASE,
   I3D_CLID_ANIMATION_SET,
   I3D_CLID_INTERPOLATOR,
   I3D_CLID_ANIM_INTERPOLATOR,
   I3D_CLID_PROCEDURAL,

   I3D_CLID_LAST
};

//----------------------------

struct S_vertex_flare{
   float x, y;
   float u, v;
};

#define FVF_VERTEX_FLARE (D3DFVF_XYZ | D3DFVF_TEX1)
#define NUM_FLARE_VERTS 4

//----------------------------

class C_LM_alloc_block;
class C_LM_surface;

//----------------------------

#define MAX_TEXTURE_STAGES 3  //max texture stages we may use simlultaneously

template<class T>
int VectorIndex(const vector<T> &v, const T &itm){
   for(int i=v.size(); i--; ) if(v[i]==itm) break;
   return i;
}


class I3D_driver{
   dword ref;
                              //drv_flags:
#define DRVF_USEZB            1
#define DRVF_CLEAR            2
#define DRVF_TEXTURECOMPRESS  4
#define DRVF_WIREFRAME        8
#define DRVF_DRAWBBOX         0x10
#define DRVF_DRAWHRBBOX       0x20
#define DRVF_DRAWHRLINK       0x40
#define DRVF_LOADMIPMAP       0x80
#define DRVF_LINFILTER        0x100
#define DRVF_MIPMAP           0x200
#define DRVF_BUMPMAPPING      0x400
#define DRVF_DRAWMIRRORS      0x800
#define DRVF_HASSTENCIL       0x1000      //stencil buffer is attached
#define DRVF_DRAWSECTORS      0x2000
#define DRVF_DRAWPORTALS      0x4000
#define DRVF_TEXTDITHER       0x8000
#define DRVF_DRAWVISUALS      0x10000
#define DRVF_USEFOG           0x20000
#define DRVF_CANRENDERSHADOWS 0x40000     //hardware is capable to render shadows
#define DRVF_USE_ANISO_FILTER 0x80000     //use anisotropic filtering (RS_ANISO_FILTERING)
#define DRVF_USESHADOWS       0x100000
#define DRVF_DEBUGDRAWSHADOWS 0x200000
#define DRVF_DEBUGDRAWSHDRECS 0x400000
#define DRVF_DEBUGDRAWBSP     0x800000
#define DRVF_DEBUGDRAWDYNAMIC 0x1000000
#define DRVF_DITHER           0x2000000
#define DRVF_USELMAPPING      0x4000000
#define DRVF_DRAWLMTEXTURES   0x8000000
#define DRVF_SINGLE_PASS_MODULATE   0x10000000  //single-pass multi-texturing for MODULATE
#define DRVF_SINGLE_PASS_MODULATE2X 0x20000000  //single-pass multi-texturing for MODULATE2X
#define DRVF_ANISOTROPY_ON    0x40000000  //anisotropy filtering is on
#define DRVF_SOUND_ENV_RESET  0x80000000  //can use D3DTA_COMPLEMENT blend stage argument

                              //drv_flags2:
#define DRVF2_DRAWCAMERAS     1
#define DRVF2_DRAWDUMMYS      2
#define DRVF2_DRAWTARGETS     4
#define DRVF2_DRAWOCCLUDERS   8
#define DRVF2_DRAWTEXTURES    0x10
#define DRVF2_DRAWVOLUMES     0x20
#define DRVF2_CAN_MODULATE2X  0x40     //can set D3DTSS_COLOROP = D3DTOP_MODULATE2X
#define DRVF2_CAN_TXTOP_ADD   0x80     //can set D3DTSS_COLOROP = D3DTOP_ADD
#define DRVF2_DRAWSOUNDS      0x100
#define DRVF2_DRAWLIGHTS      0x200
#define DRVF2_LMTRUECOLOR     0x400    //use true-color lightmaps
#define DRVF2_LMDITHER        0x800    //dither highcolor lightmaps
#define DRVF2_DRAWJOINTS      0x1000    
#define DRVF2_DIRECT_TRANSFORM 0x2000  //using direct transforming, no intermediate ProcessVertices calls
//...
#define DRVF2_USE_OCCLUSION   0x8000   //use occlusion testing
//...
#define DRVF2_ENVMAPPING      0x20000  //use environment mapping
#define DRVF2_BUMPMAPPING     0x40000  //use bump mapping
#define DRVF2_DEBUGDRAWSTATIC 0x80000  //draw frames which are collision static

//----------------------------

   LPDIRECT3D8 lpD3D8;        //if NULL, interface is uninitialized
   D3DCAPS8 d3d_caps8;
   IDirect3DDevice8 *lpDev8;
   dword drv_flags, drv_flags2;
   I3DINIT init_data;
   dword render_time;         //counter set to current time (IGraph::ReadTimer) each time a scene is rendered
   dword in_scene_count;

   int debug_draw_mat;        //material we're rendering

   bool is_hal;
public:

//----------------------------
// Check if hardware device is being used.
   inline bool IsHardware() const{ return is_hal; }

//----------------------------
// Check if device supports hardware vertex processing.
   inline bool IsDirectTransform() const{ return (drv_flags2&DRVF2_DIRECT_TRANSFORM); }

   inline const D3DCAPS8 *GetCaps() const{ return &d3d_caps8; }

   HINSTANCE h_nvlinker;      //HINSTANCE of nvlinker
   INVLink *nv_linker[2];     //[vertex_shading|pixel_shading]

#define MAX_VS_LIGHTS 6       //maximal number of lights for vertex shader
#define MAX_VS_FRAGMENTS (MAX_VS_LIGHTS+10)  //maximal number of fragments a shader may have
#define MAX_VS_BLEND_MATRICES 17

   enum E_VS_FRAGMENT{
      VSF_TRANSFORM,          //simple transformation by world-view-proj matrix
      VSF_M_PALETTE_TRANSFORM,//transformation using matrix palette blending
      VSF_TRANSFORM_MORPH1,   //transformation using weighted blending - 1 morph
      VSF_TRANSFORM_MORPH2,   //transformation using weighted blending - 2 morph2
      VSF_TRANSFORM_MIRROR,   //transformation setting z to 1.0 and w to max (clearing ZBuf)
      VSF_TEST,               //used for debugging
      VSF_COPY_UV_0,          //copy uv coordinates set #0
      VSF_COPY_UV_1,          //copy uv coordinates set #1
      VSF_COPY_UV_0TO1,       //cross-copy
      VSF_COPY_UV_1TO0,       //cross-copy
      VSF_GENERATE_ENVUV_OS_1,//generate environment mapping uv coordinates
      VSF_GENERATE_ENVUV_WS_1,//'', but in world space
      VSF_GENERATE_ENVUV_OS_2,//the same pair as above, but into UV set 2
      VSF_GENERATE_ENVUV_WS_2,//
      VSF_SHIFT_UV_0,         //compute UV shifting on set #0
      VSF_TEXT_PROJECT,       //project texture for shadow rendering
      VSF_TEXT_CLIP_OS_1,     //texture clipping coordinates
      VSF_TEXT_CLIP_OS_2,
      VSF_TEXT_CLIP_OS_3,
      VSF_TEXT_CLIP_WS_1,
      VSF_TEXT_CLIP_WS_2,
      VSF_TEXT_CLIP_WS_3,
      VSF_FOG_SIMPLE_OS,      //simple height fog
      VSF_FOG_SIMPLE_WS,      //simple height fog
      VSF_FOG_BEGIN,          //initialize fog computation
      VSF_FOG_END,            //finish fog computation, write to output reg
      VSF_FOG_END_HALVED,     //same as above, but with half fog intensity
      VSF_FOG_HEIGHT_OS,      //compute height fog, using current constant parameters (VSC_FOGPARAMS)
      VSF_FOG_HEIGHT_WS,      //'' in world space
      VSF_FOG_LAYERED_OS,     //compute layered fog in object space
      VSF_FOG_LAYERED_WS,     //'' in world space
      VSF_FOG_POINT_OS,       //compute point fog
      VSF_FOG_POINT_WS,       //compute point fog
      VSF_LIGHT_BEGIN,        //initialize lighting computation (diffuse)
      VSF_LIGHT_END,          //finish lighting computation, write to diffuse reg
      VSF_LIGHT_COPY,         //plain copy into diffuse color
      VSF_LIGHT_END_ALPHA,    //multiply current lighting temp value by vertex alpha component and write to diffuse
      VSF_SPECULAR_BEGIN,     //initialize lighting computation (specular)
      VSF_SPECULAR_END,       //finish lighting computation, write to specular reg
      VSF_LIGHT_DIR_OS,       //single directional light (object space)
      VSF_LIGHT_DIR_WS,       //''                       (world space)
      VSF_LIGHT_DIR_OS_DS,    //''                       (object space, double-sided)
      VSF_LIGHT_POINT_OS,     //single point light (object space)
      VSF_LIGHT_POINT_WS,     //''                 (world space)
      VSF_LIGHT_POINT_OS_DS,  //''                 (object space, double-sided)
      VSF_LIGHT_POINT_AMB_OS, //single point ambient light (object space)
      VSF_LIGHT_POINT_AMB_WS, //single point ambient light (world space)
      VSF_DIFFUSE_COPY,       //plain copy of vertex diffuse element
      VSF_DIFFUSE_MUL_ALPHA,  //copy vertex diffuse, modulate vertex alpha with source diffuse
      VSF_COPY_ALPHA,         //copy alpha from vertex to diffuse element
      VSF_SIMPLE_DIR_LIGHT,   //simple dp3 directional light (for debug drawing)
      VSF_LAST
   };

   dword vs_fragment_ids[VSF_LAST];
                              //VS constant registers
   enum E_VS_CONSTANT{
      VSC_CAM_LOC_POS,        //camera local position
      VSC_LIGHT_DIR_NORMAL,   //directional light, normal
      VSC_LIGHT_DIR_DIFFUSE,  //directional light, diffuse color
      VSC_LIGHT_POINT_POS_FR, //point light, position + far range
      VSC_LIGHT_POINT_DIFFUSE_RRD,  //point light, diffuse color + reciprocal of range delta
      VSC_LIGHT_LAYERED_PLANE,//layered light, plane
      VSC_LIGHT_LAYERED_COL_R,//layered light, color (xyz) and power (w)
      VSC_LIGHT_LAYERED_POS,  //layered light, position (xyz)
      VSC_LIGHT_PFOG_CONE_TOP,//pointfog light, cone top (xyz) and r_cone_cos_angle (w)
      VSC_LIGHT_PFOG_CONE_DIR,//pointfog light, cone dir (xyz) and multiplier (w)
      VSC_LIGHT_PFOG_COLOR,   //pointfog light, color (xyz) and power (w)
      VSC_TXT_CLIP_PLANE,     //texture clipping plane
      VSC_LAST
   };
   dword vs_constant_ids[VSC_LAST];
   enum E_VS_FIXED_CONSTANT{
      VSC_CONSTANTS = 0,      //constants: [0.0, 0.5, 1.0, 2.0]
      VSC_MAT_TRANSFORM_0 = 1,//transposed transformation matrix
      VSC_FOGPARAMS = 5,      //fog parameters: [fog_start, fog_end, 1.0/fog_range, 0.0]
      VSC_AMBIENT = 6,        //ambient light
      VSC_DEBUG = 7,          //value used for debugging
      VSC_CONSTANTS1 = 8,    //contants: [0.5, -0.5, ?, ?]
      VSC_MAT_TRANSFORM_1 = 9,//second matrix used for geometry blending
      VSC_MAT_TRANSFORM_2 = 13,//third matrix used for geometry blending
      VSC_UV_SHIFT = 9,       //uv shifting value
      VSC_MAT_BLEND_BASE = 9, //mase for matrix palette blending, each matrix consumes 3 slots
      VSC_MAT_BLEND_END = VSC_MAT_BLEND_BASE + MAX_VS_BLEND_MATRICES*3 - 1,
                              //last value, until this, constants are reserved (NVLink can't use them)
      VSC_FIXED_LAST          
   };


#define MAX_PS_FRAGMENTS 8

   enum E_PS_FRAGMENT{
      PSF_TEST,
      PSF_TEX_0,              //texture declarators for up to 4 textures
      PSF_TEX_01,
      PSF_TEX_012,
      PSF_TEX_0123,
      PSF_V0_COPY,            //simple copy of diffuse color
      PSF_T0_COPY,            //simple copy of texture 0
      PSF_MOD_STAGE0,         //modulation of texture and diffuse in stage 0
      PSF_LAST
   };
   dword ps_fragment_ids[PSF_LAST];
   enum E_PS_FIXED_CONSTANT{
      PSC_DEBUG = 7,          //value used for debugging
   };

                              //shader cache
   struct S_shader_entry_base{
      dword sort_value;       //value used for fas look-up
      inline bool operator <(const S_shader_entry_base &se) const{
         return (sort_value < se.sort_value);
      }
   };

                              //** vertex shader cache **
   struct S_vs_shader_entry_in: public S_shader_entry_base{
      const dword *shd_decl;  //pointer to shader declaration
      dword num_fragments;
      E_VS_FRAGMENT fragment_code[MAX_VS_FRAGMENTS];

      S_vs_shader_entry_in(const dword *sdecl):
         shd_decl(sdecl),
         num_fragments(0)
      {
                              //begin sort value with bitmask of declarator
         sort_value = (dword)shd_decl;
      }
      enum{ VSF_NUM_BITS = 5 };  //number of bits sufficiently holding all VSF_ values (except of VSF_LAST)

      inline void AddFragment(E_VS_FRAGMENT vsf){
         assert(num_fragments < MAX_VS_FRAGMENTS);
         sort_value ^= Rol(vsf, num_fragments*VSF_NUM_BITS);
         fragment_code[num_fragments++] = vsf;
      }
      inline void PopFragment(){
         assert(num_fragments);
         --num_fragments;
         sort_value ^= Rol(fragment_code[num_fragments], num_fragments*VSF_NUM_BITS);
      }
   };

   struct S_vs_shader_entry: public S_shader_entry_base{
      dword vs_handle;
      dword last_render_time; //value used for caching
                              //slots for constants:
      dword vscs_cam_loc_pos; //valid only if VSF_GENERATE_ENV_UV is present
      dword vscs_txt_clip;    //valid only if VSF_TXT_CLIP_??_? is present

      dword vscs_light_param[MAX_VS_LIGHTS * 3];   //reserve 2 constant slots per light

      S_vs_shader_entry(const S_shader_entry_base &se):
         vs_handle(0)
      {
         sort_value = se.sort_value;
      }
                              //assignment, not implmented, can't be used
      S_vs_shader_entry &operator =(const S_vs_shader_entry &se);
   };
                              //vertex-shader cache
   typedef set<S_vs_shader_entry> vs_set;
   vs_set vs_cache;

   /*
                              //released D3D textures, actual releasing posponed several frames
                              // due to common D3D/driver problems with state flushing
   struct S_D3D_texture_gone{
      C_smart_ptr<IDirect3DTexture8> d3d_tex;
      int frame_count;
      S_D3D_texture_gone(){}
      S_D3D_texture_gone(const S_D3D_texture_gone &r){ operator =(r); }
      S_D3D_texture_gone &operator =(const S_D3D_texture_gone &r){
         d3d_tex = r.d3d_tex;
         frame_count = r.frame_count;
         return *this;
      }
   };
   vector<S_D3D_texture_gone> gone_textures;
   */

//----------------------------
// Get vertex shader from provided fragment code. The input entry must have initialized
// 'shd_decl' and 'num_fragments'.
// The output entry contains all valid members.
// If the call fails, the returned value is NULL.
   const S_vs_shader_entry *GetVSHandle(const S_vs_shader_entry_in &se);


                              //** pixel shader cache **

   struct S_ps_shader_entry_in: public S_shader_entry_base{
      dword num_fragments;
      E_PS_FRAGMENT fragment_code[MAX_PS_FRAGMENTS];

      S_ps_shader_entry_in():
         num_fragments(0)
      {
         sort_value = 0;
      }
      enum{ PSF_NUM_BITS = 5 };  //number of bits sufficiently holding all PSF_ values (except of PSF_LAST)

      inline void AddFragment(E_PS_FRAGMENT psf){
         assert(num_fragments < MAX_PS_FRAGMENTS);
         sort_value ^= Rol(psf, num_fragments*PSF_NUM_BITS);
         fragment_code[num_fragments++] = psf;
      }
      inline void PopFragment(){
         assert(num_fragments);
         --num_fragments;
         sort_value ^= Rol(fragment_code[num_fragments], num_fragments*PSF_NUM_BITS);
      }
   };

   struct S_ps_shader_entry: public S_shader_entry_base{
      dword shd_handle;
      dword last_render_time; //value used for caching

      S_ps_shader_entry(const S_shader_entry_base &se):
         shd_handle(0)
      {
         sort_value = se.sort_value;
      }
                              //assignment, not implmented, can't be used
      S_ps_shader_entry &operator =(const S_ps_shader_entry &se);
   };
                              //vertex-shader cache
   typedef set<S_ps_shader_entry> ps_set;
   ps_set ps_cache;

//----------------------------
// Get shader from provided fragment code. The input entry must have initialized
// 'shd_decl' and 'num_fragments'.
// The output entry contains all valid members.
// If the call fails, the returned value is NULL.
   const S_ps_shader_entry *GetPSHandle(const S_ps_shader_entry_in &se);

//----------------------------
// Set pixel shader from currently set values (debugging).
   void SetLegacyPixelShader();


                              //callback
   dword cb_flags;
   PI3D_SCNCALLBACK cb_proc;
                              //statistics
                              //let other interfaces write to the stats
   enum{
      DRV_COL_STATS_S_S, DRV_COL_STATS_S_R, DRV_COL_STATS_S_B, DRV_COL_STATS_S_F,
      DRV_COL_STATS_L_S, DRV_COL_STATS_L_R, DRV_COL_STATS_L_B, DRV_COL_STATS_L_F,
      DRV_COL_STATS_LAST
   };
   dword col_stats[DRV_COL_STATS_LAST];

   I3D_stats_render render_stats;
   I3D_stats_scene scene_stats;
   I3D_stats_bsp bsp_stats;
   I3D_stats_voltree voltree_stats;

   dword other_memory_size;   //size of memory allocated by resources

   int force_lod_index;       //-1 = default
   float lod_scale;
   float shadow_range_n, shadow_range_f;

                              //multiple VB managers for various vertex formats
   vector<C_smart_ptr<C_RB_manager<IDirect3DVertexBuffer8> > > vb_manager_list;
   vector<C_smart_ptr<C_RB_manager<IDirect3DIndexBuffer8> > > ib_manager_list;
   C_smart_ptr<IDirect3DVertexBuffer8> vb_flare;   //initialized vertex buffer for flares

                              //program database
   C_database dbase;

private:
//----------------------------
                              //render targets
   C_smart_ptr<IDirect3DSurface8> curr_render_target, curr_z_sten;

                              //VS declarator cache, never released
   vector<vector<dword> *> vs_decl_cache;

   dword texture_sort_id;     //texture is, assigned to texture at creation, used for sorting
public:
   dword *RegisterVSDeclarator(const vector<dword> &decl);

                              //shadow rendering:
   C_smart_ptr<I3D_texture> tp_shadow[3]; //[rendertarget|blurred|range]
   C_smart_ptr<IDirect3DSurface8> zb_shadow;
   C_smart_ptr<IDirect3DIndexBuffer8> ib_shadow;
   int ib_shadow_base_index;
   C_smart_ptr<I3D_texture> tp_dyn_light_circle;   //dynamic light's circle
                              //bluring rectangles
   C_smart_ptr<IDirect3DVertexBuffer8> vb_blur;
   C_smart_ptr<IDirect3DIndexBuffer8> ib_blur;
   struct S_blur_vertex{
      S_vectorw v;
      dword diffuse;
      I3D_text_coor uv;
   };
   enum{ NUM_SHADOW_BLUR_SAMPLES = 9 };

                              //clipping by texture:
   C_smart_ptr<I3D_texture> tp_clip;   //alpha texture
   int clip_txt_stage;
   S_plane pl_clip;


   void SetRenderTarget(IDirect3DSurface8 *rt, IDirect3DSurface8 *z_sten){
                              //setup new values
      HRESULT hr;
      hr = lpDev8->SetRenderTarget(curr_render_target = rt, curr_z_sten = z_sten);
      CHECK_D3D_RESULT("SetRenderTarget", hr);
      drv_flags &= ~DRVF_HASSTENCIL;
      if(curr_z_sten){
         D3DSURFACE_DESC desc;
         curr_z_sten->GetDesc(&desc);
         switch(desc.Format){
         case D3DFMT_D15S1: case D3DFMT_D24S8: case D3DFMT_D24X4S4: drv_flags |= DRVF_HASSTENCIL; break;
         }
      }
   }

   void SaveRenderTarget(C_smart_ptr<IDirect3DSurface8> &save_rt, C_smart_ptr<IDirect3DSurface8> &save_z_sten) const{
      save_rt = curr_render_target;
      save_z_sten = curr_z_sten;
   }
private:
   dword render_view_id;

public:
   inline dword GetRenderviewID(){
      if(!render_view_id) ++render_view_id;
      return render_view_id++;
   }

//----------------------------
// Update D3D viewport.
   void UpdateViewport(int l, int t, int r, int b);

private:
   void RefreshAfterReset();

//----------------------------
   bool specular_enable;
                              //fog

   bool fog_enable;               //fog in 3D API (DirectX, OpenGL) enable flag
   dword d3d_fog_color;
public:
   inline void EnableFog(bool on){
      if(fog_enable != on){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_FOGENABLE, fog_enable = on);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }
   inline bool IsFogEnabled() const{ return fog_enable; }

   inline void SetFogColor(dword color){
      if(d3d_fog_color!=color){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_FOGCOLOR, d3d_fog_color = color);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }
   inline dword GetFogColor() const{ return d3d_fog_color; }

   inline void EnableSpecular(bool on){
      if(specular_enable!=on){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_SPECULARENABLE, specular_enable = on);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

private:
//----------------------------
   D3DBLEND blend_table[2][14];    //[src/dest][D3DBLEND]
                              //blending cache
   dword srcblend, dstblend;
#define ALPHAMODE_ENABLED     0x10000
#define ALPHAMODE_TEXTURE     1
#define ALPHAMODE_FACTOR      2
#define ALPHAMODE_VERTALPHA   4
   dword alpha_mode;
public:
   inline dword GetAlphaMode() const{ return alpha_mode; }
   inline dword GetSrcBlend() const{ return srcblend; }
   inline dword GetDstBlend() const{ return dstblend; }
private:
   dword stage0_alpha_arg2;

   vector<S_pixelformat> texture_formats[2]; //[texture|RTT texture]
   vector<vector<D3DFORMAT> > rtt_zbuf_fmts; //depth surfaces usable for each rendertarget

//----------------------------
// Initialize D3D and all associated resources.
   void InitD3DResources();

//----------------------------
   void CloseD3DResources();

//----------------------------
// Initialize light-map pixel format convertor.
   void InitLMConvertor();

//----------------------------
// Init all resources associated with real-time shadow rendering.
   void InitShadowResources();

                              //internal resource lists
   vector<PI3D_texture> managed_textures;
   vector<PI3D_sound> sounds;

                              //list of hardware vertex buffers and their references
   struct S_VB_use_ref{
      IDirect3DVertexBuffer8 *vb;
      int vertex_index;
      IDirect3DVertexBuffer8 **ref;
   };
   vector<S_VB_use_ref> hw_vb_list;

//----------------------------
                              //directories
   vector<C_string> dirs[I3DDIR_LAST];

public:
   const vector<C_string> &GetDirs(I3D_DIRECTORY_TYPE dt) const{ return dirs[dt]; }
   dword last_dir_index[I3DDIR_LAST];     //index of last directory

private:
//----------------------------
                              //debugging stats
   dword interface_count[I3D_CLID_LAST];

   friend I3D_scene;
   friend I3D_sound;
   friend I3D_light;

   C_string dll_filename;

   I3D_driver(const char *dll_filename);
   friend I3D_RESULT I3DAPI __stdcall I3DCreate(I3D_driver**, const char*);

public:
                              //callback used when device is reset
   enum E_RESET_MSG{
      RESET_RELEASE,          //notification before device reset (all resources must be released now)
      RESET_RECREATE,         //notification after device reset (resources may be recreated)
   };
   typedef void t_ResetCallback(E_RESET_MSG, void *context);

   void RegResetCallback(t_ResetCallback *cb, void *context){
      reset_callbacks.push_back(pair<t_ResetCallback*, void*>(cb, context));
   }
   void UnregResetCallback(t_ResetCallback *cb, void *context){
      pair<t_ResetCallback*, void*> p(cb, context);
      for(int i=reset_callbacks.size(); i--; ){
         if(reset_callbacks[i]==p){
            reset_callbacks[i] = reset_callbacks.back(); reset_callbacks.pop_back();
            break;
         }
      }
   }
private:
                              //all registered callbacks
   vector<pair<t_ResetCallback*, void*> > reset_callbacks;

public:
   ~I3D_driver();
//----------------------------
                              //cache states
   dword last_blend_flags;
   dword last_vs_handle;      //last vertex shader handle
   dword last_texture_factor;
   dword last_alpha_ref;
   bool alpharef_enable;
   float last_bump_scale[MAX_TEXTURE_STAGES];
   dword last_txt_coord_index[MAX_TEXTURE_STAGES];
   D3DTEXTUREOP stage_txt_op[MAX_TEXTURE_STAGES], stage_alpha_op[MAX_TEXTURE_STAGES];

   bool stage_anisotropy[MAX_TEXTURE_STAGES];
   D3DTEXTUREFILTERTYPE stage_minf[MAX_TEXTURE_STAGES], min_filter[2];    //[normal|aniso]

   PI3D_texture_base last_texture[MAX_TEXTURE_STAGES];   //multi-texturing
   bool d3d_clipping;
                              //stream0 cache:
   IDirect3DVertexBuffer8 *last_stream0_vb;
   dword last_stream0_stride;

   inline void SetStreamSource(IDirect3DVertexBuffer8 *vb, dword vs){
      if(last_stream0_vb!=vb || last_stream0_stride!=vs){
         HRESULT hr;
         hr = lpDev8->SetStreamSource(0, last_stream0_vb=vb, last_stream0_stride=vs);
         CHECK_D3D_RESULT("SetStreamSource", hr);
      }
   }
   inline void ResetStreamSource(){ last_stream0_vb = 0; }

   inline void EnableAnisotropy(int stage, bool b){
      if(stage_anisotropy[stage]!=b){
         stage_anisotropy[stage] = b;
         D3DTEXTUREFILTERTYPE minf = min_filter[b];
         if(stage_minf[stage]!=minf){
            HRESULT hr;
            hr = lpDev8->SetTextureStageState(stage, D3DTSS_MINFILTER, stage_minf[stage] = minf);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
         }
      }
   }

private:
   bool b_cullnone;           //true when no culling
   bool zw_enable;            //true when z-writing is enabled
   D3DCULL cull_default;      //default culling order
   //set<int> used_palettes;    //palette number allocation manager

public:
                              //image used for light-map construction - cached here
   C_string map_channel_2_img_name;
   C_smart_ptr<IImage> map_channel_2_img;
   dword map_channel_2_time[2];

                              //vertex lighting mode
   D3DTEXTUREOP vertex_light_blend_op;    //MODULATE or MODULATE2X, depending on caps
   float vertex_light_mult;               //multiplier for vertex lighting

//----------------------------
// Set alpha reference value.
   inline void I3D_driver::SetAlphaRef(dword ar){

      if(last_alpha_ref!=ar){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_ALPHAREF, last_alpha_ref = ar);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------
// Enable/disable alpha testing.
   inline void I3D_driver::SetupAlphaTest(bool on){

      if(alpharef_enable!=on){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_ALPHATESTENABLE, alpharef_enable = on);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

   inline void SetTexture1(dword i, PI3D_texture_base tb){
      LPDIRECT3DTEXTURE8 new_t = tb ? tb->GetD3DTexture() : NULL;
      if(last_texture[i]!=tb){
         last_texture[i] = tb;
         HRESULT hr;
         hr = lpDev8->SetTexture(i, new_t);
         CHECK_D3D_RESULT("SetTexture", hr);
      }
   }

//----------------------------

   inline void SetDefaultCullMode(D3DCULL cm){

      cull_default = cm;
      if(!b_cullnone){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_CULLMODE, cull_default);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------

   inline void EnableNoCull(bool b){

      if(b_cullnone != b){
         b_cullnone = b;
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_CULLMODE, b_cullnone ? D3DCULL_NONE : cull_default);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------

   inline bool IsCullNone() const{ return b_cullnone; }

//----------------------------

   inline void EnableZWrite(bool b){

      if(zw_enable!=b){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_ZWRITEENABLE, zw_enable = b);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------

   inline bool IsZWriteEnabled() const{ return zw_enable; }

//----------------------------

   inline void SetRenderMat(CPI3D_material mat, int txt_stage = 0){

      PI3D_texture_base tb = mat->GetTexture1(MTI_DIFFUSE);
      SetTexture1(txt_stage, tb);

                              //setup culling
      EnableNoCull(mat->Is2Sided1());
                              //setup alpha ref for translucent materials
      if(mat->IsCkeyAlpha1() && tb)
         SetAlphaRef(mat->GetAlphaRef());
   }

//----------------------------

   inline void SetTextureFactor(dword tf){

      if(last_texture_factor != tf){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_TEXTUREFACTOR, last_texture_factor = tf);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------

   inline void SetBumpScale(int stage, float bs){

      if(last_bump_scale[stage] != bs){
         last_bump_scale[stage] = bs;
         HRESULT hr;
         hr = lpDev8->SetTextureStageState(stage, D3DTSS_BUMPENVMAT00, FloatAsInt(bs));
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         hr = lpDev8->SetTextureStageState(stage, D3DTSS_BUMPENVMAT11, FloatAsInt(bs));
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
   }

//----------------------------

   inline void SetTextureCoordIndex(int stage, int index){

      if(last_txt_coord_index[stage] != index){
         assert(!IsDirectTransform());
         HRESULT hr;
         hr = lpDev8->SetTextureStageState(stage, D3DTSS_TEXCOORDINDEX,
            D3DTSS_TCI_PASSTHRU | (last_txt_coord_index[stage]=index));
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
   }

//----------------------------

   inline int NumSimultaneousTextures() const{
      return Min(d3d_caps8.MaxTextureBlendStages, d3d_caps8.MaxSimultaneousTextures);
   }

//----------------------------
// Allocate new palette number. This number is passed do D3D for palette identification.
   //int AllocatePaletteNumber();

//----------------------------
// Free previously allocated palette.
   //void FreePaletteNumber(int);

//----------------------------

   inline void SetVertexShader(dword handle){

      if(last_vs_handle != handle){
         HRESULT hr;
         hr = lpDev8->SetVertexShader(last_vs_handle = handle);
         CHECK_D3D_RESULT("SetVertexShader", hr);
      }
   }

//----------------------------

   inline dword GetFlags() const{ return drv_flags; }
   inline dword GetFlags2() const{ return drv_flags2; }

//----------------------------

   inline void SetupTextureStage(int stage, D3DTEXTUREOP op){

      if(stage_txt_op[stage] != op){
         stage_txt_op[stage] = op;
         HRESULT hr;
         hr = lpDev8->SetTextureStageState(stage, D3DTSS_COLOROP, op);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
      }
   }

//----------------------------

   inline void DisableTextureStage(int stage){

      assert(stage>0);
      if(stage_txt_op[stage] != D3DTOP_DISABLE){
         HRESULT hr;
         hr = lpDev8->SetTextureStageState(stage, D3DTSS_COLOROP, stage_txt_op[stage] = D3DTOP_DISABLE);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         SetTexture1(stage, NULL);
                              //reset also coord index
         SetTextureCoordIndex(stage, stage);
                              //disable all other enabled stages, some drivers have problems
                              // with this
         while(++stage < MAX_TEXTURE_STAGES){
            if(stage_txt_op[stage] == D3DTOP_DISABLE)
               break;
            hr = lpDev8->SetTextureStageState(stage, D3DTSS_COLOROP, stage_txt_op[stage] = D3DTOP_DISABLE);
            CHECK_D3D_RESULT("SetTextureStageState", hr);
            SetTexture1(stage, NULL);
            SetTextureCoordIndex(stage, stage);
         }
      }
   }

//----------------------------
// Clipping planes - using either D3D clipping plane, or emulated clipping by alpha texture.
   void SetClippingPlane(const S_plane &pl, const S_matrix &m_view_proj_hom);
   void DisableClippingPlane();

   inline dword GetTextureSortID(){
      return texture_sort_id++;
   }

//----------------------------
   
   inline void SetClipping(bool b){

      if(d3d_clipping != b){
         HRESULT hr;
         hr = lpDev8->SetRenderState(D3DRS_CLIPPING, d3d_clipping = b);
         CHECK_D3D_RESULT("SetRenderState", hr);
      }
   }

//----------------------------
   //inline bool IsClippingOn() const{ return d3d_clipping; }
   
   inline dword GetRenderTime() const{ return render_time; }
   
                              //debugging
   int debug_int[8];
   float debug_float[8];

   void PRINT(const char *cp, bool still=false) const{ if(cb_proc) cb_proc(CBM_MESSAGE, (dword)cp, still, NULL); }
   void PRINT(float f, bool still=false) const{ PRINT(C_fstring("%.2f", f), still); }
   void PRINT(int i, bool still=false) const{ PRINT(C_fstring("%i", i), still); }
   void PRINT(const S_vector &v, bool still=false) const{ PRINT(C_fstring("[%.2f, %.2f, %.2f]", v.x, v.y, v.z), still); }

   void DEBUG(const char *cp, int type = 0) const{ if(cb_proc) cb_proc(CBM_DEBUG_MESSAGE, (dword)cp, type, NULL); }
   void DEBUG(float f, bool persist = false) const{ DEBUG(C_fstring("%.2f", f), persist); }
   void DEBUG(int i, bool persist = false) const{ DEBUG(C_fstring("%i", i), persist); }
   void DEBUG(const S_vector &v, bool persist = false) const{ DEBUG(C_fstring("[%.2f, %.2f, %.2f]", v.x, v.y, v.z), persist); }

   void DEBUG_CODE(dword code, dword data) const{ if(cb_proc) cb_proc(CBM_DEBUG_CODE, code, data, NULL); }
private:
//----------------------------
                              //light-mapping

                              //keep all LMap textures here for further allocation,
                              // no AddRef kept on these
   vector<C_LM_surface*> lm_textures;
   int max_lm_memory;         //max bytes for LM textures
public:
   int lm_create_aa_ratio;    //LM creation - antialias sub-pixel divider
   C_LM_alloc_block *LMAlloc(vector<I3D_rectangle>*);
   void LMFree(C_LM_alloc_block*);
   bool UnregLMTexture(PI3D_texture);
   void DrawLMaps(int viewport[4]);

                              //texture formats
   S_pixelformat pf_light_maps;
   class C_rgb_conversion *rgbconv_lm;

   void PurgeLMTextures();

//----------------------------

   dword GraphCallback(dword msg, dword par1, dword par2);

   static GRAPHCALLBACK GraphCallback_thunk;
   
   void RegisterManagedTexture(PI3D_texture);
   void UnregisterManagedTexture(PI3D_texture);
   const vector<PI3D_texture> &GetManagedTextures() const{ return managed_textures; }

//----------------------------
                              //creation

   I3D_RESULT CreateTexture1(const I3D_CREATETEXTURE*, PI3D_texture*, PI3D_LOAD_CB_PROC, void *cb_context);

//----------------------------
   inline IDirect3DDevice8 *GetDevice1() const{ return lpDev8; }

                              //resource buffers
   bool AllocVertexBuffer(dword fvf_flags, D3DPOOL mem_pool, dword d3d_usage, int num_verts,
      IDirect3DVertexBuffer8 **vb_ret, int *beg_indx_ret, bool allow_cache = true);
   bool FreeVertexBuffer(IDirect3DVertexBuffer8 *vb, int vertex_indx);

   bool AllocIndexBuffer(D3DPOOL mem_pool, dword d3d_usage, int num_faces,
      IDirect3DIndexBuffer8 **ib_ret, int *beg_face_ret);
   bool FreeIndexBuffer(IDirect3DIndexBuffer8 *vb, int face_indx);

//----------------------------
                              //debug interface counting
   void AddCount(I3D_CLID id){ ++interface_count[id]; }
   void DecCount(I3D_CLID id){ --interface_count[id]; }

   void SetupBlend(dword flags);
//----------------------------
                              //procedurals:
private:

   vector<class I3D_procedural_base*> procedural_cache;
   void UnregisterProcedural(I3D_procedural_base*);
   friend I3D_procedural_base;
public:
//----------------------------
// Get/create procedural from file definition.
   I3D_procedural_base *GetProcedural(const char *filename,
      const struct S_load_context&, const char *frame_name, const char *material_name);

//----------------------------
                              //editing support
private:
   vector<PI3D_material> icon_materials;
public:
   PI3D_material GetIcon(int res_id);

//----------------------------
// Get valid format of Zbuffer/Stencil surface for current device and specified render target format.
// Returns D3DFMT_UNKNOWN is no such format exists.
   D3DFORMAT FindZStenFormat(const S_pixelformat*) const;


   typedef map<int, S_sound_env_properties> t_env_properties;
private:
   t_env_properties env_properties;

public:
   inline const t_env_properties &GetEnvProperties() const{ return env_properties; }
   const S_sound_env_properties *GetEnvProperty(int id){
      t_env_properties::const_iterator it = env_properties.find(id);
      return (it!=env_properties.end()) ? &(*it).second : NULL;
   }
   inline void MakeEnvSettingDirty(){ drv_flags |= DRVF_SOUND_ENV_RESET; }

                              //exported functions
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

                              //--- initialization ---
   I3DMETHOD(Init)(CPI3DINIT);
   I3DMETHOD(Close)();
   I3DMETHOD_(bool,IsInit)() const{ return (lpD3D8!=NULL); }
   I3DMETHOD(GetStats)(I3DSTATSINDEX, void *data);

                              //--- directories ---
   I3DMETHOD(AddDir)(I3D_DIRECTORY_TYPE dt, const char *cp){
      if(dt >= I3DDIR_LAST) return I3DERR_INVALIDPARAMS;
      dirs[dt].push_back(cp);
      return I3D_OK;
   }

   I3DMETHOD(DeleteDir)(I3D_DIRECTORY_TYPE dt, const char *cp){
      if(dt >= I3DDIR_LAST) return I3DERR_INVALIDPARAMS;
      int i = VectorIndex(dirs[dt], C_string(cp));
      if(i==-1) return I3DERR_OBJECTNOTFOUND;
      dirs[dt].erase(dirs[dt].begin()+i);
      return I3D_OK;
   }

   I3DMETHOD(ClearDirs)(I3D_DIRECTORY_TYPE dt){
      if(dt >= I3DDIR_LAST) return I3DERR_INVALIDPARAMS;
      dirs[dt].clear();
      return I3D_OK;
   }

   I3DMETHOD_(int,NumDirs)(I3D_DIRECTORY_TYPE dt) const{
      if(dt >= I3DDIR_LAST) return -1;
      return dirs[dt].size();
   }

   I3DMETHOD_(const char*,GetDir)(I3D_DIRECTORY_TYPE dt, int indx) const{
      if(dt >= I3DDIR_LAST) return NULL;
      if(indx<0 || indx>=dirs[dt].size())
         return NULL;
      return dirs[dt][indx];
   }

                              //--- interfaces ---
   I3DMETHOD_(class ISND_driver*,GetSoundInterface)() const{ return init_data.lp_sound; }
   I3DMETHOD_(class IGraph*,GetGraphInterface)() const{ return init_data.lp_graph; }

                              //--- Render state ---
   I3DMETHOD(SetState)(I3D_RENDERSTATE, dword value);
   I3DMETHOD_(dword,GetState)(I3D_RENDERSTATE) const;
   I3DMETHOD(FindTextureFormat)(S_pixelformat&, dword flags) const;
   I3DMETHOD(SetTexture)(PI3D_texture);
                              //rendering
   I3DMETHOD(SetViewport)(dword l, dword t, dword r, dword b);
   I3DMETHOD(BeginScene)();
   I3DMETHOD(EndScene)();
   I3DMETHOD(Clear)(dword color = 0);
                              //--- Creation - frames ---
   I3DMETHOD_(PI3D_scene,CreateScene)();
   I3DMETHOD_(PI3D_user,CreateUserFrame)();
   I3DMETHOD_(PI3D_dummy,CreateDummy)();
   I3DMETHOD_(PI3D_target,CreateTarget)();
   I3DMETHOD_(PI3D_visual,CreateVisual)(dword visual_type);
   I3DMETHOD_(PI3D_model,CreateModel)();
   I3DMETHOD_(PI3D_light,CreateLight)();
   I3DMETHOD_(PI3D_camera,CreateCamera)();
   I3DMETHOD_(PI3D_volume,CreateVolume)();
   I3DMETHOD_(PI3D_joint,CreateJoint)();
   I3DMETHOD_(PI3D_frame,CreateFrame)(dword frm_type, dword sub_type = 0);
                              //--- Creation - other types ---
   I3DMETHOD(CreateTexture)(const I3D_CREATETEXTURE *ct, PI3D_texture *tp1){
      return CreateTexture1(ct, tp1, NULL, NULL);
   }
   I3DMETHOD_(PI3D_material,CreateMaterial)();
   I3DMETHOD_(PI3D_mesh_level,CreateMeshLevel)();
   I3DMETHOD_(PI3D_mesh_object,CreateMesh)();
   I3DMETHOD_(PI3D_animation_base,CreateAnimation)(I3D_ANIMATION_TYPE);
   I3DMETHOD_(PI3D_animation_set,CreateAnimationSet)();
   I3DMETHOD_(PI3D_interpolator,CreateInterpolator)();
                              //visual types
   I3DMETHOD(EnumVisualTypes)(I3D_ENUMVISUALPROC*, dword user) const;

   I3DMETHOD_(PI3D_SCNCALLBACK,SetCallback)(PI3D_SCNCALLBACK cb, dword cbf);
   I3DMETHOD(SetDataBase)(const char *db_filename, dword size, int max_keep_days);
   I3DMETHOD(FlushDataBase)();

   I3DMETHOD_(dword,NumProperties)(dword visual_type) const;
   I3DMETHOD_(const char*,GetPropertyName)(dword visual_type, dword index) const;
   I3DMETHOD_(I3D_PROPERTYTYPE,GetPropertyType)(dword visual_type, dword index) const;
   I3DMETHOD_(bool, CollisionResponse)(I3D_collision_data&) const;
   I3DMETHOD_(void,DisableLeakWarning)() const;

                              //sound properties
   I3DMETHOD(SetSoundEnvData)(int set_id, I3D_SOUND_ENV_DATA_INDEX, dword data);
   I3DMETHOD_(dword, GetSoundEnvData)(int set_id, I3D_SOUND_ENV_DATA_INDEX) const;

                              //internal:
   I3DMETHOD_(void,Dummy)(){}
   I3DMETHOD(GetVersion)(dword&,dword&);
};
//----------------------------

#endif
