/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.
   All rights reserved.

   File: billbrd.cpp
   Content: I3D_object_billboard - billboard frame, frame
      inherited from I3D_object, making billboard effect
      by 'looking' always towards active camera.
--------------------------------------------------------*/

#include "all.h"
#include "flare.h"
#include "scene.h"
#include "camera.h"
#include "object.h"

//#define DEBUG_NO_COLLISIONS   //don't test collisions

//----------------------------

#define CHECK_COL_TIME 400


//----------------------------

static struct S_auto_init_flare{
   static I3D_visual *Create(PI3D_driver drv){
      return new I3D_flare(drv);
   }
   S_auto_init_flare(){
      RegisterVisual(Create, I3D_VISUAL_FLARE, "Flare");
   }
} auto_init_flare;

//----------------------------
//----------------------------

I3D_flare::I3D_flare(PI3D_driver d):
   I3D_visual(d),
   scale_begin_dist(5.0f),
   rotation_ratio(.02f),
   fade_normal_angle(0.0f),
   occlusion_fade_time(250),
   last_check_col_time(drv->GetRenderTime() - CHECK_COL_TIME - S_int_random(CHECK_COL_TIME)),
   occlusion_fade_count(-1),
   pivot(0.0f)
{
   visual_type = I3D_VISUAL_FLARE;
}

//----------------------------

I3D_RESULT I3D_flare::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);
   PI3D_visual vis = I3DCAST_VISUAL(frm);
   switch(vis->GetVisualType()){
   case I3D_VISUAL_FLARE:
      {
         PI3D_flare flr = (PI3D_flare)vis;
         SetMaterial(flr->GetMaterial());
         SetScaleDisance(flr->GetScaleDisance());
         SetRotationRatio(flr->GetRotationRatio());
         SetNormalAngle(flr->GetNormalAngle());
         SetOcclusionFade(flr->GetOcclusionFade());
      }
      break;
   default:
      {
         I3D_mesh_base *mb = vis->GetMesh();
         if(mb){
            SetMaterial(mb->GetFGroups1()->mat);
         }else{
            PI3D_material mat = drv->CreateMaterial();
            SetMaterial(mat);
            mat->Release();
         }
      }
   }
   return I3D_visual::Duplicate(vis);
}

//----------------------------

const float FLARE_BSPHERE_RADIUS = .70711f;

bool I3D_flare::ComputeBounds(){

   bound.bound_local.bbox.min = S_vector(-FLARE_BSPHERE_RADIUS*.5f, -FLARE_BSPHERE_RADIUS*.5f, -FLARE_BSPHERE_RADIUS*.5f);
   bound.bound_local.bbox.max = S_vector( FLARE_BSPHERE_RADIUS*.5f,  FLARE_BSPHERE_RADIUS*.5f,  FLARE_BSPHERE_RADIUS*.5f);
   bound.bound_local.bsphere.pos.Zero();
   bound.bound_local.bsphere.radius = FLARE_BSPHERE_RADIUS*.5f;

   return I3D_visual::ComputeBounds();
}

//----------------------------

void I3D_flare::AddPrimitives(S_preprocess_context &pc){

   if(pc.mode!=RV_NORMAL)
      return;

   if(!mat)
      return;
   if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES))
      return;

   PI3D_camera cam = pc.scene->GetActiveCamera1();
   const S_matrix &m_camera = cam->GetMatrix();

   S_vector dir_to_cam = m_camera(3) - matrix(3);
                              //check distance to camera
   float dist_to_cam = dir_to_cam.Magnitude();
   float dist_to_near = -m_camera(2).Dot(dir_to_cam) - cam->GetNCP();

   if(dist_to_near <= pivot)
      return;

   float alpha = mat->GetAlpha();

   if((dword&)fade_normal_angle){
                              //check angle to direction normal
      float angle = (m_camera(3) - matrix(3)).AngleTo(matrix(2));
      if(angle >= fade_normal_angle){
         occlusion_fade_count = 0;
         return;
      }
      alpha *= 1.0f - angle / fade_normal_angle;
   }

#ifndef DEBUG_NO_COLLISIONS
   if(occlusion_fade_time){
                              //check collisions now
      int cct_delta = drv->GetRenderTime() - last_check_col_time;
      if(cct_delta > CHECK_COL_TIME){
         last_check_col_time = drv->GetRenderTime();

         struct S_hlp{
            static bool I3DAPI cbResp2(I3D_collision_data &cd){
                              //don't detect with specially marked collisions
               assert(cd.GetHitFrm());
               return !(cd.GetHitFrm()->GetFrameFlags()&I3D_FRMF_NO_FLARE_COLLISION);
            }
         };

         I3D_collision_data cd(matrix(3), dir_to_cam, I3DCOL_FORCE2SIDE, NULL, S_hlp::cbResp2);
         bool col = pc.scene->TestCollision(cd);        
         if(col){
            flare_flags |= FLRF_COLLIDING;
            if(occlusion_fade_count == -1)
               occlusion_fade_count = occlusion_fade_time;
         }else{
            flare_flags &= ~FLRF_COLLIDING;
            if(!occlusion_fade_count)
               ++occlusion_fade_count;
         }
      }

                              //process occlusion fading
      int render_time_delta = drv->GetRenderTime() - last_render_time;
   
      if(!(flare_flags&FLRF_COLLIDING)){
         if(occlusion_fade_count != -1){
                                 //fade on
            occlusion_fade_count += render_time_delta;
            if(occlusion_fade_count >= occlusion_fade_time)
               occlusion_fade_count = -1;
            else
               alpha *= (float)occlusion_fade_count / (float)occlusion_fade_time;
         }
      }else{
                                 //fade off
         occlusion_fade_count = Max(0, occlusion_fade_count - render_time_delta);
         if(!occlusion_fade_count)
            return;
         alpha *= (float)occlusion_fade_count / (float)occlusion_fade_time;
      }
   }
#endif//!DEBUG_NO_COLLISIONS

                              //add to sort list
   pc.prim_list.push_back(S_render_primitive(0));
   S_render_primitive &p = pc.prim_list.back();
   p.vis = this;
   p.clip = pc.clip;
   p.sector = pc.curr_sector;
   p.scene = pc.scene;
                              //determine blending mode
   p.blend_mode = I3DBLEND_VERTEXALPHA;
   if(occlusion_fade_time){
      p.sort_value = PRIM_SORT_ALPHA_NOZ;
      ++pc.alpha_noz;
   }else{
      p.sort_value = PRIM_SORT_ALPHA_NOZWRITE;
      ++pc.alpha_nozwrite;
   }

   bool is_self_alpha = false;
   if(!mat->IsTransl()){
      p.blend_mode |= I3DBLEND_ALPHABLEND;
   }else
   if(mat->IsCkeyAlpha1()){
      p.blend_mode |= I3DBLEND_ALPHABLEND;
   }else{
      PI3D_texture_base tb = mat->GetTexture1();
      if(tb && (tb->GetTxtFlags()&TXTF_SELF_ALPHA)){
         p.blend_mode |= I3DBLEND_ADD;
         is_self_alpha = true;
      }else
         p.blend_mode |= I3DBLEND_ALPHABLEND;
   }
   drv->render_stats.vert_trans += NUM_FLARE_VERTS;

                              //create user context, to be passed to DrawPrimitive
   S_user_context *uc = new S_user_context;
   p.user = (dword)uc;
   uc->g_matrix(0, 3) = 0.0f;
   uc->g_matrix(1, 3) = 0.0f;
   uc->g_matrix(2, 3) = 0.0f;
   uc->g_matrix(3, 3) = 1.0f;
                              //scale alpha by distance
   if((dist_to_near - pivot) < scale_begin_dist){
      alpha *= (dist_to_near - pivot) / scale_begin_dist;
   }

   {                          //compute matrix
      float scale = matrix(0).Magnitude();
      if(dist_to_near < scale_begin_dist){
         scale *= .5f + (.5f * dist_to_near / scale_begin_dist);
      }
      scale *= (dist_to_near - pivot) / dist_to_near;

      const S_matrix *mp = &m_camera;
      if((dword&)rotation_ratio){
                              //apply rotation by Z
         S_matrix m_tmp;
         m_tmp.RotationZ(dist_to_near * rotation_ratio * 2.0f * PI);
         m_tmp = m_tmp % m_camera;
         mp = &m_tmp;
      }

      uc->g_matrix(0) = (*mp)(0) * scale;
      uc->g_matrix(1) = (*mp)(1) * scale;
      uc->g_matrix(2) = (*mp)(2) * scale;
      uc->g_matrix(3) = matrix(3);
      if(!occlusion_fade_time){
                              //apply pivot
         uc->g_matrix(3) += dir_to_cam * (1.0f / dist_to_cam) * pivot;
      }
   }

                              //compute dest color,
                              // which will be assigned during the rendering
   uc->diffuse = mat->GetDiffuse();
   if(is_self_alpha)
      uc->diffuse *= alpha;
   uc->diffuse.w = alpha;
}

//----------------------------

static const dword vs_decl[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT2),   //position
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

void I3D_flare::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   S_user_context *uc = (S_user_context*)rp.user;
   assert(uc);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

   dword new_blend = rp.blend_mode;
   if(drv->last_blend_flags!=new_blend)
      drv->SetupBlend(new_blend);
   drv->SetupTextureStage(0, D3DTOP_MODULATE);

   I3D_driver::S_vs_shader_entry_in se(vs_decl);
   se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
   PrepareVertexShader(NULL, se, rp, VSPREP_TRANSFORM | VSPREP_FEED_MATRIX | VSPREP_COPY_UV, &uc->g_matrix);

   drv->SetClipping(rp.clip);
   drv->DisableTextureStage(1);
   //drv->SetRenderMat(mat);
   drv->SetTexture1(0, mat->GetTexture1(MTI_DIFFUSE));
   drv->EnableNoCull(true);   //use no culling due to mirrors

   drv->EnableFog(vis_flags&VISF_FOG_VALID);

                              //setup flare's color
   HRESULT hr;
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &uc->diffuse, 1);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

   drv->SetStreamSource(drv->vb_flare, sizeof(S_vertex_flare));

   hr = lpDev8->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
   CHECK_D3D_RESULT("DrawPrimitive", hr);

   delete uc;
}

//----------------------------
//----------------------------


