/*--------------------------------------------------------
   Copyright (c) 1999, 2000 Michal Bacik. 
   All rights reserved.

   File: Joint.cpp
   Content: Joint frame.
--------------------------------------------------------*/

#include "all.h"
#include "joint.h"
#include "sng_mesh.h"
#include "scene.h"


//----------------------------

I3D_joint::I3D_joint(PI3D_driver d):
   I3D_frame(d),
   icon_size(.5f),
   joint_flags(0)
{
   bbox.min = S_vector(-.1f, -.1f, -.1f);
   bbox.max = S_vector( .1f,  .1f,  .1f);

   drv->AddCount(I3D_CLID_JOINT);
   type = FRAME_JOINT;
   enum_mask = ENUMF_JOINT;
}

//----------------------------

I3D_joint::~I3D_joint(){
   drv->DecCount(I3D_CLID_JOINT);
}

//----------------------------

I3D_RESULT I3D_joint::Duplicate(CPI3D_frame frm){

   if(frm->GetType1()==FRAME_JOINT){
      PI3D_joint bp = I3DCAST_JOINT(frm);

      joint_flags = bp->joint_flags;
      bbox_matrix = bp->bbox_matrix;
      bbox = bp->bbox;
      volume_box_matrix = bp->volume_box_matrix;
      volume_box = bp->volume_box;
      icon_size = bp->icon_size;
   }
   return I3D_frame::Duplicate(frm);
}

//----------------------------

I3D_RESULT I3D_joint::SetPos(const S_vector &v){

   if(joint_flags&JOINTF_FIXED_POSITION)
      return I3DERR_UNSUPPORTED;
   return I3D_frame::SetPos(v);
}

//----------------------------

void I3D_joint::SetBBox(const I3D_bbox &bb1, const S_matrix &m1){

   bbox = bb1;
   bbox_matrix = m1;
   joint_flags |= JOINTF_BBOX_VALID;
   icon_size = (bbox.max-bbox.min).Magnitude() * bbox_matrix.GetScale().Sum() * .4f;
}

//----------------------------

static const S_vector color(1.0f, .7f, 0.0f);
static const S_vector volume_color(.7f, .3f, .3f);
static const word sides_i[] = {
   1, 3, 7,  7, 5, 1,
   4, 6, 2,  2, 0, 4,
   2, 6, 7,  7, 3, 2,
   4, 0, 1,  1, 5, 4,

   0, 2, 3,  3, 1, 0,
   7, 6, 5,  6, 4, 5,
};

//----------------------------

void I3D_joint::Draw1(PI3D_scene scene, bool strong) const{

   byte alpha = strong ? 0x00 : 0xd0;

   float size = icon_size * matrix(0).Magnitude();

   scene->DrawIcon(GetWorldPos(), 11, alpha, size);

                              //draw connection to parent
   if(GetParent1()){
      const S_vector &v0 = GetWorldPos();
      const S_vector &v1 = GetParent1()->GetWorldPos();
      scene->SetRenderMatrix(GetIdentityMatrix());
      scene->DrawLine(v0, v1, color, alpha);
   }

   S_vector bbox_full[8];

   {
      if(IsBBoxValid()){
         GetBBox1().Expand(bbox_full);
         scene->SetRenderMatrix(GetBBoxMatrix1() * GetMatrix());
                              //draw sides
         scene->DrawTriangles(bbox_full, 8, sides_i, 3*8,//sizeof(sides_i)/sizeof(word),
            color, sizeof(S_vector), strong ? 0xe0 : 0xe8);

                              //draw edges
         scene->DebugDrawBBox(bbox_full, color, alpha);

         {                     //draw direction line
            S_vector v[2];
            v[0].Zero();
            v[0].z = GetBBox1().max.z;
            v[1] = v[0];
            v[1].z += GetBBox1().max.z - GetBBox1().min.z;
            static const word dir_list[2] = {0, 1};
            scene->DrawLines(v, 2, dir_list, 2, 
               S_vector(1, 0, 1), 
               sizeof(S_vector), alpha);
         }
      }
   }

   if(drv->GetFlags2()&DRVF2_DRAWVOLUMES)
   if(IsVolumeBoxValid()){
      GetVolumeBox1().Expand(bbox_full);
      scene->SetRenderMatrix(GetVolumeBoxMatrix1() * GetMatrix());

      const S_vector &color = strong ?
         S_vector(1.0, 0.0, 0.0) : volume_color;
                              //draw sides
      scene->DrawTriangles(bbox_full, 8, sides_i, sizeof(sides_i)/sizeof(word),
         color, sizeof(S_vector), strong ? 0x80 : 0x80);

      byte alpha = strong ? 0x00 : 0x80;
                              //draw edges
      scene->DebugDrawBBox(bbox_full, color, alpha);
   }
}

//----------------------------

I3D_RESULT I3D_joint::DebugDraw(PI3D_scene scn) const{

   Draw1(scn, true);
   return I3D_OK;
}

//----------------------------

void I3D_joint::UpdateMatrixInternal(){

   I3D_frame::UpdateMatrixInternal();

   if(!drv->IsDirectTransform()){
                              //affect parent SM - reset light
      for(PI3D_frame frm = GetParent1(); frm; frm=frm->GetParent1()){
         if(frm->GetType1()==FRAME_VISUAL && I3DCAST_VISUAL(frm)->GetVisualType()==I3D_VISUAL_SINGLEMESH){
            I3DCAST_VISUAL(frm)->vis_flags &= ~VISF_DEST_LIGHT_VALID;
            break;
         }
      }
   }
}

//----------------------------
//----------------------------
