/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.
   All rights reserved.

   File: Light.cpp
   Content: Light frame.
--------------------------------------------------------*/

#include "all.h"
#include "object.h"
#include "light.h"
#include "scene.h"
#include "camera.h"
#include <malloc.h>           //alloca


#define FOG_FADE_OUT_RATIO 1.0f  //ratio between fog's distance to camera and its radius, below which fog starts to fade out

//----------------------------

inline void DEBUGLINE(PI3D_driver drv, int type, const S_vector &from, const S_vector &to, const S_vector &color = S_vector(1, 1, 1)){
   S_vector v[3] = {from, to, color};
   drv->DEBUG_CODE(type, (dword)v);
}

//----------------------------

                              //light class
I3D_light::I3D_light(PI3D_driver d):
   I3D_frame(d),
   lflags(I3DLIGHTMODE_VERTEX | I3DLIGHTMODE_LIGHTMAP),
   last_render_time(-1),
   range_n(5.0f),
   range_f(10.0f),
   power(1.0f),
   ltype(I3DLIGHT_NULL),
   normalized_dir(0, 0, 0),
   dw_color(0),
   range_n_scaled(0),
   range_f_scaled(0),
   range_f_scaled_2(0),
   range_delta_scaled(0)
{
   drv->AddCount(I3D_CLID_LIGHT);

   //ld.cone_in = 20.0f*PI/180.0f;
   //ld.cone_out = 40.0f*PI/180.0f;
   //ld.specular = S_vector(1.0f, 1.0f, 1.0f);
   //ld.specular_power = 0.5f;

   type = FRAME_LIGHT;
   enum_mask = ENUMF_LIGHT;

   color.Zero();              //avoid SetColor to be too smart
   SetColor(S_vector(1.0f, 1.0f, 1.0f));
}

//----------------------------

I3D_light::~I3D_light(){

                              //unlink all light sectors
   while(light_sectors.size())
      light_sectors[0]->DeleteLight(this);

   drv->DecCount(I3D_CLID_LIGHT);
}

//----------------------------

I3D_RESULT I3D_light::Duplicate(CPI3D_frame frm){

   if(frm->GetType1()==FRAME_LIGHT){
      PI3D_light lp = I3DCAST_LIGHT(frm);

      while(NumLightSectors())
         GetLightSectors()[0]->DeleteLight(this);

      SetLightType(lp->GetLightType1());

      for(int i=lp->NumLightSectors(); i--; )
         lp->GetLightSectors()[i]->AddLight(this);

      color = lp->color;
      power = lp->power;
      range_n = lp->range_n;
      range_f = lp->range_f;
      dw_color = lp->dw_color;
      drv = lp->drv;
      lflags = lp->lflags;
      lflags |= LFLG_DIRTY | LFLG_COLOR_DIRTY;
   }
   return I3D_frame::Duplicate(frm);
}

//----------------------------

I3D_RESULT I3D_light::SetLightType(I3D_LIGHTTYPE t){

   if(ltype==t)
      return I3D_OK;

   switch(t){
   case I3DLIGHT_NULL:
   case I3DLIGHT_POINT:
   //case I3DLIGHT_SPOT:
   case I3DLIGHT_DIRECTIONAL:
   case I3DLIGHT_AMBIENT:
   case I3DLIGHT_FOG:
   case I3DLIGHT_POINTAMBIENT:
   case I3DLIGHT_POINTFOG:
   case I3DLIGHT_LAYEREDFOG:
      break;
   default:
      return I3DERR_INVALIDPARAMS;
   }

   MarkDirty();

   bool was_fog = IsFogType();
   ltype = t;
   bool is_fog = IsFogType();

   if(is_fog != was_fog){
                              //change list in sector's lists
      for(int i=light_sectors.size(); i--; ){
         PI3D_sector sct = light_sectors[i];
         sct->sct_flags |= SCTF_LIGHT_LIST_RESET;
         vector<PI3D_light> *l[2] = { &sct->light_list, &sct->fog_light_list };
         int pi;
         pi = FindPointerInArray((void**)l[was_fog]->begin(), l[was_fog]->size(), this);
         assert(pi!=-1);
         (*l[was_fog])[pi] = l[was_fog]->back(); l[was_fog]->pop_back();
         assert(-1 == FindPointerInArray((void**)l[is_fog]->begin(), l[is_fog]->size(), this));
         l[is_fog]->push_back(this);
      }
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_light::SetMode(dword m){

   if(m&(~LFLG_MODE_MASK))
      return I3DERR_INVALIDPARAMS;
   if(GetMode()!=m){
      MarkDirty();
      lflags &= ~LFLG_MODE_MASK;
      lflags |= m;
      MarkDirty();
   }
   return I3D_OK;
}

//----------------------------

void lightUpdColor(PI3D_driver drv, const S_vector &color, float power,
   S_vector *col, dword *dw_color, bool surf_match){

   *col = (color * power);
                              //match dword color to current pixel format
   int c[3] = {
      FloatToInt((*col)[0]*255.0f),
      FloatToInt((*col)[1]*255.0f),
      FloatToInt((*col)[2]*255.0f)
   };
   if(surf_match)
   {
      const S_pixelformat &pf = *drv->GetGraphInterface()->GetPixelFormat();
      for(int i=0; i<3; i++){
         dword dw;
         const dword &mask = (&pf.r_mask)[i];
         dw = FindLastBit(mask);
         dw = Ror(mask, dw-7);
         c[i] += ((~(signed char)dw)+1)/2;
         c[i] = Min(255, c[i]) & dw;
      }
   }
   for(int i=0; i<3; i++) c[i] = Max(0, Min(255, c[i]));

   *dw_color = 0xff000000 | (c[0]<<16) | (c[1]<<8) | c[2];
}

//----------------------------

I3D_RESULT I3D_light::SetColor(const S_vector &v){

                              //check if some change occured
   float delta = fabs(color.x-v.x) + fabs(color.y-v.y) + fabs(color.z-v.z);
   if(delta<MRG_ZERO)
      return I3D_OK;

   color = v;
   MarkDirty();
   lflags |= LFLG_COLOR_DIRTY;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_light::SetPower(float p){
                              //check if some change occured
   float delta = fabs(power-p);
   if(delta<MRG_ZERO) return I3D_OK;

   power = p;
   MarkDirty();
   lflags |= LFLG_COLOR_DIRTY;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_light::SetRange(float rn, float rf){

                              //check if some change occured
   float delta = fabs(range_n-rn) + fabs(range_f-rf);
   if(delta<MRG_ZERO)
      return I3D_OK;

   range_n = rn;
   range_f = Min(1e+10f, rf);

   MarkDirty();
   return I3D_OK;
}

//----------------------------

void I3D_light::SetOn(dword flg){

   if(flg){
      if(!(frm_flags&FRMFLAGS_ON)){
         I3D_frame::SetOn(flg);
         MarkDirty();
      }
   }else{
      if(frm_flags&FRMFLAGS_ON){
         MarkDirty();
         I3D_frame::SetOn(flg);
      }
   }
}

//----------------------------

void I3D_light::MarkDirty(){

   if(!(frm_flags&FRMFLAGS_ON))
      return;
   lflags |= LFLG_DIRTY;

   if(lflags&(I3DLIGHTMODE_VERTEX|I3DLIGHTMODE_DYNAMIC_LM)){
      for(int i=light_sectors.size(); i--; )
         light_sectors[i]->frm_flags |= FRMFLAGS_HR_LIGHT_RESET;
   }
}

//----------------------------

void I3D_light::Update(){

   if(GetFrameFlags()&FRMFLAGS_UPDATE_NEEDED)
      MarkDirty();
   I3D_frame::Update();
}

//----------------------------

void I3D_light::UpdateLight(){

   if(lflags & (LFLG_DIRTY | LFLG_COLOR_DIRTY)){
      if(lflags&LFLG_COLOR_DIRTY){
                              //update color
         switch(ltype){
         case I3DLIGHT_FOG:
         case I3DLIGHT_LAYEREDFOG:
         case I3DLIGHT_AMBIENT:
            {
               S_vector dst_color = color * power;
                              //make dword color
               int c[3] = {
                  FloatToInt(dst_color[0]*255.0f),
                  FloatToInt(dst_color[1]*255.0f),
                  FloatToInt(dst_color[2]*255.0f)
               };
               if(ltype==I3DLIGHT_FOG){
                              //match dword color to current pixel format
                  const S_pixelformat &pf = *drv->GetGraphInterface()->GetPixelFormat();
                  for(int i=0; i<3; i++){
                     dword dw;
                     const dword &mask = (&pf.r_mask)[i];
                     dw = FindLastBit(mask);
                     dw = Ror(mask, dw-7);
                     c[i] = Min(255, c[i] + ((~(signed char)dw)+1)/2);
                     c[i] &= dw;
                  }
               }
                              //clamp
               for(int i=0; i<3; i++) c[i] = Max(0, Min(255, c[i]));
               dw_color = 0xff000000 | (c[0]<<16) | (c[1]<<8) | c[2];
            }
            break;
         }
      }

      switch(ltype){
      case I3DLIGHT_NULL:
      case I3DLIGHT_AMBIENT:
      case I3DLIGHT_FOG:
         break;
      case I3DLIGHT_POINTFOG:
      case I3DLIGHT_LAYEREDFOG:
      case I3DLIGHT_POINT:
      //case I3DLIGHT_SPOT:
      case I3DLIGHT_DIRECTIONAL:
      case I3DLIGHT_POINTAMBIENT:
         {
                                 //update characteristics
            const S_matrix &m = GetMatrix();
            float scale = m(2).Magnitude();
            if(!IsAbsMrgZero(scale))
               normalized_dir = m(2) / scale;
            else
               normalized_dir.Zero();

                              //compute run-time cached values
            range_n_scaled = range_n * scale;
            range_f_scaled = range_f * scale;
            range_f_scaled_2 = range_f_scaled * range_f_scaled;
            range_delta_scaled = range_f_scaled - range_n_scaled;
         }
         break;
      default: assert(0);
      }
      lflags &= ~(LFLG_DIRTY | LFLG_COLOR_DIRTY);
   }
}

//----------------------------

#define SPHERE_NUM_VERTS 24
static word sphere_c_list[SPHERE_NUM_VERTS*2];

void I3D_light::Draw1(PI3D_scene scene, const S_view_frustum *vf, bool strong) const{

                              //sphere values
   static struct S_auto_init{
      S_auto_init(){
                                 //init volume sphere
         for(int i=0; i<SPHERE_NUM_VERTS; i++){
            sphere_c_list[i*2+0] = i + 0;
            sphere_c_list[i*2+1] = i + 1;
         }
         sphere_c_list[i*2-1] = 0;
      }
   } auto_init;

   static const S_vector color_cone_out(.75f, .75f, .2f),
      color_cone_in(.5f, .5f, .13f);

   byte alpha(strong ? 0 : 0xa0);

   dword lt = GetLightType();
   switch(lt){

   case I3DLIGHT_NULL:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 1, alpha);
      break;

   case I3DLIGHT_POINT:
   case I3DLIGHT_POINTAMBIENT:
   case I3DLIGHT_POINTFOG:
   case I3DLIGHT_LAYEREDFOG:
      {
         float nr, fr;
         GetRange(nr, fr);
         switch(lt){
         case I3DLIGHT_LAYEREDFOG:
            if(strong)
               scene->DebugDrawCylinder(GetMatrix(), fr, fr, nr, color_cone_out, alpha);
            scene->DebugDrawArrow(GetMatrix(), nr, color_cone_in, alpha, 1);
            break;
         case I3DLIGHT_POINTFOG:
            scene->DebugDrawSphere(GetMatrix(), fr, color_cone_out, alpha);
            scene->DebugDrawSphere(GetMatrix(), nr, color_cone_in, alpha);
            break;
         default:
            scene->DebugDrawSphere(GetMatrix(), fr, color_cone_out, alpha);
            scene->DebugDrawSphere(GetMatrix(), nr, color_cone_in, alpha);
         }

         if(vf){              //check clipping
            bool clip;
            if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
               return;
         }

         switch(lt){
         case I3DLIGHT_POINT: scene->DrawIcon(GetWorldPos(), 1, alpha); break;
         case I3DLIGHT_POINTAMBIENT: scene->DrawIcon(GetWorldPos(), 6, alpha); break;
         case I3DLIGHT_POINTFOG: scene->DrawIcon(GetWorldPos(), 12, alpha); break;
         case I3DLIGHT_LAYEREDFOG: scene->DrawIcon(GetWorldPos(), 13, alpha); break;
         }
      }
      break;

   case I3DLIGHT_DIRECTIONAL:
   //case I3DLIGHT_SPOT:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DebugDrawArrow(GetMatrix(), 1.0f, color_cone_out, alpha);
      scene->DrawIcon(GetWorldPos(), lt==I3DLIGHT_DIRECTIONAL ? 4 : 7, alpha);
      break;

      /*
   case I3DLIGHT_SPOT:
      {
         float cone[2];
         GetCone(cone[0], cone[1]);
         cone[0] *= .5f; cone[1] *= .5f;
         float rng[2];
         GetRange(rng[0], rng[1]);
         S_matrix m(GetIdentityMatrix());
         const int sphere_num_verts = 24;
         S_vector sphere_circle[sphere_num_verts+1];
                     //outer range
         float cosc = cos(cone[1]);
         float sinc = sin(cone[1]);
         for(int j=0; j<sphere_num_verts; j++){
            sphere_circle[j].z = cosc;
            float f = (float)j * GetPI2() / (float)sphere_num_verts;
            sphere_circle[j].x = sinc * cos(f);
            sphere_circle[j].y = sinc * sin(f);
         }
         sphere_circle[sphere_num_verts].Zero();
         for(int ii=0; ii<2; ii++){
            m(0, 0) = m(1, 1) = m(2, 2) = rng[ii];
            drv->SetWorldMatrix(m*GetMatrix());
            drv->DrawLines(sphere_circle, sphere_num_verts, sphere_c_list, sphere_num_verts*2,
               !ii ? color_cone_in : color_cone_out, sizeof(S_vector), alpha);
                     //connecting lines
            static const word indx[]={
               sphere_num_verts, 0,
               sphere_num_verts, sphere_num_verts/4,
               sphere_num_verts, sphere_num_verts/2,
               sphere_num_verts, sphere_num_verts*3/4,
            };
            drv->DrawLines(sphere_circle, sphere_num_verts+1,
               indx, sizeof(indx)/sizeof(word), color_cone_out, sizeof(S_vector), alpha);
         }
         scene->DrawIcon(GetWorldPos(), 5, alpha);
      }
      break;
      */

   case I3DLIGHT_AMBIENT:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 6, alpha);
      break;

   case I3DLIGHT_FOG:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 7, alpha);
      break;
   }
}

//----------------------------

I3D_RESULT I3D_light::DebugDraw(PI3D_scene scene) const{

   Draw1(scene, NULL, true);
   return I3D_OK;
}

//----------------------------
//----------------------------
