/*--------------------------------------------------------
   Copyright (c) 2000, 2001 Lonely Cat Games.
   All rights reserved.

   File: Lmap.cpp
   Content: Lit-object frame.
--------------------------------------------------------*/

#include "all.h"
#include "lmap.h"
#include "light.h"
#include "scene.h"
#include "camera.h"
#include <c_cache.h>
#include <conv_rgb.hpp>
#include <compress.h>
#include <malloc.h>



#define USE_ASM               //use inline assembly parts of code

#define LM_REQUEST_TEXTURE_SIZE 256
#define LM_MAX_COMPUTE_SIZE 256  //maximal size of lightmap we may compute

#define MAX_LM_PIXELS (256*256*16)  //max # of pixels we may allocate for lightmapping

//#define LMAP_RANDOM_INIT      //init by random contents
//#define FADE_SHADOWS

#define EDGE_OUT_DISTANCE 3.1f   //lumel's distance from edge, behind which it is considered not to affect particular rectangle
#define INSIDE_PUSH_RATIO .0075f  //ratio at which we push reference point inside of closest rectangle during light computation - this causes proper collision detection
#define LM_CONSTR_PLANAR_THRESH (PI*.6f) //angle at which faces are considered planar (between same smoothgroup)
#define EXPAND_LUMEL_SIZE .5f //number of lumels to expand edges on each side of rectangle

#define MIN_RESOLUTION .001f  //minimal resolution user may set

                              //debugging:
//#define DEBUG_REAL_TIME       //recompute lighting in real-time

#define LM_ALLOC_ERR(driver,msg) driver->DEBUG(msg, 2)

#ifdef _DEBUG

#define LM_ALLOC_LOG(driver,n)
//#define LM_ALLOC_LOG(driver,msg) driver->PRINT(msg)

//#define DEBUG_DRAW_LUMEL_NORMAL
//#define DEBUG_DRAW_RAYS

#else
#define LM_ALLOC_LOG(driver,n)
#endif

//----------------------------
// Fill rectangle in a texture by white (debugging purposes).
static void FillTextureByWhite(PI3D_texture tp, const I3D_rectangle *rc = NULL){

   D3DLOCKED_RECT lrc;
   const IImage *img = tp->GetSysmemImage();
   byte bpp = img->GetPixelFormat()->bytes_per_pixel;
   HRESULT hr = img->GetDirecXTexture()->LockRect(0, &lrc, (LPRECT)rc, 0);
   if(FAILED(hr))
      return;

   int sx = !rc ? img->SizeX() : (rc->r - rc->l);
   int sy = !rc ? img->SizeY() : (rc->b - rc->t);
   for(int i=sy; i--; ){
      memset(lrc.pBits, -1, sx*bpp);
      (byte*&)lrc.pBits += lrc.Pitch;
   }
   img->GetDirecXTexture()->UnlockRect(0);
   tp->SetDirty();
}

/*--------------------------------------------------------
   C_LM_alloc_block
   Abstract:
      Allocation handle of multiple LM rectangles - used
      by single LM object to assure that all LMaps are
      on same texture.
--------------------------------------------------------*/

struct I3D_LM_rect{
   I3D_rectangle rect;
   bool dirty;
};

class C_LM_alloc_block{
public:
   class C_LM_surface *lms;   //for use by LM manager only
   vector<I3D_LM_rect> rect_list;
   dword last_render_time;

   C_LM_alloc_block(){}
   C_LM_alloc_block(int num_rects):
      rect_list(num_rects),
      last_render_time(0),
      lms(NULL)
   {}
};

/*--------------------------------------------------------
   C_rect_allocator
   Abstract:
      Memory allocation manager of rectanguar areas on a surface.
--------------------------------------------------------*/

class C_rect_allocator{
public:
   struct S_block{
      C_LM_alloc_block *owner_id; //identificator of owner
      I3D_rectangle rect;     //position+size
   };
private:
   vector<S_block> block_list;//currently allocated blocks
   int sx, sy;                //size of rectangle

                              //allocation help:
   int size_free;
   short *alloc_map_x_y;      //X * Y per-block allocation info, highest bit marks if block is used
   short *line_info;          //info per each line - # of blocks free
public:
   C_rect_allocator(int sx1, int sy1):
      sx(sx1),
      sy(sy1),
      size_free(sx1*sy1)
   {
      alloc_map_x_y = new short[sx * sy + sy];
      line_info = alloc_map_x_y + (sx * sy);
      for(int y=sy; y--; ){
         line_info[y] = sx;
         short *alloc_map_x = alloc_map_x_y + (y * sx);
         for(int x=sx+1; --x; ++alloc_map_x)
            *alloc_map_x = x;
      }
   }
   ~C_rect_allocator(){
#ifdef _DEBUG
      assert(size_free == sx*sy);
                              //assert that alloc maps are valid upon exit
      for(int y=0; y<sy; y++){
         assert(line_info[y] == sx);
         for(int x=0; x<sx; x++)
            assert(alloc_map_x_y[y*sx + x] == sx-x);
      }
#endif
      delete[] alloc_map_x_y;
   }
   bool Allocate(int req_size_x, int req_size_y, C_LM_alloc_block *owner, I3D_rectangle &out_rc);
   void Free(C_LM_alloc_block *owner);
   inline int GetFreeSize() const{ return size_free; }
   inline int GetTotalSize() const{ return sx * sy; }

   inline int NumBlocks() const{ return block_list.size(); }
   inline const S_block &GetBlock(int i) const{ return block_list[i]; }
};

//----------------------------

bool C_rect_allocator::Allocate(int req_size_x, int req_size_y, C_LM_alloc_block *owner,
   I3D_rectangle &out_rc){

                              //check invalid requirements
   if(req_size_x>sx || req_size_y>sy) return false;

   if(!block_list.size()){
                              //take 1st position
      out_rc.l = 0;
      out_rc.t = 0;
   }else{
                              //first scan vertical list to find appropriate area
      int y;
      int i;

      for(y=0; y <= sy-req_size_y; y++){
         if(line_info[y] >= req_size_x){
                              //scan next n lines to see if their size matches
            for(i=1; i<req_size_y; i++){
               if(line_info[y+i] < req_size_x)
                  break;
            }
            if(i==req_size_y){
               while(true){
                              //lines found, try columns
                  short *alloc_map_x = alloc_map_x_y + (y * sx);
                  for(int x=0; x!=sx; ){
                     assert(x<sx);
                     int bl_free = alloc_map_x[x];
                              //block allocated or not satisfying, skip
                     if(bl_free<req_size_x){
                              //move past the failed area
                        x += bl_free&0x7fff;
                        continue;
                     }
                              //block ok, try to fit on the area,
                              // but consult fail tries first
                     {
                              //try down
                        for(i=1; i<req_size_y; i++){
                           if(alloc_map_x_y[(y+i)*sx + x] < req_size_x)
                              break;
                        }
                        if(i==req_size_y){
                              //alloc ok, take this block
                           out_rc.l = x;
                           out_rc.t = y;
                           goto ok;
                        }
                              //this column failed, move past the failed area
                        x += bl_free;
                     }
                  }
                              //alloc failed, move on to next line
                  ++y;
                  if(y > (sy-req_size_y))
                     break;
                  if(line_info[y+req_size_y-1] < req_size_x)
                     break;
                              //ok, continue with this segment
               }
            }else{
                              //move past the failed area
               y += i;
            }
         }
      }
                              //no satisfying block found, alloc failed
      return false;
   }
ok:
                              //save block
   out_rc.r = out_rc.l + req_size_x;
   out_rc.b = out_rc.t + req_size_y;
   block_list.push_back(S_block());
   block_list.back().owner_id = owner;
   block_list.back().rect = out_rc;
                              //count free size
   size_free -= req_size_x * req_size_y;

                              //adjust alloc maps
   {
                              //mark all lines
      for(int y=out_rc.t; y<out_rc.b; y++){
                              //lower free space on line
         line_info[y] -= req_size_x;
                              //create block
         short *alloc_map_x = alloc_map_x_y + (y * sx);
         for(int x=out_rc.r; --x >= out_rc.l; )
            alloc_map_x[x] = (out_rc.r - x) | 0x8000;

         if(x>=0){
            if(alloc_map_x[x]&0x8000){
               ++x;
                              //adjust used block
               while(x-- && (alloc_map_x[x]&0x8000))
                  alloc_map_x[x] = (out_rc.l - x) | 0x8000;
            }else{
               ++x;
                              //adjust unused block
               while(x-- && !(alloc_map_x[x]&0x8000))
                  alloc_map_x[x] = out_rc.l - x;
            }
         }
      }
   }
#ifdef _DEBUG
                              //assert that blocks never overlap
   {
      for(int i=block_list.size()-1; i--; ){
         const I3D_rectangle &rc = block_list[i].rect;
         bool overlap = (rc.r > out_rc.l && rc.l < out_rc.r &&
            rc.b > out_rc.t && rc.t < out_rc.b);
         assert(!overlap);
      }
   }
#endif

   return true;
}

//----------------------------

void C_rect_allocator::Free(C_LM_alloc_block *owner){

                              //free all blocks of the 'owner'
   for(int i=block_list.size(); i--; ){
      S_block &bl = block_list[i];
      if(bl.owner_id==owner){
                           //count free size
         const I3D_rectangle &rc = bl.rect;
         size_free += (rc.r - rc.l) * (rc.b - rc.t);
                           //adjust alloc maps
         int req_size_x = rc.r - rc.l;
         for(int y=rc.t; y<rc.b; y++){
                              //adjust free space on line
            line_info[y] += req_size_x;
            short *alloc_map_x = alloc_map_x_y + (y * sx);

            int free_base = 0;
            if(rc.r != sx && alloc_map_x[rc.r] > 0)
               free_base = alloc_map_x[rc.r];

                              //kill block
            for(int x=rc.r; --x >= rc.l; )
               alloc_map_x[x] = free_base + (rc.r - x);

            if(x>=0){
               if(alloc_map_x[x]&0x8000){
                  ++x;
                              //adjust used block
                  while(x-- && (alloc_map_x[x]&0x8000))
                     alloc_map_x[x] = (rc.l - x) | 0x8000;
               }else{
                  ++x;
                              //adjust unused block
                  while(x-- && !(alloc_map_x[x]&0x8000))
                     alloc_map_x[x] = free_base + (rc.r - x);
               }
            }
         }
                           //erase block
         bl = block_list.back(); block_list.pop_back();
      }
   }
}

/*--------------------------------------------------------
   C_LM_surface
   Abstract:
      Logical surface which allows allocation of small rectangular areas used as light-maps.
      - contains a texture which provides physical surface
      - manages allocation
--------------------------------------------------------*/

class C_LM_surface: public C_unknown{
public:
   C_rect_allocator alloc;
   C_smart_ptr<I3D_texture> txt;

   C_LM_surface(PI3D_texture txt1):
      alloc(txt1->SizeX1(), txt1->SizeY1()),
      txt(txt1)
   { }
};

//----------------------------

static const S_visual_property props[] = {
   I3DPROP_FLOAT, "Resolution",
};

static I3D_visual *CreateLitObject(PI3D_driver drv){
   return new I3D_lit_object(drv);
}

static struct S_auto_init_lmap{
   S_auto_init_lmap(){
      RegisterVisual(CreateLitObject, I3D_VISUAL_LIT_OBJECT, "Lit-object",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_lmap;

//----------------------------
//----------------------------


I3D_lit_object::I3D_lit_object(PI3D_driver d):
   I3D_visual(d),
   lmap_flags(0),
   curr_resolution(1.0f),
   any_rect_dirty(true),
   //lmap_uv_set(-1),
   lm_alloc_block(NULL)
{
   visual_type = I3D_VISUAL_LIT_OBJECT;
}

//----------------------------

I3D_lit_object::~I3D_lit_object(){

   if(lm_alloc_block)
      drv->LMFree(lm_alloc_block);
}

//----------------------------

I3D_RESULT I3D_lit_object::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);

   Destruct();

   PI3D_visual vis = I3DCAST_VISUAL(frm);

   switch(vis->GetVisualType()){
   case I3D_VISUAL_LIT_OBJECT:
      {
         PI3D_lit_object lmap = (PI3D_lit_object)vis;
         lm_rects = lmap->lm_rects;
         for(int i=lm_rects.size(); i--; ){
            lm_rects[i]->dirty = true;
         }
         face_lmap_use = lmap->face_lmap_use;
         any_rect_dirty = true;
         lmap_flags &= ~LMAPF_CONSTRUCTED;
         if(lmap->lmap_flags&LMAPF_CONSTRUCTED)
            lmap_flags |= LMAPF_CONSTRUCTED;
      }
      break;
   }
   SetMesh(vis->GetMesh());

   return I3D_visual::Duplicate(frm);
}

//----------------------------

void I3D_lit_object::PrepareDestVB(I3D_mesh_base *mb, dword){

   mesh->ComputeBoundingVolume();

   dword numv = mb->NumVertices1();
   dword fvf_flags = (2<<D3DFVF_TEXCOUNT_SHIFT);

   if(drv->IsDirectTransform())
      vertex_buffer.CreateD3DVB(drv, D3DFVF_XYZ | D3DFVF_NORMAL | fvf_flags, numv, true, true);
   else
      vertex_buffer.CreateD3DVB(drv, fvf_flags | D3DFVF_DIFFUSE, numv, false);

                              //copy data
   D3D_lock<byte> v_src(mesh->vertex_buffer.GetD3DVertexBuffer(), mesh->vertex_buffer.D3D_vertex_buffer_index, numv,
      mesh->GetSizeOfVertex(), D3DLOCK_READONLY);
   D3D_lock<byte> v_dst(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.D3D_vertex_buffer_index, numv, vertex_buffer.GetSizeOfVertex(), 0);
   v_src.AssureLock();
   v_dst.AssureLock();
   const I3D_vertex_lmap *src = (const I3D_vertex_lmap*)(byte*)v_src;

   if(drv->IsDirectTransform()){
      struct S_vertex{
         S_vector xyz;
         S_vector normal;
         I3D_text_coor tex[2];
      };
      S_vertex *dst = (S_vertex*)(byte*)v_dst;

      for(int i=numv; i--; ++src, dst=(S_vertex*)((byte*)dst+vertex_buffer.GetSizeOfVertex())){
         dst->xyz = src->xyz;
         dst->normal = src->normal;
         dst->tex[1] = src->tex[0];
      }
   }else{
                              //copy only base UV from source into dest UV1
      I3D_text_coor *dst = (I3D_text_coor*)(((byte*)v_dst) + vertex_buffer.GetTexCoordOffset(1));
      for(int i=numv; i--; ++src, dst=(I3D_text_coor*)((byte*)dst+vertex_buffer.GetSizeOfVertex()))
         (*dst) = src->tex[0];
   }

   vis_flags &= ~VISF_DEST_UV0_VALID;
}

//----------------------------

struct S_face{
   int group_index;           //face-group index
   int face_index;            //face index within the group
   I3D_triface face;
   bool free;
   S_vector normal;
   byte smooth_group;
};

I3D_RESULT I3D_lit_object::Construct(PI3D_scene scene, dword flags, PI3D_LOAD_CB_PROC cb_proc, void *cb_context){

   if(!mesh)
      return I3DERR_GENERIC;
   assert(curr_resolution >= MIN_RESOLUTION*.5f);

   Destruct();

   assert(!(lmap_flags&LMAPF_CONSTRUCTED));

   any_rect_dirty = true;

   int i, j;

   const I3D_vertex_lmap *verts = (const I3D_vertex_lmap*)mesh->vertex_buffer.Lock(D3DLOCK_READONLY);
   int num_verts = mesh->vertex_buffer.NumVertices();

                              //keeping remapped faces
   vector<S_face> face_list;

   for(i = 0; i<mesh->NumFGroups1(); i++){
      face_lmap_use.push_back(vector<word>());
      face_lmap_use.back().assign(mesh->GetFGroups1()[i].num_faces, -1);
   }

   {
                              //perform vertex mapping - this causes proper
                              // computation of meshes, for which
                              // verts have been duplicated because of
                              // texture mapping
      vector<word> v_map(num_verts);
      {
         const vector<word> &v_map1 = mesh->vertex_buffer.GetVertexSpreadMap();
         int num_orig_verts = num_verts - v_map1.size();
         for(int i=0; i<num_orig_verts; i++)
            v_map[i] = i;
         for(; i<num_verts; i++)
            v_map[i] = v_map1[i - num_orig_verts];
      }

      PI3D_triface faces = mesh->index_buffer.Lock(D3DLOCK_READONLY);
      const byte *smooth_groups = mesh->GetSmoothGroups();
                              //collect all faces
                              // + build face_lmap_use (uninitialized)
      for(i = 0; i<mesh->NumFGroups1(); i++){
         const I3D_face_group *fg = &mesh->GetFGroups1()[i];
         for(j=0; j<fg->num_faces; j++){
            face_list.push_back(S_face());
            S_face &f = face_list.back();
            f.group_index = i;
            f.face_index = j;
            f.face = *faces++;
            f.face[0] = v_map[f.face[0]];
            f.face[1] = v_map[f.face[1]];
            f.face[2] = v_map[f.face[2]];
            f.free = true;
            f.normal.GetNormal(verts[f.face[0]].xyz, verts[f.face[1]].xyz, verts[f.face[2]].xyz);
            if(f.normal.IsNull()){
                              //discard this face (invalid)
               face_lmap_use[i][j] = 0;
               f.free = false;
               continue;
            }
            assert(!f.normal.IsNull());
            f.normal.Normalize();
            f.smooth_group = smooth_groups ? (*smooth_groups++) : 0xff;
         }
      }
      mesh->index_buffer.Unlock();
   }

                              //process all free faces
   for(i=face_list.size(); i--; ){
      if(face_list[i].free){
         S_face &face = face_list[i];
         face.free = false;

                              //faces (indicies) of this lmap
         vector<int> face_indx;
         byte lmap_index = lm_rects.size();
         face_indx.push_back(i);
         face_lmap_use[face.group_index][face.face_index] = lmap_index;

                              //resulting normal of all faces in thresh
         S_vector rc_normal;
         rc_normal = face.normal;

                              //*** collect coplanar adjacent faces ***
         //if(0)
         {
                              //edge list - contains re-mapped vert indicies
            vector<I3D_edge> egde_list;
                              //smooth group for each edge
            vector<byte> smooth_group_list;
            {
               const I3D_triface &fc = face.face;
               egde_list.push_back(I3D_edge(fc[0], fc[1]));
               egde_list.push_back(I3D_edge(fc[1], fc[2]));
               egde_list.push_back(I3D_edge(fc[2], fc[0]));
            
               smooth_group_list.push_back(face.smooth_group);
               smooth_group_list.push_back(face.smooth_group);
               smooth_group_list.push_back(face.smooth_group);
            }

            while(egde_list.size()){
               I3D_edge e = egde_list.back(); egde_list.pop_back();
               byte sg = smooth_group_list.back(); smooth_group_list.pop_back();

               for(j=i; j--; )
               if(face_list[j].free){
                  S_face &face1 = face_list[j];
                  byte sg1 = face1.smooth_group;
                              //check smoothgroups
                  if(!(sg&sg1))
                     continue;
                  const I3D_triface &fc1 = face1.face;

                              //check if these faces share same edge
                  dword w0 = fc1[0];
                  for(int k=3; k--; ){
                     dword w1 = w0;
                     w0 = fc1[k];
                     if(w0==e[1] && w1==e[0])
                        break;
                  }
                  if(k!=-1){
                              //check angle against all added faces
                     for(int l=face_indx.size(); l--; ){
                        int fci = face_indx[l];
                        float a = face1.normal.AngleTo(face_list[fci].normal);
                        if(a > LM_CONSTR_PLANAR_THRESH)
                           break;
                     }
                     if(l==-1)
                     {
                              //add this face
                        face_indx.push_back(j);
                        face1.free = false;
                        face_lmap_use[face1.group_index][face1.face_index] = lmap_index;
                              //add its edges (except the shared one)

                        egde_list.push_back(I3D_edge(fc1[(k+1)%3], fc1[(k+2)%3]));
                        egde_list.push_back(I3D_edge(fc1[(k+2)%3], fc1[k]));

                        smooth_group_list.push_back(sg1);
                        smooth_group_list.push_back(sg1);

                        rc_normal += face1.normal;
                              //don't process other faces, 
                              // assuming edges are always only among 2 faces
                        break;
                     }
                  }
               }
            }
            assert(!smooth_group_list.size());
         }

         rc_normal.Normalize();

                              //*** make transformation matrix ***
                              //collect all points of faces of this lmap
         vector<dword> pt_list;
         for(j=face_indx.size(); j--; ){
            const I3D_triface &fc = face_list[face_indx[j]].face;
            for(int k=0; k<3; k++){
               dword vi = fc[k];
                              //don't add same vertex twice
               for(int l=pt_list.size(); l--; ) 
                  if(pt_list[l]==vi)
                     break;
               if(l==-1) 
                  pt_list.push_back(vi);
            }
         }

                              //find smallest possible rectangle containing all points 
                              // (in it 2D transformation)
         S_plane lm_plane;
         lm_plane.normal = rc_normal;
         lm_plane.d = -verts[face.face[0]].xyz.Dot(lm_plane.normal);

         float best_size = 1e+16f;
         I3D_bbox best_bbox;
         S_matrix best_mat;

                              //proposed transformation matrix
         S_matrix prop_matrix;
         prop_matrix.Identity();
         prop_matrix(2) = -lm_plane.normal;

         S_vector *v_rect = new S_vector[pt_list.size()];
         S_vector *v_trans = new S_vector[pt_list.size()];
         for(j=pt_list.size(); j--; )
            v_rect[j] = verts[pt_list[j]].xyz;

         for(j = pt_list.size(); j--; ){
                              //get 1st vertex and let it lay on the plane
            const S_vector &v0a = v_rect[j];
            S_vector v0 = v0a - lm_plane.normal * v0a.DistanceToPlane(lm_plane);

            for(int k=j; k--; ){
                              //get 2nd vertex and let it lay on the plane
               const S_vector &v1a = v_rect[k];
               S_vector v1 = v1a - lm_plane.normal * v1a.DistanceToPlane(lm_plane);

               S_vector dir_v0_v1 = v1 - v0;
                              //if verts are too close (identical), skip this vertex
               if(IsMrgZero(dir_v0_v1.Dot(dir_v0_v1)))
                  continue;
                              //compute proposed matrix
               prop_matrix(0) = S_normal(dir_v0_v1);
               prop_matrix(1) = prop_matrix(0).Cross(prop_matrix(2));

               S_matrix prop_matrix_inv = ~prop_matrix;

                              //process all points, build bounds (projected to 2D)
               I3D_bbox rc_bbox;
               rc_bbox.Invalidate();

               TransformVertexArray(v_rect, sizeof(S_vector), pt_list.size(),
                  v_trans, sizeof(S_vector), prop_matrix_inv);

               for(int l=pt_list.size(); l--; ){
                  const S_vector &vt = v_trans[l];
                  rc_bbox.min.Minimal(vt);
                  rc_bbox.max.Maximal(vt);
               }
                              //save result, if better than previous
               float curr_size = (rc_bbox.max.x-rc_bbox.min.x) * (rc_bbox.max.y-rc_bbox.min.y);
               if(best_size > curr_size){
                  best_size = curr_size;
                  best_bbox = rc_bbox;
                  best_mat = prop_matrix;
               }
            }
         }
         delete[] v_trans;
         delete[] v_rect;

                              //*** now we have computed smallest rectangle ***
         lm_rects.push_back(new S_lmap_rectangle());
         S_lmap_rectangle *lm_rect = lm_rects.back();
         lm_rect->Release();
                              //initialize LM rectangle
         lm_rect->m = best_mat;

                              //expand bbox by some value
                              // - helps in phong shading, also it's safer for edges
#if 1
         {
            float f = (1.0f/curr_resolution) * EXPAND_LUMEL_SIZE;
            f /= GetWorldScale();
            best_bbox.min.x -= f;
            best_bbox.min.y -= f;
            best_bbox.max.x += f;
            best_bbox.max.y += f;
         }
#endif
         lm_rect->m(3) = best_bbox.min.RotateByMatrix(lm_rect->m);
         S_vector bbox_diag = best_bbox.max - best_bbox.min;
                              //scale
         {
            for(int i=0; i<3; i++){
               lm_rect->m(0, i) *= bbox_diag.x;
               lm_rect->m(1, i) *= bbox_diag.y;
            }
         }
         lm_rect->inv_m = lm_rect->m;
         lm_rect->m.Invert();
         lm_rect->z_delta = best_bbox.max.z - best_bbox.min.z;
#if 0                         //debug draw lm rextangle
         {
            for(int i=pt_list.size(); i--; ){
               const S_vector &v1 = verts[pt_list[i]].xyz;
               S_vector vt = v1 * lm_rect->m;
               drv->PRINT(vt);
               { S_vector v[2] = {v1,
                  v1 + lm_plane.normal}; drv->DEBUG_CODE(0, (dword)v); }

            }
         }
#endif

         S_vector scale = matrix.GetScale();
                              //init lm surface
         float res = curr_resolution;
         {
                              //compute destination resolution
            dword size_x = bbox_diag.x * res * scale.x + 1;
            dword size_y = bbox_diag.y * res * scale.z + 1;

                              //let size be always at least 2, 
                              // so that we have phong shading,
                              // and avoid problems with too small lm surfaces
                              // (width or height of 1 causes divide by zero, 
                              //    and other problems during light computations)
            size_x = Max(size_x, 2ul);
            size_y = Max(size_y, 2ul);

                              //limit to max possible lmap size
            size_x = Min(size_x, (dword)LM_MAX_COMPUTE_SIZE);
            size_y = Min(size_y, (dword)LM_MAX_COMPUTE_SIZE);

            lm_rect->size_x = size_x;
            lm_rect->size_y = size_y;
            lm_rect->dst_map = new S_lmap_pixel[size_x*size_y];
#ifdef LMAP_RANDOM_INIT       //init by random contents
            {
               memset(dlod.dst_map, 0, sizeof(S_lmap_pixel)*dlod.size_x*dlod.size_y);
               int ii = 0;
               for(j=dlod.size_y-1; j--; ){
                  for(int i=dlod.size_x-1; i--; ++ii){
                     dlod.dst_map[ii].r = Random(256);
                     dlod.dst_map[ii].g = Random(256);
                     dlod.dst_map[ii].b = Random(256);
                  }
                  ++ii;
               }
            }
#endif
         }
         lm_rect->dirty = true;
      }
   }
   mesh->vertex_buffer.Unlock();

   lmap_flags |= LMAPF_CONSTRUCTED;
   CheckUV();

   I3D_RESULT ir = MakeLight(scene, flags, cb_proc, cb_context);
   if(I3D_FAIL(ir))
      Destruct();

   return ir;
}

//----------------------------

I3D_RESULT I3D_lit_object::Destruct(dword){

   if(lm_alloc_block){
      drv->LMFree(lm_alloc_block);
      lm_alloc_block = NULL;
   }
   lm_rects.clear();
   face_lmap_use.clear();
   any_rect_dirty = true;

   lmap_flags &= ~LMAPF_CONSTRUCTED;
   return I3D_OK;
}

//----------------------------

#define LM_SAVE_VERSION 17    //format version

#pragma pack(push,1)
struct S_lobj_data_header{
   byte version;              //current format version, equals to LM_SAVE_VERSION
   byte mode;                 //holds I3D_lit_object::GetMode()
   int num_lods;              //number of levels of detail, 0 for unconstructed saved data
   float ignored;
};

struct S_lmap_header{         //used with light-mapping
   word num_LM_rectangles, num_face_groups;
};

#pragma pack(pop)

//----------------------------

I3D_RESULT I3D_lit_object::Load(int handle){

   if(IsConstructed())
      Destruct();

   if(!mesh)
      return I3DERR_NOTINITIALIZED;

   C_cache *cp = (C_cache*)handle;
                              //read header
   S_lobj_data_header hdr;
   cp->read((byte*)&hdr, sizeof(hdr));
   if(hdr.version != LM_SAVE_VERSION){
      if(hdr.version < 13)
         return I3DERR_INVALIDPARAMS;
   }
   if(!hdr.num_lods){
                              //not constructed
      return I3D_OK;
   }

                              //older versions supported multiple LODs
                              // now it's only 1
   if(hdr.num_lods != 1)
      return I3DERR_INVALIDPARAMS;

   int i;
   any_rect_dirty = true;

   if(hdr.version < 14)
   {                          
                              //load vertex-light information
      word num_vertices;
      cp->read((char*)&num_vertices, sizeof(num_vertices));
      cp->seekg(cp->tellg() + sizeof(dword) * num_vertices);
   }

                              //load light-map information
   {
      S_lmap_header hdr1;
      cp->read((byte*)&hdr1, sizeof(hdr1));

      PI3D_mesh_base mb = mesh;
                              //check if data are valid for current mesh
      if(hdr1.num_face_groups != mb->NumFGroups1())
         return I3DERR_INVALIDPARAMS;
                              //read LM rectangles
      lm_rects.reserve(hdr1.num_LM_rectangles);
      for(i=0; i<hdr1.num_LM_rectangles; i++){
         lm_rects.push_back(new S_lmap_rectangle());
         S_lmap_rectangle *lm_rect = lm_rects.back();
         lm_rect->Release();

         cp->read((byte*)&lm_rect->m, sizeof(S_matrix));
         lm_rect->inv_m = ~lm_rect->m;
         bool is_single_color = false;
         dword single_color;
         if(hdr.version<16){
                              //read is_single_color
            cp->read((char*)&is_single_color, sizeof(byte));
                              //read single color (if present)
            if(is_single_color)
               cp->read((char*)&single_color, sizeof(single_color));
         }
                              //read num of LODs
         dword num_lods;
         cp->read((char*)&num_lods, sizeof(dword));
         if(num_lods!=1)
            goto fail;
         //for(int j=0; j<num_lods; j++)
         {
            cp->read((char*)&lm_rect->size_x, sizeof(dword)*2);
            if(!is_single_color){
                              //read compressed data 
               dword num_elements;
               Decompress(cp, &lm_rect->dst_map, &num_elements);
               if(hdr.version==14){ //compatibility - swap RGB
                  for(int i=lm_rect->size_x*lm_rect->size_y; i--; )
                     swap(lm_rect->dst_map[i].r, lm_rect->dst_map[i].b);
               }
            }else{
                              //only alloc data
               lm_rect->dst_map = new S_lmap_pixel[lm_rect->size_x * lm_rect->size_y];

                              //compatibility - fill with single color
               for(int i=lm_rect->size_x * lm_rect->size_y; i--; )
                  lm_rect->dst_map[i] = *(S_lmap_pixel*)&single_color;
            }
         }
      }
                              //read face use
      face_lmap_use.reserve(hdr1.num_face_groups);
      for(i=0; i<hdr1.num_face_groups; i++){
         dword nf;
         cp->read((byte*)&nf, sizeof(dword));

                              //check mesh consistency
         const I3D_face_group &fg = mb->GetFGroups1()[i];
         if(nf != fg.NumFaces1())
            goto fail;
                              //read data
         face_lmap_use.push_back(vector<word>(nf));
         cp->read((byte*)face_lmap_use.back().begin(), sizeof(word)*nf);
      }
      lmap_flags |= LMAPF_CONSTRUCTED;
   }

   CheckUV();
   lmap_flags |= LMAPF_CONSTRUCTED;

   return I3D_OK;

fail:
   Destruct();

   return I3DERR_INVALIDPARAMS;
}

//----------------------------

I3D_RESULT I3D_lit_object::Save(int handle){

   C_cache *cp = (C_cache*)handle;

   S_lobj_data_header hdr;
   hdr.version = LM_SAVE_VERSION;
   hdr.ignored = 0.0f;
   hdr.num_lods = IsConstructed() ? 1 : 0;
                              //save header
   cp->write((char*)&hdr, sizeof(hdr));

                              //unconstructed saves only header
   if(!IsConstructed())
      return I3D_OK;

   int i;
                              //save light-map information
   S_lmap_header hdr1;
   hdr1.num_LM_rectangles = lm_rects.size();
   hdr1.num_face_groups = face_lmap_use.size();
   cp->write((char*)&hdr1, sizeof(hdr1));

                              //save LM rectangles info
   for(i=0; i<lm_rects.size(); i++){
      const S_lmap_rectangle *lm_rect = lm_rects[i];
                              //save:
                              // - matrix
                              // - dword num_lods (always 1)
                              // - size
                              // - bitmap
      cp->write((char*)&lm_rect->m, sizeof(S_matrix));

      dword num_lods = 1;
      cp->write((char*)&num_lods, sizeof(dword));
      cp->write((char*)&lm_rect->size_x, sizeof(dword)*2);
                              //write compressed data 
      vector<byte> v;
      Compress(lm_rect->dst_map, lm_rect->size_x*lm_rect->size_y, v);
      cp->write((char*)v.begin(), v.size());
   }     
                              //save face use
   for(i=0; i<face_lmap_use.size(); i++){
                              //save:
                              // - num_faces (byte)
                              // - face use indicies (byte)
      dword nf = face_lmap_use[i].size();
      cp->write((char*)&nf, sizeof(dword));
      cp->write((char*)face_lmap_use[i].begin(), sizeof(word)*nf);
   }
   return I3D_OK;
}

//----------------------------

void I3D_lit_object::CheckUV(){

   if(!IsConstructed())
      return;

   bool vertex_added = false;

   int num_verts = mesh->vertex_buffer.NumVertices();
   vector<byte> vert_lm_use(num_verts, 0xff);
   vector<word> vertex_dup_info; vertex_dup_info.reserve(num_verts);

   PI3D_triface faces = mesh->index_buffer.Lock(0);
   int i, j;
   for(i=0; i<face_lmap_use.size(); i++){
      for(j=0; j<face_lmap_use[i].size(); j++){
         I3D_triface &face = *faces++;
         dword lmap_index = face_lmap_use[i][j];
         for(int k=0; k<3; k++){
            word vi = face[k];
            if(vert_lm_use[vi]==0xff || vert_lm_use[vi]==lmap_index){
               vert_lm_use[vi] = lmap_index;
            }else{
                              //save duplicated vertex index, make copy out of loop
               vertex_dup_info.push_back(vi);
               vert_lm_use.push_back(lmap_index);

               face[k] = num_verts;
               num_verts++;
            }
         }
      }
   }
   if(vertex_dup_info.size()){
                             //duplicate all vertices now at once
      mesh->vertex_buffer.DuplicateVertices(vertex_dup_info.begin(), vertex_dup_info.size());
      vertex_added = true;
   }
   if(vertex_added){
      vertex_buffer.DestroyD3DVB(drv);
      vis_flags &= ~VISF_DEST_UV0_VALID;
   }
   mesh->index_buffer.Unlock();
}

//----------------------------

void I3D_lit_object::MapVertices(){

                              //setup light-map texture
   int vertex_count = mesh->vertex_buffer.NumVertices();
                              //re-map UV coords of this face group
   const I3D_vertex_lmap *v_src = (const I3D_vertex_lmap*)mesh->vertex_buffer.Lock(D3DLOCK_READONLY);
   D3D_lock<byte> v_dst(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.D3D_vertex_buffer_index, vertex_count, vertex_buffer.GetSizeOfVertex(), 0);
   v_dst.AssureLock();
                              //array of 'vertex processed' flags, to avoid duplicating work
   vector<bool> v_ok;
   v_ok.assign(vertex_count, false);

   assert(lm_alloc_block);
   float tx = lm_alloc_block->lms->txt->SizeX1();
   float ty = lm_alloc_block->lms->txt->SizeY1();
   float r_tx = 1.0f / tx;
   float r_ty = 1.0f / ty;
   float half_texel_x = .5f * r_tx;
   float half_texel_y = .5f * r_ty;

                              //make all face-groups
   const I3D_triface *faces = mesh->index_buffer.Lock(D3DLOCK_READONLY);
   for(int j = 0; j<mesh->NumFGroups1(); j++){
      const I3D_face_group *fg = &mesh->GetFGroups1()[j];
      const vector<word> &face_lmaps = face_lmap_use[j];

      for(int i = 0; i<fg->num_faces; i++){
         int lm_index = face_lmaps[i];
         const I3D_triface &face = *faces++;
         const S_lmap_rectangle *lm_rect = lm_rects[lm_index];

                              //compute texture coordinates in light-map texture
         const I3D_LM_rect &t_rc = lm_alloc_block->rect_list[lm_index];

         for(int j=0; j<3; j++){
            int vertex_index = face[j];
            if(v_ok[vertex_index])
               continue;
            v_ok[vertex_index] = true;

            float u, v;
            {
               const S_vector &vs = v_src[vertex_index].xyz;
               const S_matrix &m = lm_rect->m;
                              //transform into 2D uniform rectangle space
               S_vector2 vt;
               vt.x = vs.x * m.m[0][0] + vs.y * m.m[1][0] + vs.z * m.m[2][0] + m.m[3][0];
               vt.y = vs.x * m.m[0][1] + vs.y * m.m[1][1] + vs.z * m.m[2][1] + m.m[3][1];
                              //transform to real LM texture space
               u = (t_rc.rect.l + (t_rc.rect.r-t_rc.rect.l-1) * vt.x) * r_tx + half_texel_x;
               v = (t_rc.rect.t + (t_rc.rect.b-t_rc.rect.t-1) * vt.y) * r_ty + half_texel_y;
            }
            byte *vp = ((byte*)v_dst) + vertex_index * vertex_buffer.GetSizeOfVertex();
            I3D_text_coor &txt = *(I3D_text_coor*)(vp + vertex_buffer.GetTexCoordOffset(0));
            txt.u = u;
            txt.v = v;
         }
      }
   }
   mesh->index_buffer.Unlock();
   mesh->vertex_buffer.Unlock();
   v_dst.Unlock();
}

//----------------------------

struct S_lm_face{
   I3D_triface fc;
   int face_index;//index into original faces
};

//----------------------------

const float unit_light = 128.0f;

#define MIN_PROGRESS_LUMELS 40   //how often progress indicator is called (lumels)
#define MIN_PROGRESS_DELAY 100   //how often we show progress
#define MIN_CANCEL_CHECK_DELAY 250   //how often we check for cancellation

I3D_RESULT I3D_lit_object::MakeLight(PI3D_scene scene, dword flags,
   PI3D_LOAD_CB_PROC cb_proc, void *cb_context){

   I3D_RESULT ir = I3D_OK;
   int i;

   if(lm_alloc_block){
      drv->LMFree(lm_alloc_block);
      lm_alloc_block = NULL;
   }


   const vector<PI3D_light> *sct_lights = NULL;
                              //collect lights which apply to this sector
   PI3D_frame frm_sect = this;
   while(frm_sect && frm_sect->GetType1()!=FRAME_SECTOR) frm_sect = frm_sect->GetParent();
   if(frm_sect){
      PI3D_sector sct = I3DCAST_SECTOR(frm_sect);
      sct_lights = &sct->GetLights1();
   }
                              
                              //get sum of lumels (due to progress indicator)
   int num_lumels = 0;
   for(i=lm_rects.size(); i--; ){
      const S_lmap_rectangle *lm_rect = lm_rects[i];
      num_lumels += lm_rect->size_x * lm_rect->size_y;
   }
   LPIGraph igraph = drv->GetGraphInterface();

                              //progress indicator support
   int last_cancel_check_time = igraph->ReadTimer() - MIN_CANCEL_CHECK_DELAY;
   int last_progress_time = igraph->ReadTimer();
   int last_progress_lumel = 0;
   int progress_lumel_count = 0;
   int progress_state = 0;

   bool inv_face_reported = false;

                              //process all lm_rects
   for(int lmrc_i = 0; lmrc_i != lm_rects.size(); lmrc_i++){
      S_lmap_rectangle *lm_rect = lm_rects[lmrc_i];

      //S_matrix mat_lm = (~lm_rect->m) * matrix;  //MSVC bug - bad compile in release version
      //m = (~lm_rect->m);
      //m = m * matrix;
      //S_matrix mat_lm = lm_rect->m; mat_lm.Invert();
      S_matrix mat_lm = lm_rect->inv_m;
      mat_lm = mat_lm * matrix;

                              //get lm rect plane
      S_plane pl;
      pl.normal = -mat_lm(2);
      pl.normal.Normalize();
      pl.d = -(mat_lm(3).Dot(pl.normal));
      float z_delta = lm_rect->z_delta;

                              //collect all lights which may illuminate the surface
      vector<PI3D_light> rc_lights;
      S_vector lumel_base;    //ambient lightness
      lumel_base.Zero();

      if(sct_lights)
      for(int i=sct_lights->size(); i--; ){
         PI3D_light lp = (*sct_lights)[i];
         lp->UpdateLight();

         I3D_LIGHTTYPE lt;
         if(lp->IsOn1() && (lp->I3D_light::GetMode()&I3DLIGHTMODE_LIGHTMAP) && !IsAbsMrgZero(lp->power))
         switch(lt=lp->GetLightType()){
         case I3DLIGHT_LAYEREDFOG: case I3DLIGHT_FOG:
         case I3DLIGHT_POINTFOG: case I3DLIGHT_NULL:
            break;
         case I3DLIGHT_AMBIENT:
            {
               S_vector color = lp->color * lp->power;
               swap(color.x, color.z);
               lumel_base += color;
            }
            break;

         default:
            {
               const S_matrix &m = lp->GetMatrix();
                              //based on type, compute several values
               switch(lt){
                  /*
               case I3DLIGHT_DIRECTIONAL:
                  {
                           //compute direction to light
                     float dir_cos = -(lp->world_dir.Dot(pl.normal));
                           //optimizations: remove lights shining from back
                     if(dir_cos <= 0.0f) continue;
                  }
                  break;

               case I3DLIGHT_SPOT:
                  {
                           //fast reject lights pointing away
                     float inner, outer;
                     lp->I3D_light::GetCone(inner, outer);
                     float a = acos(lp->world_dir.Dot(pl.normal)) + outer * .5f;
                     if(a < PI*.5f)
                        continue;
                  }
                           //flow...
                  */
               case I3DLIGHT_POINT:
                  {
                              //fast reject lights too far from object
                     {
                        const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
                        S_vector l_dist = bsphere.pos - m(3);
                        float dist_2 = l_dist.Dot(l_dist);
                        float r_sum = lp->range_f_scaled + bsphere.radius;
                        if((r_sum*r_sum) <= dist_2)
                           continue;
                     }

                              //fast reject lights behind plane
                     float dist_to_plane = pl.normal.Dot(m(3)) + pl.d;
                     if((dist_to_plane + z_delta) <= -lp->range_f_scaled)
                        continue;
                              //fast reject lights too far from plane
                     if(dist_to_plane >= lp->range_f_scaled)
                        continue;
                  }
                  break;

               case I3DLIGHT_POINTAMBIENT:
                  {
                              //fast reject lights too far from object
                     const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
                     S_vector l_dist = bsphere.pos - m(3);
                     float dist_2 = l_dist.Dot(l_dist);
                     float r_sum = lp->range_f_scaled + bsphere.radius;
                     if((r_sum*r_sum) <= dist_2)
                        continue;
                  }
                  break;
               }
                              //definitelly add this light
               rc_lights.push_back(lp);
            }
         }
      }

      {
         int size_x = lm_rect->size_x;
         int size_y = lm_rect->size_y;
         assert(size_x>1 && size_y>1);

                              //1. collect all triangles of this LM rectangle
                              //2. transform all verts into LM rect's space (2D)
         const S_vector &mat_diffuse = mesh->GetFGroups1()[0].GetMaterial1()->GetMaterialData().diffuse;

                              //check if additional mapping channel present
         const C_map_channel *mc2 = mesh->GetMapChannel2();
         LPIImage mc2_img = NULL;
         if(mc2 && !mc2->bitmap_name.Size())
            mc2 = NULL;
         if(mc2){
                              //try to load specified texture (image)
            if(drv->map_channel_2_img && drv->map_channel_2_img_name.Matchi(mc2->bitmap_name)){
                              //check if time matches
               const vector<C_string> &dirs = drv->GetDirs(I3DDIR_MAPS);
               for(i=dirs.size(); i--; ){
                  C_fstring fname("%s\\%s", (const char*)dirs[i], (const char*)mc2->bitmap_name);
                  int h = dtaOpen(fname);
                  if(h!=-1){
                     i = -1;
                     dword time[2];
                     if(dtaGetTime(h, time) && !memcmp(time, drv->map_channel_2_time, sizeof(time))){
                        i = -2;
                     }
                     dtaClose(h);
                     break;
                  }
               }
               if(i!=-2){
                              //release cached version
                  drv->map_channel_2_img = NULL;
                  drv->map_channel_2_img_name = NULL;
               }
            }

            if(drv->map_channel_2_img && drv->map_channel_2_img_name.Matchi(mc2->bitmap_name)){
                              //use cached version
               mc2_img = drv->map_channel_2_img;
            }else{
               drv->map_channel_2_img_name = NULL;
               drv->map_channel_2_img = NULL;
                              //try to load the image
               drv->GetGraphInterface()->CreateImage(&mc2_img);
                              //use 24-bit RGB format
               S_pixelformat pf;
               memset(&pf, 0, sizeof(pf));
               pf.bytes_per_pixel = 3;
               pf.r_mask = 0xff0000;
               pf.g_mask = 0x00ff00;
               pf.b_mask = 0x0000ff;
               pf.flags = 0;
                              //look in all bitmap directories
               const vector<C_string> &dirs = drv->GetDirs(I3DDIR_MAPS);
               for(i=dirs.size(); i--; ){
                  C_fstring fname("%s\\%s", (const char*)dirs[i], (const char*)mc2->bitmap_name);
                  bool b = mc2_img->Open(fname,
                     IMGOPEN_SYSMEM, 0, 0, &pf);
                  if(b){
                     drv->map_channel_2_img_name = mc2->bitmap_name;
                     drv->map_channel_2_img = mc2_img;
                     int h = dtaOpen(fname);
                     assert(h!=-1);
                     if(h!=-1){
                        dtaGetTime(h, drv->map_channel_2_time);
                        dtaClose(h);
                     }
                     break;
                  }
               }
               mc2_img->Release();
               if(i==-1){
                              //failed to open the image, report problem to user
                  if(cb_proc)
                     (*cb_proc)(CBM_ERROR, (dword)(const char*)C_fstring("failed to open image: '%s'", (const char*)mc2->bitmap_name), 0, cb_context);
                  mc2 = NULL;
                  mc2_img = NULL;
               }
            }
         }

         const I3D_vertex_lmap *verts = (const I3D_vertex_lmap*)mesh->vertex_buffer.Lock(D3DLOCK_READONLY);

         PI3D_triface faces = mesh->index_buffer.Lock(D3DLOCK_READONLY);
         
         vector<S_lm_face> lm_faces;
         vector<S_vector2> verts_trans;
         {
            for(int fgi=0, face_i = 0; fgi<face_lmap_use.size(); fgi++){
               for(int fi=0; fi<face_lmap_use[fgi].size(); fi++, face_i++, faces++){
                  if(face_lmap_use[fgi][fi] == lmrc_i){
                     lm_faces.push_back(S_lm_face());
                     lm_faces.back().fc = *faces;
                     lm_faces.back().face_index = face_i;
                  }
               }
            }

            int i = mesh->vertex_buffer.NumVertices();
            verts_trans.assign(i, S_vector2());
            while(i--){
               S_vector v = verts[i].xyz * lm_rect->m;
               verts_trans[i].x = v.x * (size_x-1);
               verts_trans[i].y = v.y * (size_y-1);
            }
         }
         mesh->index_buffer.Unlock(); faces = NULL;

         int x, y;
                              //alloc temp mem
         S_vector *lumel_map = new S_vector[size_x * size_y];
         S_vector *mc2_map = new S_vector[size_x * size_y];
         S_vector *lumel = lumel_map;
         S_vector *mc2_lumel = mc2_map;

         bool compress_err = false;

         float aa_pixel_offset = (.5f/(float)drv->lm_create_aa_ratio - .5f);
         float aa_pixel_shift = (1.0f/drv->lm_create_aa_ratio);
         float r_num_aa_levels = (1.0f/(float)(drv->lm_create_aa_ratio*drv->lm_create_aa_ratio));
                              //process all lumens
         for(y=0; y<size_y; y++){
                              //position on 2D LM rectangle, for which we compute lightness
                              // it may have also non-zero fractional part (sub-pixels)
            S_vector2 f_pt;

            for(x=0; x<size_x; x++, ++lumel, ++mc2_lumel){

               lumel->Zero();
               mc2_lumel->Zero();

               bool is_valid_lumel = true;

               f_pt.y = y + aa_pixel_offset;
               for(int aa_i_y=0; aa_i_y<drv->lm_create_aa_ratio && is_valid_lumel; aa_i_y++,
                  f_pt.y += aa_pixel_shift){

                  f_pt.x = x + aa_pixel_offset;
                  for(int aa_i_x=0; aa_i_x<drv->lm_create_aa_ratio; aa_i_x++,
                     f_pt.x += aa_pixel_shift){

                              //determine point on a triangle and normal (phong shading)
               S_vector pt_xyz, pt_normal;
               S_vector col_pt_xyz; //point from which collision testing occurs

                              //face, which this point is definitely inside of
               int in_face_i = -1;
                              //in case lumel is not inside of any face, this is closest face
                              // lumel may be in, with closes edge
               int out_face_i = -1, out_edge_i;
                              //distance of lumel to out-face
               float out_face_dist = 1e+16f;

               {
                              //choose face we should operate on
                  for(int i=lm_faces.size(); i--; ){
                     const I3D_triface &fc = lm_faces[i].fc;
                     
                     const S_vector2 *v[3] = {&verts_trans[fc[0]],
                        &verts_trans[fc[1]],
                        &verts_trans[fc[2]] };

                              //check if point is in triangle,
                              // eventually compute dist to closest edge
                     bool in_face = true;
                     int out_edge_i1 = -1;
                     float out_face_dist1 = 1e+16f;
                     float out_face_dist_sum = 0.0f;

                              //check all 3 edges of the face
                     dword vi0 = 0;
                     for(int k=3; k--; ){
                        dword vi1 = vi0;
                        vi0 = k;

                        float d = 
                           f_pt.DistanceToLine(*v[vi0], *v[vi1] - *v[vi0]);
                        bool out_edge = (d < 0.001f);
                              //since bluring and filtering may be applied, 
                              // include also lumels beyong edges
                        if(out_edge){
                           in_face = false;
                           if(d > -EDGE_OUT_DISTANCE){
                              d = -d;
                              if(out_face_dist1 > d){
                                 out_face_dist1 = d;
                                 out_face_dist_sum += d;
                                 out_edge_i1 = k;
                              }
                           }else{
                              //too far, can't be in this face
                              out_edge_i1 = -1;
                              break;
                           }
                        }
                     }
                     if(in_face){
                              //found face which we're in, stop looking for more
                        in_face_i = i;
                        break;
                     }else
                     if(out_edge_i1!=-1 && out_face_dist>out_face_dist_sum){
                              //choose triangle which is closer to the point,
                              // use sum of 'out-of-edge' values to determine this
                        out_face_dist = out_face_dist_sum;
                        out_edge_i = out_edge_i1;
                        out_face_i = i;
                     }
                  }
                              //check if such face exists in mesh
                  int face_i = in_face_i != -1 ? in_face_i : out_face_i;
                  if(face_i == -1){
                                          //set 'outside' lumels to special color (red)
                     lumel->x = 0; lumel->y = 0; lumel->z += 2.0f;
                     if(mc2_img){
                        mc2_lumel->x += 1.0f; mc2_lumel->y += 1.0f; mc2_lumel->z += 1.0f;
                     }
                     is_valid_lumel = false;
                     //continue;
                     break;
                  }
                              //found in one of faces, or near
                  I3D_triface fc = lm_faces[face_i].fc;
                     
                  const S_vector2 *v[3] = {
                     &verts_trans[fc[0]],
                     &verts_trans[fc[1]],
                     &verts_trans[fc[2]] 
                  };

                  S_vector2 edge_dir = *v[2] - *v[1];
                  S_vector2 p_dir = f_pt - *v[0];
                              //get intersection of those 2 lines
                  float u1, u2;

                              //todo: rewrite to 2D intersection test (performance)
                  bool b = LineIntersection(S_vector(v[0]->x, v[0]->y, 0),
                     S_vector(p_dir.x, p_dir.y, 0),
                     S_vector(v[1]->x, v[1]->y, 0), 
                     S_vector(edge_dir.x, edge_dir.y, 0),
                     u1, u2);
                  if(!b || fabs(u1)<MRG_ZERO){
                              //shuffle a bit and try again
                     swap(v[0], v[1]); swap(v[0], v[2]);
                     swap(fc[0], fc[1]); swap(fc[0], fc[2]);

                     edge_dir = *v[2] - *v[1];
                     p_dir = f_pt - *v[0];

                     bool b = LineIntersection(S_vector(v[0]->x, v[0]->y, 0),
                        S_vector(p_dir.x, p_dir.y, 0),
                        S_vector(v[1]->x, v[1]->y, 0), 
                        S_vector(edge_dir.x, edge_dir.y, 0),
                        u1, u2);
                     if(!b || fabs(u1)<MRG_ZERO){
                              //shuffle a bit and try again
                        swap(v[0], v[1]); swap(v[0], v[2]);
                        swap(fc[0], fc[1]); swap(fc[0], fc[2]);

                        edge_dir = *v[2] - *v[1];
                        p_dir = f_pt - *v[0];

                        b = LineIntersection(S_vector(v[0]->x, v[0]->y, 0),
                           S_vector(p_dir.x, p_dir.y, 0),
                           S_vector(v[1]->x, v[1]->y, 0), 
                           S_vector(edge_dir.x, edge_dir.y, 0),
                           u1, u2);
                        if(!b || fabs(u1)<MRG_ZERO){
                              //invalid face
                              //set lumel to special color (green)
                           lumel->x = 0; lumel->y = 2.0f; lumel->z = 0;
                                          //report problem to user
                           if(!inv_face_reported){
                              inv_face_reported = true;
                              if(cb_proc)
                                 (*cb_proc)(CBM_ERROR, (dword)"invalid face", 0, cb_context);
                           }
                           is_valid_lumel = false;
                           //continue;
                           break;
                        }
                     }
                  }

                  {           
                              //get lumel's point, in world coords,
                              // and its normal
                     S_vector edge_xyz = verts[fc[1]].xyz +
                        (verts[fc[2]].xyz - verts[fc[1]].xyz) * u2;
                     pt_xyz = verts[fc[0]].xyz + (edge_xyz - verts[fc[0]].xyz) / u1;
                     pt_xyz *= matrix;

                     S_vector edge_normal = verts[fc[1]].normal * (1.0f-u2) +
                        verts[fc[2]].normal * u2;
                     float u1a = 1.0f / u1;
                     pt_normal = verts[fc[0]].normal * (1.0f-u1a) + edge_normal * u1a;
                     pt_normal = pt_normal.RotateByMatrix(matrix);
                     pt_normal.Normalize();
                  }

                              //get collision point, usually the same as lumel's point,
                              // but it's different for out-of-face lumels
                  if(in_face_i == -1){
                              //collisions in this case must be handled
                              // from the closest edge point, to avoid
                              // 'out of area' collision testing
                     assert(out_face_i!=-1);
                     
                     const I3D_triface &fc = lm_faces[out_face_i].fc;
                     word vi0 = fc[out_edge_i], vi1 = fc[(out_edge_i+1)%3];
                              //get edge's dir
                     S_vector2 edge_dir(verts_trans[vi1] - verts_trans[vi0]);

                              //get direct point on edge
                     float u = edge_dir.Dot(edge_dir);
                     if(u>MRG_ZERO){
                        u = edge_dir.Dot(f_pt-verts_trans[vi0]) / u;

                              //put slightly towards triangle
                              // - helps to collision testing to avoid missing something
                              // this computation is in 3D
                              //clamp to line 
                        u = Max(INSIDE_PUSH_RATIO*2.0f, Min(1.0f-INSIDE_PUSH_RATIO*2.0f, u));
                        col_pt_xyz = verts[vi0].xyz + (verts[vi1].xyz - verts[vi0].xyz) * u;
                        col_pt_xyz = col_pt_xyz * matrix;
                        {
                           const S_vector *vp[3] = {&verts[fc[0]].xyz, &verts[fc[1]].xyz, &verts[fc[2]].xyz };
                           S_vector face_normal;
                           face_normal.GetNormal(*vp[0], *vp[1], *vp[2]);

                           S_vector edge_normal;
                           edge_normal.GetNormal(verts[vi0].xyz, verts[vi1].xyz, verts[vi0].xyz-face_normal);
                           edge_normal = edge_normal.RotateByMatrix(matrix);
                           float mag = edge_normal.Magnitude();
                           if(mag>MRG_ZERO)
                              col_pt_xyz += edge_normal * INSIDE_PUSH_RATIO / mag;
                           //{ S_vector v[2] = {pt_xyz, pt_xyz+pt_normal }; drv->DEBUG_CODE(0, (dword)v); }
                        }
                     }else{//failcase, but also rather ok (anyway it shouldn't happen)
                        col_pt_xyz = pt_xyz;
                     }
                  }else
                     col_pt_xyz = pt_xyz;
                           //debug - breakpoint on specified lumel
                  /*
                  if(lmrc_i==0 && x==6 && y==2){
                     (*lumel)[2] = 2.0f;
                     { S_vector v[2] = {pt_xyz, pt_xyz+pt_normal }; drv->DEBUG_CODE(0, (dword)v); }
                  }
                  */

                              //light computation goes from collision point,
                              // not recommended - just test
                  //pt_xyz = col_pt_xyz;

#ifdef DEBUG_DRAW_LUMEL_NORMAL
                  { S_vector v[3] = {pt_xyz, pt_xyz+pt_normal, S_vector(0, 0, 1)};
                     drv->DEBUG_CODE(20, (dword)v); 
                  }
#endif
                  if(mc2_img){
                     assert(!IsAbsMrgZero(u1));
                                    //determine uv coordinates of map channel 2
                                    // for the lumel
                     S_vector2 pp = *v[0] + p_dir * u1;
                                    //get uv on edge point
                     float e_len = edge_dir.Sum();
                     assert(!IsAbsMrgZero(e_len));

                     const I3D_triface &uv_fc = mc2->uv_faces[lm_faces[face_i].face_index];

                     u2 = (pp - *v[1]).Sum() / e_len;
                     I3D_text_coor tex;
                     tex.u = mc2->uv_verts[uv_fc[1]].u + (mc2->uv_verts[uv_fc[2]].u - mc2->uv_verts[uv_fc[1]].u) * u2;
                     tex.v = mc2->uv_verts[uv_fc[1]].v + (mc2->uv_verts[uv_fc[2]].v - mc2->uv_verts[uv_fc[1]].v) * u2;

                     u1 = 1.0f / u1;
                     tex.u = mc2->uv_verts[uv_fc[0]].u + (tex.u - mc2->uv_verts[uv_fc[0]].u) * u1;
                     tex.v = mc2->uv_verts[uv_fc[0]].v + (tex.v - mc2->uv_verts[uv_fc[0]].v) * u1;

                     tex.u = fmod(tex.u, 1.0f);
                     tex.v = fmod(tex.v, 1.0f);
                     if(tex.u<0.0f) tex.u += 1.0f;
                     if(tex.v<0.0f) tex.v += 1.0f;

                                    //map UV to image's pixel
                     dword sx = mc2_img->SizeX(), sy = mc2_img->SizeY();
                     int x = Min((int)(sx * tex.u), (int)sx-1);
                     int y = Min((int)(sy * tex.v), (int)sy-1);
                     x = x;
                     y = y;

                     struct S_rgb{
                        byte b, g, r;
                     } *p_rgb;
                     dword pitch;
                     byte bpp;
                     bool b = mc2_img->Lock((void**)&p_rgb, &pitch, &bpp);
                     assert(b);
                     if(b){
                        p_rgb = (S_rgb*)(((byte*)p_rgb) + pitch * y + x * sizeof(S_rgb));
                        mc2_lumel->x += (float)p_rgb->b / 255.0f;
                        mc2_lumel->y += (float)p_rgb->g / 255.0f;
                        mc2_lumel->z += (float)p_rgb->r / 255.0f;
                        mc2_img->Unlock();
                     }
                  }
               }

                              //compute all lights
               for(int i=rc_lights.size(); i--; ){
                  PI3D_light lp = rc_lights[i];
                  const S_matrix &m_light = lp->GetMatrixDirect();
                  const S_vector &light_pos = m_light(3);
                  S_vector ldir = light_pos - pt_xyz;
                  dword col_test_flags = 0;

                  float A = lp->power;

                  switch(lp->I3D_light::GetLightType()){
                  case I3DLIGHT_DIRECTIONAL:
                     {
                        float dir_cos = -(lp->normalized_dir.Dot(pt_normal));
                        if(CHECK_ZERO_LESS(dir_cos))
                           continue;
                        A *= dir_cos;
                        ldir = -lp->normalized_dir * ldir.Magnitude();
                        col_test_flags |= I3DGEOCOL_RAY;
                     }
                     break;

/*
                  case I3DLIGHT_SPOT:
                     {
                              //compute angle, fast reject out-going rays
                        float Rd = ldir.Dot(pt_normal);
                        if(Rd<0.0f) continue;
                              //get distance
                        float ldist_2 = ldir.Dot(ldir);
                        if(ldist_2>=lp->range_f_scaled_2) continue;
                        float ldist = sqrt(ldist_2);
                              //compute attenuation
                        if(ldist>lp->range_n_scaled) A *= (lp->range_f_scaled - ldist) / lp->range_delta_scaled;
                              //cone
                        float r_dist = 1.0f/ldist;
                        {
                           float a = -lp->world_dir.Dot(ldir) * r_dist;
                           if(a<=lp->outer_cos) continue;
                           if(a<lp->inner_cos) A *= (a - lp->outer_cos) / lp->cone_cos_delta;
                        }
                        A *= Rd * r_dist;

                        ldir = light_pos - col_pt_xyz;
                     }
                     break;
*/

                  case I3DLIGHT_POINT:
                     {
                              //compute angle, fast reject out-going rays
                        float Rd = ldir.Dot(pt_normal);
                        if(Rd<0.0f) continue;
                              //get distance
                        float ldist_2 = ldir.Dot(ldir);
                        if(ldist_2>=lp->range_f_scaled_2) continue;
                        float ldist = sqrt(ldist_2);
                              //compute attenuation
                        if(ldist>lp->range_n_scaled) A *= (lp->range_f_scaled - ldist) / lp->range_delta_scaled;
                        A *= Rd / ldist;

                        ldir = light_pos - col_pt_xyz;
                     }
                     break;

                  case I3DLIGHT_POINTAMBIENT:
                     {
                              //compute angle, fast reject out-going rays
                        //float Rd = ldir.Dot(pt_normal);
                        //if(Rd<0.0f) continue;
                              //get distance
                        float ldist_2 = ldir.Dot(ldir);
                        if(ldist_2>=lp->range_f_scaled_2) continue;
                        float ldist = sqrt(ldist_2);
                              //compute attenuation
                        if(ldist>lp->range_n_scaled) A *= (lp->range_f_scaled - ldist) / lp->range_delta_scaled;
                        //A *= Rd / ldist;

                        ldir = light_pos - col_pt_xyz;
                     }
                     break;
                  }
                              //optimization: don't further process rays with minimal intensity
                  if(IsAbsMrgZero(A))
                     continue;
                              //check collisions
                              // note: PointAmbient lights do not check collisions
                  S_vector color = lp->color;
                  swap(color.x, color.z);
                  if(flags&(I3D_LIGHT_GEO_COLLISIONS|I3D_LIGHT_VOL_COLLISIONS) &&
                     lp->I3D_light::GetLightType() != I3DLIGHT_POINTAMBIENT){

                     float len = ldir.Magnitude();
                     float dist;
                     bool b;
                           //put a bit in front of face, 
                           // don't allow colliding with itself
                     S_vector x_pos1 = col_pt_xyz + pl.normal*INSIDE_PUSH_RATIO;
#ifdef DEBUG_DRAW_RAYS
                     //{ S_vector v[2] = {x_pos1, x_pos1+ldir }; drv->DEBUG_CODE(0, (dword)v); }
#endif
                     if(flags&I3D_LIGHT_GEO_COLLISIONS){
                        struct S_hlp{
                           S_vector opacity;
                           S_vector self_add;
                           bool collided;
                           bool collision_or;
                           float collision_dist;
                           bool texture_alpha;
                           bool zero_opacity;
                           PI3D_driver drv;
                           bool compress_err;

                           static bool I3DAPI cbcol(const S_vector &collision_normal,
                              float collision_dist, PI3D_frame frm, void *context, void*){

                              PI3D_frame frm1 = frm;
                              do{
                                 if(frm1->GetFrameFlags()&I3D_FRMF_NOSHADOW)
                                    return false;
                              }while(frm1 = frm1->GetParent1());
                              return true;
                           }
                           static bool I3DAPI cbTexel(const S_vector &collision_normal,
                              float collision_dist, PI3D_frame frm, void *context, void *special){

                              PI3D_frame frm1 = frm;
                              do{
                                 if(frm1->GetFrameFlags()&I3D_FRMF_NOSHADOW)
                                    return false;
                              }while(frm1 = frm1->GetParent1());

         
                              S_hlp *hp = (S_hlp*)context;
                              if(hp->zero_opacity)
                                 return true;
                                                //check texel of texture
                              I3D_texel_collision *tc = (I3D_texel_collision*)special;

                              S_vector self_color;
                              if(tc->mat){
                                 self_color = tc->mat->GetEmmisive();
                              }

                              if(!IsAbsMrgZero(tc->mat->GetAlpha()) ||
                                 tc->tb->GetTxtFlags()&(TXTF_ALPHA|TXTF_SELF_ALPHA)){
                                                //get color of texel
                                 int x = tc->u * tc->tb->SizeX1();
                                 int y = tc->v * tc->tb->SizeY1();
                                 const IImage *img = tc->tb->GetSysmemImage();
                                 if(!img)
                                    return true;
                                 void *area;
                                 dword pitch;
                                 byte bpp;

                                 if(img->GetPixelFormat()->flags&PIXELFORMAT_COMPRESS){
                                    hp->compress_err = true;
                                    //AssertMsg(0, "Can't make lightmaps with compressed textures");
                                    hp->collided = true;
                                    hp->collision_or = true;
                                    return true;
                                 }
                                 if(((LPIImage)img)->Lock(&area, &pitch, &bpp)){
                                    dword texel;
                                    byte *bp = ((byte*)area) + pitch*y + bpp*x;
                                    switch(bpp){
                                    case 1: texel = *bp; break;
                                    case 2: texel = *(word*)bp; break;
                                    case 3:
                                       *(word*)&texel = *(word*)bp;
                                       *(((byte*)texel)+2) = *(((byte*)bp)+2);
                                       break;
                                    case 4: texel = *(dword*)bp; break;
                                    }
                                    const S_pixelformat *pf = img->GetPixelFormat();
                                    LPC_rgb_conversion rgb_conv;
                                    hp->drv->GetGraphInterface()->CreateRGBConv(&rgb_conv);
                                    byte r, g, b, a;
                                    rgb_conv->InverseRGB(texel, &r, &g, &b, &a, pf, (byte*)img->GetPal());
                                    rgb_conv->Release();
                                    ((LPIImage)img)->Unlock();
                                                         //get color of texel
                                    S_vector color = S_vector(r/255.0f, g/255.0f, b/255.0f);
                                                         //get alpha of texel
                                    float alpha;
                                    if(tc->tb->GetTxtFlags()&TXTF_SELF_ALPHA){
                                       alpha = color.Dot(color) / 3.0f;
                                    }else{
                                       int num_alpha_bits = CountBits(pf->a_mask);
                                       if(num_alpha_bits){
                                          dword max_alpha = 0xff - ((1<<(8-num_alpha_bits)) - 1);
                                          alpha = (float)a / (float)max_alpha;
                                       }else
                                          alpha = 1.0f;
                                    }
                                    if(tc->mat)
                                       alpha *= tc->mat->GetAlpha();

                                    hp->opacity *= (1.0f - alpha);
                                    hp->opacity += (hp->opacity * color * alpha);
                                    if(hp->opacity.IsNull())
                                       hp->zero_opacity = true;

                                    hp->collided = true;
                                    hp->collision_or = true;
                                    hp->texture_alpha = true;
                                    hp->collision_dist = collision_dist;
                                 }
                              }
                              return true;
                           }
                        } hlp;
                        hlp.opacity = S_vector(1.0f, 1.0f, 1.0f);
                        hlp.self_add.Zero();
                        hlp.collision_or = false;
                        hlp.collided = false;
                        hlp.zero_opacity = false;
                        hlp.compress_err = false;
                        hlp.drv = drv;

                        b = false;
                        while(true){
                           hlp.collided = false;
                           hlp.texture_alpha = false;
                           b = scene->TestColGeoInternal(x_pos1, ldir,
                              col_test_flags | I3DGEOCOL_FACE | I3DGEOCOL_TEXTURE | I3DGEOCOL_FORCE2SIDE,
                              NULL, NULL, S_hlp::cbcol, &hlp, S_hlp::cbTexel);
                           if(!hlp.collided)
                              break;
                           if(hlp.zero_opacity)
                              break;
                           dist = hlp.collision_dist + .01f;
                           if(len<dist)
                              break;
                           S_vector flied = ldir * (dist/len);
                           x_pos1 += flied;
                           ldir -= flied;
                           if(IsAbsMrgZero(ldir.Magnitude()))
                              break;
                        }
                        if(hlp.collision_or){
                           color *= hlp.opacity;
                           (*lumel) += hlp.self_add;
                        }
                        if(hlp.compress_err)
                           compress_err = true;
                     }else{
                        I3D_collision_data cd(x_pos1, ldir, 0);
                        b = scene->TestCollision(cd);
                        if(b)
                           dist = cd.GetClosestHit();
                     }

#ifdef DEBUG_DRAW_RAYS
                     { S_vector v[2] = {x_pos1,
                        x_pos1 + (!b ? ldir : (ldir*dist/ldir.Magnitude())) }; drv->DEBUG_CODE(0, (dword)v); }
#endif

                     if(b){
                        //{ S_vector v[2] = {x_pos, x_pos1+ldir*dist/ldir.Magnitude() }; drv->DEBUG_CODE(0, (dword)v); }
#ifdef FADE_SHADOWS
                        //float f = dist/(len*.6666f);
                              //collision detected, attenuate light amount by collision distance
                        float f = dist/(len*.8f);
                        f = Min(1.0f, f);
                        A *= f;
#else
                        continue;
#endif
                     }
                  }
                  (*lumel) += color * A;
               }
               }
               }
               if(is_valid_lumel){
                  *lumel *= r_num_aa_levels;
                  *lumel += lumel_base;
                  *lumel *= mat_diffuse;

                  if(mc2_img)
                     *mc2_lumel *= r_num_aa_levels;
               }
            }                 //row loop
                              //time for progress indicator
            {
               if(flags&I3D_LIGHT_ALLOW_CANCEL){
                  int curr_time = igraph->ReadTimer();
                  if((curr_time-last_cancel_check_time) >= MIN_CANCEL_CHECK_DELAY){
                     last_cancel_check_time = curr_time;
                              //check cancellation
                     char k = igraph->ReadKey(true);
                     if(k==K_ESC){
                        ir = I3DERR_CANCELED;
                        break;
                     }
                  }
               }
               progress_lumel_count += lm_rect->size_x;
               if((progress_lumel_count-last_progress_lumel) >= MIN_PROGRESS_LUMELS){
                  last_progress_lumel = progress_lumel_count;

                  int curr_time = igraph->ReadTimer();
                  if((curr_time-last_progress_time) >= MIN_PROGRESS_DELAY){
                     last_progress_time = curr_time;
                              //call-back
                     if(cb_proc){
                        float f = (float)progress_lumel_count / (float)num_lumels;
                        (*cb_proc)(CBM_PROGRESS, FloatAsInt(f), progress_state, cb_context);
                        if(!progress_state) progress_state = 1;
                     }
                  }
               }
            }
         }                    //line loop
         mesh->vertex_buffer.Unlock(); verts = NULL;

         if(compress_err)
            if(cb_proc)
               (*cb_proc)(CBM_ERROR, (dword)"cannot build lightmap with compressed textures", 0, cb_context);

         if(I3D_FAIL(ir)){
            delete[] lumel_map;
            delete[] mc2_map;
            break;
            //goto break_opp;
         }
         {
                              //blur light-map
            if(flags&I3D_LIGHT_BLUR){
               int sx = lm_rect->size_x, sy = lm_rect->size_y;
               S_vector *v_map1 = new S_vector[sx*sy];

                              //make blur by 3x3 matrix:
                              // 1, 2, 1,
                              // 2, 4, 2,
                              // 1, 2, 1
               static const float blur_mat[3][3] = {
                  1, 2, 1,
                  2, 4, 2,
                  1, 2, 1
               };
               const float r_blur_sum = 1.0f / 16.0f;

               S_vector *lumel1 = v_map1;

               for(y=0; y<sy; y++){
                  for(x=0; x<sx; x++, ++lumel1){
                     (*lumel1).Zero();
                     for(int yy = -1; yy<2; yy++){
                        for(int xx = -1; xx<2; xx++){
                           S_vector *src = lumel_map +
                              Max(0, Min((int)sy-1, y+yy)) * sx +
                              Max(0, Min((int)sx-1, x+xx));
                           float f = blur_mat[yy+1][xx+1];
                           (*lumel1)[0] += src->x * f;
                           (*lumel1)[1] += src->y * f;
                           (*lumel1)[2] += src->z * f;
                        }
                     }
                     *lumel1 *= r_blur_sum;
                  }
               }
               delete[] lumel_map;
               lumel_map = v_map1;
            }

            if(mc2_img){
                           //mix map-channel2-map with light-map
               S_vector *lmap = lumel_map, *mc2 = mc2_map;
               for(i=size_x*size_y; i--; ++lmap, ++mc2){
                  *lmap *= *mc2;
                  *lmap *= 2.0f;
               }
            }
                              //final step - convert to RGB surface
            S_lmap_pixel *dstp = lm_rect->dst_map;

#if defined _MSC_VER && defined USE_ASM && 1
            //int sx = lmr->data_lods[dlod_i].size_x, sy = lmr->data_lods[dlod_i].size_y;
            int sx = lm_rect->size_x, sy = lm_rect->size_y;

            __asm{
               push ecx
                              //ecx = counter
                              //esi = src
                              //edi = dst
                              //ST(0) = unit_light
               mov esi, lumel_map
               mov edi, dstp
               fld unit_light
               mov ecx, sx
               imul ecx, sy
               push eax
            lx:
                                 //Red
               fld dword ptr[esi+0]
               fmul st, st(1)
               fistp dword ptr[esp]
               mov eax, [esp]
               test eax, 0xffffff00
               jz ok1
               test eax, eax
               jns ok2
               xor eax, eax
               jmp ok1
            ok2:
               mov al, 0xff
            ok1:
               mov [edi+0],al
                                 //Green
               fld dword ptr[esi+4]
               fmul st, st(1)
               fistp dword ptr[esp]
               mov eax, [esp]
               test eax, 0xffffff00
               jz ok3
               test eax, eax
               jns ok4
               xor eax, eax
               jmp ok3
            ok4:
               mov al, 0xff
            ok3:
               mov [edi+1],al
                                 //Red
               fld dword ptr[esi+8]
               fmul st, st(1)
               fistp dword ptr[esp]
               mov eax, [esp]
               test eax, 0xffffff00
               jz ok5
               test eax, eax
               jns ok6
               xor eax, eax
               jmp ok5
            ok6:
               mov al, 0xff
            ok5:
               mov [edi+2],al

               add esi, size S_vector
               add edi, size S_lmap_pixel

               dec ecx
               jnz lx
                                 //clear FPU
               fstp dword ptr[esp]
               pop eax

               pop ecx
            }
#else //_MSC_VER
            int size_x = lmr->data_lods[dlod_i].size_x, size_y = lmr->data_lods[dlod_i].size_y;
            S_vector *src = lumel_map;
            for(y=0; y<size_y; y++)
            for(x=0; x<size_x; x++, ++dstp, ++src){
               dstp->b = Max(0, Min(255, FloatToInt((*src)[0] * unit_light)));
               dstp->g = Max(0, Min(255, FloatToInt((*src)[1] * unit_light)));
               dstp->r = Max(0, Min(255, FloatToInt((*src)[2] * unit_light)));
            }
#endif   //!_MSC_VER
         }

         delete[] lumel_map;
         delete[] mc2_map;
      }
      if(I3D_FAIL(ir))
         break;
      lm_rect->dirty = true;
   }
   any_rect_dirty = true;

   if(I3D_FAIL(ir))
      Destruct();

   if(cb_proc){
      static const float f = 1.0f;
      (*cb_proc)(CBM_PROGRESS, FloatAsInt(f), 2, cb_context);
   }
   return ir;
}

//----------------------------

void I3D_lit_object::SetMesh(PI3D_mesh_base mb){

   mesh = NULL;
   if(mb){
      mesh = new I3D_mesh_lmap(drv);
      mesh->Clone(mb);
      mesh->Release();
   }
}

//----------------------------

void I3D_lit_object::Update(){

   if(frm_flags&FRMFLAGS_UPDATE_NEEDED){
      I3D_visual::Update();
   }
}

//----------------------------

bool I3D_lit_object::ComputeBounds(){

   if(!mesh){
      bound.bound_local.bbox.Invalidate();
      bound.bound_local.bsphere.pos.Zero();
      bound.bound_local.bsphere.radius = 0.0f;
      return true;
   }
   mesh->ComputeBoundingVolume();
   bound.bound_local = mesh->GetBoundingVolume();
                              //expand bound-box
   //bound.bound_local.bbox.Expand(bound.bbox_local_full);

   S_vector bbox_diag_half = (bound.bound_local.bbox.max - bound.bound_local.bbox.min) * .5f;
   bound.bound_local.bsphere.pos = (bound.bound_local.bbox.min + bbox_diag_half);
   bound.bound_local.bsphere.radius = bbox_diag_half.Magnitude();

   vis_flags |= VISF_BOUNDS_VALID;
   frm_flags &= ~FRMFLAGS_BSPHERE_TRANS_VALID;

   return true;
}

//----------------------------

I3D_RESULT I3D_lit_object::SetProperty(dword index, dword value){

   switch(index){
   case 0:
      {
         float f = IntAsFloat(value);
         if(f < MIN_RESOLUTION)
            return I3DERR_INVALIDPARAMS;
         curr_resolution = f;
      }
      break;
   default: return I3DERR_INVALIDPARAMS;
   }
   return I3D_OK;
}

//----------------------------

dword I3D_lit_object::GetProperty(dword index) const{

   switch(index){
   case 0: return FloatAsInt(curr_resolution);
   }
   return 0xcdcdcdcd;
}

//----------------------------

void I3D_lit_object::UploadLMaps(){

                              //if lm_alloc_block dirty (texture released), delete it
   if(lm_alloc_block && (!lm_alloc_block->lms || (lm_alloc_block->lms->txt->GetTxtFlags()&TXTF_UPLOAD_NEEDED))){
      drv->LMFree(lm_alloc_block);
      lm_alloc_block = NULL;
      for(int i=lm_rects.size(); i--; )
         lm_rects[i]->dirty = true;
      any_rect_dirty = true;
   }
                              //if lm_alloc_block not allocated, do it now
   if(!lm_alloc_block){
      while(true){
         vector<I3D_rectangle> v;
         v.reserve(lm_rects.size());

         for(int i=0; i<lm_rects.size(); i++){
            const S_lmap_rectangle *lm_rect = lm_rects[i];

            v.push_back(I3D_rectangle());
            I3D_rectangle &rc = v.back();
            rc.r = lm_rect->size_x;
            rc.b = lm_rect->size_y;
         }
         lm_alloc_block = drv->LMAlloc(&v);
         if(!lm_alloc_block){
                           //serious problem - we don't have space for lightmap
            return;
         }
         MapVertices();
         break;
      }
   }

   lm_alloc_block->last_render_time = drv->GetRenderTime();

   if(any_rect_dirty){

      PI3D_texture tp_lmap = lm_alloc_block->lms->txt;

      IDirect3DTexture8 *dst = tp_lmap->GetSysmemImage()->GetDirecXTexture();
                           //update all lm_rects
      for(int i=lm_rects.size(); i--; ){
         S_lmap_rectangle *lm_rect = lm_rects[i];
         if(lm_rect->dirty){
            const I3D_rectangle &rc_dst = lm_alloc_block->rect_list[i].rect;

            D3DLOCKED_RECT lrc;
            HRESULT hr = dst->LockRect(0, &lrc, (LPRECT)&rc_dst, 0);
            CHECK_D3D_RESULT("LockRect", hr);
            if(FAILED(hr))
               continue;

            void *dstp = lrc.pBits;

            {
               drv->rgbconv_lm->Convert(
                  lm_rect->dst_map,
                  dstp,
                  lm_rect->size_x, lm_rect->size_y,
                  lm_rect->size_x*sizeof(S_lmap_pixel),  //srcpitch
                  lrc.Pitch,              //dstpitch
                  sizeof(S_lmap_pixel),   //srcbpp
                  NULL,          //pal
                  ((drv->GetFlags2()&DRVF2_LMDITHER) ? MCONV_DITHER : 0),
                  0);
            }
            hr = dst->UnlockRect(0);
            CHECK_D3D_RESULT("UnlockRect", hr);
            lm_rect->dirty = false;
         }
      }
      tp_lmap->SetDirty();
      any_rect_dirty = false;
   }
}

//----------------------------

static const dword vs_decl[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(2, D3DVSDT_FLOAT3),   //normal
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

static const dword vs_decl1[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(2, D3DVSDT_FLOAT3),   //normal
   D3DVSD_REG(1, D3DVSDT_FLOAT4),   //UV coordinates (set 0 & 1 in one reg)
   D3DVSD_END()
};

//----------------------------

void I3D_lit_object::AddPrimitives(S_preprocess_context &pc){

   if(pc.mode==RV_SHADOW_CASTER)
      return;

   if(!mesh)
      return;

#ifdef DEBUG_REAL_TIME
   MakeLight(ps.scene,
      I3D_LIGHT_GEO_COLLISIONS |
      I3D_LIGHT_BLUR |
      0, NULL, 0);
#endif
   PI3D_camera cam = pc.scene->GetActiveCamera1();

   const vector<C_auto_lod> &auto_lods = mesh->GetAutoLODs();

   int curr_auto_lod = -1;

                              //get LOD's vertex count and face source
   const vector<I3D_face_group> *fgrps;
   int vertex_count;
   if(curr_auto_lod < 0){
      fgrps = &mesh->GetFGroupVector();
      vertex_count = mesh->vertex_buffer.NumVertices();
   }else{
      fgrps = &auto_lods[curr_auto_lod].fgroups;
      vertex_count = auto_lods[curr_auto_lod].vertex_count;
   }
   if(!vertex_count)
      return;

   if(!vertex_buffer.GetD3DVertexBuffer()) 
      PrepareDestVB(mesh);

   {
      bool use_lmapping = (drv->GetFlags()&DRVF_USELMAPPING);
      bool is_lmapping = (lmap_flags&LMAPF_DRAW_LIGHT_MAP);
      if(use_lmapping != is_lmapping){
         lmap_flags &= ~LMAPF_DRAW_LIGHT_MAP;
         vis_flags &= ~VISF_DEST_LIGHT_VALID;
         if(use_lmapping){
            lmap_flags |= LMAPF_DRAW_LIGHT_MAP;
         }else{
                              //release light-map resources
            if(lm_alloc_block){
               drv->LMFree(lm_alloc_block);
               lm_alloc_block = NULL;
               for(int i=lm_rects.size(); i--; )
                  lm_rects[i]->dirty = true;
               any_rect_dirty = true;
            }
         }
      }
                              //upload light-maps into textures
      if((lmap_flags&LMAPF_CONSTRUCTED) && (lmap_flags&LMAPF_DRAW_LIGHT_MAP)){
         UploadLMaps();
      }
   }

   drv->render_stats.vert_trans += vertex_count;

                              //add to sort list
   int num_fg = fgrps->size();

   for(int ii=0; ii<num_fg; ii++){
      const I3D_face_group &fg = (*fgrps)[ii];

      pc.prim_list.push_back(S_render_primitive(curr_auto_lod));
      S_render_primitive &p = pc.prim_list.back();
      p.vis = this;
      p.clip = pc.clip;
      p.sector = pc.curr_sector;
      p.scene = pc.scene;
      p.user = ii;

      bool is_alpha = false;
      dword sort_value;
                              //determine blending mode
      PI3D_material mp = fg.GetMaterial1();
      ++pc.opaque;
      sort_value = PRIM_SORT_OPAQUE;

      if(!is_alpha){
                              //sort by material
#if 1
         sort_value |= (mp->GetSortID()&PRIM_SORT_MAT_MASK)<<PRIM_SORT_MAT_SHIFT;
         if(lm_alloc_block)
            sort_value |= (lm_alloc_block->lms->txt->GetSortID() << 15) & PRIM_SORT_MAT_MASK;
#else
         const S_matrix &m_view = cam->GetViewMatrix1();
         const S_vector &pos = matrix(3);
         float depth = 
            pos[0] * m_view(0, 2) +
            pos[1] * m_view(1, 2) +
            pos[2] * m_view(2, 2) +
                     m_view(3, 2);
         sort_value |= PRIM_SORT_DIST_MASK - (((FloatToInt(depth * 100.0f))&PRIM_SORT_DIST_MASK) << PRIM_SORT_DIST_SHIFT);
#endif
      }else{
                              //sort by distance
         const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
         const S_matrix &m = cam->GetViewMatrix1();
         float depth =
            bsphere.pos[0]*m(0, 2) +
            bsphere.pos[1]*m(1, 2) +
            bsphere.pos[2]*m(2, 2) +
                           m(3, 2);
         sort_value |= ((FloatToInt(depth * 100.0f))&PRIM_SORT_DIST_MASK) << PRIM_SORT_DIST_SHIFT;
      }
      p.sort_value = sort_value;

                              //driver stats
      drv->render_stats.triangle += fg.NumFaces1();
   }
}

//----------------------------

void I3D_lit_object::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   const I3D_face_group *fg;
   dword vertex_count;
   int base_index;

   PI3D_mesh_base mb = mesh;

   if(rp.curr_auto_lod < 0){
      fg = &mesh->GetFGroupVector()[rp.user];
      vertex_count = mesh->vertex_buffer.NumVertices();
      base_index = mb->index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(mb->index_buffer.GetD3DIndexBuffer(),
         vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }else{
      const C_auto_lod &al = mb->GetAutoLODs()[rp.curr_auto_lod];
      fg = &al.fgroups[rp.user];
      vertex_count = al.vertex_count;
      base_index = al.index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(al.index_buffer.GetD3DIndexBuffer(),
         vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }
   bool single_pass_mode = (drv->GetFlags()&DRVF_SINGLE_PASS_MODULATE2X);
   PI3D_material mat = fg->GetMaterial1();

   assert(pc.mode!=RV_SHADOW_CASTER);

   drv->SetClipping(rp.clip);
   drv->EnableNoCull(mat->Is2Sided1());
   drv->SetupBlend(I3DBLEND_OPAQUE);
   PI3D_texture_base tp_base = mat->GetTexture1(MTI_DIFFUSE);

   dword face_count = fg->NumFaces1();

   if(drv->IsDirectTransform()){
      last_auto_lod = rp.curr_auto_lod;
      drv->SetStreamSource(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.GetSizeOfVertex());

      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX;
      I3D_driver::S_vs_shader_entry_in se(vs_decl1);

      if(lmap_flags&LMAPF_DRAW_LIGHT_MAP){
         if(drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD)
            prep_flags |= VSPREP_MAKELIGHTING | VSPREP_LIGHT_LM_DYNAMIC;
      }else
         prep_flags |= VSPREP_MAKELIGHTING;

      if((lmap_flags&(LMAPF_DRAW_LIGHT_MAP|LMAPF_CONSTRUCTED)) != (LMAPF_DRAW_LIGHT_MAP|LMAPF_CONSTRUCTED)){
         se.AddFragment(I3D_driver::VSF_COPY_UV_1TO0);
      }else{
         se.AddFragment(I3D_driver::VSF_COPY_UV_0);
         if((drv->GetFlags2()&DRVF2_DRAWTEXTURES) && mat->GetTexture1()){
            se.AddFragment(I3D_driver::VSF_COPY_UV_1);
         }
      }
      PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);
      drv->EnableFog(vis_flags&VISF_FOG_VALID);

      if(!(lmap_flags&LMAPF_CONSTRUCTED)){
                                 //debugging render, not optimized
         drv->DisableTextureStage(1);
         drv->SetupTextureStage(0, D3DTOP_MODULATE);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TFACTOR);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         drv->SetTextureFactor(0xff00ffff);
         drv->SetTexture1(0, tp_base);
                                 //render primitive
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         return;
      }

      if(!(lmap_flags&LMAPF_DRAW_LIGHT_MAP)){
         drv->DisableTextureStage(1);
                              //debug mode - without lightmap
         if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES)){
                              //use factor (50% gray) instead of texture
            drv->SetTexture1(0, NULL);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
            drv->SetTextureFactor(0xff808080);
            hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count, (base_index + fg->base_index) * 3, fg->num_faces);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
            return;
         }
         drv->SetupTextureStage(0, drv->vertex_light_blend_op);
         drv->SetTexture1(0, tp_base);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
         return;
      }
                              //add diffuse color for additional lighting
      drv->SetupTextureStage(0, (drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD) ? D3DTOP_ADD : D3DTOP_SELECTARG1);

      if(!lm_alloc_block)
         return;

                              //setup lightmap texture into stage 0
      drv->SetTexture1(0, lm_alloc_block->lms->txt);
      drv->EnableAnisotropy(0, false);

      if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES) || !mat->GetTexture1()){
                                 //debug mode - without primary texture
                                 // or no base texture
         drv->DisableTextureStage(1);
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
         return;
      }

      if(single_pass_mode){
         drv->SetTexture1(1, tp_base);
         drv->SetupTextureStage(1, D3DTOP_MODULATE2X);
         drv->DisableTextureStage(2);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
      }else{
                              //render 1st texture
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);

                                 //setup 2nd texture
                                 //absolute failback - should be never executed on most cards
         drv->SetTexture1(0, tp_base);
         drv->EnableAnisotropy(0, true);
         if(drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD)
            drv->SetupTextureStage(0, D3DTOP_SELECTARG1);

                                 //setup blending mode
         dword srcb = drv->GetSrcBlend();
         dword dstb = drv->GetDstBlend();
         bool aen = (drv->GetAlphaMode()&ALPHAMODE_ENABLED);
         if(srcb != D3DBLEND_DESTCOLOR)
            lpDev8->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
         if(dstb != D3DBLEND_SRCCOLOR)
            lpDev8->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_SRCCOLOR);
         if(!aen)
            lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
         if(vis_flags&VISF_FOG_VALID)
            drv->SetFogColor(0x80808080);

         {
            I3D_driver::S_vs_shader_entry_in se(vs_decl1);
            se.AddFragment(I3D_driver::VSF_COPY_UV_1TO0);
            PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, VSPREP_TRANSFORM | VSPREP_FEED_MATRIX);
         }

                                 //render 2nd texture
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
                                 //put state back
         if(srcb != D3DBLEND_DESTCOLOR) lpDev8->SetRenderState(D3DRS_SRCBLEND, srcb);
         if(dstb != D3DBLEND_SRCCOLOR) lpDev8->SetRenderState(D3DRS_DESTBLEND, dstb);
         if(!aen) lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, false);
      }

   }else{

      if(dest_vb_trans_id!=pc.render_view_id){
         dest_vb_trans_id = pc.render_view_id;

         if(last_auto_lod != rp.curr_auto_lod){
            last_auto_lod = rp.curr_auto_lod;
            vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
         }
         dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX;

         I3D_driver::S_vs_shader_entry_in se(vs_decl);
         if(!(vis_flags&VISF_DEST_LIGHT_VALID)){

            if(lmap_flags&LMAPF_DRAW_LIGHT_MAP){
               if(drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD)
               prep_flags |= VSPREP_MAKELIGHTING | VSPREP_LIGHT_LM_DYNAMIC;
            }else
               prep_flags |= VSPREP_MAKELIGHTING;
            vis_flags |= VISF_DEST_LIGHT_VALID;
         }

         PrepareVertexShader(mb->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

         drv->DisableTextureStage(1);

         drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_lmap));
         hr = lpDev8->ProcessVertices(mb->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
      drv->EnableFog(vis_flags&VISF_FOG_VALID);

      drv->SetStreamSource(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.GetSizeOfVertex());
      drv->SetVertexShader(vertex_buffer.GetFVFlags());
   
      if(!(lmap_flags&LMAPF_CONSTRUCTED)){
                                 //debugging render, not optimized
         drv->DisableTextureStage(1);
         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TFACTOR);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         drv->SetupTextureStage(0, D3DTOP_MODULATE);
         drv->SetTextureFactor(0xff00ffff);
         drv->SetTextureCoordIndex(0, 1);
         drv->SetTexture1(0, tp_base);
                                 //render primitive
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);

         hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
         CHECK_D3D_RESULT("SetTextureStageState", hr);
         drv->SetTextureCoordIndex(0, 0);
         return;
      }
   
      if(!(lmap_flags&LMAPF_DRAW_LIGHT_MAP)){
                              //debug mode - without lightmap
         drv->DisableTextureStage(1);

         if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES)){
                              //use factor (50% gray) instead of texture
            drv->SetTexture1(0, NULL);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
            drv->SetTextureFactor(0xff808080);
            hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count, (base_index + fg->base_index) * 3, fg->num_faces);
            hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
            return;
         }
         drv->SetTextureCoordIndex(0, 1);
         drv->SetupTextureStage(0, drv->vertex_light_blend_op);
         drv->SetTexture1(0, tp_base);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
         drv->SetTextureCoordIndex(0, 0);
         return;
      }
                              //add diffuse color for additional lighting
      drv->SetupTextureStage(0, (drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD) ? D3DTOP_ADD : D3DTOP_SELECTARG1);

      if(!lm_alloc_block)
         return;

                                 //setup lightmap texture into stage 0
      drv->SetTexture1(0, lm_alloc_block->lms->txt);
      drv->EnableAnisotropy(0, false);

      if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES) || !mat->GetTexture1()){
                                 //debug mode - without primary texture
                                 // or no base texture
         drv->DisableTextureStage(1);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
         return;
      }

      if(single_pass_mode){
         drv->SetTexture1(1, tp_base);
         drv->SetupTextureStage(1, D3DTOP_MODULATE2X);
         drv->DisableTextureStage(2);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
      }else{
                                 //render 1st texture
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
                                 //setup 2nd texture
         drv->SetTextureCoordIndex(0, 1);
         drv->SetTexture1(0, tp_base);
         drv->EnableAnisotropy(0, true);
         if(drv->GetFlags2()&DRVF2_CAN_TXTOP_ADD)
            drv->SetupTextureStage(0, D3DTOP_SELECTARG1);

                                 //setup blending mode
         dword srcb = drv->GetSrcBlend();
         dword dstb = drv->GetDstBlend();
         bool aen = (drv->GetAlphaMode()&ALPHAMODE_ENABLED);
         if(srcb != D3DBLEND_DESTCOLOR)
            lpDev8->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
         if(dstb != D3DBLEND_SRCCOLOR)
            lpDev8->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_SRCCOLOR);
         if(!aen)
            lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
         if(vis_flags&VISF_FOG_VALID)
            drv->SetFogColor(0x80808080);

                                 //render 2nd texture
         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, face_count);
         CHECK_D3D_RESULT("DrawIP", hr);
                                 //put state back
         drv->SetTextureCoordIndex(0, 0);
         if(srcb != D3DBLEND_DESTCOLOR) lpDev8->SetRenderState(D3DRS_SRCBLEND, srcb);
         if(dstb != D3DBLEND_SRCCOLOR) lpDev8->SetRenderState(D3DRS_DESTBLEND, dstb);
         if(!aen) lpDev8->SetRenderState(D3DRS_ALPHABLENDENABLE, false);
      }
   }
}

//----------------------------

void I3D_lit_object::RenderSolidMesh(PI3D_scene scene, bool debug, const I3D_cylinder *vf){

   I3D_driver::S_vs_shader_entry_in se(vs_decl);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_TEXT_PROJECT);
   if(debug)
      se.AddFragment(I3D_driver::VSF_SIMPLE_DIR_LIGHT);
   drv->SetVertexShader(drv->GetVSHandle(se)->vs_handle);

   I3D_visual::RenderSolidMesh(scene, debug, vf);
}

//----------------------------
//----------------------------

struct S_rect{
   int indx;
   I3D_rectangle rc;
   S_rect(){}
   S_rect(const I3D_rectangle &rc1, int i):
      indx(i),
      rc(rc1)
   {}
   bool operator <(const S_rect &r2) const{
      return ((rc.r*rc.b) < (r2.rc.r*r2.rc.b));
   }
};

//----------------------------

struct S_surf{
   C_LM_surface *surf;
   int sort_value;
   S_surf(){}
   S_surf(C_LM_surface *s1):
      surf(s1),
      sort_value(s1->alloc.GetFreeSize())
   {}
   bool operator <(const S_surf &s2) const{
      return (sort_value < s2.sort_value);
   }
};

//----------------------------

C_LM_alloc_block *I3D_driver::LMAlloc(vector<I3D_rectangle> *rect_list){

   int i;
   int req_size = 0;
   int max_x = 0;
   int max_y = 0;
                              //sort input rectangles by size
   vector<S_rect> rl;
   rl.reserve(rect_list->size());
   for(i=rect_list->size(); i--; ){
      rl.push_back(S_rect((*rect_list)[i], i));
      const S_rect &rc = rl.back();
      req_size += rc.rc.r * rc.rc.b;
      max_x = Max(max_x, (int)rc.rc.r);
      max_y = Max(max_y, (int)rc.rc.b);
   }
   sort(rl.begin(), rl.end());


   C_LM_alloc_block *ab = new C_LM_alloc_block(rect_list->size());

   C_LM_surface *lms = NULL;
   bool alloc_ok = false;

   vector<S_surf> sl;

   if(!lm_textures.size())
      goto alloc_new;

                              //check if requirements not out of possibilities
                              // assume all lmaps are the same size
   if(max_x>lm_textures.back()->txt->SizeX1() || max_y>lm_textures.back()->txt->SizeY1()){
      LM_ALLOC_ERR(this, "LightMap allocate: a block too big, can't allocate");
      return NULL;
   }

                              //Strategy:
                              // 1. search existing LM_surface-s for free space
                              // 2. search existing LM_surface-s, ignoring too old blocks
                              // 3. determine ratio (LM_surfaces:free_vidmem), if underlimit, alloc new
                              // 4. alloc on LM_surface with best (free_space:old_blocks) ratio

                              //0. sort existing surfaces by free space
   sl.reserve(lm_textures.size());
   for(i=lm_textures.size(); i--; ) sl.push_back(S_surf(lm_textures[i]));
   sort(sl.begin(), sl.end());

                              //1. search all existing
#if 0                         //try to fit first in those with smallest free space
   for(i=0; i<sl.size(); i++){
      if(sl[i].sort_value < req_size*2) continue;
#else                         //try to fit first in those with largest free space
   for(i=sl.size(); i--; ){
      if(sl[i].sort_value < req_size*2) break;
#endif
      LM_ALLOC_LOG(this, "lm allocate: trying in existing texture...");
      C_LM_surface *lms1 = sl[i].surf;
                              //try to allocate all surfaces here
      for(int j=rl.size(); j--; ){
         bool b = lms1->alloc.Allocate(rl[j].rc.r, rl[j].rc.b, ab, ab->rect_list[rl[j].indx].rect);
         if(!b) break;
      }
      if(j==-1){              //all surfaces allocated OK, quit
         LM_ALLOC_LOG(this, "alloc OK");
         lms = lms1;
         lms->AddRef();
         alloc_ok = true;
         break;
      }
      LM_ALLOC_LOG(this, "failed to alloc");
      lms1->alloc.Free(ab);   //release surfaces
   }

   if(!lms){
      dword curr_lm_pixels = lm_textures.back()->txt->SizeX1() * lm_textures.back()->txt->SizeY1() * sl.size();
      if(curr_lm_pixels < MAX_LM_PIXELS)
      {                       //3. if we're still in limit with texture memory, go on and allocate new
                              // assume all lmaps are the same size
         int single_size = lm_textures.back()->txt->SizeX1()*lm_textures.back()->txt->SizeY1() *
            lm_textures.back()->txt->GetSysmemImage()->GetPixelFormat()->bytes_per_pixel;
         int current_lm_size = lm_textures.size() * single_size;
         if((current_lm_size+single_size/2) < max_lm_memory)
            goto alloc_new;
      }

      const int optimal_expire_time = 20000;
                              //sort by (free_space:old_blocks) ratio
      if(sl.size()>=1)
      for(i=sl.size(); i--; ){
         C_LM_surface *lms1 = sl[i].surf;

         int sort_value = 0;
         for(int j=0; j<lms1->alloc.NumBlocks(); j++){
            const C_rect_allocator::S_block &bl = lms1->alloc.GetBlock(j);
            int block_size = (bl.rect.r-bl.rect.l) * (bl.rect.b-bl.rect.t);
            int age = render_time - bl.owner_id->last_render_time;
            sort_value += Min(optimal_expire_time, age) * block_size;
         }
         sl[i].sort_value = sl[i].sort_value * (optimal_expire_time/10) + sort_value;
      }
      sort(sl.begin(), sl.end());

      LM_ALLOC_LOG(this, "allocate: choosing lm texture with largest oldest block(s)");
      lms = sl.back().surf;
      lms->AddRef();
   }


alloc_new:
   if(!lms){
      LM_ALLOC_LOG(this, "creating new LM texture and allocating here");
                              //3. alloc new LM surface
      I3D_CREATETEXTURE ct;
      memset(&ct, 0, sizeof(ct));
      ct.flags = TEXTMAP_NOMIPMAP | TEXTMAP_USEPIXELFORMAT | TEXTMAP_HINTDYNAMIC;
      ct.pf = &pf_light_maps;
      ct.size_x = LM_REQUEST_TEXTURE_SIZE;
      ct.size_y = LM_REQUEST_TEXTURE_SIZE;

      PI3D_texture tp;
      I3D_RESULT ir;
      ir = CreateTexture(&ct, &tp);
      assert(I3D_SUCCESS(ir));
      if(GetFlags()&DRVF_DRAWLMTEXTURES)
         FillTextureByWhite(tp);
      lms = new C_LM_surface(tp);
      lm_textures.push_back(lms);

      tp->Release();
   }


   if(!alloc_ok)
   while(true){
                              //try to allocate all rectangles
      for(i=rl.size(); i--; ){
         bool b = lms->alloc.Allocate(rl[i].rc.r, rl[i].rc.b, ab, ab->rect_list[rl[i].indx].rect);
         if(!b) break;
      }
      if(i==-1){
                              //all surfaces allocated OK, quit
         LM_ALLOC_LOG(this, "alloc OK");
         break;
      }
      LM_ALLOC_LOG(this, "failed to alloc, trying next texture");
      lms->alloc.Free(ab);    //release surfaces

      if(lms->alloc.NumBlocks()){
                              //free oldest block
         int max_age = 0;     //don't accept recent blocks
         C_LM_alloc_block *oldest_block = NULL;
         for(int j=0; j<lms->alloc.NumBlocks(); j++){
            const C_rect_allocator::S_block &bl = lms->alloc.GetBlock(j);
            int age = render_time - bl.owner_id->last_render_time;
            if(max_age<age){
               max_age = age;
               oldest_block = bl.owner_id;
            }     
         }
         if(!oldest_block){
            LM_ALLOC_ERR(this, "LightMap: failed to alloc (no oldest block)");
            lms->Release();
            delete ab;
            return NULL;
         }
                              //discard oldest block
         if(GetFlags()&DRVF_DRAWLMTEXTURES){
                              //fill all discarded rects by white
            for(int j=lms->alloc.NumBlocks(); j--; ){
               const C_rect_allocator::S_block &bl = lms->alloc.GetBlock(j);
               if(bl.owner_id==oldest_block){
                  LM_ALLOC_LOG(this, "filling block by white");
                  FillTextureByWhite(lms->txt, &bl.rect);
               }
            }
         }
         LM_ALLOC_LOG(this, "freeing (old) block to gain space");
         lms->alloc.Free(oldest_block);
         oldest_block->lms->Release();
         oldest_block->lms = NULL;
      }else{
         LM_ALLOC_ERR(this, "LightMap: alloc failed, quitting");
         ab->lms = lms;
         LMFree(ab);
         return NULL;
      }
   }
   ab->lms = lms;

   return ab;
}

//----------------------------

void I3D_driver::LMFree(C_LM_alloc_block *ab){

   if(GetFlags()&DRVF_DRAWLMTEXTURES)
   if(ab->lms){               //fill all discarded rects by white
      for(int j=ab->lms->alloc.NumBlocks(); j--; ){
         const C_rect_allocator::S_block &bl = ab->lms->alloc.GetBlock(j);
         if(bl.owner_id==ab){
            LM_ALLOC_LOG(this,
               C_fstring("I3D_driver::LMFree: filling blocks by white (%i, %i, %i, %i)", 
               bl.rect.l, bl.rect.t, bl.rect.r, bl.rect.b));
            FillTextureByWhite(ab->lms->txt, &bl.rect);
         }
      }
   }

   C_LM_surface *lms = ab->lms;
                              //might be NULL, if released by surface manager (LMAlloc)
   if(lms){
                              //free all blocks on this LM texture
      lms->alloc.Free(ab);

      if(!lms->Release()){    //texture is empty, it may be released
         int i = VectorIndex(lm_textures, lms);
         if(i!=-1){              //it might not be in the list because it may be lost
            lm_textures[i] = lm_textures.back();
            lm_textures.pop_back();
         }
      }
   }
   delete ab;
}

//----------------------------

bool I3D_driver::UnregLMTexture(PI3D_texture tp){

   for(int i=lm_textures.size(); i--; ){
      if(lm_textures[i]->txt==tp){
         lm_textures[i] = lm_textures.back(); lm_textures.pop_back();
         return true;
      }
   }
   return false;
}

//----------------------------

void I3D_driver::DrawLMaps(int viewport[4]){
                              //draw all lmaps on screen
   const int NUM_X = 8, NUM_Y = 6;

   //float sx = (float)GetGraphInterface()->Scrn_sx() / NUM_X;
   float sx = (float)(viewport[2] - viewport[0]) / NUM_X;
   //float sy = (float)GetGraphInterface()->Scrn_sy() / NUM_Y;
   float sy = (float)(viewport[3] - viewport[1]) / NUM_Y;

   bool is_filter = (drv_flags&DRVF_LINFILTER);
   if(is_filter) SetState(RS_LINEARFILTER, false);
   SetupBlend(I3DBLEND_OPAQUE);
   SetupTextureStage(0, D3DTOP_SELECTARG1);

   bool is_zb = (drv_flags&DRVF_USEZB);
   if(is_zb) SetState(RS_USEZB, false);

   int i = 0;
   for(int y = 0; y<NUM_Y; y++)
   for(int x = 0; x<NUM_X; x++, i++){
      if(i>=lm_textures.size()) break;

      struct S_vertex{
         S_vector xyz;
         float rhw;
         float u, v;
      } verts[4];
      for(int j=0; j<4; j++){
         verts[j].xyz.z = 0.0f;
         verts[j].rhw = 1.0f;
         verts[j].u = (j&1);
         verts[j].v = (j&2)>>1;
         verts[j].xyz.x = viewport[0] + x * sx;
         verts[j].xyz.y = viewport[1] + y * sy;
         if(j&1) verts[j].xyz.x += sx - 1;
         if(j&2) verts[j].xyz.y += sy - 1;
      }
      SetTexture(lm_textures[i]->txt);

      SetVertexShader(D3DFVF_XYZRHW | D3DFVF_TEX1);
      HRESULT hr;
      hr = lpDev8->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, verts, sizeof(S_vertex));
      CHECK_D3D_RESULT("DrawPrimitiveUP", hr);
   }
   if(is_zb) SetState(RS_USEZB, true);
   if(is_filter) SetState(RS_LINEARFILTER, true);
}

//----------------------------
//----------------------------


