#ifndef __MATERIAL_H
#define __MATERIAL_H

#include <vector>
#include <d3d8.h>
#include <c_string.hpp>


//----------------------------


                              //material
class I3D_material_base{
public:
   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(void,SetName)(const char *cp) = 0;
   I3DMETHOD_(const char*,GetName)() const = 0;

   I3DMETHOD(SetAmbient)(const S_vector&) = 0;
   I3DMETHOD(SetDiffuse)(const S_vector&) = 0;
   I3DMETHOD(SetSpecular)(const S_vector&) = 0;
   I3DMETHOD(SetEmissive)(const S_vector&) = 0;
   I3DMETHOD(SetAlpha)(float) = 0;
   I3DMETHOD(SetTexture)(PI3D_texture_base) = 0;
   I3DMETHOD_(void,Update)() = 0;
   I3DMETHOD(Duplicate)(CPI3D_material) = 0;

   I3DMETHOD(Set2Sided)() = 0;

   I3DMETHOD_(const S_vector&,GetAmbient)() const = 0;
   I3DMETHOD_(const S_vector&,GetDiffuse)() const = 0;
   I3DMETHOD_(const S_vector&,GetSpecular)() const = 0;
   I3DMETHOD_(const S_vector&,GetEmmisive)() const = 0;
   I3DMETHOD_(float,GetAlpha)() const = 0;

   I3DMETHOD_(PI3D_texture,GetTexture)(I3D_MATERIAL_TEXTURE_INDEX) const = 0;
   I3DMETHOD_(bool,IsCkey)() const = 0;
};

//----------------------------

struct I3D_material_data{
   S_vector diffuse;
   float diffuse_alpha;
   S_vector ambient;
   S_vector emissive;
   float power;
   bool two_sided;
};

//----------------------------
                              //enviromental effect
enum E_ENVIROMENTAL_EFFECT{
   EE_NULL,
   EE_BLENDFACTOR,            //linear blend by a factor
   EE_ALPHABLEND,             //linear blend using texture alpha
   EE_HILIGHT,                //modulate
   EE_HILIGHT2X,              //modulate multiplied
   EE_ADD,                    //add
   EE_ALPHABLENDALPHA,        //similar to EE_ALPHABLEND, but using alpha channel also for diffuse texture
   EE_LAST
};

//----------------------------

class I3D_material{

#define MATF_CKEY_ZERO_REF 1  //colorkeyed with zero alpha ref
#define MATF_CKEY_ALPHA    2
#define MATF_TXT_ALPHA     4
#define MATF_SELF_ALPHA    8
#define MATF_DIFFUSE_ALPHA 0x10

   I3D_material_data md;
                              //DirectX structure
   dword sort_id;             //sorting identifier
   dword mat_flags;
   dword alpharef;            //alpha reference value used for alpha-blended items

   friend I3D_driver;        //creation
protected:
   dword ref;
                              //all texture maps
   C_smart_ptr<I3D_texture_base> tp[MTI_LAST];

   float refl_opacity;        //opacity of reflection maps
   float bump_opacity;        //opacity of bump maps

   E_ENVIROMENTAL_EFFECT env_effect;
   PI3D_driver drv;

   I3D_material(PI3D_driver);
   I3D_material(const I3D_material& m);
public:
   C_string name;             //identification

   const I3D_material& operator =(const I3D_material&);
   I3D_material(){}
   ~I3D_material();
                              //fast access functions
   inline const C_string &GetName1() const{ return name; }
   inline bool IsTransl() const{ return mat_flags&(MATF_DIFFUSE_ALPHA|MATF_TXT_ALPHA|MATF_CKEY_ALPHA|MATF_SELF_ALPHA); }
   inline bool IsDiffuseAlpha() const{ return (mat_flags&MATF_DIFFUSE_ALPHA); }
   inline bool IsTextureAlpha() const{ return (mat_flags&MATF_TXT_ALPHA); }

   inline void SetEnvOpacity(float f){ refl_opacity = f; }
   inline float GetEnvOpacity() const{ return refl_opacity; }
   inline void SetBumpOpacity(float f){ bump_opacity = f; }
   inline float GetBumpOpacity() const{ return bump_opacity; }
   inline const I3D_material_data &GetMaterialData() const{ return md; }
   inline bool IsCkeyAlpha1() const{ return (mat_flags&MATF_CKEY_ALPHA); }
   inline bool Is2Sided1() const{ return md.two_sided; }
   inline void SetTexture1(I3D_MATERIAL_TEXTURE_INDEX mti, PI3D_texture_base tp1){
      if(mti==MTI_DIFFUSE)
         SetTexture(tp1);
      else
         tp[mti] = tp1;
   }
   inline PI3D_texture_base GetTexture1(I3D_MATERIAL_TEXTURE_INDEX mti = MTI_DIFFUSE) const{
      return tp[mti];
   }
   inline void SetEnvironmetEffect(E_ENVIROMENTAL_EFFECT ee){ env_effect = ee; }
   inline E_ENVIROMENTAL_EFFECT GetEnvironmetEffect() const{ return env_effect; }
   inline float GetAlpha1() const{ return md.diffuse_alpha; }

   inline dword GetSortID() const{ return sort_id; }
   inline dword GetAlphaRef() const{ return alpharef; }

   inline dword GetMatGlags() const{ return mat_flags; }
   inline void SetMatGlags(dword dw){ mat_flags = dw; }

public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(void,SetName)(const char *cp){ name=cp; }
   I3DMETHOD_(const char*,GetName)() const{ return name; }

   I3DMETHOD(SetAmbient)(const S_vector&);
   I3DMETHOD(SetDiffuse)(const S_vector&);
   I3DMETHOD(SetSpecularColor)(const S_vector&);
   I3DMETHOD(SetSpecularPower)(float);
   I3DMETHOD(SetSpecularAngle)(float);
   I3DMETHOD(SetEmissive)(const S_vector&);
   I3DMETHOD(SetAlpha)(float);
   I3DMETHOD(SetTexture)(PI3D_texture_base);
   I3DMETHOD_(void,Update)(){}
   I3DMETHOD(Duplicate)(CPI3D_material);

   I3DMETHOD(Set2Sided)();

   I3DMETHOD_(const S_vector&,GetAmbient)() const{ return md.ambient; }
   I3DMETHOD_(const S_vector&,GetDiffuse)() const{ return md.diffuse; }
   I3DMETHOD_(const S_vector&,GetSpecularColor)() const{ static S_vector v(0, 0, 0); return v; }
   I3DMETHOD_(float,GetSpecularPower)() const{ return 0.0f; }
   I3DMETHOD_(float,GetSpecularAngle)() const{ return 0.0f; }
   I3DMETHOD_(const S_vector&,GetEmmisive)() const{ return md.emissive; }
   I3DMETHOD_(float,GetAlpha)() const{ return md.diffuse_alpha; }

   I3DMETHOD_(PI3D_texture_base,GetTexture)(I3D_MATERIAL_TEXTURE_INDEX mti) const{
      return GetTexture1(mti);
   }
   I3DMETHOD_(bool,IsCkey)() const{ return (mat_flags&MATF_CKEY_ALPHA); }
};

//----------------------------

#endif
