/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Mesh.cpp
   Content: Mesh classes and mesh utility functions.
--------------------------------------------------------*/

#include "all.h"
#include "mesh.h"
#include <malloc.h>           //alloca
#include "vb_mgr.h"
#include "visual.h"

//----------------------------

                              
const float SAFE_BOUND_THRESH = .001f; //make bounding volume slightly greater, to be safe
const float BOUND_EXPAND = .01f;       //during expansion, add this value to minimize further expands

                              //debugging:
//#define PAINT_VERTEX_NORMALS

//----------------------------
// Compute 2D area of specified rectangle
//----------------------------
float AreaOfTriangle(const S_vector *verts, dword vertex_pitch, const I3D_triface &face){

                              //use equation:
                              // area = triangle_base * line_perp_to_base / 2

   const S_vector *v[3] = {
      (S_vector*)(((byte*)verts) + vertex_pitch * face[0]),
      (S_vector*)(((byte*)verts) + vertex_pitch * face[1]),
      (S_vector*)(((byte*)verts) + vertex_pitch * face[2])
   };

                              //base is between points 0 and 1
   S_vector base = *v[1] - *v[0];
   float base_len = base.Magnitude();

   float f = base.Dot(base);
   if(IsAbsMrgZero(f)) return 0.0f;
   float u = (*v[2] - *v[0]).Dot(base) / f;
   float t = ((*v[0] + base * u) - *v[2]).Magnitude();

   float area = base_len * t * .5f;
   return area;
}

//----------------------------
// Determine which vertices are identical (by examining position) with others
//    in the pool, and create mapping array                                    
//----------------------------
void MakeVertexMapping(const S_vector *vp, dword pitch,
   int numv, word *v_map, float thresh){

   if(!numv) return;
#ifndef _MSC_VER
//#if 1

   for(int i=0; i<numv; i++){
      const S_vector &v1 = *(S_vector*)(((byte*)vp) + pitch * i);
      for(int j=0; j<i; j++){
         const S_vector &v2 = *(S_vector*)(((byte*)vp) + pitch * j);
         if((fabs(v1.x-v2.x) + fabs(v1.y-v2.y) + fabs(v1.z-v2.z)) < thresh)
            break;
      }
      v_map[i] = j;
   }
#else

   __asm{
      push ecx

      mov ecx, 0
      mov esi, vp
      mov ebx, v_map
      push eax
      mov eax, thresh
   l1:
      mov edx, 0
      cmp edx, ecx
      jz sk1
      mov edi, vp
   l2:
      fld dword ptr[esi+0]
      fsub dword ptr[edi+0]
      fabs
      
      fld dword ptr[esi+4]
      fsub dword ptr[edi+4]
      fabs
      faddp st(1), st
      
      fld dword ptr[esi+8]
      fsub dword ptr[edi+8]
      fabs
      faddp st(1), st
      fstp dword ptr[esp]

      cmp eax, [esp]
      ja sk1

      add edi, pitch
      inc edx
      cmp edx, ecx
      jnz l2
   sk1:
      mov [ebx+ecx*2], dx
      add esi, pitch
      inc ecx
      cmp ecx, numv
      jne l1
      pop eax

      pop ecx
   }
#endif
}

//----------------------------
// Check if set of vertices are planar, use specified thresh to determine
//----------------------------
bool CheckIfVerticesPlanar(const S_vector *verts, dword num_verts, float thresh){
                              
   assert(num_verts >= 3);
   float planear_err = 0.0f;
   S_vector normal;
   normal.GetNormal(verts[0], verts[1], verts[2]);
   for(int j=1; j<num_verts; j++){
      S_vector n1;
      n1.GetNormal(verts[j], verts[(j+1)%num_verts], verts[(j+2)%num_verts]);
      planear_err = Max(planear_err, (float)fabs(normal.AngleTo(n1)));
      normal = n1;
   }
   return (planear_err <= thresh);
}

//----------------------------
// Build normals of list of faces
//----------------------------
void BuildFaceNormals(const S_vector *verts,
   const I3D_triface *faces, dword num_faces, S_vector *normals){

   for(int i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      S_vector &n = normals[i];
      n.GetNormal(verts[fc[0]], verts[fc[1]], verts[fc[2]]);
      n.Normalize();
   }
}

//----------------------------
// Find specified face (given by 2 indicies) in list, return index,
// or -1 if not found, if found, fill third index.
//----------------------------
int FindFace(const I3D_triface *faces, dword num_faces, word e1, word e2, word &e3){

   for(int i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      if(fc[0]==e1 && fc[1]==e2){ e3=fc[2]; break; }
      if(fc[1]==e1 && fc[2]==e2){ e3=fc[0]; break; }
      if(fc[2]==e1 && fc[0]==e2){ e3=fc[1]; break; }
   }
   return i;
}
//----------------------------
// Find specified face (given by 3 indicies) in list, return index,
// or -1 if not found.
//----------------------------
int FindFace(const I3D_triface *faces, dword num_faces, const I3D_triface &fc1){

   for(int i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      if(fc[0]==fc1[0] && fc[1]==fc1[1] && fc[2]==fc1[2]) break;
      if(fc[0]==fc1[1] && fc[1]==fc1[2] && fc[2]==fc1[0]) break;
      if(fc[0]==fc1[2] && fc[1]==fc1[0] && fc[2]==fc1[1]) break;
   }
   return i;
}

//----------------------------

void CleanMesh(vector<S_vector> &verts, vector<I3D_triface> &faces){

   int i;
                              //mark used vertices
   vector<word> v_map(verts.size());
   vector<bool> v_use(verts.size(), false);

   for(i=faces.size(); i--; ){
      for(int j=3; j--; )
         v_use[faces[i][j]] = true;
   }
   int new_count = verts.size();
                              //process all verts, remove unused
   for(i=0; i<new_count; i++){
      v_map[i] = i;           //by default vertex stays where it is
      if(!v_use[i]){
                              //vertex from back is moved forward
         --new_count;
         v_map[new_count] = i;
         verts[i] = verts[new_count];
         v_use[i] = v_use[new_count];
         --i;
      }
   }
                              //strip verts
   verts.erase(verts.begin()+new_count, verts.end());
                              //remap faces
   for(i=faces.size(); i--; ){
      for(int j=3; j--; )
         faces[i][j] = v_map[faces[i][j]];
   }
}

//----------------------------

void CleanMesh(vector<S_vector> &verts, vector<I3D_face> &faces){

   int i;
                              //mark used vertices
   vector<word> v_map(verts.size());
   vector<bool> v_use(verts.size(), false);

   for(i=faces.size(); i--; ){
      const I3D_face &fc = faces[i];
      for(int j=fc.num_points; j--; )
         v_use[fc[j]] = true;
   }
   int new_count = verts.size();
                              //process all verts, remove unused
   for(i=0; i<new_count; i++){
      v_map[i] = i;           //by default vertex stays where it is
      if(!v_use[i]){
                              //vertex from back is moved forward
         --new_count;
         v_map[new_count] = i;
         verts[i] = verts[new_count];
         v_use[i] = v_use[new_count];
         --i;
      }
   }
                              //strip verts
   verts.erase(verts.begin()+new_count, verts.end());
                              //remap faces
   for(i=faces.size(); i--; ){
      I3D_face &fc = faces[i];
      for(int j=fc.num_points; j--; )
         fc[j] = v_map[fc[j]];
   }
}

//----------------------------

void CloneVertexBuffer(const void *src1, dword src_fvf, const void *dst1, dword dst_fvf,
   dword num_vertices){

   const byte *src = (const byte*)src1;
   byte *dst = (byte*)dst1;

   dword src_pos_bytes = GetSizeOfVertex(src_fvf&D3DFVF_POSITION_MASK);
   dword dst_pos_bytes = GetSizeOfVertex(dst_fvf&D3DFVF_POSITION_MASK);
   dword pos_bytes = Min(src_pos_bytes, dst_pos_bytes);
   dword src_uv_bytes = ((src_fvf&D3DFVF_TEXCOUNT_MASK)>>D3DFVF_TEXCOUNT_SHIFT) * sizeof(I3D_text_coor);
   dword dst_uv_bytes = ((dst_fvf&D3DFVF_TEXCOUNT_MASK)>>D3DFVF_TEXCOUNT_SHIFT) * sizeof(I3D_text_coor);
   dword uv_coord_bytes = Min(src_uv_bytes, dst_uv_bytes);
   for(int i=num_vertices; i--; ){
                           //copy position
      memcpy(dst, src, pos_bytes);
      src += src_pos_bytes;
      dst += dst_pos_bytes;
                           //now copy others elements of the vertex
      if(dst_fvf&D3DFVF_NORMAL){
         if(src_fvf&D3DFVF_NORMAL)
            *(S_vector*)dst = *(S_vector*)src;
         dst += sizeof(S_vector);
      }
      if(src_fvf&D3DFVF_NORMAL)
         src += sizeof(S_vector);

      if(dst_fvf&D3DFVF_DIFFUSE){
         if(src_fvf&D3DFVF_DIFFUSE)
            *(dword*)dst = *(dword*)src;
         dst += sizeof(dword);
      }
      if(src_fvf&D3DFVF_DIFFUSE)
         src += sizeof(dword);

      if(dst_fvf&D3DFVF_SPECULAR){
         if(src_fvf&D3DFVF_SPECULAR)
            *(dword*)dst = *(dword*)src;
         dst += sizeof(dword);
      }
      if(src_fvf&D3DFVF_SPECULAR)
         src += sizeof(dword);

      memcpy(dst, src, uv_coord_bytes);
      src += src_uv_bytes;
      dst += dst_uv_bytes;
   }
}

//----------------------------

#define EDGE_MODE

//----------------------------
// Auto-LOD helper func: find vertex in pool, which is best for remove
// The returned value is vertex index to be removed 'v_out', and index of vertex
//    which the removed vertex splits to.
// Note: the func considers only faces which are valid - invalidated faces are considerd
//    to be already optimized out.
//----------------------------
bool FindNextPointToRemove(const S_vector *verts, dword vertex_pitch, dword num_verts,
   const I3D_triface *faces, dword num_faces,
   word &v_out, word &v_map_to){

#ifdef EDGE_MODE
   bool found = false;
                           //find smallest edge to be removed
   float best_size_2 = 1e+8f;
   for(int i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      if(!fc.IsValid())
         continue;

                              //process all 3 edges of triangle
      dword w0 = fc[0];
      for(int j=3; j--; ){
         dword w1 = w0;
         w0 = fc[j];
                           //don't process the same vertex twice
         if(w0>w1)
            continue;

                              //get length of the edge
         const S_vector &v0 = *(S_vector*)(((byte*)verts) + w0 * vertex_pitch);
         const S_vector &v1 = *(S_vector*)(((byte*)verts) + w1 * vertex_pitch);
         S_vector dir = (v0 - v1);
         float size_2 = dir.Dot(dir);
         if(best_size_2 > size_2){
                              //smaller edge found, save it
            best_size_2 = size_2;
            found = true;
            v_map_to = w0;
            v_out = w1;
         }
      }
   }
   if(!found)
      return false;
                              //choose vertex which is closer to center
   const S_vector &v0 = *(S_vector*)(((byte*)verts) + v_out * vertex_pitch);
   const S_vector &v1 = *(S_vector*)(((byte*)verts) + v_map_to * vertex_pitch);
   if(v0.Dot(v0) > v1.Dot(v1))
      swap(v_out, v_map_to);
   return true;

#endif   //EDGE_MODE
}

//----------------------------

void CreateLODsGeneric(float min_vertex_ratio, float min_area_ratio,
   float ratio_change_step, float min_dist, float max_dist,
   PI3D_driver drv, 
   const S_vector *verts, dword vstride, dword num_verts,
   I3D_source_index_buffer &index_buffer, const vector<I3D_face_group> &fgroups,
   vector<C_auto_lod> &out_lods, float out_lod_dists[2],
   byte *out_verts, vector<word> &out_v_map){

   int i, j;

                              //collect all faces into single list
                              // we make operations on this (remove, remap, etc)
   vector<I3D_triface> all_faces;
   {
      const I3D_triface *faces = index_buffer.Lock();
      all_faces.assign(faces, faces + index_buffer.NumFaces1());
      index_buffer.Unlock();
   }

   vector<word> v_map(num_verts);
   MakeVertexMapping(verts, vstride, num_verts, v_map.begin());

   vector<word> vertex_shuffle;
   vertex_shuffle.reserve(num_verts);

   typedef vector<vector<I3D_triface> > t_fgroups;
   vector<t_fgroups> lod_fgroups;

   C_auto_lod curr_al(drv);
   t_fgroups curr_fgroups;

   curr_al.ratio = 1.0f;
   curr_al.vertex_count = num_verts;

   for(i=0; i<fgroups.size(); i++){
      curr_fgroups.push_back(vector<I3D_triface>(
         &all_faces[fgroups[i].base_index],
         &all_faces[fgroups[i].base_index + fgroups[i].num_faces]));
   }

   float total_area = 0.0f;
                              //remap faces to orig verts
                              //also compute total area
   for(i=all_faces.size(); i--; ){
      I3D_triface &fc = all_faces[i];
      for(j=0; j<3; j++){
         fc[j] = v_map[fc[j]];
      }
      total_area += AreaOfTriangle(verts, vstride, all_faces[i]);
   }
   float curr_area = total_area;

   vector<bool> vertex_used(num_verts, true);

   while(true){
      word v_out, v_map_to;
      bool b = FindNextPointToRemove(verts, vstride,
         num_verts,
         all_faces.begin(), all_faces.size(),
         v_out, v_map_to);
      if(!b)
         break;

      word v_map_to_orig = v_map_to;
      if(!vertex_used[v_map_to]){
                              //find next vertex mapped to this
         for(i=num_verts; i--; )
         if(vertex_used[i] && v_map[i]==v_map_to){
            v_map_to = i;
            break;
         }
         assert(i!=-1);
      }

      vector<bool> vertex_now_used(num_verts, false);
                              //remove all faces using this
                              // remove vertices not belonging to any face
      for(i=curr_fgroups.size(); i--; ){
         vector<I3D_triface> &faces = curr_fgroups[i];
         for(j=faces.size(); j--; ){
            I3D_triface &fc = faces[j];
                              //check if face contains the edge
            word v0 = v_map[fc[0]];
            for(int k=3; k--; ){
               word v1 = v0;
               v0 = v_map[fc[k]];
               if((v0==v_out && v1==v_map_to_orig) || (v1==v_out && v0==v_map_to_orig))
                  break;
            }
            if(k!=-1){
                              //remove the face - it contains the removed edge
               curr_area -= AreaOfTriangle(verts, vstride, fc);
               fc = faces.back(); faces.pop_back();
            }else{
                              //adjust face if it contains removed vertex
               for(k=3; k--; ){
                  if(v_map[fc[k]] == v_out){
                     curr_area -= AreaOfTriangle(verts, vstride, fc);
                     fc[k] = v_map_to;
                     curr_area += AreaOfTriangle(verts, vstride, fc);
                     vertex_now_used[v_map_to] = true;
                     while(k--){
                        vertex_now_used[fc[k]] = true;
                     }
                     break;
                  }
                  vertex_now_used[fc[k]] = true;
               }
            }
         }
      }
                              //remove vertices lost by this LOD (put to back)
      for(i=num_verts; i--; ){
         if(!vertex_now_used[i] && vertex_used[i]){

            vertex_shuffle.push_back(i);
            --curr_al.vertex_count;
            vertex_used[i] = false;
         }
      }

                              //adjust temp face list - invalidate removed faces,
                              // reorder face indicies,
      for(i=all_faces.size(); i--; ){
         I3D_triface &fc = all_faces[i];
         if(!fc.IsValid())
            continue;
         if(FaceContainsIndicies(fc, v_out, v_map_to_orig)){
            fc.Invalidate();
         }else{
            ReplaceFaceIndex(fc, v_out, v_map_to_orig);
         }
      }

                              //compute ratio of this LOD
      float area_ratio = curr_area / total_area;
      if(area_ratio < min_area_ratio)
         break;
      float ratio = (float)curr_al.vertex_count / (float)num_verts;
      if(ratio < min_vertex_ratio)
         break;
      if(curr_al.ratio - ratio >= ratio_change_step){
         curr_al.ratio = ratio;
                              //save this LOD
         out_lods.push_back(curr_al);
         lod_fgroups.push_back(curr_fgroups);
                              //don't go below required minimum
         if(ratio-ratio_change_step < min_vertex_ratio)
            break;
      }
   }

   {
                              //reorganize vertices, so that first removed are at back
      vector<word> v_remap(num_verts, 0xffff);

                              //put vertices to back, in order in which they were removed
      for(i=0, j=num_verts-1; i<vertex_shuffle.size(); i++, --j){
         int vi = vertex_shuffle[i];
         memcpy(j*vstride + (byte*)out_verts, vi*vstride + (const byte*)verts, vstride);
         assert(v_remap[vi]==0xffff);  //make sure it's not assigned yet
         v_remap[vi] = j;
      }
      for(j=j, i=num_verts; i--; ){
         if(v_remap[i]==0xffff){
            memcpy(j*vstride + (byte*)out_verts, i*vstride + (const byte*)verts, vstride);
            v_remap[i] = j;
            --j;
         }
      }
      assert(j==-1);          //make sure we've put all vertices back into list

                              //adjust faces
      for(int l=out_lods.size(); l--; ){
         C_auto_lod &al = out_lods[l];
         t_fgroups &fg = lod_fgroups[l];
         assert(fg.size() == fgroups.size());

         al.fgroups.reserve(fg.size());
         vector<I3D_triface> new_faces;
         for(i=fg.size(); i--; ){
            vector<I3D_triface> &faces = fg[i];
            for(j=faces.size(); j--; ){
               I3D_triface &fc = faces[j];
               for(int k=0; k<3; k++){
                  fc[k] = v_remap[fc[k]];
                  assert(fc[k] < al.vertex_count); //make sure we don't reference vertex beyond boundary
               }
            }
            //assert(faces.size());   //test - removed fgroup
            if(faces.size()){
               al.fgroups.push_back(I3D_face_group());
               I3D_face_group &al_fg = al.fgroups.back();
               al_fg.base_index = new_faces.size();
               al_fg.num_faces = faces.size();
               al_fg.mat = fgroups[i].mat;
               new_faces.insert(new_faces.end(), faces.begin(), faces.end());
            }
         }
         al.index_buffer.SetFaces(new_faces.begin(), new_faces.size());
      }

                              //adjust main set of faces
      vector<I3D_triface> new_faces(index_buffer.NumFaces1());
      {
         const I3D_triface *faces = index_buffer.Lock();
         for(i=index_buffer.NumFaces1(); i--; ){
            const I3D_triface &src = faces[i];
            I3D_triface &dst = new_faces[i];
            dst[0] = v_remap[src[0]];
            dst[1] = v_remap[src[1]];
            dst[2] = v_remap[src[2]];
         }
         index_buffer.Unlock();
      }
      index_buffer.SetFaces(new_faces);
   }

   out_lod_dists[0] = min_dist;
   out_lod_dists[1] = max_dist;

   {
      out_v_map.assign(num_verts);
                              //re-build vertex-map info
      MakeVertexMapping((const S_vector*)out_verts, vstride,
         num_verts, out_v_map.begin());
                              //determine number of original vertices
      for(int i=out_v_map.size(); i--; ){
         if(v_map[i] >= i){
            ++i;
            break;
         }
      }
      out_v_map.erase(out_v_map.begin(), out_v_map.begin()+i);
   }
}

//----------------------------
                              //sometimes replace the shit above (CreateLODsGeneric)
                              // by this, this is cleaner
/*
class C_lod_build{
                              //iniitialization values:
   const float min_area_ratio;
   const float min_vertex_ratio;
   const float ratio_change_step;
   const I3D_vertex_singlemesh *verts;
   dword num_verts;
   vector<word> v_map;
public:
                              //precomputed constants:
   float total_area;

                              //run-time:
   float curr_area;           //current area, lowered each time face is removed
   int vertex_count;          //current vertex count, lowered each time vertex is removed
   vector<I3D_triface> faces; //working set of faces, mapped to original set of vertices; invalidated as we create LODs
   vector<vector<I3D_triface> > fgroups;  //same as '', but organized in face groups, and invalid faces are removed (empty groups are kept due to assigning material later)
                              
   vector<word> removed_verts;//removed vertices, first are put in front
   vector<word> num_removed_verts_per_lod;   //# of vertices 'out' for each LOD
   vector<bool> vertex_used;  //flag specifying if vertex is still used in mesh
   float curr_quality;

                              //result:
   vector<float> lods_quality;   //quality per each LOD
   vector<vector<vector<I3D_triface> > > lods_fgroups;   //fgroups per each LOD
   vector<word> lods_vertex_count;


   C_lod_build(float mar, float mvr, float rcs):
      min_area_ratio(mar),
      min_vertex_ratio(mvr),
      ratio_change_step(rcs)
   {
   }

   void Initialize(const I3D_vertex_singlemesh *verts1, dword nv1,
      const I3D_triface *faces1, dword num_faces1,
      const vector<I3D_face_group> &fgroups1,
      vector<S_SM_bone_segment> &bone_segments){

      verts = verts1;
      num_verts = nv1;
      v_map.assign(num_verts);
      MakeVertexMapping(&verts->xyz, sizeof(I3D_vertex_singlemesh),
         num_verts, v_map.begin());

      removed_verts.reserve(num_verts);
      vertex_used.assign(num_verts, true);

      int i;
      faces.assign(faces1, faces1 + num_faces1);
      for(i=0; i<fgroups1.size(); i++){
         fgroups.push_back(vector<I3D_triface>(
            &faces[fgroups1[i].base_index],
            &faces[fgroups1[i].base_index + fgroups1[i].num_faces]));
      }
      for(i=faces.size(); i--; ){
         I3D_triface &fc = faces[i];
         fc[0] = v_map[fc[0]];
         fc[1] = v_map[fc[1]];
         fc[2] = v_map[fc[2]];
      }
      curr_quality = 1.0f;
      vertex_count = num_verts;
   }

//----------------------------
// Compute total area of all mesh's faces.
   void ComputeTotalArea(){

      total_area = 0.0f;
      for(int i=faces.size(); i--; ){
         total_area += AreaOfTriangle(&verts->xyz, sizeof(I3D_vertex_singlemesh), faces[i]);
      }
      curr_area = total_area;
   }

//----------------------------
// Return current number of computed LODs.
   inline int NumLODs() const{ return lods_quality.size(); }

//----------------------------
// Splits vertex with other one, removing the edge between them. All faces containing the
// edge are erased. Empty face groups are erased.
   void SplitVertex(word v_out, word v_map_to){

      word v_map_to_orig = v_map_to;
      if(!vertex_used[v_map_to]){
                              //find next used vertex mapped to this
         for(int vi = num_verts; vi--; ){
            if(vertex_used[vi] && v_map[vi]==v_map_to){
               v_map_to = vi;
               break;
            }
         }
         assert(vi!=-1);
      }

      vector<bool> vertex_now_used(num_verts, false);
                              //remove all faces using this edge
                              // remove vertices not belonging to any face
      for(int fgi = fgroups.size(); fgi--; ){
         vector<I3D_triface> &faces = fgroups[fgi];
         for(int fi = faces.size(); fi--; ){
            I3D_triface &fc = faces[fi];
                              //check if face contains the edge
            word v0 = v_map[fc[0]];
            for(int vi = 3; vi--; ){
               word v1 = v0;
               v0 = v_map[fc[vi]];
               if((v0==v_out && v1==v_map_to_orig) || (v1==v_out && v0==v_map_to_orig))
                  break;
            }
            if(vi!=-1){
                              //remove the face - it contains the removed edge
               curr_area -= AreaOfTriangle(&verts->xyz, sizeof(I3D_vertex_singlemesh), fc);
               fc = faces.back(); faces.pop_back();
            }else{
                              //adjust face if it contains removed vertex
               for(vi=3; vi--; ){
                  if(v_map[fc[vi]] == v_out){
                     curr_area -= AreaOfTriangle(&verts->xyz, sizeof(I3D_vertex_singlemesh), fc);
                     fc[vi] = v_map_to;
                     curr_area += AreaOfTriangle(&verts->xyz, sizeof(I3D_vertex_singlemesh), fc);
                     vertex_now_used[v_map_to] = true;
                     while(vi--){
                        vertex_now_used[fc[vi]] = true;
                     }
                     break;
                  }
                  vertex_now_used[fc[vi]] = true;
               }
            }
         }
      }
                              //remove vertices lost by this LOD (put to list)
      for(int vi = num_verts; vi--; ){
         if(!vertex_now_used[vi] && vertex_used[vi]){

            removed_verts.push_back(vi);
            --vertex_count;
            vertex_used[vi] = false;
         }
      }
                              //adjust temp face list - invalidate removed faces,
                              // reorder face indicies,
      for(int fi = faces.size(); fi--; ){
         I3D_triface &fc = faces[fi];
         if(!fc.IsValid())
            continue;
         if(FaceContainsIndicies(fc, v_out, v_map_to_orig)){
            fc.Invalidate();
         }else{
            ReplaceFaceIndex(fc, v_out, v_map_to_orig);
         }
      }
   }

//----------------------------
// Create single level of detail. If returned value is false,
// the last level have been created.
   bool CreateSingleLOD(int num_segments){

      bool ret;
                              //remove vertices until we're satisfied with the LOD
      while(true){
         word v_out, v_map_to;
                              //find vertex to be removed
         bool b = FindNextPointToRemove(&verts->xyz, sizeof(I3D_vertex_singlemesh),
            num_verts,
            faces.begin(), faces.size(),
            v_out, v_map_to);
         if(!b)
            return false;

         SplitVertex(v_out, v_map_to);

                                 //compute ratio of this LOD
         float ratio = (float)vertex_count / (float)num_verts;
         float area_ratio = curr_area / total_area;
         if(area_ratio < min_area_ratio || ratio < min_vertex_ratio){
            //curr_al.ratio = ratio;
            curr_quality = ratio;
            ret = false;
            break;
         }
         if(curr_quality - ratio >= ratio_change_step){
            curr_quality = ratio;
                              //don't go below required minimum
            ret = (ratio-ratio_change_step > min_vertex_ratio);
#if 0
                                 //debug: create only one LOD
            ret = false;
#endif
            break;
         }
#if 0
                              //debug: out after 1st removed vertex
         curr_quality = ratio;
         ret = false;
         break;
#endif
      }
      num_removed_verts_per_lod.push_back(removed_verts.size());
      lods_quality.push_back(curr_quality);
      lods_fgroups.push_back(fgroups);
      return ret;
   }

//----------------------------
// Reorder vertices, so that first removed are at back. This must be done on per-SM-segment
// basis, since each segment is processed separately.
   void ReorderVertices(I3D_vertex_singlemesh *new_verts,
      const I3D_triface *orig_faces, I3D_triface *new_faces);

//----------------------------
// Store resulting LODs info into provided vector.
   void StoreLODs(vector<C_auto_lod> &auto_lods, PI3D_driver drv,
      vector<I3D_face_group> &orig_fgroups) const{

      auto_lods.assign(NumLODs(), C_auto_lod(drv));
      for(int lod_i=NumLODs(); lod_i--; ){
         C_auto_lod &al = auto_lods[lod_i];
         al.ratio = lods_quality[lod_i];
         al.vertex_count = lods_vertex_count[lod_i];

                              //convert faces
         vector<I3D_triface> new_faces;
         al.fgroups.reserve(orig_fgroups.size());
         const vector<vector<I3D_triface> > &lod_fgroups = lods_fgroups[lod_i];
         for(int fgi=orig_fgroups.size(); fgi--; ){
            const vector<I3D_triface> &lod_faces = lod_fgroups[fgi];
            if(lod_faces.size()){
               al.fgroups.push_back(I3D_face_group());
               I3D_face_group &al_fg = al.fgroups.back();
               al_fg.base_index = new_faces.size();
               al_fg.num_faces = lod_faces.size();
               al_fg.mat = orig_fgroups[fgi].mat;
               new_faces.insert(new_faces.end(), lod_faces.begin(), lod_faces.end());
            }
         }
         al.index_buffer.SetFaces(new_faces.begin(), new_faces.size());
      }
   }
};

//----------------------------

void C_lod_build::ReorderVertices(I3D_vertex_singlemesh *new_verts,
   const I3D_triface *orig_faces, I3D_triface *new_faces){

   vector<word> v_remap(num_verts, 0xffff);
   int num_lods = NumLODs();

                              //put removed verts to back
   int dst_i = num_verts;
   for(int i=0; i<removed_verts.size(); i++){
      --dst_i;
      word vi = removed_verts[i];
      new_verts[dst_i] = verts[vi];
      v_remap[vi] = dst_i;
   }
                              //put rest verts
   for(i=num_verts; i--; ){
      if(v_remap[i]==0xffff){
         --dst_i;
         new_verts[dst_i] = verts[i];
         v_remap[i] = dst_i;
      }
   }
   assert(dst_i==0);

                              //assign counters
   lods_vertex_count.assign(num_lods);
   for(i=num_lods; i--; ){
      lods_vertex_count[i] = num_verts - num_removed_verts_per_lod[i];
   }

   {
                              //remap faces
      for(int lod_i = num_lods; lod_i--; ){
         vector<vector<I3D_triface> > &fgroups = lods_fgroups[lod_i];
         for(int fgi = fgroups.size(); fgi--; ){
            vector<I3D_triface> &faces = fgroups[fgi];
            for(int fi = faces.size(); fi--; ){
               I3D_triface &fc = faces[fi];
               for(int vi = 3; vi--; ){
                  fc[vi] = v_remap[fc[vi]];
               }
            }
         }
      }

      for(int fi = faces.size(); fi--; ){
         const I3D_triface &src = orig_faces[fi];
         I3D_triface &dst = new_faces[fi];
         for(int vi = 3; vi--; ){
            dst[vi] = v_remap[src[vi]];
         }
      }
   }
}

//----------------------------

void I3D_singlemesh::CreateLODs(float min_vertex_ratio, float min_area_ratio,
   float ratio_change_step, float min_dist, float max_dist){


   C_lod_build lod_build(min_area_ratio, min_vertex_ratio, ratio_change_step);

   const I3D_vertex_singlemesh *verts = (const I3D_vertex_singlemesh*)vertex_buffer.Lock();
   dword num_verts = NumVertices1();

   {
      const I3D_triface *faces = index_buffer.Lock();
      lod_build.Initialize(verts, num_verts, faces, NumFaces1(), fgroups, bone_segments);
      index_buffer.Unlock();
   }
   lod_build.ComputeTotalArea();

                              //create all LODs
   while(lod_build.CreateSingleLOD(bone_segments.size()));

   vector<I3D_vertex_singlemesh> new_verts(num_verts);
   {
      vector<I3D_triface> new_faces(NumFaces1());
      const I3D_triface *faces = index_buffer.Lock();
      lod_build.ReorderVertices(new_verts.begin(), faces, new_faces.begin());
      index_buffer.Unlock();
      index_buffer.SetFaces(new_faces);
   }
   lod_build.StoreLODs(auto_lods, drv, fgroups);

   auto_lod_dists[0] = min_dist;
   auto_lod_dists[1] = max_dist;

   vertex_buffer.Unlock();
   vertex_buffer.SetVertices(new_verts);
}
*/

//----------------------------
//----------------------------

void I3D_source_vertex_buffer::AllocVertices(dword num_verts1){

   dword d3d_usage_flags = 0;
   d3d_usage_flags |= D3DUSAGE_DONOTCLIP;     //bug in DX ? must include clipping flags in source VB
      //D3DUSAGE_WRITEONLY |
   D3DPOOL d3d_pool_flags = D3DPOOL_SYSTEMMEM;

   if(drv->IsDirectTransform()){
      //d3d_pool_flags = D3DPOOL_DEFAULT;
   }else{
      d3d_usage_flags |= D3DUSAGE_SOFTWAREPROCESSING;
   }

                              //alloc vertex buffer
   if(!D3D_vertex_buffer || num_vertices != num_verts1){
      if(D3D_vertex_buffer){
         drv->FreeVertexBuffer(D3D_vertex_buffer, D3D_vertex_buffer_index);
         D3D_vertex_buffer = NULL;
         D3D_vertex_buffer_index = 0;
      }
                           //alloc raw VB
                           //// specify size in dwords (???)
      bool b;
      b = drv->AllocVertexBuffer(fvflags, d3d_pool_flags, d3d_usage_flags,
         //num_verts1 * vstride,
         num_verts1,
         (IDirect3DVertexBuffer8**)&D3D_vertex_buffer, &D3D_vertex_buffer_index);
      assert(b);
      num_vertices = num_verts1;
   }
}

//----------------------------

void I3D_source_vertex_buffer::DuplicateVertices(word *dvi, dword num_dup_verts){

   dword d3d_usage_flags = 0;
   d3d_usage_flags |= D3DUSAGE_DONOTCLIP;     //bug in DX ? must include clipping flags in source VB
      //D3DUSAGE_WRITEONLY |
   D3DPOOL d3d_pool_flags = D3DPOOL_SYSTEMMEM;

   if(drv->IsDirectTransform()){
      //d3d_pool_flags = D3DPOOL_DEFAULT;
   }else{
      d3d_usage_flags |= D3DUSAGE_SOFTWAREPROCESSING;
   }

   int num_orig_verts = num_vertices - vertex_map.size();
   vertex_map.reserve(num_vertices + num_dup_verts);

   for(int i=0; i<num_dup_verts; i++){
      int index = dvi[i];
      if(index < num_orig_verts)
         vertex_map.push_back(index);
      else
         vertex_map.push_back(vertex_map[index - num_orig_verts]);
   }
                              //alloc (new) vertex buffer
   IDirect3DVertexBuffer8 *D3D_vertex_buffer1;
   int D3D_vertex_buffer_index1;

   bool b;
   b = drv->AllocVertexBuffer(fvflags, d3d_pool_flags, d3d_usage_flags, num_vertices + num_dup_verts,
      (IDirect3DVertexBuffer8**)&D3D_vertex_buffer1, &D3D_vertex_buffer_index1);
   assert(b);

   if(D3D_vertex_buffer){
                              //feed vertices
      const void *vp = Lock();
      assert(vp);
      if(vp){
         void *vp1;
         dword vstride = ::GetSizeOfVertex(fvflags);
         HRESULT hr = D3D_vertex_buffer1->Lock(D3D_vertex_buffer_index1 * vstride,
            (num_vertices + num_dup_verts) * vstride,
            (byte**)&vp1,
            0);
         if(SUCCEEDED(hr)){
            memcpy(vp1, vp, num_vertices * vstride);
            for(i=0; i<num_dup_verts; i++){
               int index = dvi[i];
               memcpy((byte*)vp1 + vstride*(num_vertices+i),
                  (byte*)vp + vstride*index, vstride);
               //((I3D_vertex_mesh*)vp1)[num_vertices+i] = ((I3D_vertex_mesh*)vp)[index];
            }

            D3D_vertex_buffer1->Unlock();
         }
         Unlock();
      }
      drv->FreeVertexBuffer(D3D_vertex_buffer, D3D_vertex_buffer_index);
      D3D_vertex_buffer = D3D_vertex_buffer1;
      D3D_vertex_buffer_index = D3D_vertex_buffer_index1;
   }
   num_vertices += num_dup_verts;
}

//----------------------------
//----------------------------

I3D_source_index_buffer::I3D_source_index_buffer(PI3D_driver drv1):
   drv(drv1),
   D3D_index_buffer(NULL),
   num_faces(0)
{
}

//----------------------------

I3D_source_index_buffer::~I3D_source_index_buffer(){

   if(D3D_index_buffer)
      drv->FreeIndexBuffer(D3D_index_buffer, D3D_index_buffer_index);
}

//----------------------------

void I3D_source_index_buffer::AllocFaces(int num){

                              //alloc vertex buffer
   if(!D3D_index_buffer || num_faces != num){
      if(D3D_index_buffer){
         drv->FreeIndexBuffer(D3D_index_buffer, D3D_index_buffer_index);
         D3D_index_buffer = NULL;
         D3D_index_buffer_index = 0;
      }
      if(num){
         dword d3d_usage_flags = 0;
         D3DPOOL d3d_pool_flags = D3DPOOL_SYSTEMMEM;

         if(drv->IsDirectTransform()){
                              //don't use default now, we need a static copy!
            //d3d_pool_flags = D3DPOOL_DEFAULT;
         }else{
            d3d_usage_flags |= D3DUSAGE_SOFTWAREPROCESSING;
         }
         
         bool b;
         b = drv->AllocIndexBuffer(d3d_pool_flags, d3d_usage_flags,
            num, &D3D_index_buffer, &D3D_index_buffer_index);
         assert(b);
      }
      num_faces = num;
   }
}

//----------------------------

void I3D_source_index_buffer::SetFaces(const I3D_triface *faces, dword num_faces1){

   AllocFaces(num_faces1);
   if(num_faces1){
                              //feed vertices
      PI3D_triface ip = Lock();
      assert(ip);
      if(!ip)
         return;
      memcpy(ip, faces, num_faces * sizeof(I3D_triface));
      Unlock();
   }
}

//----------------------------

PI3D_triface I3D_source_index_buffer::Lock(dword d3d_lock_flags){

   if(!num_faces)
      return NULL;
   assert(D3D_index_buffer);

   PI3D_triface ip;
   HRESULT hr = D3D_index_buffer->Lock(D3D_index_buffer_index * sizeof(I3D_triface),
      num_faces * sizeof(I3D_triface),
      (byte**)&ip,
      d3d_lock_flags);
   CHECK_D3D_RESULT("Lock", hr);
   if(FAILED(hr))
      return NULL;
   return ip;
}

//----------------------------

const I3D_triface *I3D_source_index_buffer::LockForReadOnly() const{

   return ((I3D_source_index_buffer*)this)->Lock(D3DLOCK_READONLY);
}

//----------------------------

bool I3D_source_index_buffer::Unlock() const{

   if(!num_faces)
      return false;
   HRESULT hr = D3D_index_buffer->Unlock();
   CHECK_D3D_RESULT("Unlock", hr);
   return (SUCCEEDED(hr));
}

//----------------------------
//----------------------------
                              //meshes

I3D_mesh_base::I3D_mesh_base(PI3D_driver d1, dword fvf1):
   ref(1),
   drv(d1),
   vertex_buffer(drv, fvf1),
   index_buffer(drv)
{
   auto_lod_dists[0] = 0.0f;
   auto_lod_dists[1] = 100.0f;

   drv->AddCount(I3D_CLID_MESH_BASE);
}

//----------------------------

I3D_mesh_base::~I3D_mesh_base(){
   drv->DecCount(I3D_CLID_MESH_BASE);
}

//----------------------------

bool I3D_mesh_base::SaveCachedInfo(C_cache *ck, word version) const{

                           //write version
   ck->write((char*)&version, sizeof(version));

   word num;
   {                          //write vertices
      num = vertex_buffer.NumVertices();
      ck->write((char*)&num, sizeof(num));
      const void *verts = const_cast<PI3D_mesh_base>(this)->vertex_buffer.Lock();
      ck->write((char*)verts, num * vertex_buffer.vstride);
      const_cast<PI3D_mesh_base>(this)->vertex_buffer.Unlock();
   }
   {                       //write spread map
      const vector<word> &vs_map = vertex_buffer.GetVertexSpreadMap();
      num = vs_map.size();
      ck->write((char*)&num, sizeof(num));
      ck->write((char*)vs_map.begin(), num*sizeof(word));
   }
   {                       //write faces
      num = index_buffer.NumFaces1();
      ck->write((char*)&num, sizeof(num));
      const I3D_triface *faces = index_buffer.LockForReadOnly();
      ck->write((char*)faces, num * sizeof(I3D_triface));
      index_buffer.Unlock();
   }
   {                       //write faces groups
      num = fgroups.size();
      ck->write((char*)&num, sizeof(num));
      for(int i=0; i<num; i++){
         const I3D_face_group &fg = fgroups[i];
         word base = fg.BaseIndex1();
         word num_f = fg.NumFaces1();
         ck->write((char*)&base, sizeof(base));
         ck->write((char*)&num_f, sizeof(num_f));
                              //write name of material
         const C_string &mat_name = fg.mat->GetName1();
         ck->write((const char*)mat_name, mat_name.Size() + 1);
      }
   }
   {                       //write auto-lod info
      word num = auto_lods.size();
      ck->write((char*)&num, sizeof(num));
      for(int i=0; i<num; i++){
         const C_auto_lod &al = auto_lods[i];

         ck->write((const char*)&al.ratio, sizeof(al.ratio));
                           //write vertex count
         ck->write((const char*)&al.vertex_count, sizeof(word));

                           //write faces of this LOD
         word num1 = al.index_buffer.NumFaces1();
         ck->write((char*)&num1, sizeof(num));
         const I3D_triface *faces = al.index_buffer.LockForReadOnly();
         ck->write((char*)faces, num1 * sizeof(I3D_triface));
         al.index_buffer.Unlock();

                           //write face groups
         num1 = al.fgroups.size();
         ck->write((char*)&num1, sizeof(num1));
         for(int fgi = 0; fgi < num1; fgi++){
            const I3D_face_group &fg = al.fgroups[fgi];

            word base = fg.BaseIndex1();
            word num_f = fg.NumFaces1();
            ck->write((char*)&base, sizeof(base));
            ck->write((char*)&num_f, sizeof(num_f));
                           //write name of material
            const C_string &mat_name = fg.mat->GetName1();
            ck->write((const char*)mat_name, mat_name.Size() + 1);
         }
      }
      ck->write((const char*)&auto_lod_dists, sizeof(auto_lod_dists));
   }
   return true;
}

//----------------------------

bool I3D_mesh_base::LoadCachedInfo(C_cache *ck, S_load_context &lc, word version){

   {                          //read and check version
      word v = 0;
      ck->read((char*)&v, sizeof(v));
      if(v!=version)
         return false;
   }
   word num;
   vector<byte> vertices;
   dword num_vertices;
   vector<I3D_triface> faces;
   vector<word> vs_map;
   int rl;
   {                       //read vertices
      rl = ck->read((char*)&num, sizeof(num));
      if(rl!=sizeof(num))
         return false;
      num_vertices = num;
      vertices.assign(num * vertex_buffer.vstride);
      rl = ck->read((char*)vertices.begin(), num * vertex_buffer.vstride);
      if(rl!=num * vertex_buffer.vstride)
         return false;
   }
   {                       //read spread map
      rl = ck->read((char*)&num, sizeof(num));
      if(rl!=sizeof(num))
         return false;
      vs_map.assign(num);
      rl = ck->read((char*)vs_map.begin(), num*sizeof(word));
      if(rl!=num*sizeof(word))
         return false;
   }
   {                       //read faces
      rl = ck->read((char*)&num, sizeof(num));
      if(rl!=sizeof(num))
         return false;
      faces.assign(num);
      rl = ck->read((char*)faces.begin(), num * sizeof(I3D_triface));
      if(rl!=num * sizeof(I3D_triface))
         return false;
   }
   {                          //read fgroups
      rl = ck->read((char*)&num, sizeof(num));
      if(rl!=sizeof(num))
         return false;
      fgroups.assign(num);
      for(int i=0; i<num; i++){
                              //read number of faces
         word base, num_f;
         rl = ck->read((char*)&base, sizeof(base));
         if(rl!=sizeof(base))
            return false;
         rl = ck->read((char*)&num_f, sizeof(num_f));
         if(rl!=sizeof(num_f))
            return false;
                              //create and init face-group
         I3D_face_group &fg = fgroups[i];
         fg.base_index = base;
         fg.num_faces = num_f;
                              //read material name
         char mat_name[256];
         int j = 0;
         do{
            rl = ck->read(&mat_name[j++], sizeof(char));
            if(rl!=sizeof(char))
               return false;
         }while(mat_name[j-1]);

         map<C_string, S_mat_info>::iterator it = lc.mat_list.find(mat_name);
         if(it==lc.mat_list.end()){
                              //only default mats may cause this,
                              // cached info invalidated
            fgroups.clear();
            return false;
         }
         fg.SetMaterial((*it).second.GetMaterial(lc));
      }
   }
   {                          //read auto-lod info
      word num;
      rl = ck->read((char*)&num, sizeof(num));
      if(rl!=sizeof(num))
         return false;
      auto_lods.assign(num, C_auto_lod(drv));
      for(int i=0; i<num; i++){
         C_auto_lod &al = auto_lods[i];
         rl = ck->read((char*)&al.ratio, sizeof(al.ratio));
         if(rl!=sizeof(al.ratio))
            return false;
         word num1;
                              //read vertex counts
         rl = ck->read((char*)&al.vertex_count, sizeof(word));
         if(rl!=sizeof(word))
            return false;

                              //read faces of this LOD
         rl = ck->read((char*)&num1, sizeof(num1));
         if(rl!=sizeof(num1))
            return false;
         al.index_buffer.AllocFaces(num1);
         PI3D_triface faces = al.index_buffer.Lock(0);
         rl = ck->read((char*)faces, num1 * sizeof(I3D_triface));
         if(rl!=num1 * sizeof(I3D_triface))
            return false;
         al.index_buffer.Unlock();

                              //read face groups
         rl = ck->read((char*)&num1, sizeof(num1));
         if(rl!=sizeof(num1))
            return false;
         al.fgroups.assign(num1);
         for(int fgi = 0; fgi < num1; fgi++){
            I3D_face_group &fg = al.fgroups[fgi];

            word base, num_f;
            rl = ck->read((char*)&base, sizeof(base));
            if(rl!=sizeof(base))
               return false;
            rl = ck->read((char*)&num_f, sizeof(num_f));
            if(rl!=sizeof(num_f))
               return false;
            fg.base_index = base;
            fg.num_faces = num_f;

                              //read material name
            char mat_name[256];
            int j = 0;
            do{
               rl = ck->read(&mat_name[j++], sizeof(char));
               if(rl!=sizeof(char))
                  return false;
            }while(mat_name[j-1]);

            map<C_string, S_mat_info>::iterator it = lc.mat_list.find(mat_name);
            if(it==lc.mat_list.end()){
                              //only default mats may cause this,
                              // cached info invalidated, so that user sees the message
               auto_lods.clear();
               return false;
            }
            fg.SetMaterial((*it).second.GetMaterial(lc));
         }
      }
      rl = ck->read((char*)&auto_lod_dists, sizeof(auto_lod_dists));
      if(rl!=sizeof(auto_lod_dists))
         return false;
   }
   vertex_buffer.SetVertices(vertices.begin(), num_vertices);
   index_buffer.SetFaces(faces);
   vertex_buffer.SetVertexSpreadMap(vs_map);

   return true;
}

//----------------------------


void I3D_mesh_base::CreateLODs(float min_vertex_ratio, float min_area_ratio, float ratio_change_step,
   float min_dist, float max_dist){

                              //don't compute LODs twice (for case of multiple-referenced meshes)
   if(auto_lods.size())
      return;

   const void *verts = vertex_buffer.Lock();
   dword num_verts = NumVertices1();

   vector<word> new_v_map;
   vector<byte> new_verts(num_verts * vertex_buffer.vstride);

   CreateLODsGeneric(min_vertex_ratio, min_area_ratio, ratio_change_step, min_dist, max_dist,
      drv, (const S_vector*)verts, vertex_buffer.vstride, num_verts,
      index_buffer, fgroups, auto_lods, auto_lod_dists,
      (byte*)new_verts.begin(), new_v_map);

   vertex_buffer.Unlock();
   vertex_buffer.SetVertices(new_verts.begin(), num_verts);
   vertex_buffer.SetVertexSpreadMap(new_v_map.begin(), new_v_map.size());
}

//----------------------------

I3D_RESULT I3D_mesh_level::Build2(const I3D_vertex_mesh *verts, dword num_verts,
   const I3D_triface *faces, dword num_faces,
   const I3D_face_group *fgroups, dword num_fgroups,
   dword build_flags){

                              //make data validity checks
   if(!num_verts || !num_faces || !num_fgroups)
      return I3DERR_INVALIDPARAMS;

#ifdef _DEBUG
   int i;
                              //check if fgroups ok
   for(i=num_fgroups; i--; ){
      if((fgroups[i].base_index + fgroups[i].num_faces) > num_faces){
         assert(0);
         return I3DERR_INVALIDPARAMS;
      }
   }
                              //check if indicies ok
   for(i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      if(fc[0]>num_verts || fc[1]>num_verts || fc[2]>num_verts){
         assert(0);
         return I3DERR_INVALIDPARAMS;
      }
   }
#endif

                              //assign vertices
   vertex_buffer.SetVertices(verts, num_verts);
                              //assign faces
   index_buffer.SetFaces(faces, num_faces);
                              //assign fgroups
   SetFGroups(fgroups, num_fgroups);

   return I3D_OK;
}

//----------------------------

void I3D_mesh_base::Clone(CPI3D_mesh_base mb){

                           //copy faces
   SetFGroups(mb->GetFGroups1(), mb->NumFGroups1());
   const I3D_triface *faces = mb->index_buffer.LockConst();
   index_buffer.SetFaces(faces, mb->index_buffer.NumFaces1());
   mb->index_buffer.Unlock();
                           //copy auto-lod info
   //auto_lods = mb->GetAutoLODs();
   const vector<C_auto_lod> &als = mb->GetAutoLODs();
   auto_lods.assign(als.size(), C_auto_lod(drv));
   for(int i=als.size(); i--; ){
      const C_auto_lod &s = als[i];
      C_auto_lod &d = auto_lods[i];
      d.ratio = s.ratio;
      d.vertex_count = s.vertex_count;
      d.fgroups = s.fgroups;
      const I3D_triface *fcs = ((C_auto_lod&)s).index_buffer.Lock(D3DLOCK_READONLY);
      d.index_buffer.SetFaces(fcs, s.index_buffer.NumFaces1());
      s.index_buffer.Unlock();
   }
   memcpy(auto_lod_dists, mb->GetAutoLODDists(), sizeof(auto_lod_dists));
                           //copy vertices - only matching elements, ignore others
   vector<byte> new_vertices(mb->vertex_buffer.NumVertices() * vertex_buffer.vstride);
   const void *src = mb->vertex_buffer.LockConst();
   CloneVertexBuffer(src, mb->vertex_buffer.fvflags, new_vertices.begin(), vertex_buffer.fvflags, mb->vertex_buffer.NumVertices());
   mb->vertex_buffer.UnlockConst();
   vertex_buffer.SetVertexSpreadMap(mb->vertex_buffer.GetVertexSpreadMap());
   SetClonedVertices(new_vertices);
                              //copy rest
   smooth_groups = mb->smooth_groups;
   map_channel_2 = mb->map_channel_2;
}

//----------------------------

void I3D_mesh_base::ComputeBoundingVolume(){

   bvolume.bbox.Invalidate();
   if(!NumVertices1())
      return;

   const S_vector *src = (const S_vector*)vertex_buffer.Lock(D3DLOCK_READONLY);
   assert(src);
   if(!src)
      return;

   const S_vector *src1 = src;
   for(int i=vertex_buffer.NumVertices(); i--; src1 = (const S_vector*)(((byte*)src1) + vertex_buffer.vstride)){
      bvolume.bbox.min.Minimal(*src1);
      bvolume.bbox.max.Maximal(*src1);
   }
                              //expand by safe thresh
   for(i=0; i<3; i++){
      bvolume.bbox.min[i] -= SAFE_BOUND_THRESH;
      bvolume.bbox.max[i] += SAFE_BOUND_THRESH;
   }
                              //bounding sphere
   S_vector bbox_diagonal = bvolume.bbox.max - bvolume.bbox.min;
   bvolume.bsphere.pos = (bvolume.bbox.min + bbox_diagonal*.5f);

                              //try to make bsphere as small as possible
   bvolume.bsphere.radius = 0.0f;
   for(i = vertex_buffer.NumVertices(); i--; src1 = (const S_vector*)((byte*)src1) + vertex_buffer.vstride){
      S_vector v = *src - bvolume.bsphere.pos;
      float dist2 = v.Dot(v);
      bvolume.bsphere.radius = Max(bvolume.bsphere.radius, dist2);
   }
   if(bvolume.bsphere.radius>MRG_ZERO){
      bvolume.bsphere.radius = SqrtFast(bvolume.bsphere.radius);
                              //safe expansion
      bvolume.bsphere.radius += SAFE_BOUND_THRESH;
   }else
      bvolume.bsphere.radius = 0.0f;

   vertex_buffer.Unlock();
}

//----------------------------

#include <deque>


static bool operator <(const I3D_edge &e1, const I3D_edge &e2){
   return ( (*(dword*)&e1) < (*(dword*)&e2) );
}

//----------------------------

class C_mesh_optimizer{
   //const S_vector *vertices;
   //dword vertex_stride;
   //dword num_verts;
   const I3D_triface *faces;
   dword num_faces;
   const vector<I3D_face_group> &fgroups;
   vector<I3D_triface> &out_faces;
   PI3D_driver drv;

//----------------------------
// Finds best face adjacent to given face. This is determined from provided edge map.
   static const I3D_triface *FindBestAdjacendFace(const I3D_triface &fc,
      map<I3D_edge, I3D_triface> &edge_map, int &delta){

      int fc_val = fc[0] + fc[1] + fc[2];
      delta = 0;
      const I3D_triface *best_face = NULL;
      map<I3D_edge, I3D_triface>::const_iterator it;

                              //test edge 0, 1
      it = edge_map.find(I3D_edge(fc[1], fc[0]));
      if(it!=edge_map.end()){
         const I3D_triface &edge_fc = (*it).second;
         int curr_delta = (edge_fc[0] + edge_fc[1] + edge_fc[2]) - fc_val;
         if(!best_face || abs(delta) > abs(curr_delta)){
            best_face = &edge_fc;
            delta = curr_delta;
         }
      }
                              //test edge 1, 2
      //if(!best_face){
      it = edge_map.find(I3D_edge(fc[2], fc[1]));
      if(it!=edge_map.end()){
         const I3D_triface &edge_fc = (*it).second;
         int curr_delta = (edge_fc[0] + edge_fc[1] + edge_fc[2]) - fc_val;
         if(!best_face || abs(delta) > abs(curr_delta)){
            best_face = &edge_fc;
            delta = curr_delta;
         }
      }
      //}
                              //test edge 2, 0
      //if(!best_face){
      it = edge_map.find(I3D_edge(fc[0], fc[2]));
      if(it!=edge_map.end()){
         const I3D_triface &edge_fc = (*it).second;
         int curr_delta = (edge_fc[0] + edge_fc[1] + edge_fc[2]) - fc_val;
         if(!best_face || abs(delta) > abs(curr_delta)){
            best_face = &edge_fc;
            delta = curr_delta;
         }
      }
      //}

      return best_face;
   }

//----------------------------

   static const I3D_triface *FindBestAdjacendFace(const I3D_triface &fc,
      map<I3D_edge, I3D_triface> &edge_map){

      map<I3D_edge, I3D_triface>::const_iterator it;
                              //test edge 0, 1
      it = edge_map.find(I3D_edge(fc[1], fc[0]));
      if(it!=edge_map.end())
         return &(*it).second;
                              //test edge 1, 2
      it = edge_map.find(I3D_edge(fc[2], fc[1]));
      if(it!=edge_map.end())
         return &(*it).second;
                              //test edge 2, 0
      it = edge_map.find(I3D_edge(fc[0], fc[2]));
      if(it!=edge_map.end())
         return &(*it).second;

      return NULL;
   }

//----------------------------
// Remove all edges of specified face from map.
   static void RemoveFaceEdgesFromMap(const I3D_triface &fc, map<I3D_edge, I3D_triface> &edge_map){

      map<I3D_edge, I3D_triface>::iterator it;

      it = edge_map.find(I3D_edge(fc[0], fc[1]));
      if(it != edge_map.end())
         edge_map.erase(it);

      it = edge_map.find(I3D_edge(fc[1], fc[2]));
      if(it != edge_map.end())
         edge_map.erase(it);

      it = edge_map.find(I3D_edge(fc[2], fc[0]));
      if(it != edge_map.end())
         edge_map.erase(it);
   }


//----------------------------
// Sorts region of faces, makes strips as possible.
// Result is appended to out_faces.
// The returned value is number of strips made.
   int SortFaceRegion(const I3D_triface *faces, dword num_faces, int &debug_countdown){

      int num_strips = 0;
      int i;

      deque<I3D_triface> sorted_faces;
      vector<I3D_triface> fg_faces(faces, faces + num_faces);

                              //make edge mapping
      map<I3D_edge, I3D_triface> edge_map;
      for(i=num_faces; i--; ){
         const I3D_triface &fc = fg_faces[i];
         edge_map[I3D_edge(fc[0], fc[1])] = fc;
         edge_map[I3D_edge(fc[1], fc[2])] = fc;
         edge_map[I3D_edge(fc[2], fc[0])] = fc;
      }

      while(fg_faces.size()){
         deque<I3D_triface> strip;
                              //add 1st face
         I3D_triface fc_front = fg_faces.back(); fg_faces.pop_back();
         RemoveFaceEdgesFromMap(fc_front, edge_map);

                              //find connected face with lowest delta
         //int delta;
         //const I3D_triface *best_face = FindBestAdjacendFace(fc_front, edge_map, delta);
         const I3D_triface *best_face = FindBestAdjacendFace(fc_front, edge_map);
         if(!best_face){
            sorted_faces.push_back(fc_front);
            ++num_strips;
            continue;
         }
         int fi = FindFace(fg_faces.begin(), fg_faces.size(), *best_face);
         assert(fi!=-1);
         fg_faces[fi] = fg_faces.back(); fg_faces.pop_back();

         I3D_triface fc_back = *best_face;
         RemoveFaceEdgesFromMap(fc_back, edge_map);
         //if(delta > 0) swap(fc_front, fc_back);

                              //put them both into list
         strip.push_back(fc_front);
         strip.push_back(fc_back);

                              //iterate upwards and backwards, build strip
         while(true){
            const I3D_triface *best_face = FindBestAdjacendFace(fc_back, edge_map);
            if(!best_face)
               break;
            int fi = FindFace(fg_faces.begin(), fg_faces.size(), *best_face);
            assert(fi!=-1);
            fg_faces[fi] = fg_faces.back(); fg_faces.pop_back();

            fc_back = *best_face;
            RemoveFaceEdgesFromMap(fc_back, edge_map);

            strip.push_back(fc_back);
         }
         while(true){
            const I3D_triface *best_face = FindBestAdjacendFace(fc_front, edge_map);
            if(!best_face)
               break;
            int fi = FindFace(fg_faces.begin(), fg_faces.size(), *best_face);
            assert(fi!=-1);
            fg_faces[fi] = fg_faces.back(); fg_faces.pop_back();

            fc_front = *best_face;
            RemoveFaceEdgesFromMap(fc_front, edge_map);

            strip.push_front(fc_front);
         }
#if 0
         {
            deque<I3D_triface>::const_iterator it = strip.begin();
            int count = 5;

            while(debug_countdown){
               --debug_countdown;
               const I3D_triface &f0 = *it;
               ++it;
               if(it==strip.end())
                  break;
               const I3D_triface &f1 = *it;
               int se = FindSharedEdge(f0, f1);
               assert(se!=-1);

               S_vector vv[3] = {verts[f0[se]].xyz, verts[f0[(se+1)%3]].xyz};
               vv[2].Zero();
               if(count==5) vv[2] = S_vector(1, 1, 1);
               else vv[2][count] = 1.0f;
               ++count %= 3;
               drv->DEBUG_CODE(20, (dword)vv);
            }
         }
#endif

                              //end of strip build,
                              // put it into global list
         sorted_faces.insert(sorted_faces.end(), strip.begin(), strip.end());
         ++num_strips;
      }

      for(deque<I3D_triface>::const_iterator it = sorted_faces.begin(); it!=sorted_faces.end(); ++it){
         out_faces.push_back(*it);
      }
      return num_strips;
   }

//----------------------------
// Sorts region of faces, makes strips as possible.
// Result is appended to out_faces.
// The returned value is number of strips made.
   int SortFaceRegion1(const I3D_triface *faces, dword num_faces, int &debug_countdown){

      int num_strips = 0;

      deque<I3D_triface> sorted_faces;
      vector<I3D_triface> fg_faces(faces, faces + num_faces);
                              //put 1st face
      I3D_triface fc_back = fg_faces.back(); fg_faces.pop_back();
      I3D_triface fc_front = fc_back;
                              //start 1st strip
      sorted_faces.push_back(fc_back);
      ++num_strips;
      while(fg_faces.size()){
                              //find face which would fit either to front or to back
         for(int fi=fg_faces.size(); fi--; ){
            I3D_triface &fc = fg_faces[fi];
            if(FindSharedEdge(fc, fc_back) != -1){
               sorted_faces.push_back(fc);
               fc_back = fc;
               fc = fg_faces.back(); fg_faces.pop_back();
               break;
            }else
            if(FindSharedEdge(fc, fc_front) != -1){
               sorted_faces.push_front(fc);
               fc_front = fc;
               fc = fg_faces.back(); fg_faces.pop_back();
               break;
            }
         }
         if(fi==-1){
                              //no such face found, generate another strip
            fc_back = fg_faces.back(); fg_faces.pop_back();
            sorted_faces.push_back(fc_back);
            ++num_strips;
         }
      }
      for(deque<I3D_triface>::const_iterator it = sorted_faces.begin(); it!=sorted_faces.end(); ++it){
         out_faces.push_back(*it);
      }
      return num_strips;
   }
   
//----------------------------
// Sorts region of faces, makes strips as possible.
// Result is appended to out_faces.
// The returned value is number of strips made.
   int SortFaceRegion2(const I3D_triface *faces, dword num_faces, int &debug_countdown){

      vector<I3D_triface> fg_faces(faces, faces + num_faces);
      deque<word> vertex_cache;

      //reverse(fg_faces.begin(), fg_faces.end());

      const int CACHE_SIZE = 8;

      while(fg_faces.size()){
                              //find face which best fits in cache
         int best_face, best_score = 0;

         for(int i=fg_faces.size(); i--; ){
            const I3D_triface &fc = fg_faces[i];
            deque<word>::const_iterator it0 = find(vertex_cache.begin(), vertex_cache.end(), fc[0]);
            deque<word>::const_iterator it1 = find(vertex_cache.begin(), vertex_cache.end(), fc[1]);
            deque<word>::const_iterator it2 = find(vertex_cache.begin(), vertex_cache.end(), fc[2]);
            int num_verts_in_cache =
               int(it0!=vertex_cache.end()) + int(it1!=vertex_cache.end()) + int(it2!=vertex_cache.end());
            if(num_verts_in_cache==3){
                              //3 verts in cache, perfect hit
               best_score = 3;
               best_face = i;
               break;
            }
            if(best_score < num_verts_in_cache){
               best_score = num_verts_in_cache;
               best_face = i;
            }
         }
         if(!best_score){
                              //no face found which would utilize cache, use random
            best_face = fg_faces.size() - 1;
         }
         I3D_triface &fc = fg_faces[best_face];
                              //output this face
         out_faces.push_back(fc);
                              //add non-cached verts into cache
         for(i=0; i<3; i++){
            deque<word>::const_iterator it = find(vertex_cache.begin(), vertex_cache.end(), fc[i]);
            if(it==vertex_cache.end()){
               vertex_cache.push_back(fc[i]);
                              //use FIFO system
               if(vertex_cache.size() > CACHE_SIZE)
                  vertex_cache.pop_front();
            }
         }
                              //remove from temp list
         fc = fg_faces.back(); fg_faces.pop_back();
      }
      return 0;
   }

public:
   C_mesh_optimizer(PI3D_driver d1,
      //const S_vector *verts1, dword vs, dword num_verts1,
      const I3D_triface *faces1, dword num_faces1,
      const vector<I3D_face_group> &fg1,
      vector<I3D_triface> &out_faces1):

      drv(d1),
      //vertices(verts1), vertex_stride(vs), num_verts(num_verts1),
      fgroups(fg1),
      faces(faces1), num_faces(num_faces1),
      out_faces(out_faces1)
   {}

   void Optimize(PI3D_driver drv, const char *mesh_name){

      out_faces.reserve(num_faces);

      static int cntdwn_base = 0;
      int debug_countdown = cntdwn_base++;

      for(int fgi=0; fgi<fgroups.size(); fgi++){
         const I3D_face_group &fg = fgroups[fgi];
         //int ns =
            SortFaceRegion2(faces + fg.base_index, fg.num_faces, debug_countdown);
         //drv->PRINT(C_fstring("%s: s: %i (v: %i, f: %i)", mesh_name, ns, num_verts, num_faces));
      }
      assert(out_faces.size() == num_faces);
   }
};

//----------------------------

void I3D_mesh_base::Optimize(const S_load_context &lc){

#if 0
   {
                              //debug: shuffle faces
      const I3D_triface *faces = index_buffer.Lock();
      dword num_faces = NumFaces1();

      vector<I3D_triface> out_faces; out_faces.reserve(num_faces);

      for(int fgi=0; fgi<fgroups.size(); fgi++){
         const I3D_face_group &fg = fgroups[fgi];

         C_sort_list<const I3D_triface*> sl;
         sl.Reserve(fg.num_faces);
         for(int fi=fg.num_faces; fi--; ){
            const I3D_triface &fc = faces[fg.base_index + fi];
            //sl.Add(&fc, fc[0] + fc[1] + fc[2]);
            sl.Add(&fc, S_int_random(-1));
         }
         sl.Sort();
         dword base_index = out_faces.size();
         out_faces.insert(out_faces.end(), fg.num_faces, I3D_triface());
         for(fi=0; fi<fg.num_faces; fi++){
            out_faces[base_index + fi] = *sl[fi];
         }
      }
      index_buffer.Unlock();
      index_buffer.SetFaces(out_faces);
   }
#endif


#if 0
   {
                              //re-sort faces
      vector<I3D_triface> out_faces;
      C_mesh_optimizer mesh_optimizer(drv,
         index_buffer.Lock(), index_buffer.NumFaces1(), fgroups, out_faces);
      mesh_optimizer.Optimize(drv, lc.recent_instance_name);
      index_buffer.Unlock();
      index_buffer.SetFaces(out_faces);
                              //process all LODs
      for(int i=auto_lods.size(); i--; ){
         C_auto_lod &al = auto_lods[i];

         vector<I3D_triface> out_faces;
         C_mesh_optimizer mesh_optimizer(drv,
            al.index_buffer.Lock(), al.index_buffer.NumFaces1(),
            al.fgroups, out_faces);
         mesh_optimizer.Optimize(drv, lc.recent_instance_name);
         al.index_buffer.Unlock();
         al.index_buffer.SetFaces(out_faces);
      }
   }
#endif
}

//----------------------------
//----------------------------

I3D_mesh_level::I3D_mesh_level(PI3D_driver d):
   I3D_mesh_base(d, FVF_VERTEX_MESH),
   lod_factor(0.0f),
   mesh_flags(0)
{
}

//----------------------------

/*
void I3D_mesh_level::SetSmoothGroups(const vector<byte> &sg){

   assert(!sg.size() || sg.size() == NumFaces1());
   if(sg.size()){
      smooth_groups.Assign((const char*)sg.begin(), sg.size());
   }else{
      smooth_groups = NULL;
   }
}
*/

//----------------------------

void I3D_mesh_level::ApplyPivotAndNUScale(const S_vector &p, const S_vector &s){

   if(mesh_flags&MESHF_PIVOT_AND_NU_APPLIED)
      return;

   if(!p.IsNull() || s != S_vector(1.0f, 1.0f, 1.0f)){
      PI3D_vertex_mesh verts = (PI3D_vertex_mesh)vertex_buffer.Lock();
      for(int i=vertex_buffer.NumVertices(); i--; ){
         verts[i].xyz -= p;
         verts[i].xyz *= s;
      }
      vertex_buffer.Unlock();
   }

   mesh_flags |= MESHF_PIVOT_AND_NU_APPLIED;
}

//----------------------------
#define DBASE_VERSION 0x001c

bool I3D_mesh_level::SaveCachedInfo(C_cache *ck, word) const{

                              //write core
   if(!((I3D_mesh_level*)this)->I3D_mesh_base::SaveCachedInfo(ck, DBASE_VERSION))
      return false;
   {                          //write smooth groups
      byte has_sg_list = (smooth_groups.Size() != 0);
      ck->write((char*)&has_sg_list, sizeof(byte));
      if(has_sg_list)
         ck->write((const char*)smooth_groups, NumFaces1() * sizeof(byte));
   }
   {                          //write additional map channel
      byte mc2_present = (map_channel_2 != NULL);
      ck->write((const char*)&mc2_present, sizeof(byte));
      if(mc2_present){
                                 //write vertices
         word num = map_channel_2->uv_verts.size();
         ck->write((const char*)&num, sizeof(num));
         ck->write((const char*)map_channel_2->uv_verts.begin(), num * sizeof(I3D_text_coor));
                                 //write faces
         num = map_channel_2->uv_faces.size();
         ck->write((const char*)&num, sizeof(num));
         ck->write((const char*)map_channel_2->uv_faces.begin(), num * sizeof(I3D_triface));
                                 //write texture name
         num = map_channel_2->bitmap_name.Size();
         ck->write((const char*)&num, sizeof(num));
         ck->write((const char*)map_channel_2->bitmap_name, num);
      }
   }
   return true;
}

//----------------------------

bool I3D_mesh_level::LoadCachedInfo(C_cache *ck, S_load_context &lc, word){

                              //read core
   if(!I3D_mesh_base::LoadCachedInfo(ck, lc, DBASE_VERSION))
      return false;
   {                          //read smooth groups
      byte has_sg_list;
      ck->read((char*)&has_sg_list, sizeof(byte));
      if(has_sg_list){
         smooth_groups.Assign(NULL, index_buffer.NumFaces1());
         ck->read((char*)(const char*)smooth_groups, index_buffer.NumFaces1() * sizeof(byte));
      }else{
         smooth_groups = NULL;
      }
   }
                              //read additional map channel
   byte mc2_present = false;
   ck->read((char*)&mc2_present, sizeof(byte));
   if(mc2_present){
      word num;
      map_channel_2 = new C_map_channel; map_channel_2->Release();
                              //read vertices
      ck->read((char*)&num, sizeof(num));
      map_channel_2->uv_verts.assign(num);
      ck->read((char*)map_channel_2->uv_verts.begin(), num * sizeof(I3D_text_coor));
                              //read faces
      ck->read((char*)&num, sizeof(num));
      map_channel_2->uv_faces.assign(num);
      ck->read((char*)map_channel_2->uv_faces.begin(), num * sizeof(I3D_triface));
                              //read texture name
      ck->read((char*)&num, sizeof(num));
      if(num){
         char *cp = (char*)alloca(num+1);
         ck->read(cp, num);
         cp[num] = 0;
         map_channel_2->bitmap_name = cp;
      }
   }
   //mesh_flags |= MESHF_PIVOT_AND_NU_APPLIED;;
   return true;
}

//----------------------------

I3D_RESULT I3D_mesh_level::Duplicate2(CPI3D_mesh_level ml){

   if(ml==this)
      return I3D_OK;

   mesh_flags = ml->mesh_flags;
   lod_factor = ml->lod_factor;
   {                          //duplicate vertex buffer
      const I3D_vertex_mesh *verts = (const I3D_vertex_mesh*)((PI3D_mesh_level)ml)->vertex_buffer.Lock(D3DLOCK_READONLY);
      vertex_buffer.SetVertices(verts, ml->NumVertices1());
      ((PI3D_mesh_level)ml)->vertex_buffer.Unlock();
   }
   {                          //duplicate faces
      const I3D_triface *faces = ((PI3D_mesh_level)ml)->index_buffer.Lock(D3DLOCK_READONLY);
      index_buffer.SetFaces(faces, ml->NumFaces1());
      ((PI3D_mesh_level)ml)->index_buffer.Unlock();
   }
   smooth_groups = ml->smooth_groups;

   fgroups = ml->fgroups;
   map_channel_2 = ml->map_channel_2;

   return I3D_OK;
}

//----------------------------
//----------------------------

I3D_mesh_object::I3D_mesh_object(PI3D_driver d):
   ref(1),
   drv(d),
   mo_flags(0)
   //num_lods(0)
{
   drv->AddCount(I3D_CLID_MESH_OBJECT);
   /*
   bvolume.bbox.min.Zero();
   bvolume.bbox.max.Zero();
   bvolume.bsphere.pos.Zero();
   bvolume.bsphere.radius = 0.0f;
   */
}

//----------------------------

I3D_mesh_object::~I3D_mesh_object(){

   drv->DecCount(I3D_CLID_MESH_OBJECT);
}

//----------------------------

/*
void I3D_mesh_object::ComputeBoundingVolume(){

   if(mo_flags&MOF_BOUNDSBUILD)
      return;

   mo_flags |= MOF_BOUNDSBUILD;

   int i, j;

                              //create bounding box
                              // use all levels of detail
   bvolume.bbox.Invalidate();
   for(j=num_lods; j--; ){
      PI3D_mesh_level ml = lod_meshes[j];
      ml->ComputeBoundingVolume();
      bvolume.bbox.min.Minimal(ml->bvolume.bbox.min);
      bvolume.bbox.max.Maximal(ml->bvolume.bbox.max);
   }
                              //bounding sphere
   S_vector bbox_diagonal = bvolume.bbox.max - bvolume.bbox.min;
   bvolume.bsphere.pos = (bvolume.bbox.min + bbox_diagonal*.5f);

                              //try to make bsphere as small as possible
   bvolume.bsphere.radius = 0.0f;
   for(j=num_lods; j--; ){
      PI3D_mesh_level ml = lod_meshes[j];
      const I3D_vertex_mesh *vp = (const I3D_vertex_mesh*)ml->vertex_buffer.Lock(D3DLOCK_READONLY);
      for(i = ml->vertex_buffer.NumVertices(); i--; ){
         S_vector v(vp[i].xyz - bvolume.bsphere.pos);
         float dist2 = v.Dot(v);
         bvolume.bsphere.radius = Max(bvolume.bsphere.radius, dist2);
      }
      ml->vertex_buffer.Unlock();
   }
   if(bvolume.bsphere.radius>MRG_ZERO){
      bvolume.bsphere.radius = SqrtFast(bvolume.bsphere.radius);
                              //safe expansion
      bvolume.bsphere.radius += SAFE_BOUND_THRESH;
   }else bvolume.bsphere.radius = 0.0f;
}

//----------------------------
*/

const I3D_bound_volume &I3D_mesh_object::GetBoundVolume() const{ 
   
   //assert(mo_flags&MOF_BOUNDSBUILD);
   //return bvolume; 
   return lod_mesh->GetBoundingVolume();
}

//----------------------------

dword I3D_mesh_object::NumVertices(int lod_index) const{

   assert(!lod_index);
   if(lod_index)
      return 0;
   return lod_mesh->NumVertices1();
}

//----------------------------

PI3D_vertex_mesh I3D_mesh_object::LockVertices(int lod_index, dword flags){

   assert(!lod_index);
   if(lod_index)
      return NULL;

   if(flags&I3DLOCK_VERTEX){
      if(flags&(I3DLOCK_XYZ | I3DLOCK_NORMAL)){
         if(!(flags&I3DLOCK_NO_LIGHT_RESET))
            mo_flags |= MOF_LOCK_VERTS_RESET_LIGHTING;
      }
      if(flags&I3DLOCK_UV)
         mo_flags |= MOF_LOCK_VERTS_RESET_UV;

      if(flags&I3DLOCK_XYZ){
         mo_flags |= MOF_LOCK_VERTS_RESET_XYZ;
         //mo_flags &= ~MOF_BOUNDSBUILD;
      }
   }
   return (PI3D_vertex_mesh)lod_mesh->vertex_buffer.Lock();
}

//----------------------------

I3D_RESULT I3D_mesh_object::UnlockVertices(int lod_index){

   assert(!lod_index);
   if(lod_index)
      return I3DERR_INVALIDPARAMS;

   lod_mesh->vertex_buffer.Unlock();
   if(mo_flags&(MOF_LOCK_VERTS_RESET_XYZ|MOF_LOCK_VERTS_RESET_LIGHTING|MOF_LOCK_VERTS_RESET_UV)){
                              //make all visuals dirty:
                              // - force up-loading of certices if T&L device
                              // - recompute bounding volumes
      /*
      if(!(mo_flags&MOF_BOUNDSBUILD))
         ComputeBoundingVolume();
         */
      mo_flags &= ~(MOF_LOCK_VERTS_RESET_XYZ | MOF_LOCK_VERTS_RESET_LIGHTING | MOF_LOCK_VERTS_RESET_UV);
   }

   return I3D_OK;
}

//----------------------------

PI3D_triface I3D_mesh_object::LockFaces(int lod_index, dword flags){

   assert(!lod_index);
   if(lod_index)
      return NULL;

   return (PI3D_triface)lod_mesh->LockFaces();
}

//----------------------------

I3D_RESULT I3D_mesh_object::UnlockFaces(int lod_index){

   assert(!lod_index);
   if(lod_index)
      return I3DERR_INVALIDPARAMS;

   lod_mesh->UnlockFaces();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_mesh_object::SetLOD(PI3D_mesh_level ml, int indx){

   if(indx!=0)
      return I3DERR_INVALIDPARAMS;
   lod_mesh = ml;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_mesh_object::Duplicate(CPI3D_mesh_object mo){

   if(mo==this)
      return I3D_OK;

   lod_mesh = mo->lod_mesh;
   drv = mo->drv;
   mo_flags = mo->mo_flags;

   return I3D_OK;
}

//----------------------------
//----------------------------
