#ifndef __MESH_H
#define __MESH_H

#include <vector>
#include <d3d8.h>
#include <c_unknwn.hpp>
#include <c_string.hpp>
#include "load_3ds.h"
#include "common.h"


#define FVF_VERTEX_MESH (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1)

//----------------------------
// Check if face contains vertex (index)
inline bool FaceContainsIndex(const I3D_triface &fc, word v){
   return (fc[0]==v || fc[1]==v || fc[2]==v);
}

//----------------------------
// Check if face contains edge
inline bool FaceContainsIndicies(const I3D_triface &fc, word v0, word v1){
   return (FaceContainsIndex(fc, v0) && FaceContainsIndex(fc, v1));
}

//----------------------------
//    Replace vertex in face by other vertex
inline void ReplaceFaceIndex(I3D_triface &fc, word v_orig, word v_new){
   if(fc[0]==v_orig) fc[0] = v_new;
   if(fc[1]==v_orig) fc[1] = v_new;
   if(fc[2]==v_orig) fc[2] = v_new;
}

//----------------------------
// Compute 2D area of specified rectangle
float AreaOfTriangle(const S_vector *verts, dword vertex_pitch, const I3D_triface &face);

//----------------------------
// Auto-LOD helper func: find vertex in pool, which is best for remove
bool FindNextPointToRemove(const S_vector *verts, dword vertex_pitch, dword num_verts,
   const I3D_triface *faces, dword num_faces,
   word &v_out, word &v_map_to);

//----------------------------
// Check if set of vertices are planar, use specified thresh to determine
bool CheckIfVerticesPlanar(const S_vector *verts, dword num_verts, float thresh = 1e-2f);

//----------------------------
// Build normals of list of faces
void BuildFaceNormals(const S_vector *verts,
   const I3D_triface *faces, dword num_faces, S_vector *normals);

//----------------------------
// Find specified face (given by 2 indicies) in list, return index,
// or -1 if not found, if found, fill third index.
int FindFace(const I3D_triface *faces, dword num_faces, word e1, word e2, word &e3);

//----------------------------
// Find specified face (given by 3 indicies) in list, return index,
// or -1 if not found.
int FindFace(const I3D_triface *faces, dword num_faces, const I3D_triface &fc);

//----------------------------
// Find edge shared among two faces. If one is found, the returned value is index
// of 1st point of the edge in the 1st of faces. The other edge vertex is the next 
// vertex in the face.
// If shared edge is not found, the returned value is -1.
inline int FindSharedEdge(const I3D_triface &fc1, const I3D_triface &fc2){
   if(FaceContainsIndicies(fc2, fc1[0], fc1[1])) return 0;
   if(FaceContainsIndicies(fc2, fc1[1], fc1[2])) return 1;
   if(FaceContainsIndicies(fc2, fc1[2], fc1[0])) return 2;
   return -1;
}

//----------------------------
// Find specified edge (given by 2 indicies) in list, return index,
// or -1 if not found.
inline int FindEdge(const I3D_edge *edges, dword num_edges, const I3D_edge &e){
   return FindPointerInArray((void**)edges, num_edges, *(void**)&e);
}

//----------------------------
// Find specified vertex index in face, return index, 
// or -1 if not found.
inline int FindVertexInFace(const I3D_triface &fc, word w){
   for(int i=3; i--; )
      if(fc[i]==w) break;
   return i;
}

//----------------------------
// Determine which vertices are identical (by examining position) with others
//    in the pool, and create mapping array                                    
void MakeVertexMapping(const S_vector *vp, dword pitch,
   int numv, word *v_map, float thresh = .001f);

//----------------------------
// Remove specified edge from list, if it is found, the return value is true.
inline bool RemoveEdge(vector<I3D_edge> &edges, const I3D_edge &e){
   int i = FindEdge(edges.begin(), edges.size(), e);
   if(i==-1) return false;
   edges[i] = edges.back(); edges.pop_back();
   return true;
}

//----------------------------
// Add edge into list, or split with inversed edge (remove those from list),
// return value is true if edge is added, or false if it is split.
inline bool AddOrSplitEdge(vector<I3D_edge> &edge_list, const I3D_edge &e){
   if(RemoveEdge(edge_list, I3D_edge(e[1], e[0])))
      return false;
   edge_list.push_back(e);
   return true;
}

//----------------------------
// Clean mesh - remove unused vertices and remap faces.
void CleanMesh(vector<S_vector> &verts, vector<I3D_triface> &faces);
void CleanMesh(vector<S_vector> &verts, vector<I3D_face> &faces);

//----------------------------
// Create multiple LODs on any vertex type.
void CreateLODsGeneric(float min_vertex_ratio, float min_area_ratio,
   float ratio_change_step, float min_dist, float max_dist,
   PI3D_driver drv, 
   const S_vector *verts, dword vstride, dword num_verts,
   class I3D_source_index_buffer &index_buffer, const vector<I3D_face_group> &fgroups,
   vector<struct C_auto_lod> &out_lods, float out_lod_dists[2],
   byte *out_verts, vector<word> &out_v_map);

//----------------------------
// Make clone of src VB - copy common elements into dst, leave others.
void CloneVertexBuffer(const void *src, dword src_fvf, const void *dst, dword dst_fvf,
   dword num_vertices);

//----------------------------

class I3D_face_group{
public:
   C_smart_ptr<I3D_material> mat;
   dword base_index;
   dword num_faces;

   I3D_face_group(): num_faces(0), base_index(0)
   {}
   I3D_face_group(const I3D_face_group& fg): num_faces(0)
   { *this = fg; }
   const I3D_face_group &operator =(const I3D_face_group &fg){
      num_faces = fg.num_faces;
      base_index = fg.base_index;
      mat = fg.mat;
      return (*this);
   }
public:
   inline dword NumFaces1() const{ return num_faces; }
   inline dword BaseIndex1() const{ return base_index; }
                              //material
   void SetMaterial(PI3D_material mp){ mat = mp; }
   inline PI3D_material GetMaterial1() const{ return mat; }
};

typedef I3D_face_group *PI3D_face_group;

//----------------------------

class I3D_source_vertex_buffer{
   IDirect3DVertexBuffer8 *D3D_vertex_buffer;

   C_smart_ptr<I3D_driver> drv;
   dword num_vertices;        //number of vertices in this buffer
public:
   dword fvflags;
   dword vstride;             //stride of single vertex, in bytes

   int D3D_vertex_buffer_index;
   vector<word> vertex_map;

   I3D_source_vertex_buffer(PI3D_driver drv1, dword fvf1):
      fvflags(fvf1),
      vstride(::GetSizeOfVertex(fvf1)),
      drv(drv1),
      num_vertices(0),
      D3D_vertex_buffer(NULL)
   {}
   ~I3D_source_vertex_buffer(){
      if(D3D_vertex_buffer)
         drv->FreeVertexBuffer(D3D_vertex_buffer, D3D_vertex_buffer_index);
   }
   I3D_source_vertex_buffer(const I3D_source_vertex_buffer &vb)
   {
      operator =(vb);
   }
   I3D_source_vertex_buffer &operator =(const I3D_source_vertex_buffer &vb){
      D3D_vertex_buffer = vb.D3D_vertex_buffer;
      D3D_vertex_buffer_index = vb.D3D_vertex_buffer_index;
      vertex_map = vb.vertex_map;
      num_vertices = vb.num_vertices;
      drv = vb.drv;
      return *this;
   }

   inline int NumVertices() const{ return num_vertices; }
   inline dword GetFVFlags() const{ return fvflags; }
   inline dword GetSizeOfVertex() const{ return vstride; }

   inline IDirect3DVertexBuffer8 *GetD3DVertexBuffer(){ return D3D_vertex_buffer; }

   void *Lock(dword d3d_lock_flags = 0){

      if(!num_vertices)
         return NULL;
      assert(D3D_vertex_buffer);
      void *vp;
      HRESULT hr = D3D_vertex_buffer->Lock(D3D_vertex_buffer_index * vstride,
         num_vertices * vstride,
         (byte**)&vp,
         d3d_lock_flags);
      CHECK_D3D_RESULT("Lock", hr);
      if(FAILED(hr))
         return NULL;
      return vp;
   }
   inline const void *LockConst() const{ return (const void*)((I3D_source_vertex_buffer*)this)->Lock(D3DLOCK_READONLY); }

   bool Unlock(){
      if(!num_vertices)
         return false;
      assert(D3D_vertex_buffer);
      HRESULT hr = D3D_vertex_buffer->Unlock();
      CHECK_D3D_RESULT("Unlock", hr);
      return (SUCCEEDED(hr));
   }

   inline void UnlockConst() const{ ((I3D_source_vertex_buffer*)this)->Unlock(); }

   void AllocVertices(dword num_verts1);

   void SetVertices(const void *src_vp, dword nv){

      AllocVertices(nv);
                              //feed vertices
      void *vp = Lock();
      assert(vp);
      if(!vp)
         return;
      memcpy(vp, src_vp, num_vertices * vstride);
      Unlock();
   }
   template<class T>
   void SetVertices(const vector<T> &vv){
      SetVertices(vv.begin(), vv.size());
   }
   inline const vector<word> &GetVertexSpreadMap() const{ return vertex_map; }
   void SetVertexSpreadMap(const vector<word> &vs_map){ vertex_map = vs_map; }
   void SetVertexSpreadMap(const word *beg, dword num){ vertex_map.assign(beg, beg + num); }

   void DuplicateVertices(word *vi, dword num_verts);
};

//----------------------------

class I3D_source_index_buffer{
   IDirect3DIndexBuffer8 *D3D_index_buffer;

   C_smart_ptr<I3D_driver> drv;
   dword num_faces;

public:
   int D3D_index_buffer_index;      //in sizeof(I3D_triface)

   I3D_source_index_buffer(PI3D_driver);
   I3D_source_index_buffer();
   ~I3D_source_index_buffer();
   I3D_source_index_buffer(const I3D_source_index_buffer &ib){
      operator =(ib);
   }
   I3D_source_index_buffer &operator =(const I3D_source_index_buffer &ib){
      D3D_index_buffer = ib.D3D_index_buffer;
      D3D_index_buffer_index = ib.D3D_index_buffer_index;
      drv = ib.drv;
      num_faces = ib.num_faces;
      return *this;
   }
   inline int NumFaces1() const{ return num_faces; }
   inline LPDIRECT3DINDEXBUFFER8 GetD3DIndexBuffer() const{ return D3D_index_buffer; }

   PI3D_triface Lock(dword d3d_lock_flags = 0);
   inline CPI3D_triface LockConst() const{ return (CPI3D_triface)((I3D_source_index_buffer*)this)->Lock(D3DLOCK_READONLY); }

   const I3D_triface *LockForReadOnly() const;
   bool Unlock() const;
   void AllocFaces(int);
   void SetFaces(const I3D_triface*, dword num_faces);
   void SetFaces(const vector<I3D_triface> &fv){ SetFaces(fv.begin(), fv.size()); }
};

//----------------------------

struct C_auto_lod{
public:
   float ratio;               //quality ratio of this LOD, range 0.0f ... 1.0f
   word vertex_count;
   vector<I3D_face_group> fgroups;
   I3D_source_index_buffer index_buffer;

   C_auto_lod();
   C_auto_lod(PI3D_driver drv): index_buffer(drv){}
   C_auto_lod(const C_auto_lod &al): index_buffer(NULL){ operator =(al); }
   C_auto_lod &operator =(const C_auto_lod &al){
      ratio = al.ratio;
      vertex_count = al.vertex_count;
      fgroups = al.fgroups;
      index_buffer = al.index_buffer;
      return *this;
   }
};

//----------------------------
                              //additional uv mapping channel info
class C_map_channel: public C_unknown{
public:
   vector<I3D_text_coor> uv_verts;
   vector<I3D_triface> uv_faces;
   C_string bitmap_name;
};

//----------------------------
                              //mesh base - class used for higher-level mesh-based sources
                              //it contains:
                              // - vertices and faces (in vertex and index buffers)
                              // - auto-LOD info
class I3D_mesh_base{
protected:
   dword ref;

   PI3D_driver drv;

                              //face groups
   vector<I3D_face_group> fgroups;
                              //automatic LOD
   vector<C_auto_lod> auto_lods;
                              //distance range at which computed LODs are distributed,
                              // valid at camera's FOV = 65 degrees, otherwise scaled
   float auto_lod_dists[2];   //{min, max}

                              //bounding volume of all vertices;
   I3D_bound_volume bvolume;

   C_string smooth_groups;
   C_smart_ptr<C_map_channel> map_channel_2;
public:
   I3D_mesh_base(PI3D_driver d1, dword fvf1);
   ~I3D_mesh_base();

   inline const byte *GetSmoothGroups() const{
      return !smooth_groups.Size() ? NULL :
         (const byte*)(const char*)smooth_groups;
   }
   void SetSmoothGroups(const vector<byte> &sg){
      if(sg.size()){
         assert(sg.size() == NumFaces1());
         smooth_groups.Assign((const char*)sg.begin(), sg.size());
      }else{
         smooth_groups = NULL;
      }
   }

   inline const C_map_channel *GetMapChannel2() const{ return map_channel_2; }

   I3D_source_vertex_buffer vertex_buffer;
   I3D_source_index_buffer index_buffer;

   inline dword NumVertices1() const{ return vertex_buffer.NumVertices(); }
   inline dword NumFaces1() const{ return index_buffer.NumFaces1(); }

   inline dword NumFGroups1() const{ return fgroups.size(); }
   inline const I3D_face_group *GetFGroups1() const{ return &fgroups[0]; }
   inline const vector<I3D_face_group> &GetFGroupVector() const{ return fgroups; }
   inline void SetFGroups(const I3D_face_group *fgps, dword num_fgroups){
      fgroups.assign(fgps, fgps+num_fgroups);
   }

   inline const vector<C_auto_lod> &GetAutoLODs() const{ return auto_lods; }
   inline const float *GetAutoLODDists() const{ return auto_lod_dists; }
   inline const I3D_bound_volume &GetBoundingVolume() const{ return bvolume; }

//----------------------------
// Create automatic levels of detail.
   void CreateLODs(float min_vertex_ratio, float min_area_ratio, float ratio_change_step,
      float min_dist, float max_dist);

//----------------------------
// Optimize mesh for faster rendering.
   void Optimize(const struct S_load_context &lc);

//----------------------------
// Constructor for cloning, using artibrary source mesh.
   void Clone(CPI3D_mesh_base mb);

//----------------------------
// Setup cloned verts.
   void SetClonedVertices(vector<byte> &new_vertices){
      vertex_buffer.SetVertices(new_vertices.begin(), new_vertices.size()/vertex_buffer.vstride);
   }

//----------------------------
// Compute bounding volume of mesh.
   void ComputeBoundingVolume();

public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
                              //vertices
   I3DMETHOD_(dword,NumVertices)() const{ return vertex_buffer.NumVertices(); }
   I3DMETHOD_(const S_vector*,LockVertices)(){ return (const S_vector*)vertex_buffer.Lock(); }
   I3DMETHOD_(void,UnlockVertices)(){ vertex_buffer.Unlock(); }
   I3DMETHOD_(dword,GetSizeOfVertex)() const{ return vertex_buffer.GetSizeOfVertex(); }
                              //faces and groups
   I3DMETHOD_(dword,NumFGroups)() const{ return fgroups.size(); }
   I3DMETHOD_(dword,NumFaces)() const{ return index_buffer.NumFaces1(); }
   I3DMETHOD_(PI3D_face_group,GetFGroup)(int indx){ return &fgroups[indx]; }
   I3DMETHOD_(CPI3D_face_group,GetFGroup)(int indx) const{ return &fgroups[indx]; }
   I3DMETHOD_(const I3D_triface*,LockFaces)(){ return index_buffer.Lock(); }
   I3DMETHOD_(void,UnlockFaces)(){ index_buffer.Unlock(); }

   //I3DMETHOD_(void,VisReserved00)() = 0;
   virtual bool SaveCachedInfo(C_cache *ck, word version = 0) const;
   //I3DMETHOD_(void,VisReserved01)() = 0;
   virtual bool LoadCachedInfo(C_cache *ck, S_load_context &lc, word version = 0);
   I3DMETHOD_(void,VisReserved02)(){}
   I3DMETHOD_(void,VisReserved03)(){}
   I3DMETHOD_(void,VisReserved04)(){}
   I3DMETHOD_(void,VisReserved05)(){}
   I3DMETHOD_(void,VisReserved06)(){}
   I3DMETHOD_(void,VisReserved07)(){}
};

//----------------------------
                              //one level of mesh detail
class I3D_mesh_level: public I3D_mesh_base{

#define MESHF_PIVOT_AND_NU_APPLIED  8  //pivot and non-uniform scale applied to mesh

   dword mesh_flags;

   float lod_factor;

   I3D_mesh_level(PI3D_driver);

   friend class I3D_driver;   //creation
   friend class I3D_mesh_object;

public:
   inline float GetLODFactor() const{ return lod_factor; }
   inline void SetMapChannel2(C_map_channel *mc){ map_channel_2 = mc; }

//----------------------------
// Apply pivot offset and non-uniform scale to vertices.
   void ApplyPivotAndNUScale(const S_vector &p, const S_vector &s);

//----------------------------
// Save/Load mesh - vertices and faces after optimization, with vertex normals.
   virtual bool SaveCachedInfo(class C_cache*, word = 0) const;
   virtual bool LoadCachedInfo(class C_cache*, struct S_load_context &lc, word = 0);

public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(Build2)(const I3D_vertex_mesh *verts, dword num_verts,
      const I3D_triface *faces, dword num_faces,
      const I3D_face_group *fgroups, dword num_fgroups, dword flags = 0);
   I3DMETHOD(Duplicate2)(CPI3D_mesh_level);
                              
                              //will go out - no access to geometry from outside
   I3DMETHOD_(const I3D_vertex_mesh*,LockVertices2)(){ return (const I3D_vertex_mesh*)vertex_buffer.Lock(); }
   I3DMETHOD_(void,UnlockVertices2)(){ vertex_buffer.Unlock(); }
};

//----------------------------
//#define MAX_LEVELS_OF_DETAIL  12

class I3D_mesh_object{
protected:
   dword ref;
private:
                              //vertices reset - valid only between call to Lock/Unlock methods
#define MOF_LOCK_VERTS_RESET_XYZ       2
#define MOF_LOCK_VERTS_RESET_LIGHTING  4
#define MOF_LOCK_VERTS_RESET_UV        8

   C_smart_ptr<I3D_mesh_level> lod_mesh;

   PI3D_driver drv;
   dword mo_flags;

   friend I3D_driver;
   I3D_mesh_object(PI3D_driver);
public:

   ~I3D_mesh_object();

   inline PI3D_mesh_level GetLOD1() const{ return lod_mesh; }

public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(PI3D_mesh_level,GetLOD)(int indx) const{
      if(indx!=0)
         return NULL;
      return lod_mesh;
   }
   I3DMETHOD(SetLOD)(PI3D_mesh_level, int indx = 0);
   I3DMETHOD_(dword,NumLevels)() const{ return 1; }
   I3DMETHOD(Duplicate)(CPI3D_mesh_object);

   I3DMETHOD_(const I3D_bound_volume&,GetBoundVolume)() const;

   I3DMETHOD_(dword,NumVertices)(int lod_index = 0) const;
   I3DMETHOD_(PI3D_vertex_mesh,LockVertices)(int lod_index = 0, dword flags = 0);
   I3DMETHOD(UnlockVertices)(int lod_index = 0);

   I3DMETHOD_(PI3D_triface,LockFaces)(int lod_index = 0, dword flags = I3DLOCK_INDEX);
   I3DMETHOD(UnlockFaces)(int lod_index = 0);
};

//----------------------------

#endif
