/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Mirror.cpp
   Content: Mirror frame.
--------------------------------------------------------*/

#include "all.h"
#include "object.h"
#include "camera.h"

//----------------------------
//----------------------------

                              //debugging:
//#define DEBUG_SPECIAL_BACKGROUND //fill mirror's background to special color

//----------------------------
//----------------------------

class I3D_object_mirror: public I3D_object{
   S_plane loc_plane, world_plane;
   float depth;
   bool loc_plane_valid;
   float alpha;
   dword alpha_factor;

public:
   I3D_object_mirror(PI3D_driver);

   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive &rp);
   void ComputeLocalPlane();
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
   I3DMETHOD(SetProperty)(dword index, dword value);
   I3DMETHOD_(dword,GetProperty)(dword index) const;

   inline const S_plane &GetPlane() const{ return world_plane; }
};

//----------------------------

I3D_object_mirror::I3D_object_mirror(PI3D_driver d):
   I3D_object(d),
   depth(0.0f),
   alpha(.5f),
   loc_plane_valid(false)
{
   visual_type = I3D_VISUAL_MIRROR;
}

//----------------------------

I3D_RESULT I3D_object_mirror::SetProperty(dword index, dword value){

   switch(index){
   case 0: depth = IntAsFloat(value); break;
   case 1: alpha = IntAsFloat(value);
      alpha_factor = Max(0, Min(255, FloatToInt(alpha*255.0f)));
      break;
   default:
      return I3DERR_INVALIDPARAMS;
   }
   return I3D_OK;
}

//----------------------------

dword I3D_object_mirror::GetProperty(dword index) const{

   switch(index){
   case 0: return FloatAsInt(depth);
   case 1: return FloatAsInt(alpha);
   }
   return 0;
}

//----------------------------

void I3D_object_mirror::ComputeLocalPlane(){

   if(!mesh)
      return;

   PI3D_mesh_base mb = GetMesh();
   assert(mb);
                              //time to compute the plane
   const void *vp = mb->vertex_buffer.Lock(D3DLOCK_READONLY);
   dword vstride = mb->vertex_buffer.GetSizeOfVertex();
   PI3D_triface faces = mb->index_buffer.Lock(D3DLOCK_READONLY);
   const S_vector *v[3] = {
      (const S_vector*)(((byte*)vp) + vstride * faces[0][0]),
      (const S_vector*)(((byte*)vp) + vstride * faces[0][1]),
      (const S_vector*)(((byte*)vp) + vstride * faces[0][2])
   };

   loc_plane.normal.GetNormal(*v[0], *v[1], *v[2]);
   loc_plane.normal.Normalize();
   loc_plane.d = -(*v[0]).Dot(loc_plane.normal);

   mb->index_buffer.Unlock();
   mb->vertex_buffer.Unlock();

   loc_plane_valid = true;
}

//----------------------------

extern const dword vs_decl_vertex_mesh[];

//----------------------------

void I3D_object_mirror::AddPrimitives(S_preprocess_context &pc){

   if(pc.mode==RV_SHADOW_CASTER)
      return;

                              //if this mirror recursively renders, skip it
   if(pc.curr_mirror == this)
      return;

   PI3D_mesh_base mb = GetMesh();
   if(!mb)
      return;

   bool allow_mirrors = ((drv->GetFlags()&DRVF_HASSTENCIL) &&
      (drv->GetFlags()&DRVF_DRAWMIRRORS));

   bool in_mirror_mode = (pc.curr_mirror!=NULL);

   if(!in_mirror_mode && allow_mirrors){
      if(!loc_plane_valid){
         ComputeLocalPlane();
         assert(loc_plane_valid);
      }
      world_plane = loc_plane * matrix;
                           //check if it is visible
      float d = world_plane.normal.Dot(world_plane.normal);
      if(d <= 0.0f)
         return;
   }

   const I3D_face_group &fg = mb->GetFGroupVector().front();
   dword vertex_count = mb->vertex_buffer.NumVertices();
   assert(vertex_count);
   word face_count = fg.NumFaces1();
   drv->render_stats.triangle += face_count;
   drv->render_stats.vert_trans += vertex_count;
   PI3D_material mat = fg.GetMaterial1();

   pc.prim_list.push_back(S_render_primitive(0));
   S_render_primitive &p = pc.prim_list.back();
   
   p.vis = this;
   p.clip = pc.clip;
   p.sector = pc.curr_sector;
   p.scene = pc.scene;
   p.user = (!in_mirror_mode && allow_mirrors);
                              
   p.sort_value = PRIM_SORT_OPAQUE;
   ++pc.opaque;
   p.sort_value |= (mat->GetSortID()&PRIM_SORT_MAT_MASK)<<PRIM_SORT_MAT_SHIFT;

   if(!vertex_buffer.GetD3DVertexBuffer())
      PrepareDestVB(mb);
}

//----------------------------

void I3D_object_mirror::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   HRESULT hr;
   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

   PI3D_mesh_base mb = mesh->GetLOD1();
   const I3D_face_group *fg = &mb->GetFGroupVector()[0];
   PI3D_material mat = fg->GetMaterial1();

   bool active_mirror = rp.user;

   if(active_mirror){
      bool is_wire = (drv->GetFlags()&DRVF_WIREFRAME);
      if(is_wire) lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
                              //render to stencil buffer only
      I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);
      PrepareVertexShader(mat, se, rp, VSPREP_TRANSFORM | VSPREP_FEED_MATRIX);
      
      drv->SetupBlend(I3DBLEND_OPAQUE);
      drv->SetupTextureStage(0, D3DTOP_SELECTARG1);

      drv->SetTexture1(0, NULL);
      drv->DisableTextureStage(1);

      drv->SetClipping(rp.clip);
      drv->EnableFog(false);

      drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), mb->vertex_buffer.GetSizeOfVertex());
      hr = lpDev8->SetIndices(mb->index_buffer.GetD3DIndexBuffer(),
         mb->vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);

                              //setup D3D to clear color-buffer to background color
                              // (for mirror rendering),
                              // as well as write to stencil buffer
      lpDev8->SetRenderState(D3DRS_STENCILENABLE, true);
      lpDev8->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_ALWAYS);
      lpDev8->SetRenderState(D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP);
      lpDev8->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
      lpDev8->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_REPLACE);
      lpDev8->SetRenderState(D3DRS_STENCILREF, 1);

      dword clear_color = pc.scene->GetBgndClearColor();
#ifdef DEBUG_SPECIAL_BACKGROUND
      clear_color = 0xffff0000;
#endif
      drv->SetTextureFactor(clear_color);

      drv->SetupTextureStage(0, D3DTOP_SELECTARG1);
      lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
                              //render bits into stencil buffer,
                              // no z-writes
      bool zw_enabled = drv->IsZWriteEnabled();
      if(zw_enabled) drv->EnableZWrite(false);
      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,
         mb->vertex_buffer.NumVertices(),
         (mb->index_buffer.D3D_index_buffer_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);
      if(zw_enabled) drv->EnableZWrite(true);

                              //setup stencil-testing
      lpDev8->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_EQUAL);
      lpDev8->SetRenderState(D3DRS_STENCILREF, 1);
      lpDev8->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);

                              //clear the z-buffer in mirrored area -
                              // draw pixels only into stenciled area
                              // with zero z and w set to far clipping plane
      {
         lpDev8->SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);

         const int *viewport = pc.scene->GetViewport1();
         float rhw = 1.0f / pc.scene->GetActiveCamera1()->GetFCP();
         S_vectorw v[4];
         for(int i=0; i<4; i++){
            v[i].w = rhw;
            v[i].z = .999f;
            v[i].x = viewport[(i&1) << 1];
            v[i].y = viewport[(i&2) + 1];
         }
         static const word indx[] = {0, 1, 2, 2, 1, 3};

         drv->SetVertexShader(D3DFVF_XYZRHW);
         hr = lpDev8->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, v, sizeof(S_vectorw));
         CHECK_D3D_RESULT("DrawPrimitiveUP", hr);
         drv->ResetStreamSource();
         lpDev8->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
      }

      lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);

      lpDev8->SetRenderState(D3DRS_STENCILENABLE, false);
      lpDev8->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);
      if(is_wire) lpDev8->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

                              //render mirrored image now
                              //setup camera to look from behind the mirror
      PI3D_camera cam = drv->CreateCamera();
      C_smart_ptr<I3D_camera> cam_save = pc.scene->GetActiveCamera1();
      cam->SetFOV(cam_save->GetFOV1());
      cam->SetRange(cam_save->GetNCP(), cam_save->GetFCP());
      pc.scene->SetActiveCamera(cam);
      cam->SetPos(cam_save->GetWorldPos());
      cam->SetRot(cam_save->GetWorldRot());
      cam->Update();
      cam->Mirror(world_plane);
      for(PI3D_frame sct = this; sct = sct->GetParent(); ){
         if(sct->GetType1()==FRAME_SECTOR){
            cam->SetCurrSector(I3DCAST_SECTOR(sct));
            break;
         }
      }
      drv->SetDefaultCullMode(D3DCULL_CW);

      lpDev8->SetRenderState(D3DRS_STENCILENABLE, true);
      lpDev8->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_EQUAL);
      lpDev8->SetRenderState(D3DRS_STENCILREF, 1);

      S_plane add_planes[2];
      dword num_add_planes = 1;
      add_planes[0] = -world_plane;
      if(!IsMrgZero(depth)){
         add_planes[1] = world_plane;
         add_planes[1].d -= depth;
         ++num_add_planes;
      }
      pc.scene->RenderView(pc.render_context, pc.render_flags, RV_MIRROR, this, add_planes, num_add_planes);

      drv->SetDefaultCullMode(D3DCULL_CCW);
      pc.scene->SetActiveCamera(cam_save);

      pc.scene->SetMViewProj(cam_save->GetViewMatrix1(), cam_save->GetProjectionMatrix1());
      cam->Release();
                              //setup stencil to clear to zero again
      lpDev8->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_EQUAL);
      lpDev8->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_DECR);
   }
   {
      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX | VSPREP_MAKELIGHTING | VISF_DEST_UV0_VALID;

      I3D_driver::S_vs_shader_entry_in se(vs_decl_vertex_mesh);
      PrepareVertexShader(mat, se, rp, prep_flags);
      drv->SetStreamSource(mb->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_mesh));
      drv->SetClipping(rp.clip);

      hr = lpDev8->SetIndices(mb->index_buffer.GetD3DIndexBuffer(),
         mb->vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);

                              //mirrors disabled, use standard path
      drv->SetRenderMat(mat);
      dword blend_mode = I3DBLEND_OPAQUE;
      if(active_mirror){
         blend_mode |= I3DBLEND_ALPHABLEND | I3DBLEND_ALPHAFACTOR;
         if(mat->IsDiffuseAlpha())
            blend_mode |= I3DBLEND_VERTEXALPHA;
      }
      drv->SetupBlend(blend_mode);
      drv->SetupTextureStage(0, drv->vertex_light_blend_op);
      drv->SetTextureFactor(alpha_factor<<24);
      drv->EnableAnisotropy(0, true);
                              //apply environment effect
      E_ENVIROMENTAL_EFFECT ee = mat->GetEnvironmetEffect();
      if(ee!=EE_NULL && envmap_uv_set!=-1)
         SetupEnvMapping(ee, mat, envmap_uv_set);
      else
         drv->DisableTextureStage(1);

      if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES))
         drv->SetTexture1(0, NULL);

      drv->EnableFog(vis_flags&VISF_FOG_VALID);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, mb->vertex_buffer.NumVertices(),
         (mb->index_buffer.D3D_index_buffer_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);
   }
                              //finally, disable stenciling
   if(active_mirror)
      lpDev8->SetRenderState(D3DRS_STENCILENABLE, false);
}

//----------------------------

static const S_visual_property props[] = {
   I3DPROP_FLOAT, "Depth",
   I3DPROP_FLOAT, "Alpha",
};

static I3D_visual *CreateMirror(PI3D_driver drv){
   return new I3D_object_mirror(drv);
}

static struct S_auto_init_mirror{
   S_auto_init_mirror(){
      RegisterVisual(CreateMirror, I3D_VISUAL_MIRROR, "Mirror",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_mirror;

//----------------------------
//----------------------------


