/*--------------------------------------------------------
   Copyright (c) 1999, 2000 Michal Bacik. 
   All rights reserved.

   File: Model.cpp
   Content: Model frame.
--------------------------------------------------------*/

#include "all.h"
#include "model.h"
#include "particle.h"
#include "scene.h"

//----------------------------
                              //virtual table for container callback to model
static const S_container_vcall vc_container_model = {
   (S_container_vcall::t_Open)I3D_model::Open,
   (S_container_vcall::t_Close)I3D_model::Close,
   (S_container_vcall::t_AddFrame)I3D_model::AddFrame,
   (S_container_vcall::t_DeleteFrame)I3D_model::DeleteFrame,
};

//----------------------------
                              //model class
I3D_model::I3D_model(PI3D_driver d):
   I3D_frame(d),
   in_container(NULL),
   shadow_info(NULL)
{
   drv->AddCount(I3D_CLID_MODEL);
   type = FRAME_MODEL;
   enum_mask = ENUMF_MODEL;

   cont.master_interface = (C_unknown*)this;
   cont.vcall = &vc_container_model;
}

//----------------------------

I3D_model::~I3D_model(){

   Close();
   delete shadow_info;
   drv->DecCount(I3D_CLID_MODEL);
}

//----------------------------

void I3D_model::Close(){

                              //release all our frames
   while(cont.frames.size())
      DeleteFrame(cont.frames.back());
                              //unlink all child frames
   while(children.size())
      children.back()->LinkTo(NULL);
   anim_set = NULL;
}

//----------------------------

static void modelSetSingleName(const C_string &name, PI3D_frame fp){

   const C_string &str = fp->GetName1();
   int j=str.Size();
   const char *base;
   if(j){
      while(j && str[j-1]!='.') --j;
      base = &str[j];
   }else base="";
   fp->SetName(C_fstring("%s.%s", (const char*)name, base));
}

//----------------------------

I3D_RESULT I3D_model::AddFrame(PI3D_frame frm){

   switch(frm->GetType1()){
   case FRAME_VISUAL:
      if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE) break;
      cont.AddAnimFrame(frm);
      if(in_container)
         in_container->AddAnimFrame(frm);
      break;
   case FRAME_MODEL:
      {
                              //check if it's already in some other scene
         PI3D_model mod = I3DCAST_MODEL(frm);
         if(mod->in_container){
                              //can't be in 2 scenes at the same time
            return I3DERR_GENERIC;
         }
         mod->in_container = &cont;
                              //if it contains animation, create interpolator
         if(mod->cont.NumInterpolators1()){
            cont.AddAnimFrame(frm);
            if(in_container)
               in_container->AddAnimFrame(frm);
         }
      }
      break;
   }
   cont.frames.push_back(frm);
   frm->frm_flags |= FRMFLAGS_SUBMODEL;
   modelSetSingleName(name, frm);

   return I3D_OK;
}

//----------------------------

static void modelUnsetSingleName(PI3D_frame fp){

   const C_string &str=fp->GetName1();
   int j=str.Size();
   while(j && str[j-1]!='.') --j;
   fp->SetName(&str[j]);
}

//----------------------------

I3D_RESULT I3D_model::DeleteFrame(PI3D_frame frm){

   for(int i=0; i<cont.frames.size(); i++)
   if(cont.frames[i]==frm){
      frm->AddRef();          //released later
      cont.frames[i] = cont.frames.back(); cont.frames.pop_back();

      switch(frm->GetType1()){
      case FRAME_VISUAL:
         if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE)
            break;
         cont.DeleteAnimFrame(frm);
         if(in_container && !cont.NumInterpolators1())
            in_container->DeleteAnimFrame(frm);
         break;

      case FRAME_MODEL:
         {
            PI3D_model mod = I3DCAST_MODEL(frm);
            if(mod->in_container==&cont){
               if(mod->cont.NumInterpolators1()){
                  cont.DeleteAnimFrame(frm);
                  if(in_container && !cont.NumInterpolators1())
                     in_container->DeleteAnimFrame(frm);
               }
               mod->in_container = NULL;
            }
         }
         break;
      }
      frm->frm_flags &= ~FRMFLAGS_SUBMODEL;
      if(frm->Release()){     //don't change name on destroyed frames
         modelUnsetSingleName(frm);
      }
      return I3D_OK;
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------

void I3D_model::SetName(const char *cp){

   I3D_frame::SetName(cp);
   for(int i=cont.frames.size(); i--; )
      modelSetSingleName(name, cont.frames[i]);
}

//----------------------------

static void cbModelDup(PI3D_model orig, PI3D_model newm, PI3D_driver drv){

   int i;

   const vector<C_smart_ptr<I3D_frame> > &chld = orig->GetContainer1()->GetFrameVector();
                              //sort list by number of partents to root
   C_sort_list<int> slist;
   for(i=chld.size(); i--; ){
      PI3D_frame frm = chld[i];
      for(int nump=0; frm=frm->GetParent(); nump++ );
      slist.Add(i, nump);
   }
   slist.Sort();

   map<PI3D_frame, PI3D_frame> link_map;
   map<PI3D_frame, PI3D_frame>::iterator it;

   link_map[orig] = newm;

   for(i=0; i<chld.size(); i++){
      int si = slist[i];
      PI3D_frame orig = chld[si];
      PI3D_frame dupc = NULL;

      switch(orig->GetType1()){
      case FRAME_VISUAL:
         dupc = drv->CreateVisual(I3DCAST_VISUAL(orig)->GetVisualType());
         break;

      case FRAME_SECTOR:
         break;

      case FRAME_SOUND:
         {
            PI3D_sound snd = I3DCAST_SOUND(orig);
            dupc = snd->GetScene()->CreateSound();
         }
         break;

      default:
         dupc = drv->CreateFrame(orig->GetType1());
         assert(dupc);
      }
      if(dupc){
         it = link_map.find(orig->GetParent());
         if(it!=link_map.end())
            dupc->LinkTo((*it).second);

         newm->AddFrame(dupc);
         dupc->Release();

         link_map[orig] = dupc;
      }
   }
   for(i=chld.size(); i--; ){
      int si = slist[i];
      PI3D_frame orig = chld[si];
      PI3D_frame dupc = link_map[orig];
      dupc->Duplicate(orig);
      dupc->SetFrameFlags(dupc->GetFrameFlags()|FRMFLAGS_SUBMODEL);
      dupc->Update();
   }
}

//----------------------------

I3D_RESULT I3D_model::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()==FRAME_MODEL){
      PI3D_model mp = I3DCAST_MODEL(frm);

      Close();

      cont.filename = mp->cont.filename;

      cbModelDup(mp, this, drv);
      anim_set = mp->anim_set;
                              //must set flags this way - special processing inside
      SetFlags(frm->GetFlags(), -1);
   }
   return I3D_frame::Duplicate(frm);
}

//----------------------------
typedef map<C_string, PI3D_frame> t_fmap;

I3D_RESULT I3D_model::SetAnimation(dword stage, PI3D_animation_set ap,
   dword options, float scale_factor, float blend_factor, float speed){

   if(!ap)
      ap = anim_set;
   if(!ap)
      return I3DERR_INVALIDPARAMS;

   StopAnimation(stage);
                              //allow assigning animations only on frames in our container
   t_fmap frm_map;
   const vector<C_smart_ptr<I3D_frame> > &own_frames = cont.GetFrameVector();
   for(int i=own_frames.size(); i--; ){
      PI3D_frame frm = own_frames[i];
      frm_map[frm->GetOrigName()] = frm;
   }

   const I3D_animation_set::S_anim_link *lp = ap->GetAnimLinks();
   for(i=0; i<ap->NumLinks(); i++){
      const I3D_animation_set::S_anim_link &al = lp[i];
      t_fmap::const_iterator it = frm_map.find(al.link);
      if(it!=frm_map.end()){
         PI3D_frame frm = (*it).second;
                              //find interpolator of this frame, if it has one in our list
         PI3D_interpolator aip;
         const PI3D_interpolator *intps = cont.GetInterpolators1();
         for(int j = cont.NumInterpolators1(); j--; )
         if(intps[j]->GetFrame1()==frm && intps[j]->IsPRS()){
            aip = intps[j];
            break;
         }
         if(j==-1){
                              //we don't have this frame in list,
                              // create interpolator for it now.
            aip = drv->CreateInterpolator();
            aip->SetFrame(frm);
            cont.AddInterpolator(aip);
            aip->Release();
         }
         aip->SetAnimation(stage, al.anim, options, scale_factor, blend_factor, speed);
      }
   }
   if(in_container && cont.NumInterpolators1())
      in_container->AddAnimFrame(this);
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_model::SetAnimScaleFactor(dword stage, float scale_factor){

   vector<C_smart_ptr<I3D_interpolator> > iv = cont.GetInterpolatorVector();
   for(int i=cont.NumInterpolators1(); i--; ){
      PI3D_interpolator intp = iv[i];
      intp->SetScaleFactor(stage, scale_factor);
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_model::SetAnimBlendFactor(dword stage, float blend_factor){

   vector<C_smart_ptr<I3D_interpolator> > iv = cont.GetInterpolatorVector();
   for(int i=cont.NumInterpolators1(); i--; ){
      PI3D_interpolator intp = iv[i];
      intp->SetBlendFactor(stage, blend_factor);
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_model::SetAnimSpeed(dword stage, float speed_mult){

   vector<C_smart_ptr<I3D_interpolator> > iv = cont.GetInterpolatorVector();
   for(int i=cont.NumInterpolators1(); i--; ){
      PI3D_interpolator intp = iv[i];
      intp->SetSpeed(stage, speed_mult);
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_model::StopAnimation(dword stage){ 

   if(cont.NumInterpolators1()){
      vector<C_smart_ptr<I3D_interpolator> > iv = cont.GetInterpolatorVector();
      for(int i=cont.NumInterpolators1(); i--; ){
         PI3D_interpolator intp = iv[i];
         if(!intp->IsPRS())
            continue;
         intp->SetAnimation(stage, NULL);
         if(!intp->NumActiveStages())
            cont.DeleteInterpolator(intp);
      }
      if(in_container && !cont.NumInterpolators1())
         in_container->DeleteAnimFrame(this);
   }

   return I3D_OK; 
};

//----------------------------

dword I3D_model::SetFlags(dword new_flags, dword flags_mask){

   if(flags_mask&I3D_FRMF_SHADOW_CAST){
      if((frm_flags^new_flags)&I3D_FRMF_SHADOW_CAST){
         if(frm_flags&I3D_FRMF_SHADOW_CAST){
            assert(shadow_info);
            delete shadow_info;
            shadow_info = NULL;
         }else{
            assert(!shadow_info);
            shadow_info = new S_shadow_info;
         }
      }
   }
   return I3D_frame::SetFlags(new_flags, flags_mask);
}

//----------------------------

const char *I3D_model::GetFileName() const{ return cont.GetFileName(); }
void I3D_model::StoreFileName(const char *fn){ cont.StoreFileName(fn); }
dword I3D_model::NumFrames() const{ return cont.NumFrames1(); }
const PI3D_frame *I3D_model::GetFrames() const{ return cont.GetFrames(); }
I3D_RESULT I3D_model::AddInterpolator(PI3D_interpolator i){ return cont.AddInterpolator(i); }
I3D_RESULT I3D_model::DeleteInterpolator(PI3D_interpolator i){ return cont.DeleteInterpolator(i); }
dword I3D_model::NumInterpolators() const{ return cont.NumInterpolators1(); }
const PI3D_interpolator *I3D_model::GetInterpolators() const{ return cont.GetInterpolators(); }

I3D_RESULT I3D_model::Tick(int time){
   return cont.Tick(time);
}

//----------------------------
//----------------------------

