/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Morph.cpp
   Content: Morph frame.
--------------------------------------------------------*/

#include "all.h"
#include "mesh.h"
#include "procedural.h"
#include "visual.h"
#include "light.h"
#include "camera.h"
#include "scene.h"
#include <malloc.h>

//----------------------------
//----------------------------

bool IsPointInBBox(const S_vector &v, const I3D_bbox &bbox, const S_matrix &m,
   float &weight);

//----------------------------

                              //use 2 weights, one for applying morph ratio,
                              // the other for source selection (index)
#define FVF_VERTEX_MORPH (D3DFVF_XYZB2 | D3DFVF_NORMAL | D3DFVF_TEX1)

struct I3D_vertex_morph{
   S_vector xyz;
   float weight;
   float index;
   S_vector normal;
   I3D_text_coor tex[1];
};

//----------------------------
                              //morphing mesh
class I3D_morph_mesh: public I3D_mesh_base{

#define MMSHF_RESET_VERTICES     1     //reset weight info in vertices

   dword mesh_flags;

   struct S_region{
      I3D_bbox bbox;
      S_matrix tm;
   };
   vector<S_region> regions;

public:
   I3D_morph_mesh(PI3D_driver d):
      I3D_mesh_base(d, FVF_VERTEX_MORPH),
      mesh_flags(0)
   {}

                              //source of morphing
   C_smart_ptr<I3D_procedural_morph> procedural;
//----------------------------
// Save/Load mesh - vertices and faces after optimization, vertex normals,
// bone segments.
   //virtual bool SaveCachedInfo(class C_cache*, word = 0) const;
   //virtual bool LoadCachedInfo(class C_cache*, struct S_load_context &lc, word = 0);

//----------------------------
// Attach procedural object.
   bool AttachProcedural(PI3D_procedural_base pb){

                                 //check if procedural is of expected type
      if(pb->GetID() == PROCID_MORPH_3D){
         procedural = (I3D_procedural_morph*)pb;
         mesh_flags |= MMSHF_RESET_VERTICES;
                                 //init vertices - setup weights
         return true;
      }
      return false;
   }

   inline I3D_procedural_morph *GetProcedural() const{ return procedural; }

//----------------------------
// Add region into list.
   bool AddRegion(const I3D_bbox &bb, const S_matrix &m){

      regions.push_back(S_region());
      S_region &rg = regions.back();
      rg.bbox = bb;
      rg.tm = m;
      return true;
   }

   inline int NumRegions() const{ return regions.size(); }

//----------------------------
// Prepare morphing - setup weights, etc. This is done before 1st rendering,
// only if MMSHF_RESET_VERTICES is set.
   void PrepareMorphing(){

      dword num_verts = NumVertices1();
      vector<word> v_map(num_verts);

      I3D_vertex_morph *verts = (I3D_vertex_morph*)vertex_buffer.Lock();

      MakeVertexMapping(&verts->xyz, sizeof(I3D_vertex_morph), num_verts, v_map.begin());

      if(!procedural){
                                 //just reset to zero
         for(int i=num_verts; i--; ){
            I3D_vertex_morph &v = verts[i];
            v.weight = 0.0f;
            v.index = 0.0f;
         }
      }else{
         int num_channels = procedural->GetNumChannels();
         for(int i=0; i<num_verts; i++){
            I3D_vertex_morph &v = verts[i];
            if(v_map[i]==i){
                                 //original vertex
               if(regions.size()){
                  v.weight = 0.0f;
                                    //check if vertex is in any region
                  for(int ri=regions.size(); ri--; ){
                     if(IsPointInBBox(v.xyz, regions[ri].bbox, regions[ri].tm, v.weight))
                        break;
                  }
               }else{
                  v.weight = 1.0f;
               }
               v.index = S_int_random(num_channels);
            }else{
                                 //shared vertex
               const I3D_vertex_morph &vs = verts[v_map[i]];
               v.weight = vs.weight;
               v.index = vs.index;
            }
         }     
      }
      vertex_buffer.Unlock();
   }

public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

//----------------------------
//----------------------------

class I3D_object_morph: public I3D_visual{
   C_smart_ptr<I3D_morph_mesh> mesh;

   enum E_MODE{
      MODE_SINGLE_CHANNEL,    //all verts morphed by single channel (#0)
      MODE_PER_REGION_CHANNEL,//each region morphed by random channel
      MODE_PER_VERTEX_CHANNEL,//each vertex morphed by random channel
      MODE_LAST
   } mode;

   int channel;               //procedural channel we're using if MODE_SINGLE_CHANNEL
   float power;
public:
   I3D_object_morph(PI3D_driver d):
      I3D_visual(d),
      power(1.0f),
      mode(MODE_SINGLE_CHANNEL),
      channel(-1)
   {
      visual_type = I3D_VISUAL_MORPH;
   }

   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD_(I3D_mesh_base*,GetMesh)() const{ return mesh; }
   I3DMETHOD(SetProperty)(dword index, dword value);
   I3DMETHOD_(dword,GetProperty)(dword index) const;

   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   virtual bool ComputeBounds();

//----------------------------

   virtual bool AttachProcedural(PI3D_procedural_base pb){

      if(mesh && pb->GetID()==PROCID_MORPH_3D){
         if(!mesh->AttachProcedural(pb))
            return false;
         channel = S_int_random(mesh->procedural->GetNumChannels());
         return true;
      }
      return I3D_visual::AttachProcedural(pb);
   }

//----------------------------

   virtual bool AddRegion(const I3D_bbox &bb, const S_matrix &m, int index){

      if(mesh)
         return mesh->AddRegion(bb, m);
      return false;
   }

//----------------------------

   virtual void PrepareDestVB(I3D_mesh_base *mb, dword num_txt_sets){
      mesh->PrepareMorphing();
      I3D_visual::PrepareDestVB(mb, num_txt_sets);
   }
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

//----------------------------
//----------------------------

bool I3D_object_morph::ComputeBounds(){

   if(!mesh){
      bound.bound_local.bbox.Invalidate();
      bound.bound_local.bsphere.pos.Zero();
      bound.bound_local.bsphere.radius = 0.0f;
   }else{
      mesh->ComputeBoundingVolume();
      bound.bound_local = mesh->GetBoundingVolume();
                              //expand by maximum of morphing radius
      if(mesh->procedural){
         S_vector proc_scale = mesh->procedural->GetMaxMorphScale();
         I3D_bbox &bbox = bound.bound_local.bbox;
         bbox.min -= proc_scale;
         bbox.max += proc_scale;
         bound.bound_local.bsphere.radius += proc_scale.Magnitude();
      }
   }
   return I3D_visual::ComputeBounds();
}

//----------------------------

I3D_RESULT I3D_object_morph::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);

   PI3D_visual vis = I3DCAST_VISUAL(frm);
   switch(vis->GetVisualType()){
   case I3D_VISUAL_MORPH:
      mesh = (I3D_morph_mesh*)(((I3D_object_morph*)frm)->GetMesh());
      channel = mesh->procedural ? S_int_random(mesh->procedural->GetNumChannels()) : -1;
      break;
   default:
      {
         mesh = NULL;
         I3D_mesh_base *mb = vis->GetMesh();
         if(mb){
            mesh = new I3D_morph_mesh(drv);
            mesh->Clone(mb);
            mesh->Release();
         }
      }
   }
   return I3D_visual::Duplicate(vis);
}

//----------------------------

I3D_RESULT I3D_object_morph::SetProperty(dword index, dword value){

   switch(index){
   case 0: power = IntAsFloat(value); break;
   case 1:
      if(value >= MODE_LAST) 
         return I3DERR_INVALIDPARAMS;
      mode = (E_MODE)value;
      break;
   default: return I3DERR_INVALIDPARAMS;
   }
   return I3D_OK;
}

//----------------------------

dword I3D_object_morph::GetProperty(dword index) const{

   switch(index){
   case 0: return FloatAsInt(power);
   case 1: return mode;
   }
   return 0xcdcdcdcd;
}

//----------------------------

static const dword vs_decl[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(3, D3DVSDT_FLOAT2),   //weight + index
   D3DVSD_REG(2, D3DVSDT_FLOAT3),   //normal
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

void I3D_object_morph::AddPrimitives(S_preprocess_context &pc){

   if(!mesh)
      return;
   AddPrimitives1((I3D_morph_mesh*)mesh, pc);
}

//----------------------------

void I3D_object_morph::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   drv->SetClipping(rp.clip);
   HRESULT hr;

   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      dest_vb_trans_id = pc.render_view_id;
      LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }
      dword vertex_count;
      if(rp.curr_auto_lod < 0)
         vertex_count = mesh->vertex_buffer.NumVertices();
      else
         vertex_count = mesh->GetAutoLODs()[rp.curr_auto_lod].vertex_count;

                              //transform vertices
      dword prep_flags = VSPREP_FEED_MATRIX;

      if(drv->IsDirectTransform()){
         if(pc.mode!=RV_SHADOW_CASTER)
            prep_flags |= VSPREP_MAKELIGHTING;
         prep_flags |= VSPREP_COPY_UV;
      }else{
         if(pc.mode!=RV_SHADOW_CASTER){
#ifdef DEBUG_NO_CACHE_LIGHT
            vis_flags &= ~VISF_DEST_LIGHT_VALID;  
#endif
            if(!(vis_flags&VISF_DEST_LIGHT_VALID)){
               prep_flags |= VSPREP_MAKELIGHTING;
               vis_flags |= VISF_DEST_LIGHT_VALID;
            }
         }
         if(!(vis_flags&VISF_DEST_UV0_VALID)){
            prep_flags |= VSPREP_COPY_UV;
            vis_flags |= VISF_DEST_UV0_VALID;
         }
      }
      if(pc.mode==RV_SHADOW_CASTER)
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;


      I3D_driver::S_vs_shader_entry_in se(vs_decl);

      I3D_procedural_morph *proc = (I3D_procedural_morph*)mesh->GetProcedural();
      if(proc){
         proc->Evaluate();
         switch(mode){
         case MODE_SINGLE_CHANNEL:
         case MODE_PER_REGION_CHANNEL: //to do!
            {
               se.AddFragment(I3D_driver::VSF_TRANSFORM_MORPH1);
                                    //setup morphing constants
               assert(channel!=-1);
               S_vectorw vw = proc->GetMorphingValues()[channel];
               vw *= power;
               hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_BLEND_BASE, &vw, 1);
               CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
            }
            break;
         case MODE_PER_VERTEX_CHANNEL:
            {
               se.AddFragment(I3D_driver::VSF_TRANSFORM_MORPH2);
               int numch = proc->GetNumChannels();
               S_vectorw *vw = (S_vectorw*)alloca(numch * sizeof(S_vectorw));
               memcpy(vw, proc->GetMorphingValues(), numch * sizeof(S_vectorw));
               for(int i=numch; i--; ){
                  vw[i] *= power;
               }
               hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_BLEND_BASE, vw, numch);
               CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
            }
            break;
         default: assert(0);
         }
      }else{
         se.AddFragment(I3D_driver::VSF_TRANSFORM);
      }
      PrepareVertexShader(mesh->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      if(!drv->IsDirectTransform()){
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mesh->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_morph));
         hr = lpDev8->ProcessVertices(mesh->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }
   DrawPrimitive1((I3D_morph_mesh*)mesh, pc, rp);
}

//----------------------------
//----------------------------
static const S_visual_property props[] = {
   I3DPROP_FLOAT, "Morph power",
   I3DPROP_ENUM, "Channel\0Single\0Regions\0Vertices\0",
};

static struct S_auto_init_morph{
   static I3D_visual *CreateMorph(PI3D_driver drv){
      return new I3D_object_morph(drv);
   }
   S_auto_init_morph(){
      RegisterVisual(CreateMorph, I3D_VISUAL_MORPH, "Morph",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_morph;

//----------------------------
//----------------------------

