/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Particle.cpp
   Content: Particle frame.
--------------------------------------------------------*/

#include "all.h"
#include "particle.h"
#include "camera.h"
#include "scene.h"
#include "object.h"

//----------------------------


                              //when too big time comes to Tick, slice it to smaller pieces
                              //+ nicer animation at lower frame rates
                              //- slows down
//#define ALLOW_SLICE_TIME

#define MAX_TICK_COUNT 500

//----------------------------

class I3D_particle_imp: public I3D_particle{

#define PRTF_DO_EMIT_BASE     1  //schedule to emit base elements in next Tick
#define PRTF_BASE_DIR_WORLD   2  //base dir in world coordinates
#define PRTF_DEST_DIR_WORLD   4  //dest dir in world coordinates
#define PRTF_OPTIMIZED        8  //ticked only when bounding box is visible

   enum E_MODE{
      MODE_BILLBOARD,
      MODE_XY,
      MODE_XZ,
      MODE_YZ,
   };

   mutable dword prt_flags;

   struct S_elem{
                              //life:
      int life_cnt;
      float r_life_time;
                              //current:
      S_vector pos;
      float scale;
      float base_opacity, opacity;
      float rotation_angle;
                              //direction:
      S_vector base_dir, dest_dir;
      float r_affect_time;
      int affect_cnt;
      float scale_dir;
      float base_scale;
      float rotation_speed;   //<0 ... left, 0 ... no, >0 ... right
      S_matrix matrix;
      bool operator <(const S_elem &e){ return (dword)this < (dword)&e; }
      bool operator ==(const S_elem &e){ return (dword)this == (dword)&e; }
      bool operator !=(const S_elem &e){ return (dword)this != (dword)&e; }
   };
   list<S_elem> elems;
   C_smart_ptr<I3D_material> mat;

                              //particle param block
   struct S_data{
      E_MODE mode;
      int out_time[2];
      int counter;            //how many elements we'll produce
      int life_len[2];
      float element_scale;
      float scale_range[2];
      float init_pos_disp;
      float init_dir_disp;
      float affect_dir_dips;
      int affect_time[2];
      float opt_range[2];
      float opacity;
      S_vector init_dir;
      S_vector dest_dir;
      byte blend_mode;
      float rotation_speed_min;
      float rotation_speed_max;
      bool emit_base;
      //float aspect_ratio;
   } data;
                              //run-time
   int out_time_count;
   int count_down;            //how many elements we still have to produce (-1 = stopped)

   void Reset();

   void TickInternal(int time);

//----------------------------
// Get time delta since last render. The returned value is not grater than some optimal value,
// so that we don't waste time on updating unnecessary elements.
   int GetTickTime() const{
      int count = drv->GetRenderTime();
      int delta = count - last_render_time;
      return Max(0, Min(MAX_TICK_COUNT, delta));
   }
public:
   I3D_particle_imp(PI3D_driver d):
      I3D_particle(d),
      prt_flags(0)
   {
      visual_type = I3D_VISUAL_PARTICLE;
      data.mode = MODE_BILLBOARD;
      Reset();
   }

   virtual void Tick(int time){
      if(!(prt_flags&PRTF_OPTIMIZED) || (prt_flags&PRTF_DO_EMIT_BASE))
         TickInternal(time);
   }
   I3D_particle_imp &operator =(const I3D_particle_imp&);

public:
   I3DMETHOD_(void,Update)(){
      I3D_visual::Update();
      vis_flags &= ~VISF_BOUNDS_VALID;
      frm_flags &= ~(FRMFLAGS_HR_BOUND_VALID | FRMFLAGS_SM_BOUND_VALID);
   }

   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(SetProperty)(dword index, dword param);
   I3DMETHOD_(dword,GetProperty)(dword index) const;
                              //visual methods
   virtual bool ComputeBounds();
   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
   
   I3DMETHOD(SetMaterial)(PI3D_material mat1){
      mat = mat1;
      if(!mat)
         SetOn(false);
      return I3D_OK;
   }
   I3DMETHOD_(PI3D_material,GetMaterial)() const{ return mat; }
};

//----------------------------

void I3D_particle_imp::Reset(){

   out_time_count = 1;
   elems.clear();
                              //setup default particle
   SetProperty(PRT_OUT_TIME, 100);
   SetProperty(PRT_OUTT_THRESH, 60);
   SetProperty(PRT_LIFE_LEN, 1800);
   SetProperty(PRT_LIFEL_THRESH, 600);
   SetProperty(PRT_ELEMENT_SCALE, FloatAsInt(1.0f));
   SetProperty(PRT_SCALE_GROW_MIN, FloatAsInt(-.25f));
   SetProperty(PRT_SCALE_GROW_MAX, FloatAsInt(.6f));
   SetProperty(PRT_INIT_POS_THRESH, FloatAsInt(.2f));
   SetProperty(PRT_INIT_DIR_THRESH, FloatAsInt(1.0f));
   SetProperty(PRT_AFFECT_TIME, 400);
   SetProperty(PRT_AFFECTT_THRESH, 100);
   SetProperty(PRT_AFFECT_DIR_THRESH, FloatAsInt(1.0f));
   SetProperty(PRT_OPACITY_IN, FloatAsInt(.1f));
   SetProperty(PRT_OPACITY_OUT, FloatAsInt(.1f));
   SetProperty(PRT_OPACITY, FloatAsInt(1.0f));
   SetProperty(PRT_BASE_DIR_X, FloatAsInt(0.0f));
   SetProperty(PRT_BASE_DIR_Y, FloatAsInt(1.0f));
   SetProperty(PRT_BASE_DIR_Z, FloatAsInt(0.0f));
   SetProperty(PRT_DEST_DIR_X, FloatAsInt(.2f));
   SetProperty(PRT_DEST_DIR_Y, FloatAsInt(.8f));
   SetProperty(PRT_DEST_DIR_Z, FloatAsInt(0.0f));
   SetProperty(PRT_BASE_DIR_WORLD, false);
   SetProperty(PRT_DEST_DIR_WORLD, false);
   SetProperty(PRT_ROT_SPEED_MIN, 0);
   SetProperty(PRT_ROT_SPEED_MAX, 0);
   SetProperty(PRT_DRAW_MODE, I3DBLEND_ADD);
   //SetProperty(PRT_DRAW_MODE, I3DBLEND_ALPHABLEND);
   //SetProperty(PRT_DRAW_MODE, I3DBLEND_OPAQUE);
   SetProperty(PRT_SETCOUNT, 0);
   SetProperty(PRT_EMIT_BASE, false);
   SetProperty(PRT_MODE, 0);
   //SetProperty(PRT_ASPECT_RATIO, FloatAsInt(1.0f));
}

//----------------------------

I3D_RESULT I3D_particle_imp::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);
   PI3D_visual vis = I3DCAST_VISUAL(frm);
   switch(vis->GetVisualType()){
   case I3D_VISUAL_PARTICLE:
      {
         I3D_particle_imp *pp = (I3D_particle_imp*)vis;
         memcpy(&data, &pp->data, sizeof(data));
         count_down = pp->count_down;
         out_time_count = pp->out_time_count;

         elems = pp->elems;
         prt_flags = pp->prt_flags;
         SetMaterial(pp->GetMaterial());
      }
      break;
   default:
      {
         I3D_mesh_base *mb = vis->GetMesh();
         if(mb){
            SetMaterial(mb->GetFGroups1()->mat);
         }else{
            PI3D_material mat = drv->CreateMaterial();
            SetMaterial(mat);
            mat->Release();
         }
      }
   }
   return I3D_visual::Duplicate(vis);
}

//----------------------------

I3D_RESULT I3D_particle_imp::SetProperty(dword index, dword param){

   switch(index){
   case PRT_OUT_TIME:         data.out_time[0]    = param; break;
   case PRT_OUTT_THRESH:      data.out_time[1]    = param; break;
   case PRT_LIFE_LEN:         data.life_len[0]    = param; break;
   case PRT_LIFEL_THRESH:     data.life_len[1]    = param; break;
   case PRT_AFFECT_TIME:      data.affect_time[0] = Max(1, (int)param); break;
   case PRT_AFFECTT_THRESH:   data.affect_time[1] = param; break;
   case PRT_DRAW_MODE:        data.blend_mode     = param; break;
   case PRT_SETCOUNT:         count_down = data.counter = param; break;

   case PRT_ELEMENT_SCALE:    data.element_scale  = *(float*)&param; break;
   case PRT_SCALE_GROW_MIN:   data.scale_range[0] = *(float*)&param; break;
   case PRT_SCALE_GROW_MAX:   data.scale_range[1] = *(float*)&param; break;
   case PRT_INIT_POS_THRESH:  data.init_pos_disp  = *(float*)&param; break;
   case PRT_INIT_DIR_THRESH:  data.init_dir_disp  = *(float*)&param; break;
   case PRT_AFFECT_DIR_THRESH:data.affect_dir_dips= *(float*)&param; break;
   case PRT_OPACITY_IN:       data.opt_range[0]   = *(float*)&param; break;
   case PRT_OPACITY_OUT:      data.opt_range[1]   = *(float*)&param; break;
   case PRT_OPACITY:          data.opacity        = *(float*)&param; break;
   //case PRT_ASPECT_RATIO:     data.aspect_ratio   = *(float*)&param; break;
   case PRT_BASE_DIR_X:       data.init_dir.x     = *(float*)&param; break;
   case PRT_BASE_DIR_Y:       data.init_dir.y     = *(float*)&param; break;
   case PRT_BASE_DIR_Z:       data.init_dir.z     = *(float*)&param; break;
   case PRT_DEST_DIR_X:       data.dest_dir.x     = *(float*)&param; break;
   case PRT_DEST_DIR_Y:       data.dest_dir.y     = *(float*)&param; break;
   case PRT_DEST_DIR_Z:       data.dest_dir.z     = *(float*)&param; break;
   case PRT_BASE_DIR_WORLD:
      prt_flags &= ~PRTF_BASE_DIR_WORLD;
      if(param) prt_flags |= PRTF_BASE_DIR_WORLD;
      break;
   case PRT_DEST_DIR_WORLD:
      prt_flags &= ~PRTF_DEST_DIR_WORLD;
      if(param) prt_flags |= PRTF_DEST_DIR_WORLD;
      break;
   case PRT_ROT_SPEED_MIN: data.rotation_speed_min = IntAsFloat(param); break;
   case PRT_ROT_SPEED_MAX: data.rotation_speed_max = IntAsFloat(param); break;

   case PRT_EMIT_BASE:
      {
         data.emit_base = param;
         if(data.emit_base)
            prt_flags |= PRTF_DO_EMIT_BASE;
      }
      break;
   case PRT_MODE:             data.mode           = (E_MODE)param; break;
   case PRT_OPTIMIZED:
      prt_flags &= ~PRTF_OPTIMIZED;
      if(param)
         prt_flags |= PRTF_OPTIMIZED;
      break;
   case PRT_ASPECT_RATIO:
      break;
   default:
      assert(0);
   }
   return I3D_OK;
}

//----------------------------

dword I3D_particle_imp::GetProperty(dword index) const{

   switch(index){

   case PRT_OUT_TIME:         return data.out_time[0];
   case PRT_OUTT_THRESH:      return data.out_time[1];
   case PRT_LIFE_LEN:         return data.life_len[0];
   case PRT_LIFEL_THRESH:     return data.life_len[1];
   case PRT_AFFECT_TIME:      return data.affect_time[0];
   case PRT_AFFECTT_THRESH:   return data.affect_time[1];
   case PRT_DRAW_MODE:        return data.blend_mode;
   case PRT_SETCOUNT:         return data.counter;

   case PRT_ELEMENT_SCALE:    return FloatAsInt(data.element_scale);
   case PRT_SCALE_GROW_MIN:   return FloatAsInt(data.scale_range[0]);
   case PRT_SCALE_GROW_MAX:   return FloatAsInt(data.scale_range[1]);
   case PRT_INIT_POS_THRESH:  return FloatAsInt(data.init_pos_disp);
   case PRT_INIT_DIR_THRESH:  return FloatAsInt(data.init_dir_disp);
   case PRT_AFFECT_DIR_THRESH:return FloatAsInt(data.affect_dir_dips);
   case PRT_OPACITY_IN:       return FloatAsInt(data.opt_range[0]);
   case PRT_OPACITY_OUT:      return FloatAsInt(data.opt_range[1]);
   case PRT_OPACITY:          return FloatAsInt(data.opacity);
   //case PRT_ASPECT_RATIO:     return FloatAsInt(data.aspect_ratio);
   case PRT_BASE_DIR_X:       return FloatAsInt(data.init_dir.x);
   case PRT_BASE_DIR_Y:       return FloatAsInt(data.init_dir.y);
   case PRT_BASE_DIR_Z:       return FloatAsInt(data.init_dir.z);
   case PRT_DEST_DIR_X:       return FloatAsInt(data.dest_dir.x);
   case PRT_DEST_DIR_Y:       return FloatAsInt(data.dest_dir.y);
   case PRT_DEST_DIR_Z:       return FloatAsInt(data.dest_dir.z);
   case PRT_BASE_DIR_WORLD:   return (prt_flags&PRTF_BASE_DIR_WORLD);
   case PRT_DEST_DIR_WORLD:   return (prt_flags&PRTF_DEST_DIR_WORLD);
   case PRT_ROT_SPEED_MIN:    return FloatAsInt(data.rotation_speed_min);
   case PRT_ROT_SPEED_MAX:    return FloatAsInt(data.rotation_speed_max);

   case PRT_NUMELEMENTS:      return elems.size();
   case PRT_EMIT_BASE:        return data.emit_base;
   case PRT_MODE:             return data.mode;
   case PRT_OPTIMIZED:        return (prt_flags&PRTF_OPTIMIZED);
   default:
      assert(0);
   }
   return 0;
}

//----------------------------
                              //which axis goes off for particular particle mode
static const int mode_off_axis[] = {
   -1, 2, 1, 0
};

//----------------------------

void I3D_particle_imp::TickInternal(int time){

   if(prt_flags&PRTF_DO_EMIT_BASE){
      prt_flags &= ~PRTF_DO_EMIT_BASE;
      //int time = 2000;
      int time = 0;
      if(!time) time = data.out_time[0] + data.out_time[1] + data.life_len[0] + data.life_len[1];
      int tick_time = Max(25, data.out_time[0]);
      for(int i=0; i<time; i+=tick_time){
         TickInternal(tick_time);
      }
   }


   if(!IsOn1())
      return;
#ifdef ALLOW_SLICE_TIME
                              //slice time
   while(time > Max(10, data.out_time[0]*2)){
      int t1 = Max(5, data.out_time[0]);
      TickInternal(t1);
      time -= t1;
   }
#endif//ALLOW_SLICE_TIME

   const S_matrix &mat = GetMatrix();
   if(count_down!=-1){
                              //creation of new elements enabled
                              //check if it's right time to do so
      out_time_count -= time;
      while(out_time_count <= 0){
                              //comp delay to emit new element
         int add_time = data.out_time[0];
         if(data.out_time[1]) add_time += S_int_random(data.out_time[1]);
         if(!add_time) break; //problem - neverending loop detected - invalid params?!
         out_time_count += add_time;

                              //create element info & file
         elems.push_back(S_elem());
         S_elem &si = elems.back();
         si.matrix.Identity();
                              //get inherited scale
         si.base_scale = si.scale = data.element_scale * mat(0).Magnitude();
                              //setup life length
         si.life_cnt = data.life_len[0];
         if(data.life_len[1])
            si.life_cnt += S_int_random(data.life_len[1]);
         si.r_life_time = 1.0f / (float)si.life_cnt;

         si.base_opacity = data.opacity;

                              //setup direction
         /*
         S_vector i_dir;
         if(!(prt_flags&PRTF_BASE_DIR_WORLD)) i_dir = init_dir % mat;
         else i_dir = init_dir * si.base_scale;

         float idd = init_dir_disp * si.base_scale;
         si.dest_dir = S_vector(
            i_dir.x - idd * .5f + S_float_random() * idd,
            i_dir.y - idd * .5f + S_float_random() * idd,
            i_dir.z - idd * .5f + S_float_random() * idd);
         if(mode) si.dest_dir[mode_off_axis[mode]] = 0.0f;
         */
         {
            S_vector &dd = si.dest_dir;
            dd = data.init_dir;
            //float idd = init_dir_disp * si.base_scale;
            float idd = data.init_dir_disp;
            dd.x += S_float_random() * idd - idd * .5f;
            dd.y += S_float_random() * idd - idd * .5f;
            dd.z += S_float_random() * idd - idd * .5f;

            if(data.mode) dd[mode_off_axis[data.mode]] = 0.0f;
            if(!(prt_flags&PRTF_BASE_DIR_WORLD)) dd %= mat;
            //else i_dir = init_dir * si.base_scale;
         }

         float ipd = data.init_pos_disp;
         S_vector disp(
            ipd/2.0f - S_float_random() * ipd,
            ipd/2.0f - S_float_random() * ipd,
            ipd/2.0f - S_float_random() * ipd);
         if(data.mode) disp[mode_off_axis[data.mode]] = 0.0f;
         si.matrix(3) = disp * mat;
                              //setup scale
         si.scale_dir = data.scale_range[0] + S_float_random() * (data.scale_range[1] - data.scale_range[0]);
         si.affect_cnt = 0;

                              //setup rotation
         si.rotation_speed = data.rotation_speed_min + S_float_random() * (data.rotation_speed_max - data.rotation_speed_min);
         if(!IsAbsMrgZero(si.rotation_speed)){
            si.rotation_angle = S_float_random() * GetPI2();
         }else
            *(dword*)&si.rotation_speed = 0;

         if(count_down && !--count_down){
            count_down = -1;
            break;
         }
      }
   }

   float tsec = (float)time * .001f;
                              //move all elements
   list<S_elem>::iterator it, it_next, it_end = elems.end();
   for(it=elems.begin(); it!=it_end; ){

      it_next = it;
      ++it_next;
      S_elem &si = (*it);
                              //lifecounf
      if((si.life_cnt -= time) <= 0){
                              //dead
         elems.erase(it);
         it_end = elems.end();
         it = it_next;
         continue;
      }
#ifdef USE_PREFETCH
      if(it_next!=it_end){
         Prefetch64Bytes(&(*it_next));
      }
#endif
                              //compute opacity
      float ft = 1.0f - ((float)si.life_cnt * si.r_life_time);
      si.opacity = ft<data.opt_range[0] ? ft/data.opt_range[0] :
         ft<=data.opt_range[1] ? 1.0f :
         1.0f - (ft-data.opt_range[1])/(1.0f-data.opt_range[1]);
      si.opacity *= si.base_opacity;
                              //compute affect direction
      if((si.affect_cnt -= time) <= 0){
                              //don't accept less than 1, we divide by this later
         int affect_time = data.affect_time[0];
         if(data.affect_time[1]) affect_time += S_int_random(data.affect_time[1]);
         assert(affect_time);
         si.affect_cnt = affect_time;
         si.r_affect_time = 1.0f / (float)affect_time;

                              //compute new dest dir
         {
            S_vector &dd = si.dest_dir;
            si.base_dir = dd;
            float mag;
            float dot = dd.Dot(dd);
            if(dot <= (MRG_ZERO*MRG_ZERO)) mag = 0.0f;
            else mag = sqrt(dot);

            S_vector disp1 = data.dest_dir;
            if(data.mode) disp1[mode_off_axis[data.mode]] = 0.0f;
            if(!(prt_flags&PRTF_DEST_DIR_WORLD))
               disp1 %= mat;
            dd += disp1;

                              //keep similar size of direction
            dot = dd.Dot(dd);
            float new_mag;
            if(dot <= (MRG_ZERO*MRG_ZERO))
               new_mag = 0.0f;
            else{
               new_mag = sqrt(dot);
               dd *= mag / new_mag;
            }

            float add = data.affect_dir_dips;
            S_vector disp(
               add*.5f - S_float_random() * add,
               add*.5f - S_float_random() * add,
               add*.5f - S_float_random() * add);
            if(data.mode) disp[mode_off_axis[data.mode]] = 0.0f;
            if(!(prt_flags&PRTF_DEST_DIR_WORLD)) disp %= mat;

            dd += disp;
         }
      }
                              //compute new position
      float fd = (float)si.affect_cnt * si.r_affect_time;
      S_vector dir = si.base_dir * fd + si.dest_dir * (1.0f - fd);
      si.matrix(3) += dir * tsec;
                              //adjust scale
      si.scale *= (1.0f + si.scale_dir * tsec);

                              //compute new rotation angle
      if(FloatAsInt(si.rotation_speed) != 0){
         si.rotation_angle += si.rotation_speed * tsec;
      }
      it = it_next;
   }
   ComputeBounds();
}

//----------------------------

const float PARTICLE_BSPHERE_RADIUS = .8660255f;

const float PARTICLE_EXPAND_RATIO = .3f;  //ratio of expanding cached bbox on both min & max sides
//----------------------------

bool I3D_particle_imp::ComputeBounds(){


                              //compute real bounding box
   I3D_bbox bbox;
   bbox.Invalidate();

   if(elems.size()){

      const S_matrix &m_inv = GetInvMatrix1();
      const float m_inv_scale = m_inv(2).Magnitude();
      float radius = PARTICLE_BSPHERE_RADIUS * m_inv_scale;

      for(list<S_elem>::const_iterator it=elems.begin(); it!=elems.end(); it++){
         const S_elem &si = (*it);

         const S_vector pos = si.matrix(3) * m_inv;
         float e_radius = radius * si.scale;
                              //expand bounding box
         bbox.min.x = Min(bbox.min.x, pos.x-e_radius);
         bbox.min.y = Min(bbox.min.y, pos.y-e_radius);
         bbox.min.z = Min(bbox.min.z, pos.z-e_radius);

         bbox.max.x = Max(bbox.max.x, pos.x+e_radius);
         bbox.max.y = Max(bbox.max.y, pos.y+e_radius);
         bbox.max.z = Max(bbox.max.z, pos.z+e_radius);
      }
   }
                              //compare it with cached version
   if(vis_flags&VISF_BOUNDS_VALID){
      const I3D_bbox &vbb = bound.bound_local.bbox;
      if(vbb.IsValid()){
         bool is_greater = (bbox.min.x<vbb.min.x || bbox.min.y<vbb.min.y || bbox.min.z<vbb.min.z ||
            bbox.max.x>vbb.max.x || bbox.max.y>vbb.max.y || bbox.max.z>vbb.max.z);
                              //if new bbox is greater than cached one, invalidate
         if(is_greater){
            vis_flags &= ~VISF_BOUNDS_VALID;
         }else{
                              //check how smaller is the new one,
                              // if too small, invalidate too
            S_vector diag1 = vbb.max - vbb.min;
            if(diag1.x<.001f || diag1.y<.001f || diag1.z<.001f){
               vis_flags &= ~VISF_BOUNDS_VALID;
            }else{
               S_vector diag = bbox.max - bbox.min;
               float min_ratio = Min(diag.x/diag1.x, Min(diag.y/diag1.y, diag.z/diag1.z));
               if(min_ratio < (1.0f/(1.0f+PARTICLE_EXPAND_RATIO*4.0f))){
                              //new bbox significantly smaller, use it
                  vis_flags &= ~VISF_BOUNDS_VALID;
               }
            }
         }
      }else{
                              //cached bbox was not valid,
                              // use new if it is valid
         if(bbox.IsValid())
            vis_flags &= ~VISF_BOUNDS_VALID;
      }
   }

   if(!(vis_flags&VISF_BOUNDS_VALID)){
                              //expand our bbox by certain amount
      S_vector scaled_diag = (bbox.max - bbox.min) * PARTICLE_EXPAND_RATIO;
      bbox.min -= scaled_diag;
      bbox.max += scaled_diag;
                              //store into cached one
      bound.bound_local.bbox = bbox;
                              //expand real bbox by some value and store in real bbox
                              //invalidate hierarchy bounds
      frm_flags &= ~FRMFLAGS_HR_BOUND_VALID;
      {
         PI3D_frame frm1a = this;
         do{
            frm1a->SetFrameFlags(frm1a->GetFrameFlags() & ~(FRMFLAGS_HR_BOUND_VALID | FRMFLAGS_SM_BOUND_VALID));
         }while(frm1a=frm1a->GetParent1());
      }
      return I3D_visual::ComputeBounds();
   }
   return true;
}

//----------------------------
#define NUM_PARTICLE_VERTS 4

static const I3D_text_coor particle_uv[NUM_PARTICLE_VERTS] = {
   I3D_text_coor(0.0f, 0.0f),
   I3D_text_coor(0.0f, 1.0f),
   I3D_text_coor(1.0f, 0.0f),
   I3D_text_coor(1.0f, 1.0f)
};

static const S_vector particle_coor[NUM_PARTICLE_VERTS] = {
   S_vector(-.5f,  .5f, 0.0f),
   S_vector( .5f,  .5f, 0.0f),
   S_vector(-.5f, -.5f, 0.0f),
   S_vector( .5f, -.5f, 0.0f)
};

//----------------------------

static const dword vs_decl[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT2),   //position
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

void I3D_particle_imp::AddPrimitives(S_preprocess_context &pc){

   if(prt_flags&PRTF_OPTIMIZED){
      int time = GetTickTime();
      if(time)
         TickInternal(time);
   }
                              //don't allow in shadow casting
   if(pc.mode==RV_SHADOW_CASTER)
      return;

   if(!elems.size() || !mat)
      return;

   PI3D_camera cam = pc.scene->GetActiveCamera1();

   pc.prim_list.push_back(S_render_primitive(0));
   S_render_primitive &p = pc.prim_list.back();
   p.vis = this;
   p.clip = pc.clip;
   p.blend_mode = data.blend_mode | I3DBLEND_VERTEXALPHA;
   p.sector = pc.curr_sector;
   p.scene = pc.scene;

                              //compute sorting depth
   const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
   const S_vector &w_pos = bsphere.pos;
   const S_matrix &m_view = cam->GetViewMatrix1();
   float depth =
         w_pos[0] * m_view(0, 2) +
         w_pos[1] * m_view(1, 2) +
         w_pos[2] * m_view(2, 2) +
                    m_view(3, 2);

   dword sort_base = ((FloatToInt(depth * 100.0f))&PRIM_SORT_DIST_MASK) << PRIM_SORT_DIST_SHIFT;
   p.sort_value = PRIM_SORT_ALPHA_NOZWRITE | sort_base;
   ++pc.alpha_nozwrite;

   drv->render_stats.vert_trans += NUM_PARTICLE_VERTS * elems.size();
}

//----------------------------

void I3D_particle_imp::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   if(drv->last_blend_flags!=rp.blend_mode)
      drv->SetupBlend(rp.blend_mode);
   drv->SetupTextureStage(0, D3DTOP_MODULATE);

   drv->SetClipping(rp.clip);
   drv->DisableTextureStage(1);
   //drv->SetRenderMat(mat);
   drv->SetTexture1(0, mat->GetTexture1(MTI_DIFFUSE));

                              //debug mode - without textures
   if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES))
      drv->SetTexture1(0, NULL);

   drv->EnableNoCull(true);   //use no culling due to mirrors

   C_sort_list<const S_elem*> sort_list(elems.size());
                              //preprocess elements - sort by distance
   {
      PI3D_camera cam = rp.scene->GetActiveCamera1();
      const S_matrix &m_camera = cam->GetMatrixDirect();
      const S_matrix &mat = GetMatrixDirect();
      S_matrix m_rot;
      m_rot.Zero();
      m_rot(3, 3) = 1.0f;
      switch(data.mode){
      case MODE_BILLBOARD:
      case MODE_XY: m_rot(2, 2) = 1.0f; break;
      case MODE_XZ: m_rot(2, 1) = 1.0f; break;
      case MODE_YZ: m_rot(2, 1) = 1.0f; break;
      }

      list<S_elem>::iterator it, it_end = elems.end();
      for(it=elems.begin(); it!=it_end; it++){
         S_elem &si = (*it);

         if(FloatAsInt(si.rotation_speed) != 0){
            float c = cos(si.rotation_angle);
            float s = sin(si.rotation_angle);
            s *= si.scale;
            c *= si.scale;

            S_matrix m;
            switch(data.mode){
            case MODE_BILLBOARD:
               m_rot(0, 0) = c;
               m_rot(1, 1) = c;
               m_rot(0, 1) = -s;
               m_rot(1, 0) = s;
               m = m_rot % m_camera;
               break;
            case MODE_XY:
               m_rot(0, 0) = c;
               m_rot(1, 1) = c;
               m_rot(0, 1) = -s;
               m_rot(1, 0) = s;
               m = m_rot % mat;
               break;
            case MODE_XZ:
               m_rot(0, 0) = c;
               m_rot(0, 2) = s;
               m_rot(1, 0) = s;
               m_rot(1, 2) = -c;
               m = m_rot % mat;
               break;
            case MODE_YZ:
               m_rot(0, 1) = s;
               m_rot(0, 2) = -c;
               m_rot(1, 1) = -c;
               m_rot(1, 2) = -s;
               m = m_rot % mat;
               break;
            }
            si.matrix(0) = m(0);
            si.matrix(1) = m(1);
            si.matrix(2) = m(2);
         }else{
            S_matrix m;
            switch(data.mode){
            case MODE_BILLBOARD:
               m_rot(0, 0) = si.scale;
               m_rot(1, 1) = si.scale;
               m = m_rot % m_camera;
               break;
            case MODE_XY:
               m_rot(0, 0) = si.scale;
               m_rot(1, 1) = si.scale;
               m = m_rot % mat;
               break;
            case MODE_XZ:
               m_rot(0, 0) = si.scale;
               m_rot(1, 2) = -si.scale;
               m = m_rot % mat;
               break;
            case MODE_YZ:
               m_rot(0, 2) = -si.scale;
               m_rot(1, 1) = -si.scale;
               m = m_rot % mat;
               break;
            }
            si.matrix(0) = m(0);
            si.matrix(1) = m(1);
            si.matrix(2) = m(2);
         }
                                 //compute distance to camera
         const S_matrix &m_view = cam->GetViewMatrix1();
         const S_vector &opos = si.matrix(3);
         float depth =
            opos[0] * m_view(0, 2) +
            opos[1] * m_view(1, 2) +
            opos[2] * m_view(2, 2) +
                      m_view(3, 2);
         sort_list.Add(&si, FloatToInt(depth * 100.0f));
      }
      sort_list.Sort();
   }

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

   HRESULT hr;

   I3D_driver::S_vs_shader_entry_in se(vs_decl);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
   PrepareVertexShader(NULL, se, rp, VSPREP_COPY_UV);

   if(pc.mode!=RV_SHADOW_CASTER){
      drv->EnableFog(vis_flags&VISF_FOG_VALID);
   }

   drv->SetStreamSource(drv->vb_flare, sizeof(S_vertex_flare));

   float mat_alpha = mat->GetAlpha1();
   bool is_self_alpha = ((data.blend_mode&0xffff) == I3DBLEND_ADD);
   for(int i=sort_list.Count(); i--; ){
      const S_elem &si = *sort_list[i];
                              //prepare transformation matrix
      {
         S_matrix mt;
         mt.Make4X4Transposed(si.matrix, rp.scene->GetViewProjHomMat());
         hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0, &mt, 4);
         CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
      }
                              //setup color of element
      {
         S_vectorw diffuse = mat->GetDiffuse();
         float alpha = si.opacity * mat_alpha;
         if(is_self_alpha)
            diffuse *= alpha;
         diffuse.w = alpha;
         hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &diffuse, 1);
         CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
      }
   
      hr = lpDev8->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
      CHECK_D3D_RESULT("DrawPrimitive", hr);
   }
   sort_list.Clear();
}

//----------------------------

static const S_visual_property props[] = {
   I3DPROP_INT, "Out time",
   I3DPROP_INT, "Out time thresh",
   I3DPROP_INT, "Life len",
   I3DPROP_INT, "Life len thresh",
   I3DPROP_FLOAT, "Scale min",
   I3DPROP_FLOAT, "Scale max",
   I3DPROP_FLOAT, "Init pos thresh",
   I3DPROP_FLOAT, "Init dir thresh",
   I3DPROP_INT, "Affect time",
   I3DPROP_INT, "Affect thresh",
   I3DPROP_FLOAT, "Affect dir",
   I3DPROP_FLOAT, "Opacity in",
   I3DPROP_FLOAT, "Opacity out",
   I3DPROP_FLOAT, "Base dir X",
   I3DPROP_FLOAT, "Base dir Y",
   I3DPROP_FLOAT, "Base dir Z",
   I3DPROP_FLOAT, "Dest dir X",
   I3DPROP_FLOAT, "Dest dir Y",
   I3DPROP_FLOAT, "Dest dir Z",
   I3DPROP_ENUM, "Draw mode\0Opaque\0Blend\0Add\0",
   I3DPROP_INT, "Count",
   I3DPROP_BOOL, "Base dir world",
   I3DPROP_BOOL, "Dest dir world",
   I3DPROP_BOOL, "Emit base",
   I3DPROP_FLOAT, "Rot speed min",
   I3DPROP_FLOAT, "Rot speed max",
   I3DPROP_FLOAT, "Element scale",
   I3DPROP_FLOAT, "Opacity",
   I3DPROP_ENUM, "Mode\0Billboard\0XY\0XZ\0YZ\0",
   I3DPROP_BOOL, "Optimized",
   //I3DPROP_FLOAT, "Aspect",
};

static I3D_visual *CreateParticle(PI3D_driver drv){
   return new I3D_particle_imp(drv);
}

static struct S_auto_init_prt{

   S_auto_init_prt(){
      RegisterVisual(CreateParticle, I3D_VISUAL_PARTICLE, "Particle",
         props, sizeof(props)/sizeof(S_visual_property));
   }
} auto_init_prt;

//----------------------------
//----------------------------
