/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Scene.cpp
   Content: Scene class.
--------------------------------------------------------*/

#include "all.h"
#include <isound2.h>
#include "object.h"
#include "light.h"
#include "camera.h"
#include "scene.h"
#include "model.h"
#include "volume.h"
//----------------------------


#define USE_ASM


int debug_count[4];

//----------------------------

static bool NameMatch(const C_string &s, const char *cp){

   size_t l1 = s.Size(), l2 = strlen(cp);
   if(!l1 || !l2){
      if(l1!=0 && s[0] == '*')
         return true;
      return (l1==l2);
   }

   for(const char *cp1 = s; ; ++cp1, ++cp){
      switch(*cp1){
      case 0:
         return !(*cp);

      case '*':
         return true;

      case '?':
         if(!*cp)
            return false;
         break;

      default:
         if(*cp != *cp1)
            return false;
      }
   }
}

//----------------------------

void S_expanded_view_frustum::Make(const S_view_frustum &vf){

   for(int i=vf.num_clip_planes; i--; ){
      const S_plane &pl = vf.clip_planes[i];
      nx[i] = pl.normal.x;
      ny[i] = pl.normal.y;
      nz[i] = pl.normal.z;
      d[i]  = pl.d;
   }
   num_clip_planes = vf.num_clip_planes;
   int ncp = (num_clip_planes + 3) & -4;
                              //pad last elements with the 1st one
   i = ncp - vf.num_clip_planes;
   const S_plane &pl = vf.clip_planes[0];
   while(i--){
      nx[vf.num_clip_planes + i] = pl.normal.x;
      ny[vf.num_clip_planes + i] = pl.normal.y;
      nz[vf.num_clip_planes + i] = pl.normal.z;
      d[vf.num_clip_planes + i] = pl.d;
   }
}

//----------------------------
//----------------------------
                              //virtual table for container callback to scene
static const S_container_vcall vc_container_scene = {
   (S_container_vcall::t_Open)I3D_scene::Open,
   (S_container_vcall::t_Close)I3D_scene::Close,
   (S_container_vcall::t_AddFrame)I3D_scene::AddFrame,
   (S_container_vcall::t_DeleteFrame)I3D_scene::DeleteFrame,
};

//----------------------------
                              //scene
I3D_scene::I3D_scene(PI3D_driver d):
   ref(1),
   primary_sector(NULL),
   backdrop_sector(NULL),
   backdrop_camera(NULL),
   ssp_count_down(0),
   last_render_time(0),
   proposed_num_shd_receivers(0),
   drv(d),
   cb_flags(0),
   cb_proc(NULL),
   //scn_flags(SCNF_RESET_VOL_TREE | SCNF_RESET_DYNAMIC_VOL_TREE),
   scn_flags(SCNF_RESET_DYNAMIC_VOL_TREE),
   last_dynamic_tree_build_time(0)
{
   drv->AddCount(I3D_CLID_SCENE);

   cont.master_interface = (C_unknown*)this;
   cont.vcall = &vc_container_scene;

   primary_sector = new I3D_sector(this, true);
   primary_sector->SetName("Primary sector");

   backdrop_sector = new I3D_sector(this, true);
   backdrop_sector->SetName("Backdrop sector");

   backdrop_camera = GetDriver1()->CreateCamera();
   backdrop_camera->SetName("backdrop camera");
   backdrop_camera->SetCurrSector(backdrop_sector);
   backdrop_camera->SetRange(1.0f, 20.0f);
   backdrop_camera->UpdateMatrices(.75f);
   bsp_builder.Init(this);

   shd_camera = drv->CreateCamera();
   shd_camera->Release();
}

//----------------------------

I3D_scene::~I3D_scene(){

   Close();

   if(primary_sector) primary_sector->Release();
   if(backdrop_sector) backdrop_sector->Release();
   if(backdrop_camera) backdrop_camera->Release();

   drv->DecCount(I3D_CLID_SCENE);
   drv = NULL;

   bsp_builder.Close();
   for(int i=sounds.size(); i--; ){
      assert(sounds[i]->owning_scene==this);
      sounds[i]->owning_scene = NULL;
   }
}

//----------------------------

bool I3D_scene::Init(PI3D_driver drv1){

   if(!drv1 || !drv1->IsInit()) return false;
   drv=drv1;

   SetBgndColor(S_vector(0, 0, 0));
   LPIGraph ig = drv->GetGraphInterface();
   SetViewport(0, 0,
      ig->Scrn_sx(), ig->Scrn_sy(),
      !(drv->GetGraphInterface()->GetFlags()&IG_FULLSCREEN));

                              //init sound listener
   PISND_driver isp = drv->GetSoundInterface();
   if(isp){
      isp->SetListenerRolloffFactor(0.0f);
      {
#ifdef USE_EAX
         /*
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_ENVIRONMENT,
         //EAX_ENVIRONMENT_BATHROOM);
         EAX_ENVIRONMENT_DRUGGED);
         //EAX_ENVIRONMENT_GENERIC);
         //EAX_ENVIRONMENT_PADDEDCELL);
         */
#endif
      }
   }
   return true;
}

//----------------------------

PI3D_SCNCALLBACK I3D_scene::SetCallback(PI3D_SCNCALLBACK cb, dword cbf){

   PI3D_SCNCALLBACK ret = cb_proc;
   if(cb){
      if(cb_proc!=cb) cb_flags=0;
      cb_proc = cb;
      cb_flags |= cbf;
   }else{
      cb_flags &= ~cbf;
      if(!cb_flags){
         cb_proc = NULL;
      }
   }
   return ret;
}

//----------------------------
void lightUpdColor(PI3D_driver drv, const S_vector &color, float power,
   S_vector *col, dword *dw_color, bool surf_match);

//----------------------------

I3D_RESULT I3D_scene::SetBgndColor(const S_vector &c){

   color_bgnd = c;
   S_vector v;
   lightUpdColor(drv, c, 1.0f, &v, &dw_color_bgnd, true);

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::SetGlobalColor(const S_vector &v){

   assert(0);
   return I3DERR_UNSUPPORTED;
}

//----------------------------
static const float default_material[] = {
   1.0f, 1.0f, 1.0f, 1.0f,    //diffuse
   1.0f, 1.0f, 1.0f, 1.0f,    //ambient
   0.0f, 0.0f, 0.0f, 1.0f, 0.0f, //specular
   0.0f, 0.0f, 0.0f, 0.0f, 0.0f  //emmisive
};

S_vector I3D_scene::GetLightness(const S_vector &pos, const S_vector *norm, dword light_flags) const{

   const I3D_bsphere bsphere(pos, 0.0f);
   PI3D_sector sct = GetSector(pos);
   if(!sct)
      return S_vector(0, 0, 0);
   S_vector norm_tmp;

   S_vector ret(0, 0, 0);

   const vector<PI3D_light> &lights = sct->GetLights1();
   for(int i=lights.size(); i--; ){
      PI3D_light lp = lights[i];
      if(!lp->IsOn1())
         continue;
      if(!(lp->GetLightFlags()&light_flags))
         continue;

      switch(lp->GetLightType1()){
      case I3DLIGHT_AMBIENT:
         ret += lp->color * lp->power;
         break;

      case I3DLIGHT_POINT:
      case I3DLIGHT_POINTAMBIENT:
         {
            const S_vector &lpos = lp->GetMatrix()(3);
            S_vector dir_to_l = lpos - pos;
            float dist = dir_to_l.Magnitude();
            const float &nr = lp->range_n_scaled, &fr = lp->range_f_scaled;
            if(dist>=fr)
               continue;
            float p = lp->power;
            if(norm){
               dir_to_l /= dist;
               float dt = norm->Dot(dir_to_l);
               if(dt <= 0.0f)
                  continue;
               p *= dt;
            }
            if(dist>nr)
               p *= (fr-dist) / (fr-nr);
            ret += lp->color * p;
         }
         break;

      case I3DLIGHT_DIRECTIONAL:
         {
            float p = lp->power;
            if(norm){
               float dt = norm->Dot(-lp->normalized_dir);
               if(dt <= 0.0f)
                  continue;
               p *= dt;
            }
            ret += lp->color * p;
         }
         break;
      }
   }
   return ret;
}

//----------------------------

void I3D_scene::SetBackdropRange(float n1, float f1){
   assert(backdrop_camera);
   backdrop_camera->SetRange(n1, f1);
   backdrop_camera->Update();
}

//----------------------------

void I3D_scene::GetBackdropRange(float &n1, float &f1) const{
   assert(backdrop_camera);
   backdrop_camera->GetRange(n1, f1);
}

//----------------------------

void I3D_scene::Close(){

   assert(drv->IsInit());
   DestroyDynamicVolTree();

   int i;

   for(i=0; i<cont.frames.size(); i++){
      PI3D_frame frm = cont.frames[i];
      switch(frm->GetType1()){
      case FRAME_MODEL:
         {
            PI3D_model mod = I3DCAST_MODEL(frm);
            if(mod->in_container==&cont){
               if(mod->cont.NumInterpolators1())
                  cont.DeleteAnimFrame(frm);
               mod->in_container = NULL;
            }
         }
         break;
      case FRAME_SOUND:
         I3DCAST_SOUND(frm)->VirtualPause();
         break;
      }
      frm->LinkTo(NULL);
   }
   cont.frames.clear();

   AttenuateSounds();

   if(current_camera)
      current_camera->SetCurrSector(NULL);
   SetActiveCamera(NULL);
                              //release sectors
   for(i=sectors.size(); i--; ){
      sectors[i]->Disconnect(NULL);
   }
   sectors.clear();
   current_camera = NULL;
   primary_sector->Disconnect(NULL);
   primary_sector->SetEnvironmentID(0);
   backdrop_sector->Disconnect(NULL);
   backdrop_sector->SetEnvironmentID(0);

                              //free lights of sectors
   while(primary_sector->NumLights()){
      I3D_RESULT ir;
      ir = primary_sector->DeleteLight(primary_sector->GetLight(0));
      assert(I3D_SUCCESS(ir));
   }
   while(backdrop_sector->NumLights()){
      I3D_RESULT ir;
      ir = backdrop_sector->DeleteLight(backdrop_sector->GetLight(0));
      assert(I3D_SUCCESS(ir));
   }

                              //release animations
   cont.ClearInterpolators();

   SetBgndColor(S_vector(0, 0, 0));

   bsp_builder.Clear();
}

//----------------------------

PI3D_sector I3D_scene::GetSector(const S_vector &v, PI3D_sector suggest_sector) const{

   if(suggest_sector){
      if(suggest_sector!=primary_sector){
         if(suggest_sector->CheckPoint(v))
            return suggest_sector;
      }else
         suggest_sector = NULL;
   }

   for(int i=sectors.size(); i--; ){
      PI3D_sector sct = sectors[i];
      if(sct==suggest_sector || sct->IsSimple())
         continue;
      if(sct->CheckPoint(v))
         return sct;
   }

   return primary_sector;
}

//----------------------------

I3D_RESULT I3D_scene::SetFrameSectorPos(PI3D_frame frm, const S_vector &pos){

   PI3D_frame prnt = frm->GetParent1();
                              //in order to change sector, frame must be
                              //already linked to one, or to NULL
   if(prnt)
   switch(prnt->GetType1()){
   case FRAME_SECTOR:
      if(prnt == GetBackdropSector())
         return I3DERR_CANTSETSECTOR;
      break;
   default:
      return I3DERR_CANTSETSECTOR;
   }
   PI3D_frame sct = GetSector(pos);
   {
                              //check if this sector is not our child
      PI3D_frame scr_prnt = sct;
      while(scr_prnt=scr_prnt->GetParent1())
      if(scr_prnt==frm){
         sct = GetPrimarySector1();
         break;
      }
   }
   if(sct != prnt){
      frm->LinkTo(sct, I3DLINK_UPDMATRIX);
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::SetFrameSector(PI3D_frame frm){

   const S_vector &pos = frm->GetWorldPos();

   return SetFrameSectorPos(frm, pos);
}

//----------------------------
                              //enumeration

                              //version without name checking
static I3DENUMRET EnumRecursiveProc(const vector<PI3D_frame> &root,
   I3D_ENUMPROC *proc, dword user, dword flags){

   for(int i=0; i<root.size(); ){
      PI3D_frame frm = root[i];

      if(frm->GetEnumMask()&flags){
         I3DENUMRET er = proc(frm, user);
         switch(er){
         case I3DENUMRET_SKIPCHILDREN:
            if(i>=root.size()) return I3DENUMRET_OK;
            if(frm!=root[i]) continue;
            goto next;
         case I3DENUMRET_CANCEL: return I3DENUMRET_CANCEL;
         default:
            if(i>=root.size()) return I3DENUMRET_OK;
            if(frm!=root[i]) continue;
         }
      }
                              //enum all children, if any
      if(frm->NumChildren1()){
         if(EnumRecursiveProc(frm->GetChildren1(), proc, user, flags)==I3DENUMRET_CANCEL)
            return I3DENUMRET_CANCEL;
      }
next:
      i++;
   }
   return I3DENUMRET_OK;
}

//----------------------------
                              //version with name checking
static I3DENUMRET EnumRecursiveProc(const vector<PI3D_frame> &parent,
   I3D_ENUMPROC *proc, dword user, dword flags,
   const C_string &mask){

   for(int i=0; i<parent.size(); ){
      PI3D_frame frm = parent[i];

      if(frm->GetEnumMask()&flags){
         const char *fname = frm->GetName1();
         if((flags&ENUMF_MODEL_ORIG) && (frm->GetFrameFlags()&FRMFLAGS_SUBMODEL)){
            const char *cp = strchr(fname, '.');
            if(cp) fname=cp+1;
         }
         bool do_this = (flags&ENUMF_WILDMASK) ?
            NameMatch(mask, fname) :
            !strcmp(fname, mask);
         if(do_this){
            I3DENUMRET er=proc(frm, user);
            switch(er){
            case I3DENUMRET_SKIPCHILDREN:
               if(i>=parent.size()){
                  return I3DENUMRET_OK;
               }
               if(frm!=parent[i]) continue;
               goto next;
            case I3DENUMRET_CANCEL: return I3DENUMRET_CANCEL;
            default:
               if(i>=parent.size()){
                  return I3DENUMRET_OK;
               }
               if(frm!=parent[i]) continue;
            }
         }
      }
                              //enum all children, if any
      if(frm->NumChildren1()){
         if(EnumRecursiveProc(frm->GetChildren1(), proc, user, flags, mask/*, use_list*/)==I3DENUMRET_CANCEL){
            return I3DENUMRET_CANCEL;
         }
      }
next:
      i++;
   }
   return I3DENUMRET_OK;
}

//----------------------------

I3D_RESULT I3D_frame::EnumFrames(I3D_ENUMPROC *proc, dword user,
   dword flags, const char *mask) const{

   if(frm_flags&FRMFLAGS_SUBMODEL)
      flags |= ENUMF_MODEL_ORIG;

   I3DENUMRET er;
   if(!mask){
      er = EnumRecursiveProc(GetChildren1(), proc, user, flags);
   }else{
      C_string str(mask);
      er = EnumRecursiveProc(GetChildren1(), proc, user, flags, str);
   }
   if(er==I3DENUMRET_CANCEL) return I3DERR_CANCELED;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::EnumFrames(I3D_ENUMPROC *proc, dword user,
   dword flags, const char *mask) const{

   I3D_RESULT ir =
      primary_sector->EnumFrames(proc, user, flags, mask);

   if(SUCCEEDED(ir))
   if(flags&ENUMF_BACKDROP)
      ir = backdrop_sector->EnumFrames(proc, user, flags, mask);
   return ir;
}

//----------------------------

static I3DENUMRET I3DAPI cbFindFrm(PI3D_frame frm, dword user){

   (*(PI3D_frame*)user)=frm;
   return I3DENUMRET_CANCEL;
}

//----------------------------

PI3D_frame I3D_scene::FindFrame(const char *name, dword f) const{

   PI3D_frame frm = NULL;
   EnumFrames(cbFindFrm, (dword)&frm, f, name);
   if(!frm){
      if(name){
         if(f&ENUMF_WILDMASK){
            if(primary_sector->GetName1().Match(name)) frm = primary_sector;
            else
            if(f&ENUMF_BACKDROP){
               if(backdrop_sector->GetName1().Match(name)) 
                  frm = backdrop_sector;
            }
         }else{
            if(!strcmp(primary_sector->GetName1(), name)) frm = primary_sector;
            else
            if(f&ENUMF_BACKDROP){
               if(!strcmp(backdrop_sector->GetName1(), name)) 
                  frm = backdrop_sector;
            }
         }
      }
   }
   return frm;
}

//----------------------------
                              //scene frames
I3D_RESULT I3D_scene::AddFrame(PI3D_frame frm){

   switch(frm->GetType1()){
   case FRAME_VISUAL:
      if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE) break;
      cont.AddAnimFrame(frm);
      break;
   case FRAME_MODEL:
      {
                              //check if it's already in some other scene
         PI3D_model mod = I3DCAST_MODEL(frm);
         if(mod->in_container){
                              //can't be in 2 scenes at the same time
            return I3DERR_GENERIC;
         }
         mod->in_container = &cont;
                              //if it contains animation, create interpolator
         if(mod->cont.NumInterpolators1())
            cont.AddAnimFrame(frm);
      }
      break;
   }
   cont.frames.push_back(frm);

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::DeleteFrame(PI3D_frame frm){

   for(int i=0; i<cont.frames.size(); i++)
   if(cont.frames[i]==frm){

      switch(frm->GetType1()){
      case FRAME_VISUAL:
         if(I3DCAST_VISUAL(frm)->GetVisualType() != I3D_VISUAL_PARTICLE) break;
         cont.DeleteAnimFrame(frm);
         break;
      case FRAME_MODEL:
         {
            PI3D_model mod = I3DCAST_MODEL(frm);
            if(mod->in_container==&cont){
               if(mod->cont.NumInterpolators1())
                  cont.DeleteAnimFrame(frm);
               mod->in_container = NULL;
            }
         }
         break;
      case FRAME_SECTOR:
         {
            for(int j=sectors.size(); j--; )
               if(sectors[j]==frm) break;
            if(j!=-1){
               sectors[j] = sectors.back(); sectors.pop_back();
            }
            PI3D_sector sct = I3DCAST_SECTOR(frm);
            sct->Disconnect(this);
            if(current_camera && current_camera->GetCurrSector()==sct)
               current_camera->SetCurrSector(NULL);
         }
         break;
      }
      cont.frames[i] = cont.frames.back(); cont.frames.pop_back();
      return I3D_OK;
   }
   return I3DERR_OBJECTNOTFOUND;
}

//----------------------------
                              //scene sectors
PI3D_sector I3D_scene::CreateSector(){

   PI3D_sector sct = new I3D_sector(this, false);
   sectors.push_back(sct);
   return sct;
}

//----------------------------

PI3D_sound I3D_scene::CreateSound(){

   PI3D_sound sp = new I3D_sound(this);
   assert(sp);
   sounds.push_back(sp);
   drv->sounds.push_back(sp);
   return sp;
}

//----------------------------

I3D_RESULT I3D_scene::SetActiveCamera(PI3D_camera cam){

   current_camera = cam;
   return I3D_OK;
}

//----------------------------
                              //nested sectors not supported by this function!
void I3D_scene::SetupCameraSector(void *context){

                              //if no camera, there's nothing to update
   if(!current_camera)
      return;

   PI3D_sector old_sct = current_camera->GetCurrSector();
                              //determine sector we're in
   const S_vector &cpos = current_camera->GetWorldPos();

#if 1
   if(old_sct){
                              //try old sector first, maybe we're lucky
      if(old_sct->CheckPoint(cpos)){
         if(!current_camera->GetCurrSector()->IsPrimary1())
            return;
      }
   }
                              //determine sector
   for(int i=sectors.size(); i--; ){
      PI3D_sector sct = sectors[i];
      if(sct->IsSimple())
         continue;
      if(sectors[i]->CheckPoint(cpos))
         break;
   }
                              //note: cast to PI3D_sector, so that compiler doesn't generate smart-pointers
   PI3D_sector curr_sct = i!=-1 ? (PI3D_sector)sectors[i] : (PI3D_sector)primary_sector;
   current_camera->SetCurrSector(curr_sct);
                              //call-back user if requested
   if((cb_flags&SCNCALLBACK_SECTSWITCH) && old_sct!=current_camera->GetCurrSector())
      cb_proc(CBM_SECTORSWITCH, (dword)old_sct, (dword)(PI3D_sector)current_camera->GetCurrSector(),
      context);

#else
                              //nested version
   struct S_hlp{
      static I3DENUMRET I3DAPI enum_cam(PI3D_frame frm, dword dwcam){

         PI3D_camera cam = (PI3D_camera)dwcam;
         PI3D_sector sct = I3DCAST_SECTOR(frm);
         if(sct->CheckPoint(cam->GetWorldPos())){
                                    //maybe this
            cam->SetCurrSector(sct);
            return I3DENUMRET_OK;   //enum further - maybe in sub-sector
         }
         return I3DENUMRET_SKIPCHILDREN;
      }
   };
   if(old_sct){
                              //try old sector first, maybe we're lucky
      if(old_sct->CheckPoint(cpos)){
         if(current_camera->GetCurrSector()->IsPrimary1())
            EnumFrames(S_hlp::enum_cam, (dword)(PI3D_camera)current_camera, ENUMF_SECTOR);
         else{
                              //nested sectors
            old_sct->EnumFrames(S_hlp::enum_cam, (dword)(PI3D_camera)current_camera, ENUMF_SECTOR, NULL);
         }
         goto report;
      }
      current_camera->SetCurrSector(NULL);
   }
                              //determine sector
   EnumFrames(S_hlp::enum_cam, (dword)(PI3D_camera)current_camera, ENUMF_SECTOR);
   if(!current_camera->curr_sector)
      current_camera->SetCurrSector(primary_sector);
report:
                              //call-back user if requested
   if((cb_flags&SCNCALLBACK_SECTSWITCH) && old_sct!=current_camera->curr_sector)
      cb_proc(CBM_SECTORSWITCH, (dword)old_sct, (dword)(PI3D_sector)current_camera->curr_sector);

#endif
}

//----------------------------
                              //scene animation
void I3D_scene::AddAnimation(PI3D_animation_base ap, PI3D_frame fp, dword options){

   assert(ap);
   PI3D_interpolator intp = drv->CreateInterpolator();
   intp->SetFrame(fp);
   if(ap)
      intp->SetAnimation(0, ap, options);
   AddInterpolator(intp);
   intp->Release();
}

//----------------------------

void I3D_scene::DelAnimation(PI3D_frame frm){

   const PI3D_interpolator *intps = cont.GetInterpolators1();
   for(int j=cont.NumInterpolators1(); j--; )
   if(intps[j]->GetFrame() == frm){
      DeleteInterpolator(intps[j]);
   }
}

//----------------------------

I3D_RESULT I3D_scene::UnmapScreenPoint(int x1, int y1, S_vector &pos, S_vector &dir) const{

   if(!current_camera)
      return I3DERR_NOTINITIALIZED;

   if(x1<viewport[0] || x1>=viewport[2] || y1<viewport[1] || y1>=viewport[3])
      return I3DERR_INVALIDPARAMS;

   const S_matrix &cm = current_camera->GetMatrix();
                              //get camera position
   pos = cm(3);

                              //get centered screen position
   float x =  (((float)x1) - viewport_center[0]) / scrn_width2;
   float y = -(((float)y1) - viewport_center[1]) / scrn_width2;

   if(!current_camera->GetOrthogonal()){
                              //adjust height
      if(!(scn_flags&SCNF_VIEWPORT_ASPECT)) y *= (scrn_width2/-scrn_height2*.75f);
      y *= 1.333333f;
                              //perspective camera
      S_matrix m = current_camera->GetProjectionMatrix1();
      m(0, 3) = 0.0f; m(1, 3) = 0.0f; m(2, 3) = 0.0f; m(3, 3) = 1.0f;
      m = current_camera->GetViewMatrix1() * m;
      m.Invert();


      float z;
      {
         float ncp, fcp;
         current_camera->GetRange1(ncp, fcp);
         z = 0.0f;
         if(fcp>MRG_ZERO){
            float delta = fcp - ncp;
            if(delta>MRG_ZERO){
               z = fcp / delta;
            }
         }
      }
      dir = S_vector(x, y, z).RotateByMatrix(m);
   }else{
                              //orthogonal camera
      float s = 1.0f/current_camera->GetOrthoScale1();
      pos += cm(0) * (x*s) + cm(1) * (y*s);
      dir = cm(2);
   }

   dir.Normalize();

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::TransformPoints(const S_vector *in, S_vectorw *out, dword num) const{

   if(!current_camera)
      return I3DERR_GENERIC;

   current_camera->UpdateMatrices(aspect_ratio);
   ((I3D_scene*)this)->SetMViewProj(current_camera->GetViewMatrix1(), current_camera->GetProjectionMatrix1());

   ((I3D_scene*)this)->SetWorldMatrix(NULL);
   TransformVertsClipped(in, num, out, NULL, NULL);

   return I3D_OK;
}

//----------------------------

void I3D_scene::SetMViewProj(const S_matrix &m_view, const S_matrix &m_proj){

   m_view_proj_hom = m_view.Mult4X4(m_proj);
   mt_view_proj_hom = m_view_proj_hom; mt_view_proj_hom.Transpose();

   m_view_proj = m_view_proj_hom;
   for(int i=0; i<4; i++){
      m_view_proj(i, 0) *= scrn_width2;
      m_view_proj(i, 1) *= scrn_height2;
   }
}

//----------------------------

void I3D_scene::SetWorldMatrix(const S_matrix *m_world){

   if(!m_world)
      m_world_view_proj = m_view_proj;
   else
      m_world_view_proj.Make4X4(*m_world, m_view_proj);
}

//----------------------------

I3D_RESULT I3D_scene::SetViewport(dword l, dword t, dword r, dword b, bool aspect){

   viewport[0] = l;
   viewport[1] = t;
   viewport[2] = r;
   viewport[3] = b;

   scrn_width2 = (r - l) * .5f;
   scrn_height2 = -((int)b - (int)t) * .5f;
   
   viewport_center[0] = viewport[0] + scrn_width2;
   viewport_center[1] = viewport[1] - scrn_height2;

   if(aspect){
      scn_flags |= SCNF_VIEWPORT_ASPECT;
      aspect_ratio = (float)(b-t) / (float)(r-l);
   }else{
      scn_flags &= ~SCNF_VIEWPORT_ASPECT;
      aspect_ratio = .75f;
   }

   for(int i=0; i<4; i++){
      float f = viewport[i];
      fi_viewport[i] = FLOAT_BITMASK(f);
   }

   if(current_camera){
      current_camera->ResetProjectionMatrix();
      current_camera->ResetViewMatrix();
   }

   return I3D_OK;
}

//----------------------------
                              //computing 6 viewing frustum planes
                              //in parametric equations
                              //normals of planes have size 1.0

void I3D_scene::PrepareViewFrustum(S_preprocess_context &pc){

   S_view_frustum &vf = pc.view_frustum;
   I3D_bsphere &vf_sphere = pc.vf_sphere;
   vf.num_clip_planes = 6;

   const S_matrix &m = current_camera->matrix;
   const S_vector &cam_pos = m(3);
   float fcp = current_camera->GetFCP();
   float inv_aspect = (scn_flags&SCNF_VIEWPORT_ASPECT) ?
      (-scrn_height2 / scrn_width2) : .75f;

   int i;

   if(!current_camera->GetOrthogonal()){
      const float &fov = current_camera->GetFOV1();

      float sin_fov = sin(fov*.5f);
      float cos_fov = cos(fov*.5f);

      static const E_VIEW_FRUSTUM_SIDE side_map[4] = {
         VF_LEFT, VF_RIGHT, VF_BOTTOM, VF_TOP
      };

                              //get all planes
      for(i=0; i<2; i++){
         E_VIEW_FRUSTUM_SIDE side_max = side_map[i*2+1];
         E_VIEW_FRUSTUM_SIDE side_min = side_map[i*2];

         S_vector &n_max = vf.clip_planes[side_max].normal;
         float &d_max = vf.clip_planes[side_max].d;
         S_vector &n_min = vf.clip_planes[side_min].normal;
         float &d_min = vf.clip_planes[side_min].d;

         if(!i){              //x
            n_max.x = cos_fov;
            n_min.x = -n_max.x;
            n_max.y = 0.0f;
            n_min.y = 0.0f;
            n_max.z = -sin_fov;
            n_min.z = n_max.z;
         }else{               //y
            n_max.x = 0.0f;
            n_min.x = 0.0f;
            n_max.y = cos_fov;
            n_max.z = -sin_fov;
            n_max.z *= inv_aspect;
            n_max.Normalize();

            n_min.y = -n_max.y;
            n_min.z = n_max.z;
         }
                              //rotate by view
         n_max = n_max.RotateByMatrix(current_camera->matrix);
         n_min = n_min.RotateByMatrix(current_camera->matrix);

         d_max = - n_max.Dot(cam_pos);
         d_min = - n_min.Dot(cam_pos);
      }
   }else{                     //orthogonal
      float o_scale = 1.0f / current_camera->GetOrthoScale1();
                              //horizontal
      vf.clip_planes[VF_RIGHT].normal = m(0);
      vf.clip_planes[VF_RIGHT].d = -cam_pos.Dot(vf.clip_planes[VF_RIGHT].normal);

      vf.clip_planes[VF_LEFT].normal = -m(0);
      vf.clip_planes[VF_LEFT].d = -vf.clip_planes[VF_RIGHT].d;

      vf.clip_planes[VF_RIGHT].d -= o_scale;
      vf.clip_planes[VF_LEFT].d -= o_scale;
                              //vertical
      vf.clip_planes[VF_TOP].normal = m(1);
      vf.clip_planes[VF_TOP].d = -cam_pos.Dot(vf.clip_planes[VF_TOP].normal);

      vf.clip_planes[VF_BOTTOM].normal = -m(1);
      vf.clip_planes[VF_BOTTOM].d = -vf.clip_planes[VF_TOP].d;

      vf.clip_planes[VF_TOP].d -= o_scale * inv_aspect;
      vf.clip_planes[VF_BOTTOM].d -= o_scale * inv_aspect;
   }
   {                       //z
      S_vector &n_max = vf.clip_planes[VF_BACK].normal;
      float &d_max = vf.clip_planes[VF_BACK].d;
      S_vector &n_min = vf.clip_planes[VF_FRONT].normal;
      float &d_min = vf.clip_planes[VF_FRONT].d;

                           //we look along z axis, so use direction vector
      n_max = m(2);
      n_min = -n_max;

      d_max = - n_max.Dot(cam_pos);
      d_min = - d_max;// + current_camera->GetNCP();
      d_max -= fcp;
   }

                              //prepare bounding sphere
   if(!current_camera->GetOrthogonal()){
      {
                              //get intersection of 2 planes
         const S_plane &pl1 = vf.clip_planes[VF_RIGHT];
         const S_plane &pl2 = vf.clip_planes[VF_TOP];
         S_vector pt, dir;
         bool b = pl2.IntersectionLine(pl1, pt, dir);
         assert(b);
                              //get intersection of line and far plane
         float u = vf.clip_planes[VF_BACK].normal.Dot(dir);
         assert(!IsAbsMrgZero(u));
         u = -(vf.clip_planes[VF_BACK].normal.Dot(pt) + vf.clip_planes[VF_BACK].d) / u;
         S_vector v_fcp = pt + dir * u;

         S_vector l_half = cam_pos + (v_fcp - cam_pos) * .5f;

         S_vector n; n.GetNormal(cam_pos + m(2)*fcp, cam_pos, v_fcp); n.Normalize();
         n.GetNormal(cam_pos + n, cam_pos, l_half); n.Normalize();
         float u1, u2;
         b = LineIntersection(cam_pos, m(2), l_half, n, u1, u2);
         assert(b);
         if(u1 >= fcp){
            vf_sphere.pos = cam_pos + m(2) * fcp;
            vf_sphere.radius = (vf_sphere.pos - v_fcp).Magnitude();
         }else{
            vf_sphere.pos = cam_pos + m(2) * u1;
            vf_sphere.radius = (cam_pos - vf_sphere.pos).Magnitude();
         }
      }
   }else{
                              //todo: compute proper bounding sphere
      vf_sphere.pos.Zero();
      vf_sphere.radius = 1e+16f;
   }
                              //get 4 edge points on fcp
                              // for portal transforms
   static const int plane_index[] = { 
      VF_BOTTOM,
      VF_LEFT,
      VF_TOP,
      VF_RIGHT,
      VF_BOTTOM,
   };
                              //unoptimized...
   for(i=0; i<4; i++){
                              //get intersection of 2 planes
      S_plane pl1, pl2;
      pl1.normal = vf.clip_planes[plane_index[i]].normal;
      pl1.d = vf.clip_planes[plane_index[i]].d;
      pl2.normal = vf.clip_planes[plane_index[i+1]].normal;
      pl2.d = vf.clip_planes[plane_index[i+1]].d;
      S_vector pt, dir;
      pl2.IntersectionLine(pl1, pt, dir);
                              //get intersection of line and plane
      const S_plane &pl = vf.clip_planes[VF_BACK];
      float u = pl.normal.Dot(dir);
      assert(!IsAbsMrgZero(u));
      u = -(pl.normal.Dot(pt) + pl.d) / u;
   }
   //pc.exp_frustum.Make(vf);
}

//----------------------------

void I3D_scene::GetViewport(dword &l, dword &t, dword &r, dword &b){

   l = viewport[0];
   t = viewport[1];
   r = viewport[2];
   b = viewport[3];
}

//----------------------------

static S_vectorw trv_dummy_v;
static const float f1 = 1.0f;

void I3D_scene::TransformVertsClipped(const S_vector *verts, dword num_verts,
   S_vectorw *verts_out, dword *clip_union1, dword *clip_intersection1) const{

   int i;

   S_vectorw *vp;
   dword v_stride;
                              //determine if user wants output
   if(verts_out){
      vp = verts_out;
      v_stride = sizeof(S_vectorw);
   }else{
      vp = &trv_dummy_v;
      v_stride = 0;
   }
   const S_vector *in_v = verts;
                              //determine, if clip tests required
   dword clip_union = 0;
   dword clip_intersection = I3DCLIP_ALL;

   if(i=num_verts){

#if defined _MSC_VER && defined USE_ASM

      __asm{
         push ecx
                              //setup registers:
                              //eax - free
                              //ebx - m_current
                              //ecx - loop
                              //edx - fi_viewport[]
                              //esi - in_vp
                              //edi - vp
         mov edi, this
         lea eax, [edi].viewport_center
         fld dword ptr [eax]
         fld dword ptr [eax+4]
         lea ebx, [edi].m_world_view_proj
         lea edx, [edi].fi_viewport
         mov ecx, i
         mov esi, in_v
         mov edi, vp
         push eax             //make local variable

      main_loop:
                           //rhw
         fld dword ptr  [esi + 0x00]
         fmul dword ptr [ebx + 0x0c]
         fadd dword ptr [ebx + 0x3c]
         fld dword ptr  [esi + 0x04]
         fmul dword ptr [ebx + 0x1c]
         faddp st(1), st
         fld dword ptr  [esi + 0x08]
         fmul dword ptr [ebx + 0x2c]
         faddp st(1), st
         fabs
         fst dword ptr[esp]
         cmp dword ptr[esp], 0x322BCC77   //bitmask of MRG_ZERO (1e-8)
         ja ok1
         fstp dword ptr[esp]
         mov dword ptr [edi + 0x00], 0
         mov dword ptr [edi + 0x04], 0
         mov dword ptr [edi + 0x08], 0
         mov dword ptr [edi + 0x0c], 0
         jmp clip
      ok1:
         fld1
         fdivrp st(1), st
                              //store rhw
         fst dword ptr[edi + 0x0c]

                           //x
         fld dword ptr  [esi + 0x00]
         fmul dword ptr [ebx + 0x00]
         fld dword ptr  [esi + 0x04]
         fmul dword ptr [ebx + 0x10]
         faddp st(1), st
         fld dword ptr  [esi + 0x08]
         fmul dword ptr [ebx + 0x20]
         faddp st(1), st
         fadd dword ptr  [ebx + 0x30]
         fmul st, st(1)
         fadd st, st(3)
         fstp dword ptr [edi + 0x00]

                           //y
         fld dword ptr  [esi + 0x00]
         fmul dword ptr [ebx + 0x04]
         fld dword ptr  [esi + 0x04]
         fmul dword ptr [ebx + 0x14]
         faddp st(1), st
         fld dword ptr  [esi + 0x08]
         fmul dword ptr [ebx + 0x24]
         faddp st(1), st
         fadd dword ptr  [ebx + 0x34]
         fmul st, st(1)
         fadd st, st(2)
         fstp dword ptr [edi + 0x04]

                           //z
         fld dword ptr  [esi + 0x00]
         fmul dword ptr [ebx + 0x08]
         fld dword ptr  [esi + 0x04]
         fmul dword ptr [ebx + 0x18]
         faddp st(1), st
         fld dword ptr  [esi + 0x08]
         fmul dword ptr [ebx + 0x28]
         faddp st(1), st
         fadd dword ptr [ebx + 0x38]
         fmulp st(1), st
         fstp dword ptr [edi + 0x08]

                           //clipping

                           //clip x
      clip:
         mov eax, [edi]
         cmp eax, [edx + 0x00]
         jg skipx1
         or clip_union, I3DCLIP_LEFT
         and clip_intersection, ~I3DCLIP_RIGHT
         jmp skipx2
      skipx1:
         and clip_intersection, ~I3DCLIP_LEFT
         cmp eax, [edx + 0x08]
         jl skipx3
         or clip_union, I3DCLIP_RIGHT
         jmp skipx2
      skipx3:
         and clip_intersection, ~I3DCLIP_RIGHT
      skipx2:

                           //clip y
         mov eax, [edi + 0x04]
         cmp eax, [edx + 0x04]
         jg skipy1
         or clip_union, I3DCLIP_TOP
         and clip_intersection, ~I3DCLIP_BOTTOM
         jmp skipy2
      skipy1:
         and clip_intersection, ~I3DCLIP_TOP
         cmp eax, [edx + 0x0c]
         jl skipy3
         or clip_union, I3DCLIP_BOTTOM
         jmp skipy2
      skipy3:
         and clip_intersection, ~I3DCLIP_BOTTOM
      skipy2:

                           //clip z
         mov eax, [edi + 0x08]
         test eax, eax
         jg skipz1
         or clip_union, I3DCLIP_FRONT
         and clip_intersection, ~I3DCLIP_BACK
         jmp skipz2
      skipz1:
         and clip_intersection, ~I3DCLIP_FRONT
         cmp eax, 0x3f800000  //bitmask of 1.0f
         jl skipz3
         or clip_union, I3DCLIP_BACK
         jmp skipz2
      skipz3:
         and clip_intersection, ~I3DCLIP_BACK
      skipz2:

                           //next vertex
         add esi, 12
         add edi, v_stride

                           //loop
         dec ecx
         jne main_loop

                              //unload values from FPU
         fstp dword ptr[esp]
         fstp dword ptr[esp]
         pop eax

         pop ecx
      }

#else
//----------------------------
      do{
         float rhw = fabs(in_v->x*m_current.m[0][3] + in_v->y*m_current.m[1][3] +
            in_v->z*m_current.m[2][3] + m_current.m[3][3]);
         if(IsMrgZero(rhw)){
            vp->xyz.x = 0.0f;
            vp->xyz.y = 0.0f;
            vp->xyz.z = 0.0f;
            vp->w = 0.0f;
            //goto clip;
         }else{
            rhw = 1.0f / rhw;

            vp->w = rhw;

            vp->xyz.x = (in_v->x*m_current.m[0][0] + in_v->y*m_current.m[1][0] +
               in_v->z*m_current.m[2][0] + m_current.m[3][0])
               * rhw
               + viewport_center[0];

            vp->xyz.y = (in_v->x*m_current.m[0][1] + in_v->y*m_current.m[1][1] +
               in_v->z*m_current.m[2][1] + m_current.m[3][1]) 
               * rhw
               + viewport_center[1];

            vp->xyz.z = (in_v->x*m_current.m[0][2] + in_v->y*m_current.m[1][2] +
               in_v->z*m_current.m[2][2] + m_current.m[3][2]) 
               * rhw;
         }

                           //clip X
         int x = FLOAT_BITMASK(vp->xyz.x);
         if(x <= fi_viewport[0]){
            clip_union |= I3DCLIP_LEFT;
            clip_intersection &= ~I3DCLIP_RIGHT;
         }else{
            clip_intersection &= ~I3DCLIP_LEFT;
            if(x >= fi_viewport[2]) clip_union |= I3DCLIP_RIGHT;
            else clip_intersection &= ~I3DCLIP_RIGHT;
         }

                           //clip Y
         int y = FLOAT_BITMASK(vp->xyz.y);
         if(y <= fi_viewport[1]){
            clip_union |= I3DCLIP_TOP;
            clip_intersection &= ~I3DCLIP_BOTTOM;
         }else{
            clip_intersection &= ~I3DCLIP_TOP;
            if(y >= fi_viewport[3]) clip_union |= I3DCLIP_BOTTOM;
            else clip_intersection &= ~I3DCLIP_BOTTOM;
         }

                           //clip Z
         int z = FLOAT_BITMASK(vp->xyz.z);
         if(z <= 0){
            clip_union |= I3DCLIP_FRONT;
            clip_intersection &= ~I3DCLIP_BACK;
         }else{
            clip_intersection &= ~I3DCLIP_FRONT;
            if(z >= FLOAT_BITMASK(f1)) clip_union |= I3DCLIP_BACK;
            else clip_intersection &= ~I3DCLIP_BACK;
         }
         

         ++in_v;
         vp = (S_vectorw*)(((byte*)vp)+v_stride);
      }while(--i);
#endif

   }
   if(clip_union1) *clip_union1 = clip_union;
   if(clip_intersection1) *clip_intersection1 = clip_intersection;
}

//----------------------------
#define SECTOR_SOUND_FADE 1000

void I3D_scene::UpdateSoundProperties(int time, PI3D_sector prev_sector){

   PISND_driver isp = drv->GetSoundInterface();
   if(!isp)
      return;
                              //check current sector, update if necessary
   assert(current_camera);
   PI3D_sector sct = current_camera->GetCurrSector();
   assert(sct);

   bool sector_switch = (prev_sector!=sct);
   if(sector_switch){
      if(!prev_sector)
         ssp_count_down = 0;
      else{
                              //start fade
         if(!ssp_count_down){
                              //setup source fade values
            const S_sound_env_properties *ep = drv->GetEnvProperty(prev_sector->GetSoundEnvID());
            if(!ep)
               ep = drv->GetEnvProperty(0);
            ssp_fade = *ep;
         }
         ssp_count_down = SECTOR_SOUND_FADE;
      }
   }

   if(ssp_count_down || (drv->GetFlags()&DRVF_SOUND_ENV_RESET)){

      drv->drv_flags &= ~DRVF_SOUND_ENV_RESET;

#if defined USE_EAX && 1

      const S_sound_env_properties *ssp_dst = drv->GetEnvProperty(sct->GetSoundEnvID());
      if(!ssp_dst)
         ssp_dst = drv->GetEnvProperty(0);

                              //compute faded values
      S_sound_env_properties ssp;
      if(ssp_count_down){
         if((ssp_count_down -= time) <= 0){
            ssp_count_down = 0;
         }
         ssp.MakeBlend(*ssp_dst, ssp_fade, (float)ssp_count_down/(float)SECTOR_SOUND_FADE);
      }else
         ssp = *ssp_dst;

                              //setup room (lf, hf)
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_ROOM,
         EAXLISTENER_MINROOM + ssp.room_level_lf * (EAXLISTENER_MAXROOM - EAXLISTENER_MINROOM));
      
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_ROOMHF,
         EAXLISTENER_MINROOMHF + ssp.room_level_hf * (EAXLISTENER_MAXROOMHF - EAXLISTENER_MINROOMHF));
      
                              //setup decay (time, hf)
      float decay_time = Max(EAXLISTENER_MINDECAYTIME, Min(EAXLISTENER_MAXDECAYTIME,
         (float)ssp.decay_time*.001f));
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_DECAYTIME, FloatAsInt(decay_time));

      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_DECAYHFRATIO,
         FloatAsInt(EAXLISTENER_MINDECAYHFRATIO + ssp.decay_hf_ratio * (EAXLISTENER_MAXDECAYHFRATIO - EAXLISTENER_MINDECAYHFRATIO)));

                              //setup reflections (amount, delay)
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_REFLECTIONS,
         EAXLISTENER_MINREFLECTIONS + ssp.reflections * (EAXLISTENER_MAXREFLECTIONS - EAXLISTENER_MINREFLECTIONS));

      float refl_delay = Max(EAXLISTENER_MINREFLECTIONSDELAY, Min(EAXLISTENER_MAXREFLECTIONSDELAY,
         (float)ssp.reflections_delay*.001f));
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_REFLECTIONSDELAY, FloatAsInt(refl_delay));

                              //setup reverb (amount, delay)
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_REVERB,
         EAXLISTENER_MINREVERB + ssp.reverb * (EAXLISTENER_MAXREVERB - EAXLISTENER_MINREVERB));

      float rev_delay = Max(EAXLISTENER_MINREVERBDELAY, Min(EAXLISTENER_MAXREVERBDELAY,
         (float)ssp.reverb_delay*.001f));
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_REVERBDELAY, FloatAsInt(rev_delay));
      
                              //setup diffusion
      isp->SetEAXProperty(DSPROPERTY_EAXLISTENER_ENVIRONMENTDIFFUSION, FloatAsInt(ssp.env_diffusion));

#endif //USE_EAX
   }
   const S_matrix &m = current_camera->GetMatrixDirect();
   isp->SetListenerPos(m(3));
   isp->SetListenerDir(m(2), m(1));

   isp->UpdateListener();
}

//----------------------------

void I3D_scene::AttenuateSounds(){

   struct S_hlp{
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword){

         I3DCAST_SOUND(frm)->VirtualPause();
         return I3DENUMRET_OK;
      }
   };
   EnumFrames(S_hlp::cbEnum, 0, ENUMF_SOUND | ENUMF_BACKDROP);
}

//----------------------------

I3D_RESULT I3D_scene::SetRenderMatrix(const S_matrix &m){

   S_matrix mt;
   mt.Make4X4Transposed(m, m_view_proj_hom);
   drv->GetDevice1()->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0, &mt, 4);

   return I3D_OK;
}

//----------------------------

static const dword vs_decl_line[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_END()
}, *vs_decl_triangles = vs_decl_line;

//----------------------------

I3D_RESULT I3D_scene::DrawLine(const S_vector &v1, const S_vector &v2,
   const S_vector &c, dword flags){

   if(flags&I3DDRAW_TRANPARENCY_MASK)
      drv->SetupBlend(I3DBLEND_ALPHABLEND);
   else
      drv->SetupBlend(I3DBLEND_OPAQUE);
   drv->SetupTextureStage(0, D3DTOP_MODULATE);

   drv->DisableTextureStage(1);
   drv->SetTexture(NULL);
   drv->SetClipping(true);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   I3D_driver::S_vs_shader_entry_in se(vs_decl_line);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
   drv->SetVertexShader(drv->GetVSHandle(se)->vs_handle);
   S_vectorw c4 = c;
   c4.w = 1.0f - (float)(flags&I3DDRAW_TRANPARENCY_MASK)/255.0f;
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &c4, 1);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

   S_vector v[2] = {v1, v2};
   hr = lpDev8->DrawPrimitiveUP(D3DPT_LINELIST, 1, v, sizeof(S_vector));
   CHECK_D3D_RESULT("DrawPrimitiveUP", hr);

   drv->ResetStreamSource();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::DrawLines(const S_vector *v, dword numv,
   const word *indx, dword numi, const S_vector &c, dword v_stride,
   dword flags){

   if(!numi)
      return I3D_OK;

   if(flags&I3DDRAW_TRANPARENCY_MASK)
      drv->SetupBlend(I3DBLEND_ALPHABLEND);
   else
      drv->SetupBlend(I3DBLEND_OPAQUE);
   drv->SetupTextureStage(0, D3DTOP_MODULATE);

   drv->DisableTextureStage(1);
   drv->SetTexture(NULL);
   drv->SetClipping(true);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   I3D_driver::S_vs_shader_entry_in se(vs_decl_line);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
   drv->SetVertexShader(drv->GetVSHandle(se)->vs_handle);
   S_vectorw c4 = c;
   c4.w = 1.0f - (float)(flags&I3DDRAW_TRANPARENCY_MASK)/255.0f;
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &c4, 1);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

   hr = lpDev8->DrawIndexedPrimitiveUP(D3DPT_LINELIST, 0, numv,
      numi / 2, indx, D3DFMT_INDEX16, v, v_stride);
   CHECK_D3D_RESULT("DrawIndexedPrimitiveUP", hr);

   drv->ResetStreamSource();
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_scene::DrawTriangles(const void *v, dword numv, const word *indx,
   dword numi, const S_vector &c, dword v_stride, dword flags, PI3D_text_coor txc){

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   byte alpha = 255 - (flags&I3DDRAW_TRANPARENCY_MASK);

   if(flags&I3DDRAW_TRANPARENCY_MASK){
      drv->SetupAlphaTest(true);
      drv->SetAlphaRef(0x20 * alpha / 255);
      drv->SetupBlend(I3DBLEND_ALPHABLEND | I3DBLEND_VERTEXALPHA);
   }else
      drv->SetupBlend(I3DBLEND_OPAQUE);
   drv->SetupTextureStage(0, D3DTOP_MODULATE);

   drv->DisableTextureStage(1);
   drv->SetClipping(true);

   if(!(flags&I3DDRAW_TRANSFORMED)){

      I3D_driver::S_vs_shader_entry_in se(vs_decl_triangles);
      se.AddFragment(I3D_driver::VSF_TRANSFORM);
      se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
      drv->SetVertexShader(drv->GetVSHandle(se)->vs_handle);
      S_vectorw c4 = c;
      c4.w = (float)alpha / 255.0f;
      hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &c4, 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

      hr = lpDev8->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, numv,
         numi / 3, indx, D3DFMT_INDEX16, v, v_stride);
      CHECK_D3D_RESULT("DrawIndexedPrimitiveUP", hr);
   }else{
      dword dc = D3DCOLOR_ARGB(alpha,
         FloatToInt(255.0f*c[0]), FloatToInt(255.0f*c[1]), FloatToInt(255.0f*c[2]));

      const S_vectorw *vp4 = (const S_vectorw*)v;

      struct S_vertex{
         S_vectorw v;
         dword diffuse;
         I3D_text_coor txc;
      } *vp = new S_vertex[numv];
      for(int i=0; i<numv; i++){
         vp[i].v = *vp4;
         vp[i].diffuse = dc;
         vp4 = (S_vectorw*)(((byte*)vp4) + v_stride);
         if(txc){
            vp[i].txc = (*txc);
            txc = (PI3D_text_coor)(((byte*)txc) + v_stride);
         }
      }
      HRESULT hr;

      drv->SetVertexShader(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);

      hr = lpDev8->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, numv,
         numi / 3, indx, D3DFMT_INDEX16, vp, sizeof(S_vertex));
      CHECK_D3D_RESULT("DrawIndexedPrimitiveUP", hr);
      delete[] vp;
   }

   if(flags&I3DDRAW_TRANPARENCY_MASK)
      drv->SetupAlphaTest(false);

   drv->ResetStreamSource();
   return I3D_OK;
}

//----------------------------

static const dword vs_decl_icon[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

I3D_RESULT I3D_scene::DrawSprite(const S_vector &pos, PI3D_material mat,
   byte inv_alpha, float size){

   if(!current_camera)
      return I3DERR_INVALIDPARAMS;
   const S_matrix &m = current_camera->GetMatrix();

   struct S_vertex{
      S_vector v;
      I3D_text_coor tex;
   } verts[4];

   byte alpha = 255 - inv_alpha;
   const S_vector &mat_color = mat->GetMaterialData().diffuse;

   for(int j=0; j<4; j++){
      S_vector &v = verts[j].v;
      verts[j].tex.u = (j&1);
      verts[j].tex.v = (float)((j&2)>>1);
      v = pos;
      v -= m(0) * (size*.5f);
      v -= m(1) * (size*.5f);
      if(j&1)
         v += m(0) * size;
      if(!(j&2))
         v += m(1) * size;
   }
   drv->SetRenderMat(mat);
   drv->SetupBlend(inv_alpha ? (I3DBLEND_ALPHABLEND | I3DBLEND_VERTEXALPHA) : (mat->IsTransl() ? I3DBLEND_ALPHABLEND : I3DBLEND_OPAQUE));
   drv->SetupTextureStage(0, D3DTOP_MODULATE);
   drv->DisableTextureStage(1);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   drv->SetupAlphaTest(true);
   dword ar = (0x20 * alpha)>>8;
   drv->SetAlphaRef(ar);
   
   drv->SetClipping(true);

                              //setup shader
   I3D_driver::S_vs_shader_entry_in se(vs_decl_icon);
   se.AddFragment(I3D_driver::VSF_TRANSFORM);
   se.AddFragment(I3D_driver::VSF_LIGHT_COPY);
   se.AddFragment(I3D_driver::VSF_COPY_UV_0);
   drv->SetVertexShader(drv->GetVSHandle(se)->vs_handle);
   S_vectorw c4 = mat_color;
   c4.w = alpha / 255.0f;
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &c4, 1);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0, &mt_view_proj_hom, 4);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

   hr = lpDev8->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, verts, sizeof(S_vertex));
   CHECK_D3D_RESULT("DrawPrimitiveUP", hr);

   drv->SetupAlphaTest(false);
   drv->ResetStreamSource();
   return true;
}

//----------------------------

void I3D_scene::DrawIcon(const S_vector &pos, int res_id, byte inv_alpha, float size){

   PI3D_material mat = drv->GetIcon(res_id);
   if(mat)
      DrawSprite(pos, mat, inv_alpha, size);
}

//----------------------------

#define SPHERE_NUM_VERTS 24
static S_vector sphere_circle[3][SPHERE_NUM_VERTS];
static word sphere_indicies[3][SPHERE_NUM_VERTS*2];
static word cylinder_indicies[3][SPHERE_NUM_VERTS*2];

static struct S_sphere_auto_init{
   S_sphere_auto_init(){
                              //init volume sphere
      for(int j=0; j<3; j++){
         for(int i=0; i<SPHERE_NUM_VERTS; i++){
            sphere_indicies[j][i*2+0] = SPHERE_NUM_VERTS*j+i+0;
            sphere_indicies[j][i*2+1] = SPHERE_NUM_VERTS*j+i+1;

            float s = sin(GetPI2()*i/SPHERE_NUM_VERTS),
                  c = cos(GetPI2()*i/SPHERE_NUM_VERTS);

            sphere_circle[j][i].x = (j ? s : 0);
            sphere_circle[j][i].y = (!j ? s : j==2 ? c : 0);
            sphere_circle[j][i].z = ((j<2) ? c : 0);
         }
         sphere_indicies[j][i*2-1] = SPHERE_NUM_VERTS*j;
      }
      memcpy(cylinder_indicies[0], sphere_indicies[0], sizeof(cylinder_indicies[0])*2);
      for(int i=SPHERE_NUM_VERTS; i--; ){
         cylinder_indicies[2][i*2+0] = i;
         cylinder_indicies[2][i*2+1] = SPHERE_NUM_VERTS + i;
      }
   }
} sphere_auto_init;

//----------------------------

void I3D_scene::DebugDrawSphere(const S_matrix &m1, float radius, const S_vector &color, byte alpha){

                              //scale by radius
   S_matrix m = GetIdentityMatrix();
   m(0, 0) = m(1, 1) = m(2, 2) = radius;
   SetRenderMatrix(m * m1);
   DrawLines(sphere_circle[0], SPHERE_NUM_VERTS*3, sphere_indicies[0], SPHERE_NUM_VERTS*2*3, color, sizeof(S_vector), alpha);
}

//----------------------------

void I3D_scene::DebugDrawCylinder(const S_matrix &m1, float radius1, float radius2, float height, const S_vector &color, byte alpha){

   S_vector circle[2][SPHERE_NUM_VERTS];
   for(int i=SPHERE_NUM_VERTS; i--; ){
      circle[0][i].x = sphere_circle[2][i].x * radius1;
      circle[0][i].z = sphere_circle[2][i].y * radius1;
      circle[0][i].y = 0.0f;

      circle[1][i].x = sphere_circle[2][i].x * radius2;
      circle[1][i].z = sphere_circle[2][i].y * radius2;
      circle[1][i].y = height;
   }
   SetRenderMatrix(m1);
   DrawLines(circle[0], SPHERE_NUM_VERTS*2, cylinder_indicies[0], SPHERE_NUM_VERTS*2*3, color, sizeof(S_vector), alpha);
}

//----------------------------

void I3D_scene::DebugDrawArrow(const S_matrix &m1, float lenght, const S_vector &color, byte alpha,
   int axis){

   static const S_vector dir_lines[3][7] = {
                              //x-oriented
      S_vector(0, 0, 0), S_vector(1, 0, 0),
      S_vector(1.3f, 0.0f, 0.0f),
      S_vector(1.0f, .1f, .1f), S_vector(1.0f, -.1f, .1f), S_vector(1.0f, -.1f, -.1f), S_vector(1.0f, .1f, -.1f),
                              //y-oriented
      S_vector(0, 0, 0), S_vector(0, 1, 0),
      S_vector(0.0f, 1.3f, 0.0f),
      S_vector(.1f, 1.0f, .1f), S_vector(-.1f, 1.0f, .1f), S_vector(-.1f, 1.0f, -.1f), S_vector(.1f, 1.0f, -.1f),
                              //z-oriented
      S_vector(0, 0, 0), S_vector(0, 0, 1),
      S_vector(0.0f, 0.0f, 1.3f),
      S_vector(.1f, .1f, 1.0f), S_vector(-.1f, .1f, 1.0f), S_vector(-.1f, -.1f, 1.0f), S_vector(.1f, -.1f, 1.0f),
   };
   static const word indx[] = {
      0, 1,
      2, 3, 2, 4, 2, 5, 2, 6,
      3, 4, 4, 5, 5, 6, 6, 3
   };
   S_matrix m = GetIdentityMatrix();
   m(axis, axis) = lenght;
   SetRenderMatrix(m * m1);
   DrawLines(dir_lines[axis], sizeof(dir_lines[0])/sizeof(S_vector),
      indx, sizeof(indx)/sizeof(word), color, sizeof(S_vector), alpha);
}

//----------------------------

void I3D_scene::DebugDrawBBox(const S_vector *bbox_full, const S_vector &color, byte alpha){

   static word bbox_list[] = {
      0, 1, 2, 3, 4, 5, 6, 7,
      0, 2, 1, 3, 4, 6, 5, 7,
      0, 4, 1, 5, 2, 6, 3, 7
   };
   DrawLines(bbox_full, 8, bbox_list, sizeof(bbox_list)/sizeof(word), color, sizeof(S_vector), alpha);
}

//----------------------------

void I3D_scene::DebugDrawContour(const S_vector *contour_points, dword num_cpts, const S_vector &color, byte alpha){

   for(int i=0; i<num_cpts; i++){
      DrawLine(contour_points[i], contour_points[(i+1)%num_cpts], color, alpha);
   }
}

//----------------------------

I3D_RESULT I3D_scene::CreateBspTree(I3D_BSP_BUILD_CB_PROC *cbP, void *context){

   I3D_RESULT ir = bsp_builder.Build(cbP, context);
   bsp_builder.GetStats(drv->bsp_stats);
   return ir;
}

//----------------------------

void I3D_scene::DestroyBspTree(){

   bsp_builder.Clear();
   bsp_builder.SetAbsentVolumesPickable();
}

//----------------------------
bool I3D_scene::SaveBsp(C_cache *cp){

   I3D_RESULT ir = I3D_OK;
   ir = bsp_builder.Save(cp);
   return I3D_SUCCESS(ir);
}

//----------------------------

bool I3D_scene::LoadBsp(C_cache *cp, I3D_LOAD_CB_PROC *cbP, void *context, dword check_flags){

   I3D_RESULT ir = I3D_OK;
   ir = bsp_builder.Load(cp, cbP, context, check_flags);
   if(I3D_FAIL(ir)){
      bsp_builder.Clear();
   }
   bsp_builder.GetStats(drv->bsp_stats);
   return I3D_SUCCESS(ir);
}

//----------------------------

void I3D_scene::ResetDynamicVolTreeEx(){

   ResetDynamicVolTree();
}

//----------------------------

bool I3D_scene::IsBspBuild(){

   return bsp_builder.IsValid();
}

//----------------------------

void I3D_scene::StoreFileName(const char *fn){ cont.StoreFileName(fn); }
const char *I3D_scene::GetFileName() const{ return cont.GetFileName(); }
dword I3D_scene::NumFrames() const{ return cont.NumFrames1(); }
const PI3D_frame *I3D_scene::GetFrames() const{ return cont.GetFrames(); }
I3D_RESULT I3D_scene::AddInterpolator(PI3D_interpolator i){ return cont.AddInterpolator(i); }
I3D_RESULT I3D_scene::DeleteInterpolator(PI3D_interpolator i){ return cont.DeleteInterpolator(i); }
dword I3D_scene::NumInterpolators() const{ return cont.NumInterpolators1(); }
const PI3D_interpolator *I3D_scene::GetInterpolators() const{ return cont.GetInterpolators1(); }
I3D_RESULT I3D_scene::Tick(int time){ return cont.Tick(time); }

//----------------------------
//----------------------------
//----------------------------
