/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Sng_mesh.cpp
   Content: Single-mesh frame.
--------------------------------------------------------*/

#include "all.h"
#include "sng_mesh.h"
#include "camera.h"
#include "scene.h"
#include <malloc.h>           //alloca


                              //make bounding volume slightly greater,
                              // for safe purposes
const float SAFE_BOUND_THRESH = .001f;

                              //debugging:
//#define DEBUG_NO_CACHE_LIGHT  //don't cache light info in vertices
//#define DEBUG_PAINT_VERTEX_ASSIGNMENT
//#define DEBUG_PAINT_ASSIGN_WAVE

//----------------------------

static struct S_auto_init_smesh{
   static I3D_visual *Create(PI3D_driver drv){
      return new I3D_object_singlemesh(drv);
   }
   S_auto_init_smesh(){
      RegisterVisual(Create, I3D_VISUAL_SINGLEMESH, "SingleMesh");
   }
} auto_init_smesh;

//----------------------------
//----------------------------

I3D_singlemesh::I3D_singlemesh(PI3D_driver d):
   I3D_mesh_base(d, FVF_VERTEX_SINGLEMESH)
{
   drv->AddCount(I3D_CLID_SINGLE_MESH);
}

//----------------------------

I3D_singlemesh::~I3D_singlemesh(){

   drv->DecCount(I3D_CLID_SINGLE_MESH);
}

//----------------------------
// check if vertex is in a bounding-box,
// also compute weight value, based on distance
// in bbox's Z coord
// if test succeeded, weight is in range 0.0f ... 1.0f
bool IsPointInBBox(const S_vector &v, const I3D_bbox &bbox, const S_matrix &m, float &weight){

   I3D_bbox bbox1;
   bbox1.min = bbox.min * m;
   bbox1.max = bbox.max * m;
                              //process 3 axes of bound-box
   for(int i=0; i<3; i++){
      float v_dot = v.Dot(m(i));
                              //check max side
      float dx = bbox1.max.Dot(m(i));
      if(v_dot <= dx){
                              //check min side
         float dn = bbox1.min.Dot(m(i));
         if(v_dot >= dn){
            if(i==2){
                              //compute weight
               float f = (dx - dn);
               assert(!IsMrgZero(f));
               weight = (v_dot - dn) / f;
            }
            continue;
         }
      }
                              //can't be in, don't bother further
      return false;
   }
   return true;
}

//----------------------------

struct S_SM_joint_info{
   S_matrix m_loc;            //joint's bbox matrix in local coords (relative to single-mesh)
   PI3D_joint joint;
   word first_vertex, num_vertices; //segment of vertices in VB to be processed
};

//----------------------------

struct S_SM_vertex_info{
   int joint_i;               //index of joint, to which this vertex belongs

                              //ratio, by which vertex belongs to this joint,
                              // rather than to its parent
   float weight;

   bool assigned;             //set to true as soon as vertex is assigned to joint

   word dest_index;           //for face-remapping after verts are re-organized

   S_SM_vertex_info():
      dest_index(0xffff),
      assigned(false)
   {}
};

//----------------------------

#ifdef _DEBUG
static void PaintEdge(PI3D_driver drv, const vector<S_vector> &verts, const I3D_edge &e){
   S_vector vv[3] = {verts[e[0]], verts[e[1]]};
   drv->DEBUG_CODE(0, (dword)vv);
}

static void PaintNormal(PI3D_driver drv, const S_vector &v, const S_vector &n){
   S_vector vv[3] = {v, v+n};
   drv->DEBUG_CODE(0, (dword)vv);
}
#endif

//----------------------------
// Collect all valid joint children from provided hierarchy tree.
// Also alloc joint segments and init matrices plus names.
static void PreprocessJoints(PI3D_visual root, vector<S_SM_joint_info> &joint_info,
   vector<S_SM_bone_segment> &bone_segments,
   const S_matrix &m_inv){

                              //collect all joints
   struct S_hlp{
      const S_matrix *m_inv;
      vector<S_SM_joint_info> *joint_info;
      static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

         switch(frm->GetType1()){
         case FRAME_JOINT:
            {
               PI3D_joint joint = I3DCAST_JOINT(frm);
               if(joint->IsBBoxValid()){
                  S_hlp *hp = (S_hlp*)c;
                  hp->joint_info->push_back(S_SM_joint_info());
                  S_SM_joint_info &ji = hp->joint_info->back();
                  ji.m_loc = joint->GetBBoxMatrix1() * joint->GetMatrix() * (*hp->m_inv);
                  ji.joint = joint;
#if 0                         //debug paint direction of bounding-box
                              // (not up-to-date!)
                  {
                     S_vector vv[2] = {joint->GetWorldPos(),
                        joint->GetWorldPos() + (ji.m_loc * root->GetMatrix())(2)
                     };
                     joint->GetDriver1()->DEBUG_CODE(0, (dword)vv);
                  }
#endif
               }
               frm->EnumFrames(cbEnum, c, ENUMF_ALL);
            }
            break;
         }
         return I3DENUMRET_SKIPCHILDREN;
      }
   } hlp;
   hlp.m_inv = &m_inv;
   hlp.joint_info = &joint_info;
   root->EnumFrames(S_hlp::cbEnum, (dword)&hlp, ENUMF_ALL);

                              //init bone segments
   bone_segments.clear();
   bone_segments.assign(joint_info.size());
   for(int i=joint_info.size(); i--; ){
      S_SM_joint_info &ji = joint_info[i];
      PI3D_joint joint = ji.joint;
                              //init segment
      S_SM_bone_segment &bs = bone_segments[i];
      
      bs.joint_name = joint->GetName1();
      bs.orig_mat_inv = ~(joint->GetMatrix() * m_inv);
   }
}

//----------------------------
// Assign unique vertices inside of joint's bounding boxes.
static void AssignUniqueInsideVertices(vector<S_SM_vertex_info> &vertex_info,
   const I3D_vertex_singlemesh *verts, const word *vertex_map,
   const vector<S_SM_joint_info> &joint_info, PI3D_frame root){

                              //process all unique vertices
   for(int i=vertex_info.size(); i--; ){
                              //if not unique, skip it
      if(vertex_map[i] != i)
         continue;

      const I3D_vertex_singlemesh &v = verts[i];
      S_SM_vertex_info &vi = vertex_info[i];

      vi.joint_i = -1;
      float closest_joint_dist_2 = 1e+16f;
      for(int j = joint_info.size(); j--; ){
         const S_SM_joint_info &ji = joint_info[j];
         PI3D_joint joint = ji.joint;

         assert(joint->IsBBoxValid());

         const S_matrix &m = ji.m_loc;
         float weight;
         if(IsPointInBBox(v.xyz, joint->GetBBox1(), m, weight)){
                              //check closest distance
            S_vector dir = v.xyz - ji.m_loc(3);
            float dist_2 = dir.Dot(dir);
            if(closest_joint_dist_2 > dist_2){
               closest_joint_dist_2 = dist_2;
               vi.joint_i = j;
               vi.weight = weight;

               vi.assigned = true;
            }
         }
#if defined DEBUG_PAINT_VERTEX_ASSIGNMENT || 0
         if(vi.assigned){
            S_vector vv[3] = {v.xyz * root->GetMatrix(),
               joint_info[vi.joint_i].joint->GetWorldPos() };
            vv[2].x = vi.weight;
            drv->DEBUG_CODE(10, (dword)vv);
         }
#endif                        //DEBUG_PAINT_VERTEX_ASSIGNMENT
      }
   }
}

//----------------------------
// Assign all vertices, which are not directly in one of joints' bounding boxes.
// This is done by traversing the edges and determining, which to assigned vertices
// given vertex is connected to.
static void AssignUnassignedVertices(vector<S_SM_vertex_info> &vertex_info,
   const I3D_vertex_singlemesh *verts, const word *vertex_map,
   const I3D_triface *faces, dword num_faces,
   const vector<S_SM_joint_info> &joint_info, PI3D_frame root){

   dword num_verts = vertex_info.size();
   int i;

   vector<vector<word> > vertex_face_use(num_verts);
   for(i = num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      vertex_face_use[vertex_map[fc[0]]].push_back(i);
      vertex_face_use[vertex_map[fc[1]]].push_back(i);
      vertex_face_use[vertex_map[fc[2]]].push_back(i);
   }

   struct S_hlp{
      static void SpreadVertex(word vertex_i, const I3D_triface *all_faces,
         vector<S_SM_vertex_info> &vertex_info, const word *v_map,
         const vector<S_SM_joint_info> &joint_info,
         const I3D_vertex_singlemesh *verts,
#ifdef DEBUG_PAINT_ASSIGN_WAVE
         PI3D_driver drv, const S_matrix &mat,
#endif
         const vector<vector<word> > &vertex_face_use, int &num_done){

         const vector<word> &face_use = vertex_face_use[vertex_i];
                           //process all faces accessing this
         for(int i=face_use.size(); i--; ){
            const I3D_triface &fc = all_faces[face_use[i]];
                           //this is our face, spread it further
            for(int k=3; k--; ){
               word vertex_i1 = fc[k];
                           //process only unique verts
               vertex_i1 = v_map[vertex_i1];
                           //don't process this vertex
               if(vertex_i1 == vertex_i)
                  continue;

               S_SM_vertex_info &vi = vertex_info[vertex_i1];
                           //don't process processed ones
               if(vi.assigned)
                  continue;
                           //set vertex to be like the processed one
               vi.joint_i = vertex_info[vertex_i].joint_i;
               const S_SM_joint_info &ji = joint_info[vi.joint_i];

#ifdef DEBUG_PAINT_ASSIGN_WAVE
               {
                  S_vector vv[2] = {verts[vertex_i].xyz*mat, verts[vertex_i1].xyz*mat};
                  drv->DEBUG_CODE(0, (dword)vv);
               }
#endif

               PI3D_joint joint = ji.joint;
               if(joint->IsBBoxValid()){
                  const S_matrix &m = ji.m_loc;
                  float d = verts[vertex_i1].xyz.Dot(m(2));
                  float d1 = m(3).Dot(m(2));
                  vi.weight = (d < d1) ?  0.0f : 1.0f;
               }else{
                           //shouldn't happen!
                  vi.weight = 0.0f;
               }
               vi.assigned = true;
               ++num_done;
                           //spread this vertex further
               SpreadVertex(vertex_i1, all_faces, vertex_info, v_map,
                  joint_info, verts,
#ifdef DEBUG_PAINT_ASSIGN_WAVE
                  drv, mat,
#endif
                  vertex_face_use, num_done);
            }
         }
      }
   };

   int num_done;
   do{
      num_done = 0;
      for(i = num_verts; i--; ){    
                              //process only unique vertices
         if(vertex_map[i] != i)
            continue;
         S_SM_vertex_info &vi = vertex_info[i];
         if(!vi.assigned){
            continue;
         }
                              //spread accross all faces
         S_hlp::SpreadVertex(i, faces, vertex_info, vertex_map,
            joint_info, verts,
#ifdef DEBUG_PAINT_ASSIGN_WAVE
            root->GetDriver1(), root->GetMatrix(),
#endif
            vertex_face_use, num_done);
      }
   }while(num_done);
}

//----------------------------
// Re-order vertices in array, so that each segment has contiguous vertex segment.
// Write segments into 'bone_segments'.
// Also assign joint and weight to yet unassigned vertices, which is actually
// wrong assignment, but for the rest of pipeline, each vertex must belong somewhere.
static void OrderVerticesAndRemapFaces(vector<S_SM_vertex_info> &vertex_info,
   vector<S_SM_joint_info> &joint_info,
   const I3D_vertex_singlemesh *verts, dword num_verts, const word *vertex_map,
   vector<I3D_vertex_singlemesh> &verts_out,
   I3D_triface *faces, dword num_faces){

   int i;

   verts_out.reserve(num_verts);

   for(i=joint_info.size(); i--; ){
      S_SM_joint_info &ji = joint_info[i];
                              //save vertex segment
      ji.first_vertex = verts_out.size();
      ji.num_vertices = 0;

                              //find index of parent joint
      PI3D_frame j_prnt = ji.joint->GetParent();
      int prnt_index = joint_info.size();
      if(j_prnt->GetType1() == FRAME_JOINT){
         for(prnt_index=0; prnt_index<joint_info.size(); prnt_index++){
            if(joint_info[prnt_index].joint == j_prnt)
               break;
         }
      }
                              //each matrix consumes 3 slots, so multiply by 3
      float f_matrix_index = (float)i * 3.0f;
      float f_prnt_matrix_index = (float)prnt_index * 3.0f;

                              //assign all verts of this bone/joint to this segment
      for(int j = num_verts; j--; ){
         word vertex_index = vertex_map[j];
         S_SM_vertex_info &vi = vertex_info[vertex_index];
                              //each vertex should be assigned now, except those
                              // not connected with rest of geometry, and not inside
                              // of any joint
         if(!vi.assigned){
                              //assign yet unassigned vertices
            vi.weight = 0.0f;
            vi.joint_i = 0;
            vertex_info[j].dest_index = 0xffff;
         }
                              //add the vertex to this segment
         if(vi.joint_i==i){
            assert(vertex_info[j].dest_index==0xffff);
            vertex_info[j].dest_index = verts_out.size();
            verts_out.push_back(verts[j]);
            I3D_vertex_singlemesh &v = verts_out.back();
            v.weight = vi.weight;
                              //generate matrix palette indicies
            //v.res = 0;
            v.indicies[0] = f_matrix_index;
            v.indicies[1] = f_prnt_matrix_index;
            ++ji.num_vertices;
         }
      }
   }

                              //remap faces to fit into new vertices
   for(i = num_faces; i--; ){
      I3D_triface &face = faces[i];
      for(int k=0; k<3; k++){
         if(vertex_info[face[k]].dest_index==0xffff)
            vertex_info[face[k]].dest_index = 0;
         face[k] = vertex_info[face[k]].dest_index;
      }
   }
}

//----------------------------

static void InitBoundingVolumes(I3D_bound_volume &bvol_root,
   vector<S_SM_bone_segment> &bone_segments,
   const vector<S_SM_joint_info> &joint_info,
   const I3D_vertex_singlemesh *verts, PI3D_frame root){

   int i;
   bvol_root.bbox.Invalidate();
                              //create bounding-boxes on segments
   for(i=bone_segments.size(); i--; ){
      const S_SM_joint_info &ji = joint_info[i];
      S_SM_bone_segment &bs = bone_segments[i];
      bs.bvol.bbox.Invalidate();
                              //expand by all our verts
      for(int j=ji.num_vertices; j--; ){
         const I3D_vertex_singlemesh &v = verts[ji.first_vertex + j];
                              //consider this vertex if it has at least something from us
         if(ji.joint->GetParent()==root && v.weight!=1.0f){
                              //put into root bound
            bvol_root.bbox.min.Minimal(v.xyz);
            bvol_root.bbox.max.Maximal(v.xyz);
         }
         if(v.weight==0.0f){
            continue;
         }
         bs.bvol.bbox.min.Minimal(v.xyz);
         bs.bvol.bbox.max.Maximal(v.xyz);
      }

                              //expand by vertices of all other segments, which connect to us
                              // or by parent's joint
      for(int k=bone_segments.size(); k--; ){
         const S_SM_joint_info &ji1 = joint_info[k];
         float extreme_weight = 1.0f;
         if(ji1.joint==ji.joint->GetParent()){
            continue;
                              //expand by parent's joint
            extreme_weight = 0.0f;
         }else{
            if(ji1.joint->GetParent()!=ji.joint)
               continue;
         }

         //const S_SM_bone_segment &bs1 = bone_segments[k];
         for(int j=ji1.num_vertices; j--; ){
            const I3D_vertex_singlemesh &v = verts[ji1.first_vertex + j];
                              //consider this vertex if it has at least something from us
            if(v.weight==extreme_weight)
               continue;
            bs.bvol.bbox.min.Minimal(v.xyz);
            bs.bvol.bbox.max.Maximal(v.xyz);
         }
      }

      for(j=0; j<3; j++){
         bs.bvol.bbox.min[j] -= SAFE_BOUND_THRESH;
         bs.bvol.bbox.max[j] += SAFE_BOUND_THRESH;
      }

                              //make bounding sphere
      S_vector bbox_half_diagonal = (bs.bvol.bbox.max - bs.bvol.bbox.min) * .5f;
      bs.bvol.bsphere.pos = bs.bvol.bbox.min + bbox_half_diagonal;
      bs.bvol.bsphere.radius = bbox_half_diagonal.Magnitude();
   }
                              //make bounding sphere
   S_vector bbox_half_diagonal = (bvol_root.bbox.max - bvol_root.bbox.min) * .5f;
   bvol_root.bsphere.pos = bvol_root.bbox.min + bbox_half_diagonal;
   bvol_root.bsphere.radius = bbox_half_diagonal.Magnitude();
}

//----------------------------

I3D_singlemesh::E_IBS_RESULT I3D_singlemesh::InitBoneSegments(PI3D_visual root){
   
   assert(NumVertices1() && NumFaces1());
   if(!NumVertices1())
      return IBS_NO_VERTICES;
   if(!NumFaces1())
      return IBS_NO_FACES;

                              //get root's inversed matrix, so that we operate
                              // in it's local coords
   const S_matrix &m_inv = root->GetInvMatrix1();

                              //info about all collected joints
   vector<S_SM_joint_info> joint_info;

                              //collect all joint children
   PreprocessJoints(root, joint_info, bone_segments, m_inv);
   if(!bone_segments.size())
      return IBS_NO_JOINTS;

                              //get access to vertices and faces
   I3D_triface *faces = index_buffer.Lock(0);
   dword num_faces = index_buffer.NumFaces1();
   const I3D_vertex_singlemesh *verts = (const I3D_vertex_singlemesh*)vertex_buffer.Lock(D3DLOCK_READONLY);
   dword num_verts = NumVertices1();

                              //determine vertices singularity
   vector<word> v_map(num_verts, 0);
   MakeVertexMapping(&verts[0].xyz, sizeof(I3D_vertex_singlemesh),
      num_verts, v_map.begin());


   vector<S_SM_vertex_info> vertex_info(num_verts);

   AssignUniqueInsideVertices(vertex_info, verts, v_map.begin(),
      joint_info, root);

   AssignUnassignedVertices(vertex_info, verts, v_map.begin(), faces, num_faces,
      joint_info, root);

                              //new vertex pool
   vector<I3D_vertex_singlemesh> verts_out;
   OrderVerticesAndRemapFaces(vertex_info, joint_info, verts, num_verts,
      v_map.begin(), verts_out, faces, num_faces);

   index_buffer.Unlock();
   faces = NULL;

   vertex_buffer.Unlock();
   verts = NULL;

   InitBoundingVolumes(bvol_root, bone_segments, joint_info, verts_out.begin(), root);

                              //modify vertex pool
   vertex_buffer.SetVertices(verts_out);

   return IBS_OK;
}

//----------------------------

#define DBASE_VERSION 0x001a

bool I3D_singlemesh::SaveCachedInfo(C_cache *ck, word) const{

                              //write core
   if(!((I3D_singlemesh*)this)->I3D_mesh_base::SaveCachedInfo(ck, DBASE_VERSION))
      return false;
   {                          //write bone segments
      word num = bone_segments.size();
      ck->write((char*)&num, sizeof(num));
      for(int i=0; i<num; i++){
         const S_SM_bone_segment &bs = bone_segments[i];
                              //write joint's name
         ck->write((const char*)bs.joint_name, bs.joint_name.Size() + 1);
                              //other info
         ck->write((const char*)&bs.orig_mat_inv, sizeof(bs.orig_mat_inv));
         ck->write((const char*)&bs.bvol, sizeof(bs.bvol));
      }
      ck->write((const char*)&bvol_root, sizeof(bvol_root));
   }
   return true;
}

//----------------------------

bool I3D_singlemesh::LoadCachedInfo(C_cache *ck, S_load_context &lc, word){

                              //read core
   if(!I3D_mesh_base::LoadCachedInfo(ck, lc, DBASE_VERSION))
      return false;
   {                          //read bone segments
      word num;
      ck->read((char*)&num, sizeof(num));
      bone_segments.assign(num);
      for(int i=0; i<num; i++){
         S_SM_bone_segment &bs = bone_segments[i];
                              //read joint's name
         char name[256];
         int j = 0;
         do{
            ck->read(&name[j++], sizeof(char));
         }while(name[j-1]);
         bs.joint_name = name;
                              //other info
         ck->read((char*)&bs.orig_mat_inv, sizeof(bs.orig_mat_inv));
         ck->read((char*)&bs.bvol, sizeof(bs.bvol));
      }
      ck->read((char*)&bvol_root, sizeof(bvol_root));
   }

   return true;
}

//----------------------------

I3D_RESULT I3D_object_singlemesh::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);

   PI3D_visual vis = I3DCAST_VISUAL(frm);
   I3D_RESULT ir = I3D_visual::Duplicate(vis);
   switch(vis->GetVisualType()){
   default:
      {
         mesh = NULL;
         I3D_mesh_base *mb = vis->GetMesh();
         if(mb){
            mesh = new I3D_singlemesh(drv);
            mesh->Clone(mb);
            mesh->Release();
         }
      }
      break;
   case I3D_VISUAL_SINGLEMESH:
      {
         PI3D_object_singlemesh sm = (PI3D_object_singlemesh)vis;
         SetMesh(sm->GetSingleMesh());
         smesh_flags = sm->smesh_flags;
         volume_box_matrix = sm->volume_box_matrix;
         volume_box = sm->volume_box;
      }
      break;
   }

   return ir;
}

//----------------------------

I3D_RESULT I3D_object_singlemesh::SetMesh(PI3D_singlemesh m1){

   mesh = m1;
                              //init bone segments
   bone_segments.clear();
   if(mesh){
      bone_segments.reserve(mesh->GetBoneSegments().size());
      for(int i=0; i<mesh->GetBoneSegments().size(); i++){
         PI3D_joint joint = I3DCAST_JOINT(FindChildFrame(mesh->GetBoneSegments()[i].joint_name,
            ENUMF_ALL | ENUMF_MODEL_ORIG));
         assert(joint);
         bone_segments.push_back(S_SM_object_bone_segment());
         S_SM_object_bone_segment &bs = bone_segments.back();
         bs.joint = joint;
      }
   }
                              //destroy old vertex buffers (if any)
   vertex_buffer.DestroyD3DVB(drv);

   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
   frm_flags &= ~(FRMFLAGS_HR_BOUND_VALID | FRMFLAGS_SM_BOUND_VALID);
   return I3D_OK;
}

//----------------------------

bool I3D_object_singlemesh::ComputeBounds(){

   if(frm_flags&FRMFLAGS_SM_BOUND_VALID)
      return true;

   if(!mesh){
                              //invalidate bounding volume
      bound.bound_local.bbox.Invalidate();
      bound.bound_local.bsphere.pos.Zero();
      bound.bound_local.bsphere.radius = 0.0f;
   }else{
      I3D_bound_volume &bvol = bound.bound_local;

      const S_matrix &m_inv = GetInvMatrix1();

      dword num_segments = bone_segments.size();
      I3D_bsphere *bs_list = (I3D_bsphere*)alloca(sizeof(I3D_bsphere)*(num_segments+1));
      dword num_bspheres = 0;

      bvol.bbox = mesh->GetRootBounds().bbox;
      bvol.bsphere = mesh->GetRootBounds().bsphere;
      
      bs_list[num_bspheres++] = mesh->GetRootBounds().bsphere;

      if(!(frm_flags&FRMFLAGS_SM_BOUND_VALID))
         PrepareBlendMatrices();
                              //expand bbox by all joints' bboxes
      for(int i=num_segments; i--; ){
         const S_SM_bone_segment &bs = mesh->GetBoneSegments()[i];

         if(!bs.bvol.bbox.IsValid())
            continue;

         const S_SM_object_bone_segment &obs = bone_segments[i];

         const S_matrix &m_joint = obs.mat;
         S_matrix m_joint_inv = m_joint * m_inv;

                              //transform bound-box to SM's local coords
         for(int j=8; j--; ){
#ifndef USE_D3DX
            S_vector v = S_vector(bs.bvol.bbox[j&1].x, bs.bvol.bbox[(j&2)/2].y, bs.bvol.bbox[(j&4)/4].z) * m_joint_inv;
#else
            S_vectorw v = S_vector(bs.bvol.bbox[j&1].x, bs.bvol.bbox[(j&2)/2].y, bs.bvol.bbox[(j&4)/4].z);
            D3DXVec3Transform((D3DXVECTOR4*)&v, (D3DXVECTOR3*)&v, (D3DXMATRIX*)&m_joint_inv);
#endif
            bvol.bbox.min.Minimal(v);
            bvol.bbox.max.Maximal(v);
         }
         
         {
            I3D_bsphere &bs1 = bs_list[num_bspheres++];
                                 //transform bound-sphere
            float scale = sqrt(m_joint_inv(0, 0)*m_joint_inv(0, 0) + m_joint_inv(1, 0)*m_joint_inv(1, 0) + m_joint_inv(2, 0)*m_joint_inv(2, 0));
            bs1.radius = bs.bvol.bsphere.radius * scale;
            bs1.pos = bs.bvol.bsphere.pos * m_joint_inv;
         }
      }
                              //compute bounding sphere
      S_vector bbox_half_diagonal = (bvol.bbox.max - bvol.bbox.min) * .5f;
      bvol.bsphere.pos = bvol.bbox.min + bbox_half_diagonal;
      bvol.bsphere.radius = bbox_half_diagonal.Magnitude();

                              //don't make bsphere greater than sum of bound spheres
      float max_dist = 0.0f;
      for(i=num_bspheres; i--; ){
         float dist = (bvol.bsphere.pos - bs_list[i].pos).Magnitude() + bs_list[i].radius;
         max_dist = Max(max_dist, dist);
      }
      bvol.bsphere.radius = Min(bvol.bsphere.radius, max_dist);

                              //don't let bound-box be greater than bound-sphere
      for(i=0; i<3; i++){
         bvol.bbox.min[i] = Max(bvol.bbox.min[i], bvol.bsphere.pos[i] - bvol.bsphere.radius);
         bvol.bbox.max[i] = Min(bvol.bbox.max[i], bvol.bsphere.pos[i] + bvol.bsphere.radius);
      }

   }
   frm_flags &= ~FRMFLAGS_BSPHERE_TRANS_VALID;
   return true;
}

//----------------------------

static const word sides_i[] = {
   1, 3, 7,  7, 5, 1,
   4, 6, 2,  2, 0, 4,
   2, 6, 7,  7, 3, 2,
   4, 0, 1,  1, 5, 4,

   0, 2, 3,  3, 1, 0,
   7, 6, 5,  6, 4, 5,
};

//----------------------------

I3D_RESULT I3D_object_singlemesh::DebugDraw(PI3D_scene scene) const{

   if(!mesh)
      return I3DERR_INVALIDPARAMS;
   int i;

   static const S_vector color(1.0f, .7f, 0.0f);
   byte alpha = 0xe0;
   S_vector bbox_full[8];
                              //paint root
   scene->SetRenderMatrix(matrix);
   mesh->GetRootBounds().bbox.Expand(bbox_full);
   scene->DebugDrawBBox(bbox_full, color, alpha);
   {
      S_matrix m = matrix;
      m(3) += mesh->GetRootBounds().bsphere.pos.RotateByMatrix(m);
      scene->DebugDrawSphere(m, mesh->GetRootBounds().bsphere.radius, color, alpha);
   }

   for(i=mesh->GetBoneSegments().size(); i--; ){
      const S_SM_bone_segment &bs = mesh->GetBoneSegments()[i];
      const S_SM_object_bone_segment &obs = bone_segments[i];

      scene->SetRenderMatrix(obs.mat);

      bs.bvol.bbox.Expand(bbox_full);
      scene->DebugDrawBBox(bbox_full, color, alpha);

      {
         S_matrix m = obs.mat;
         m(3) += bs.bvol.bsphere.pos.RotateByMatrix(m);
         scene->DebugDrawSphere(m, bs.bvol.bsphere.radius, color, alpha);
      }
   }

   if(drv->GetFlags2()&DRVF2_DRAWVOLUMES)
   if(IsVolumeBoxValid()){
      GetVolumeBox1().Expand(bbox_full);
      scene->SetRenderMatrix(GetVolumeBoxMatrix1() * GetMatrix());

      const S_vector &color = //strong ?
         S_vector(1.0, 0.0, 0.0);
                              //draw sides
      scene->DrawTriangles(bbox_full, 8, sides_i, sizeof(sides_i)/sizeof(word),
         color, sizeof(S_vector),
         0x80);

      byte alpha = 0x80;
                              //draw edges
      scene->DebugDrawBBox(bbox_full, color, alpha);
   }
   return I3D_OK;
}

//----------------------------

void I3D_object_singlemesh::PrepareBlendMatrices(){

                              //process all bone segments - prepare matrices
   const vector<S_SM_bone_segment> &bsgs1 = mesh->GetBoneSegments();
   const S_SM_bone_segment *mesh_bs_ptr = bsgs1.begin();
   S_SM_object_bone_segment *sm_bs_ptr = bone_segments.begin();

   for(int i=bsgs1.size(); i--; ){
#ifdef USE_PREFETCH
      S_SM_object_bone_segment *obs_next;
      if(i){
         obs_next = &sm_bs_ptr[i-1];
         Prefetch1(obs_next);
      }
#endif
      const S_SM_bone_segment &bs = mesh_bs_ptr[i];
      S_SM_object_bone_segment &obs = sm_bs_ptr[i];
      PI3D_joint joint = obs.joint;
                              //resulting transformation matrix:
                              // - rotate by joint's matrix (m_joint and m_joint_prnt)
                              // - translation to joint's current world position (m_joint(3))
      const S_matrix &m_joint = joint->GetMatrix();
      S_matrix &m = obs.mat;
      m = bs.orig_mat_inv % m_joint;
      m(3) += m_joint(3);
#ifdef USE_PREFETCH
      if(i){
         Prefetch1(obs_next->joint);
      }
#endif
      S_matrix mt = m;
      mt.Transpose();
      drv->GetDevice1()->SetVertexShaderConstant(I3D_driver::VSC_MAT_BLEND_BASE + i*3,
         &mt, 3);
   }
                              //basic matrix
   S_matrix mt = GetMatrix();
   mt.Transpose();
   drv->GetDevice1()->SetVertexShaderConstant(I3D_driver::VSC_MAT_BLEND_BASE + bone_segments.size()*3,
      &mt, 3);

   frm_flags |= FRMFLAGS_SM_BOUND_VALID;
}

//----------------------------

static const dword vs_decl[] = {
   D3DVSD_STREAM(0),
   D3DVSD_REG(0, D3DVSDT_FLOAT3),   //position
   D3DVSD_REG(3, D3DVSDT_FLOAT3),   //weight (1) + indicies (2)
   D3DVSD_REG(2, D3DVSDT_FLOAT3),   //normal
   D3DVSD_REG(1, D3DVSDT_FLOAT2),   //UV coordinates
   D3DVSD_END()
};

//----------------------------

void I3D_object_singlemesh::AddPrimitives(S_preprocess_context &pc){

   if(!mesh)
      return;
   AddPrimitives1(mesh, pc);
}

//----------------------------

void I3D_object_singlemesh::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   drv->SetClipping(rp.clip);

   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      dest_vb_trans_id = pc.render_view_id;

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }

      dword vertex_count;
      const vector<C_auto_lod> &auto_lods = mesh->GetAutoLODs();
      if(rp.curr_auto_lod < 0){
         vertex_count = mesh->vertex_buffer.NumVertices();
      }else{
         vertex_count = auto_lods[rp.curr_auto_lod].vertex_count;
      }
                              //transform & light vertices
      LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();

      dword prep_flags = VSPREP_WORLD_SPACE | VSPREP_FEED_MATRIX;

      if(drv->IsDirectTransform()){
         if(pc.mode!=RV_SHADOW_CASTER)
            prep_flags |= VSPREP_MAKELIGHTING;
         prep_flags |= VSPREP_COPY_UV;
      }else{
         if(pc.mode!=RV_SHADOW_CASTER){
#ifdef DEBUG_NO_CACHE_LIGHT
            vis_flags &= ~VISF_DEST_LIGHT_VALID;  
#endif
            if(!(vis_flags&VISF_DEST_LIGHT_VALID)){
               prep_flags |= VSPREP_MAKELIGHTING;
               vis_flags |= VISF_DEST_LIGHT_VALID;
            }
         }
         if(!(vis_flags&VISF_DEST_UV0_VALID)){
            prep_flags |= VSPREP_COPY_UV;
            vis_flags |= VISF_DEST_UV0_VALID;
         }
      }
      if(pc.mode==RV_SHADOW_CASTER)
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;

      PrepareBlendMatrices();

      I3D_driver::S_vs_shader_entry_in se(vs_decl);
      se.AddFragment(I3D_driver::VSF_M_PALETTE_TRANSFORM);
      PrepareVertexShader(mesh->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      HRESULT hr;
                              //fill-in transposed view-projection matrix
      hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0,
         &pc.scene->GetViewProjHomMatTransposed(), 4);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

      if(!drv->IsDirectTransform()){
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mesh->vertex_buffer.GetD3DVertexBuffer(), sizeof(I3D_vertex_singlemesh));
         hr = lpDev8->ProcessVertices(mesh->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }
   DrawPrimitive1(mesh, pc, rp);
}

//----------------------------
//----------------------------


