#ifndef __SNG_MESH_H
#define __SNG_MESH_H

#include "mesh.h"
#include "joint.h"
#include "visual.h"

//----------------------------

                              //use 3 weights, actually only one is used,
                              // others 2 are indicies into 2 matrices
#define FVF_VERTEX_SINGLEMESH (D3DFVF_XYZB3 | D3DFVF_NORMAL | D3DFVF_TEX1)

//----------------------------

struct I3D_vertex_singlemesh{
   S_vector xyz;
   float weight;
   float indicies[2];
   S_vector normal;
   I3D_text_coor tex[1];
};

//----------------------------
                              //bone segment, lists vertices affected by single bone segment
                              // in I3D_singlemesh
                              //I3D_singlemesh contains one or more such segments
                              //Note: joint is referenced by name
struct S_SM_bone_segment{
   C_string joint_name;
   S_matrix orig_mat_inv;     //joint's original matrix, in mesh's local coords, inverted
   I3D_bound_volume bvol;

   S_SM_bone_segment(){}
   S_SM_bone_segment(const S_SM_bone_segment &ref){ operator =(ref); }
   S_SM_bone_segment &operator =(const S_SM_bone_segment &bs){
      joint_name = bs.joint_name;
      orig_mat_inv = bs.orig_mat_inv;
      bvol = bs.bvol;
      return (*this);
   }
};

//----------------------------
                              //single-mesh mesh
class I3D_singlemesh: public I3D_mesh_base{

   vector<S_SM_bone_segment> bone_segments;
   I3D_bound_volume bvol_root;//bounding volume of vertices affected by root matrix

   friend class I3D_driver;   //creation

public:
   I3D_singlemesh(PI3D_driver);
   ~I3D_singlemesh();

   enum E_IBS_RESULT{
      IBS_OK,
      IBS_NO_VERTICES,
      IBS_NO_FACES,
      IBS_NO_JOINTS
   };
//----------------------------
// Initialize bone segments from provided hierarchy, where joints are scanned.
   E_IBS_RESULT InitBoneSegments(PI3D_visual root);

   inline const vector<S_SM_bone_segment> &GetBoneSegments() const{ return bone_segments; }

   inline const I3D_bound_volume &GetRootBounds() const{ return bvol_root; }

//----------------------------
// Save/Load mesh - vertices and faces after optimization, vertex normals,
// bone segments.
   virtual bool SaveCachedInfo(class C_cache*, word = 0) const;
   virtual bool LoadCachedInfo(class C_cache*, struct S_load_context &lc, word = 0);

public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

typedef I3D_singlemesh *PI3D_singlemesh;
typedef const I3D_singlemesh *CPI3D_singlemesh;

//----------------------------
                              //bone segment, contains joint (the other joint is the joint's
                              // parent)
                              //also contains matrices of the joints - very temporary
struct S_SM_object_bone_segment{
   C_smart_ptr<I3D_joint> joint;

                              //joint's matrix, positioned at joint's world pos,
                              // rotated to trasnform vertices vertices
   S_matrix mat;              

   S_SM_object_bone_segment(){
      mat.Identity();
   }
   S_SM_object_bone_segment(const S_SM_object_bone_segment &ref){ operator =(ref); }
   S_SM_object_bone_segment &operator =(const S_SM_object_bone_segment &ref){
      joint = ref.joint;
      mat = ref.mat;
      return (*this);
   }
};

//----------------------------

class I3D_object_singlemesh: public I3D_visual{
#define SMESH_FLAGS_VBOX_VALID    1       //volume-box is valid

   dword smesh_flags;

   S_matrix volume_box_matrix;
   I3D_bbox volume_box;      //volume-box - used for collision detection

   C_smart_ptr<I3D_singlemesh> mesh;
   vector<S_SM_object_bone_segment> bone_segments;

   void PrepareBlendMatrices();

public:
   I3D_object_singlemesh(PI3D_driver d): 
      I3D_visual(d),
      smesh_flags(0)
   {
      visual_type = I3D_VISUAL_SINGLEMESH;
   }
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(DebugDraw)(PI3D_scene) const;
   I3DMETHOD_(I3D_mesh_base*,GetMesh)() const{ return mesh; }

   void SetVolumeBox(const I3D_bbox &bb1, const S_matrix &m1){
      volume_box = bb1;
      volume_box_matrix = m1;
      smesh_flags |= SMESH_FLAGS_VBOX_VALID;
   }
   inline bool IsVolumeBoxValid() const{ return (smesh_flags&SMESH_FLAGS_VBOX_VALID); }
   inline const I3D_bbox &GetVolumeBox1() const{ return volume_box; }
   inline const S_matrix &GetVolumeBoxMatrix1() const{ return volume_box_matrix; }

   I3D_RESULT SetMesh(PI3D_singlemesh);
   inline PI3D_singlemesh GetSingleMesh() const{ return mesh; }

   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   virtual bool ComputeBounds();

public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
};

typedef I3D_object_singlemesh *PI3D_object_singlemesh;
typedef const I3D_object_singlemesh *CPI3D_object_singlemesh;

//----------------------------

#endif//__SNG_MESH_H