/*--------------------------------------------------------
   Copyright (c) 1999, 2000 Michal Bacik. 
   All rights reserved.

   File: Sound.cpp
   Content: Sound frame.
--------------------------------------------------------*/

#include "all.h"
#include "soundi3d.h"
#include "scene.h"
#include "camera.h"
#include <isound2.h>
#include <mmreg.h>            //WAVEFORMAT
#include <dsound.h>


//#define FORCE_EMULATION
//#define FORCE_SOFTWARE

#define EMULATE_ATTENUATION   //compute attenuation (by setting volume) ourselves
//#define EMULATE_3DPOSITIONING //compute 3D positioning (by setting panning) ourselves


#ifdef _DEBUG

#define SND_LOG(n)
#define SND_PRINT(msg)
//#define SND_LOG(msg) drv->DEBUG(msg)
//#define SND_PRINT(msg) drv->PRINT(msg)

#else

#define SND_LOG(n)
#define SND_PRINT(msg)

#endif


//----------------------------
//----------------------------

void S_sound_env_properties::SetupDefaults(){

                              //setup sound data defaults (EAX defaults)
   memset(this, 0, sizeof(S_sound_env_properties));
#ifdef USE_EAX
   room_level_lf = (float)(EAXLISTENER_DEFAULTROOM-EAXLISTENER_MINROOM) / (float)(EAXLISTENER_MAXROOM-EAXLISTENER_MINROOM);
   room_level_hf = (float)(EAXLISTENER_DEFAULTROOMHF-EAXLISTENER_MINROOMHF) / (float)(EAXLISTENER_MAXROOMHF-EAXLISTENER_MINROOMHF);
   decay_time = EAXLISTENER_DEFAULTDECAYTIME * 1000.0f;
   decay_hf_ratio = (EAXLISTENER_DEFAULTDECAYHFRATIO-EAXLISTENER_MINDECAYHFRATIO) / (EAXLISTENER_MAXDECAYHFRATIO-EAXLISTENER_MINDECAYHFRATIO);
   reflections = (float)(EAXLISTENER_DEFAULTREFLECTIONS-EAXLISTENER_MINREFLECTIONS) / (float)(EAXLISTENER_MAXREFLECTIONS-EAXLISTENER_MINREFLECTIONS);
   reflections_delay = EAXLISTENER_DEFAULTREFLECTIONSDELAY * 1000.0f;
   reverb = (float)(EAXLISTENER_DEFAULTREVERB-EAXLISTENER_MINREVERB) / (float)(EAXLISTENER_MAXREVERB-EAXLISTENER_MINREVERB);
   reverb_delay = EAXLISTENER_DEFAULTREVERBDELAY * 1000.0f;
   env_diffusion = EAXLISTENER_DEFAULTENVIRONMENTDIFFUSION;
#endif                        //USE_EAX
   name = NULL;
   emulation_volume = 1.0f;
}

//----------------------------

void S_sound_env_properties::MakeBlend(const S_sound_env_properties &ssp1, const S_sound_env_properties &ssp2,
   float b2, bool use_eax){

   float b1 = (1.0f - b2);

   if(use_eax){
      room_level_lf = b1 * ssp1.room_level_lf + b2 * ssp2.room_level_lf;
      room_level_hf = b1 * ssp1.room_level_hf + b2 * ssp2.room_level_hf;
      decay_time = b1 * ssp1.decay_time + b2 * ssp2.decay_time;
      decay_hf_ratio = b1 * ssp1.decay_hf_ratio + b2 * ssp2.decay_hf_ratio;
      reflections = b1 * ssp1.reflections + b2 * ssp2.reflections;
      reflections_delay = b1 * ssp1.reflections_delay + b2 * ssp2.reflections_delay;
      reverb = b1 * ssp1.reverb + b2 * ssp2.reverb;
      reverb_delay = b1 * ssp1.reverb_delay + b2 * ssp2.reverb_delay;
      env_diffusion = b1 * ssp1.env_diffusion + b2 * ssp2.env_diffusion;
   }else{
      emulation_volume = b1 * ssp1.emulation_volume + b2 * ssp2.emulation_volume;
   }
}

//----------------------------
//----------------------------
                              //sounds
I3D_sound::I3D_sound(PI3D_scene s):
   snd_src(NULL),
   voice(NULL),
   owning_scene(s),
   I3D_frame(s->GetDriver()),
                              //set D3D defaults
   volume_out(0.0f),
   cone_out(PI*.5f), cone_in(0.0f), cos_cone_in(1.0f), cos_cone_out(.7071f),
   snd_type(I3DSOUND_NULL),
   frequency(1.0f),
   virtual_volume(1.0f),
   state(STOP),
   snd_flags(SNDF_UPDATEPOS | SNDF_UPDATEDIR | SNDF_UPDATEDIST | SNDF_UPDATESCALE),
   fade_time(0),
   fade_count_down(0),
   volume(1.0f)
{
   drv->AddCount(I3D_CLID_SOUND);
   type = FRAME_SOUND;
   enum_mask = ENUMF_SOUND;
   SetRange(1.0f, 10.0f, 0.0f);
}

//----------------------------

I3D_sound::~I3D_sound(){

   Close();
                              //unregister from driver list
   for(int i=drv->sounds.size(); i--; ){
      if(drv->sounds[i]==this){
         drv->sounds[i]=drv->sounds.back(); drv->sounds.pop_back();
         break;
      }
   }
   assert(i!=-1);
   while(sound_sectors.size())
      sound_sectors[0]->DeleteSound(this);
   drv->DecCount(I3D_CLID_SOUND);

   if(owning_scene){
      i = FindPointerInArray((void**)owning_scene->sounds.begin(), owning_scene->sounds.size(), this);
      if(i!=-1){
         owning_scene->sounds[i] = owning_scene->sounds.back();
         owning_scene->sounds.pop_back();
      }
   }
}

//----------------------------

I3D_RESULT I3D_sound::Duplicate(CPI3D_frame frm){

   if(type==frm->GetType1()){
      PI3D_sound s1 = I3DCAST_SOUND(frm);
      Close();

      SetSoundType(s1->snd_type);
      SetVolume(s1->volume);
      SetOutVol(s1->volume_out);
      SetCone(s1->cone_in, s1->cone_out);
      SetRange(s1->range_n, s1->range_f, s1->ambient_factor);
      SetLoop(s1->IsLoop());
      SetStreaming(s1->IsStreaming());
      Open(s1->GetFileName(), s1->open_flags);
   }

   return I3D_frame::Duplicate(frm);
}

//----------------------------

I3D_RESULT I3D_sound::SetSoundType(I3D_SOUNDTYPE st){

   if(st>I3DSOUND_POINTAMBIENT)
      return I3DERR_INVALIDPARAMS;

   FreeVoice();
   snd_type = st;
   snd_flags |= SNDF_UPDATEOUTVOL;     //we may simulate DSound point sounds by setting out vol=1.0

   return ISND_OK;
}

//----------------------------

I3D_RESULT I3D_sound::Open(const char *fname, dword flags, I3D_LOAD_CB_PROC *cb_proc,
   void *context, void *special){

   Close();

   file_name = fname;
                              //make filename lower-case (for comparision)
   file_name.ToLower();
   open_flags = flags;
   frequency = 1.0f;

   int i;
                              //try to reuse existing sample
   for(i=drv->sounds.size(); i--; ){
      PI3D_sound snd = drv->sounds[i];
                              //try to duplicate sound
      if(snd->file_name==fname && snd->snd_src){
         snd_src = snd->snd_src;
         if(snd_src) snd_src->AddRef();
         if(IsOn()) SetOn(true);
         return I3D_OK;
      }
   }

   PISND_driver lpis = drv->GetSoundInterface();
   if(!lpis)
      return I3DERR_NOTINITIALIZED;
                              //create sound source
   lpis->CreateSource(&snd_src);

   ISND_RESULT ir;
                              //look to all dirs
   const vector<C_string> &sounds_dirs = drv->GetDirs(I3DDIR_SOUNDS);
   for(i=0; i<sounds_dirs.size(); i++){
      C_string str;
      str = sounds_dirs[i].Size() ? C_fstring("%s\\%s", (const char*)sounds_dirs[i], fname) :
         fname;

      dword src_open_flags = ISND_PRELOAD;
      if(snd_flags&SNDF_STREAMING)
         src_open_flags = ISND_STREAMING;
      ir = snd_src->Open(str, src_open_flags);
      if(ISND_SUCCESS(ir)){
         SND_PRINT(C_fstring("Opened source from location '%s'.", (const char*)str));
         if(IsOn())
            SetOn(true);
         return I3D_OK;
      }
      if(ir==ISNDERR_OUTOFMEM)
         goto fail;
   }
   ir = ISNDERR_NOFILE;
fail:
   SND_PRINT(C_fstring("Failed to open source '%s'.", fname));
   snd_src->Release();
   snd_src = NULL;
                              //open error
   switch(ir){
   case ISNDERR_OUTOFMEM: return I3DERR_OUTOFMEM;
   case ISNDERR_NOFILE: return I3DERR_NOFILE;
   default: return I3DERR_GENERIC;
   }
}

//----------------------------

void I3D_sound::Close(){

   if(IsPlaying()){
      if(voice)
         voice->Stop();
      state = STOP;
   }

   if(snd_src){
      snd_src->Release();
      snd_src = NULL;
   }
   file_name = NULL;

   FreeVoice();
   state = STOP;
}

//----------------------------

void I3D_sound::SetOn(dword flg){

   I3D_frame::SetOn(flg);

   if(flg){
      if(voice && IsPlaying())
         voice->Stop();
                              //use real time, it may be long since we last rendered
      int timer = drv->GetGraphInterface()->ReadTimer();
      if(voice){
         voice->SetCurrTime(pause_pos = 0);
         pause_time = timer;
      }else{
         pause_pos = 0;
         pause_time = timer;
      }
      state = PAUSE;
   }else{
      if(IsPlaying()){
         if(voice)
            voice->Stop();
         state = STOP;
      }
   }
}

//----------------------------

I3D_RESULT I3D_sound::SetLoop(bool b){

   if(b) snd_flags |= SNDF_LOOPED;
   else snd_flags &= ~SNDF_LOOPED;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sound::SetStreaming(bool b){

   snd_flags &= ~SNDF_STREAMING;
                              //possible crashes and other problems with streaming
                              // don't do this now
   //if(b) snd_flags |= SNDF_STREAMING;
   return I3D_OK;
}

//----------------------------

bool I3D_sound::IsPlaying() const{

   if(!snd_src) return false;
                              //playing means PLAY or PAUSE (PAUSE is internal)
   return (state!=STOP);
}

//----------------------------

dword I3D_sound::GetPlayTime() const{

   if(!snd_src)
      return 0;
   const S_wave_format &wf = *snd_src->GetFormat();
   return wf.size * 1000 / (wf.samples_per_sec*wf.bytes_per_sample);
}

//----------------------------

dword I3D_sound::GetCurrTime() const{

   if(!snd_src) return 0;

   if(state==PAUSE){
                              //use real time, it may be long since we last rendered
      int timer = drv->GetGraphInterface()->ReadTimer();
      int ret = pause_pos + timer - pause_time;
      int len = GetPlayTime();
      if(ret >= len)          //check overflow
      if(snd_flags&SNDF_LOOPED) ret %= len;  //loop
      else ret = len;         //clamp
      return ret;
   }
   if(!voice) return 0;
   return voice->GetCurrTime();
}

//----------------------------

I3D_RESULT I3D_sound::SetCurrTime(dword pos){

   if(!snd_src) return I3DERR_NOTINITIALIZED;

   switch(state){
   case PAUSE:
      {
                              //use real time, it may be long since we last rendered   
         int timer = drv->GetGraphInterface()->ReadTimer();
         int virt = pause_pos + timer - pause_time;
         int delta = pos - virt;
         pause_time -= delta;
      }
      break;
   case PLAY:
      assert(voice);
      voice->SetCurrTime(pos);
      break;
   }
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sound::SetFrequency(float fr){

   if(!(open_flags&I3DLOAD_SOUND_FREQUENCY)){
      return I3DERR_UNSUPPORTED;
   }
   frequency = fr;
   snd_flags |= SNDF_UPDATEFREQ;

   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sound::SetRange(float min, float max, float af){

   range_n = min;
   range_f = max;
   ambient_factor = af;

   snd_flags |= SNDF_UPDATESCALE;
   snd_flags |= SNDF_UPDATEDIST;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sound::SetCone(float ia1, float oa1){

   cone_in = ia1;
   cone_out = oa1;
   cos_cone_in = cos(cone_in*.5f);
   cos_cone_out = cos(cone_out*.5f);
   snd_flags |= SNDF_UPDATECONE;
   return I3D_OK;
}

//----------------------------

I3D_RESULT I3D_sound::SetOutVol(float v){

   volume_out = v;
   snd_flags |= SNDF_UPDATEOUTVOL;
   return I3D_OK;
}

//----------------------------

void I3D_sound::Update(){

   if(frm_flags&FRMFLAGS_UPDATE_NEEDED){
      snd_flags |= SNDF_UPDATEPOS | SNDF_UPDATEDIR | SNDF_UPDATECONE | SNDF_UPDATESCALE | SNDF_UPDATEDIST;
   }
   I3D_frame::Update();
}

//----------------------------

void I3D_sound::MakeVoice(PI3D_camera cam){

   dword flags = open_flags;
   bool use_3d_hw = true;
#ifdef FORCE_EMULATION
   use_3d_hw = false;
#endif

   dword isnd_open_flags = 0;

   PISND_driver lpis = drv->GetSoundInterface();

   dword caps[6];
   lpis->GetCaps(caps);

                              //check if we don't have any 3D hardware,
                              // go with emulation
   if(!caps[ISNDCAPS_NUM_3D_ALL_BUFFERS])
      use_3d_hw = false;

                              //use hardware buffer, if available
#ifndef FORCE_SOFTWARE
   bool use_hw = caps[ISNDCAPS_NUM_HW_ALL_BUFFERS];
#else
   bool use_hw = false;
   isnd_open_flags |= ISND_SYSMEM;
#endif

                              //streaming sounds always in software
   /*
   if(snd_flags&SNDF_STREAMING){
      use_3d_hw = false;
      use_hw = false;
      isnd_open_flags |= ISND_SYSMEM;
   }
   */

   switch(snd_type){
   case I3DSOUND_AMBIENT:
   case I3DSOUND_VOLUME:
                              //not need for 3D sound
      use_3d_hw = false;
      break;
   case I3DSOUND_POINTAMBIENT:
      isnd_open_flags |= ISND_SYSMEM;
      use_hw = false;
      use_3d_hw = false;
      break;
   }

   if(flags&I3DLOAD_SOUND_SOFTWARE){
      isnd_open_flags |= ISND_SYSMEM;
      use_hw = false;
      use_3d_hw = false;
   }

   if(flags&I3DLOAD_SOUND_FREQUENCY){
                              //frequency-adjustable sounds, deal with them special way -
                              // force software mixing
      isnd_open_flags |= ISND_FREQUENCY | ISND_SYSMEM;
      use_hw = false;
      use_3d_hw = false;
   }

   const S_wave_format *wf = snd_src->GetFormat();
   if(wf->num_channels > 1){
                              //check if ambient
      if(snd_type!=I3DSOUND_AMBIENT && snd_type!=I3DSOUND_POINTAMBIENT)
         return;
                              //stereo sounds can't use 3D acceleration
      use_3d_hw = false;
   }

   if(use_hw && use_3d_hw){
      isnd_open_flags |= ISND_3DBUFFER;
                              //no panning on real 3D buffers
      isnd_open_flags |= ISND_NOPANNING;
   }

   snd_flags &= ~SNDF_EMULATED;


                              //create sound voice
   ISND_RESULT ir;
   ir = lpis->Create3DSound(&voice);
   if(ISND_FAIL(ir))
      return;

   if(use_hw){                 //locate on card
      isnd_open_flags |= ISND_CARDMEM;
   }

                              //try to open until we succeed
   while(true){
      ir = voice->Open(snd_src, isnd_open_flags);

      if(ISND_SUCCESS(ir)){
         SND_PRINT(C_fstring("Alloc voice ok."));
         if(!use_3d_hw)       //emulate if not a 3D sound buffer
            snd_flags |= SNDF_EMULATED;
         if(use_hw)
            snd_flags |= SNDF_HARDWARE;
         break;
      }
      SND_PRINT(C_fstring("Alloc voice - failed, free farthest."));
      if(use_3d_hw || use_hw){
         const S_vector &cam_pos = cam->GetMatrix()(3);
                              //hardware requested, try to free most far paused hw buffer
         float max_dist = 0.0f;
         PI3D_sound fahrest_voice = NULL;
         for(int i=drv->sounds.size(); i--; ){
            PI3D_sound snd = drv->sounds[i];
            if(snd!=this && snd->voice && snd->state==PAUSE){
               bool check_this_voice = false;
               if(use_3d_hw && !(snd->snd_flags&SNDF_EMULATED))
                  check_this_voice = true;
               else
               if(use_hw && (snd->snd_flags&SNDF_HARDWARE))
                  check_this_voice = true;

               if(check_this_voice){
                              //paused hw buffer, determine distance
                  float d = (snd->GetWorldPos()-cam_pos).Magnitude();
                  d -= snd->scaled_range_f;
                  if(max_dist<d){
                     max_dist = d;
                     fahrest_voice = snd;
                  }
               }
            }
         }
         if(fahrest_voice){
            SND_PRINT(C_fstring("Alloc voice - freeing '%s'.", fahrest_voice->GetName()));
                              //a voice to be freed found, free it and continue trying
            fahrest_voice->FreeVoice();
            continue;
         }
                              //failed to free voice
         if(use_3d_hw){
                              //we're out of 3D buffers, use non-3d
            use_3d_hw = false;
            isnd_open_flags &= ~(ISND_3DBUFFER | ISND_NOPANNING);
            SND_PRINT(C_fstring("Out of 3D buffers, use non-3D."));
         }else{
                              //we're out of hw buffers, use software
            use_hw = false;
            isnd_open_flags &= ~ISND_CARDMEM;
            SND_PRINT(C_fstring("Out of hw buffers, use sw."));
         }
      }else{
                              //unknown error - cannot allocate sysmem voice?!?
         SND_PRINT(C_fstring("Can't allocate - unknown error."));
         voice->Release();
         voice = NULL;
         break;
      }
   }
   pause_pos = 0;
   return;
}

//----------------------------

void I3D_sound::FreeVoice(){

   VirtualPause();
   if(voice){
      SND_PRINT(C_fstring("Releasing voice."));
      voice->Release();
      voice = NULL;
      snd_flags |= (SNDF_UPDATEPOS | SNDF_UPDATEDIR | SNDF_UPDATEDIST | SNDF_UPDATECONE | SNDF_UPDATEOUTVOL | SNDF_UPDATEFREQ);
   }
}

//----------------------------

void I3D_sound::VirtualPause(){

   if(state==PLAY){
                              //pause sound and save pause time
      voice->Stop();
      state = PAUSE;
                              //keep pause position and time, so that we may restore
                              // playing from actual position at the time
      pause_pos = voice->GetCurrTime();
                              //use render time, we're called from rendering
      int timer = drv->render_time;

      pause_time = timer;
      fade_count_down = 0;

      SND_PRINT(C_fstring("Virtual pause, pos: %i, time: %i.", pause_pos, pause_time));
   }
}

//----------------------------

void I3D_sound::VirtualPlay(){

   if(state == PAUSE){
                              //advance play position
      int timer = drv->render_time;
                              //use render time, we're called from rendering
      dword pos = pause_pos + timer - pause_time;
      if(pos >= snd_src->GetPlayTime()){
                              //wrapping around - non-looped samples will be stopped
         if(!(snd_flags&SNDF_LOOPED)){
            SND_PRINT(C_fstring("Virtual resume: non-looped voice stopped."));
            if(voice)
               voice->Stop();
            state = STOP;
            return;
         }
         pos %= snd_src->GetPlayTime();
      }
      voice->SetCurrTime(pos);
                              //unpause sound
      state = PLAY;
      SND_PRINT(C_fstring("Virtual resume, pos: %i.", pos));
   }
}

//----------------------------

bool I3D_sound::FadeVoice(int time, bool in, float &new_volume){

   assert(fade_count_down > 0);

   if(in){
      if((fade_count_down += time) >= fade_time){
         //SND_PRINT(C_fstring("Fade in finished."));
         fade_count_down = 0;
         return true;
      }
   }else{
      if((fade_count_down -= time) <= 0){
         SND_PRINT(C_fstring("Fade off finished."));
         fade_count_down = 0;
         return true;
      }
   }
   assert(fade_time);
   if(!fade_time)
      return true;

   float fade_ratio = (float)fade_count_down / (float)fade_time;
   SND_LOG(C_fstring("fade ratio: %.3f.", fade_ratio));
   new_volume *= fade_ratio;
   return false;
}

//----------------------------

bool I3D_sound::CheckIfAudible(int time, PI3D_camera cam, float &new_volume,
   float &curr_dist, float &dir_cos){

   bool sector_ok = true;
   {
      dword num_sectors = sound_sectors.size();
      if(num_sectors){
                              //check if camera is in one of sectors we're in
         int i = FindPointerInArray((void**)sound_sectors.begin(), num_sectors,
            cam->GetCurrSector());
         if(i!=-1){
                              //sector found
            //fade_time = cam->GetCurrSector()->GetSoundProps().switch_fade_time;
            fade_time = 1000;
         }else{
                              //not in any sector
            sector_ok = false;
         }
      }
   }

   if(!sector_ok){
      if(state!=PLAY)
         return false;
                              //if currently playing, setup fade-out
      if(!fade_count_down){
         SND_PRINT(C_fstring("not in any audible sector, starting fade down."));
         assert(fade_time >= 0);
         fade_count_down = fade_time;
         if(!fade_count_down)
            return false;
      }
      if(FadeVoice(time, false, new_volume)){
                              //fading finished
         return false;
      }
                              //flow... (continue in attenuation computation)
   }else{
                              //sector ok, start fading on if sound's off
      if(state!=PLAY){
         if(!fade_count_down){
            SND_PRINT(C_fstring("in audible sector, starting fade on."));
            fade_count_down = 1;
         }
      }
   }

   switch(snd_type){

   case I3DSOUND_AMBIENT:
      if(sector_ok && fade_count_down)
         FadeVoice(time, true, new_volume);
      new_volume *= volume;
      return true;

   case I3DSOUND_POINT:
   case I3DSOUND_SPOT:
   case I3DSOUND_POINTAMBIENT:
      {
                              //get listener/sound positions
         const S_matrix &m = GetMatrixDirect();
         const S_matrix &cm = cam->GetMatrixDirect();
         const S_vector &snd_pos = m(3);
         const S_vector &cam_pos = cm(3);
                              //get direction vector
         S_vector dir_to_cam = cam_pos - snd_pos;
         float curr_dist_2 = dir_to_cam.Dot(dir_to_cam);

                              //check for maximal distance
         if(curr_dist_2<scaled_range_f_2){
            curr_dist = sqrt(curr_dist_2);
            bool audible = false;

            if(snd_type==I3DSOUND_SPOT){
               if(!IsMrgZero(curr_dist)){
                              //check outside cone
                  dir_to_cam /= curr_dist;

                  S_vector snd_dir = m(2);
                  snd_dir.Normalize();
                  dir_cos = dir_to_cam.Dot(snd_dir);
                  if(volume_out || dir_cos>cos_cone_out){
                     audible = true;
                  }
               }else
                  audible = true;
            }else
               audible = true;

            if(audible){
               if(sector_ok && fade_count_down)
                  FadeVoice(time, true, new_volume);

                              //apply sound's volume
               new_volume *= volume;
               return true;
            }
         }
      }
      break;
   }
   return false;
}

//----------------------------

void I3D_sound::ComputeAttenuation(PI3D_frame listener, float curr_dist, float dir_cos,
   float &new_volume){

   switch(snd_type){
   case I3DSOUND_SPOT:
                              //apply cone volume
      if(dir_cos<cos_cone_in){
         if(dir_cos>cos_cone_out){
                              //apply cone volume
            new_volume *=
               volume_out + (1.0f-volume_out)*(dir_cos-cos_cone_out) / (cos_cone_in-cos_cone_out);
         }else
            new_volume *= volume_out;
      }
                              //flow...
   case I3DSOUND_POINT:
   case I3DSOUND_POINTAMBIENT:
                                 //apply distance
      if(curr_dist > scaled_range_n){
         assert(curr_dist <= scaled_range_f);
         float delta = scaled_range_f - scaled_range_n;
         if(!IsMrgZero(delta)){
            float f = (scaled_range_f - curr_dist) / delta;
            new_volume *= f;
         }
      }
      break;
   }
}

//----------------------------

void I3D_sound::ComputePanning(PI3D_frame listener){

   switch(snd_type){
   case I3DSOUND_SPOT:
   case I3DSOUND_POINT:
      {
                              //emulate position by panning
         const S_matrix &m = GetMatrixDirect();
         const S_matrix &lm = listener->GetMatrixDirect();
         const S_vector &l_dir = lm(2);
         S_vector dir_to_cam = lm(3) - m(3);
         float ca = (!IsAbsMrgZero(l_dir.z)) ? atan(l_dir.x/l_dir.z) : PI*.5f;
         if(l_dir.z<0.0f) ca += GetPI();
         if(l_dir.x<0.0f) ca += GetPI2();
         float sa = (!IsAbsMrgZero(dir_to_cam.z)) ? atan(dir_to_cam.x/dir_to_cam.z) : PI*.5f;
         if(dir_to_cam.z>=0.0f) sa += GetPI();
         if(dir_to_cam.x>=0.0f) sa += GetPI2();
         float diff = sa - ca;
         float pan = sin(diff);
         voice->SetPanning(pan);
      }
      break;
   }
}

//----------------------------

void I3D_sound::UpdateSound(int time){

   if(frm_flags&FRMFLAGS_CHANGE_MAT){
      snd_flags |= (SNDF_UPDATEPOS | SNDF_UPDATEDIR | SNDF_UPDATESCALE);
   }

   if(!snd_src || (state==STOP))
      return;

   const S_matrix &m = GetMatrix();

   SND_LOG(C_fstring("Sound '%s':", GetName()));

                              //update scale first, because it affects our distance computations
   if(snd_flags&SNDF_UPDATESCALE){
      float f = m(0).Magnitude();
      scaled_range_n = f * range_n;
      scaled_range_f = f * range_f;

                           //run-time values
      scaled_range_f_2 = scaled_range_f * scaled_range_f;
      f = scaled_range_f - scaled_range_n;

      snd_flags &= ~SNDF_UPDATESCALE;
   }

   bool audible = false;

   float new_volume = 1.0f;

   PI3D_camera cam = owning_scene->GetActiveCamera1();
   float curr_dist;
   float dir_cos;             //valid and used only for spot sounds
   audible = CheckIfAudible(time, cam, new_volume, curr_dist, dir_cos);
   SND_LOG(C_fstring("audible: %i", audible));

                              //toggle sound pause based on if audible
   if(!audible){
      switch(state){
      case PLAY:
         VirtualPause();
         break;
      case PAUSE:             
                              //check non-looped sample for finish
         if(!(snd_flags&SNDF_LOOPED)){
            int timer = drv->render_time;
                              //use render time, we're called from rendering
            dword pos = pause_pos + timer - pause_time;
            if(pos >= snd_src->GetPlayTime()){
               SND_PRINT(C_fstring("Virtual pase - loop over."));
               SetOn(false);
            }
         }
         break;
      }
   }else{
      bool start = false;

      if(!voice){
                              //create and upload sample
         MakeVoice(cam);
         if(!voice)           //failed to create voice?
            return;
      }
                              //update frequency
      if(snd_flags&SNDF_UPDATEFREQ){
         voice->SetFrequency(frequency);
         snd_flags &= ~SNDF_UPDATEFREQ;
      }
      if(snd_flags&SNDF_EMULATED){
         ComputeAttenuation(cam, curr_dist, dir_cos, new_volume);
         ComputePanning(cam);
      }else{
#ifdef EMULATE_ATTENUATION
         ComputeAttenuation(cam, curr_dist, dir_cos, new_volume);
#else
         if(snd_flags&SNDF_UPDATEDIR)
            voice->SetDir(m(2));
         if(snd_flags&SNDF_UPDATEDIST)
            voice->SetDistances(scaled_range_n, scaled_range_f);
         if(snd_flags&SNDF_UPDATEOUTVOL)
            voice->SetOutVolume((snd_type==I3DSOUND_SPOT) ? volume_out : 1.0f);
         if(snd_flags&SNDF_UPDATECONE)
            voice->SetCone(cone_in, cone_out);
#endif//!EMULATE_ATTENUATION

#ifdef EMULATE_3DPOSITIONING
         //EmulatePanning(cam);
#else
                              //update 3D hw sound
         if(snd_flags&SNDF_UPDATEPOS)
            voice->SetPos(m(3));
#endif

         snd_flags &= ~(SNDF_UPDATEPOS | SNDF_UPDATEDIR | SNDF_UPDATEDIST | SNDF_UPDATECONE | SNDF_UPDATEOUTVOL);
      }

      if(state == I3D_sound::PAUSE){
         VirtualPlay();
                              //looped voices won't get play
         if(state!=PLAY)
            return;

         start = true;
      }else{                  
                              //check non-looped sample for finish
         if(!(snd_flags&SNDF_LOOPED) && !voice->IsPlaying()){
            if(voice){
               SND_PRINT(C_fstring("Loop over, stopping sound."));
               voice->Stop();
            }
            state = STOP;
            return;
         }
      }
                              //update volume
      SND_LOG(C_fstring("volume: %.3f.", new_volume));
      if(!IsAbsMrgZero(virtual_volume-new_volume)){
         virtual_volume = new_volume;
         //SND_PRINT(C_fstring("Setting volume: %.3f.", new_volume));
         voice->SetVolume(new_volume);
      }

      if(start){
         ISND_RESULT ir;
         ir = voice->Play((snd_flags&SNDF_LOOPED) ? ISND_LOOP : 0);
         assert(ISND_SUCCESS(ir));
      }
                              //stats
      ++drv->scene_stats.frame_count[FRAME_SOUND];
   }
}

//----------------------------

void I3D_sound::GetRange(float &min, float &max, float &af) const{

   min = range_n, max = range_f;
   af = ambient_factor;
}

//----------------------------

void I3D_sound::GetCone(float &in_angle, float &out_angle) const{

   in_angle = cone_in;
   out_angle = cone_out;
}

//----------------------------

I3D_RESULT I3D_sound::SetVolume(float v){

   volume = v;
   return I3D_OK;
}

//----------------------------
#define SPHERE_NUM_VERTS 24
static word sphere_c_list[SPHERE_NUM_VERTS*2];


void I3D_sound::Draw1(PI3D_scene scene, const S_view_frustum *vf, bool strong) const{

                              //sphere values
   static struct S_auto_init{
      S_auto_init(){
                                 //init volume sphere
         for(int i=0; i<SPHERE_NUM_VERTS; i++){
            sphere_c_list[i*2+0] = i + 0;
            sphere_c_list[i*2+1] = i + 1;
         }
         sphere_c_list[i*2-1] = 0;
      }
   } auto_init;

   static const S_vector color_cone_out(.0, .75f, .8f), color_cone_in(.0f, .5f, .5f);

   byte alpha(strong ? 0 : 0xa0);

   switch(snd_type){

   case I3DSOUND_NULL:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 3, alpha);
      break;

   case I3DSOUND_POINT:
   case I3DSOUND_POINTAMBIENT:
      scene->DebugDrawSphere(GetMatrix(), range_f, color_cone_out, alpha);
      scene->DebugDrawSphere(GetMatrix(), range_n, color_cone_in, alpha);
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 3, alpha);
      break;

   case I3DSOUND_SPOT:
      {
         float cone[2];
         GetCone(cone[0], cone[1]);
         cone[0] *= .5f; cone[1] *= .5f;
         float rng[2] = { range_n, range_f };
         S_matrix m = GetIdentityMatrix();
         const int sphere_num_verts = 24;
         S_vector sphere_circle[sphere_num_verts+1];
                              //outer range
         float cosc = cos(cone[1]);
         float sinc = sin(cone[1]);
         for(int j=0; j<sphere_num_verts; j++){
            sphere_circle[j].z = cosc;
            float f((float)j*GetPI2()/(float)sphere_num_verts);
            sphere_circle[j].x = sinc * cos(f);
            sphere_circle[j].y = sinc * sin(f);
         }
         sphere_circle[sphere_num_verts].Zero();
         for(int ii=0; ii<2; ii++){
            m(0, 0) = m(1, 1) = m(2, 2) = rng[ii];
            scene->SetRenderMatrix(m*GetMatrix());
            scene->DrawLines(sphere_circle, sphere_num_verts, sphere_c_list, sphere_num_verts*2,
               !ii ? color_cone_in : color_cone_out, sizeof(S_vector), alpha);
                                 //connecting lines
            static const word indx[]={
               sphere_num_verts, 0,
               sphere_num_verts, sphere_num_verts/4,
               sphere_num_verts, sphere_num_verts/2,
               sphere_num_verts, sphere_num_verts*3/4,
            };
            scene->DrawLines(sphere_circle, sphere_num_verts+1,
               indx, sizeof(indx)/sizeof(word), color_cone_out, sizeof(S_vector), alpha);
         }
         if(vf){                 //check clipping
            bool clip;
            if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
               return;
         }
         scene->DrawIcon(GetWorldPos(), 3, alpha);
      }
      break;

   case I3DSOUND_AMBIENT:
      if(vf){                 //check clipping
         bool clip;
         if(!SphereInVF(*vf, I3D_bsphere(GetWorldPos(), .5f), clip))
            return;
      }
      scene->DrawIcon(GetWorldPos(), 8, alpha);
      break;
   }
}

//----------------------------

I3D_RESULT I3D_sound::DebugDraw(PI3D_scene scene) const{

   Draw1(scene, NULL, true);
   return I3D_OK;
}

//----------------------------
//----------------------------
