#ifndef __SOUNDI3D_H
#define __SOUNDI3D_H

#include <vector>
#include "frame.h"


//----------------------------

class I3D_sound: public I3D_frame{

#define SNDF_LOOPED        1
#define SNDF_EMULATED      2        //sample 3D positioning emulated by volume and panning

#define SNDF_HARDWARE      8        //hardware buffer used
#define SNDF_STREAMING     0x10

#define SNDF_UPDATEFREQ    0x100
                              //following flags used only by non-emulated hw sounds
#define SNDF_UPDATEPOS     0x10000
#define SNDF_UPDATEDIR     0x20000
#define SNDF_UPDATEDIST    0x40000
#define SNDF_UPDATECONE    0x80000
#define SNDF_UPDATEOUTVOL  0x100000
#define SNDF_UPDATESCALE   0x400000

   I3D_SOUNDTYPE snd_type;
   float volume, volume_out;  //volume
   float cone_in, cone_out;   //cone
   C_string file_name;        //currently-open filename
   dword open_flags;          //flags by which file is opened
   class ISND_source *snd_src;//open wave data
   float range_n, range_f;    //sound range
   float ambient_factor;
   float frequency;           //frequency multiplier

   dword snd_flags;
   PI3D_scene owning_scene;

   vector<class I3D_sector*> sound_sectors;

                              //run-time:
   class ISND_3Dsound *voice; //NULL if voice not allocated
   float virtual_volume;      //virtual volume to avoid frequent volume changes
   float scaled_range_n, scaled_range_f;  //range scaled by inherited scale
   float scaled_range_f_2;    //scaled_range_f ^ 2
   float cos_cone_in, cos_cone_out; //cos(cone_???)
   //float fade_factor;         //reciprocal of delta of scaled_range_f and scaled_fade_dist

                              //virtual pausing:
   enum{ STOP, PLAY, PAUSE } state;
   int pause_time, pause_pos;

                              //fading:
   int fade_time;             //current fade time
   int fade_count_down;       //0 = not fading

   I3D_sound(PI3D_scene);

   friend I3D_driver;         //creation
   friend I3D_scene;          //management
   friend I3D_sector;         //sound in sector
public:
   ~I3D_sound();
   void Close();

   inline vector<class I3D_sector*> &GetSectorsVector() { return sound_sectors; }

//----------------------------
// Update sound - to be called each frame to emulate sounds, manage sound audibility,
// release finished voices, etc.
   void UpdateSound(int time);

//----------------------------
// Create physical voice and upload sample into it.
   void MakeVoice(PI3D_camera); 

//----------------------------
// Pause sound (virtually, calling VirtualPause), and release resources.
   void FreeVoice();       

//----------------------------
// Pause sound (physically stop), keep time when it get paused, so that when it's
// audible again, it continues playing from that position.
   void VirtualPause();

//----------------------------
// Continue in playing of virtually-paused voice.
   void VirtualPlay();

//----------------------------
// Check if voice audible (check distance to camera), if true, also compute
// attenuation.
// As a side-effect, compute current distance to camera and
// direction cosine (audible point or spot sounds only).
   bool CheckIfAudible(int time, PI3D_camera, float &volume, float &curr_dist, float &dir_cos);

//----------------------------
// Compute attenuation based on distance to listener (camera).
   void ComputeAttenuation(PI3D_frame listener, float curr_dist, float dir_cos, float &new_volume);

//----------------------------
// Compute and apply panning effect based on position of sound relative to listener.
   void ComputePanning(PI3D_frame listener);

//----------------------------
// Compute fading. If fading is finished, return true, otherwise multiply volume
// by fade ratio.
   bool FadeVoice(int time, bool in, float &volume);

   inline PI3D_scene GetScene() const{ return owning_scene; }

   I3D_sound &operator =(const I3D_sound &);

//----------------------------
// Draw light - for debugging purposes. If 'vf' is not NULL, light's icon is clipped
// against view frustum and fast rejected if out.
   void Draw1(PI3D_scene, const S_view_frustum *vf, bool strong) const;
public:
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD(DebugDraw)(PI3D_scene) const;

   I3DMETHOD_(void,Update)();
   I3DMETHOD_(void,SetOn)(dword);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD(Open)(const char *fname, dword flags, I3D_LOAD_CB_PROC* = NULL,
      void *context = NULL, void *special = NULL);
   I3DMETHOD_(bool,IsPlaying)() const;
                              //setup
   I3DMETHOD(SetSoundType)(I3D_SOUNDTYPE);
   I3DMETHOD(SetRange)(float min, float max, float ambient_facrot);
   I3DMETHOD(SetCone)(float in_angle, float out_angle);
   I3DMETHOD(SetOutVol)(float);
   I3DMETHOD(SetVolume)(float);
   I3DMETHOD(SetLoop)(bool);
   I3DMETHOD(SetStreaming)(bool);
   I3DMETHOD(SetCurrTime)(dword);
   I3DMETHOD(SetFrequency)(float);

   I3DMETHOD_(I3D_SOUNDTYPE,GetSoundType)() const{ return snd_type; }
   I3DMETHOD_(void,GetRange)(float &min, float &max, float &ambient_facrot) const;
   I3DMETHOD_(void,GetCone)(float &in_angle, float &out_angle) const;
   I3DMETHOD_(float,GetOutVol)() const{ return volume_out; }
   I3DMETHOD_(float,GetVolume)() const{ return volume; }
   I3DMETHOD_(bool,IsLoop)() const{ return (snd_flags&SNDF_LOOPED); }
   I3DMETHOD_(bool,IsStreaming)() const{ return (snd_flags&SNDF_STREAMING); }
   I3DMETHOD_(dword,GetCurrTime)() const;
   I3DMETHOD_(dword,GetPlayTime)() const;
   I3DMETHOD_(float,GetFrequency)() const{ return frequency; }

   I3DMETHOD_(const char*,GetFileName)() const{ return file_name; }
   I3DMETHOD_(dword,GetOpenFlags)() const{ return open_flags; }
   I3DMETHOD_(class ISND_source*,GetSoundSource)() const{ return snd_src; }

   I3DMETHOD_(dword,NumSoundSectors)() const{ return sound_sectors.size(); }
   I3DMETHOD_(class I3D_sector* const*,GetSoundSectors)() const{ return sound_sectors.begin(); }
};

//----------------------------
//----------------------------
#endif
