#ifndef __TEXTURE_H
#define __TEXTURE_H

#include "common.h"


//----------------------------
                              //Class: basic texture, providing pointer to D3D
                              // texture interface on request, managing the contents
                              // in a transparent manner.
                              //This class also specifies rendering mode (alpha),
                              // size of texture and unique sorting value.
class I3D_texture_base{

#define TXTF_CKEY_ALPHA    2     //colorkey info in (possibly) 1-bit alpha
#define TXTF_ALPHA         4     //alpha pixels
#define TXTF_SELF_ALPHA    8     //used for ADD blending mode
#define TXTF_MIPMAP        0x10  //has mipmap levels

protected:
   PI3D_driver drv;
   mutable dword txt_flags;
   const dword sort_id;       //rendering sort identifier
   dword size_x, size_y;

public:   
   I3D_texture_base(PI3D_driver);

   I3DMETHOD_(dword,AddRef)() = 0;
   I3DMETHOD_(dword,Release)() = 0;

   I3DMETHOD_(const char*,GetFileName)(int index = 0) const{ return NULL; }
   I3DMETHOD_(dword,SizeX)() const{ return size_x; }
   I3DMETHOD_(dword,SizeY)() const{ return size_y; }
   I3DMETHOD_(const IImage*,GetSysmemImage)() const{ return NULL; }
public:

//----------------------------
// Get pointer to D3D texture. Inherited classes may animate contents of texture or
// swap among multiple pointers in this call.
// It is legal to return NULL by this function (in case of internal failure), in which case
// nothing will be rendered.
   virtual IDirect3DTexture8 *GetD3DTexture() = 0;

//----------------------------
// Get RLE compression mask keeping colorkey info. Optional.
   virtual const byte *GetRLEMask() const{ return NULL; }

//----------------------------
// Read access to internal data.
   inline dword SizeX1() const{ return size_x; }
   inline dword SizeY1() const{ return size_y; }
   inline dword GetTxtFlags() const{ return txt_flags; }
   inline dword GetSortID() const{ return sort_id; }
};

typedef I3D_texture_base *PI3D_texture_base;
typedef const I3D_texture_base *CPI3D_texture_base;

//----------------------------
                              //Class: classical static texture, keeping static bitmap.
                              //When dirty (TXTF_UPLOAD_NEEDED set), the sysmem copy is
                              // updated to vidmem copy, using dirty rectangles of
                              // D3D sysmem texture.
class I3D_texture: public I3D_texture_base{

#define TXTF_UPLOAD_NEEDED 0x10000  //need uploading (Manage)

protected:
   dword ref;
   C_smart_ptr<IImage> sysmem_image;   //static copy of texture data
   C_smart_ptr<IImage> vidmem_image;   //vidmem image - placeholder for D3D texture
   C_smart_ptr<IDirect3DTexture8> d3d_tex;

   dword img_open_flags;      //IMGOPEN_??? flags, used for creation of vidmem image
   mutable dword last_render_time;  //used for caching

                              //creation params, used when texture lost
   C_string filenames[3];     //pic_name, opt_name, bump_name;
   dword create_flags;

   I3D_texture(const I3D_texture&);
   const I3D_texture& operator =(const I3D_texture&);

public:
   I3D_texture(PI3D_driver);
   ~I3D_texture();

   inline void SetDirty(){ txt_flags |= TXTF_UPLOAD_NEEDED; }
   inline void SetTxtFlags(dword dw){ txt_flags = dw; }
   inline bool IsCkeyAlpha() const{ return txt_flags&TXTF_CKEY_ALPHA; }

   enum E_manage_cmd{
      MANAGE_FREE,            //free of video memory, release resources
      MANAGE_UPLOAD,          //load to video memory
      MANAGE_UPLOAD_IF_FREE,  //load to video memory, if there's some space
      MANAGE_CREATE_ONLY,     //only create vidmem image, no upload
   };
   void Manage(E_manage_cmd cmd);
   I3D_RESULT Open(const struct I3D_CREATETEXTURE*, I3D_LOAD_CB_PROC*, void *cb_context);
   bool Match(const struct I3D_CREATETEXTURE*);
   bool Reload();

   virtual IDirect3DTexture8 *GetD3DTexture();
   virtual const byte *GetRLEMask() const{
      return (const byte*)sysmem_image->GetColorkeyInfo();
   }
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(const char*,GetFileName)(int index = 0) const{
      if(index<0 || index>2)
         return NULL;
      return filenames[index].Size() ? (const char*)filenames[index] : NULL;
   }
   I3DMETHOD_(const IImage*,GetSysmemImage)() const{ return sysmem_image; }
};

//----------------------------
                              //Class: animated texture sequence.
class I3D_animated_texture: public I3D_texture_base{
protected:
   dword ref;
   int anim_delay;            //in ms

                              //run-time:
   mutable dword last_render_time;
   mutable int anim_count;
   mutable int curr_phase;
   vector<C_smart_ptr<I3D_texture> > texture_list;
public:
   I3D_animated_texture(PI3D_driver d1);

   inline void AddTexture(PI3D_texture tp){
      if(!texture_list.size())
         txt_flags = tp->GetTxtFlags();
      texture_list.push_back(tp);
   }
   inline void SetAnimSpeed(int as){ anim_delay = as; }

   virtual IDirect3DTexture8 *GetD3DTexture();
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
   I3DMETHOD_(const char*,GetFileName)(int i = 0) const{
      return texture_list.size() ? texture_list.front()->GetFileName(i) : NULL;
   }
};

//----------------------------

class I3D_camera_texture: public I3D_texture_base{
   C_smart_ptr<IImage> vidmem_image;   //vidmem image - placeholder for D3D texture
protected:
   dword ref;
public:
   I3D_camera_texture(PI3D_driver d1);
public:
   I3DMETHOD_(dword,AddRef)(){ return ++ref; }
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }
   virtual IDirect3DTexture8 *GetD3DTexture();
};

//----------------------------

#endif
