/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Visual.cpp
   Content: Visual frame.
--------------------------------------------------------*/

#include <D3dx8core.h>
#pragma comment(lib, "advapi32")

#include "all.h"
#include "visual.h"
#include "vb_mgr.h"
#include "light.h"
#include "scene.h"
#include "camera.h"
#include <malloc.h>

//----------------------------

#define ENABLE_ADVANCED_FOG
#define USE_SHARED_DEST_VB    //allocate dest VB on portions of huge VBs (VB alloc manager)


                              //report if too many lights shine on visual
                              // undef to disable also in debug build
#if defined _DEBUG || 1
# define DEBUG_REPORT_TOO_MANY_LIGHTS
#endif

//----------------------------
//----------------------------

void I3D_dest_vertex_buffer::CreateD3DVB(PI3D_driver drv, dword fvf_flags_in,
   int max_vertices_in, bool allow_hw_trans, bool force_xyz_format){

   DestroyD3DVB(drv);

   max_vertices = max_vertices_in;
   fvf_flags = fvf_flags_in;
   dword d3d_usage = 0;
   d3d_usage |= D3DUSAGE_WRITEONLY;
   //d3d_usage |= D3DUSAGE_DYNAMIC;
   D3DPOOL mem_pool = D3DPOOL_DEFAULT;

   allow_hw_trans = (allow_hw_trans && drv->IsDirectTransform());

   if(!force_xyz_format){
                              //include transformed or untransformed 
                              // destination vertex format
      if(allow_hw_trans){
         bool use_b1 = ((fvf_flags&D3DFVF_POSITION_MASK) == D3DFVF_XYZB1);
         fvf_flags &= ~D3DFVF_POSITION_MASK;
         if(!use_b1) 
            fvf_flags |= D3DFVF_XYZ;
         else
            fvf_flags |= D3DFVF_XYZB1;
      }else{
         fvf_flags &= ~D3DFVF_POSITION_MASK;
         fvf_flags |= D3DFVF_XYZRHW;

                              //include specular component (due to vertex fog)
                              // (and possibly specular effects)
         fvf_flags |= D3DFVF_SPECULAR;
      }
   }
   
   if(!drv->IsHardware()){
      if(!drv->IsDirectTransform())
         d3d_usage |= D3DUSAGE_SOFTWAREPROCESSING;
   }

   if(!allow_hw_trans){
                              //software transforming will be used
      d3d_usage |= D3DUSAGE_SOFTWAREPROCESSING;
      mem_pool = D3DPOOL_SYSTEMMEM;
   }

   int offs = 0;

   xyz_o = -1;
   xyzrhw_o = -1;

   switch(fvf_flags&D3DFVF_POSITION_MASK){
   case D3DFVF_XYZ:
      xyz_o = offs;
      offs += sizeof(S_vector);
      break;
   case D3DFVF_XYZB1:
      xyz_o = offs;
      offs += sizeof(S_vector) + sizeof(float);
      break;
   case D3DFVF_XYZB2:
      xyz_o = offs;
      offs += sizeof(S_vector) + 2*sizeof(float);
      break;
   case D3DFVF_XYZB3:
      xyz_o = offs;
      offs += sizeof(S_vector) + 3*sizeof(float);
      break;
   case D3DFVF_XYZRHW:
      xyzrhw_o = offs;
      offs += sizeof(S_vectorw);
      break;
   default:
      assert(0);
   }
   if(fvf_flags&D3DFVF_NORMAL)
      offs += sizeof(S_vector);

   diff_o = -1;
   if(fvf_flags&D3DFVF_DIFFUSE){
      diff_o = offs;
      offs += sizeof(dword);
   }
   
   spec_o = -1;
   if(fvf_flags&D3DFVF_SPECULAR){
      spec_o = offs;
      offs += sizeof(dword);
   }
   
   txt_o = -1;
   num_txtc = (fvf_flags&D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT;
   if(num_txtc){
      txt_o = offs;
      offs += num_txtc * sizeof(I3D_text_coor);
   }
   size_of_vertex = offs;

#ifdef USE_SHARED_DEST_VB
   drv->AllocVertexBuffer(fvf_flags, mem_pool, d3d_usage,
      max_vertices,
      &D3D_vertex_buffer, &D3D_vertex_buffer_index, true);
#else
   HRESULT hr;
   hr = drv->GetDevice1()->CreateVertexBuffer(max_vertices*size_of_vertex,
      d3d_usage, fvf_flags, mem_pool, (IDirect3DVertexBuffer8**)&D3D_vertex_buffer);
   CHECK_D3D_RESULT("CreateVertexBuffer", hr);
   D3D_vertex_buffer_index = 0;
#endif
}

//----------------------------

void I3D_dest_vertex_buffer::DestroyD3DVB(PI3D_driver drv){

   if(D3D_vertex_buffer){
#ifdef USE_SHARED_DEST_VB
      drv->FreeVertexBuffer(D3D_vertex_buffer, D3D_vertex_buffer_index);
#else
      D3D_vertex_buffer->Release();
#endif
      D3D_vertex_buffer = NULL;
      D3D_vertex_buffer_index = 0;
   }
}

//----------------------------
//----------------------------
                              
I3D_visual::I3D_visual(PI3D_driver d):
   I3D_frame(d),
   last_render_time(0),
   dest_vb_trans_id(0),
   vis_flags(0),
   material_id(0),
   envmap_uv_set(-1),
   hide_dist(0.0f),
   last_auto_lod(-1)
{
   drv->AddCount(I3D_CLID_VISUAL);
   type = FRAME_VISUAL;

   enum_mask = ENUMF_VISUAL;

   bound.bound_local.bbox.Invalidate();
   bound.bound_local.bsphere.pos.Zero();
   bound.bound_local.bsphere.radius = 0.0f;
   hr_bound.bound_local.bbox.Invalidate();

}

//----------------------------

I3D_visual::~I3D_visual(){

   drv->DecCount(I3D_CLID_VISUAL);
   vertex_buffer.DestroyD3DVB(drv);
}

//----------------------------

bool I3D_visual::ComputeBounds(){

   if(!(vis_flags&VISF_BOUNDS_VALID)){
      if(bound.bound_local.bbox.IsValid()){
                              //compute bounding sphere
         S_vector bbox_diagonal = (bound.bound_local.bbox.max - bound.bound_local.bbox.min) * .5f;
         bound.bound_local.bsphere.pos = bound.bound_local.bbox.min + bbox_diagonal;
         bound.bound_local.bsphere.radius = bbox_diagonal.Magnitude();
      }else{
         bound.bound_local.bsphere.pos.Zero();
         bound.bound_local.bsphere.radius = 0.0f;
      }

      vis_flags |= VISF_BOUNDS_VALID;
      frm_flags &= ~FRMFLAGS_BSPHERE_TRANS_VALID;
   }

   return true;
}

//----------------------------

bool I3D_visual::AttachProcedural(PI3D_procedural_base pb){

   return false;
}

//----------------------------

I3D_RESULT I3D_visual::DebugDraw(PI3D_scene) const{
   return I3DERR_UNSUPPORTED;
}

//----------------------------

void I3D_visual::Update(){

   if(frm_flags&FRMFLAGS_UPDATE_NEEDED){
      vis_flags &= ~VISF_DEST_LIGHT_VALID;
      I3D_frame::Update();
   }
}

//----------------------------

I3D_RESULT I3D_visual::Duplicate(CPI3D_frame frm){

   switch(frm->GetType1()){
   case FRAME_VISUAL:
      {
         PI3D_visual vis = I3DCAST_VISUAL(frm);
         material_id = vis->material_id;
         hide_dist = vis->hide_dist;
      }
      break;
   }
   I3D_RESULT ir = I3D_frame::Duplicate(frm);
   return ir;
}

//----------------------------

void I3D_visual::PrepareDestVB(I3D_mesh_base *mb, dword num_txt_sets){

   dword numv = mb->NumVertices1();
   if(!numv)
      return;

   envmap_uv_set = -1;
   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);

   if(drv->NumSimultaneousTextures() > num_txt_sets){
                           //search for any env-mapped material
      for(int j=mb->NumFGroups1(); j--; ){
         PI3D_material mat = mb->GetFGroups1()[j].mat;
         if(mat->GetEnvironmetEffect() != EE_NULL){
            envmap_uv_set = num_txt_sets++;
            break;
         }
      }
   }

   if(drv->IsDirectTransform()){

      vertex_buffer.CreateD3DVB(drv, mb->vertex_buffer.GetFVFlags(), numv, true, true);

      assert(vertex_buffer.GetSizeOfVertex() == mb->vertex_buffer.GetSizeOfVertex());
                              //plain copy src buffer to dest
      D3D_lock<byte> v_src(mb->vertex_buffer.GetD3DVertexBuffer(), mb->vertex_buffer.D3D_vertex_buffer_index,
         numv, mb->vertex_buffer.GetSizeOfVertex(), D3DLOCK_READONLY);
      D3D_lock<byte> v_dst(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.D3D_vertex_buffer_index, numv,
         vertex_buffer.GetSizeOfVertex(), 0);
      v_src.AssureLock();
      v_dst.AssureLock();
      memcpy(v_dst, v_src, numv*vertex_buffer.GetSizeOfVertex());

   }else{
                              //create dest VB for ProcessVertices
      vertex_buffer.CreateD3DVB(drv, D3DFVF_DIFFUSE | (num_txt_sets<<D3DFVF_TEXCOUNT_SHIFT),
         mb->NumVertices1(), false);
   }
}

//----------------------------

int I3D_visual::GetAutoLOD(const S_preprocess_context &pc, const float *auto_lods, dword al_stride,
   dword num_auto_lods, const float *dist_range) const{

   assert(auto_lods && dist_range);
                              //get distance to camera
   float factor = (pc.viewer_pos - matrix(3)).Magnitude();
                              //multiply by object's scale
   float scale = matrix(0).Magnitude();
   if(scale<MRG_ZERO)
      return 0;
   factor /= scale;
   factor *= pc.LOD_factor;

   if(factor<dist_range[0])
      return -1;
   if(factor>=dist_range[1])
      factor = 1.0f;
   else
      factor = (factor - dist_range[0]) / (dist_range[1] - dist_range[0]);

   const float &lowest_quality = *(float*)(((byte*)auto_lods)+(num_auto_lods-1)*al_stride);
   factor *= (1.0f - lowest_quality);
   factor = 1.0f - Max(0.0f, Min(1.0f, factor));
   factor -= lowest_quality*.5f;

   for(int i=0; i<num_auto_lods; i++, auto_lods = (float*)(((byte*)auto_lods)+al_stride)){
      if(*auto_lods < factor)
         break;
   }
   assert(i!=-1);
   --i;
   return i;
}

//----------------------------

const I3D_driver::S_vs_shader_entry *I3D_visual::PrepareVertexShader(CPI3D_material mat,
   I3D_driver::S_vs_shader_entry_in &se_in, const S_render_primitive &rp, dword flags, const S_matrix *m_trans) const{

#ifdef DEBUG_REPORT_TOO_MANY_LIGHTS
# define DEBUG_CHECK_LIGHT_COUNT \
   if(light_count==MAX_VS_LIGHTS){ \
      drv->DEBUG(C_fstring("vis %s: too many lights", (const char*)name), 2); \
      i=0; \
      continue; \
   } 
#else
//# define DEBUG_CHECK_LIGHT_COUNT
# define DEBUG_CHECK_LIGHT_COUNT \
   if(light_count==MAX_VS_LIGHTS){ \
      i=0; continue; \
   } 
#endif

   HRESULT hr;

   bool alt_m_trans = (m_trans!=NULL);
   if(!m_trans)
      m_trans = &matrix;

                              //add transform code fragment
   if(flags&VSPREP_TRANSFORM)
      se_in.AddFragment(I3D_driver::VSF_TRANSFORM);

   bool object_space = !(flags&VSPREP_WORLD_SPACE);

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
                           //lighting:
   typedef float t_c_data[4];
   t_c_data light_params[MAX_VS_LIGHTS * 3];
   dword num_light_params = 0;

   dword light_count = 0;
   vis_flags &= ~VISF_FOG_VALID;

   const I3D_material_data &md = mat->GetMaterialData();

                              //process ligting
   if(flags&VSPREP_MAKELIGHTING){
      const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);

      S_matrix m_tmp;
      float f_tmp, *inv_scale = NULL;
      if(alt_m_trans)
         m_tmp = ~(*m_trans);
      const S_matrix &inv_matrix = !alt_m_trans ? GetInvMatrix1() : m_tmp;

      S_vectorw light_base = md.emissive;
      light_base.w = md.diffuse_alpha;

      if(light_base.x >= 1.0f && light_base.y >= 1.0f && light_base.z >= 1.0f){
         if(!(flags&VSPREP_LIGHT_MUL_ALPHA))
            se_in.AddFragment(I3D_driver::VSF_LIGHT_COPY);
         else{
            se_in.AddFragment(I3D_driver::VSF_LIGHT_BEGIN);
            se_in.AddFragment(I3D_driver::VSF_LIGHT_END_ALPHA);
         }
      }else{

      se_in.AddFragment(I3D_driver::VSF_LIGHT_BEGIN);

      dword light_flags = I3DLIGHTMODE_VERTEX;
      if(flags&VSPREP_LIGHT_LM_DYNAMIC)
         light_flags = I3DLIGHTMODE_DYNAMIC_LM;

      //const vector<PI3D_light> &sct_lights = rp.sector->GetLights();
      const vector<PI3D_light> &sct_lights = !(flags&VSPREP_LIGHT_LM_DYNAMIC) ?
         rp.sector->GetLights_vertex() : rp.sector->GetLights_dun_LM();
      const PI3D_light *lpptr = sct_lights.begin();
      for(int i=sct_lights.size(); i--; ){
         PI3D_light lp = *lpptr++;
#ifdef USE_PREFETCH
         if(i) Prefetch1(&(*lpptr)->ltype);
#endif
         assert(!lp->IsFogType());

                              //check if light is on
         if(!lp->IsOn1())
            continue;

         I3D_LIGHTTYPE lt = lp->GetLightType1();
         switch(lt){

         case I3DLIGHT_AMBIENT:
            light_base += lp->color * md.diffuse * lp->power;
            break;

         case I3DLIGHT_POINTAMBIENT:
            {
               const S_matrix &lmat = lp->GetMatrixDirect();

               S_vector dir = lmat(3) - (*m_trans)(3);
               float dist_2 = dir.Dot(dir);
               if(dist_2 > lp->range_f_scaled_2)
                  break;
               float a;
               if(dist_2 < (lp->range_n_scaled*lp->range_n_scaled)){
                  a = 1.0f;
               }else{
                  if(IsAbsMrgZero(lp->range_delta_scaled))
                     break;
                  float dist = sqrt(dist_2);
                  a = (lp->range_f_scaled - dist) / lp->range_delta_scaled;
               }
               light_base += lp->color * md.diffuse * (a * lp->power);
            }
            break;

         case I3DLIGHT_DIRECTIONAL:
         case I3DLIGHT_POINT:
         //case I3DLIGHT_LAYEREDFOG:
         //case I3DLIGHT_POINTAMBIENT:
            {
               //S_vector color = lp->diffuse * md.diffuse * lp->power;
               S_vector color = lp->color * md.diffuse * (lp->power * drv->vertex_light_mult);
               if(color.IsNull())
                  break;
               switch(lt){
               case I3DLIGHT_DIRECTIONAL:
                  {
                     /*
                     if(md && md->two_sided){
                        light_base += lp->color * md->diffuse * (lp->power * .5f);
                        continue;
                     }
                     */

                     DEBUG_CHECK_LIGHT_COUNT;

                     S_vectorw *vw = (S_vectorw*)&light_params[num_light_params];
                     if(object_space){
                        vw[0] = lp->normalized_dir % inv_matrix;
                        vw[0].Normalize();
                     }else{
                        vw[0] = lp->normalized_dir;
                     }
                     vw[1] = color;
                     vw[0].w = 0.0f;
                     vw[1].w = 0.0f;
                     num_light_params += 2;
                     se_in.AddFragment(object_space ?
                        I3D_driver::VSF_LIGHT_DIR_OS :
                        I3D_driver::VSF_LIGHT_DIR_WS);
                  }
                  break;

               case I3DLIGHT_POINT:
               //case I3DLIGHT_POINTAMBIENT:
                  {
                     const S_matrix &m = lp->GetMatrix();

                                 //fast reject lights too far from object
                     S_vector l_dist = bsphere.pos - m(3);
                     float dist_2 = l_dist.Dot(l_dist);
                     float r_sum = lp->range_f_scaled + bsphere.radius;
                     if((r_sum*r_sum) <= dist_2)
                        continue;

                     DEBUG_CHECK_LIGHT_COUNT;

                     S_vectorw *vw = (S_vectorw*)&light_params[num_light_params];

                     const S_matrix &lmat = lp->GetMatrixDirect();
                     S_vector &c_color = vw[1];
                     c_color = color;
                     float range_delta = lp->range_delta_scaled;
                     float &c_far_range = vw[0].w;
                     I3D_driver::E_VS_FRAGMENT fid;
                     if(object_space){
                        if(!inv_scale){
                           f_tmp = inv_matrix(0).Magnitude();
                           inv_scale = &f_tmp;
                        }
                        vw[0] = lmat(3) * inv_matrix;
                        c_far_range = lp->range_f_scaled * (*inv_scale);
                        range_delta *= (*inv_scale);

                        if(lt==I3DLIGHT_POINT){
                           /*
                           if(md && md->two_sided){
                                       //use ambient for 2-sided mats,
                                       // but scaled to half
                              fid = I3D_driver::VSF_LIGHT_POINT_AMB_OS;
                              c_color *= .5f;
                           }else
                              */
                              fid = I3D_driver::VSF_LIGHT_POINT_OS;
                        }else
                           fid = I3D_driver::VSF_LIGHT_POINT_AMB_OS;
                     }else{
                        vw[0] = lmat(3);
                        c_far_range = lp->range_f_scaled;

                        fid = (lt==I3DLIGHT_POINT) ?
                           I3D_driver::VSF_LIGHT_POINT_WS :
                           I3D_driver::VSF_LIGHT_POINT_AMB_WS;
                     }
                     se_in.AddFragment(fid);

                     float &c_r_range_delta = vw[1].w;
                     if(!IsAbsMrgZero(range_delta))
                        c_r_range_delta = 1.0f / range_delta;
                     else
                        c_r_range_delta = 1e+8f;

                     num_light_params += 2;
                  }
                  break;
               }
                              //check for lignt count overflow
               ++light_count;
               /*
#ifndef DEBUG_REPORT_TOO_MANY_LIGHTS
               if(light_count==MAX_VS_LIGHTS) i = 0;
#endif
               */
            }
            break;
         }
      }
      if(flags&VSPREP_LIGHT_MUL_ALPHA)
         se_in.AddFragment(I3D_driver::VSF_LIGHT_END_ALPHA);
      else
         se_in.AddFragment(I3D_driver::VSF_LIGHT_END);

      }

      light_base *= drv->vertex_light_mult;
      hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_AMBIENT, &light_base, 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   }

   bool do_env_mapping = ((envmap_uv_set!=-1) &&
      !(flags&VSPREP_NO_ENVMAP) &&
      (drv->GetFlags()&DRVF_SINGLE_PASS_MODULATE2X));

   if(((flags&VSPREP_COPY_UV) || do_env_mapping) && !(flags&VSPREP_NO_COPY_UV))
      se_in.AddFragment(I3D_driver::VSF_COPY_UV_0);

   if(do_env_mapping){
      dword env_code = I3D_driver::VSF_GENERATE_ENVUV_WS_1 - (int)object_space;
                              //for bumpmapping in hw, we must shuffle UVs a bit
      if(drv->IsDirectTransform()){
         PI3D_texture_base tb_bump = mat->GetTexture1(MTI_BUMP);
         if(tb_bump && (drv->GetFlags2()&DRVF2_BUMPMAPPING)){
            se_in.AddFragment(I3D_driver::VSF_COPY_UV_0TO1);
                              //we'll compute envmap into UV #2
            env_code += 2;
         }
      }
                              //add envmap computation
      se_in.AddFragment((I3D_driver::E_VS_FRAGMENT)env_code);
   }

   if((drv->GetFlags()&DRVF_USEFOG) && !(flags&VSPREP_NO_FOG)){
#ifdef ENABLE_ADVANCED_FOG
      const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);

      S_matrix m_tmp;
      const S_matrix *m_inv = NULL;

      se_in.AddFragment(I3D_driver::VSF_FOG_BEGIN);

      const vector<PI3D_light> &sct_fogs = rp.sector->GetFogs();
      const PI3D_light *fpptr = sct_fogs.begin();
      for(int i=sct_fogs.size(); i--; ){
         PI3D_light lp = *fpptr++;
         assert(lp->IsFogType());

         if(!(lp->IsOn1() && (lp->GetLightFlags()&I3DLIGHTMODE_VERTEX)))
            continue;

         I3D_LIGHTTYPE lt = lp->GetLightType1();
         switch(lt){
         case I3DLIGHT_LAYEREDFOG:
            {
               const S_matrix &m = lp->GetMatrix();

                                 //fast reject lights too far from object
               S_vector l_dist = bsphere.pos - m(3);
               float dist_2 = l_dist.Dot(l_dist);
               float r_sum = lp->range_f_scaled + bsphere.radius;
               if((r_sum*r_sum) <= dist_2)
                  continue;
               {              //todo: optimize - use this plane later
                  S_plane pl1;
                  pl1.normal = -m(1);
                  float light_scale = pl1.normal.Magnitude();
                  float height = lp->range_n * light_scale;
                  pl1.normal /= light_scale;
                  pl1.d = -m(3).Dot(pl1.normal) + height;
                  float dist_to_plane = bsphere.pos.DistanceToPlane(pl1);
                  if(dist_to_plane < -bsphere.radius || dist_to_plane > bsphere.radius+height)
                     continue;
               }

               DEBUG_CHECK_LIGHT_COUNT;

               float r_light_scale;
               S_vectorw *vw = (S_vectorw*)&light_params[num_light_params];
               num_light_params += 3;

               float power = lp->power;
               S_plane &pl = *(S_plane*)&vw[0];
               if(object_space){
                  se_in.AddFragment(I3D_driver::VSF_FOG_LAYERED_OS);
                  if(!m_inv){
                                          //get inverse matrix now
                     if(alt_m_trans){
                        m_tmp = ~(*m_trans);
                        m_inv = &m_tmp;
                     }else
                        m_inv = &GetInvMatrix1();
                  }

                  S_vector loc_pos = m(3) * (*m_inv);
                  pl.normal = -m(1) % (*m_inv);
                  float height = lp->range_n * pl.normal.Magnitude();
                  r_light_scale = !IsAbsMrgZero(height) ? (1.0f / height) : 1e+8f;
                  pl.normal *= power;
                  pl.d = -(loc_pos.Dot(pl.normal)) + height * power;

                  vw[2] = loc_pos;
                  vw[2].w = lp->range_f_scaled * (*m_inv)(2).Magnitude();
               }else{
                  se_in.AddFragment(I3D_driver::VSF_FOG_LAYERED_WS);
                  pl.normal = -m(1);
                  float height = lp->range_n * pl.normal.Magnitude();
                  r_light_scale = !IsAbsMrgZero(height) ? (1.0f / height) : 1e+8f;
                  pl.normal *= power;
                  pl.d = -(m(3).Dot(pl.normal)) + height * power;

                  vw[2] = m(3);
                  vw[2].w = lp->range_f_scaled;
               }
                                       //keep range squared
               vw[2].w *= vw[2].w;
                                 //scale
               pl.normal *= r_light_scale;
               pl.d *= r_light_scale;

               vw[1] = lp->color * lp->power;
               vw[1].w = lp->power;


               vis_flags |= VISF_FOG_VALID;
                              //check for lignt count overflow
               if(++light_count==MAX_VS_LIGHTS) i = 0;
            }
            break;

         case I3DLIGHT_FOG:
            {
               se_in.AddFragment(object_space ?
                  I3D_driver::VSF_FOG_HEIGHT_OS :
                  I3D_driver::VSF_FOG_HEIGHT_WS);
               vis_flags |= VISF_FOG_VALID;
                                    //setup fog constants
               S_vectorw v4;
               v4.w = 0.0f;
               float &fog_range_n = v4.x, &fog_range_f = v4.y, &fog_range_r_delta = v4.z;
               fog_range_f = lp->range_f;
               if(IsAbsMrgZero(fog_range_f))
                  fog_range_f = rp.scene->GetActiveCamera1()->GetFCP();
               fog_range_n = lp->range_n * fog_range_f;
               fog_range_r_delta = 1.0f / Max(.01f, fog_range_f - fog_range_n);
               hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_FOGPARAMS, &v4, 1);
               CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
            }
            break;

         default: continue;
         }
      }
      if(vis_flags&VISF_FOG_VALID){
                              //ok, write result of fog
         se_in.AddFragment((flags&VSPREP_HALVE_FOG) ?
            I3D_driver::VSF_FOG_END_HALVED : I3D_driver::VSF_FOG_END);
      }else{
                              //no fog light, remove preparation instruction
         se_in.PopFragment();
      }
#else//ENABLE_ADVANCED_FOG
      PI3D_light lp = rp.sector->GetFogLight();
      if(lp && lp->IsOn1() && (lp->I3D_light::GetMode1()&I3DLIGHTMODE_VERTEX)){
         se_in.AddFragment(object_space ?
            I3D_driver::VSF_FOG_SIMPLE_OS :
            I3D_driver::VSF_FOG_SIMPLE_WS);
         vis_flags |= VISF_FOG_VALID;
                              //setup fog constants
         S_vectorw v4;
         float &fog_range_n = v4.x, &fog_range_f = v4.y, &fog_range_r_delta = v4.z;
         v4.w = 0.0f;
         fog_range_f = lp->range_f;
         if(IsAbsMrgZero(fog_range_f))
            fog_range_f = rp.scene->GetActiveCamera1()->GetFCP();
         fog_range_n = lp->range_n * fog_range_f;
         fog_range_r_delta = 1.0f / Max(.01f, fog_range_f - fog_range_n);
         hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_FOGPARAMS, &v4, 1);
         CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
      }
#endif//!ENABLE_ADVANCED_FOG
   }

                              //create/get shader and set to D3D
   const I3D_driver::S_vs_shader_entry *se = drv->GetVSHandle(se_in);
   if(!se){
      abort();
      return NULL;
   }
   drv->SetVertexShader(se->vs_handle);

                              //fill-in constants
   if(flags&VSPREP_FEED_MATRIX){
                              //fill-in transformation matrix
      S_matrix m;
      m.Make4X4Transposed(*m_trans, rp.scene->GetViewProjHomMat());
      hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0, &m, 4);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   }

                              //feed lighting constants
   for(int i=0; i<num_light_params; i++){
      assert(se->vscs_light_param[i] < 96);
      hr = lpDev8->SetVertexShaderConstant(se->vscs_light_param[i], light_params[i], 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   }

   if(do_env_mapping){
                              //setup constants
      const S_matrix *inv_matrix = &GetInvMatrix1();
                              //try find helper animation dummy
      const vector<PI3D_frame> &children = GetChildren1();
      for(int i=children.size(); i--; ){
         PI3D_frame frm = children[i];
         if(frm->GetType1()==FRAME_DUMMY && !strncmp(frm->GetOrigName(), "env*", 3)){
                              //use rotation matrix of helper dummy
            inv_matrix = &frm->GetInvMatrix1();
            break;
         }
      }
      S_vectorw cam_pos = rp.scene->GetActiveCamera1()->GetMatrixDirect()(3);
      if(object_space)
         cam_pos = cam_pos * (*inv_matrix);
      hr = lpDev8->SetVertexShaderConstant(se->vscs_cam_loc_pos, &cam_pos, 1);
      CHECK_D3D_RESULT("SetVertexShaderConstant", hr);
   }

   return se;
}

//----------------------------

void I3D_visual::SetupEnvMapping(E_ENVIROMENTAL_EFFECT ee, CPI3D_material mat, dword stage){

   if((drv->GetFlags2()&DRVF2_ENVMAPPING) && (drv->GetFlags()&DRVF_SINGLE_PASS_MODULATE2X)){

      if(drv->NumSimultaneousTextures() >= 3){
                           //setup bump-map
         PI3D_texture_base tb_bump = mat->GetTexture1(MTI_BUMP);
         if(tb_bump && (drv->GetFlags2()&DRVF2_BUMPMAPPING)){
                           //use primary UV coords for bump mapping
            if(!drv->IsDirectTransform())
               drv->SetTextureCoordIndex(stage, 0);

            drv->SetupTextureStage(stage, D3DTOP_BUMPENVMAP);
            drv->SetTexture1(stage, tb_bump);
   
            drv->SetBumpScale(stage, mat->GetBumpOpacity());
                           //use next stage for environment map
            ++stage;
         }else
            drv->DisableTextureStage(2);
      }
                              //use secondary UVs for environment
      if(!drv->IsDirectTransform())
         drv->SetTextureCoordIndex(stage, 1);
      drv->SetupTextureStage(stage, D3DTOP_MODULATE2X);

      PI3D_texture_base tb = mat->GetTexture1(MTI_ENVIRONMENT);
      drv->SetTexture1(stage, tb);
   }else{
      drv->DisableTextureStage(stage);
   }
}

//----------------------------

int I3D_visual::AddPrimitives1(I3D_mesh_base *mb, S_preprocess_context &pc){

   int curr_auto_lod = -1;
   const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
   {
      curr_auto_lod = drv->force_lod_index;
      if(curr_auto_lod==-1 && auto_lods.size()){
                              //compute automatic LOD based on distance
         curr_auto_lod = GetAutoLOD(pc, &auto_lods[0].ratio, sizeof(C_auto_lod),
            auto_lods.size(), mb->GetAutoLODDists());
      }else{
         curr_auto_lod = Min(curr_auto_lod, (int)auto_lods.size()-1);
      }
   }
                           //get LOD's vertex count and face source
   const vector<I3D_face_group> *fgrps;
   int vertex_count;
   if(curr_auto_lod < 0){
      fgrps = &mb->GetFGroupVector();
      vertex_count = mb->vertex_buffer.NumVertices();
   }else{
      fgrps = &auto_lods[curr_auto_lod].fgroups;
      vertex_count = auto_lods[curr_auto_lod].vertex_count;
   }
   if(!vertex_count)
      return 0;

                           //make sure destination VB is allocated
   if(!vertex_buffer.GetD3DVertexBuffer())
      PrepareDestVB(mb);

                           //add primitives to sort list
   for(int ii=fgrps->size(); ii--; ){
      const I3D_face_group &fg = (*fgrps)[ii];
      word face_count = fg.NumFaces1();
      assert(face_count);

                           //driver stats
      drv->render_stats.triangle += face_count;

      pc.prim_list.push_back(S_render_primitive(curr_auto_lod));
      S_render_primitive &p = pc.prim_list.back();
   
      p.vis = this;
      p.clip = pc.clip;
      p.sector = pc.curr_sector;
      p.scene = pc.scene;
      p.user = ii;

      bool is_alpha = false;
      dword sort_value;
                           //determine blending mode
      p.blend_mode = 0;
      PI3D_material mp = fg.GetMaterial1();

      if(!mp->IsTransl()){
         p.blend_mode |= I3DBLEND_OPAQUE;
         sort_value = PRIM_SORT_OPAQUE;
         ++pc.opaque;
      }else
      if(mp->IsCkeyAlpha1()){
         p.blend_mode |= I3DBLEND_ALPHABLEND;
         sort_value = PRIM_SORT_CKEY_ALPHA;
         ++pc.ckey_alpha;
         is_alpha = true;
      }else{
         PI3D_texture_base tb = mp->GetTexture1();
         if(tb && (tb->GetTxtFlags()&TXTF_SELF_ALPHA)) p.blend_mode |= I3DBLEND_ADD;
         else p.blend_mode |= I3DBLEND_ALPHABLEND;
         is_alpha = true;
         sort_value = PRIM_SORT_ALPHA_NOZWRITE;
         ++pc.alpha_nozwrite;
      }
      if(mp->IsDiffuseAlpha()) 
         p.blend_mode |= I3DBLEND_VERTEXALPHA;

      if(!is_alpha){
                           //sort by material
         sort_value |= (mp->GetSortID()&PRIM_SORT_MAT_MASK)<<PRIM_SORT_MAT_SHIFT;
      }else{
                           //sort by distance
         const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
         const S_matrix &m_view = pc.scene->GetActiveCamera1()->GetViewMatrix1();
         float depth =
            bsphere.pos[0] * m_view(0, 2) +
            bsphere.pos[1] * m_view(1, 2) +
            bsphere.pos[2] * m_view(2, 2) +
                             m_view(3, 2);
         sort_value |= ((FloatToInt(depth * 100.0f))&PRIM_SORT_DIST_MASK) << PRIM_SORT_DIST_SHIFT;
      }
      p.sort_value = sort_value;
   }
                           //driver's stats
   drv->render_stats.vert_trans += vertex_count;

   return vertex_count;
}

//----------------------------

void I3D_visual::DrawPrimitive1(I3D_mesh_base *mb, const S_preprocess_context &pc, const S_render_primitive &rp){

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

   if(!drv->IsDirectTransform()){
      drv->SetVertexShader(vertex_buffer.GetFVFlags());
   }

   const I3D_face_group *fg;
   int vertex_count;
   int base_index;
   if(rp.curr_auto_lod < 0){
                              //direct mesh
      fg = &mb->GetFGroupVector()[rp.user];
      vertex_count = mb->vertex_buffer.NumVertices();
      base_index = mb->index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(mb->index_buffer.GetD3DIndexBuffer(), vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }else{
      const C_auto_lod &al = mb->GetAutoLODs()[rp.curr_auto_lod];
      fg = &al.fgroups[rp.user];
      vertex_count = al.vertex_count;
      base_index = al.index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(al.index_buffer.GetD3DIndexBuffer(), vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
   }

   PI3D_material mat = fg->GetMaterial1();
   if(pc.mode!=RV_SHADOW_CASTER){
      drv->SetRenderMat(mat);

      if(drv->last_blend_flags!=rp.blend_mode)
         drv->SetupBlend(rp.blend_mode);
      drv->SetupTextureStage(0, drv->vertex_light_blend_op);
                              //apply environment effect
      E_ENVIROMENTAL_EFFECT ee = mat->GetEnvironmetEffect();
      if(ee!=EE_NULL && envmap_uv_set!=-1)
         SetupEnvMapping(ee, mat, envmap_uv_set);
      else
         drv->DisableTextureStage(1);
      drv->EnableFog(vis_flags&VISF_FOG_VALID);
      drv->EnableAnisotropy(0, true);
   }else{
      if(mat->IsTextureAlpha() || mat->IsCkeyAlpha1()){
         PI3D_texture_base tb = mat->GetTexture1(MTI_DIFFUSE);
         drv->SetTexture1(0, tb);
         if(tb){
            dword ar = (tb->GetTxtFlags()&TXTF_MIPMAP) ? 0x60 : 0xe0;
            drv->SetAlphaRef(ar);
         }
      }else{
         drv->SetTexture1(0, NULL);
      }
      drv->EnableNoCull(mat->Is2Sided1());
   }

   drv->SetStreamSource(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.GetSizeOfVertex());

                              //debug mode - without textures
   if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES)){
                              //use factor (50% gray) instead of texture
      drv->SetTexture1(0, NULL);
      hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      drv->SetTextureFactor(0xff808080);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
         (base_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);
      hr = lpDev8->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      CHECK_D3D_RESULT("SetTextureStageState", hr);
      return;
   }

//----------------------------
#if 0
                              //experiment with pixel shaders
   {
      drv->SetLegacyPixelShader();

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
         (base_index + fg->base_index) * 3, fg->num_faces);
      CHECK_D3D_RESULT("DrawIP", hr);

      hr = lpDev8->SetPixelShader(NULL);
      CHECK_D3D_RESULT("SetPixelShader", hr);
      return;
   }
#endif
//----------------------------

   hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
      (base_index + fg->base_index) * 3, fg->num_faces);
   CHECK_D3D_RESULT("DrawIP", hr);
}

//----------------------------

void I3D_visual::RenderSolidMesh(PI3D_scene scene, bool debug, const I3D_cylinder *volume_cylinder){

                              //re-render all our primitives,
   PI3D_mesh_base mb = GetMesh();
   if(!mb)
      return;

   LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
   HRESULT hr;

                              //feed tranformation matrix
   S_matrix m;
   m.Make4X4Transposed(matrix, scene->GetViewProjHomMat());
   hr = lpDev8->SetVertexShaderConstant(I3D_driver::VSC_MAT_TRANSFORM_0, &m, 4);
   CHECK_D3D_RESULT("SetVertexShaderConstant", hr);

   drv->SetClipping(true);
                              //no fog in rendering (shadow, etc)
   IDirect3DVertexBuffer8 *vb_src = mb->vertex_buffer.GetD3DVertexBuffer();
   drv->SetStreamSource(vb_src, mb->vertex_buffer.GetSizeOfVertex());

   int curr_auto_lod = last_auto_lod;
   const I3D_face_group *fg;
   if(curr_auto_lod < 0){
      fg = &mb->GetFGroupVector()[0];
   }else{
      fg = &mb->GetAutoLODs()[curr_auto_lod].fgroups[0];
   }

   IDirect3DIndexBuffer8 *ib;
   dword vertex_count;
   dword ib_base;
   dword face_count;

   if(curr_auto_lod < 0){
                           //direct mesh
      ib = mb->index_buffer.GetD3DIndexBuffer();
      vertex_count = mb->vertex_buffer.NumVertices();
      ib_base = mb->index_buffer.D3D_index_buffer_index * 3;
      face_count = mb->index_buffer.NumFaces1();
   }else{
                           //automatic LOD
      const C_auto_lod &al = mb->GetAutoLODs()[curr_auto_lod];
      ib = al.index_buffer.GetD3DIndexBuffer();
      vertex_count = al.vertex_count;
      ib_base = (al.index_buffer.D3D_index_buffer_index + fg->base_index) * 3;
      face_count = fg->num_faces;
   }

   drv->EnableNoCull(fg->mat->Is2Sided1());

   if(!volume_cylinder){
      hr = lpDev8->SetIndices(ib, mb->vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count, ib_base, face_count);
      CHECK_D3D_RESULT("DrawIP", hr);
      return;
   }

   if(face_count>=8192)
      return;
                              //render only faces inside of volume cylinder

                              //transform cylinder to object's local coords
   const S_matrix &m_inv = GetInvMatrix();
   I3D_cylinder vc_loc;
   vc_loc.pos = volume_cylinder->pos * m_inv;
   vc_loc.dir = volume_cylinder->dir % m_inv;
   float inv_scale = m_inv(0).Magnitude();
   vc_loc.radius = volume_cylinder->radius * inv_scale;

   int num_hit_faces = 0;

   I3D_triface *src, *dst;
   dword vsize = mb->vertex_buffer.GetSizeOfVertex();
   S_vector *verts;

                              //lock source and dest faces and vertices
   hr = ib->Lock(ib_base*sizeof(word), face_count*sizeof(I3D_triface), (byte**)&src, D3DLOCK_READONLY);
   CHECK_D3D_RESULT("Lock", hr);
#if 0
                              //allow appending to the buffer
   if(face_count+drv->ib_shadow_base_index <= 8192){
      hr = drv->ib_shadow->Lock(drv->ib_shadow_base_index*sizeof(I3D_triface), face_count*sizeof(I3D_triface),
         (byte**)&dst, D3DLOCK_NOOVERWRITE);
   }else
#endif
   {
      drv->ib_shadow_base_index = 0;
      hr = drv->ib_shadow->Lock(0, face_count*sizeof(I3D_triface), (byte**)&dst, D3DLOCK_DISCARD);
   }
   CHECK_D3D_RESULT("Lock", hr);
   hr = vb_src->Lock(mb->vertex_buffer.D3D_vertex_buffer_index*vsize, vertex_count*vsize, (byte**)&verts, D3DLOCK_READONLY);
   CHECK_D3D_RESULT("Lock", hr);

#if 1
   const S_vector vc0 = vc_loc.pos;
   const S_vector vc1 = vc_loc.pos + vc_loc.dir;

   for(int i=face_count; i--; ){
#if 0                         //prefetching
      if(i){
         const I3D_triface &fc = src[i-1];
         Prefetch(((byte*)verts)+fc[0]*vsize);
         Prefetch(((byte*)verts)+fc[1]*vsize);
         Prefetch(((byte*)verts)+fc[2]*vsize);
      }
#endif
      const I3D_triface &fc = src[i];
      const S_vector *v[3];
      v[0] = (S_vector*)(((byte*)verts)+fc[0]*vsize);
      v[1] = (S_vector*)(((byte*)verts)+fc[1]*vsize);
      v[2] = (S_vector*)(((byte*)verts)+fc[2]*vsize);

      S_plane pl;
      pl.normal.GetNormal(*v[0], *v[1], *v[2]);
                              //skip culled faces
      float dir_dot_norm = pl.normal.Dot(vc_loc.dir);
      if(CHECK_ZERO_GREATER(dir_dot_norm))
         continue;

      pl.d = -pl.normal.Dot(*v[0]);
      float dist_vc0 = vc0.Dot(pl.normal) + pl.d;
      float dist_vc1 = vc1.Dot(pl.normal) + pl.d;
                              //if both ends out in the same hemi-space, skip face
      if(CHECK_SAME_SIGNS(dist_vc0, dist_vc1)){
         float len = sqrt(pl.normal.Square());
         float r = vc_loc.radius * len;
         if(dist_vc0>=r && dist_vc1>=r)
            continue;
         if(dist_vc0<=-r && dist_vc1<=-r)
            continue;
      }
                              //check all 3 sides of triangle
      word vi0 = 0;
      for(int j=3; j--; ){
         word vi1 = vi0;
         vi0 = j;
                              //get edge's normal
         S_vector e_dir = (*v[vi1]) - (*v[vi0]);
         S_plane pl_e;
         pl_e.normal.x = e_dir.y * pl.normal.z - e_dir.z * pl.normal.y;
         pl_e.normal.y = e_dir.z * pl.normal.x - e_dir.x * pl.normal.z;
         pl_e.normal.z = e_dir.x * pl.normal.y - e_dir.y * pl.normal.x;

         pl_e.d = -v[vi0]->Dot(pl_e.normal);
         dist_vc0 = vc0.Dot(pl_e.normal) + pl_e.d;
         dist_vc1 = vc1.Dot(pl_e.normal) + pl_e.d;
                              //if both ends out in the same hemi-space, skip face
         if(CHECK_SAME_SIGNS(dist_vc0, dist_vc1)){
            float len = sqrt(pl_e.normal.Square());
            float r = vc_loc.radius * len;
            if(dist_vc0>=r && dist_vc1>=r)
               goto skip_face;
         }
      }
                           //add face for rendering
      dst[num_hit_faces++] = fc;
   skip_face:
      {}
   }

//----------------------------
#else
   float vc_loc_radius_2 = vc_loc.radius * vc_loc.radius;
   float vc_loc_size_2 = vc_loc.dir.Magnitude();
   vc_loc_size_2 *= vc_loc_size_2;

   for(int i=face_count; i--; ){
      const I3D_triface &sf = src[i];
      bool is_hit = true;

      int num_hit_edges = 0;
      dword hit_mask = 0;
      const S_vector *tri_verts[3];
      word vi0 = sf[0];
      for(int j=3; j--; ){
         word vi1 = vi0;
         vi0 = sf[j];
         const S_vector &v0 = *(S_vector*)(((byte*)verts)+vi0*vsize);
         tri_verts[j] = &v0;
                           //check if this edge collides
         bool is_col = true;
                           //determine collision now
         const S_vector &v1 = *(S_vector*)(((byte*)verts)+vi1*vsize);
         float u1, u2;
         S_vector edge_dir = v1-v0;
         if(LineIntersection(v0, edge_dir, vc_loc.pos, vc_loc.dir, u1, u2)){
            is_col = (u2>=0.0f && u2<=1.0f);
            if(is_col){
                        //check distance
               S_vector p0 = v0 + edge_dir * u1;
               S_vector p1 = vc_loc.pos + vc_loc.dir * u2;
               float dist2 = (p0-p1).Square();
               is_col = (dist2 < vc_loc_radius_2);
            }
         }
         if(is_col){
            ++num_hit_edges;
            hit_mask |= (1<<j);
         }
      }
      switch(num_hit_edges){
      case 0:
         {
                           //not hit by edge, still may be hit
            S_plane pl;
            pl.normal.GetNormal(*tri_verts[0], *tri_verts[1], *tri_verts[2]);
                           //check if face is ok
            float d = pl.normal.Dot(vc_loc.dir);
            if(CHECK_ZERO_GREATER(d)) continue;
            //pl.normal.Normalize();
            pl.d = -pl.normal.Dot(*tri_verts[0]);
            S_vector ip;
            if(pl.Intersection(vc_loc.pos, vc_loc.dir, ip)){
               is_hit = IsPointInTriangle(ip, *tri_verts[0], *tri_verts[1], *tri_verts[2], pl.normal);
               if(is_hit){
                  float dist_2 = (ip-vc_loc.pos).Square();
                  is_hit = (dist_2 < vc_loc_size_2);
               }
            }
         }
         break;
      case 1:
         {
                           //hit by one edge, check if really hit
            S_plane pl;
            pl.normal.GetNormal(*tri_verts[0], *tri_verts[1], *tri_verts[2]);
            float d = pl.normal.Dot(vc_loc.dir);
            if(CHECK_ZERO_GREATER(d)) continue;
            pl.d = -pl.normal.Dot(*tri_verts[0]);
            S_vector ip;
            if(pl.Intersection(vc_loc.pos, vc_loc.dir, ip)){
               S_vector n;
               float d;
               const S_vector &face_normal = pl.normal;
               if(!(hit_mask&1)){
                  S_vector v((*tri_verts[0]) - (*tri_verts[1]));
                  n.x = face_normal.y * v.z - face_normal.z * v.y;
                  n.y = face_normal.z * v.x - face_normal.x * v.z;
                  n.z = face_normal.x * v.y - face_normal.y * v.x;
                  d = (*tri_verts[0]).Dot(n) - ip.Dot(n);
                  if(CHECK_ZERO_LESS(d))
                     is_hit = false;
               }
               if(is_hit && !(hit_mask&2)){
                  S_vector v((*tri_verts[1]) - (*tri_verts[2]));
                  n.x = face_normal.y * v.z - face_normal.z * v.y;
                  n.y = face_normal.z * v.x - face_normal.x * v.z;
                  n.z = face_normal.x * v.y - face_normal.y * v.x;
                  d = (*tri_verts[1]).Dot(n) - ip.Dot(n);
                  if(CHECK_ZERO_LESS(d))
                     is_hit = false;
               }
               if(is_hit && !(hit_mask&4)){
                  S_vector v((*tri_verts[2]) - (*tri_verts[0]));
                  n.x = face_normal.y * v.z - face_normal.z * v.y;
                  n.y = face_normal.z * v.x - face_normal.x * v.z;
                  n.z = face_normal.x * v.y - face_normal.y * v.x;
                  d = (*tri_verts[2]).Dot(n) - ip.Dot(n);
                  if(CHECK_ZERO_LESS(d))
                     is_hit = false;
               }
            }
         }
         break;
      }
                           //add face if hit
      if(is_hit)
         dst[num_hit_faces++] = sf;
   }
#endif

   hr = drv->ib_shadow->Unlock();
   CHECK_D3D_RESULT("Unlock", hr);
   hr = ib->Unlock();
   CHECK_D3D_RESULT("Unlock", hr);
   hr = vb_src->Unlock();
   CHECK_D3D_RESULT("Unlock", hr);

   if(num_hit_faces){
      hr = lpDev8->SetIndices(drv->ib_shadow, mb->vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);

      hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
         drv->ib_shadow_base_index*3, num_hit_faces);
      CHECK_D3D_RESULT("DrawIP", hr);
   }
   drv->ib_shadow_base_index += num_hit_faces;
}

//----------------------------
//----------------------------

