#ifndef __VISUAL_H
#define __VISUAL_H

#include <d3d8.h>
#include "frame.h"
#include "mesh.h"
#include "scene.h"
#include "camera.h"
#include "procedural.h"
#include <new.h>

//----------------------------
                              //destination vertex format may hold following items:
                              // - xyz, or xyz+rhw, zero or more betas
                              // - diffuse
                              // - specular
                              // - zero or more texture coordinates

class I3D_dest_vertex_buffer{
   LPDIRECT3DVERTEXBUFFER8 D3D_vertex_buffer;
   dword fvf_flags;
   int max_vertices;
   byte size_of_vertex;
   schar xyz_o, xyzrhw_o, diff_o, spec_o, txt_o, num_txtc;

public:
   int D3D_vertex_buffer_index;

   I3D_dest_vertex_buffer():
      D3D_vertex_buffer(NULL), 
      D3D_vertex_buffer_index(0),
      max_vertices(0),
      size_of_vertex(0), fvf_flags(0)
   {}
   void CreateD3DVB(PI3D_driver drv, dword fvf_flags, int max_vertices, bool allow_hw_trans = true, bool force_xyz_format = false);
   void DestroyD3DVB(PI3D_driver drv);

   inline int NumVertices() const{ return max_vertices; }
   inline LPDIRECT3DVERTEXBUFFER8 GetD3DVertexBuffer(){ return D3D_vertex_buffer; }

   inline int GetSizeOfVertex() const{ return size_of_vertex; }
   inline int GetXYZOffset() const{ return xyz_o; }
   inline int GetXYZRHWOffset() const{ return xyzrhw_o; }
   inline int GetDiffuseOffset() const{ return diff_o; }
   inline int GetSpecularOffset() const{ return spec_o; }
   inline int GetTexCoordOffset(int indx) const{ 
      if(txt_o!=-1 && indx<num_txtc) return txt_o + indx*8;
      return -1;
   }
   inline dword GetFVFlags() const{ return fvf_flags; }

//----------------------------

   void *Lock(dword d3d_lock_flags = 0){

      if(!max_vertices)
         return NULL;
      assert(D3D_vertex_buffer);
      void *vp;
      HRESULT hr = D3D_vertex_buffer->Lock(D3D_vertex_buffer_index * size_of_vertex,
         max_vertices * size_of_vertex,
         (byte**)&vp,
         d3d_lock_flags);
      CHECK_D3D_RESULT("Lock", hr);
      if(FAILED(hr))
         return NULL;
      return vp;
   }

//----------------------------

   bool Unlock(){
      if(!max_vertices)
         return false;
      assert(D3D_vertex_buffer);
      HRESULT hr = D3D_vertex_buffer->Unlock();
      CHECK_D3D_RESULT("Unlock", hr);
      return (SUCCEEDED(hr));
   }
};

//----------------------------

                              //flags used in call to I3D_visual::PrepareVertexShader
enum{
   VSPREP_MAKELIGHTING = 1,   //do lighting on vertices
   VSPREP_COPY_UV = 2,        //copy primary uv coordinates (set #0)
   VSPREP_WORLD_SPACE = 4,    //working in worldspace, rather than in objectspace
   VSPREP_TRANSFORM = 8,      //add simple transforming code
   VSPREP_FEED_MATRIX = 0x10, //feed transposed matrix
   VSPREP_NO_COPY_UV = 0x20,  //disable copying of primary uv
   VSPREP_NO_ENVMAP = 0x40,   //disable environment mapping
   VSPREP_NO_FOG = 0x80,      //no fog computation on vertices
   VSPREP_LIGHT_MUL_ALPHA = 0x100,  //multiply resulting diffuse alpha by in-vertex alpha (works only if VSPREP_MAKELIGHTING also included)
   VSPREP_HALVE_FOG = 0x200,  //compute halved fog - for multipass rendering
   VSPREP_LIGHT_LM_DYNAMIC = 0x400, //do lighting using LM dynamic lights (works only if VSPREP_MAKELIGHTING also included)

   VSPREP_DEBUG = 0x10000,    //debugging
};

//----------------------------

class I3D_visual: public I3D_frame{
#define VISF_BOUNDS_VALID     0x01  //bounding volume is valid
#define VISF_DEST_UV0_VALID   0x02  //cached coords in dest VB are valid (no need to copy)
#define VISF_DEST_LIGHT_VALID 0x04  //cached diffuse in dest VB is valid (no need to compute)
#define VISF_FOG_VALID        0x08  //fog info in vertices is valid (use in rendering)
#define VISF_DEST_VERTS_VALID 0x40  //cached dest processed vertices are valid

protected:
   dword visual_type;         //four-cc code
   
   I3D_dest_vertex_buffer vertex_buffer;
   dword dest_vb_trans_id;    //render view ID for which dest VB is successfully transformed
private:
                              //user-defined material ID
   dword material_id;
   friend class I3D_scene;
protected:
   dword envmap_uv_set;       //uv coord set in which we compute envmap coordinates (-1 = no)
   float hide_dist;           //distance at which visual is no more rendered (for FOV=65 and scale=1) (0=no hiding)
   int last_auto_lod;         //last LOD used to transform vertices

   dword last_render_time;

public:
   mutable dword vis_flags;
   mutable C_bound_volume bound;
   mutable C_bound_volume hr_bound;

   I3D_visual(PI3D_driver);
   ~I3D_visual();
   I3D_visual& operator =(const I3D_visual&);

   inline dword GetVisualType1() const{ return visual_type; }

   inline void SetHideDistance(float f){ hide_dist = f; }

//----------------------------
// Select level of detail in a vector of LODS, where each LOD is specified
// by a quality ration (0.0f=worst, 1.0f=best).
// The selection is made by distance of this visual to camera.
// The returned value is LOD index to use,
//    -1 if base (non-LOD) mesh should be used,
   int GetAutoLOD(const S_preprocess_context&, const float *auto_lods, dword al_stride, 
      dword num_auto_lods, const float *dist_range) const;

//----------------------------
// Determine automatic LOD and fill primitives into render list.
// Return number of vertices in this LOD, the returned value may be 0.
// Doesn't transform vertices in any way.
   int AddPrimitives1(I3D_mesh_base *mb, S_preprocess_context &pc);

//----------------------------
// Render mesh using currently set shader.
   void DrawPrimitive1(I3D_mesh_base *mb, const S_preprocess_context &pc, const S_render_primitive &rp);

//----------------------------
// Prepare VS for visual - add necessary fragments, create and setup shader,
// fill-in constants.
// If the call succeeds, the returned value is shader data. It is valid pointer
// until end of this frame.
   const I3D_driver::S_vs_shader_entry *PrepareVertexShader(CPI3D_material,
      I3D_driver::S_vs_shader_entry_in &se, const S_render_primitive &rp,
      dword flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX, const S_matrix *m_trans = NULL) const;

//----------------------------
// Sets up environment mapping params right before rendering.
   void SetupEnvMapping(E_ENVIROMENTAL_EFFECT ee, CPI3D_material mat, dword stage = 1);

public:
   I3DMETHOD(DebugDraw)(PI3D_scene) const;
   I3DMETHOD(Duplicate)(CPI3D_frame);
   I3DMETHOD_(void,Update)();
   virtual void SetMatrixDirty(){
      vis_flags &= ~VISF_DEST_LIGHT_VALID;
      I3D_frame::SetMatrixDirty();
   }
public:
   I3DMETHOD_(dword,GetVisualType)() const{ return visual_type; }

   I3DMETHOD_(const I3D_bound_volume&,GetBoundVolume)() const{
      if(!(vis_flags&VISF_BOUNDS_VALID)) ((PI3D_visual)this)->ComputeBounds();
      return bound.bound_local; 
   }
   I3DMETHOD_(const I3D_bound_volume&,GetHRBoundVolume)() const{
      if(!(frm_flags&FRMFLAGS_HR_BOUND_VALID))
         hr_bound.MakeHierarchyBounds(((PI3D_visual)this));
      return hr_bound.bound_local;

   }
   I3DMETHOD_(dword,GetLastRenderTime)() const{ return last_render_time; }
   I3DMETHOD_(I3D_mesh_base*,GetMesh)() const{ return NULL; }
   I3DMETHOD(SetProperty)(dword index, dword value){ return I3DERR_UNSUPPORTED; }
   I3DMETHOD_(dword,GetProperty)(dword index) const{ return 0xcdcdcdcd; }
   I3DMETHOD_(dword,GetCollisionMaterial)() const{ return material_id; }
   I3DMETHOD_(void,SetCollisionMaterial)(dword id){ material_id = id; }
//----------------------------
                              //non-user methods
   virtual bool ComputeBounds();
   virtual void AddPrimitives(S_preprocess_context&) = 0;
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&) = 0;

//----------------------------
// Render mesh without setting anything (for shadow rendering and debugging purposes).
   virtual void RenderSolidMesh(PI3D_scene, bool debug = false, const I3D_cylinder* = NULL);
//----------------------------
// Prepare destination VB for transforming.
   virtual void PrepareDestVB(I3D_mesh_base*, dword num_txt_sets = 1);
   virtual bool AttachProcedural(PI3D_procedural_base);
   virtual bool AddRegion(const I3D_bbox&, const S_matrix&, int index = -1){ return false; }
   I3DMETHOD_(void,VisReserved07)(){}
   I3DMETHOD_(void,VisReserved08)(){}
   I3DMETHOD_(void,VisReserved09)(){}
   I3DMETHOD_(void,VisReserved10)(){}
   I3DMETHOD_(void,VisReserved11)(){}
   I3DMETHOD_(void,VisReserved12)(){}
   I3DMETHOD_(void,VisReserved13)(){}
   I3DMETHOD_(void,VisReserved14)(){}
   I3DMETHOD_(void,VisReserved15)(){}
};

typedef I3D_visual *PI3D_visual;
typedef const I3D_visual *CPI3D_visual;

//----------------------------

#endif
