#include "all.h"
#include "visual.h"


//----------------------------

class I3D_visual_dynamic: public I3D_visual{

#define DYF_IN_LOCKED_VERTS      1     //set while vertices are locked
#define DYF_IN_LOCKED_FACES      2     //set while faces are locked

   dword dyn_flags;

   C_smart_ptr<I3D_mesh_base> mesh;
                              //VS declarator of source mesh - variable, because vertex components are also variable
   const dword *vs_decl;

//----------------------------
// Convert I3D vertex-component flags into D3D flexible vertex format flags.
   static dword ConvertFlags(dword vc_flags){
                              //deterine fvf
      dword fvf = 0;
      if(vc_flags&I3DVC_XYZ){
         if(vc_flags&I3DVC_ALPHA)
            fvf |= D3DFVF_XYZB1;
         else
            fvf |= D3DFVF_XYZ;
      }
      if(vc_flags&I3DVC_XYZRHW){
         if(vc_flags&I3DVC_ALPHA)
            return I3DERR_INVALIDPARAMS;
         fvf |= D3DFVF_XYZRHW;
      }
      if(vc_flags&D3DFVF_NORMAL) fvf |= D3DFVF_NORMAL;
      if(vc_flags&D3DFVF_DIFFUSE) fvf |= D3DFVF_DIFFUSE;
      if(vc_flags&D3DFVF_SPECULAR) fvf |= D3DFVF_SPECULAR;
      fvf |= (((vc_flags&I3DVC_TEXCOUNT_MASK) >> I3DVC_TEXCOUNT_SHIFT) << D3DFVF_TEXCOUNT_SHIFT) & D3DFVF_TEXCOUNT_MASK;

      return fvf;
   }

//----------------------------

   inline int AddPrimitives1(I3D_mesh_base *mb, S_preprocess_context &pc, bool vertex_alpha){

      int curr_auto_lod = -1;
      const vector<C_auto_lod> &auto_lods = mb->GetAutoLODs();
      if(pc.mode!=RV_SHADOW_CASTER){
         curr_auto_lod = drv->force_lod_index;
         if(curr_auto_lod==-1 && auto_lods.size()){
                                 //compute automatic LOD based on distance
            curr_auto_lod = GetAutoLOD(pc, &auto_lods[0].ratio, sizeof(C_auto_lod),
               auto_lods.size(), mb->GetAutoLODDists());
         }else{
            curr_auto_lod = Min(curr_auto_lod, (int)auto_lods.size()-1);
         }
      }
                              //get LOD's vertex count and face source
      const vector<I3D_face_group> *fgrps;
      int vertex_count;
      if(curr_auto_lod < 0){
         fgrps = &mb->GetFGroupVector();
         vertex_count = mb->vertex_buffer.NumVertices();
      }else{
         fgrps = &auto_lods[curr_auto_lod].fgroups;
         vertex_count = auto_lods[curr_auto_lod].vertex_count;
      }
      if(!vertex_count)
         return 0;

                              //add primitives to sort list
      for(int ii=fgrps->size(); ii--; ){
         const I3D_face_group &fg = (*fgrps)[ii];
         word face_count = fg.NumFaces1();
         assert(face_count);

                              //driver stats
         drv->render_stats.triangle += face_count;

         pc.prim_list.push_back(S_render_primitive(curr_auto_lod));
         S_render_primitive &p = pc.prim_list.back();
      
         p.vis = this;
         p.clip = pc.clip;
         p.sector = pc.curr_sector;
         p.scene = pc.scene;
         p.user = ii;

         bool is_alpha = false;
         dword sort_value;
                              //determine blending mode
         p.blend_mode = 0;
         PI3D_material mp = fg.GetMaterial1();

         if(!mp->IsTransl()){
            p.blend_mode |= I3DBLEND_OPAQUE;
            sort_value = PRIM_SORT_OPAQUE;
            ++pc.opaque;
         }else
         if(mp->IsCkeyAlpha1()){
            p.blend_mode |= I3DBLEND_ALPHABLEND;
            sort_value = PRIM_SORT_CKEY_ALPHA;
            ++pc.ckey_alpha;
         }else{
            PI3D_texture_base tb = mp->GetTexture1();
            if(tb && (tb->GetTxtFlags()&TXTF_SELF_ALPHA)) p.blend_mode |= I3DBLEND_ADD;
            else p.blend_mode |= I3DBLEND_ALPHABLEND;
            is_alpha = true;
            sort_value = PRIM_SORT_ALPHA_NOZWRITE;
            ++pc.alpha_nozwrite;
         }
         if(mp->IsDiffuseAlpha() || vertex_alpha)
            p.blend_mode |= I3DBLEND_VERTEXALPHA;

         if(!is_alpha){
                              //sort by material
            sort_value |= (mp->GetSortID()&PRIM_SORT_MAT_MASK)<<PRIM_SORT_MAT_SHIFT;
         }else{
                              //sort by distance
            const I3D_bsphere &bsphere = bound.GetBoundSphereTrans(this, FRMFLAGS_BSPHERE_TRANS_VALID);
            const S_matrix &m_view = pc.scene->GetActiveCamera1()->GetViewMatrix1();
            float depth =
               bsphere.pos[0] * m_view(0, 2) +
               bsphere.pos[1] * m_view(1, 2) +
               bsphere.pos[2] * m_view(2, 2) +
                                m_view(3, 2);
            sort_value |= ((FloatToInt(depth * 100.0f))&PRIM_SORT_DIST_MASK) << PRIM_SORT_DIST_SHIFT;
         }
         p.sort_value = sort_value;
      }
                              //make sure destination VB is allocated
      if(!vertex_buffer.GetD3DVertexBuffer())
         PrepareDestVB(mb);
                              //driver's stats
      drv->render_stats.vert_trans += vertex_count;

      return vertex_count;
   }
public:
   I3D_visual_dynamic(PI3D_driver d):
      I3D_visual(d),
      dyn_flags(0),
      vs_decl(NULL)
   {
      visual_type = I3D_VISUAL_DYNAMIC;
   }
   ~I3D_visual_dynamic(){
      mesh = NULL;
   }

public:
                              //specialized methods of I3D_visual:
   virtual bool ComputeBounds();
   virtual void PrepareDestVB(I3D_mesh_base*, dword num_txt_sets = 1);
   virtual void AddPrimitives(S_preprocess_context&);
   virtual void DrawPrimitive(const S_preprocess_context&, const S_render_primitive&);
   I3DMETHOD(Duplicate)(CPI3D_frame);
public:
   I3DMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   I3DMETHOD_(PI3D_mesh_base,GetMesh)() const{ return mesh; }
   I3DMETHOD_(I3D_RESULT,Build)(const void *verts, dword num_verts, dword vc_flags,
      const I3D_triface *faces, dword num_faces,
      const I3D_face_group *fgroups, dword num_fgroups, dword flags);

   I3DMETHOD_(I3D_RESULT,BuildAutoLODs)(float min_dist, float max_dist, dword num_lods, float lowest_quality){
      return I3DERR_UNSUPPORTED;
   }
   I3DMETHOD_(I3D_RESULT,DuplicateMesh)(CPI3D_mesh_base, dword vc_flags);
   I3DMETHOD_(I3D_RESULT,ResetLightInfo)(){
      if(!mesh)
         return I3DERR_NOTINITIALIZED;
      vis_flags &= ~VISF_DEST_LIGHT_VALID;
      return I3D_OK;
   }

//----------------------------

   I3DMETHOD_(void*,LockVertices)(dword lock_flags){

      if(!mesh)
         return NULL;
      if(dyn_flags&DYF_IN_LOCKED_VERTS)
         return NULL;

      void *vp;
      {
         vp = (void*)mesh->LockVertices();
      }
      if(vp){
         dyn_flags |= DYF_IN_LOCKED_VERTS;

         if(lock_flags&(I3DLOCK_XYZ | I3DLOCK_NORMAL | I3DLOCK_DIFFUSE)){
            if(!(lock_flags&I3DLOCK_NO_LIGHT_RESET))
               vis_flags &= ~VISF_DEST_LIGHT_VALID;
         }
         if(lock_flags&I3DLOCK_UV)
            vis_flags &= ~VISF_DEST_UV0_VALID;

         if(lock_flags&I3DLOCK_XYZ)
            vis_flags &= ~VISF_BOUNDS_VALID;
      }
      return vp;
   }

//----------------------------

   I3DMETHOD_(I3D_RESULT,UnlockVertices)(){

      if(!(dyn_flags&DYF_IN_LOCKED_VERTS))
         return I3DERR_INVALIDPARAMS;
      if(!mesh)
         return I3DERR_NOTINITIALIZED;
      dyn_flags &= ~DYF_IN_LOCKED_VERTS;
      if(drv->IsDirectTransform()){
         vertex_buffer.DestroyD3DVB(drv);
      }
      mesh->UnlockVertices();
      return I3D_OK;
   }

//----------------------------

   I3DMETHOD_(PI3D_triface,LockFaces)(dword lock_flags){
      if(!mesh)
         return NULL;
      if(dyn_flags&DYF_IN_LOCKED_FACES)
         return NULL;
      PI3D_triface fp = (PI3D_triface)mesh->LockFaces();
      if(fp)
         dyn_flags |= DYF_IN_LOCKED_FACES;
      return fp;
   }
   I3DMETHOD_(I3D_RESULT,UnlockFaces)(){
      if(!(dyn_flags&DYF_IN_LOCKED_FACES))
         return I3DERR_INVALIDPARAMS;
      if(!mesh)
         return I3DERR_NOTINITIALIZED;
      dyn_flags &= ~DYF_IN_LOCKED_FACES;
      mesh->UnlockFaces();
      return I3D_OK;
   }

//----------------------------

   I3DMETHOD_(I3D_RESULT,Render)(){

      if(!mesh)
         return I3DERR_INVALIDPARAMS;
                              //allow only rendering of XYZRHW formats
      dword fvf_flags = mesh->vertex_buffer.GetFVFlags();
      if((fvf_flags&D3DFVF_POSITION_MASK) != D3DFVF_XYZRHW)
         return I3DERR_INVALIDPARAMS;

      drv->SetVertexShader(fvf_flags);
      LPDIRECT3DDEVICE8 lpDev8 = drv->GetDevice1();
      HRESULT hr;

      int vertex_count = mesh->vertex_buffer.NumVertices();
      int base_index = mesh->index_buffer.D3D_index_buffer_index;
      hr = lpDev8->SetIndices(mesh->index_buffer.GetD3DIndexBuffer(),
         mesh->vertex_buffer.D3D_vertex_buffer_index);
      CHECK_D3D_RESULT("SetIndices", hr);
      drv->SetClipping(true);
      drv->SetStreamSource(mesh->vertex_buffer.GetD3DVertexBuffer(), mesh->vertex_buffer.GetSizeOfVertex());

      if(fvf_flags&D3DFVF_DIFFUSE){
         drv->SetupBlend(I3DBLEND_ALPHABLEND | I3DBLEND_VERTEXALPHA);
         drv->SetupTextureStage(0, D3DTOP_MODULATE);
      }else{
         drv->SetupBlend(I3DBLEND_ALPHABLEND);
         drv->SetupTextureStage(0, D3DTOP_SELECTARG1);
      }

      drv->EnableFog(false);
      for(int i=mesh->NumFGroups1(); i--; ){
         const I3D_face_group *fg = &mesh->GetFGroupVector()[i];
         PI3D_material mat = fg->GetMaterial1();
         if(!mat)
            return I3DERR_INVALIDPARAMS;
         drv->SetRenderMat(mat);

                           //apply environment effect
         E_ENVIROMENTAL_EFFECT ee = mat->GetEnvironmetEffect();
         if(ee!=EE_NULL && envmap_uv_set!=-1)
            SetupEnvMapping(ee, mat, envmap_uv_set);
         else
            drv->DisableTextureStage(1);

                           //debug mode - without textures
         if(!(drv->GetFlags2()&DRVF2_DRAWTEXTURES))
            drv->SetTexture1(0, NULL);

         hr = lpDev8->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, vertex_count,
            (base_index + fg->base_index) * 3, fg->num_faces);
         CHECK_D3D_RESULT("DrawIP", hr);
      }
      return I3D_OK;
   }

//----------------------------

   I3DMETHOD_(dword,NumVertices)() const{

      if(!mesh)
         return 0;
      return mesh->NumVertices1();
   }
};

//----------------------------

I3D_RESULT I3D_visual_dynamic::Build(const void *verts, dword num_verts, dword vc_flags,
   const I3D_triface *faces, dword num_faces,
   const I3D_face_group *fgroups, dword num_fgroups,
   dword flags){

                              //make data validity checks
   if(!num_verts || !num_faces || !num_fgroups)
      return I3DERR_INVALIDPARAMS;

   if((vc_flags&I3DVC_XYZ) && (vc_flags&I3DVC_XYZRHW))
      return I3DERR_INVALIDPARAMS;

   if((vc_flags&I3DVC_NORMAL) && (vc_flags&I3DVC_DIFFUSE))
      return I3DERR_INVALIDPARAMS;

#ifdef _DEBUG
   int i;
                              //check if fgroups ok
   for(i=num_fgroups; i--; ){
      if((fgroups[i].base_index + fgroups[i].num_faces) > num_faces || !fgroups[i].mat){
         assert(0);
         return I3DERR_INVALIDPARAMS;
      }
   }
                              //check if indicies ok
   for(i=num_faces; i--; ){
      const I3D_triface &fc = faces[i];
      if(fc[0]>num_verts || fc[1]>num_verts || fc[2]>num_verts){
         assert(0);
         return I3DERR_INVALIDPARAMS;
      }
   }
#endif
   dword fvf = ConvertFlags(vc_flags);

   mesh = new I3D_mesh_base(drv, fvf);
                              //assign vertices
   mesh->vertex_buffer.SetVertices(verts, num_verts);
                              //assign faces
   mesh->index_buffer.SetFaces(faces, num_faces);
                              //assign fgroups
   mesh->SetFGroups(fgroups, num_fgroups);

   mesh->Release();

   return I3D_OK;
}

//----------------------------

bool I3D_visual_dynamic::ComputeBounds(){

   if(!mesh){
      bound.bound_local.bbox.Invalidate();
      bound.bound_local.bsphere.pos.Zero();
      bound.bound_local.bsphere.radius = 0.0f;
   }else{
      mesh->ComputeBoundingVolume();
      bound.bound_local = mesh->GetBoundingVolume();
   }
   return I3D_visual::ComputeBounds();
}

//----------------------------

void I3D_visual_dynamic::PrepareDestVB(I3D_mesh_base *mb, dword num_txt_sets){

   dword numv = mb->NumVertices1();
   if(!numv)
      return;

   envmap_uv_set = -1;
   vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);

   if(drv->NumSimultaneousTextures() > num_txt_sets){
                           //search for any env-mapped material
      for(int j=mb->NumFGroups1(); j--; ){
         PI3D_material mat = mb->GetFGroups1()[j].mat;
         if(mat->GetEnvironmetEffect() != EE_NULL){
            envmap_uv_set = num_txt_sets++;
            break;
         }
      }
   }
   dword mesh_fvf = mesh->vertex_buffer.GetFVFlags();

   if(drv->IsDirectTransform()){
      vertex_buffer.CreateD3DVB(drv, mesh_fvf, numv, true, true);

      assert(vertex_buffer.GetSizeOfVertex() == mb->vertex_buffer.GetSizeOfVertex());
                                 //copy data
      D3D_lock<byte> v_src(mb->vertex_buffer.GetD3DVertexBuffer(), mb->vertex_buffer.D3D_vertex_buffer_index,
         numv, mb->vertex_buffer.GetSizeOfVertex(), D3DLOCK_READONLY);
      D3D_lock<byte> v_dst(vertex_buffer.GetD3DVertexBuffer(), vertex_buffer.D3D_vertex_buffer_index, numv,
         vertex_buffer.GetSizeOfVertex(), 0);
      v_src.AssureLock();
      v_dst.AssureLock();
      memcpy(v_dst, v_src, numv*vertex_buffer.GetSizeOfVertex());
   }else{

      dword fvflags = num_txt_sets<<D3DFVF_TEXCOUNT_SHIFT;
                              //include diffuse only if normal, diffuse or alpha present
      if((mesh_fvf&D3DFVF_NORMAL) || (mesh_fvf&D3DFVF_DIFFUSE) || ((mesh_fvf&D3DFVF_POSITION_MASK)==D3DFVF_XYZB1)){
         fvflags |= D3DFVF_DIFFUSE;
      }
                              //create dest VB
      vertex_buffer.CreateD3DVB(drv, fvflags, numv, false);
   }

                              //prepare VS declaration
   {
      vector<dword> decl;
                              //always use stream 0
      decl.push_back(D3DVSD_STREAM(0));
                              //always must contain XYZ
      switch(mesh_fvf&D3DFVF_POSITION_MASK){
      case D3DFVF_XYZ: decl.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT3)); break;
      case D3DFVF_XYZB1:
         decl.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT3));
         decl.push_back(D3DVSD_REG(3, D3DVSDT_FLOAT1));
         break;
      case D3DFVF_XYZRHW:
         decl.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT4));
         break;
      default: assert(0);
      }
      if(mesh_fvf&D3DFVF_NORMAL)
         decl.push_back(D3DVSD_REG(2, D3DVSDT_FLOAT3));
      if(mesh_fvf&D3DFVF_DIFFUSE)
         decl.push_back(D3DVSD_REG(2, D3DVSDT_D3DCOLOR));
      if(mesh_fvf&D3DFVF_SPECULAR)
         decl.push_back(D3DVSD_REG(5, D3DVSDT_D3DCOLOR));
      switch((mesh_fvf&D3DFVF_TEXCOUNT_MASK)>>D3DFVF_TEXCOUNT_SHIFT){
      case 0: break;
      case 1: decl.push_back(D3DVSD_REG(1, D3DVSDT_FLOAT2)); break;
      default: decl.push_back(D3DVSD_REG(1, D3DVSDT_FLOAT4)); break;
      }
                              //terminator
      decl.push_back(D3DVSD_END());

      vs_decl = drv->RegisterVSDeclarator(decl);
   }
}

//----------------------------

I3D_RESULT I3D_visual_dynamic::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;
   if(frm->GetType1()!=FRAME_VISUAL)
      return I3D_frame::Duplicate(frm);

   PI3D_visual vis = I3DCAST_VISUAL(frm);
   mesh = NULL;
   I3D_mesh_base *mb = vis->GetMesh();
   if(mb){
      dword fvf = D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1;
      if(vis->GetVisualType()==visual_type)
         fvf = ((I3D_visual_dynamic*)vis)->mesh->vertex_buffer.GetFVFlags();
      mesh = new I3D_mesh_base(drv, fvf);
      mesh->Clone(mb);
      mesh->Release();
   }
   return I3D_visual::Duplicate(vis);
}

//----------------------------

I3D_RESULT I3D_visual_dynamic::DuplicateMesh(CPI3D_mesh_base mb, dword vc_flags){

   mesh = new I3D_mesh_base(drv, ConvertFlags(vc_flags));
   mesh->Clone(mb);
   mesh->Release();
   return I3D_OK;
}

//----------------------------

void I3D_visual_dynamic::AddPrimitives(S_preprocess_context &pc){

   if(!mesh)
      return;
                              //don't allow in shadow casting
   if(pc.mode==RV_SHADOW_CASTER)
      return;


   dword fvf = mesh->vertex_buffer.GetFVFlags();
   bool vertex_alpha = ((fvf&D3DFVF_POSITION_MASK)==D3DFVF_XYZB1);

   AddPrimitives1(mesh, pc, vertex_alpha);
}

//----------------------------

void I3D_visual_dynamic::DrawPrimitive(const S_preprocess_context &pc, const S_render_primitive &rp){

   drv->SetClipping(rp.clip);

   if(dest_vb_trans_id!=pc.render_view_id || drv->IsDirectTransform()){
      dest_vb_trans_id = pc.render_view_id;

      if(last_auto_lod != rp.curr_auto_lod){
         last_auto_lod = rp.curr_auto_lod;
         vis_flags &= ~(VISF_DEST_LIGHT_VALID | VISF_DEST_UV0_VALID);
      }
      dword vertex_count;
      if(rp.curr_auto_lod < 0)
         vertex_count = mesh->vertex_buffer.NumVertices();
      else
         vertex_count = mesh->GetAutoLODs()[rp.curr_auto_lod].vertex_count;

                              //transform vertices
      I3D_driver::S_vs_shader_entry_in se(vs_decl);

      dword prep_flags = VSPREP_TRANSFORM | VSPREP_FEED_MATRIX;
      if(pc.mode!=RV_SHADOW_CASTER){
         dword fvf = mesh->vertex_buffer.GetFVFlags();
         bool vertex_alpha = ((fvf&D3DFVF_POSITION_MASK)==D3DFVF_XYZB1);
                              //check if lighting requested
         if(vertex_alpha || drv->IsDirectTransform() || !(vis_flags&VISF_DEST_LIGHT_VALID)){
            if(fvf&D3DFVF_NORMAL){
                                 //use lighting engine to make lighting
               prep_flags |= VSPREP_MAKELIGHTING;
               if(vertex_alpha)
                  prep_flags |= VSPREP_LIGHT_MUL_ALPHA;
            }else
            if(fvf&D3DFVF_DIFFUSE){
               if(!vertex_alpha){
                              //plain copy of diffuse
                  se.AddFragment(I3D_driver::VSF_DIFFUSE_COPY);
               }else{
                  se.AddFragment(I3D_driver::VSF_DIFFUSE_MUL_ALPHA);
               }
            }else
            if((fvf&D3DFVF_POSITION_MASK)==D3DFVF_XYZB1)
               se.AddFragment(I3D_driver::VSF_COPY_ALPHA);
            if(drv->IsDirectTransform())
               prep_flags |= VSPREP_COPY_UV;
            else
               vis_flags |= VISF_DEST_LIGHT_VALID;
         }
      }else{
         prep_flags |= VSPREP_NO_ENVMAP | VSPREP_NO_FOG;
      }
      if(!(vis_flags&VISF_DEST_UV0_VALID)){
         prep_flags |= VSPREP_COPY_UV;
         vis_flags |= VISF_DEST_UV0_VALID;
      }

      PrepareVertexShader(mesh->GetFGroups1()[0].GetMaterial1(), se, rp, prep_flags);

      if(!drv->IsDirectTransform()){
         drv->DisableTextureStage(1);

         drv->SetStreamSource(mesh->vertex_buffer.GetD3DVertexBuffer(), mesh->vertex_buffer.GetSizeOfVertex());
         HRESULT hr;
         hr = drv->GetDevice1()->ProcessVertices(mesh->vertex_buffer.D3D_vertex_buffer_index,
            vertex_buffer.D3D_vertex_buffer_index, vertex_count,
            vertex_buffer.GetD3DVertexBuffer(), D3DPV_DONOTCOPYDATA);
         CHECK_D3D_RESULT("ProcessVertices", hr);
      }
   }
   DrawPrimitive1(mesh, pc, rp);
}

//----------------------------
//----------------------------

static struct S_auto_init_dynamic{
   static I3D_visual *CreateDynamicVisual(PI3D_driver drv){
      return new I3D_visual_dynamic(drv);
   }
   S_auto_init_dynamic(){
      RegisterVisual(CreateDynamicVisual, I3D_VISUAL_DYNAMIC, "Dynamic");
   }
} auto_init_dynamic;



//----------------------------
//----------------------------
