/*--------------------------------------------------------
   Copyright (c) 1999 - 2001 Lonely Cat Games
   All rights reserved.

   File: Volume.cpp
   Content: Volume frame and collision testing functions.
--------------------------------------------------------*/

#include "all.h"
#include "object.h"
#include "mesh.h"
#include "volume.h"
#include "scene.h"
#include "material.h"
#include "sector.h"
#include "texture.h"
#include "target.h"
#include "occluder.h"
#include "model.h"
#include "sng_mesh.h"
#include "dummy.h"
#include "bsp.h"    


#define VOLUME_STATISTICS     //collect statistics about collisions
#define USE_ASM

//#define SPHERE_VOLUME_DETECT_INSIDE    //detect line-sphere collision from inside of sphere
//#define BOX_VOLUME_DETECT_INSIDE    //detect line-box collision from inside of sphere

#if defined _DEBUG 
//#define PERFORMANECE_COUNT 1  //how many times to check cols in performance count mode, undef to disable
//#define PERFORMANECE_COUNT 100  //how many times to check cols in performance count mode, undef to disable
#endif

//----------------------------
                              //for occluder picking
static const S_vector unit_bbox[2] = {
   S_vector(-1, -1, -1), S_vector(1, 1, 1)
};

//----------------------------
                              //used to reset FPU in debugger - just step over the function
                              // and debugging may continue after FPU gets invalidated
void FPUReset(){
   __asm{
      push eax
      push ecx
      push edx                                         
   }
   _fpreset();
   __asm{
      pop edx
      pop ecx
      pop eax
   }
}

//----------------------------
                              //help structure passed to collision functions,
                              // containing all data about current test

                              //offsets due to inline assembly
#define S_pick_help__from_offset 0
#define S_pick_help__normalized_dir_offset 12
#define S_pick_help__closest_hit_offset 24


struct S_pick_help{
                              //input:
   S_vector from;
   S_vector normalized_dir;
   float closest_hit;
   dword flags;

   PI3D_line_collision_response cresp_line, cresp_texel;
   void *cresp_context;

   PI3D_frame closest_frm;
   S_vector hit_normal;

   void Reset(){
      closest_hit = 1e+16f;
      closest_frm = NULL;
   }
};

//----------------------------
//----------------------------

I3D_volume::I3D_volume(PI3D_driver d):
   I3D_frame(d),
   material_id(0),
   owner(NULL),
   volume_type(I3DVOLUME_NULL),
   vol_flags(VOLF_RESET),
   non_uniform_scale(S_vector(1.0f, 1.0f, 1.0f)),
   sphere_space_min(0), sphere_space_max(0),
   //sphere_space_min(-1), sphere_space_max(-1), //uninitialized
   last_min(-1), last_max(-1),
   hr_scene(NULL)
{
   type = FRAME_VOLUME;
   enum_mask = ENUMF_VOLUME;
   drv->AddCount(I3D_CLID_VOLUME);

   g_matrix.Identity();
}

//----------------------------

I3D_volume::~I3D_volume(){

   vol_flags |= VOLF_RESET;
   drv->DecCount(I3D_CLID_VOLUME);

   if(hr_scene){
      if(last_min != -1 && last_max != -1)
         hr_scene->DeleteSphereFromSpace(this, last_min, last_max);
      hr_scene->DestroyDynamicVolTree();
   }
}

//----------------------------

I3D_RESULT I3D_volume::Duplicate(CPI3D_frame frm){

   if(frm==this)
      return I3D_OK;

   if(frm->GetType1()==FRAME_VOLUME){
      PI3D_volume vp = I3DCAST_VOLUME(frm);     
      volume_type = vp->volume_type;
      owner = NULL;
      vol_flags = vp->vol_flags;
      hr_scene = vp->hr_scene;
      non_uniform_scale = vp->non_uniform_scale;
      material_id = vp->material_id;

      if(hr_scene){
         hr_scene->ResetDynamicVolTree();
      }
      vol_flags |= VOLF_RESET;
   }
   return I3D_frame::Duplicate(frm);
}

//----------------------------

void I3D_volume::SetOn(dword flg){

   if(flg != (frm_flags&(FRMFLAGS_ON|FRMFLAGS_ON_CONDITIONAL))){
      if(hr_scene && flg){
         hr_scene->ResetDynamicVolTree();
      }
      I3D_frame::SetOn(flg);
   }
}

//----------------------------

I3D_RESULT I3D_volume::SetScene(PI3D_scene scene){

   if(hr_scene!=scene){
      if(hr_scene){
                              //remove from scene which leaving
         if(last_min != -1 && last_max != -1){
            hr_scene->DeleteSphereFromSpace(this, last_min, last_max);
            last_min = last_max = -1; //invalidate 
         }
                              //reset tree
         hr_scene->ResetDynamicVolTree();
      }
      hr_scene = scene;
      if(hr_scene){
         hr_scene->ResetDynamicVolTree();
      }
   }
   return I3D_OK;
}

//----------------------------

/*
I3D_RESULT I3D_volume::SetHierarchical(bool b){

   assert(0);
   return I3DERR_UNSUPPORTED;
}
*/

//----------------------------

I3D_RESULT I3D_volume::SetNUScale(const S_vector &scl){

                              //we don't support negative scales
   if(scl.x < 0.0f || scl.y < 0.0f || scl.z < 0.0f)
      return I3DERR_INVALIDPARAMS;

   frm_flags &= ~FRMFLAGS_G_MAT_VALID;
   frm_flags |= FRMFLAGS_UPDATE_NEEDED;
   non_uniform_scale = scl;
   return I3D_OK;
}

//----------------------------

void I3D_volume::Update(){

   if(frm_flags&FRMFLAGS_UPDATE_NEEDED)
      Reset();
   I3D_frame::Update();
}

//----------------------------

void I3D_volume::Reset(){

   vol_flags |= VOLF_RESET;
   //if(vol_flags&VOLF_HIERARCHY)
   if((GetFrameFlags()&I3D_FRMF_STATIC_COLLISION))
      drv->voltree_stats.moving_static++;
#ifndef DISABLE_HIEARCHY_FOR_NON_STATIC
   if(hr_scene) 
      hr_scene->ResetDynamicVolTree();
#endif
}

//----------------------------

void I3D_volume::GetBoundBox(PI3D_bbox bbox1) const{

   int i;
   bbox1->Invalidate();

   switch(volume_type){
   case I3DVOLUME_SPHERE:
      for(i=0; i<3; i++){
         bbox1->min[i] = Min(bbox1->min[i], w_sphere.pos[i] - w_sphere.radius);
         bbox1->max[i] = Max(bbox1->max[i], w_sphere.pos[i] + w_sphere.radius);
      }
      break;
   default:
                              //cube have 8 corners, each bbox corner enlarges itself and 3 axes
      for(i=0; i<4; i++){
         S_vector vmin = bbox.min;
         S_vector vmax = bbox.max;
         if(i<3){
            vmin -= normal[i] * (d_min[i] + d_max[i]);
            vmax += normal[i] * (d_min[i] + d_max[i]);
         }
         for(int ii=0; ii<3; ii++){
            bbox1->min[ii] = Min(bbox1->min[ii], Min(vmin[ii], vmax[ii]));
            bbox1->max[ii] = Max(bbox1->max[ii], Max(vmax[ii], vmin[ii]));
         }
      }
   }
}

//----------------------------

const S_matrix &I3D_volume::GetGeoMatrix() const{

   if(ChangedMatrix() || !(frm_flags&FRMFLAGS_G_MAT_VALID)){
      if(GetParent1()) matrix.Make(f_matrix, GetParent1()->GetMatrix());
      else matrix = f_matrix;
      frm_flags &= ~(FRMFLAGS_G_MAT_VALID | FRMFLAGS_CHANGE_MAT | FRMFLAGS_BSPHERE_TRANS_VALID);
   }
   if(!(frm_flags&FRMFLAGS_G_MAT_VALID)){
      g_matrix(0) = matrix(0) * non_uniform_scale[0];
      g_matrix(1) = matrix(1) * non_uniform_scale[1];
      g_matrix(2) = matrix(2) * non_uniform_scale[2];
      g_matrix(3) = matrix(3);
      frm_flags |= FRMFLAGS_G_MAT_VALID;
   }
   return g_matrix;
}

//----------------------------

inline void I3D_volume::SetupInSphereSpace() const{

   int curr_min = FLOAT_2_SPHERE_SPACE(sphere_space_min);
   int curr_max = FLOAT_2_SPHERE_SPACE(sphere_space_max);
   //assert(hr_scene);
   if(hr_scene){
                           //fast quit if nothing changed for this sphere
      if((last_min != curr_min) || (last_max != curr_max)){
                              //do not delete uninitialised volume
         if((last_min != -1) && (last_max != -1)){ 
            hr_scene->DeleteSphereFromSpace(this, last_min, last_max);
         }
         hr_scene->AddSphereToSpace(this, curr_min, curr_max);
      }
   }
}

//----------------------------

bool I3D_volume::Prepare() const{

   if(vol_flags&VOLF_RESET){
      int i;

      switch(volume_type){

      case I3DVOLUME_SPHERE:
         {
            const S_matrix &m = GetMatrix();
            w_sphere.pos = m(3);
            sphere_pos_dot = w_sphere.pos.Dot(w_sphere.pos);
            w_sphere.radius = m(0).Magnitude();
            radius_2 = w_sphere.radius * w_sphere.radius;
            bbox.Invalidate();
            for(i=0; i<3; i++){
               bbox.min[i] = Min(bbox.min[i], w_sphere.pos[i] - w_sphere.radius);
               bbox.max[i] = Max(bbox.max[i], w_sphere.pos[i] + w_sphere.radius);
            }
#ifndef ENABLE_SPHERE_SPACE //in case we disable sphere space, calculate values for sinle sphere speed up
            float pos_dist = IsAbsMrgZero(sphere_pos_dot) ? .0f : SqrtFast(sphere_pos_dot);
            sphere_space_min = Max(.0f, pos_dist - w_sphere.radius);
            sphere_space_max = Max(.0f, pos_dist + w_sphere.radius);
#endif
         }
         break;

      case I3DVOLUME_BOX:
         {
            const S_matrix &m =
               (volume_type==I3DVOLUME_BOX) ? GetGeoMatrix() : I3DCAST_VISUAL(GetParent())->GetMatrix();
                              //get bounding box and normals for 3 sides
            for(i=0; i<3; i++)
               normal[i] = m(i);

            for(i=0; i<3; i++){
               side[i] = normal[i] * 2.0f;
               side_dot[i] = side[i].Dot(side[i]);
               side_size_2[i] = side[i].Magnitude();
               side_size_2[i] *= side_size_2[i];
            }
            bbox.max = normal[0] + normal[1] + normal[2];
            bbox.min = -bbox.max;
            bbox.min += m(3);
            bbox.max += m(3);
            for(i=0; i<3; i++){
               side_dot[i] = side[i].Dot(side[i]);
               side_size_2[i] = side[i].Magnitude();
               side_size_2[i] *= side_size_2[i];

               normal[i].Normalize();
               d_min[i] = normal[i].Dot(bbox.min);
               d_max[i] = -normal[i].Dot(bbox.max);
            }
#if 0
                              //debug output
            {
               hr_scene->DEBUG_CODE(0, (dword)&bbox);

               S_vector vx[2]={ S_vector(0, 0, 0), normal[0]};
               hr_scene->DEBUG_CODE(0, (dword)vx);

               S_vector vy[2]={ S_vector(0, 0, 0), normal[1]};
               hr_scene->DEBUG_CODE(0, (dword)vy);

               S_vector vz[2]={ S_vector(0, 0, 0), normal[2]};
               hr_scene->DEBUG_CODE(0, (dword)vz);
            }
#endif
         }
                              //make bounding sphere from bounding box
         {
            S_vector mag_half = (bbox.max-bbox.min)*.5f;
            w_sphere.radius = mag_half.Magnitude();
            radius_2 = w_sphere.radius * w_sphere.radius;
            w_sphere.pos = bbox.min + mag_half;
            sphere_pos_dot = w_sphere.pos.Dot(w_sphere.pos);
         }
         break;

      case I3DVOLUME_RECTANGLE:
         {
            const S_matrix &m = GetGeoMatrix();
                                 //get bounding box and normal for 3 sides
            for(i=0; i<2; i++){
               normal[i] = m(i);
               side[i] = normal[i] * 2.0f;
               side_dot[i] = side[i].Dot(side[i]);
               side_size_2[i] = side[i].Magnitude();
               side_size_2[i] *= side_size_2[i];
            }
                              //z-normal is computed - avoiding non-uniform scale squashing
            normal[2].GetNormal(S_vector(0.0f, 0.0f, 0.0f), normal[0], normal[1]);


                              //z-axis is ignored
            bbox.max = normal[0] + normal[1];
            bbox.min = -bbox.max;
            bbox.min += m(3);
            bbox.max += m(3);

            for(i=0; i<3; i++){
               normal[i].Normalize();
               d_min[i] = normal[i].Dot(bbox.min);
               d_max[i] = -normal[i].Dot(bbox.max);
            }

#if 0
            {
               hr_scene->DEBUG_CODE(0, (dword)&bbox);

               S_vector vx[2]={ S_vector(0, 0, 0), normal[0]};
               hr_scene->DEBUG_CODE(0, (dword)vx);

               S_vector vy[2]={ S_vector(0, 0, 0), normal[1]};
               hr_scene->DEBUG_CODE(0, (dword)vy);

               S_vector vz[2]={ S_vector(0, 0, 0), normal[2]};
               hr_scene->DEBUG_CODE(0, (dword)vz);
            }
#endif
         }
                              //make bounding sphere from bounding box
         {
            S_vector mag_half = (bbox.max-bbox.min)*.5f;
            w_sphere.radius = mag_half.Magnitude();
            radius_2 = w_sphere.radius * w_sphere.radius;
            w_sphere.pos = bbox.min + mag_half;
            sphere_pos_dot = w_sphere.pos.Dot(w_sphere.pos);
         }
         break;
      default: 
         return true;
      }
                              //sphere space support
#ifdef ENABLE_SPHERE_SPACE
      if(!(frm_flags&I3D_FRMF_STATIC_COLLISION)){
         assert(!(vol_flags&VOLF_AUTO_GEN) || !"bound volume can't be in sphere_space - causing mutliple collisions");
         float pos_dist = IsAbsMrgZero(sphere_pos_dot) ? .0f : SqrtFast(sphere_pos_dot);
         sphere_space_min = Max(.0f, pos_dist - w_sphere.radius);
         sphere_space_max = Max(.0f, pos_dist + w_sphere.radius);
         SetupInSphereSpace();
      }
#endif

      vol_flags &= ~VOLF_RESET;
   }
   return true;
}

//----------------------------

#define L_F_DEBUG(n)
//#define L_F_DEBUG(n) drv->DEBUG(n)

//----------------------------
//----------------------------

static const word bbox_list[] = {
   0, 1, 2, 3, 4, 5, 6, 7,
   0, 2, 1, 3, 4, 6, 5, 7,
   0, 4, 1, 5, 2, 6, 3, 7
}, brect_list[]={
   0, 1, 2, 3, 0, 2, 1, 3
};

                              //sphere values
#define SPHERE_NUM_VERTS 24

static S_vector sphere_circle[3][SPHERE_NUM_VERTS];
static word sphere_c_list[3][SPHERE_NUM_VERTS*2];

static struct S_auto_init{
   S_auto_init(){
                              //init volume sphere
      for(int j=0; j<3; j++){
         for(int i=0; i<SPHERE_NUM_VERTS; i++){
            sphere_c_list[j][i*2+0]=SPHERE_NUM_VERTS*j+i+0;
            sphere_c_list[j][i*2+1]=SPHERE_NUM_VERTS*j+i+1;

            float s = sin(GetPI2()*i/SPHERE_NUM_VERTS),
                  c = cos(GetPI2()*i/SPHERE_NUM_VERTS);

            sphere_circle[j][i].x = (j ? s : 0.0f);
            sphere_circle[j][i].y = (!j ? s : j==2 ? c : 0.0f);
            sphere_circle[j][i].z = ((j<2) ? c : 0.0f);
         }
         sphere_c_list[j][i*2-1] = SPHERE_NUM_VERTS*j;
      }
   }
} auto_init;

//----------------------------

static const S_vector box_verts[] = {
   S_vector(-1, -1, -1),
   S_vector(1, -1, -1),
   S_vector(-1, 1, -1),
   S_vector(1, 1, -1),
   S_vector(-1, -1, 1),
   S_vector(1, -1, 1),
   S_vector(-1, 1, 1),
   S_vector(1, 1, 1),
};

static const word box_fill_indicies[] = {
   1, 3, 7, 1, 7, 5,
   0, 4, 6, 0, 6, 2,
   2, 6, 7, 2, 7, 3,
   0, 1, 5, 0, 5, 4,
   0, 2, 3, 0, 3, 1,
   5, 7, 6, 5, 6, 4
};

static const word box_line_indicies[] = {
   0, 1, 0, 2, 0, 4, 1, 3, 1, 5, 2, 3, 2, 6,
   3, 7, 4, 5, 4, 6, 5, 7, 6, 7
};

//----------------------------

static void DrawVolBox(const S_vector &color, PI3D_scene scene, byte alpha, bool edges = true){

   if(edges){
      scene->DrawLines(box_verts, 12, box_line_indicies, sizeof(box_line_indicies)/sizeof(word),
         color, sizeof(S_vector), alpha);
   }
   scene->DrawTriangles(box_verts, 12, box_fill_indicies, sizeof(box_fill_indicies)/sizeof(word),
      color, sizeof(S_vector), alpha);
}

//----------------------------

static void DrawVolBox1(PI3D_scene scene, const I3D_bbox &bbox,
   const LPS_vector normal, const float *d_min,
   const float *d_max, const S_vector &color, byte opt){

   scene->SetRenderMatrix(GetIdentityMatrix());
   S_vector bbox1[8];
   static const byte amap[4] = {0, 1, 2, 4};
   for(int ii=0; ii<4; ii++){
      bbox1[amap[ii]] = bbox.min;
      bbox1[7-amap[ii]] = bbox.max;
      if(ii>0){
         bbox1[amap[ii]] -= normal[ii-1] * (d_min[ii-1] + d_max[ii-1]);
         bbox1[7-amap[ii]] += normal[ii-1] * (d_min[ii-1] + d_max[ii-1]);
      }
   }
   scene->DrawLines(bbox1, 8, bbox_list, sizeof(bbox_list)/sizeof(word),
      color, sizeof(S_vector), opt);
}

//----------------------------

static const S_vector rect_edges[4] = {
   S_vector(-1.0f, -1.0f, 0.0f),
   S_vector( 1.0f, -1.0f, 0.0f),
   S_vector( 1.0f,  1.0f, 0.0f),
   S_vector(-1.0f,  1.0f, 0.0f),
};

static const word rect_fill_indicies[] = {
   0, 1, 2, 0, 2, 3, 0, 2, 1, 0, 3, 2
};

static const word rect_line_indicies[] = {
   0, 1, 1, 2, 2, 3, 3, 0
};

//----------------------------
static const S_vector draw_colors[2][2] = { //[auto_gen][trigger]
   S_vector(.7f, .3f, .3f), S_vector(1.0f, .5f, .5f),
   S_vector(.3f, .7f, .7f), S_vector(.5f, 1.0f, 1.0f)
};

void I3D_volume::Draw1(PI3D_scene scene, bool strong, bool force_bsp, const S_vector *force_color) const{

   S_vector color;
   byte alpha;
   bool auto_gen = (vol_flags&VOLF_AUTO_GEN);
   if(force_bsp){
      if(auto_gen) return;
      assert(drv->GetFlags() &DRVF_DEBUGDRAWBSP);
      static const S_vector bsp_clr(.0f, .0f, 1.0f);
      color = bsp_clr;
      alpha = 0xa0;
   }else
   if(force_color){
      color = *force_color;
      alpha = 0x40;
   }else{
      color = strong ? S_vector(1.0f, 0.0f, 0.0f) :
         draw_colors[auto_gen][bool(vol_flags&VOLF_DEBUG_TRIGGER)];
      vol_flags &= ~VOLF_DEBUG_TRIGGER;

      if(frm_flags&I3D_FRMF_STATIC_COLLISION)
         color += S_vector(.0f,.3f, .0f);
      alpha = 0xc0;
   }

   drv->SetTexture(NULL);

   switch(volume_type){

   case I3DVOLUME_SPHERE:
      scene->SetRenderMatrix(GetMatrix());
      scene->DrawLines(sphere_circle[0], SPHERE_NUM_VERTS*3, sphere_c_list[0], SPHERE_NUM_VERTS*2*3, color, sizeof(S_vector));
      break;

   case I3DVOLUME_BOX:
      if(!auto_gen){
         scene->SetRenderMatrix(GetGeoMatrix());
         DrawVolBox(color, scene, alpha, !force_bsp);
      }else{
         //if(vol_flags&VOLF_AUTO_USE_SPHERE){
            S_matrix m;
            m.Zero();
            m(0, 0) = w_sphere.radius;
            m(1, 1) = w_sphere.radius;
            m(2, 2) = w_sphere.radius;
            m(3) = w_sphere.pos;
            m(3, 3) = 1.0f;
            scene->SetRenderMatrix(m);
            scene->DrawLines(sphere_circle[0], SPHERE_NUM_VERTS*3, sphere_c_list[0], SPHERE_NUM_VERTS*2*3, color, sizeof(S_vector), 0xc0);
         //}else
            DrawVolBox1(scene, bbox, normal, d_min, d_max, color, alpha);
      }
      break;
   case I3DVOLUME_RECTANGLE:
      {
         //drv->SetWorldMatrix(GetMatrix());
         //drv->DrawLines(sphere_circle[0], SPHERE_NUM_VERTS*3, sphere_c_list[0], SPHERE_NUM_VERTS*2*3, color, sizeof(S_vector));

         scene->SetRenderMatrix(GetGeoMatrix());
         scene->DrawTriangles(rect_edges, 4, rect_fill_indicies, sizeof(rect_fill_indicies)/sizeof(word),
            color, sizeof(S_vector), alpha);
         if(!force_bsp){
            scene->DrawLines(rect_edges, 4, rect_line_indicies, sizeof(rect_line_indicies)/sizeof(word),
               color, sizeof(S_vector));
         }
      }
      break;
   }
}

//----------------------------

I3D_RESULT I3D_volume::DebugDraw(PI3D_scene scene) const{

   Prepare();
   Draw1(scene, true);
   return I3D_OK;
}

//----------------------------
//----------------------------

#define VOLUME_BOUND_LIMIT 3

/*--------------------------------------------------------
   Geometry collision testing
--------------------------------------------------------*/

#if defined _MSC_VER && defined USE_ASM

inline bool __stdcall pickBSphereTest(const I3D_bsphere &bs1, const S_pick_help &pick_data, float *hit_dist_ret = NULL){

   float tmp_var;
   __asm{
                              //ebx = bs1
                              //edi = from
      mov ebx, bs1
      mov edi, pick_data
                              //u = (bs1.pos-pick_data.from).Dot(pick_data.normalized_dir);
      fld dword ptr[ebx + 0];
      fsub dword ptr [edi + S_pick_help__from_offset + 0]
      fmul dword ptr[edi + S_pick_help__normalized_dir_offset + 0]
      fld dword ptr[ebx + 4];
      fsub dword ptr [edi + S_pick_help__from_offset + 4]
      fmul dword ptr[edi + S_pick_help__normalized_dir_offset + 4]
      faddp st(1), st
      fld dword ptr[ebx + 8];
      fsub dword ptr [edi + S_pick_help__from_offset + 8]
      fmul dword ptr[edi + S_pick_help__normalized_dir_offset + 8]
      faddp st(1), st

                              //if(u < - bs1.radius || u > (pick_data.closest_hit+bs1.radius))
                              // return false;
      fld dword ptr[ebx + 12]
      fchs
      fcomp st(1)
      fnstsw ax
      test ah, 0x41
      jz out1
      fld dword ptr[edi + S_pick_help__closest_hit_offset]
      fadd dword ptr[ebx + 12]
      fcomp st(1)
      fnstsw ax
      test ah, 1
      jnz out1
                              //determine closest point to line

                              //S_vector point_on_line = pick_data.from + pick_data.normalized_dir * u;
                              //S_vector dir = point_on_line - bs1.pos;
                              //float dist_2 = dir.Dot(dir);
      fld dword ptr[edi + S_pick_help__normalized_dir_offset + 0]
      fmul st, st(1)
      fadd dword ptr [edi + S_pick_help__from_offset + 0]
      fsub dword ptr[ebx + 0];
      fmul st, st(0)

      fld dword ptr[edi + S_pick_help__normalized_dir_offset + 4]
      fmul st, st(2)
      fadd dword ptr [edi + S_pick_help__from_offset + 4]
      fsub dword ptr[ebx + 4];
      fmul st, st(0)
      faddp st(1), st

      fld dword ptr[edi + S_pick_help__normalized_dir_offset + 8]
      fmul st, st(2)
      fadd dword ptr [edi + S_pick_help__from_offset + 8]
      fsub dword ptr[ebx + 8];
      fmul st, st(0)
      faddp st(1), st

                              //if(dist_2 >= (bs1.radius*bs1.radius))
      fld dword ptr[ebx + 12]
      fmul st, st(0)
      fcomp st(1)
      fnstsw ax
      fstp tmp_var
      test ah, 0x41
      jnz out1
      fstp tmp_var
   }
   return true;
out1:
   __asm fstp tmp_var
//fail:
   return false;
}

#else //USE_ASM

inline bool pickBSphereTest(const I3D_bsphere &bs1, const S_pick_help &pick_data, float *hit_dist_ret = NULL){

                              //check closest point on line
   float u = (bs1.pos-pick_data.from).Dot(pick_data.normalized_dir);
   if(u < - bs1.radius || u > (pick_data.closest_hit+bs1.radius))
      return false;

                              //determine closest point to line
   S_vector point_on_line = pick_data.from + pick_data.normalized_dir * u;
   S_vector dir = point_on_line - bs1.pos;
   float dist_2 = dir.Dot(dir);
   if(dist_2 >= (bs1.radius*bs1.radius))
      return false;

   return true;
}

#endif   //!USE_ASM

//----------------------------

static I3DENUMRET I3DAPI cbEnumSMJoints(PI3D_frame frm, dword dw_pick_data){

   S_pick_help &pick_data = *(S_pick_help*)dw_pick_data;
   PI3D_joint jnt = I3DCAST_JOINT(frm);
   if(!jnt->IsVolumeBoxValid())
      return I3DENUMRET_OK;

                              //make test on bounding-box
   const I3D_bbox &bbox = jnt->GetVolumeBox1();
   const S_matrix &mat = jnt->GetVolumeBoxMatrix1();

   float hit_dist;
   bool b = pickBBoxTest(mat * jnt->GetMatrix(), bbox, pick_data.from, pick_data.normalized_dir, pick_data.closest_hit, &hit_dist);
   if(b){
                              //call-back user
      if(pick_data.cresp_line){
         if(!pick_data.cresp_line(S_vector(0, 0, 0), hit_dist, frm, pick_data.cresp_context, NULL))
            return I3DENUMRET_OK;
      }
      pick_data.closest_hit = hit_dist;
      pick_data.closest_frm = frm;
      pick_data.hit_normal.Zero();
   }

   return I3DENUMRET_OK;
}

//----------------------------
                              //perform ray collision test directly on geometry,
                              // higher bounding tests already done
void I3D_scene::cbPick(PI3D_visual vis, PI3D_frame pick_frm, I3D_mesh_base *mb, S_pick_help &pick_data){

   bool collision_ok = false;
   S_vector loc_hit_normal;   //collision normal in visual's local coords
   bool loc_hit_normal_inv;
                              
   int num_fgroups = mb->NumFGroups1();
   if(!num_fgroups)
      return;
   
   const S_matrix &matrix = vis->matrix;
   const S_matrix &m_inv = vis->GetInvMatrix1();

                              //inverse-map vectors to local coords
   S_vector loc_from = pick_data.from * m_inv;
   S_vector loc_to = (pick_data.from + pick_data.normalized_dir * pick_data.closest_hit) * m_inv;
   S_vector loc_dir_normalized = loc_to - loc_from;
   float loc_hit_dist = loc_dir_normalized.Magnitude();
   if(IsMrgZero(loc_hit_dist))
      return;
   loc_dir_normalized /= loc_hit_dist;

   const S_vector *pick_verts = (const S_vector*)mb->vertex_buffer.Lock(D3DLOCK_READONLY);
   dword vstride = mb->vertex_buffer.GetSizeOfVertex();
                              //assumption: texture coordinates are at end of vertex
   const I3D_text_coor *txt_coords = (const I3D_text_coor*)(((byte*)pick_verts) + vstride - sizeof(I3D_text_coor));

   PI3D_triface faces = mb->index_buffer.Lock(D3DLOCK_READONLY);
                              //check all face groups
   const I3D_face_group *fgrps = mb->GetFGroups1();
   for(int ii=0; ii<num_fgroups; ii++){
      const I3D_face_group &fg = fgrps[ii];
      PI3D_material mat = fg.GetMaterial1();
                              //check all faces in group
      bool two_side_mat = (mat->Is2Sided1() || (pick_data.flags&I3DGEOCOL_FORCE2SIDE));

      for(int jj = 0; jj < fg.NumFaces1(); jj++, ++faces){
         bool invert_side = false;

         const I3D_triface &fc = *faces;
         const S_vector &v0 = *(const S_vector*)(((byte*)pick_verts) + fc[0] * vstride);
         const S_vector &v1 = *(const S_vector*)(((byte*)pick_verts) + fc[1] * vstride);
         const S_vector &v2 = *(const S_vector*)(((byte*)pick_verts) + fc[2] * vstride);
         S_plane pl;
         pl.normal.GetNormal(v0, v1, v2);
                              //don't normalize - it's working also without that
         //pl.normal.Normalize();
         pl.d = -v0.Dot(pl.normal);

         float f_dist = loc_from.DistanceToPlane(pl);
         L_F_DEBUG(C_fstring("distance of 'from' of plane: %.2f", f_dist));
         if(CHECK_ZERO_LESS(f_dist)){
                              //fast quit 1 - backface culling
            if(!two_side_mat){
               L_F_DEBUG("Fast quit 1 - single-side material, can't be hit");
               continue;
            }
            invert_side = true;
         }

         float u = pl.normal.Dot(loc_dir_normalized);
                              //detect parallel
         if(IsAbsMrgZero(u)){
            L_F_DEBUG("Fast quit 2 - parallel with plane");
            continue;
         }

         float curr_dist = - f_dist / u;
         L_F_DEBUG(C_fstring("point on plane at length %.2f", curr_dist));
                              //don't look back
         if(CHECK_ZERO_LESS(curr_dist)){
            continue;
         }
                              //reject more far collisions than
                              //what we already have
         if(FLOAT_BITMASK(loc_hit_dist) <= FLOAT_BITMASK(curr_dist)){
            L_F_DEBUG("Fast quit 3 - line segment off the plane");
            continue;
         }
         S_vector point_on_plane = loc_from + loc_dir_normalized * curr_dist;

         if(!IsPointInTriangle(point_on_plane, v0, v1, v2, pl.normal))
            continue;

                           //face collision succeeded
         if(pick_data.flags&I3DGEOCOL_TEXTURE){
                              //detailed texture check
            PI3D_texture_base tb = mat->GetTexture1();
            if(tb && (mat->I3D_material::IsCkey() || mat->IsCkeyAlpha1() || pick_data.cresp_texel)){
               float u1, u2;

               S_vector edge_dir = v2 - v1;

               S_vector p_dir = point_on_plane - v0;
               LineIntersection(v0, p_dir, v1, edge_dir, u1, u2);
               if(!IsAbsMrgZero(u1)){
                  S_vector pp = v0 + p_dir * u1;
                              //get uv on edge point
                  float e_len = edge_dir.Sum();
                  if(!IsAbsMrgZero(e_len)){
                     u2 = (pp - v1).Sum() / e_len;
                     I3D_text_coor tex;

                     const I3D_text_coor &uv0 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[0] * vstride);
                     const I3D_text_coor &uv1 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[1] * vstride);
                     const I3D_text_coor &uv2 = *(const I3D_text_coor*)(((byte*)txt_coords) + fc[2] * vstride);

                     tex.u = uv1.u + (uv2.u - uv1.u) * u2;
                     tex.v = uv1.v + (uv2.v - uv1.v) * u2;

                     u1 = 1.0f / u1;
                     tex.u = uv0.u + (tex.u - uv0.u) * u1;
                     tex.v = uv0.v + (tex.v - uv0.v) * u1;

                     tex.u = fmod(tex.u, 1.0f);
                     tex.v = fmod(tex.v, 1.0f);
                     if(tex.u<0.0f) tex.u += 1.0f;
                     if(tex.v<0.0f) tex.v += 1.0f;

                     if(mat->I3D_material::IsCkey() || mat->IsCkeyAlpha1()){
                                    //look on texel
                        const byte *rle = tb->GetRLEMask();
                        assert(rle);
                        dword sx = tb->SizeX1(), sy = tb->SizeY1();
                        int x = Min((int)(sx * tex.u), (int)sx-1);
                        int y = Min((int)(sy * tex.v), (int)sy-1);
                        const byte *mem = &rle[*(dword*)(&rle[y*4])];
                        bool full = false;
                        do{
                           byte b = *mem++;
                           if(!b){     //eol
                              full = false;
                              break;
                           }
                           full = (b&0x80);
                           x -= (b&0x7f);
                        }while(x>=0);
                        if(!full)
                           continue;
                     }
                                             //ask user if we furher process this
                     if(pick_data.cresp_texel){
                        I3D_texel_collision tc = {tex.u, tex.v, tb, mat};
                                             //due to this, we need to put hit data
                                             // into world coords,
                                             // performance is not critical, because this
                                             // feature is expected to be called bu light-mapping only
                        S_vector hit_normal = pl.normal.RotateByMatrix(matrix);
                        if(invert_side)
                           hit_normal.Invert();
                        S_vector col_pt = (loc_from + loc_dir_normalized * curr_dist) * matrix;
                        float hit_dist = (col_pt - pick_data.from).Magnitude();
                        if(!pick_data.cresp_texel(hit_normal, hit_dist, pick_frm, pick_data.cresp_context, &tc)){
                           mb->vertex_buffer.Unlock();
                           mb->index_buffer.Unlock();
                           return;
                        }
                     }
                  }
               }
            }
         }
                              //collision detected on this face, save data
         loc_hit_dist = curr_dist;
         collision_ok = true;
         loc_hit_normal = pl.normal;
         loc_hit_normal_inv = invert_side;
      }
   }
   mb->vertex_buffer.Unlock();
   mb->index_buffer.Unlock();

   if(collision_ok){
      S_vector hit_normal = loc_hit_normal.RotateByMatrix(matrix);
      if(loc_hit_normal_inv)
         hit_normal.Invert();

      S_vector col_pt = (loc_from + loc_dir_normalized * loc_hit_dist) * matrix;
      float hit_dist = (col_pt - pick_data.from).Magnitude();
                              //check if we want this                        
      if(pick_data.cresp_line){
         if(!pick_data.cresp_line(hit_normal, hit_dist, pick_frm, pick_data.cresp_context, NULL))
            return;
      }
                              //map collision distance back to world coords
      pick_data.closest_hit = hit_dist;
                              //map collision normal to world coords
      pick_data.hit_normal = hit_normal;
 
      pick_data.closest_frm = pick_frm;
      assert(hit_dist>=0.0f);
   }
}

//----------------------------

I3DENUMRET I3D_scene::cbGeoCol(PI3D_frame root, S_pick_help &pick_data){

   const vector<PI3D_frame> &children = root->GetChildren1();

                              //call on all children of root
   const PI3D_frame *child_p = children.begin();
   for(int i=children.size(); i--; ){
      PI3D_frame frm = child_p[i];

      if(frm->GetType1()!=FRAME_VISUAL){

         switch(frm->GetType1()){
         case FRAME_MODEL:
            if((!(pick_data.flags&I3DGEOCOL_INVISIBLE) && !frm->IsOn1()) ||
               (frm->GetFrameFlags()&I3D_FRMF_NOCOLLISION))
               continue;

                              //hierarchy test on model
            if(frm->num_hr_vis){
               PI3D_model mod = I3DCAST_MODEL(frm);
               if(!(frm->GetFrameFlags()&FRMFLAGS_HR_BOUND_VALID))
                  mod->hr_bound.MakeHierarchyBounds(frm);
               bool b = pickBSphereTest(mod->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID), pick_data);
               if(b)
                  b = pickBBoxTest(mod->GetMatrix(), mod->hr_bound.bound_local.bbox, pick_data.from, pick_data.normalized_dir, pick_data.closest_hit);
               if(!b) continue;
            }
            break;
         case FRAME_DUMMY:
            if((!(pick_data.flags&I3DGEOCOL_INVISIBLE) && !frm->IsOn1()) ||
               (frm->GetFrameFlags()&I3D_FRMF_NOCOLLISION))
               continue;
            break;
         }
      }else{
         if((!(pick_data.flags&I3DGEOCOL_INVISIBLE) && !frm->IsOn1()) ||
            (frm->GetFrameFlags()&I3D_FRMF_NOCOLLISION))
            continue;

         PI3D_visual vis = I3DCAST_VISUAL(frm);

         switch(vis->GetVisualType1()){
         case I3D_VISUAL_BILLBOARD:
            break;
                              //flow...
         default:
            {
               I3D_mesh_base *mb = vis->GetMesh();
               if(!mb)
                  break;
                              //make sure all matrices are computed
               const S_matrix &matrix = frm->GetMatrix();
                              //hierarchy bound test
               if(frm->num_hr_vis){
                  if(!(frm->GetFrameFlags()&FRMFLAGS_HR_BOUND_VALID))
                     vis->hr_bound.MakeHierarchyBounds(vis);

                  bool b = pickBSphereTest(vis->hr_bound.GetBoundSphereTrans(vis, FRMFLAGS_HR_BSPHERE_VALID),
                     pick_data);
                  if(b)
                     b = pickBBoxTest(matrix, vis->hr_bound.bound_local.bbox, pick_data.from, pick_data.normalized_dir, pick_data.closest_hit);
                  if(!b)
                     continue;
               }
               bool b = true;
               {
                                 //perform check on bounding volume
                  if(!(vis->vis_flags&VISF_BOUNDS_VALID))
                     vis->ComputeBounds();
                  b = pickBSphereTest(vis->bound.GetBoundSphereTrans(vis, FRMFLAGS_BSPHERE_TRANS_VALID), pick_data);
                  if(b)
                     b = pickBBoxTest(vis->matrix, mb->GetBoundingVolume().bbox, pick_data.from, pick_data.normalized_dir, pick_data.closest_hit);
               }
               if(b){
                  cbPick(vis, vis, mb, pick_data);
               }
            }
            break;

         case I3D_VISUAL_SINGLEMESH:
            {
               PI3D_object_singlemesh sm = (PI3D_object_singlemesh)vis;
               if(sm->IsVolumeBoxValid()){
                              //make test on bounding-box
                  const I3D_bbox &bbox = sm->GetVolumeBox1();
                  const S_matrix &mat = sm->GetVolumeBoxMatrix1();

                  float hit_dist;
                  bool b = pickBBoxTest(mat * vis->GetMatrix(), bbox, pick_data.from, pick_data.normalized_dir, pick_data.closest_hit, &hit_dist);
                  if(b){
                              //call-back user
                     if(pick_data.cresp_line){
                        if(!pick_data.cresp_line(S_vector(0, 0, 0), hit_dist, frm, pick_data.cresp_context, NULL))
                           goto skip;
                     }
                     pick_data.closest_hit = hit_dist;
                     pick_data.closest_frm = frm;
                     pick_data.hit_normal.Zero();
               skip:{}
                  }
               }
               vis->EnumFrames(cbEnumSMJoints, (dword)&pick_data, ENUMF_JOINT);
            }
            break;
         }
      }
                              //repeat on children
      if(frm->num_hr_vis)
         cbGeoCol(frm, pick_data);
   }
   return I3DENUMRET_OK;
}

//----------------------------

PI3D_frame I3D_scene::TestColGeoInternal(const S_vector &from,
   const S_vector &dir, dword flags, float *ret_dist,
   S_vector *ret_normal, PI3D_line_collision_response cresp, 
   void *cb_context, PI3D_line_collision_response textel_resp) const{

   S_pick_help pick_data;

   pick_data.Reset();
   pick_data.from = from;
   pick_data.normalized_dir = dir;
   pick_data.normalized_dir.Normalize();
   if(!(flags&I3DGEOCOL_RAY))
      pick_data.closest_hit = dir.Magnitude();
   pick_data.flags = flags;
   pick_data.cresp_line = cresp;
   pick_data.cresp_texel = textel_resp;
   pick_data.cresp_context = cb_context;

#ifdef PERFORMANECE_COUNT
   BegProf();
   for(int i=PERFORMANECE_COUNT; i--; ){
   cbGeoCol(I3D_scene::GetPrimarySector1(), pick_data);
   }
   drv->DEBUG(EndProf());
#else
   cbGeoCol(I3D_scene::GetPrimarySector1(), pick_data);
#endif

   if(flags&I3DGEOCOL_EDIT_FRAMES){

                              //try picking non-visuals 
      I3D_collision_data cd(from, dir, flags, NULL);
      cd.cresp_line_old = pick_data.cresp_line;
      cd.cresp_context = pick_data.cresp_context;
      cd.SetReturnValues(pick_data.closest_hit, pick_data.closest_frm, &pick_data.hit_normal);
      if(TestPickEditFrames(cd, pick_data.closest_hit)){
         pick_data.closest_hit = cd.GetClosestHit();
         pick_data.closest_frm  = cd.GetHitFrm();
         pick_data.hit_normal = cd.GetHitNormal();
      }
   }

   if(pick_data.closest_frm){
      if(ret_dist) *ret_dist = pick_data.closest_hit;
      if(ret_normal){
         *ret_normal = pick_data.hit_normal;
         ret_normal->Normalize();
      }
   }

   return pick_data.closest_frm;
}

//----------------------------

static const S_vector V_ZERO(0.0f, 0.0f, 0.0f);

//----------------------------

void I3D_scene::cbEnumSMandJoints(PI3D_frame root, I3D_collision_data &cd){

   const vector<PI3D_frame> &children = root->GetChildren1();
   PI3D_frame const *chp = children.begin();
   for(int i=children.size(); i--; ){
      PI3D_frame frm = *chp++;
#ifdef USE_PREFETCH
      if(i){
         Prefetch1(*chp);
      }
#endif

                              //check if this frame is ignored
      if(frm == cd.frm_ignored)
         continue;

                              // ...or invisible
      if(!frm->IsOn1()){
         if(!(cd.flags&I3DCOL_INVISIBLE))
            continue;
      }
                              // ... or no-collision
      if(frm->GetFrameFlags() & I3D_FRMF_NOCOLLISION)
         continue;

                              //determine what to do now
      switch(frm->GetType1()){
      case FRAME_VISUAL:
         {
            PI3D_visual vis = I3DCAST_VISUAL(frm);
            if(vis->GetVisualType1()==I3D_VISUAL_SINGLEMESH){
                              //make test on entire SM, skip children if failed
               if(!(vis->frm_flags&FRMFLAGS_HR_BOUND_VALID))
                  vis->hr_bound.MakeHierarchyBounds(frm);
               vis->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID);

               ++frm->drv->col_stats[I3D_driver::DRV_COL_STATS_L_S];
               if(!pickBSphereTest(vis->hr_bound.bsphere_world, cd.from, cd.normalized_dir, cd.closest_hit))
                  continue;

                                 //singlemesh is what we're interested in
               PI3D_object_singlemesh sm = (PI3D_object_singlemesh)vis;

                                 //test on SM volume box
               if(sm->IsVolumeBoxValid()){
                  ++frm->drv->col_stats[I3D_driver::DRV_COL_STATS_L_B];
                  const I3D_bbox &bbox = sm->GetVolumeBox1();
                  const S_matrix &mat = sm->GetVolumeBoxMatrix1();

                  float hit_dist;
                  bool b = pickBBoxTest(mat * frm->GetMatrix(), bbox, cd.from, cd.normalized_dir, cd.GetClosestHit(), &hit_dist);
                  if(b){
                  if(cd.cresp_face){
                              //save values (if collision refused)
                     float save_closest_hit = cd.GetClosestHit();
                     PI3D_frame save_hit_frm = cd.GetHitFrm();
                     S_vector save_hit_normal = cd.GetHitNormal();
                     cd.SetReturnValues(hit_dist, frm, &V_ZERO);
                     if(!cd.cresp_face(cd))
                        cd.SetReturnValues(save_closest_hit, save_hit_frm, &save_hit_normal);
                  }else
                     cd.SetReturnValues(hit_dist, frm, &V_ZERO);
                  }
               }
               break;
            }
         }
         break;

      case FRAME_JOINT:
         {
            PI3D_joint jnt = I3DCAST_JOINT(frm);
            if(jnt->IsVolumeBoxValid()){
               ++frm->drv->col_stats[I3D_driver::DRV_COL_STATS_L_B];
               const I3D_bbox &bbox = jnt->GetVolumeBox1();
               const S_matrix &mat = jnt->GetVolumeBoxMatrix1();
               float hit_dist;
               bool b = pickBBoxTest(mat * frm->GetMatrix(), bbox, cd.from, cd.normalized_dir, cd.GetClosestHit(), &hit_dist);
               if(b){
                  if(cd.cresp_face){
                              //save values (if collision refused)
                     float save_closest_hit = cd.GetClosestHit();
                     PI3D_frame save_hit_frm = cd.GetHitFrm();
                     S_vector save_hit_normal = cd.GetHitNormal();
                     cd.SetReturnValues(hit_dist, frm, &V_ZERO);
                     if(!cd.cresp_face(cd))
                        cd.SetReturnValues(save_closest_hit, save_hit_frm, &save_hit_normal);
                  }else
                     cd.SetReturnValues(hit_dist, frm, &V_ZERO);
               }
            }
         }
         break;

      case FRAME_MODEL:
         {
                              //hierarchy test on model bounds
            PI3D_model mod = I3DCAST_MODEL(frm);
            if(frm->num_hr_vis >= 2){
               if(!(frm->frm_flags&FRMFLAGS_HR_BOUND_VALID))
                  mod->hr_bound.MakeHierarchyBounds(frm);
               mod->hr_bound.GetBoundSphereTrans(frm, FRMFLAGS_HR_BSPHERE_VALID);
               ++frm->drv->col_stats[I3D_driver::DRV_COL_STATS_L_S];
               const I3D_bsphere &bs_world = mod->hr_bound.bsphere_world;
               float r = bs_world.radius;
               if(!FastTest_L_S(cd, bs_world.pos, r, r*r))
                  continue;
            }
         }
         break;
      }
                              //recurse only if the frame has any joints
      if(frm->num_hr_jnt)
         cbEnumSMandJoints(frm, cd);
   }
}

//----------------------------

void I3D_scene::TestVolumeSM(I3D_collision_data &cd) const{

   cbEnumSMandJoints(primary_sector, cd);
}

//----------------------------//dynamic sphere space for fast collisions sort
//Spheres are sorted by distance from 0 and radius - annulus in which reside.
//Whenewer volume reset, we check if annulus changed and move pointer into appropriet
//space represented by vector in map, identified by space id(distance to 0).

void I3D_scene::AddSphereToSpace(CPI3D_volume vol, int id_min, int id_max){

   assert(vol);
   assert(id_min <= id_max || !"AddSphereToSpace");
                              //add sphere into all space where reside(usually 1 or 2)
   for(int i = id_min; i <= id_max; i++){
      vector<CPI3D_volume> &vol_list = spheres_space[i];
      assert(FindPointerInArray((void**)vol_list.begin(), vol_list.size(), (void *)vol) == -1 || !"Sphere already in list");
      vol_list.push_back(vol);
   }
   vol->last_min = id_min;
   vol->last_max = id_max;
}

//----------------------------

void I3D_scene::DeleteSphereFromSpace(CPI3D_volume vol, int id_min, int id_max){

   assert(vol);
   assert(id_min <= id_max || !"DeleteSphereFromSpace");
                              //del sphere into all space where reside(usually 1 or 2)
   for(int i = id_min; i <= id_max; i++){
      t_spheres_space::iterator it = spheres_space.find(i);
      assert(it != spheres_space.end() || !"trying delete sphere from non exist space");
      if(it == spheres_space.end()) continue;
      vector<CPI3D_volume> &vol_list = (*it).second;
      int id = FindPointerInArray((void**)vol_list.begin(), vol_list.size(), (void *)vol);
      assert(id != -1 || !"sphere is not in list");
      if(id != -1){
         vol_list[id] = vol_list.back();
         vol_list.pop_back();
      }
                              //free unused space
      if(!vol_list.size()){
         spheres_space.erase(it);
      }
   }
   vol->last_min = vol->last_max = -1;
}


//----------------------------
//----------------------------
