#ifndef __BSP_H
#define __BSP_H

#include <vector>
#include "frame.h"

//----------------------------
// Binary sorted partion tree. 
// Features: 
//    -Can contain polygons in nodes (with same or near cooplanar normals to node's plane).
//    -can contain leaves (end nodes with list of polygons with different normals)
//    -polygons are represented as array of original scene triangles outside of tree structure (world coord)
//    -models faces can be represented as array of models faces, each model represented only one (local coord)
//    -nodes can be bounded by axis alignated boxes to posssibility skip branchech if ray not collide with box
//       (unfortunally, in trees which are not divided by axis plane bound boxes are often owerlaping, so
//        there is need of some box hit cache or selective contruction of bboxes. 
//        in fact, node bboxes can slow down entire traaverse)
//    -models could be represented by bound volume (sphere, box), positioned in space occupied by node.
//       this can speed up collision test but also expand traverse algoritm, due need of distance compare.
//    -mooving or destructable geometry and entities can be solved in several ways:
//       a)Post inserted to tree as pointer to model or volume
//       b)separate hiearchy list with supprt for sectors
//       c)Mooving or inserted in octtree as pointer to model or volume. 
//          Other opportunity to it is divide space into octtree with resonable size of nodes(boxes),
//          and into each node compile own bsp tree, and also move there entities and non static geometry.
//          adventage is only one traverse, disadventage bsp faces redundancy (1 frame can reside in more then 8 nodes)
//    -each face/volume can be cached for hit testing by ray: whenewer ray is casted throw tree,
//       result of test with current ray is stored and this face should not be tested again if reside in more nodes
//       main problemm is how reset cache for next ray. Maybe we could keep it as bit array of hit
//       faces, set and read as fast shift opearation, reset by zero mem.
//    -dynamic plane shifting & beveling:(optional)
//    -collision statistic (optional)

//cached info about ray/plane test.
// data: current ray
//       bit_array indicates which planes was tested
//       info about tested planes
// input:   plane index
// output:  ray_origin near/far side(FRONT/BACK)
//          signed distance along ray to plane

/*class C_BSP_plane_cache{

   dword *bit_field;
};*/

//----------------------------

typedef dword BSP_index;      //type used for indexing bsp arrays

#define CELL_SIZE (sizeof(dword)*8)

typedef bool I3DAPI I3D_BSP_BUILDER_CB_PROC(dword prm1, dword prm2, void *context);


//----------------------------
// collision cache
class C_bitfield{

   #define CELL_SIZE (sizeof(dword)*8)
   dword *mem;
   int num_items;
   dword alloc_size;
public:
   C_bitfield() :
      mem(NULL), 
      num_items(0),
      alloc_size(0)
   {}

   C_bitfield(int size1) :
      mem(NULL),
      num_items(0),
      alloc_size(0)
   {
      Reserve(size1);
   }

   ~C_bitfield(){
      if(mem)
         delete[] mem;
   }

   inline void Clear(){
      if(mem){
         delete[] mem;
         mem = NULL;
      }
      num_items = 0;
   }

   inline int Size() const{ return num_items; }

   inline void Reserve(int num){

      if(num > num_items){
         int new_size = num / CELL_SIZE;
         if(num % CELL_SIZE)
            new_size++;

         dword new_alloc_size = new_size*sizeof(dword);
         void *new_mem = new dword[new_size];
         memset(new_mem, 0, new_alloc_size);
         if(mem){
            memcpy(new_mem, mem, alloc_size);
            delete[] mem;
         }
         mem = (dword*)new_mem;
         num_items = num;
         alloc_size = new_alloc_size;
      }
   }

   inline void Reset(){

      if(mem){
         memset(mem, 0, alloc_size);
      }
   }

   inline void SetOn(int index, bool on_off = true){

      assert(mem && (index < Size()));
      int cell = index / CELL_SIZE;
      int bit_id = index % CELL_SIZE;
      dword bit_mask = 1 << bit_id;
      if(on_off)
         mem[cell] |= bit_mask;
      else
         mem[cell] &= ~bit_mask;
   }

   inline bool IsOn(int index){

      assert(mem && (index < Size()));
      int cell = index / CELL_SIZE;
      int bit_id = index % CELL_SIZE;
      dword bit_mask = 1 << bit_id;
      return mem[cell] & bit_mask;
   }
};

//----------------------------

struct S_bsp_origin{

   C_smart_ptr<I3D_frame> frame;  //origin frame
private:
   dword face_index;          //direct index into mesh index_buffer
   PI3D_material mat;         //original material of face from face_group
   inline bool SetGroupMaterial(); //return false if something is wrong
public:
   S_bsp_origin() : frame(NULL),
      mat(NULL)
   {}
   bool Setup(PI3D_frame frm, dword face); // return false if something is wrong
   inline PI3D_material GetMaterial() const { return mat; }
   inline const dword &GetFaceIndex() const { return face_index; }
   const S_bsp_origin &operator =(const S_bsp_origin &org);
};

//----------------------------

struct S_triindex{
   int id[3];
   S_triindex(){
      for(int i = 3; i--;)
         id[i] = -1;
   }
};

//----------------------------

#define SLIDE_EDGE_ALL 7

//----------------------------
                              //bsp face
struct S_bsp_triface{

   S_bsp_origin origin;       //info into face's frame
   BSP_index plane_id;            //index of the plane the face is parallel to
   bool plane_side;          //set if the normal is parallel to the plane normal
   S_vector verticies[3];     //face verticies
   bool expanded;
   mutable bool highlight;
   dword vertex_checksum;     //for check validity of saved tree

   bool edges_computed;
   S_vector edges[3];         //edges vectors in world coord
   S_plane edge_planes[3];    //edge planes in world coord
   byte slide_flags;          //which edges may modify collision normal in sphere tests
   S_bsp_triface() :
      edges_computed(false),
      expanded(false),
      highlight(false),
      slide_flags(SLIDE_EDGE_ALL),
      vertex_checksum(0)
   {}

   void ComputeEdges();
   bool IsPointInTriangle(const S_vector &pt)const;

   bool IsValid()const { return origin.frame != NULL; }
   bool Expand();             //expand face to world coord, compute edges and edge planes
   void DebugDraw(PI3D_scene sc, dword color, bool fill = false)const;

   inline dword GetVertexChecksum() const;

   bool GetVertexIndexes(S_triindex &ti)const;
};

//----------------------------
                              //structure used during building of bsp tree
struct S_bsp_polygon{

   BSP_index origin_face;      //point into trees face list
   vector<S_vector> verticies;//already transformed to world coord
   bool valid;

   S_bsp_polygon();
   S_bsp_polygon(const S_bsp_polygon &p);
   void SetVerticies(const S_vector *vp, int num_v, BSP_index face_idx);
   const S_bsp_polygon &operator =(const S_bsp_polygon &p);
   ~S_bsp_polygon();
};

//----------------------------

struct S_bsp_node{

private:
   BSP_index *facelist;             //polygons coincident with partion plane (indexes into tree face list)
   dword num_faces;
public:
   BSP_index plane_id;              //partioning plane
   S_bsp_node *front;
   S_bsp_node *back;

   S_bsp_node();
   ~S_bsp_node();

   I3D_RESULT Save(C_cache *cp) const;
   I3D_RESULT Load(C_cache *cp);
   dword GetNumChildren()const;
   void GetStatsRec(I3D_stats_bsp &st) const;

   inline const BSP_index *GetFaces() const { return facelist; }
   inline dword NumFaces() const { return num_faces; }
   inline void AllocFaces(int num_faces1);
   inline void SetFaces(const BSP_index *p_faces, int num_faces1);
};

//----------------------------

struct S_collision_info{

   S_plane plane;             //OPTIM: can we solve it without copy?
   CPI3D_frame hit_frm;
   dword plane_id;
   PI3D_volume vol_source;
   bool modified_face_normal;
   S_vector face_normal;
   byte slided;                //bitflag of edges which was slidet for rurrent face(modified normal)

   float in_dist;
   int face_id;               //for solving edge collision

   inline bool operator <(const S_collision_info& ci) const{
      return in_dist < ci.in_dist;
   }
   S_collision_info() : 
      plane_id(0),
      hit_frm(NULL),
      vol_source(NULL),
      face_id(-1),
      slided(0),
      modified_face_normal(false)
   {}
};

//----------------------------// structure keeping info for object collision trace

struct S_trace_help{
   S_vector from;             //trace origin
   S_vector to;               //trace end
   S_vector dir;              //(to - from)
   S_vector dir_norm;         //direction normalized
   float dir_magn;            //trace magnitude
   float dir_2_sum;           //dir.dot(dir) - used i dynamic tests
   float radius;              //sphere object
   dword flags;
   PI3D_frame frm_ignored;          //skip this frame and his children in tests
   PI3D_volume vol_source;    //pass to collision_list
   float space_min, space_max;
   int critical_dist_i;       //precomputed value for dynamic box tests
   bool space_computed;

   inline bool Prepare(){
      dir = to - from;
      dir_2_sum = dir.Dot(dir);
      if(IsMrgZero(dir_2_sum)) 
         dir_magn = .0f;
      else
         dir_magn = SqrtFast(dir_2_sum);

      if(IsMrgZero(dir_magn))
         return false;
      dir_norm = dir / dir_magn;
      return true;
   }
   inline void PrepareMoveSpace(){
      space_min = from.Magnitude();
      space_max = to.Magnitude();
      if(space_min > space_max)
         swap(space_min, space_max);
      space_min = Max(.0f, space_min - radius);
      space_max = Max(.0f, space_max + radius);
      float critical_d = dir_magn + radius; //we put this computin here 'cos radius is in Prepare undefied yet.
      critical_dist_i = FLOAT_BITMASK(critical_d);
      space_computed = true;
   }
   mutable vector<S_collision_info> collision_list;
   S_trace_help() :
      frm_ignored(NULL),
      vol_source(NULL),
      radius(.0f),
      space_min(.0f), space_max(.0f),
      critical_dist_i(0),
      space_computed(false),
      flags(0)
   {
   }
};

//----------------------------

                              //list of all verticies used by bsp tree
typedef vector<S_vector> BSP_verticies;
                              //list of all planes, used for partion or face
typedef vector<S_plane> BSP_planes;
                              //list of all faces in tree
typedef vector<S_bsp_triface> BSP_faces;


//----------------------------

class C_bsp_tree{

enum E_PLANE_SIDE{
   PS_ON,
   PS_FRONT,
   PS_BACK,
};
                              //used in load for report frames which are static but not in tree
   void ReportConsistancy(I3D_LOAD_CB_PROC *cbP, void *context, const struct S_frame_base *hlp_base = NULL)const;

friend struct S_frame_base;
protected:
   bool valid;
   PI3D_scene scene;
   BSP_planes planes;
   BSP_faces faces;
   mutable C_bitfield face_cache;

   S_bsp_node *root;

                              //traverse:
                              //ray
   float ClassifyRayInterval(BSP_index plane_indx, float ray_min, float ray_max, E_PLANE_SIDE &near_side, E_PLANE_SIDE &far_side,
      const I3D_collision_data &cd) const;
   bool FaceListItersection(const S_bsp_node &node, const float dist, I3D_collision_data &cd) const;
   bool TextureIntersection(const S_bsp_triface &tf, const float hit_dist, I3D_collision_data &cd) const;
   bool NodeIntersection(const S_bsp_node &node, float ray_min, float ray_max, I3D_collision_data &cd) const;
                              //moving sphere
   bool CheckCol_MS_F(const S_bsp_triface &tf, const S_trace_help &th, BSP_index face_id) const;
   bool Classify_MS(BSP_index plane_indx, const S_trace_help &th, float &in_dist, float &out_dist, E_PLANE_SIDE &near_side)const;
   bool FaceListItersection_MS(const S_bsp_node &node, float near_dist, float far_dist, const S_trace_help &th)const;
   bool NodeIntersection_MS(const S_bsp_node &node, float ray_min, float ray_max, const S_trace_help &th)const;
                              //static sphere
   float ClassifySphere(BSP_index plane_indx, const S_vector &pos, float radius, E_PLANE_SIDE &near_side, E_PLANE_SIDE &far_side)const;
   bool CheckCol_S_F(const S_bsp_triface &tf, I3D_collision_data &cd, float plane_dist)const;
   bool NodeIntersection_S(const S_bsp_node &node, I3D_collision_data &cd) const;
   bool FaceListItersection_S(const S_bsp_node &node, I3D_collision_data &cd, float plane_dist) const;
   byte GetSlideNormal(const S_bsp_triface &tf, const S_plane &pl, const S_vector &ip, S_vector &rn) const;
public:
   C_bsp_tree();
   ~C_bsp_tree();
   bool TraceSphere(const S_trace_help &th) const;
   void Init(PI3D_scene sc){ scene = sc; }
   inline GetNumFaces() const { return faces.size(); }

//----------------------------
   inline dword GetNumNodes() const{
      if(root)
         return root->GetNumChildren();
      else
         return 0;
   }

//----------------------------
   inline const S_plane &GetPlane(BSP_index id) const{
      assert(id < planes.size()); return planes[id]; 
   }

//----------------------------
   void Close(){
      Clear();
      scene = NULL;
   }

//----------------------------
   const S_bsp_triface *GetFace(int idx) const{
      if(idx < faces.size())
         return &faces[idx];
      else
         return NULL;
   }

//----------------------------
// Collect statistics about tree.
   void GetStats(I3D_stats_bsp &st) const;
   
//----------------------------
   virtual void Clear();
   inline bool IsValid()const { return valid; }
   bool RayIntersection(I3D_collision_data &cd) const;
   bool SphereStaticIntersection(I3D_collision_data &cd) const;
                              //find all volumes which are not currently in tree and mark it for dynamic testing
   void SetAbsentVolumesPickable() const;
   bool CompareBspChecksum(I3D_LOAD_CB_PROC*, void *context) const;

   I3DMETHOD(Save)(C_cache *cp);
   I3DMETHOD(Load)(C_cache *cp, I3D_LOAD_CB_PROC *cbP, void *context, dword check_flags);
};

//----------------------------
//----------------------------

enum E_BSP_BUILD_RESULT{

   BSP_BUILD_FAIL,
   BSP_BUILD_OK,
   BSP_BUILD_CANCEL,
};

//----------------------------

class C_bsp_tree_builder : public C_bsp_tree{

   vector<S_bsp_polygon> scene_polygons;

   struct S_partion_rating{
      int split;
      int front;
      int back;
      int on;
      float ballance_score;
      float split_score;
      void Reset(){
         split = 0; front = 0; back = 0; on = 0;
         ballance_score = .0f;
         split_score = .0f;
      }
   };

   struct S_build_stat{
      int polygon_rest;
      int curr_depth;
      void Reset();
   } build_stats;

   enum E_PARTION_RESULT;

   mutable vector<PI3D_frame> tree_frames;

   void ComputeFacesChecksum();
   void BuildSlideFlags();
   void GetFrames(vector<PI3D_frame> &frm_list)const;
                              //scene polygons

   inline bool AddFace(PI3D_frame origin_frm, int face_id);
   E_BSP_BUILD_RESULT AddPolygons(vector<PI3D_frame> &frm_list, I3D_BSP_BUILDER_CB_PROC *cbP, void *context);
   void AddFacePlane(S_bsp_triface &tf, dword context);
   E_BSP_BUILD_RESULT BuilScenePolygons(I3D_BSP_BUILDER_CB_PROC *cbP, void *context);
                              //node build support
   void MakePolygonsVolume(I3D_bbox &bb, const vector<S_bsp_polygon> &poly_list);
   
   E_PARTION_RESULT ClassifyPolygon(const S_bsp_polygon &poly, BSP_index plane_id);
   E_PARTION_RESULT Split_Polygon(const S_bsp_polygon &poly, const S_plane &partion, 
      vector<S_bsp_polygon> &front, vector<S_bsp_polygon> &back);
   

   void ClassifyPartionPlane(BSP_index plane_id, const vector<S_bsp_polygon> &poly_list, S_partion_rating &rating);

   BSP_index PickPartionPlane(const vector<S_bsp_polygon> &poly_list);
   BSP_index PickPartionPlaneRandom(const vector<S_bsp_polygon> &poly_list);

   E_BSP_BUILD_RESULT BuildNode(S_bsp_node *node, vector<S_bsp_polygon> &poly_list,
      I3D_BSP_BUILDER_CB_PROC *cbP = NULL, void *context = NULL);

   void DrawEdges(const S_preprocess_context &pc) const;
   void DebugHighlight(const S_preprocess_context &pc) const;

public:
   virtual void Clear(){
      tree_frames.clear();
      C_bsp_tree::Clear();
   }
   I3DMETHOD(Build)(I3D_BSP_BUILDER_CB_PROC *cbP = NULL, void *context = NULL);
   I3DMETHOD_(void, DebugDraw)(const S_preprocess_context &pc) const;
};

//----------------------------
//----------------------------

#endif                        //__BSP_H