#include "all.h"
#include "common.h"
#include <sortlist.hpp>

//----------------------------

class C_edit_Modify: public C_editor_item{
   virtual const char *GetName() const{ return "Modify"; }
   LPC_editor_item e_undo, e_slct;

   struct S_edit_Modify_undo{
      int num_frames;
#pragma pack(push, 1)
      struct S_frm{
         dword flags;            //flags to modify/reset
         S_vector pos;
         S_quat rot;
         float scale;
         S_vector nu_scale;
         char frm_names[1];      //frm_name [link_name]
      };
      S_frm frm_info[1];
#pragma pack(pop)
   };
   struct S_modify_item{
      dword flags;
      S_vector pos;
      S_quat rot;
      float scale;
      S_vector nu_scale;
      dword material_id;


      C_string link_name;
      C_smart_ptr<I3D_frame> frm;
                              //keep children deleted from model
      vector<C_string> model_deleted_children;

      S_modify_item():
         flags(0),
         material_id(0),
         frm(NULL)
      {}
      S_modify_item(const S_modify_item &mi):
         frm(NULL)
      {
         operator =(mi);
      }
      S_modify_item &operator =(const S_modify_item &mi){
         flags = mi.flags;
         pos = mi.pos;
         rot = mi.rot;
         scale = mi.scale;
         nu_scale = mi.nu_scale;
         link_name = mi.link_name;
         frm = mi.frm;
         model_deleted_children = mi.model_deleted_children;
         return (*this);
      }
   };
   template<class T>
   struct S_item_less{
      bool operator ()(const T &x, const T &y) const{
         return (((byte*)x) < ((byte*)y));
      }
   };
   typedef map<PI3D_frame, S_modify_item, S_item_less<PI3D_frame> > t_item_map;
   t_item_map items;
   dword default_reset_flags;

                              //list of plugins to which we send notifications
   vector<pair<LPC_editor_item, dword> > notify_list;

   void SendNotify(const pair<PI3D_frame, dword> &p){
      for(int i=0; i<notify_list.size(); i++)
         notify_list[i].first->Action(notify_list[i].second, (void*)&p);
   }

   void SendNotify(PI3D_frame frm, dword flags){
      if(notify_list.size()) SendNotify(pair<PI3D_frame, dword>(frm, flags));
   }

   void SetFrameSector(PI3D_frame frm){
                              //set by frame's center point
      I3D_bound_volume bvol;
      frm->ComputeHRBoundVolume(&bvol);
      const I3D_bbox &bbox = bvol.bbox;

      S_vector pos = frm->GetWorldPos();
      if(bbox.IsValid()) pos += bbox.min + (bbox.max-bbox.min)*.5f;
      ed->GetScene()->SetFrameSectorPos(frm, pos);
   }

public:
   C_edit_Modify():
      e_undo(NULL),
      e_slct(NULL),
      default_reset_flags(E_MODIFY_FLG_ALL)
   {}

   virtual bool Init(){
                              //find required plugins
      e_undo = ed->FindPlugin("Undo");
      e_slct = ed->FindPlugin("Selection");
      if(!e_undo || !e_slct){
         e_undo = NULL;
         e_slct = NULL;
         return false;
      }
      e_undo->AddRef();
      e_slct->AddRef();

      ed->AddShortcut(this, E_MODIFY_RESET, "%10 &Edit\\%80 Reset modify\tHome", K_HOME, 0);
      return true;
   }

   virtual void Close(){
      if(e_undo) e_undo->Release();
      if(e_slct) e_slct->Release();
      e_undo = NULL;
      e_slct = NULL;

      for(int i=notify_list.size(); i--; ) notify_list[i].first->Release();
      notify_list.clear();
   }

   //virtual void Tick(byte skeys, int time, int mouse_rx, int mouse_ry, int mouse_rz, byte mouse_butt){}

   virtual dword Action(int id, void *context){

      switch(id){
      case E_MODIFY_ADD_FLAGS:
         {
            const pair<PI3D_frame, dword> &p = *(pair<PI3D_frame, dword>*)context;
                              //check if frame already in list
            PI3D_frame frm = p.first;
            dword flags = p.second;
            S_modify_item *itm;
            t_item_map::iterator it = items.find(frm);
            if(it==items.end()){
               itm = &(items[frm] = S_modify_item());
               itm->frm = frm;
            }else 
               itm = &(*it).second;
            if((itm->flags&flags) != flags){
                              //new flags added
               dword new_flags = (itm->flags^flags)&flags;
               if(new_flags&E_MODIFY_FLG_POSITION)
                  itm->pos = frm->GetPos();
               if(new_flags&E_MODIFY_FLG_ROTATION)
                  itm->rot = frm->GetRot();
               if(new_flags&E_MODIFY_FLG_SCALE)
                  itm->scale = frm->GetScale();
               if(new_flags&E_MODIFY_FLG_NU_SCALE){
                  switch(frm->GetType()){
                  case FRAME_VOLUME: itm->nu_scale = I3DCAST_VOLUME(frm)->GetNUScale(); break;
                  default: new_flags &= ~E_MODIFY_FLG_NU_SCALE;
                  }
               }
               if(new_flags&E_MODIFY_FLG_MATERIAL){
                  switch(frm->GetType()){
                  case FRAME_VOLUME:
                     itm->material_id = I3DCAST_VOLUME(frm)->GetCollisionMaterial(); break;
                  case FRAME_VISUAL:
                     itm->material_id = I3DCAST_VISUAL(frm)->GetCollisionMaterial(); break;
                  default: new_flags &= ~E_MODIFY_FLG_MATERIAL;
                  }
               }
               if(new_flags&E_MODIFY_FLG_LINK){
                  PI3D_frame prnt = frm->GetParent();
                  if(!prnt) prnt = ed->GetScene()->GetPrimarySector();
                  itm->link_name = prnt->GetName();
               }
               itm->flags |= new_flags;
            }
            SendNotify(p);
         }
         break;

      case E_MODIFY_DEL_FLAGS:
         {
            const pair<PI3D_frame, dword> &p = *(pair<PI3D_frame, dword>*)context;
                              //check if frame already in list
            PI3D_frame frm = p.first;
            t_item_map::iterator it = items.find(frm);
            if(it!=items.end()){
               (*it).second.flags &= ~p.second;
               if(!(*it).second.flags && !(*it).second.model_deleted_children.size()){
                  items.erase(it);
               }
            }
         }
         break;

      case E_MODIFY_GET_FLAGS:
         {
            PI3D_frame frm = (PI3D_frame)context;
            t_item_map::iterator it = items.find(frm);
            if(it==items.end())
               return 0;
            return (*it).second.flags;
         }
         break;

      case E_MODIFY_DEL_FRAME:
         {
            PI3D_frame frm = (PI3D_frame)context;
            t_item_map::iterator it = items.find(frm);
            if(it==items.end())
               return false;
            items.erase(it);
            return true;
         }
         break;

      case E_MODIFY_ADD_MODEL_DEL_CHILD:
         {
                              //assure no name is listed twice
#ifndef NDEBUG
            bool b = 
#endif
               Action(E_MODIFY_DEL_MODEL_DEL_CHILD, context);
            assert(!b);
            const pair<PI3D_frame, const char*> &p = *(pair<PI3D_frame, const char*>*)context;
            PI3D_frame frm = p.first;
            S_modify_item *itm;
            t_item_map::iterator it = items.find(frm);
            if(it==items.end()){
               itm = &(items[frm] = S_modify_item());
               itm->frm = frm;
            }else 
               itm = &(*it).second;
            itm->model_deleted_children.push_back(p.second);
         }
         break;

      case E_MODIFY_DEL_MODEL_DEL_CHILD:
         {
            const pair<PI3D_frame, const char*> &p = *(pair<PI3D_frame, const char*>*)context;
            PI3D_frame frm = p.first;
            t_item_map::iterator it = items.find(frm);
            if(it==items.end())
               return false;
            if(p.second){
               for(int i=(*it).second.model_deleted_children.size(); i--; )
                  if((*it).second.model_deleted_children[i]==p.second)
                     break;
               if(i==-1) return false;
               (*it).second.model_deleted_children[i] = (*it).second.model_deleted_children.back();
               (*it).second.model_deleted_children.pop_back();
            }else
               (*it).second.model_deleted_children.clear();
            if(!(*it).second.flags && !(*it).second.model_deleted_children.size())
               items.erase(it);
            return true;
         }
         break;

      case E_MODIFY_REPLACE_FRAME:
         {
            pair<PI3D_frame, PI3D_frame> &p = *(pair<PI3D_frame, PI3D_frame>*)context;
            t_item_map::iterator it = items.find(p.first);
            if(it==items.end()) return false;
            assert((*it).second.frm==p.first);

            S_modify_item &itm = (items[p.second] = S_modify_item());
            itm = (*it).second;
            itm.frm = p.second;
            items.erase(p.first);

            return true;
         }
         break;

      case E_MODIFY_RESET:
         {
            if(!ed->CanModify()) break;
            const vector<PI3D_frame> &sel_list = *(vector<PI3D_frame>*)e_slct->Action(E_SELECTION_GETVECTOR);

            int i;
                              //prepare undo info
            vector<byte> undo_info;
            for(i=0; i<4; i++) undo_info.push_back(0);

            for(int j=sel_list.size(); j--; ){
               PI3D_frame frm = sel_list[j];
               t_item_map::iterator it = items.find(frm);
               if(it==items.end()) continue;

               S_modify_item *itm = &(*it).second;
               dword reset_flags = E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE | E_MODIFY_FLG_NU_SCALE | E_MODIFY_FLG_LINK;
               reset_flags &= itm->flags;

                              //can't reset position/rotation on created
               if(itm->flags&E_MODIFY_FLG_CREATE) reset_flags &= ~(E_MODIFY_FLG_POSITION|E_MODIFY_FLG_ROTATION);

               if(!reset_flags) continue;

               {
                              //prompt user for properties to reset
                  struct S_hlp{
                     dword reset_flags, default_reset_flags;
                     LPIGraph igraph;
                     const char *frm_name;

                     static BOOL CALLBACK dlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam){

                        switch(uMsg){
                        case WM_INITDIALOG:
                           {
                              S_hlp *hp = (S_hlp*)lParam;
                              InitDlg(hp->igraph, hwnd);
                                                            //enable selected items
                              if(!(hp->reset_flags&E_MODIFY_FLG_POSITION)) EnableWindow(GetDlgItem(hwnd, IDC_RESET_POS), false);
                              if(!(hp->reset_flags&E_MODIFY_FLG_ROTATION)) EnableWindow(GetDlgItem(hwnd, IDC_RESET_ROT), false);
                              if(!(hp->reset_flags&(E_MODIFY_FLG_SCALE|E_MODIFY_FLG_NU_SCALE))) EnableWindow(GetDlgItem(hwnd, IDC_RESET_SCL), false);
                              if(!(hp->reset_flags&E_MODIFY_FLG_LINK)) EnableWindow(GetDlgItem(hwnd, IDC_RESET_LNK), false);

                              if(hp->default_reset_flags&E_MODIFY_FLG_POSITION) CheckDlgButton(hwnd, IDC_RESET_POS, BST_CHECKED);
                              if(hp->default_reset_flags&E_MODIFY_FLG_ROTATION) CheckDlgButton(hwnd, IDC_RESET_ROT, BST_CHECKED);
                              if(hp->default_reset_flags&(E_MODIFY_FLG_SCALE|E_MODIFY_FLG_NU_SCALE)) CheckDlgButton(hwnd, IDC_RESET_SCL, BST_CHECKED);
                              if(hp->default_reset_flags&E_MODIFY_FLG_LINK) CheckDlgButton(hwnd, IDC_RESET_LNK, BST_CHECKED);
                              char win_name[256];
                              GetWindowText(hwnd, win_name, sizeof(win_name));
                              SetWindowText(hwnd, C_fstring("%s - %s", win_name, hp->frm_name));

                              SetWindowLong(hwnd, GWL_USERDATA, lParam);
                           }
                           return 1;
                        case WM_COMMAND:
                           switch(wParam){
                           case IDCANCEL:
                              EndDialog(hwnd, 0);
                              break;
                           case IDOK:
                              {
                                 S_hlp *hp = (S_hlp*)GetWindowLong(hwnd, GWL_USERDATA);
                                                            //collect results
                                 hp->default_reset_flags &= 
                                    ~(E_MODIFY_FLG_POSITION|E_MODIFY_FLG_ROTATION|E_MODIFY_FLG_SCALE|E_MODIFY_FLG_NU_SCALE|E_MODIFY_FLG_LINK);

                                 if(IsDlgButtonChecked(hwnd, IDC_RESET_POS)) hp->default_reset_flags |= E_MODIFY_FLG_POSITION;
                                 if(IsDlgButtonChecked(hwnd, IDC_RESET_ROT)) hp->default_reset_flags |= E_MODIFY_FLG_ROTATION;
                                 if(IsDlgButtonChecked(hwnd, IDC_RESET_SCL)) hp->default_reset_flags |= E_MODIFY_FLG_SCALE|E_MODIFY_FLG_NU_SCALE;
                                 if(IsDlgButtonChecked(hwnd, IDC_RESET_LNK)) hp->default_reset_flags |= E_MODIFY_FLG_LINK;
                                 hp->reset_flags &= hp->default_reset_flags;
                              }
                              EndDialog(hwnd, 1);
                              break;
                           }
                           break;
                        }
                        return 0;
                     }
                  } hlp = {reset_flags, default_reset_flags, ed->GetIGraph(), frm->GetName()};
                  int i = DialogBoxParam(GetHInstance(), "IDD_MODIFY_RESET", (HWND)ed->GetIGraph()->GetHWND(),
                     S_hlp::dlgProc, (LPARAM)&hlp);
                  if(!i) break;
                  reset_flags = hlp.reset_flags;
                  default_reset_flags = hlp.default_reset_flags;
                  if(!reset_flags)
                     continue;
               }

                              //save undo info
               {
                  int nlen = strlen(frm->GetName());
                  int len = nlen + sizeof(S_edit_Modify_undo::S_frm);
                  if(reset_flags&E_MODIFY_FLG_LINK){
                     if(frm->GetParent()) len += strlen(frm->GetParent()->GetName()) + 1;
                     else reset_flags &= ~E_MODIFY_FLG_LINK;
                  }
                  for(int i=0; i<len; i++) undo_info.push_back(0);
                  S_edit_Modify_undo::S_frm &ui = *(S_edit_Modify_undo::S_frm*)(undo_info.end() - len);
                              //store info
                  ui.flags = reset_flags;
                  if(reset_flags&E_MODIFY_FLG_POSITION) ui.pos = frm->GetPos();
                  if(reset_flags&E_MODIFY_FLG_ROTATION) ui.rot = frm->GetRot();
                  if(reset_flags&E_MODIFY_FLG_SCALE) ui.scale = frm->GetScale();
                  if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                     switch(frm->GetType()){
                     case FRAME_VOLUME: ui.nu_scale = I3DCAST_VOLUME(frm)->GetNUScale(); break;
                     }
                  }
                  if(reset_flags&E_MODIFY_FLG_LINK){
                     strcpy(&ui.frm_names[nlen+1], frm->GetParent()->GetName());
                     PI3D_frame prev_link = ed->GetScene()->FindFrame(itm->link_name, ENUMF_BACKDROP | ENUMF_ALL);
                     if(!prev_link) ed->Message(C_fstring("Reset: can't find link: '%s'", itm->link_name));
                     frm->LinkTo(prev_link);
                  }

                  strcpy(ui.frm_names, frm->GetName());
                  S_edit_Modify_undo &upi = *(S_edit_Modify_undo*)&undo_info.front();
                  ++upi.num_frames;
               }

                              //perform reset
               bool upd = false;
               if(reset_flags&E_MODIFY_FLG_POSITION){
                  frm->SetPos(itm->pos);
                  upd = true;
               }
               if(reset_flags&E_MODIFY_FLG_ROTATION){
                  frm->SetRot(itm->rot);
                  upd = true;
               }
               if(reset_flags&E_MODIFY_FLG_SCALE){
                  frm->SetScale(itm->scale);
                  upd = true;
               }
               if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                  switch(frm->GetType()){
                  case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(itm->nu_scale); break;
                  }
                  upd = true;
               }
               if(upd) frm->Update();

               SendNotify(frm, reset_flags);

                              //update flags
               itm->flags &= ~reset_flags;
               if(!itm->flags){
                  items.erase(it);
               }
               e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
               ed->SetModified();
            }
                              
            {                 //save undo info
               S_edit_Modify_undo &upi = *(S_edit_Modify_undo*)&undo_info.front();
               if(upi.num_frames){
                                    //save undo info
                  S_edit_Undo_2 eu(GetName(), E_MODIFY_ADD_NAMED, undo_info.begin(), undo_info.size());
                  e_undo->Action(E_UNDO_SAVE, &eu);

                  ed->Message(C_fstring("%i frame(s) reset", upi.num_frames));
               }
            }
         }
         break;

      case E_MODIFY_ADD_NAMED:
      case E_MODIFY_RESET_NAMED:
         {
            S_edit_Modify_undo *mu = (S_edit_Modify_undo*)context;
            S_edit_Modify_undo::S_frm *fi = mu->frm_info;

            int i;
                              //prepare undo info
            vector<byte> undo_info;
            for(i=0; i<4; i++) undo_info.push_back(0);

            for(i=0; i<mu->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_names, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                              //find in list
                  t_item_map::iterator it = items.find(frm);
                  if(it==items.end()){
                     S_modify_item &itm = (items[frm] = S_modify_item());
                     itm.frm = frm;
                  }
                  S_modify_item *itm = &items[frm];

                              //check which flags we modify/reset
                  dword reset_flags = fi->flags;

                              //save undo info
                  {
                     int nlen = strlen(frm->GetName());
                     int len = nlen + sizeof(S_edit_Modify_undo::S_frm);
                     if((reset_flags&E_MODIFY_FLG_LINK) && id==E_MODIFY_RESET_NAMED){
                        if(frm->GetParent()) len += strlen(frm->GetParent()->GetName()) + 1;
                        else reset_flags &= ~E_MODIFY_FLG_LINK;
                     }
                     for(int i=0; i<len; i++) undo_info.push_back(0);
                     S_edit_Modify_undo::S_frm &ui = *(S_edit_Modify_undo::S_frm*)(undo_info.end() - len);
                                 //store info
                     ui.flags = reset_flags;
                     if(id==E_MODIFY_RESET_NAMED){
                        if(reset_flags&E_MODIFY_FLG_POSITION) ui.pos = frm->GetPos();
                        if(reset_flags&E_MODIFY_FLG_ROTATION) ui.rot = frm->GetRot();
                        if(reset_flags&E_MODIFY_FLG_SCALE) ui.scale = frm->GetScale();
                        if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                           switch(frm->GetType()){
                           case FRAME_VOLUME: ui.nu_scale = I3DCAST_VOLUME(frm)->GetNUScale(); break;
                           }
                        }
                        if(reset_flags&E_MODIFY_FLG_LINK)
                           strcpy(&ui.frm_names[nlen+1], frm->GetParent()->GetName());
                     }else{
                        if(reset_flags&E_MODIFY_FLG_POSITION) itm->pos = frm->GetPos();
                        if(reset_flags&E_MODIFY_FLG_ROTATION) itm->rot = frm->GetRot();
                        if(reset_flags&E_MODIFY_FLG_SCALE) itm->scale = frm->GetScale();
                        if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                           switch(frm->GetType()){
                           case FRAME_VOLUME: itm->nu_scale = I3DCAST_VOLUME(frm)->GetNUScale(); break;
                           }
                        }
                        if(reset_flags&E_MODIFY_FLG_LINK)
                           itm->link_name = frm->GetParent()->GetName();
                     }

                     strcpy(ui.frm_names, frm->GetName());
                     S_edit_Modify_undo &upi = *(S_edit_Modify_undo*)&undo_info.front();
                     ++upi.num_frames;
                  }
                              //perform reset
                  bool upd = false;
                  if(id==E_MODIFY_ADD_NAMED){
                     if(reset_flags&E_MODIFY_FLG_POSITION) frm->SetPos(fi->pos), upd = true;
                     if(reset_flags&E_MODIFY_FLG_ROTATION) frm->SetRot(fi->rot), upd = true;
                     if(reset_flags&E_MODIFY_FLG_SCALE) frm->SetScale(fi->scale), upd = true;
                     if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                        switch(frm->GetType()){
                        case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(fi->nu_scale); break;
                        }
                        upd = true;
                     }
                     if(reset_flags&E_MODIFY_FLG_LINK){
                        PI3D_frame prev_link = ed->GetScene()->FindFrame(&fi->frm_names[strlen(fi->frm_names)+1], ENUMF_BACKDROP | ENUMF_ALL);
                        if(!prev_link) ed->Message("Reset: can't find link!!!");
                        frm->LinkTo(prev_link);
                     }
                  }else{
                     if(reset_flags&E_MODIFY_FLG_POSITION) frm->SetPos(itm->pos), upd = true;
                     if(reset_flags&E_MODIFY_FLG_ROTATION) frm->SetRot(itm->rot), upd = true;
                     if(reset_flags&E_MODIFY_FLG_SCALE) frm->SetScale(itm->scale), upd = true;
                     if(reset_flags&E_MODIFY_FLG_NU_SCALE){
                        switch(frm->GetType()){
                        case FRAME_VOLUME: I3DCAST_VOLUME(frm)->SetNUScale(itm->nu_scale); break;
                        }
                        upd = true;
                     }
                     if(reset_flags&E_MODIFY_FLG_LINK){
                        PI3D_frame prev_link = ed->GetScene()->FindFrame(itm->link_name, ENUMF_BACKDROP | ENUMF_ALL);
                        if(!prev_link) ed->Message(C_fstring("Reset: can't find link: '%s'", itm->link_name));
                        frm->LinkTo(prev_link);
                     }
                  }
                  if(upd) frm->Update();

                  SendNotify(frm, reset_flags);

                              //update flags
                  if(id==E_MODIFY_ADD_NAMED){
                     itm->flags |= reset_flags;
                  }else{
                     itm->flags &= ~reset_flags;
                     if(!itm->flags){
                              //remove from list
                        items.erase(it);
                     }
                  }

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_ADD_NAMED): can't find frame '%s'", fi->frm_names), 0, EM_ERROR);

               fi = (S_edit_Modify_undo::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo::S_frm) + strlen(fi->frm_names));
            }
            {                 //save undo info
               S_edit_Modify_undo &upi = *(S_edit_Modify_undo*)&undo_info.front();
               if(upi.num_frames){
                                    //save undo info
                  S_edit_Undo_2 eu(GetName(), 
                     id==E_MODIFY_RESET_NAMED ? E_MODIFY_ADD_NAMED : E_MODIFY_RESET_NAMED,
                     undo_info.begin(), undo_info.size());
                  e_undo->Action(E_UNDO_SAVE, &eu);

                  ed->Message(C_fstring("%i frame(s) %s", upi.num_frames,
                     id==E_MODIFY_RESET_NAMED ? "reset" : "modified"));
               }
            }
         }
         break;

      case E_MODIFY_GET_NUM_ITEMS:
         return items.size();

      case E_MODIFY_ENUM:
         {
            const pair<t_modify_enum*, dword> &p = *(pair<t_modify_enum*, dword>*)context;

            t_item_map::iterator it;
            for(it=items.begin(); it!=items.end(); it++){
               (*p.first)((*it).first, (*it).second.flags, p.second);
            }
         }
         break;

      case E_MODIFY_GET_ITEM_LINK:
         {
            t_item_map::iterator it = items.find((PI3D_frame)context);
            if(it==items.end()) return NULL;
            if(!((*it).second.flags&E_MODIFY_FLG_LINK)) return NULL;
            return (dword)(const char*)((*it).second.link_name);
         }
         break;

      case E_MODIFY_TRANSLATE_NAMED:
         {
            S_edit_Modify_undo_pos *up = (S_edit_Modify_undo_pos*)context;
            S_edit_Modify_undo_pos::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_POSITION);

            for(int i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  S_vector old_pos = frm->GetPos();
                  frm->SetPos(fi->pos);
                  fi->pos = old_pos;
                  frm->Update();

                  switch(frm->GetType()){
                  case FRAME_MODEL:
                     {
                        PI3D_frame prnt=frm->GetParent();
                        if(!prnt || prnt->GetType()==FRAME_SECTOR)
                           SetFrameSector(frm);
                     }
                     break;
                  }
                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_TRANSLATE_NAMED): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               fi = (S_edit_Modify_undo_pos::S_frm*)(((byte*)fi) + sizeof(S_vector)+strlen(fi->frm_name) + 1);
            }
            ed->Message(C_fstring("%i object(s) moved", up->num_frames));
                              //add undo info, reverse process: E_MODIFY_TRANSLATE_NAMED
            S_edit_Undo_2 eu(GetName(), E_MODIFY_TRANSLATE_NAMED, up, ((byte*)fi) - ((byte*)up));
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_ROTATE_NAMED:
      case E_MODIFY_ROT_TRANS_NAMED:
         {
            S_edit_Modify_undo_rot *up = (S_edit_Modify_undo_rot*)context;
            S_edit_Modify_undo_rot::S_frm *fi = up->frm_info;

            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_ROTATION);
            if(fi->use_pos) p.second |= E_MODIFY_FLG_POSITION;

            for(int i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  S_quat old_rot = frm->GetRot();
                  frm->SetRot(fi->rot);
                  fi->rot = old_rot;
                  if(fi->use_pos){
                     S_vector old_pos = frm->GetPos();
                     frm->SetPos(fi->pos);
                     fi->pos = old_pos;
                  }

                  frm->Update();
                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_ROTATE_NAMED): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               fi = (S_edit_Modify_undo_rot::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_rot::S_frm)+strlen(fi->frm_name));
            }
            ed->Message(C_fstring("%i object(s) rotated", up->num_frames));
                              //add undo info, reverse process: this action
            S_edit_Undo_2 eu(GetName(), id, up, ((byte*)fi) - ((byte*)up));
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_SCALE_NAMED:
         {
            S_edit_Modify_undo_scl *up = (S_edit_Modify_undo_scl*)context;
            S_edit_Modify_undo_scl::S_frm *fi = up->frm_info;

            pair<PI3D_frame, dword> p(NULL, 0);

            for(int i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(fi->use_non_uniform){
                     switch(frm->GetType()){
                     case FRAME_VOLUME: 
                        {
                           PI3D_volume vol = I3DCAST_VOLUME(frm);
                           S_vector old_scale = vol->GetNUScale();
                           vol->SetNUScale(fi->scl);
                           fi->scl = old_scale;
                           p.second = E_MODIFY_FLG_NU_SCALE;
                        }
                        break;
                     }
                  }else{
                     float old_scale = frm->GetScale();
                     frm->SetScale(fi->scl[0]);
                     fi->scl[0] = old_scale;
                     p.second = E_MODIFY_FLG_SCALE;
                  }
                  frm->Update();
                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_SCALE_NAMED): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               fi = (S_edit_Modify_undo_scl::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_scl::S_frm)+strlen(fi->frm_name));
            }
            ed->Message(C_fstring("%i object(s) scaled", up->num_frames));
                              //add undo info, reverse process: E_MODIFY_SCALE_NAMED
            S_edit_Undo_2 eu(GetName(), E_MODIFY_SCALE_NAMED, up, ((byte*)fi) - ((byte*)up));
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_LINK_NAMED:
         {
            S_edit_Modify_undo_lnk *up = (S_edit_Modify_undo_lnk*)context;
            S_edit_Modify_undo_lnk::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_LINK | E_MODIFY_FLG_POSITION | E_MODIFY_FLG_ROTATION | E_MODIFY_FLG_SCALE);
            int i;
                              //init undo info
            vector<byte> undo_info;
            for(i=0; i<4; i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_names, ENUMF_BACKDROP | ENUMF_ALL);
               int l1 = strlen(fi->frm_names) + 1;
               if(frm){
                              //save undo info
                  PI3D_frame curr_prnt = frm->GetParent();
                  int len = l1 + sizeof(S_edit_Modify_undo_lnk::S_frm);
                  if(curr_prnt) len += strlen(curr_prnt->GetName());
                  for(int j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_lnk::S_frm &ui = *(S_edit_Modify_undo_lnk::S_frm*)(undo_info.end() - len);
                  ui.orig_pos = frm->GetPos();
                  ui.orig_rot = frm->GetRot();
                  ui.orig_scl = frm->GetScale();
                  strcpy(ui.frm_names, frm->GetName());
                  if(curr_prnt) strcpy(&ui.frm_names[l1], curr_prnt->GetName());
                  S_edit_Modify_undo_lnk &upi = *(S_edit_Modify_undo_lnk*)&undo_info.front();
                  ++upi.num_frames;

                  frm->SetPos(fi->orig_pos);
                  frm->SetRot(fi->orig_rot);
                  frm->SetScale(fi->orig_scl);

                  PI3D_frame link_to = ed->GetScene()->FindFrame(fi->frm_names+l1, ENUMF_BACKDROP | ENUMF_ALL);
                  frm->LinkTo(link_to);
                  e_slct->Action(E_SELECTION_FLASH_FRAME, link_to);
                  frm->Update();

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_LINK_NAMED): can't find '%s'", fi->frm_names), 0, EM_ERROR);
                              //next
               int l2 = strlen(fi->frm_names + l1);
               fi = (S_edit_Modify_undo_lnk::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_lnk::S_frm) + l1 + l2);
            }
            ed->Message(C_fstring("%i object(s) linked", up->num_frames));

                              //add undo info, reverse process: E_MODIFY_LINK_NAMED
            { 
               S_edit_Modify_undo_lnk &upi = *(S_edit_Modify_undo_lnk*)&undo_info.front();
               if(upi.num_frames){
                  S_edit_Undo_2 eu(GetName(), E_MODIFY_LINK_NAMED, undo_info.begin(), undo_info.size());
                  e_undo->Action(E_UNDO_SAVE, &eu);
               }
            }
         }
         break;

      case E_MODIFY_LIGHT:
         {
            if(!ed->CanModify()) break;
            S_edit_Modify_undo_light *up = (S_edit_Modify_undo_light*)context;
            S_edit_Modify_undo_light::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_LIGHT);
            int i, j;

            vector<byte> undo_info;
            for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_names, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(frm->GetType()!=FRAME_LIGHT) continue;
                  PI3D_light lp = I3DCAST_LIGHT(frm);
                              //save undo info
                  int len = strlen(frm->GetName()) + 1 + sizeof(S_edit_Modify_undo_light::S_frm);

                              //count size of sector names
                  PI3D_sector const *spp = lp->GetLightSectors();
                  for(int i=lp->NumLightSectors(); i--; )
                     len += strlen(spp[i]->GetName()) + 1;

                  for(j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_light::S_frm &ui = *(S_edit_Modify_undo_light::S_frm*)(undo_info.end() - len);
                  ui.ltype = lp->GetLightType();
                  ui.color = lp->GetColor();
                  ui.power = lp->GetPower();
                  ui.mode = lp->GetMode();
                  lp->GetCone(ui.cone_i, ui.cone_o);
                  lp->GetRange(ui.range_n, ui.range_f);
                  ui.specular_color = lp->GetSpecularColor();
                  ui.specular_power = lp->GetSpecularPower();

                  strcpy(ui.frm_names, frm->GetName());
                  j = strlen(frm->GetName()) + 1;

                              //store names of sectors
                  for(int ii=lp->NumLightSectors(); ii--; ){
                     strcpy(ui.frm_names+j, spp[ii]->GetName());
                     j += strlen(spp[ii]->GetName()) + 1;
                  }
                  ui.frm_names[j] = 0;

                  S_edit_Modify_undo_light &upi = *(S_edit_Modify_undo_light*)&undo_info.front();
                  ++upi.num_frames;

                  lp->SetLightType(fi->ltype); 
                  lp->SetColor(fi->color); 
                  lp->SetPower(fi->power); 
                  lp->SetMode(fi->mode); 
                  lp->SetCone(fi->cone_i, fi->cone_o); 
                  lp->SetRange(fi->range_n, fi->range_f); 
                  lp->SetSpecularColor(fi->specular_color);
                  lp->SetSpecularPower(fi->specular_power);

                              //connect light to new sectors
                  while(lp->NumLightSectors())
                     lp->GetLightSectors()[0]->DeleteLight(lp);

                  j = strlen(fi->frm_names) + 1;
                  while(fi->frm_names[j]){
                     PI3D_sector sct = I3DCAST_SECTOR(ed->GetScene()->FindFrame(fi->frm_names+j, ENUMF_BACKDROP | ENUMF_SECTOR));
                     if(sct) sct->AddLight(lp);
                     j += strlen(fi->frm_names+j) + 1;
                  }

                  lp->Update();

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
                  ed->SetModified();
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_LIGHT): can't find '%s'", fi->frm_names), 0, EM_ERROR);
                              //next
               for(j=0; fi->frm_names[j]; ) j += strlen(fi->frm_names+j) + 1;
               fi = (S_edit_Modify_undo_light::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_light::S_frm)+j);
            }
            ed->Message(C_fstring("%i light(s) restored", up->num_frames));

                              //add undo info, reverse process: E_MODIFY_LIGHT
            S_edit_Undo_2 eu(GetName(), E_MODIFY_LIGHT, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_SOUND:
         {
            if(!ed->CanModify()) break;
            S_edit_Modify_undo_sound *up = (S_edit_Modify_undo_sound*)context;
            S_edit_Modify_undo_sound::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_SOUND);
            int i;

            vector<byte> undo_info;
            for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_names, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(frm->GetType()!=FRAME_SOUND) continue;
                  PI3D_sound sp = I3DCAST_SOUND(frm);
                              //save undo info
                  C_string fname = sp->GetFileName();
                              //strip extension
                  for(int ei=fname.Size(); ei--; ){
                     if(fname[ei]=='.'){
                        fname[ei] = 0;
                        break;
                     }
                  }

                  ed->GetSoundCache().GetRelativeDirectory(&fname[0]);
                  int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_sound::S_frm);
                  len += fname.Size() + 1;

                              //count size of sector names
                  PI3D_sector const *spp = sp->GetSoundSectors();
                  for(int i=sp->NumSoundSectors(); i--; )
                     len += strlen(spp[i]->GetName()) + 1;

                  for(int j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_sound::S_frm &ui = *(S_edit_Modify_undo_sound::S_frm*)(undo_info.end() - len);

                  ui.stype = sp->GetSoundType();
                  ui.vol = sp->GetVolume();
                  ui.out_vol = sp->GetOutVol();
                  sp->GetCone(ui.cone_i, ui.cone_o);
                  sp->GetRange(ui.range_n, ui.range_f, ui.fade_ratio);
                  ui.loop = sp->IsLoop();
                  ui.streaming = sp->IsStreaming();
                  ui.on = sp->IsOn();
                  strcpy(ui.frm_names, frm->GetName());
                  j = strlen(frm->GetName())+1;
                  strcpy(ui.frm_names+j, fname);
                  j += strlen(ui.frm_names+j) + 1;

                              //store names of sectors
                  for(int ii=sp->NumSoundSectors(); ii--; ){
                     strcpy(ui.frm_names+j, spp[ii]->GetName());
                     j += strlen(spp[ii]->GetName()) + 1;
                  }
                  ui.frm_names[j] = 0;

                  S_edit_Modify_undo_sound &upi = *(S_edit_Modify_undo_sound*)&undo_info.front();
                  ++upi.num_frames;

                  const char *filename = fi->frm_names + strlen(fi->frm_names) + 1;
                  sp->SetOn(false);
                  sp->SetSoundType(fi->stype); 
                  sp->SetVolume(fi->vol);
                  sp->SetOutVol(fi->out_vol);
                  sp->SetCone(fi->cone_i, fi->cone_o);
                  sp->SetRange(fi->range_n, fi->range_f, fi->fade_ratio);
                  sp->SetLoop(fi->loop);
                  sp->SetStreaming(fi->streaming);
                  ed->GetSoundCache().Open(sp, filename, ed->GetScene(), 0, OPEN_ERR_LOG(ed, filename));
                  if(fi->on) sp->SetOn(true);

                              //connect sound to new sectors
                  while(sp->NumSoundSectors())
                     sp->GetSoundSectors()[0]->DeleteSound(sp);

                  j = strlen(fi->frm_names) + 1;
                  j += strlen(fi->frm_names+j) + 1;
                  while(fi->frm_names[j]){
                     PI3D_sector sct = I3DCAST_SECTOR(ed->GetScene()->FindFrame(fi->frm_names+j, ENUMF_BACKDROP | ENUMF_SECTOR));
                     if(sct) sct->AddSound(sp);
                     j += strlen(fi->frm_names+j) + 1;
                  }

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
                  ed->SetModified();
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_SOUND): can't find '%s'", fi->frm_names), 0, EM_ERROR);
                              //next
               int len = strlen(fi->frm_names);
               len += strlen(fi->frm_names + len + 1) + 1;
               fi = (S_edit_Modify_undo_sound::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_sound::S_frm)+len);
            }
            ed->Message(C_fstring("%i sound(s) restored", up->num_frames));

                              //add undo info, reverse process: E_MODIFY_SOUND
            S_edit_Undo_2 eu(GetName(), E_MODIFY_SOUND, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_VOLUME:
         {
            S_edit_Modify_undo_volume *up = (S_edit_Modify_undo_volume*)context;
            S_edit_Modify_undo_volume::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_VOLUME);
            int i;

            vector<byte> undo_info;
            for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(frm->GetType()!=FRAME_VOLUME) continue;
                  PI3D_volume vp = I3DCAST_VOLUME(frm);
                              //save undo info
                  int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_volume::S_frm);
                  for(int j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_volume::S_frm &ui = *(S_edit_Modify_undo_volume::S_frm*)(undo_info.end() - len);

                  ui.vtype = vp->GetVolumeType();
                  ui.material_id = vp->GetCollisionMaterial();
                  strcpy(ui.frm_name, frm->GetName());
                  S_edit_Modify_undo_volume &upi = *(S_edit_Modify_undo_volume*)&undo_info.front();
                  ++upi.num_frames;

                  vp->SetVolumeType(fi->vtype); 
                  vp->SetCollisionMaterial(fi->material_id);
                  vp->Update();

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_VOLUME): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               int len = strlen(fi->frm_name);
               fi = (S_edit_Modify_undo_volume::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_volume::S_frm)+len);
            }
            ed->Message(C_fstring("%i volume(s) restored", up->num_frames));

                              //add undo info
            S_edit_Undo_2 eu(GetName(), E_MODIFY_VOLUME, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_VISUAL:
         {
            S_edit_Modify_undo_visual *up = (S_edit_Modify_undo_visual*)context;
            S_edit_Modify_undo_visual::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_VISUAL);
            int i;
            
            e_slct->Action(E_SELECTION_CLEAR);

            vector<byte> undo_info;
            for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->data, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(frm->GetType()!=FRAME_VISUAL) continue;
                  PI3D_visual vis = I3DCAST_VISUAL(frm);

                              //save undo info
                  int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_visual::S_frm);
                  dword vtype = vis->GetVisualType();
                              //compute size of all properties
                  int num_props = ed->GetDriver()->NumProperties(vtype);
                  int GetPropertySize(I3D_PROPERTYTYPE ptype, dword value);
                  for(int pi=num_props; pi--; ){
                     dword value = vis->GetProperty(pi);
                     len += GetPropertySize(ed->GetDriver()->GetPropertyType(vtype, pi), value);
                  }
                              //reserve size
                  for(int j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_visual::S_frm &ui = *(S_edit_Modify_undo_visual::S_frm*)(undo_info.end() - len);

                  ui.visual_type = vis->GetVisualType();
                  ui.material_id = vis->GetCollisionMaterial();
                  ui.data_size = 0;
                              //store name
                  strcpy(&ui.data[ui.data_size], frm->GetName());
                  ui.data_size += strlen(&ui.data[ui.data_size]) + 1;
                              //store properties
                  for(pi=0; pi<num_props; pi++){
                     I3D_PROPERTYTYPE ptype = ed->GetDriver()->GetPropertyType(vtype, pi);
                     dword value = vis->GetProperty(pi);
                     dword len = GetPropertySize(ptype, value);
                     switch(ptype){
                     case I3DPROP_STRING:
                     case I3DPROP_VECTOR3:
                     case I3DPROP_VECTOR4:
                     case I3DPROP_COLOR:
                        memcpy(&ui.data[ui.data_size], (const void*)value, len);
                        break;
                     default:
                        memcpy(&ui.data[ui.data_size], &value, len);
                     }
                     ui.data_size += len;
                  }

                  S_edit_Modify_undo_visual &upi = *(S_edit_Modify_undo_visual*)&undo_info.front();
                  ++upi.num_frames;

                  PI3D_visual new_vis;
                  if(fi->visual_type!=vtype){
                     new_vis = CastVisual(ed, vis, fi->visual_type, this, e_slct);
                  }else{
                     new_vis = vis;
                     e_slct->Action(E_SELECTION_ADD_FRAME, new_vis);
                  }
                  if(new_vis){
                              //fill-in old properties
                     dword num_props = ed->GetDriver()->NumProperties(fi->visual_type);
                     int pos = strlen(fi->data) + 1;
                     for(int pi=0; pi<num_props; pi++){
                        I3D_PROPERTYTYPE ptype = ed->GetDriver()->GetPropertyType(fi->visual_type, pi);
                        dword value;
                        switch(ptype){
                        case I3DPROP_BOOL:
                           value = *(byte*)&fi->data[pos];
                           break;
                        case I3DPROP_STRING:
                        case I3DPROP_VECTOR3:
                        case I3DPROP_VECTOR4:
                        case I3DPROP_COLOR:
                           value = (dword)&fi->data[pos];
                           break;
                        default:
                           value = *(dword*)&fi->data[pos];
                        }
                        if(ptype!=I3DPROP_NULL)
                           new_vis->SetProperty(pi, value);
                        pos += GetPropertySize(ptype, value);
                     }
                     assert(pos == fi->data_size);
                     new_vis->SetCollisionMaterial(fi->material_id);

                     e_slct->Action(E_SELECTION_FLASH_FRAME, new_vis);
                  }

                  p.first = new_vis;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_VISUAL): can't find '%s'", fi->data), 0, EM_ERROR);
                              //next
               int len = fi->data_size + sizeof(S_edit_Modify_undo_visual::S_frm) - sizeof(char);
               fi = (S_edit_Modify_undo_visual::S_frm*)(((byte*)fi) + len);
            }
            ed->Message(C_fstring("%i visual(s) restored", up->num_frames));

                              //add undo info
            S_edit_Undo_2 eu(GetName(), E_MODIFY_VISUAL, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_SECTOR:
         {
            S_edit_Modify_undo_sector *up = (S_edit_Modify_undo_sector*)context;
            S_edit_Modify_undo_sector::S_frm *fi = up->frm_info;
            pair<PI3D_frame, dword> p(NULL, E_MODIFY_FLG_SECTOR);
            int i;

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_SECTOR | ENUMF_ALL);
               if(frm){
                  PI3D_sector sct = I3DCAST_SECTOR(frm);
                              //save undo info
                  dword new_env = fi->env_id;
                  fi->env_id = sct->GetEnvironmentID();
                  sct->SetEnvironmentID(new_env);

                  e_slct->Action(E_SELECTION_FLASH_FRAME, frm);

                  p.first = frm;
                  Action(E_MODIFY_ADD_FLAGS, &p);
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_SECTOR): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               int len = strlen(fi->frm_name);
               fi = (S_edit_Modify_undo_sector::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_sector::S_frm)+len);
            }
            ed->Message(C_fstring("%i volume(s) restored", up->num_frames));

                              //add undo info
            S_edit_Undo_2 eu(GetName(), E_MODIFY_SECTOR, up, (byte*)fi - (byte*)up);
            e_undo->Action(E_UNDO_SAVE, &eu);
         }
         break;

      case E_MODIFY_FRMFLAGS:
         {
            S_edit_Modify_undo_frmflags *up = (S_edit_Modify_undo_frmflags*)context;
            S_edit_Modify_undo_frmflags::S_frm *fi = up->frm_info;
            int i;

            vector<byte> undo_info;
            for(i=0; i<sizeof(int); i++) undo_info.push_back(0);

            for(i=0; i<up->num_frames; i++){
               PI3D_frame frm = ed->GetScene()->FindFrame(fi->frm_name, ENUMF_BACKDROP | ENUMF_ALL);
               if(frm){
                  if(frm->GetFlags() == fi->frm_flags)
                     continue;
                              //save undo info
                  int len = strlen(frm->GetName()) + sizeof(S_edit_Modify_undo_frmflags::S_frm);
                  for(int j=0; j<len; j++) undo_info.push_back(0);
                  S_edit_Modify_undo_frmflags::S_frm &ui = *(S_edit_Modify_undo_frmflags::S_frm*)(undo_info.end() - len);

                  ui.frm_flags = frm->GetFlags();
                  frm->SetFlags(fi->frm_flags);

                  strcpy(ui.frm_name, frm->GetName());
                  S_edit_Modify_undo_frmflags &upi = *(S_edit_Modify_undo_frmflags*)&undo_info.front();
                  ++upi.num_frames;
               }else
                  ed->Message(C_fstring("Modify(E_MODIFY_FRMFLAGS): can't find '%s'", fi->frm_name), 0, EM_ERROR);
                              //next
               int len = strlen(fi->frm_name);
               fi = (S_edit_Modify_undo_frmflags::S_frm*)(((byte*)fi) + sizeof(S_edit_Modify_undo_frmflags::S_frm)+len);
            }
            ed->Message(C_fstring("%i frame(s) restored", up->num_frames));

                              //add undo info
            S_edit_Undo_2 eu(GetName(), E_MODIFY_FRMFLAGS, undo_info.begin(), undo_info.size());
            e_undo->Action(E_UNDO_SAVE, &eu);

                              //update property window
            LPC_editor_item e_prop = ed->FindPlugin("Properties");
            e_prop->Action(E_PROP_RESET_SHEET);
         }
         break;

      case E_MODIFY_ADD_NOTIFY:
         {
            const pair<LPC_editor_item, dword> &np = *(pair<LPC_editor_item, dword>*)context;
            for(int i=notify_list.size(); i--; )
               if(notify_list[i].first==np.first) break;
            if(i==-1){
               notify_list.push_back(np);
               np.first->AddRef();
               return true;
            }
         }
         break;

      case E_MODIFY_DEL_NOTIFY:
         {
            LPC_editor_item ei = (LPC_editor_item)context;
            for(int i=notify_list.size(); i--; )
               if(notify_list[i].first==ei) break;
            if(i!=-1){
               ei->Release();
               notify_list[i] = notify_list.back();
               notify_list.pop_back();
            }
         }
         break;

      case EDIT_ACTION_MISSION_SAVE_0:
         if(items.size()){
            C_chunk &ck = *(C_chunk*)context;
            int i;

            ck.WAscend(CT_MODIFICATIONS);
                              //sort them by number of parents due to link dependency
            C_sort_list<S_modify_item*> mod_list(items.size());
            t_item_map::iterator it;
            for(it=items.begin(); it!=items.end(); ++it){
               int nump;
               PI3D_frame frm = (*it).second.frm;
               int sort_value = 0;
                              //on the same level, save models earlier than other,
                              // because they may contain sectors, on which other
                              // things may be dependent
               if(frm->GetType()!=FRAME_MODEL)
                  ++sort_value;

               for(nump=0; frm=frm->GetParent(); nump++);
               sort_value += nump*2;
               mod_list.Add(&(*it).second, sort_value);
            }
            mod_list.Sort();

            for(i=0; i<mod_list.Count(); i++){
               const S_modify_item &mod = *mod_list[i];

                              //don't save temporary frames
               if(mod.flags&E_MODIFY_FLG_TEMPORARY)
                  continue;

               PI3D_frame frm = mod.frm;
                              //check if it's scene-frame or explicitly created frame
               bool created = (mod.flags&E_MODIFY_FLG_CREATE);

               ck.WAscend(CT_MODIFICATION);

               if(created)
                  ck.WIntChunk(CT_FRAME_TYPE, frm->GetType());

               dword flags = mod.flags;
                              //write frame's name
               ck.WStringChunk(CT_NAME, frm->GetName());

               switch(frm->GetType()){

               case FRAME_MODEL:
                  {
                              //write model's file name (without extension)
                     C_string str = I3DCAST_MODEL(frm)->GetFileName();
                     /*
                     for(int i=str.Size(); i--; ) 
                     if(str[i]=='.'){
                        str[i] = 0;
                        break;
                     }
                     */
                     ed->GetModelCache().GetRelativeDirectory(&str[0]);
                     ck.WStringChunk(CT_MODELFILENAME, str);
                              //write model's deleted frames
                     for(int i=mod.model_deleted_children.size(); i--; ){
                        ck.WStringChunk(CT_MODEL_DEL_FRAME, mod.model_deleted_children[i]);
                     }
                  }
                  break;

               case FRAME_DUMMY:
                  {
                     const I3D_bbox *bb = I3DCAST_DUMMY(frm)->GetBBox();
                     ck.WAscend(CT_MODIFY_DUMMY);
                     ck.Write(bb, sizeof(I3D_bbox));
                     ck.Descend();
                  }
                  break;

               case FRAME_LIGHT:
                  {
                     PI3D_light lp = I3DCAST_LIGHT(frm);
                              //write all light's params
                     ck.WAscend(CT_MODIFY_LIGHT);
                     {
                        ck.WIntChunk(CT_LIGHT_TYPE, lp->GetLightType());
                        ck.WVectorChunk(CT_COLOR, lp->GetColor());
                        ck.WFloatChunk(CT_LIGHT_POWER, lp->GetPower());
                        {     //cone
                           ck.WAscend(CT_LIGHT_CONE);
                           float c[2];
                           lp->GetCone(c[0], c[1]);
                           ck.Write(c, sizeof(c));
                           ck.Descend();
                        }
                        {     //range
                           ck.WAscend(CT_LIGHT_RANGE);
                           float r[2];
                           lp->GetRange(r[0], r[1]);
                           ck.Write(r, sizeof(r));
                           ck.Descend();
                        }
                        ck.WIntChunk(CT_LIGHT_MODE, lp->GetMode());
                              //write specular info
                        ck.WVectorChunk(CT_LIGHT_SPECULAR_COLOR, lp->GetSpecularColor());
                        ck.WFloatChunk(CT_LIGHT_SPECULAR_POWER, lp->GetSpecularPower());
                        //ck.WriteFloat(CT_LIGHT_SPECULAR_ANGLE, lp->GetSpecularAngle());

                              //sector names
                        for(int i=lp->NumLightSectors(); i--; ){
                           PI3D_sector sct = lp->GetLightSectors()[i];
                           ck.WStringChunk(CT_LIGHT_SECTOR, sct->GetName());
                        }
                     }
                     ck.Descend();
                  }
                  break;

               case FRAME_SOUND:
                  {
                     PI3D_sound sp = I3DCAST_SOUND(frm);
                                 //write all sound's params
                     ck.WAscend(CT_MODIFY_SOUND);
                     {
                        if(sp->IsStreaming()){  //must be before filename!
                           ck.WAscend(CT_SOUND_STREAMING);
                           ck.Descend();
                        }

                        C_string name = sp->GetFileName();
                        ed->GetSoundCache().GetRelativeDirectory(&name[0]);
                              //strip extension
                        {
                           for(int i=strlen(name); i--; )
                           if(name[i] == '.'){
                              name[i] = 0;
                              break;
                           }
                        }
                        ck.WStringChunk(CT_NAME, name);

                        ck.WIntChunk(CT_SOUND_TYPE, sp->GetSoundType());
                        ck.WFloatChunk(CT_SOUND_VOLUME, sp->GetVolume());
                        ck.WFloatChunk(CT_SOUND_OUTVOL, sp->GetOutVol());
                        {        //cone
                           ck.WAscend(CT_SOUND_CONE);
                           float c[2];
                           sp->GetCone(c[0], c[1]);
                           ck.Write(c, sizeof(c));
                           ck.Descend();
                        }
                        {        //range
                           ck.WAscend(CT_SOUND_RANGE);
                           S_vector r;
                           sp->GetRange(r[0], r[1], r[2]);
                           ck.Write(&r, sizeof(r));
                           ck.Descend();
                        }
                        if(sp->IsLoop()){
                           ck.WAscend(CT_SOUND_LOOP);
                           ck.Descend();
                        }
                        if(sp->IsOn()){
                           ck.WAscend(CT_SOUND_ENABLE);
                           ck.Descend();
                        }

                              //sector names
                        for(int i=sp->NumSoundSectors(); i--; ){
                           PI3D_sector sct = sp->GetSoundSectors()[i];
                           ck.WStringChunk(CT_SOUND_SECTOR, sct->GetName());
                        }
                     }
                     ck.Descend();
                  }
                  break;

               case FRAME_VOLUME:
                  {
                     PI3D_volume vp = I3DCAST_VOLUME(frm);
                                 //write all volume's params
                     ck.WAscend(CT_MODIFY_VOLUME);
                     {
                        ck.WIntChunk(CT_VOLUME_TYPE, vp->GetVolumeType());
                        ck.WIntChunk(CT_VOLUME_MATERIAL, vp->GetCollisionMaterial());
                     }
                     ck.Descend();
                  }
                  break;

               case FRAME_OCCLUDER:
                  {
                     PI3D_occluder occ = I3DCAST_OCCLUDER(frm);
                     ck.WAscend(CT_MODIFY_OCCLUDER);
                     switch(occ->GetOccluderType()){
                     case I3DOCCLUDER_MESH:
                        {
                           ck.WAscend(CT_OCCLUDER_VERTICES);
                           {
                              dword numv = occ->NumVertices();
                              ck.Write(&numv, sizeof(numv));
                              S_vector *vp = occ->LockVertices();
                              ck.Write(vp, numv*sizeof(S_vector));
                              occ->UnlockVertices();
                           }
                           ck.Descend();
                        }
                        break;
                     }
                     ck.Descend();
                  }
                  break;

               case FRAME_SECTOR:
                  {
                     PI3D_sector sct = I3DCAST_SECTOR(frm);
                     ck.WAscend(CT_MODIFY_SECTOR);
                     ck.WWordChunk(CT_SECTOR_ENVIRONMENT, sct->GetEnvironmentID());
                     /*
                     for(int i=I3D_SSDI_LAST; i--; ){
                        ck.WAscend(CT_SECTOR_SOUND_DATA);
                        ck.Write(&i, sizeof(word));
                        dword data = sct->GetSoundData((I3D_SECTOR_SOUND_DATA_INDEX)i);
                        ck.Write(&data, sizeof(dword));
                        ck.Descend();
                     }
                     */
                     ck.Descend();
                  }
                  break;

               }
                              //write position, rotation & scale
               if(flags&E_MODIFY_FLG_POSITION){
                  if((flags&E_MODIFY_FLG_CREATE) && !(flags&E_MODIFY_FLG_LINK))
                     ck.WVectorChunk(CT_POSITION, frm->GetWorldPos());
                  else
                     ck.WVectorChunk(CT_POSITION, frm->GetPos());
                              //save world pos, for case link would fail
                  ck.WVectorChunk(CT_WORLD_POSITION, frm->GetWorldPos());
               }
               if(flags&E_MODIFY_FLG_ROTATION){
                  if((flags&E_MODIFY_FLG_CREATE) && !(flags&E_MODIFY_FLG_LINK))
                     ck.WRotChunk(CT_QUATERNION, frm->GetWorldRot());
                  else
                     ck.WRotChunk(CT_QUATERNION, frm->GetRot());
               }
               if(flags&E_MODIFY_FLG_SCALE){
                  /*
                  if((flags&E_MODIFY_FLG_CREATE) && !(flags&E_MODIFY_FLG_LINK))
                     ck.WriteFloat(CT_USCALE, frm->GetWorldScale());
                  else
                  */
                     ck.WFloatChunk(CT_USCALE, frm->GetScale());
               }
               if(flags&E_MODIFY_FLG_NU_SCALE){
                  switch(frm->GetType()){
                  case FRAME_VOLUME: ck.WVectorChunk(CT_NUSCALE, I3DCAST_VOLUME(frm)->GetNUScale()); break;
                  }
               }
               if(flags&E_MODIFY_FLG_LINK){
                  PI3D_frame prnt = frm->GetParent();
                  if(prnt){
                     ck.WAscend(CT_LINK);
                     ck.WStringChunk(CT_NAME, prnt->GetName());
                     ck.Descend();
                  }
               }
               if(flags&E_MODIFY_FLG_HIDE){
                  ck.WAscend(CT_HIDDEN);
                  ck.Descend();
               }
               if(flags&E_MODIFY_FLG_VISUAL){
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                              //write visual type
                  ck.WIntChunk(CT_MODIFY_VISUAL, vis->GetVisualType());
                              //write visual material
                  dword mat_id = vis->GetCollisionMaterial();
                  if(mat_id)
                     ck.WIntChunk(CT_VISUAL_MATERIAL, mat_id);

                  switch(vis->GetVisualType()){
                  case I3D_VISUAL_LIT_OBJECT:
                     {
                        PI3D_lit_object lmap = (PI3D_lit_object)vis;
                        //if(lmap->IsConstructed())
                        {
                           ck.WAscend(CT_MODIFY_LIT_OBJECT);
                           lmap->Save((dword)ck.GetHandle());
                           ck.Descend();
                        }
                     }
                     break;
                  }
               }
               if(frm->GetType() == FRAME_VISUAL){
                              //write visual's properties
                  PI3D_visual vis = I3DCAST_VISUAL(frm);
                  dword vtype = vis->GetVisualType();
                  dword num_props = ed->GetDriver()->NumProperties(vtype);
                  for(int pi=0; pi<num_props; pi++){
                     I3D_PROPERTYTYPE ptype = ed->GetDriver()->GetPropertyType(vtype, pi);
                     if(ptype==I3DPROP_NULL)
                        continue;
                     ck.WAscend(CT_MODIFY_VISUAL_PROPERTY);
                     ck.Write(&ptype, sizeof(byte));
                     assert(pi < 65536);
                     ck.Write(&pi, sizeof(word));
                              //write value
                     dword value = vis->GetProperty(pi);
                     int GetPropertySize(I3D_PROPERTYTYPE, dword);   //definition
                     dword len = GetPropertySize(ptype, value);
                     switch(ptype){
                     case I3DPROP_STRING:
                     case I3DPROP_VECTOR3:
                     case I3DPROP_VECTOR4:
                     case I3DPROP_COLOR:
                        ck.Write((const void*)value, len);
                        break;
                     default:
                        ck.Write(&value, len);
                     }
                     ck.Descend();
                  }
               }
                              //write frame flags
               if(frm->GetFlags()){
                  ck.WIntChunk(CT_FRAME_FLAGS, frm->GetFlags());
               }

               ck.Descend();     //CT_FRAME
            }

                              //write frame flags of yet not written frames
            struct S_hlp{
               t_item_map *items;
               C_chunk *ck;
               static I3DENUMRET I3DAPI cbEnum(PI3D_frame frm, dword c){

                  S_hlp *hp = (S_hlp*)c;
                  if(frm->GetFlags()){
                              //check if we have already written this
                     C_edit_Modify::t_item_map::const_iterator it;
                     it = hp->items->find(frm);
                     if(it == hp->items->end()){
                        hp->ck->WAscend(CT_MODIFICATION);
                        hp->ck->WStringChunk(CT_NAME, frm->GetName());
                        hp->ck->WIntChunk(CT_FRAME_FLAGS, frm->GetFlags());
                        hp->ck->Descend();
                     }
                  }
                  switch(frm->GetType()){
                  case FRAME_TARGET:
                     {
                        PI3D_frame tgt = I3DCAST_TARGET(frm)->GetFrame();
                        if(tgt){
                           hp->ck->WAscend(CT_MODIFICATION);
                           hp->ck->WStringChunk(CT_NAME, frm->GetName());
                           hp->ck->WStringChunk(CT_MODIFY_TARGET, tgt->GetName());
                           hp->ck->Descend();
                        }
                     }
                     break;
                  }
                  return I3DENUMRET_OK;
               }
            } hlp = {&items, &ck};
            ed->GetScene()->EnumFrames(S_hlp::cbEnum, (dword)&hlp, ENUMF_ALL | ENUMF_BACKDROP);

            ck.Descend();
         }
         break;

      case EDIT_ACTION_BEFORE_FREE:
         items.clear();
         break;
         /*
                              //ignored commands:
      case EDIT_ACTION_PICK: case EDIT_ACTION_AFTER_LOAD: case EDIT_ACTION_MISSION_SAVE_1: case EDIT_ACTION_MISSION_SAVE_2: case EDIT_ACTION_MISSION_LOAD: case EDIT_ACTION_WINDOW_RESET: case EDIT_ACTION_FRAME_DELETE: case EDIT_ACTION_FRAME_DUPLICATE: case EDIT_ACTION_SET_ACTIVE:
      case EDIT_ACTION_QUERY_MODIFY: case EDIT_ACTION_VALIDATE:
         break;
      default:
         todo(C_fstring("Plugin %s: unknown command: %i", GetName(), id));
         */
      }
      return 0;
   }
   //virtual void Render(){}
   //virtual bool LoadState(class C_chunk*){ return false; }
   //virtual bool SaveState(class C_chunk*) const{ return false; }
};

//----------------------------

void CreateModify(LPC_editor ed){
   LPC_editor_item ei = new C_edit_Modify;
   ed->InstallPlugin(ei);
   ei->Release();
}

//----------------------------
