#include "all.h"
#include "igraph_i.h"
#include "loader.h"
#include <d3d.h>
#include <vector>
#include <dta_read.h>
#include <debug.h>
#include <math.h>


#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------
                              //registered loaders
vector<LPC_image_loader> load_plugins;

void RegisterImageLoader(LPC_image_loader il){
   load_plugins.push_back(il);
}

//----------------------------

void *C_image_loader::MemAlloc(dword size){
   return new byte[size];
}

//----------------------------

void C_image_loader::MemFree(void *mem){
   delete[] (byte*)mem;
}

//----------------------------
                              //image creation
bool IGraph::CreateImage(LPIImage *ip){

   LPIImage img = new IImage(this);
   *ip = img;
   return img;
}

//----------------------------

bool IGraph::GetImageInfo(const char *name, dword &sx, dword &sy, byte &bpp, dword &id){

   int h = dtaOpen(name);
   if(h==-1)
      return false;
                              //read header
   dtaRead(h, &id, sizeof(dword));

   int i;
   for(i=load_plugins.size(); i--; ){
      S_pixelformat pf;
      bool b = load_plugins[i]->GetImageInfo(h, name, &id, sizeof(id),
         &sx, &sy,  &pf);
      if(b){
         bpp = pf.bytes_per_pixel;
         break;
      }
   }
   dtaClose(h);
   return (i!=-1);
}

//----------------------------
//----------------------------
                              //image class
IImage::IImage(LPIGraph ig):
   igraph(ig),
   pal_argb(NULL),
   size_x(0),
   size_y(0)
{
}

//----------------------------

IImage::~IImage(){

   Close();
}

//----------------------------

static D3DFORMAT ConvertFormat(const S_pixelformat *pf){

   if(pf->flags&PIXELFORMAT_COMPRESS){
      switch(pf->four_cc){
      case MAKEFOURCC('D', 'X', 'T', '1'):
         return D3DFMT_DXT1;
      case MAKEFOURCC('D', 'X', 'T', '2'):
         return D3DFMT_DXT2;
      case MAKEFOURCC('D', 'X', 'T', '3'):
         return D3DFMT_DXT3;
      case MAKEFOURCC('D', 'X', 'T', '4'):
         return D3DFMT_DXT4;
      case MAKEFOURCC('D', 'X', 'T', '5'):
         return D3DFMT_DXT5;
      }
   }else
   if(pf->flags&PIXELFORMAT_BUMPMAP){
      switch(pf->bytes_per_pixel){
      case 2:
         switch(pf->u_mask){
         case 0x00ff:
            assert(pf->v_mask==0xff00);
            assert(pf->l_mask==0x0000);
            return D3DFMT_V8U8;
         case 0x001f:
            assert(pf->v_mask==0x03e0);
            assert(pf->l_mask==0xfc00);
            return D3DFMT_L6V5U5;
         default:
            assert(0);
         }
         break;
      case 4:
         switch(pf->u_mask){
         case 0x000000ff:
            assert(pf->v_mask==0x0000ff00);
            switch(pf->l_mask){
            case 0x00ff0000:
               return D3DFMT_X8L8V8U8;
            case 0x00000000:
               return D3DFMT_Q8W8V8U8;
            default:
               assert(0);
            }
            break;
         case 0x0000ffff:
            assert(pf->v_mask==0xffff0000);
            assert(pf->l_mask==0x00000000);
            return D3DFMT_V16U16;
         case 0x000003ff:
            assert(pf->v_mask==0x001ffc00);
            assert(pf->l_mask==0xffe00000);
            return D3DFMT_W11V11U10;
         default:
            assert(0);
         }
         break;
      default:
         assert(0);
      }
   }else
   switch(pf->bytes_per_pixel){

   case 1:
      if(pf->flags&PIXELFORMAT_PALETTE)
         return D3DFMT_P8;
      else
         return D3DFMT_R3G3B2;

   case 2:

      switch(pf->r_mask){
      case 0xf800:
         assert(pf->g_mask==0x07e0);
         assert(pf->b_mask==0x001f);
         return D3DFMT_R5G6B5;

      case 0x7c00:
         switch(pf->a_mask){
         case 0x0000:
            return D3DFMT_X1R5G5B5;
         case 0x8000:
            return D3DFMT_A1R5G5B5;
         }
         break;

      case 0x0f00:
         switch(pf->a_mask){
         case 0x0000:
            return D3DFMT_X4R4G4B4;
         case 0xf000:
            assert(pf->r_mask==0x0f00);
            assert(pf->g_mask==0x00f0);
            assert(pf->b_mask==0x000f);
            return D3DFMT_A4R4G4B4;
         }
         break;

      case 0x00e0:
         return D3DFMT_A8R3G3B2;

      default: assert(0);
      }
      break;

   case 3:
      assert(pf->r_mask == 0x00ff0000);
      assert(pf->g_mask == 0x0000ff00);
      assert(pf->b_mask == 0x000000ff);
      assert(pf->a_mask == 0x00000000);
      return D3DFMT_R8G8B8;

   case 4:
      assert(pf->r_mask == 0x00ff0000);
      assert(pf->g_mask == 0x0000ff00);
      assert(pf->b_mask == 0x000000ff);
      switch(pf->a_mask){
      case 0:
         return D3DFMT_X8R8G8B8;
      case 0xff000000:
         return D3DFMT_A8R8G8B8;
      }
      break;
   default: assert(0);
   }
   return D3DFMT_UNKNOWN;
}

//----------------------------

static void InitColorkeyInfo(const void *src, void *dst, int size_x, int size_y, int src_pitch, int dst_pitch,
   const S_pixelformat &pf_src, const S_pixelformat &pf_dst, vector<byte> &rle_transp){

   dword *addr = new dword[size_y];
   byte *rle_data = new byte[size_y*(size_x*2+1)], *rle_data1 = rle_data;
   union{
      byte *bp;
      word *wp;
      dword *dwp;
      void *vp;
   };
   dword not_alpha = ~pf_dst.a_mask;

   for(int y=0; y<size_y; y++){
      addr[y] = (rle_data1 - rle_data) + size_y * 4; //save address
      vp = dst;

      bool last_opaque;
      int count;
      static const byte segment_mask[] = {0, 0x80};
      for(int x=0; x<size_x; x++){
         dword pixel;
         switch(pf_src.bytes_per_pixel){
         case 1: pixel = ((byte*)src)[x]; break;
         case 2: pixel = ((word*)src)[x]; break;
         case 3:
            pixel = *(word*)(((byte*)src)+x*3);
            pixel |= ((byte*)src)[x*3+2];
            break;
         case 4: pixel = ((dword*)src)[x]; break;
         }

         bool opaque = (pixel!=0);

         if(!(pf_dst.flags&PIXELFORMAT_COMPRESS))
         switch(pf_dst.bytes_per_pixel){
         case 1:
            *bp &= not_alpha;
            if(opaque) *bp |= pf_dst.a_mask;
            break;
         case 2:
            *wp &= not_alpha;
            if(opaque) *wp |= pf_dst.a_mask;
            break;
         case 3:
            {
               *wp &= not_alpha;
               bp[2] &= not_alpha>>16;
               if(opaque){
                  *wp |= pf_dst.a_mask;
                  bp[2] |= pf_dst.a_mask>>16;
               }
            }
            break;
         case 4:
            *dwp &= not_alpha;
            if(opaque) *dwp |= pf_dst.a_mask;
            break;
         }

         if(!x){
            last_opaque = opaque;
            count = 1;
         }else
         if(last_opaque != opaque){
            while(count>=127) *rle_data1++ = (segment_mask[last_opaque] | 0x7f), count -= 127;
            if(count) *rle_data1++ = (segment_mask[last_opaque] | count);
            last_opaque = opaque;
            count = 1;
         }else
            ++count;
         bp += pf_dst.bytes_per_pixel;
      }
      if(last_opaque){
         while(count>=127) *rle_data1++ = (0x80 | 0x7f), count -= 127;
         if(count) *rle_data1++ = (0x80 | count);
      }
                              //end line by zero byte
      *rle_data1++ = 0;
      src = ((byte*)src) + src_pitch;
      dst = ((byte*)dst) + dst_pitch;
   }
   rle_transp.reserve((rle_data1 - rle_data)+size_y*4);
                              //store addresses
   for(int i=0; i<size_y; i++){
      for(dword j=0, dw=addr[i]; j<4; j++, dw >>= 8)
         rle_transp.push_back(dw);
   }
                              //store rle masks
   for(byte *bp2 = rle_data; bp2<rle_data1; )
      rle_transp.push_back(*bp2++);

   delete []addr;
   delete []rle_data;
}

//----------------------------
                              //open image
bool IImage::Open(const char *name, dword flags, dword sx1, dword sy1,
   const S_pixelformat *pf, dword num_mip_levels){

   Close();

   int i;
                              //textures needs specified pixel format
   if((flags&IMGOPEN_TEXTURE) && !pf)
      return false;

   bool texture_convert = false;
   bool create_palette = false;

   void *area = NULL;

   S_pixelformat source_pf;
   dword conv_flags = 0;

   dword ckey0;

   if(!(flags&IMGOPEN_EMPTY)){
      if(flags&IMGOPEN_MEM_PTR){    //open from memory (BMP)
         BITMAPINFO *bi = (BITMAPINFO*)name;
         int num_clr = bi->bmiHeader.biClrUsed;
         if(bi->bmiHeader.biBitCount==8 && !num_clr) num_clr = 256;
         area = (void*)bi->bmiHeader.biCompression;
         if(!area){
            area = ((byte*)bi) + sizeof(BITMAPINFO) + sizeof(RGBQUAD) * (num_clr-1);
         }
         if(bi->bmiHeader.biBitCount==8){
            pal_argb = new dword[256];
            memset(pal_argb, -1, 256 * sizeof(dword));

            byte *p_pal = (byte*)pal_argb;
            RGBQUAD *pal1 = bi->bmiColors;
            for(int i=0; i<num_clr; i++){
               p_pal[i*4+0] = pal1[i].rgbRed;
               p_pal[i*4+1] = pal1[i].rgbGreen;
               p_pal[i*4+2] = pal1[i].rgbBlue;
               p_pal[i*4+3] = 0;
            }
         }
         size_x = bi->bmiHeader.biWidth;
         size_y = bi->bmiHeader.biHeight;
         memset(&source_pf, 0, sizeof(source_pf));
         source_pf.bytes_per_pixel = bi->bmiHeader.biBitCount/8;
         if(source_pf.bytes_per_pixel==1){
            source_pf.flags = PIXELFORMAT_PALETTE;
         }else{
                              //native bitmap format has BGR ordering
            source_pf.r_mask = 0xff0000;
            source_pf.g_mask = 0x00ff00;
            source_pf.b_mask = 0x0000ff;
         }
         conv_flags |= MCONV_FLIP_Y;
      }else{
                              //open file
         int file_handle = dtaOpen(name);
         if(file_handle==-1)
            return false;
                              //read part of header
         dword header;
         i = dtaRead(file_handle, &header, sizeof(header));
         if(i!=sizeof(header)){
            dtaClose(file_handle);
            return false;
         }
                              //scan through registered image loader
         for(i=load_plugins.size(); i--; ){
            bool b = load_plugins[i]->Load(file_handle, name, &header, sizeof(header),
               &area, (void**)&pal_argb, &size_x, &size_y, &source_pf);
            if(b)
               break;
         }
         if(i==-1){
                              //cannot load file, exit
            dtaClose(file_handle);
            return false;
         }
      }
   }else{
      if(!sx1 || !sy1 || !pf)
         return false;
      size_x = sx1, size_y = sy1;
      pal_argb = new dword[256];
   }
   if(flags&IMGOPEN_NOREMAP){
      if(flags&IMGOPEN_EMPTY){
         if(!pf)
            return false;
         pixel_format = *pf;
      }else
         pixel_format = source_pf;
   }else{
      if(pf)
         pixel_format = *pf;
   }
   LPC_rgb_conversion rgb_conv = NULL;

   {
                              //create surface
      HRESULT hr;

      if(flags&IMGOPEN_TEXTURE){
                              //vidmem texture surface
         if(!(flags&IMGOPEN_EMPTY) && !(flags&IMGOPEN_NOREMAP)){
                                 //convert to specified pixel format
                                 //or let copy
            rgb_conv = igraph->GetRGBConv1(*pf);
            texture_convert = true;
                                 //get color key
            if(source_pf.bytes_per_pixel==1){
               if(pf->bytes_per_pixel==1 && (pf->flags&PIXELFORMAT_PALETTE))
               {
                                 //dest is 8-bit pal
                  ckey0 = 0;
               }else{
                                 //map palette color 0 to RGB
                  if(pal_argb){
                     byte *b_pal = (byte*)pal_argb;
                     ckey0 = rgb_conv->GetPixel(b_pal[0], b_pal[1], b_pal[2], 0);
                  }else
                     ckey0 = 0;
               }
            }else{            //source is RGB, use superblack
               ckey0 = 0;
            }
         }

         dword num_levels = 1;
         if(flags&IMGOPEN_MIPMAP){
            num_levels = num_mip_levels + 1;
         }

         D3DFORMAT d3d_fmt;
   
         if(flags&IMGOPEN_NOREMAP){
            d3d_fmt = ConvertFormat(&pixel_format);
         }else
         if(!(pf->flags&PIXELFORMAT_COMPRESS)){
            d3d_fmt = ConvertFormat(pf);
         }else{
            d3d_fmt = ConvertFormat(pf);
         }
         if(d3d_fmt==D3DFMT_P8)
            create_palette = true;

         D3DPOOL pool_type = D3DPOOL_DEFAULT;

         if(flags&IMGOPEN_SYSMEM){
            pool_type = D3DPOOL_SYSTEMMEM;
         }
         dword usage_flags = 0;

         if(flags&IMGOPEN_RENDERTARGET)
            usage_flags |= D3DUSAGE_RENDERTARGET;

         hr = igraph->GetD3DDevice()->CreateTexture(size_x, size_y, num_levels,
            usage_flags,
            d3d_fmt,
            pool_type,
            (IDirect3DTexture8**)&lpTxt8);
      }else{
         D3DFORMAT d3d_fmt;
         if((flags&IMGOPEN_NOREMAP) || pf){
            if(!pf){
               d3d_fmt = ConvertFormat(&pixel_format);
               create_palette = (pixel_format.bytes_per_pixel == 1);
            }else{
               d3d_fmt = ConvertFormat(pf);
               create_palette = (pf->bytes_per_pixel == 1);
            }
         }else{
            d3d_fmt = ConvertFormat(igraph->GetPixelFormat());
            
            create_palette = (igraph->GetPixelFormat()->bytes_per_pixel == 1);
         }

         hr = igraph->GetD3DDevice()->CreateImageSurface(size_x, size_y, d3d_fmt,
            (IDirect3DSurface8**)&lpImg8);
         CHECK_D3D_RESULT("CreateImageSurface", hr);
      }
      if(FAILED(hr))
         goto fail;

      if(!(flags&IMGOPEN_EMPTY)){
                              //fill-in surface contents
         void *mem_ptr;
         dword pitch;
         bool b = Lock(&mem_ptr, &pitch, NULL);
         if(!b)
            goto fail;
                              //copy bytes into surface
         if(texture_convert){
            i = rgb_conv->Convert(area, mem_ptr, size_x, size_y,
               size_x*source_pf.bytes_per_pixel, pitch, source_pf.bytes_per_pixel, &pal_argb,
               conv_flags |
               ((flags&IMGOPEN_DITHER) ? MCONV_DITHER : 0) |
               ((flags&IMGOPEN_COLORKEY) ? MCONV_COLORKEY : 0),
               ckey0);
            if(!i)
               goto fail;
         }else
         if(!(flags&(IMGOPEN_NOREMAP | IMGOPEN_EMPTY))){
            LPC_rgb_conversion rc;
            if(pf){
               assert(!rgb_conv);
               rgb_conv = igraph->GetRGBConv1(pixel_format);
               rc = rgb_conv;
            }else{
               rc = igraph->GetRGBConv();
            }
            i = rc->Convert(area, mem_ptr, size_x, size_y,
               size_x*source_pf.bytes_per_pixel, pitch, source_pf.bytes_per_pixel, &pal_argb,
               conv_flags |
               ((flags&IMGOPEN_DITHER) ? MCONV_DITHER : 0));
            if(!i)
               goto fail;
            pixel_format = *rc->GetPixelFormat();
         }else{
            int sy;
            for(sy=0; sy<size_y; sy++){
               memcpy(((byte*)mem_ptr) + sy*pitch,
                  ((byte*)area)+sy*size_x*pixel_format.bytes_per_pixel, size_x*pixel_format.bytes_per_pixel);
            }
         }
                              //initialize colorkey info
         if(flags&IMGOPEN_INIT_CKEY){
            const void *src = area;
            dword src_pitch = size_x*source_pf.bytes_per_pixel;
            if(conv_flags&MCONV_FLIP_Y){
               (byte*&)src += (size_y-1)*src_pitch;
               src_pitch = -src_pitch;
            }
            InitColorkeyInfo(src, mem_ptr, size_x, size_y, src_pitch, pitch,
               source_pf, pixel_format, ckey_rle_info);
         }
         Unlock();

         if(!(flags&IMGOPEN_MEM_PTR))
            delete[] area;    //free temp memory
      }
      if(flags&IMGOPEN_CLEAR){
         assert(0);
      }

      if(create_palette){
         if(pal_argb && !(flags&IMGOPEN_EMPTY)){
         }else{
            if(!pal_argb)
               pal_argb = new dword[256];
                              //setup uniform palette
            for(i=0; i<256; i++){
               ((byte*)pal_argb)[i*4 + 0] = (i) & 0xe0;
               ((byte*)pal_argb)[i*4 + 1] = (i>>3) & 0x1c;
               ((byte*)pal_argb)[i*4 + 2] = (i>>6) & 0x3;
               ((byte*)pal_argb)[i*4 + 3] = 0xff;
            }
         }
      }else{
         delete[] pal_argb;
         pal_argb = NULL;
      }
   }

   if(rgb_conv)
      rgb_conv->Release();

//----------------------------
   if(!(flags&IMGOPEN_TEXTURE)){
      open_fname = name;
      open_flags = flags;
   }

   return true;

fail:
   delete[] pal_argb;
   if(!(flags&IMGOPEN_MEM_PTR))
      delete[] area;
   lpTxt8 = NULL;
   lpImg8 = NULL;
   if(rgb_conv)
      rgb_conv->Release();

   return false;
}

//----------------------------
                              //close image
void IImage::Close(){

   if(IsInit()){
      lpTxt8 = NULL;
      lpImg8 = NULL;

      delete[] pal_argb;
      pal_argb = NULL;
   }
   open_fname = NULL;        
}

//----------------------------

bool IImage::IsInit() const{
   
   return (lpTxt8!=NULL || lpImg8!=NULL);
}

//----------------------------

const S_pixelformat *IImage::GetPixelFormat() const{
   
   return IsInit() ?
      &pixel_format :
      NULL;
}

//----------------------------

bool IImage::SetPal(dword *pal1){

   if(pal_argb && pal1){
      memcpy(pal_argb, pal1, sizeof(dword)*256);
      return true;
   }
   return false;
}

//----------------------------
                              //draw image to backbuffer
bool IImage::Draw(int dst_x, int dst_y, dword flags){

   if(!IsInit() || !lpImg8){
      return false;
   }

                              //clip
   dword src_x = 0, src_y = 0;
   if(dst_x<igraph->vport[0])
      src_x = igraph->vport[0]-dst_x, dst_x = igraph->vport[0];
   if(dst_y<igraph->vport[1])
      src_y = igraph->vport[1]-dst_y, dst_y = igraph->vport[1];
   int sx = dst_x + size_x - src_x, sy = dst_y + size_y - src_y;
   if(sx>igraph->vport[2]) sx = igraph->vport[2];
   if(sy>igraph->vport[3]) sy = igraph->vport[3];
   if((sx -= dst_x)<=0 || (sy -= dst_y)<=0)
      return true;

   RECT rc;
   rc.left = src_x;
   rc.top = src_y;
   rc.right = src_x+sx;
   rc.bottom = src_y+sy;

   HRESULT hr;

   IDirect3DDevice8 *lpDev8 = igraph->GetD3DDevice();

   IDirect3DSurface8 *lpBack;
   hr = lpDev8->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &lpBack);
   CHECK_D3D_RESULT("GetBackBuffer", hr);
   if(SUCCEEDED(hr)){
      POINT pt = {dst_x, dst_y};
      hr = lpDev8->CopyRects(lpImg8, &rc, 1, lpBack, &pt);
      lpBack->Release();
      CHECK_D3D_RESULT("CopyRects", hr);
      return true;
   }
   return false;
}

//----------------------------

bool IImage::Lock(void **area1, dword *pitch, byte *bpp1){

   if(!IsInit())
      return false;

   D3DLOCKED_RECT lrc;
   HRESULT hr;
   if(lpTxt8){
      hr = lpTxt8->LockRect(0,
         &lrc,
         NULL,
         D3DLOCK_NO_DIRTY_UPDATE | D3DLOCK_READONLY);
   }else{
      hr = lpImg8->LockRect(&lrc,
         NULL,
         D3DLOCK_NO_DIRTY_UPDATE | D3DLOCK_READONLY);
   }
   CHECK_D3D_RESULT("LockRect", hr);
   if(FAILED(hr))
      return false;
   if(area1) *area1 = lrc.pBits;
   if(pitch){
      if(!(pixel_format.flags&PIXELFORMAT_COMPRESS))
         *pitch = lrc.Pitch;
      else{
         *pitch = size_x * size_y / 2;
         if(pixel_format.four_cc != MAKEFOURCC('D', 'X', 'T', '1'))
            *pitch *= 2;
      }
   }
   if(bpp1){
      if(!(pixel_format.flags&PIXELFORMAT_COMPRESS))
         *bpp1 = pixel_format.bytes_per_pixel;
      else
         *bpp1 = 0;
   }
   return true;
}

//----------------------------

bool IImage::Unlock(){

   if(!IsInit())
      return false;

   HRESULT hr;
   if(lpTxt8)
      hr = lpTxt8->UnlockRect(0);
   else
      hr = lpImg8->UnlockRect();

   CHECK_D3D_RESULT("UnlockRect", hr);
   return (SUCCEEDED(hr));
}

//----------------------------

bool IImage::AlphaBlend(LPIImage mix_pic, dword flags){

   if(!IsInit() || !mix_pic || !mix_pic->IsInit()){
      assert(0);
      return false;
   }

   bool b = false;

   void *dst_a;
   dword dst_pitch;
   b = Lock(&dst_a, &dst_pitch, NULL);
   if(!b)
      return false;

   void *src_a;
   dword src_pitch;
   byte src_bpp;
   b = mix_pic->Lock(&src_a, &src_pitch, &src_bpp);
   if(!b){
      Unlock();
      return false;
   }
   LPC_rgb_conversion rgb_conv = igraph->GetRGBConv1(pixel_format);

                              //blend
   b = rgb_conv->AlphaBlend(dst_a, src_a, mix_pic->GetPal(),
      min(size_x, mix_pic->SizeX()), min(size_y, mix_pic->SizeY()),
      src_pitch, dst_pitch, src_bpp,
      (flags&IMGBLEND_DITHER ? MCONV_DITHER : 0) |
      (flags&IMGBLEND_INVALPHA ? MCONV_INVALPHA : 0));

   rgb_conv->Release();
   mix_pic->Unlock();
   Unlock();

   return b;
}

//----------------------------

struct S_argb{
   byte b, g, r, a;
};

static void ShrinkMipmap(S_argb *src, dword sx1, dword sy){

   S_argb *dst = src, *src1 = src + sx1 * 2;
   do{
      for(dword sx=0; sx<sx1; sx++){
         dst->r = (src[0].r + src[1].r +src1[0].r +src1[1].r) / 4;
         dst->g = (src[0].g + src[1].g +src1[0].g +src1[1].g) / 4;
         dst->b = (src[0].b + src[1].b +src1[0].b +src1[1].b) / 4;
         dst->a = (src[0].a + src[1].a +src1[0].a +src1[1].a) / 4;
         ++dst;
         src += 2;
         src1 += 2;
      }
      src += sx1*2;
      src1 += sx1*2;
   }while(--sy);
}

//----------------------------

inline int __cdecl FindLastBit(dword val){
   dword rtn;
   __asm{
      mov eax, val
      bsr eax, eax
      jnz ok
      mov eax, -1
ok:
      mov rtn, eax
   }
   return rtn;
}

inline dword __cdecl Rol(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      rol dw, cl
      pop ecx
   }
   return dw;
}
inline dword __cdecl Ror(dword dw, byte b){
   __asm{
      push ecx
      mov cl, b
      ror dw, cl
      pop ecx
   }
   return dw;
}

//----------------------------

bool IImage::MixMipmap(dword flags){

   if(!IsInit() || !lpTxt8){
      assert(0);
      return false;
   }

   dword num_levels = lpTxt8->GetLevelCount();
   if(num_levels<2)
      return true;

   HRESULT hr;

   D3DLOCKED_RECT lrc;
                              //lock surface - 1st LOD
   hr = lpTxt8->LockRect(0, &lrc, NULL, D3DLOCK_NO_DIRTY_UPDATE | D3DLOCK_READONLY);
   CHECK_D3D_RESULT("LockRect", hr);
   if(FAILED(hr))
      return false;

                              //convert source into rgba surface
   S_argb *area = new S_argb[size_x * size_y];
   if(!area){
      lpTxt8->UnlockRect(0);
      return false;
   }
   S_argb *tp = area;
   union{
      byte *src_bp;
      word *src_wp;
      void *src_vp;
   };
   src_vp = lrc.pBits;

   dword xx, yy;
   switch(pixel_format.bytes_per_pixel){
   case 1:
      if(pal_argb){
         for(yy=0; yy<size_y; yy++, src_bp+=lrc.Pitch)
         for(xx=0; xx<size_x; xx++, ++tp){
            byte *palp = (byte*)&pal_argb[src_bp[xx]];
            tp->r = palp[0];
            tp->g = palp[1];
            tp->b = palp[2];
         }
      }else{
         byte r_pos = (FindLastBit(pixel_format.r_mask)-7) & 31;
         byte g_pos = (FindLastBit(pixel_format.g_mask)-7) & 31;
         byte b_pos = (FindLastBit(pixel_format.b_mask)-7) & 31;
         byte a_pos = (FindLastBit(pixel_format.a_mask)-7) & 31;
         for(yy=0; yy<size_y; yy++, src_bp+=lrc.Pitch)
         for(xx=0; xx<size_x; xx++, ++tp){
            byte pixel = src_bp[xx];
            tp->r = Ror(pixel&pixel_format.r_mask, r_pos);
            tp->g = Ror(pixel&pixel_format.g_mask, g_pos);
            tp->b = Ror(pixel&pixel_format.b_mask, b_pos);
            tp->a = Ror(pixel&pixel_format.a_mask, a_pos);
         }
      }
      break;
   case 2:
      {
         byte r_pos = (FindLastBit(pixel_format.r_mask)-7) & 31;
         byte g_pos = (FindLastBit(pixel_format.g_mask)-7) & 31;
         byte b_pos = (FindLastBit(pixel_format.b_mask)-7) & 31;
         byte a_pos = (FindLastBit(pixel_format.a_mask)-7) & 31;
#if defined _MSC_VER && 1
         if(!(size_x&1)){
            dword r_mask = pixel_format.r_mask;
            dword g_mask = pixel_format.g_mask;
            dword b_mask = pixel_format.b_mask;
            dword a_mask = pixel_format.a_mask;
            byte *src_bp1 = src_bp;
            dword sx = size_x, sy = size_y;
            __asm{
               push ecx
                              //eax = free
                              //ebx = free
                              //ecx = free
                              //edx = src_bp
                              //esi = src_bp_1
                              //edi = tp
                              //esp[0] = counter x
                              //esp[4] = counter y
               push sy
               mov edx, src_bp1
               mov edi, tp

            loop_y:
               push sx
               mov esi, edx
               shr dword ptr[esp], 1   //process 2 pixels at a time

            loop_x:
               mov eax, [esi]

                              //RED
               mov cl, r_pos
               mov ebx, eax
               add esi, 4
               and ebx, r_mask
               ror ebx, cl
               mov [edi+2], bl

                              //GREEN
               mov cl, g_pos
               mov ebx, eax
               and ebx, g_mask
               ror ebx, cl
               mov [edi+1], bl

                              //BLUE
               mov cl, b_pos
               mov ebx, eax
               and ebx, b_mask
               ror ebx, cl
               mov [edi+0], bl

                              //ALPHA
               mov cl, a_pos
               mov ebx, eax
               and ebx, a_mask
               ror ebx, cl
               shr eax, 16    //next vertex
               mov [edi+3], bl

                              //RED
               mov cl, r_pos
               mov ebx, eax
               and ebx, r_mask
               ror ebx, cl
               mov [edi+6], bl

                              //GREEN
               mov cl, g_pos
               mov ebx, eax
               and ebx, g_mask
               ror ebx, cl
               mov [edi+5], bl

                              //BLUE
               mov cl, b_pos
               mov ebx, eax
               and ebx, b_mask
               ror ebx, cl

                              //ALPHA
               and eax, a_mask
               mov cl, a_pos
               mov [edi+4], bl
               ror eax, cl
               mov [edi+7], al


               add edi, 8
               dec dword ptr[esp]
               jnz loop_x
               pop eax
 
               add edx, lrc.Pitch
               dec dword ptr[esp]
               jnz loop_y
               pop eax

               pop ecx
            }
         }else
#endif
         {
            for(yy=0; yy<size_y; yy++, src_bp+=lrc.Pitch)
            for(xx=0; xx<size_x; xx++, ++tp){
               word pixel = src_wp[xx];
               tp->r = Ror(pixel&pixel_format.r_mask, r_pos);
               tp->g = Ror(pixel&pixel_format.g_mask, g_pos);
               tp->b = Ror(pixel&pixel_format.b_mask, b_pos);
               tp->a = Ror(pixel&pixel_format.a_mask, a_pos);
            }
         }
      }
      break;
   case 3:
      for(yy=0; yy<size_y; yy++, src_bp+=lrc.Pitch){
         byte *bp1 = src_bp;
         for(xx=0; xx<size_x; xx++, ++tp, bp1 += 3){
            tp->r=bp1[0];
            tp->g=bp1[1];
            tp->b=bp1[2];
         }
      }
      break;
   case 4:
      {
         for(yy=0; yy<size_y; yy++, tp+=size_x, src_bp+=lrc.Pitch)
            memcpy(tp, src_bp, size_x * 4);
      }
      break;
   }

                              //unlock surface
   hr = lpTxt8->UnlockRect(0);
   CHECK_D3D_RESULT("UnlockRect", hr);

   LPC_rgb_conversion conv = igraph->GetRGBConv1(pixel_format, pal_argb);

   dword sx = size_x, sy = size_y;
   for(int i=1; i<num_levels; i++){
                              //lock surface LOD
      hr = lpTxt8->LockRect(i, &lrc, NULL, 0);
      CHECK_D3D_RESULT("LockRect", hr);
      if(FAILED(hr))
         break;

      sx /= 2, sy /= 2;
      ShrinkMipmap(area, sx, sy);
                              //convert back into surface
      conv->Convert(area, lrc.pBits, sx, sy, sx*4, lrc.Pitch, 4, &pal_argb, 0);
                              //unlock surface
      hr = lpTxt8->UnlockRect(i);
      CHECK_D3D_RESULT("UnlockRect", hr);
   }
   delete[] area;
   conv->Release();
   return true;
}

//----------------------------

bool IImage::ConvertBumpMap(LPIImage src_img, dword flags){

   if(!IsInit() || !src_img || !src_img->IsInit()){
      assert(0);
      return false;
   }
   if(src_img->SizeX()!=SizeX() || src_img->SizeY()!=SizeY()){
      assert(0);
      return false;
   }
   bool b;

   void *src_a;
   dword src_pitch;
   byte src_bpp;
   b = src_img->Lock(&src_a, &src_pitch, &src_bpp);
   if(!b){
      return false;
   }
                              //convert source map to RGB format
   struct S_rgb{
      byte b, g, r;
   };
   S_rgb *src_map = new S_rgb[size_x * size_y];

   S_pixelformat pf;
   memset(&pf, 0, sizeof(pf));
   pf.bytes_per_pixel = 3;
   pf.r_mask = 0x00ff0000;
   pf.g_mask = 0x0000ff00;
   pf.b_mask = 0x000000ff;
   LPC_rgb_conversion conv = igraph->GetRGBConv1(pf);

   dword *pal = src_img->GetPal();
   conv->Convert(src_a, src_map, size_x, size_y, src_pitch, 0, src_bpp, &pal, 0, NULL);
   conv->Release();
   src_img->Unlock();

   struct S_uv{
      byte l, v, u;
   };
   S_uv *dst_map = new S_uv[size_x * size_y];

                              //convert bump map
   const S_rgb *rgb_line_base = src_map;
   S_uv *uv_line_base = dst_map;
   for(int y=0; y<size_y; y++, rgb_line_base += size_x, uv_line_base += size_x){
      const S_rgb *rgb_line = rgb_line_base;
      const S_rgb *rgb_up = !y ? (rgb_line_base + size_x * (size_y-1)) : (rgb_line_base - size_x);
      const S_rgb *rgb_down = (y==size_y-1) ? (rgb_line_base - size_x * (size_y-1)) : (rgb_line_base + size_x);
      S_uv *dst_line = uv_line_base;

      for(int x=0; x<size_x; x++, ++rgb_up, ++rgb_down, ++rgb_line, ++dst_line){
                              //get ARGB values of four neighbours
         const S_rgb *rgb_left = !x ? (rgb_line + (size_x - 1)) : (rgb_line - 1);
         const S_rgb *rgb_right = (x==size_x-1) ? (rgb_line - (size_x - 1)) : (rgb_line + 1);

         byte bright_l = (rgb_left->r + rgb_left->g + rgb_left->b) / 3;
         byte bright_t = (rgb_up->r + rgb_up->g + rgb_up->b) / 3;
         byte bright_r = (rgb_right->r + rgb_right->g + rgb_right->b) / 3;
         byte bright_b = (rgb_down->r + rgb_down->g + rgb_down->b) / 3;

         dst_line->l = 0xff;
         dst_line->u = bright_r - bright_l;
         dst_line->v = bright_b - bright_t;
      }
   }
   delete[] src_map;

   void *dst_a;
   dword dst_pitch;
   byte dst_bpp;
   b = Lock(&dst_a, &dst_pitch, &dst_bpp);
   if(!b){
      delete[] dst_map;
      return false;
   }
                              //use convertor to map UV data to texture
                              // (we fake the convertor to think it converts RGB data)
   LPC_rgb_conversion conv1 = igraph->GetRGBConv1(pixel_format);
   conv1->Convert(dst_map, dst_a, size_x, size_y, size_x*sizeof(S_uv), dst_pitch,
      3, NULL, 0, NULL);
   conv1->Release();

   delete[] dst_map;

   Unlock();

   return true;
}

//----------------------------

bool IImage::CopyStretched(LPIImage src_img){

   if(memcmp(&pixel_format, src_img->GetPixelFormat(), sizeof(pixel_format)))
      return false;

   void *area_src, *area_dst;
   dword src_pitch, dst_pitch;
   byte bpp;
   if(!src_img->Lock(&area_src, &src_pitch, &bpp))
      return false;
   if(!Lock(&area_dst, &dst_pitch, NULL)){
      src_img->Unlock();
      return false;
   }

   float ratio_x = (float)src_img->SizeX() / (float)SizeX(), count_x;
   float ratio_y = (float)src_img->SizeY() / (float)SizeY(), count_y = ratio_y * .4999f;
   const byte *src = (const byte*)area_src;

   for(int y = SizeY(); y--; area_dst = (byte*)area_dst + dst_pitch){
      assert(src < (((const byte*)area_src) + src_pitch * src_img->SizeY()));

      const byte *src_x = src;
      byte *dst_x = (byte*)area_dst;
      count_x = ratio_x * .4999f;
      for(int x = SizeX(); x--; dst_x += bpp){

         switch(bpp){
         case 1: *dst_x = *src_x; break;
         case 2: *(word*)dst_x = *(word*)src_x; break;
         case 3: *(word*)dst_x = *(word*)src_x; ((byte*)dst_x)[2] = ((byte*)src_x)[2]; break;
         case 4: *(dword*)dst_x = *(dword*)src_x; break;
         }
                              //next row
         count_x += ratio_x;
         int i = FloatToInt(floor(count_x));
         src_x += bpp * i;
         count_x -= (float)i;
      }
                              //next line
      count_y += ratio_y;
      int i = FloatToInt(floor(count_y));
      src += src_pitch * i;
      count_y -= (float)i;
   }
   Unlock();
   src_img->Unlock();

   return true;
}

//----------------------------
//----------------------------

