#include <vector>
#include <c_string.hpp>
#include <i3d\i3d_math.h>     //due to S_vector
#include <c_cache.h>
#ifdef _DEBUG
#include <exception>
#endif

#ifdef _MSC_VER
using namespace std;
#endif

//----------------------------
//----------------------------

#ifdef _DEBUG

#define THROW(what) throw(exception(what))

#else

#define THROW(what)

#endif

//----------------------------

typedef byte ENUM_TYPE;

//----------------------------

                              //table template definition
class C_table_template{
public:
   byte guid[16];
   char *caption;
   char *help_text;
   C_table_element te[1];

   const void *GetGUID() const{ return guid; }
   const C_table_element *GetElements() const{ return te; }
   const char *GetCaption() const{ return caption; }
   bool IsDataType(int indx) const{
      switch(te[indx].type){
      case TE_NULL: case TE_BRANCH: case TE_ARRAY: return false;
      default: return true;
      }
   };
   int GetItemTableIndex(int indx) const{
      return te[indx].tab_index;
   }
   int ArraySize(int indx) const{ return te[indx].min_array_size; }
   int BranchDepth(int indx) const{ return te[indx].branch_depth; }
   int NumItems(bool count_all) const{
      int i=0, num=0;
      while(te[i].type!=TE_NULL){
         if(count_all || IsDataType(i)) ++num;
         ++i;
      }
      return num;
   };
   E_TABLE_ELEMENT_TYPE GetItemType(int indx) const{ return te[indx].type; }
   int SizeOf(int indx) const{
      static const byte sizes[TE_LAST]={
         0, sizeof(bool), sizeof(int), sizeof(float),
         sizeof(ENUM_TYPE), sizeof(char), sizeof(byte)*3, sizeof(float)*3
      };
      switch(te[indx].type){
      case TE_ARRAY: return te[indx].min_array_size * SizeOf(indx+1);
      case TE_STRING: return te[indx].default_string_size;
      case TE_BRANCH: return 0;
      case TE_NULL: return 0;
      default: return sizes[te[indx].type];
      }
   };
   dword GetDefaultVal(int indx){
      return te[indx].int_default;
   }
   void *operator new(size_t sz, int i){
      return new byte[sz + sizeof(C_table_element)*(i-1)];
   }
};

//----------------------------

class C_table_base{
public:
   TABMETHOD_(dword,AddRef)() = 0;
   TABMETHOD_(dword,Release)() = 0;

   TABMETHOD_(bool,Open)(dword dw_open, dword flags) = 0;
   TABMETHOD_(bool,Save)(dword dw_save, dword flags) = 0;
   TABMETHOD_(bool,IsOpen)() const = 0;

   TABMETHOD_(int,NumItems)() const = 0;
   TABMETHOD_(int,ArrayLen)(int index) const = 0;
   TABMETHOD_(int,StringSize)(int index) const = 0;
   TABMETHOD_(int,SizeOf)(int index) const = 0;
   TABMETHOD_(void*,Item)(int index) = 0;
                              //get various items
   TABMETHOD_(bool&,ItemB)(int index) = 0;
   TABMETHOD_(int&,ItemI)(int index) = 0;
   TABMETHOD_(float&,ItemF)(int index) = 0;
   TABMETHOD_(byte&,ItemE)(int index) = 0;
   TABMETHOD_(char*,ItemS)(int index) = 0;
   TABMETHOD_(struct S_vector&,ItemV)(int index) = 0;
                              //get various items from arrays
   TABMETHOD_(bool&,ItemB)(int index, int ai) = 0;
   TABMETHOD_(int&,ItemI)(int index, int ai) = 0;
   TABMETHOD_(float&,ItemF)(int index, int ai) = 0;
   TABMETHOD_(byte&,ItemE)(int index, int ai) = 0;
   TABMETHOD_(char*,ItemS)(int index, int ai, int str_size) = 0;
   TABMETHOD_(struct S_vector&,ItemV)(int index, int ai) = 0;

   TABMETHOD_(E_TABLE_ELEMENT_TYPE,GetItemType)(int index) const = 0;
   TABMETHOD_(void*,GetGUID)() = 0;

                              //table editor - returns HWND
   TABMETHOD_(void*,Edit)(const class C_table_template*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL) = 0;

   TABMETHOD_(void*,Edit)(C_table*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL) = 0;
   TABMETHOD_(bool,IsModified)() const = 0;
};

//----------------------------

class C_table: public C_table_base{
                              //C_unk:
   dword ref;

   bool lock;
   bool modified;
   dword num_items;
   byte guid[16];
#pragma pack(push,1)
   struct S_desc_item{
      dword offset;
      byte type;              //real type: E_TABLE_ELEMENT_TYPE
      union{
         byte max_string_size;   //for TE_STRING
      };
      word array_len;
      S_desc_item():
         offset(-1),
         type(TE_NULL),
         max_string_size(0),
         array_len(0)
      {}
   } *desc;
#pragma pack(pop)
   byte *data;                //pointer to table data
   dword data_size;
                              //save header
#define HDRFLAGS_COMPRESS     1
#define HDRFLAGS_MODIFY_DATA  2     //format of modify data: null-terminated string, word year, month, day, hour, minute
   struct S_file_header{
      dword ID;
      bool lock;
      dword num_items, data_size;
      dword flags;
      dword reserved;
   };

   friend LPC_table __declspec(dllexport) TABAPI CreateTable();

   C_table():
      ref(1),
      lock(false),
      modified(false),
      num_items(0),
      desc(NULL),
      data(NULL)
   {
      memset(&guid, 0, sizeof(guid));
   }
   ~C_table(){
      Close();
   }

   void Close();
   int InsertItem(LPC_table_template tt, int templ_indx, int &storage,
      vector<byte> &v_defaults, int array_len=0);

public:
   TABMETHOD_(dword,AddRef)(){ return ++ref; }
   TABMETHOD_(dword,Release)(){ if(--ref) return ref; delete this; return 0; }

   TABMETHOD_(bool,Open)(dword dw_open, dword flags);
   TABMETHOD_(bool,Save)(dword dw_save, dword flags);
   TABMETHOD_(bool,IsOpen)() const{
      return (desc && data);
   }

   TABMETHOD_(int,NumItems)() const{ return num_items; }
   TABMETHOD_(int,ArrayLen)(int indx) const{ return !desc ? -1 : indx>=num_items ? 0 : desc[indx].array_len; }
   TABMETHOD_(int,StringSize)(int indx) const{
      return !desc ? 0 : indx>=num_items ? 0 :
      desc[indx].type==TE_STRING ? desc[indx].max_string_size : 0;
   }
   TABMETHOD_(int,SizeOf)(int indx) const{
      if(!desc || indx>=num_items) return 0;
      static const byte sizes[TE_LAST]={
         0, sizeof(bool), sizeof(int), sizeof(float),
         sizeof(ENUM_TYPE), sizeof(char), sizeof(byte)*3, sizeof(float)*3
      };
      E_TABLE_ELEMENT_TYPE type = (E_TABLE_ELEMENT_TYPE)desc[indx].type;
      return type==TE_STRING ? desc[indx].max_string_size : sizes[type];
   }

   TABMETHOD_(void*,Item)(int indx){
      assert(desc && data);
      if(!desc || !data){
         THROW(C_fstring("C_table::Item(int): reading from uninitialized table"));
         return NULL;
      }
      assert(indx>=0 && indx<num_items);
      if(indx>=num_items){
         THROW(C_fstring("C_table::Item(int): reading out of range (index = %i, num_items = %i)", indx, num_items));
         return NULL;
      }
      int offset = desc[indx].offset;
      if(offset==-1){
         THROW(C_fstring("C_table::Item(int): reading from uninitialized index (%i)", indx));
         return NULL;
      }
      return data + offset;
   }
                              //get various items
   bool  &TABAPI ItemB(int i){ return *(bool*)C_table::Item(i); }
   int   &TABAPI ItemI(int i){ return *(int*)C_table::Item(i); }
   float &TABAPI ItemF(int i){ return *(float*)C_table::Item(i); }
   byte  &TABAPI ItemE(int i){ return *(byte*)C_table::Item(i); }
   char  *TABAPI ItemS(int i){ return (char*)C_table::Item(i); }
   S_vector &TABAPI ItemV(int i){ return *(S_vector*)C_table::Item(i); }

                           //get various items from arrays
   bool &TABAPI ItemB(int i, int ai){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemB(int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(bool*)NULL;
      }
      return ((bool*)vp)[ai];
   }
   int &TABAPI ItemI(int i, int ai){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemI(int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(int*)NULL;
      }
      return ((int*)vp)[ai];
   }
   float &TABAPI ItemF(int i, int ai){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemF(int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(float*)NULL;
      }
      return ((float*)vp)[ai];
   }
   byte &TABAPI ItemE(int i, int ai){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemE(int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(byte*)NULL;
      }
      return ((byte*)vp)[ai];
   }
   char *TABAPI ItemS(int i, int ai, int str_size){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemS(int, int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(char**)NULL;
      }
      return ((char*)vp)+ai*str_size;
   }
   S_vector &TABAPI ItemV(int i, int ai){
      void *vp = Item(i);
      if(!vp || ai>=desc[i].array_len){
         THROW(C_fstring("C_table::ItemV(int, int): reading out of range (index = %i, array_len = %i)", ai, desc[i].array_len));
         return *(S_vector*)NULL;
      }
      return ((S_vector*)vp)[ai];
   }


   TABMETHOD_(E_TABLE_ELEMENT_TYPE,GetItemType)(int indx) const{
      return !IsOpen() ? TE_NULL : indx>=num_items ? TE_NULL : (E_TABLE_ELEMENT_TYPE)desc[indx].type;
   }
   TABMETHOD_(void*,GetGUID)(){ return guid; }

#ifdef TABBLER_EDIT_SUPPORT
                              //table editor - returns HWND
   TABMETHOD_(void*,Edit)(const class C_table_template*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL);

   TABMETHOD_(void*,Edit)(C_table*, void *hwnd_parent,
      table_callback cb_proc, dword cb_user, dword create_flags = 0,
      const char *def_table = NULL, int *pos_size = NULL);
#else
   TABMETHOD_(void*,Edit)(const class C_table_template*, void*,
      table_callback, dword, dword, const char*, int*){
      return NULL;
   }

   TABMETHOD_(void*,Edit)(C_table*, void*, table_callback, dword, dword,
      const char *, int*){
      return NULL;
   }
#endif

   TABMETHOD_(bool,IsModified)() const;

//----------------------------

                              //history - just skip data in file (no more saved)
   void ReadModifyData(C_cache *ch, const S_file_header &hdr){

      if(hdr.flags&HDRFLAGS_MODIFY_DATA){
         char buf[256];
         char c;
         int i = 0;
         do{
            ch->read(&c, 1);
            buf[i++] = c;
         }while(c);

         struct{
            word wYear, wMonth, wDay, wHour, wMinute;
         } modify_time;
         ch->read((byte*)&modify_time, sizeof(modify_time));
      }
   }

   inline void SetModify(){ modified = true; }
};

//----------------------------
//----------------------------
