#include <inet2.h>
#include <vector>

//----------------------------
//----------------------------
                              //debugging support
#ifdef _DEBUG
#define NET_LOG(n) if(net) net->Log(n)
#else
#define NET_LOG(n)
#endif

//----------------------------
#define NET_MAX_PLAYERS 4

//----------------------------
// Connect to network.
bool NetConnect();

//----------------------------
// Send network messages.
// If 'immediate' is specified, no throtling is done and messages are sent now.
bool NetSend(bool immediate = true);

//----------------------------
// Process network messages - receive, dispatch, send out.
void NetProcessMessages();

//----------------------------
// Update network statistics and pass to editor.
void NetStatUpdate();

//----------------------------
// Disconnect from network game, release all resources.
void NetCloseAll(bool allow_suspend = true);

//----------------------------
                              //global interfaces
extern PI_net net;
extern PC_net_out_msg net_msg;


//----------------------------

                              //network synchronization class - new implementation
class C_net_sync: public C_unknown{
public:
   enum E_SYNC_STATE{
      SS_NULL,                //normal state
      SS_PROGRESS,            //in progress - solving sync event
      SS_RESOLVED,            //event solved
   };
//----------------------------
// Get current state of class.
// If returned value is SS_RESOLVED, the 'winner_id' contains winning player's net ID,
// and 'event_id' contains event ID of the winner.
   virtual E_SYNC_STATE GetState(dword *winner_pid = NULL, dword *event_id = NULL) const = 0;

//----------------------------
// Request sync event.
// The request fails if:
//    - class is locked
//    - state is not SS_NULL
   virtual bool BeginEvent(word event_id = 0) = 0;

//----------------------------
// Reset after event was resolved. This is actually a request, which must be synced by all participants.
   virtual void Reset() = 0;

//----------------------------
// Lock/unlock - disable further request
   virtual void Lock(bool) = 0;

//----------------------------
// Message processing.
   virtual void NetIn(PC_net_in_msg in_msg, bool skip_chunk_header = true) = 0;

//----------------------------
// Player disconnection.
   virtual void OnPlayerDisconnect(dword pid) = 0;

//----------------------------
// Get PIDs of all non-local players which already answered to recent query (works only in SS_PROGRESS mode).
   virtual bool GetAnsweredPIDs(vector<dword> &pids) = 0;
};

//----------------------------
// Create network synchronization class.
// Parameters:
//    net_id ... chunk ID to store data into (may be 0, in which case data are not stored in chunk)
//    hdr_bits ... header code (put before data)
//    num_hdr_bits ... number of valid bits in 'hdr_bits' (may be zero, in which case no header is being inserted)
C_net_sync *CreateNetSync(word net_id = 0, dword hdr_bits = 0, byte num_hdr_bits = 0);

//----------------------------
                              //chat class
class C_net_chat{
   enum{ CHAT_BUFFER_SIZE = 33 };
   enum{
      MODE_NULL,
      MODE_ENTER,
      MODE_KEY_RELEASE
   } mode;
   char chat_buffer[CHAT_BUFFER_SIZE];
   int chat_buf_pos;          //-1 means not in chat mode
   int chat_cursor_count;
   bool chat_cursor_phase;
   C_smart_ptr<C_text> text;
   word chat_msg_id;          //ID of chat chunk

   void HideText(){ text = NULL; }
   void ShowText(const char *text, bool reset_cursor = false);

public:
   C_net_chat(word msg_id):
      mode(MODE_NULL),
      chat_msg_id(msg_id)
   {}

//----------------------------
// Open/Close chatting.
   void Open();
   void Close();

   void Tick(const S_tick_context &tc, bool auto_open = false);

   bool IsChatMode() const{ return (mode != MODE_NULL); }

   void NetIn(class C_net_in_msg *in_msg);
};

//----------------------------
                              //network player sorting
struct S_net_player{
   C_string name;
   dword pid;
   S_net_player(){}
   S_net_player(const C_string &n, dword p):
      name(n),
      pid(p)
   {}
   S_net_player(const S_net_player &np){ operator =(np); }
   S_net_player &operator =(const S_net_player &np){
      name = np.name;
      pid = np.pid;
      return *this;
   }
   inline bool operator <(const S_net_player &np){ return (name < np.name); }
};

void MakeSortedNetList(vector<S_net_player> &v, bool including_ours);

//----------------------------
//----------------------------
                              //H&D network IDs
enum E_NETWORK_MESSAGE{

   NM_BEGIN_MISSION = 0x5000,
      //string mission_name
      //dword mission_crc

   NM_TEXT_CHAT = 0x5100,
      //string chat_text

   NM_GAME_MENU = 0x5200, //[+menu_id]
      //data for game menu

   NM_BRIEFING = 0x5300,
      //data for briefing

   NM_CAMP_ANIM = 0x5400,
      //data for campaign animation

   NM_GAME_PROFILE = 0x5500,
      //game profile

   NM_GAME = 0x5600,
      //game data
};

extern PI_net net;
extern PC_net_out_msg net_msg;

//----------------------------
//----------------------------
