#include "all.h"
#include "net.h"
#include "tables.h"
#include "actors.h"
#include "smokeray.h"


void TableToParticle(LPC_table tp, PI3D_particle sp);
//----------------------------
                              //rockets over network:
                              // - primary rocket:
                              //    - evaluation of hit
                              // - network rocket:
                              //    - display particle (movement)
                              // - sent over network:
                              //    - byte id
                              //    - S_vector position
                              //    - S_vector direction
                              //    - word shooter_id

#define ROCKET_ACCEL_TIME 4000
#define ROCKET_MAX_TIME 6000

class C_rocket: public C_actor{
   S_vector pos, dir;
   float speed;               //m/sec
   int fly_count;
   int expl_type;
   LPC_actor shooter;
   bool use_ballistic;
   enum{
      MODE_FLY,
      MODE_FINISH
   } mode;

   C_smoke_ray *smoke_ray;

   void SetupModel(){

      PI3D_frame frm = driver->CreateDummy();
      SetFrame(frm);
      frm->Release();

      delete smoke_ray;
      S_smoke_ray_init si;
      si.scene = mission.GetScene();
      si.start_pos = pos;
      si.start_dir = dir;
      si.tail_fade_time = 2000;
      si.tile_tile = 3.0f;
      si.dest_scale = 5.0f;
      si.tail_name = "tail04";
      smoke_ray = GreateSmokeRay(si);
   }
public:
   C_rocket(C_game_mission &m, void *init_data):
      C_actor(m, ACTOR_ROCKET),
      mode(MODE_FLY),
      fly_count(0),
      shooter(NULL),
      smoke_ray(NULL),
      speed(1.0f),
      expl_type(EXPLOSION_PANZER),
      use_ballistic(true)
   {
                              //allow init data to be NULL, used during LoadGame
      if(!init_data)
         return;
      const S_rocket_init &ri = *(S_rocket_init*)init_data;

      pos = ri.pos;
      dir = ri.dir;

      SetupModel();

      if(ri.id==34){             //tank missile
         speed = 80.0f;
         expl_type = EXPLOSION_TANK_MISSILE;
         use_ballistic = false;
      }

      frame->SetPos(pos);
      frame->SetDir(dir, 0.0f);
      frame->Update();
      mission.GetScene()->SetFrameSector(frame);
                              //play sound
      const char *snd_name = tab_inventory->ItemS(TAB_S16_INV_USESOUND, ri.id, 16);
      if(*snd_name)
         //PlaySound(C_fstring("%s.wav", (const char*)snd_name), model,
         mission.PlaySound(snd_name, frame,
            tab_game_cfg->ItemF(TAB_F_GAME_SNDR_ROCKET_MIN),
            tab_game_cfg->ItemF(TAB_F_GAME_SNDR_ROCKET_MAX));

      shooter = ri.shooter;

      network_actor = ri.netw_actor;
                              //send over network
      if(net && !network_actor){
         bool g = false;
         net_msg->AddWord(C_game_mission::NM_GAME_CREATE_ROCKET, g);
         net_msg->AddByte(ri.id, g);
         net_msg->AddVector(ri.pos, g);
         net_msg->AddVector(ri.dir, g);
         net_msg->AddWord(shooter ? shooter->GetNetID() : 0, g);
      }
   }

   ~C_rocket(){
      delete smoke_ray;
   }

//----------------------------

   dword Tick(const S_tick_context &tc){

      switch(mode){
      case MODE_FLY:
         {
            float tsec = tc.time*.001f;
            fly_count += tc.time;
            if(fly_count<ROCKET_ACCEL_TIME){
               speed += tsec*100.0f;
            }else
            if(fly_count>ROCKET_MAX_TIME){
                                    //disappear silently
               mission.DestroyActor(this);
               return -1;
            }
            S_vector delta(dir*speed*tsec);
                                    //check collisions
            I3D_collision_data cd(pos, delta, I3DCOL_TEXTURE | I3DCOL_JOINT_VOLUMES, NULL, FlyColRespGeo);
            if(mission.GetScene()->TestCollision(cd)){
               const S_vector &ret_pos = cd.GetHitPosition();
               const S_vector &ret_norm = cd.GetHitNormal();
               if(!network_actor){
                                    //make explosion
                  S_explosion_init ei;
                  ei.pos = ret_pos + ret_norm * .5f;
                  ei.id = expl_type;
                  ei.actor_reason = shooter;
                  mission.CreateActor(ACTOR_EXPLOSION, (dword)&ei);
               }

               smoke_ray->Finish();
               mode = MODE_FINISH;
            }else{
               pos += delta;
               frame->SetPos(pos);
               frame->Update();
               if(use_ballistic){
                              //fall down slowly
                  dir.y -= tsec*.025f;
                  dir.Normalize();
               }
               smoke_ray->UpdateTrajectory(pos, dir);
            }
            smoke_ray->UpdateFade(tc.time);
         }
         break;

      case MODE_FINISH:
         smoke_ray->UpdateFade(tc.time);
                              //check for all elements off
         if(smoke_ray->IsOver())
            mission.DestroyActor(this);
         break;
      }
      return 0;
   }

//----------------------------

   virtual bool MaySaveGame() const{
      return (mode != MODE_FINISH);
   }

//----------------------------

   virtual void SaveGame(C_chunk *ck) const{

      //SG_SaveModelInfo(ck);
                           //write data
      ck->Write(&pos, sizeof(pos));
      ck->Write(&dir, sizeof(dir));
      ck->Write(&speed, sizeof(speed));
      ck->Write(&fly_count, sizeof(fly_count));
      ck->Write(&expl_type, sizeof(expl_type));
      ck->Write(&use_ballistic, sizeof(use_ballistic));
      ck->Write(&mode, sizeof(byte));
   }


//----------------------------

   virtual bool LoadGame(C_chunk *ck){

      //SG_LoadModelInfo(ck);
                           //read data
      ck->Read(&pos, sizeof(pos));
      ck->Read(&dir, sizeof(dir));
      ck->Read(&speed, sizeof(speed));
      ck->Read(&fly_count, sizeof(fly_count));
      ck->Read(&expl_type, sizeof(expl_type));
      ck->Read(&use_ballistic, sizeof(use_ballistic));
      ck->Read(&mode, sizeof(byte));
      SetupModel();
      return true;
   }
};

//----------------------------

LPC_actor CreateRocket(C_game_mission &gm, void *data){

   return new C_rocket(gm, data);
}

//----------------------------
