#include "all.h"
#include "smokeray.h"
#include <deque>


//----------------------------

#define MAX_RAY_CURVATURE_ANGLE (PI*.02f)
#define MIN_ELEM_SIZE   .1f   //minimal size of element
#define ORIGIN_FADE_DIST 1.0f  //distance from origin, in which tail fades down (to hide cross at near camera
#define TILED_FIRST_ELEM_SIZE 2.0f //max size of 1st elem size - due to initial alpha

//----------------------------
                              //C_smoke_ray - class for smoke-ray effect following flying object.

class C_smoke_ray_imp: public C_smoke_ray{

   float tail_fade;           //in seconds
   struct S_vertex_info{
      S_vector orig_pos;      //original position of vertex, without z, which we compute
      bool back_side;         //at back side of element, closer to shooting origin
   };
   vector<S_vertex_info> vertex_info;

   struct S_vertex_gunshoot{
      S_vector xyz;
      dword diffuse;
      I3D_text_coor tex;
   };
   float mat_alpha;
   S_vector mat_color;
                              //added references:
   PI3D_scene scene;

                              //added (need to be updated)
   S_vector curr_pos;
   S_vector curr_dir;

   enum{
      SR_MODE_FLYING,            //regular fly
      SR_MODE_FINISHING,         //fading out all rays, destroying when they're done
   } mode;

                              //elements:
   C_smart_ptr<I3D_mesh_base> mb_ray_element;
   struct S_ray_element{
      C_smart_ptr<I3D_visual_dynamic> dyn;
      S_vector start_pos;     //world position of back of element
      S_normal start_dir;     //world direction of back of element
      S_vertex_gunshoot *locked_verts;
      float back_fade;        //fade ratio at the back of element
      float dist_from_origin; //distance of back of element to the origin of shooting

      S_ray_element(): locked_verts(NULL), back_fade(1.0f){}
      S_ray_element(const S_ray_element &e){ operator =(e); }
      S_ray_element &operator =(const S_ray_element &e){
         dyn = e.dyn;
         start_pos = e.start_pos;
         start_dir = e.start_dir;
         locked_verts = e.locked_verts;
         back_fade = e.back_fade;
         dist_from_origin = e.dist_from_origin;
         return *this;
      }
      ~S_ray_element(){
         if(locked_verts){
            dyn->UnlockVertices();
         }
      }
   };
                              //new elems are added to back of the list,
                              // so elems closer to shoot origin are at the front
   deque<S_ray_element> rays;
   float finish_fade;         //when mode==MODE_FINISHING, this is the main fade

   float dest_scale;          //scale at ray's life-end, may be positive or negative
   float tile_tile;           //tile distance of v texture coord, 0.0f means it's shifted
   bool tile_first_done;      //tiled mode - set to true when 1st element is out (it's shorter than others due to alpha)

   void PlaceRay(float ray_len);

   void InitRayElement(const char *tail_name);

   bool CreateUniqueRayElement(PI3D_visual_dynamic*);

   void ComputeBackMatrix(S_matrix &m, const float last_fade,
      const S_ray_element &re, const S_ray_element *re_next) const;

   void ComputeFrontMatrix(S_matrix &m, const float last_fade,
      const S_ray_element &re, const S_vector &front_pos, const S_vector &front_dir) const;

   void TransformElement(const S_ray_element &re, const float last_fade,
      const S_matrix *m_back, const S_matrix *m_front, const S_ray_element *re_next,
      bool first_elem) const;

public:
   C_smoke_ray_imp(const S_smoke_ray_init &is);
   virtual ~C_smoke_ray_imp(){}

   virtual void Finish(){ mode = SR_MODE_FINISHING; }
   virtual bool IsFinishing() const { return mode ==SR_MODE_FINISHING; }
   virtual void UpdateTrajectory(const S_vector &new_pos, const S_vector &new_dir);
   virtual void UpdateFade(int time);
   virtual bool IsOver() const{ return (mode==SR_MODE_FINISHING && (!mb_ray_element || !rays.size())); }
};

//----------------------------

C_smoke_ray_imp::C_smoke_ray_imp(const S_smoke_ray_init &is):
   scene(is.scene),
   curr_pos(is.start_pos),
   curr_dir(is.start_dir),
   dest_scale(is.dest_scale),
   finish_fade(1.0f),
   tail_fade((float)is.tail_fade_time * .001f),
   tile_tile(is.tile_tile), 
   mode(SR_MODE_FLYING),
   mat_alpha(1.0f),
   mat_color(1, 1, 1),
   tile_first_done(false)
{
   InitRayElement(is.tail_name);
}

//----------------------------

void C_smoke_ray_imp::InitRayElement(const char *tail_name){

   if(!tail_name)
      return;
   PI3D_model mod_ray_element = driver->CreateModel();

   model_cache.Open(mod_ray_element, C_fstring("Tails\\%s", tail_name), scene, 0, ErrReport, editor);
                           //find first visual in the model
   PI3D_visual vis_ray_element = I3DCAST_VISUAL(mod_ray_element->FindChildFrame(NULL, ENUMF_VISUAL));
   if(vis_ray_element){
      mb_ray_element = vis_ray_element->GetMesh();
      int numv = mb_ray_element->NumVertices();
      vertex_info.assign(numv);
      assert(mb_ray_element->GetSizeOfVertex()==sizeof(I3D_vertex_mesh));
      PI3D_vertex_mesh verts = (PI3D_vertex_mesh)mb_ray_element->LockVertices();
      assert(verts);
      if(verts){
                           //first-time init, mark vertices on beginning and end
         for(int i=numv; i--; ){
            S_vector &v = verts[i].xyz;
            float z = v.z;
            vertex_info[i].back_side = (z < .1f);
                           //discard z coordinate, we compute it ourselves
            vertex_info[i].orig_pos = S_vector(v.x, v.y, 0.0f);
         }
         mb_ray_element->UnlockVertices();
      }
      PI3D_material mat = mb_ray_element->GetFGroup(0)->GetMaterial();
      mat_alpha = mat->GetAlpha();
      mat_color = mat->GetDiffuse();
   }else{
      ErrReport(C_fstring("can't find visual in loaded ray model: %s", tail_name), editor);
   }
   mod_ray_element->Release();
}

//----------------------------

bool C_smoke_ray_imp::CreateUniqueRayElement(PI3D_visual_dynamic *ret_vis){

   assert(mb_ray_element);
   PI3D_visual_dynamic dyn = (PI3D_visual_dynamic)driver->CreateVisual(I3D_VISUAL_DYNAMIC);
#ifdef EDITOR
   dyn->SetName("C_smoke_ray element");
#endif
   dyn->SetFlags(I3D_FRMF_NOCOLLISION, I3D_FRMF_NOCOLLISION);

   *ret_vis = dyn;

   dyn->DuplicateMesh(mb_ray_element, I3DVC_XYZ | I3DVC_DIFFUSE | (1<<I3DVC_TEXCOUNT_SHIFT));
   return true;
}

//----------------------------

void C_smoke_ray_imp::ComputeBackMatrix(S_matrix &m, const float last_fade,
   const S_ray_element &re, const S_ray_element *re_next) const{

                     //our tex coords would end up negative, so we rather adjust
                     // vertex position
   m.Identity();
   S_vector new_pos = re.start_pos, new_dir = re.start_dir;
   float scale;

   if(re.back_fade < 0.0f){
                     //apply position change
      float d = -re.back_fade / (last_fade - re.back_fade);

      if(re_next){
         new_pos += (re_next->start_pos - re.start_pos) * d;
         new_dir = new_dir * (1.0f - d) + re_next->start_dir * d;
      }else{
         new_pos += (curr_pos - re.start_pos) * d;
         new_dir = new_dir * (1.0f - d) + curr_dir * d;
      }
      scale = dest_scale;
   }else{
      scale = 1.0f + (dest_scale - 1.0f) * (1.0f - re.back_fade);
   }

   m.SetDir(new_dir, 0.0f);
   m(3) = new_pos;

   if(dest_scale != 1.0f){
                     //apply scale on vertices
      m(0) *= scale;
      m(1) *= scale;
      m(2) *= scale;
   }
}

//----------------------------

void C_smoke_ray_imp::ComputeFrontMatrix(S_matrix &m, const float last_fade,
   const S_ray_element &re, const S_vector &front_pos, const S_vector &front_dir) const{

                              //scale front of element
   m.Identity();
   m.SetDir(front_dir, 0.0f);
   m(3) = front_pos;

   float scale = 1.0f + (dest_scale - 1.0f) * (1.0f - last_fade);
                     
   m(0) *= scale;
   m(1) *= scale;
   m(2) *= scale;
}

//----------------------------

void C_smoke_ray_imp::TransformElement(const S_ray_element &re, const float last_fade,
   const S_matrix *m_start, const S_matrix *m_end, const S_ray_element *re_next,
   bool first_elem) const{

   bool use_tile = (*((dword*)&tile_tile)!=0);
                              //front and back v - used in tiled mode
   float front_v, back_v;
                              //front and back alpha, used in tiled mode
   float front_alpha, back_alpha;

   if(use_tile){
                              //compute front and back values
      if(re.back_fade < 0.0f){
                              //back part is the end of tail

                              //use zero alpha
         back_alpha = 0.0f;
                              
                              //compute proper v tile
         float d = -re.back_fade / (last_fade - re.back_fade);
         float dist = re.dist_from_origin +
            (re_next ? (re_next->dist_from_origin-re.dist_from_origin) :
            ((curr_pos-re.start_pos).Magnitude())) * d;
            back_v = dist / tile_tile;
      }else{
                              //back part is fading down
         back_alpha = re.back_fade;
      }
                              //apply alpha when too close to origin
      if(re.dist_from_origin < ORIGIN_FADE_DIST)
         back_alpha *= re.dist_from_origin / ORIGIN_FADE_DIST;


      if(!re_next){
                              //this is the most front part,
                              // compute distance from origin
         float dist = re.dist_from_origin + (curr_pos-re.start_pos).Magnitude();
                              //compute v tile based on this distance
         front_v = dist / tile_tile;

                              //compute alpha at current position
                              // (usualy 1.0 unles we've hit wall
         front_alpha = last_fade;

                              //apply alpha when too close to origin
         if(dist < ORIGIN_FADE_DIST)
            front_alpha *= dist / ORIGIN_FADE_DIST;
      }else{
                              //this is mid part, use current fade value in this point
         front_alpha = last_fade;

                              //apply alpha when too close to origin
         if(re_next->dist_from_origin < ORIGIN_FADE_DIST)
            front_alpha *= re_next->dist_from_origin / ORIGIN_FADE_DIST;
      }
   }

                              //process all vertices
   S_vertex_gunshoot *verts = re.locked_verts;
   for(int ii=re.dyn->GetMesh()->NumVertices(); ii--; ){
      S_vertex_gunshoot &v = verts[ii];
                              //check if vertex belongs to front or back
      if(vertex_info[ii].back_side){
         if(!use_tile){
                              //v-shifting used

            if(first_elem || re.back_fade < 0.0f){
                              //back part is the end of tail,
                              // or this is very first element
               v.tex.v = 1.0f;
            }else{
                              //v tile is inverse of fade value on this position
               v.tex.v = 1.0f - re.back_fade;
            }
         }else{
                              //v-tiling
            if(re.back_fade < 0.0f){
                              //back part is the end of tail,
                              // v tile is in movement (also coords are)
               v.tex.v = back_v;
            }
                              //alpha value equals to fade of back
            v.diffuse &= 0xffffff;
            v.diffuse |= ((int)(back_alpha*mat_alpha*255.0f)) << 24;
         }
                              //optionally transform vertex (pos/scale)
         if(m_start)
            verts[ii].xyz = vertex_info[ii].orig_pos * (*m_start);
      }else{
         if(!use_tile){
                              //v-shifting
                              // v tile is inverse of fade value on this position
            v.tex.v = 1.0f - last_fade;
         }else{
            if(!re_next){
                              //this is the most front part,
               v.tex.v = front_v;
            }
            v.diffuse &= 0xffffff;
            v.diffuse |= ((int)(front_alpha*mat_alpha*255.0f)) << 24;
         }
                              //optionally transform vertex (pos/scale)
         if(m_end)
            verts[ii].xyz = vertex_info[ii].orig_pos * (*m_end);
      }
   }
}


//----------------------------

void C_smoke_ray_imp::UpdateTrajectory(const S_vector &new_pos, const S_vector &new_dir){

   bool create_new_element = false;
   if(!mb_ray_element)
      return;

                              //store last values for new element
   S_vector last_pos = curr_pos;
   S_vector last_dir = curr_dir;

                              //update position and dir from references
   curr_pos = new_pos;
   curr_dir = new_dir;

                              //tail processing
   if(!rays.size())
      create_new_element = true;
   else{
                              //check angle with previous element
      float angle = rays.back().start_dir.AngleTo(curr_dir);
      if(angle >= MAX_RAY_CURVATURE_ANGLE)
         create_new_element = true;
      else{
         float last_elem_size = (rays.back().start_pos-curr_pos).Magnitude();
         if(last_elem_size > MIN_ELEM_SIZE){
#ifdef DEBUG_MAX_ELEM_SIZE
                              //debug:
            if(last_elem_size >= DEBUG_MAX_ELEM_SIZE)
               create_new_element = true;
#endif
            if(*(dword*)&tile_tile){
               float last_size = (rays.back().start_pos-curr_pos).Magnitude();
#ifdef MAX_ELEM_SIZE
               if(last_size >= MAX_ELEM_SIZE)
                  create_new_element = true;
               else
#endif
               if(!tile_first_done && last_size>=TILED_FIRST_ELEM_SIZE){
                  tile_first_done = true;
                  create_new_element = true;
               }
            }
         }
      }
   }

   if(create_new_element){
      rays.push_back(S_ray_element());
      S_ray_element &re = rays.back();
      re.start_pos = last_pos;
      re.start_dir = last_dir;
                        //compute logical distance from origin
      if(rays.size()>1){
         const S_ray_element &re1 = rays[rays.size()-2];
         re.dist_from_origin =
            re1.dist_from_origin + (re.start_pos-re1.start_pos).Magnitude();
      }else{
         re.dist_from_origin = 0.0f;
      }

      PI3D_visual_dynamic dyn;
      CreateUniqueRayElement(&dyn);
      re.dyn = dyn; dyn->Release();
      //mission.GetScene()->SetFrameSectorPos(re.obj, curr_pos);

      {                       //update ray's vertices
         S_matrix m_start; m_start.Identity();
         m_start.SetDir(last_dir, 0.0f);
         m_start(3) = last_pos;

         //if(rays.size() >= 2) DEBUG_LINE(rays[rays.size()-2].start_pos, last_pos);
         //DEBUG_LINE(curr_pos, curr_pos+curr_dir/curr_dir.Magnitude());

         if(!re.locked_verts){
            re.locked_verts = (S_vertex_gunshoot*)re.dyn->LockVertices(I3DLOCK_VERTEX);
         }
         S_vertex_gunshoot *verts = re.locked_verts;
         assert(verts);
         if(verts){
            bool use_tile = *(dword*)&tile_tile;
                              //first-time init, adjust original position
            for(int i=re.dyn->GetMesh()->NumVertices(); i--; ){
               S_vertex_gunshoot &v = verts[i];
               v.diffuse = int(mat_color.x*255.0f)<<16 | int(mat_color.y*255.0f)<<8 | int(mat_color.z*255.0f);
               float a = 1.0f;
               if(vertex_info[i].back_side){
                  v.xyz = vertex_info[i].orig_pos * m_start;
                  if(use_tile){
                     verts[i].tex.v = re.dist_from_origin / tile_tile;
                     if(re.dist_from_origin < ORIGIN_FADE_DIST)
                        a *= re.dist_from_origin / ORIGIN_FADE_DIST;
                  }
               }
               v.diffuse |= ((int)(a*mat_alpha*255.0f)) << 24;
            }
         }
      }
   }
   {
                              //move recent element to new position
      assert(rays.size());
      S_ray_element &re = rays.back();
      assert(re.dyn);
                        //put to appropriate sector
      {                              
         S_vector center_pos = curr_pos + (re.start_pos - curr_pos) * .5f;
         scene->SetFrameSectorPos(re.dyn, center_pos);
      }

      S_matrix m_end; m_end.Identity();
      m_end.SetDir(curr_dir, 0.0f);
      m_end(3) = curr_pos;
 
      if(!re.locked_verts){
         re.locked_verts = (S_vertex_gunshoot*)re.dyn->LockVertices(I3DLOCK_VERTEX);
      }
      S_vertex_gunshoot *verts = re.locked_verts;
      assert(verts);
      if(verts){
                           //first-time init, mark vertices on front side
         for(int i=re.dyn->GetMesh()->NumVertices(); i--; ){
            S_vector &v = verts[i].xyz;
            if(!vertex_info[i].back_side){
               v = vertex_info[i].orig_pos * m_end;
            }
         }
      }
   }
                              //unlock all meshes
   for(deque<S_ray_element>::iterator it = rays.begin(); it!=rays.end(); ++it){
      if(it->locked_verts){
         I3D_RESULT ir;
         ir = it->dyn->UnlockVertices();
         assert(I3D_SUCCESS(ir));
         it->locked_verts = NULL;
      }
   }
}

//----------------------------

void C_smoke_ray_imp::UpdateFade(int time){

   if(!tail_fade)
      return;

   float fade_down = time * .001f / tail_fade;

   if(mode == SR_MODE_FINISHING){
      finish_fade -= fade_down;
   }
                              //fade beginnings of all meshes, adjust uv coordinates
                              // and vertex position of tail element, remove unused elements
   if(mb_ray_element && rays.size()){

      //assert(rays.size());

      //if(!rays.size())        //just for safe, remove if you are sure
         //return;
                              //compute fade at the front of tail
                              // this is 1 until ray hits something
      //float last_fade = (mode==MODE_FINISHING) ? rays.back().back_fade : 1.0f;
      float last_fade = finish_fade;

      bool do_scale = (dest_scale != 1.0f);

                              //process all elements back to front,
                              // stop when fade of element would be negative
      const S_ray_element *re_next = NULL;
      for(int i = rays.size(); i-- && last_fade >= 0.0f; ){
         S_ray_element &re = rays[i];
         re.back_fade -= fade_down;

                              //re-transform verts if position or scale changes
         S_matrix m_back, m_front;
         dword lock_flags = I3DLOCK_UV;

         bool use_m_back = (re.back_fade < 0.0f || do_scale);
         if(use_m_back){
            ComputeBackMatrix(m_back, last_fade, re, re_next);
            lock_flags |= I3DLOCK_XYZ;
         }
         bool use_m_front = (do_scale);
         if(use_m_front){
            if(re_next)
               ComputeFrontMatrix(m_front, last_fade, re, re_next->start_pos, re_next->start_dir);
            else
               ComputeFrontMatrix(m_front, last_fade, re, curr_pos, curr_dir);
         }
                              //get access to vertices and compute values
         if(!re.locked_verts)
            re.locked_verts = (S_vertex_gunshoot*)re.dyn->LockVertices(lock_flags);

         TransformElement(re, last_fade, use_m_back ? &m_back : NULL, use_m_front ? &m_front : NULL,
            re_next, (i==0));

         last_fade = re.back_fade;
         re_next = &re;
      }
      rays.erase(rays.begin(), rays.begin() + i + 1);

                              //unlock all meshes
      for(deque<S_ray_element>::iterator it = rays.begin(); it!=rays.end(); ++it){
         if(it->locked_verts){
            I3D_RESULT ir;
            ir = it->dyn->UnlockVertices();
            assert(I3D_SUCCESS(ir));
            it->locked_verts = NULL;
         }
      }
   }
}

//----------------------------

C_smoke_ray *GreateSmokeRay(const S_smoke_ray_init &is){

   return new C_smoke_ray_imp(is);
}

//----------------------------
//----------------------------
